'''
Created on Jul 16, 2013

@author: marten
'''
from django.utils.translation import ugettext_lazy as _
from sqlalchemy import func, and_

from basesoft.web.contrib.license.models import License
from basesoft.web.contrib.file.models import File
from sqlalchemy.orm.session import Session
from basesoft.util.virtual_date import today
from basesoft.web.contrib.notification.models import Notification


class PvcOperatorLicense(License):
    """
    @DynamicAttrs
    """
    NAME_I18N = _("PVC Operator")
    REQUIRED_PERMISSIONS = ['pvc_operator_license']

    def __get_controller(self):
        from pvcycle.core.controllers import PvcOperatorController
        return PvcOperatorController

    controller = property(__get_controller)

    def __repr__(self):
        return '<PvcOperatorLicense id=%s>' % self.id


class PvcAccountantLicense(License):
    """
    @DynamicAttrs
    """
    NAME_I18N = _('PVC Accountant')
    REQUIRED_PERMISSIONS = ['pvc_accountant_license']

    def __repr__(self):
        return '<PvcAccountantLicense id=%s>' % self.id


class PvcTbsLicense(License):
    """
    @DynamicAttrs
    """
    NAME_I18N = _('PVC Take Back Scheme')
    REQUIRED_PERMISSIONS = ['pvc_tbs_license']

    TBS_TYPE_SUBSIDIARY = 1 # implies invoicing.
    TBS_TYPE_LOCAL = 2 # only aggregated data import.

    DECLARATION_PERIOD_DEFAULT = 0  # used on overriding by TBS Members
    DECLARATION_PERIOD_MONTHLY = 1
    DECLARATION_PERIOD_YEARLY = 2
    DECLARATION_PERIOD_QUARTERLY = 3

    def __get_controller(self):
        from pvcycle.core.controllers import PvcTBSController
        return PvcTBSController

    controller = property(__get_controller)

    def __repr__(self):
        return '<PvcTbsLicense id=%s>' % self.id

class PvcTbsServiceLevel(object):
    """
    @DynamicAttrs
    """

    STATUS_ACTIVE = 1
    STATUS_DISABLED_FOR_NEW_REGISTRATIONS = 2

    def get_current_administration_fee(self):
        session = Session.object_session(self)
        return session.query(func.pvcGetCurrentAdministrationFee(self.id, today().year)).one()[0]

    def get_rounded_administration_fee(self):
        from decimal import Decimal
        return self.get_current_administration_fee().quantize(Decimal('.01'))

    current_admin_fee = property(get_rounded_administration_fee)

    def get_current_contribution_fee(self):
        session = Session.object_session(self)
        return session.query(func.pvcGetCurrentContributionFee(self.id, today())).one()[0]

    def get_rounded_contribution_fee(self):
        from decimal import Decimal
        return self.get_current_contribution_fee().quantize(Decimal('.01'))

    current_contrib_fee = property(get_rounded_contribution_fee)

    def get_current_fee_start_year(self):
        return today().year

    def __repr__(self):
        return '<PvcTbsServiceLevel id="%s">' % self.id

class PvcTbsServiceLevelAdministrationFee(object):
    """
    @DynamicAttrs
    """
    pass

class PvcTbsServiceLevelContributionFee(object):
    """
    @DynamicAttrs
    """
    pass

class PvcAccountantTbs(object):
    """
    @DynamicAttrs
    """
    pass

class PvcMemberLicense(License):
    """
    @DynamicAttrs
    """
    NAME_I18N = _('PVC Member License')
    REQUIRED_PERMISSIONS = ['pvc_tbs_member_manager'] # this can also be created from PUBLIC or by Operator?.

    def __get_controller(self):
        from pvcycle.core.controllers import PvcMemberController
        return PvcMemberController

    controller = property(__get_controller)

    def __repr__(self):
        return '<PvcMemberLicense id=%s>' % self.id


class PvcTbsMember(object):
    """
    @DynamicAttrs
    """
    STATUS_INCOMPLETE = 0
    STATUS_PENDING = 1
    STATUS_APPROVED = 2
    STATUS_REJECTED = 3
    STATUS_SUSPENDED = 4

    STATUS_RESIGNING = 5
    STATUS_RESIGNED = 6

    SELLING_TECHNIQUE_MANUFACTURE_AND_SELL = 1
    SELLING_TECHNIQUE_DISTRIBUTE_OR_RESELL_REBRANDED = 2
    SELLING_TECHNIQUE_IMPORTING = 4
    SELLING_TECHNIQUE_INTERNET_SALES_OWN_COUNTRY = 8
    SELLING_TECHNIQUE_INTERNET_SALES_OTHER_COUNTRIES = 16

    def _get_active_service_level(self, date=None):

        if date is None:
            date = today()

        session = Session.object_session(self)
        return session.query(PvcTbsMemberServiceLevel).filter(
            and_(
                PvcTbsMemberServiceLevel.tbs_member==self,
                PvcTbsMemberServiceLevel.start_date<=date,
                PvcTbsMemberServiceLevel.status==PvcTbsMemberServiceLevel.STATUS_APPROVED
            )
        ).order_by(PvcTbsMemberServiceLevel.start_date.desc()).first()

    active_service_level = property(_get_active_service_level)

    def get_service_level_for_date(self, date):
        return self._get_active_service_level(date)

    def _get_upcoming_service_level(self):
        session = Session.object_session(self)
        return session.query(PvcTbsMemberServiceLevel).filter(
            and_(
                PvcTbsMemberServiceLevel.tbs_member==self,
                PvcTbsMemberServiceLevel.start_date >= today()
            )
        ).order_by(PvcTbsMemberServiceLevel.start_date.desc()).first()

    upcoming_service_level = property(_get_upcoming_service_level)


    def _get_active_membership(self):
        session = Session.object_session(self)
        return session.query(PvcTbsMemberMembership).filter(
            and_(
                PvcTbsMemberMembership.tbs_member==self,
                PvcTbsMemberMembership.start_date <= today()
            )
        ).order_by(PvcTbsMemberMembership.start_date.desc()).first()

    active_membership = property(_get_active_membership)

    def _get_declaration_periodicity(self):
        if self.declaration_periodicity != PvcTbsLicense.DECLARATION_PERIOD_DEFAULT:
            return self.declaration_periodicity
        # we are still dealing with a declaration change on TBS level.
        if self.upcoming_periodicity_change is not None:
            return self.tbs.get_config('pvc.tbs.previous.declaration.period')
        return self.tbs.get_config('pvc.tbs.declaration.period')

    declaration_period = property(_get_declaration_periodicity)

    def _get_closing_day(self):
        if self.declaration_closing_day is not None:
            return self.declaration_closing_day
        return self.tbs.get_config('pvc.tbs.declaration.closing.day')

    closing_day = property(_get_closing_day)

    def __repr__(self):
        return '<PvcTbsMember id=%s>' % self.id

class PvcTbsMemberServiceLevel(object):
    """
    @DynamicAttrs
    """
    BACKREF = 'service_level'

    STATUS_PENDING = 1
    STATUS_APPROVED = 2

    def get_current_administration_fees(self):
        rv = []

        for x in ('service_level', 'pv_service_level'):

            if getattr(self, x, None) is not None:
                sl = getattr(self, x)
                rv.append((
                    sl.name,
                    sl.get_current_administration_fee(),
                    sl.get_current_fee_start_year()
                ))

        return rv



class PvcTbsMemberContractFile(File):
    """
    @DynamicAttrs
    """

    TYPE_CONTRACT_FULL = 1
    TYPE_CONTRACT_ANNEX = 2

    def __repr__(self):
        return '<PvcTbsMemberContractFile id=%s>' % self.id


class PvcTbsCertificate(object):
    """
    @DynamicAttrs
    """

class PvcTbsMemberCertificate(object):
    """
    @DynamicAttrs
    """

    STATUS_PAID = 0
    STATUS_APPROVED = 1
    STATUS_REJECTED = 2
    STATUS_REVOKED = 3


    def __get_issue_date(self):
        return self.created.date()

    issue_date = property(__get_issue_date)

class PvcTbsMembership(object):
    """
    @DynamicAttrs
    """
    def get_current_fee(self):
        session = Session.object_session(self)
        return session.query(func.pvcGetCurrentMembershipFee(self.id, today().year)).one()[0]

    # compatibility with PvcTbsServiceLevel
    get_current_administration_fee = get_current_fee

    def get_current_fee_start_year(self):
        session = Session.object_session(self)
        return session.query(func.pvcGetCurrentMembershipStartYear(self.id, today().year)).one()[0]

    def get_rounded_fee(self):
        from decimal import Decimal
        return self.get_current_fee().quantize(Decimal('.01'))

    current_admin_fee = property(get_rounded_fee)

    def __repr__(self):
        return '<PvcTbsMembership id="%s">' % self.id

class PvcTbsMembershipFee(object):
    """
    @DynamicAttrs
    """
    pass

class PvcTbsMemberMembership(object):
    """
    @DynamicAttrs
    """
    BACKREF = 'membership'

    def get_current_administration_fees(self):

        return [(
            self.membership.name,
            self.membership.get_current_administration_fee(),
            self.membership.get_current_fee_start_year()
        )]


class PvcROECountry(object):
    """
    @DynamicAttrs
    """
    pass

class PvcRegistrationNotification(Notification):
    """
    @DynamicAttrs
    """

    name_i18n = _('Registration')

    def __repr__(self):
        return '<PvcRegistrationNotification id=%s>' % self.id

class PvcTbsMemberProofFinancialGuaranteeFile(File):
    """
    @DynamicAttrs
    """

    def __repr__(self):
        return '<PvcTbsMemberProofFinancialGuaranteeFile id=%s>' % self.id