# Copyright (c) 2015 BaseSoft B.V.
# See LICENSE for details.
from types import NoneType
from basesoft.web.contrib.file.controllers import UploadFileController, ProcessFileController
from basesoft.web.contrib.file.models import File
from basesoft.web.contrib.notification.controllers import NotificationController
from basesoft.web.core.exceptions import SecurityException

from django.utils.translation import ugettext as _

from basesoft.web.contrib.contact.controllers import SimpleContactController
from basesoft.web.core.util.controllers import AbstractController, state_to_obj, lazy_assert, transaction_decorator

from pvcycle.core.controllers import AbstractPvcController
from pvcycle.logistics.models import PvcTreatmentCenter, PvcDeliveryRequest, PvcPickupRequest, PvcHolderLicense, \
    PvcPickupRequestNotification, PvcCarrierLicense
from sqlalchemy.sql.elements import and_


class PvcLogisticsController(AbstractPvcController, AbstractController):

    #
    # common
    #

    def get_holders(self):
        """
        returns a list of all PvcHolderLicenses the current user can access
        """
        # TODO: access control
        return self.sa_session.query(PvcHolderLicense).order_by(PvcHolderLicense.name).all()

    def get_carriers(self):
        """
        returns a list of all PvcHolderLicenses the current user can access
        """
        # TODO: access control
        return self.sa_session.query(PvcCarrierLicense).order_by(PvcCarrierLicense.name).all()

    def get_treatment_centers_state(self):
        rv = []
        tcs = self.sa_session.query(PvcTreatmentCenter)

        scc = SimpleContactController(self.sa_session, account_id=self.get_account().id)
        for tc in tcs:
            rv.append(scc.load_company_state(tc))

        return rv

    #
    # pickup request
    #

    def get_pickup_request(self, pickup_request_id):
        """
        returns an instance of the given pickup_request_id
        """
        lazy_assert(pickup_request_id, int)

        # TODO: access control
        return self.sa_session.query(PvcPickupRequest).filter(PvcPickupRequest.id == pickup_request_id).one()

    def get_pending_pickup_requests(self):
        """
        returns a list of all PvcPickupRequest the current user can access with status pending
        """
        # TODO: access control
        return self.sa_session.query(PvcPickupRequest) \
            .filter(PvcPickupRequest.status==PvcPickupRequest.STATUS_PENDING) \
            .order_by(PvcPickupRequest.id).all()

    def get_pickup_requests_from_delivery(self, delivery_request_id):
        """
        returns a list of all PvcPickupRequest belonging to a delivery_request
        """
        # TODO: access control
        delivery_request = self.get_delivery_request(delivery_request_id)
        return self.sa_session.query(PvcPickupRequest) \
            .filter(PvcPickupRequest.delivery_request==delivery_request) \
            .order_by(PvcPickupRequest.id).all()

    @transaction_decorator
    def save_or_update_pickup_request(self, pickup_request_id, form_values):

        # TODO: access control

        if pickup_request_id is None:
            pickup_request = PvcPickupRequest()
            self.sa_session.add(pickup_request)
        else:
            pickup_request = self.sa_session.query(PvcPickupRequest) \
                .filter(PvcPickupRequest.id == pickup_request_id).one()

        if form_values['weight_tc'] == '':
            form_values['weight_tc'] = None

        pickup_request = state_to_obj(pickup_request, form_values, (
            ('holder_id', int),
            ('number_modules', int),
            ('weight_holder', int),
            ('weight_tc', (NoneType, int) ),
        ))

        if form_values['photo_file']:

            file_tmp = UploadFileController().get_hash_temp_files(form_values['photo_file'])[0]

            fc = ProcessFileController(self.sa_session, account_id=self.account_id)
            category = fc.get_category_by_label(_("Pickup request"))
            if category is None:
                category = fc.save_or_update_file_category(None, "Pickup request")

            photo_file = fc.save_files_from_temp([file_tmp], category)[0]
            self.sa_session.add(photo_file)

            pickup_request.photo_file = photo_file

            # FIXME: access control
            self.grant_privilege_to(photo_file, File.PRIV_VIEW, domain=True, target_group=self.get_default_group(), check_own_privs=False)

        return pickup_request

    def submit_pickup_request(self, pickup_request_id):

        pickup_request = self.get_pickup_request(pickup_request_id)
        pickup_request.status = PvcPickupRequest.STATUS_PENDING

        # notification to key-user to make a DeliveryRequest with this PickupRequested selected
        nc = NotificationController(self.sa_session, request=self.request, account_id=self.get_account().id)
        notification = nc.set_notification(_('New pickup request'), _('A new pickup request has been made.'),
                                           instance_class=PvcPickupRequestNotification, callback={
            'js_module': 'pvc_delivery_request_details',
            'parameters': [None, {
                'pickup_requests': [pickup_request.id]
            }]
        })
        recipients = self.get_groups_in_domain_with_permissions('pvc_logistics_key_user')
        nc.create_notification(notification, recipients)

    def load_pickup_request(self, pickup_request_id):
        pickup_request = self.get_pickup_request(pickup_request_id)

        photo_url = None
        if pickup_request.photo_file:
            photo_url = '/basesoft.web.contrib.file/view/'+pickup_request.photo_file.file_hash

        return {
            'id': pickup_request.id,
            'holder_id': pickup_request.holder_id,
            'delivery_request_id': pickup_request.delivery_request_id,
            'technology': pickup_request.technology,
            'condition': pickup_request.condition,
            'number_modules': pickup_request.number_modules,
            'weight_holder': pickup_request.weight_holder,
            'weight_tc': pickup_request.weight_tc,
            'status': pickup_request.status,
            'photo_url': photo_url
        }

    #
    # delivery request
    #

    def get_delivery_request(self, delivery_request_id):
        """
        returns an instance of the given delivery_request_id
        """
        lazy_assert(delivery_request_id, int)

        # TODO: access control
        return self.sa_session.query(PvcDeliveryRequest).filter(PvcDeliveryRequest.id == delivery_request_id).one()

    @transaction_decorator
    def save_or_update_delivery_request(self, delivery_request_id, form_values):

        # TODO: access control

        if delivery_request_id is None:
            delivery_request = PvcDeliveryRequest()
            self.sa_session.add(delivery_request)
        else:
            delivery_request = self.sa_session.query(PvcDeliveryRequest) \
                .filter(PvcDeliveryRequest.id == delivery_request_id).one()

        # FIXME: a NumberInput should post None instead of ''
        if form_values['total_kilometers'] == '':
            form_values['total_kilometers'] = None

        delivery_request = state_to_obj(delivery_request, form_values, (
            ('carrier_id', int),
            ('treatment_center_id', int),
            ('total_units', int),
            ('total_kilometers', (NoneType, int) ),
        ))

        delivery_request.pickup_requests = []
        for pickup_request_id in form_values['pickup_requests']:
            pickup_request = self.get_pickup_request(int(pickup_request_id))

            # assign delivery request to pickup request
            pickup_request.delivery_request = delivery_request
            pickup_request.status = PvcPickupRequest.STATUS_PENDING

        # when a delivery request with status accepted is saved it will become 'completed'
        # then process weight-by-tc and uploaded documents
        if delivery_request.status == PvcDeliveryRequest.STATUS_ACCEPTED or delivery_request.status == PvcDeliveryRequest.STATUS_COMPLETED:

            if form_values['annex7_file'] != None and form_values['annex7_file'] != '':

                file_tmp = UploadFileController().get_hash_temp_files(form_values['annex7_file'])[0]

                fc = ProcessFileController(self.sa_session, account_id=self.account_id)
                category = fc.get_category_by_label(_("Pickup request"))
                if category is None:
                    category = fc.save_or_update_file_category(None, "Pickup request")

                annex7_file = fc.save_files_from_temp([file_tmp], category)[0]
                self.sa_session.add(annex7_file)

                delivery_request.annex7_file = annex7_file

                # FIXME: access control
                self.grant_privilege_to(annex7_file, File.PRIV_VIEW, domain=True, target_group=self.get_default_group(), check_own_privs=False)


            for row in form_values['pickup_rowcontrol']:

                pickup_request = self.get_pickup_request(row['pickup_request_id'])

                # FIXME: enable this check. Currently disabled for demo because the same PR could be attached
                # to multiple DR's
                #assert pickup_request.delivery_request == delivery_request

                pickup_request.weight_tc = row['weight_tc']

                if row['wtf_file'] != None and row['wtf_file'] != '':

                    file_tmp = UploadFileController().get_hash_temp_files(row['wtf_file'])[0]

                    fc = ProcessFileController(self.sa_session, account_id=self.account_id)
                    category = fc.get_category_by_label(_("Pickup request"))
                    if category is None:
                        category = fc.save_or_update_file_category(None, "Pickup request")

                    wtf_file = fc.save_files_from_temp([file_tmp], category)[0]
                    self.sa_session.add(wtf_file)

                    pickup_request.wtf_file = wtf_file

                    # FIXME: access control
                    self.grant_privilege_to(wtf_file, File.PRIV_VIEW, domain=True, target_group=self.get_default_group(), check_own_privs=False)


        return delivery_request

    def load_delivery_request(self, delivery_request_id):
        delivery_request = self.get_delivery_request(delivery_request_id)

        return {
            'id': delivery_request.id,
            'carrier_id': delivery_request.carrier_id,
            'treatment_center_id': delivery_request.treatment_center_id,
            'pickup_requests': [x.id for x in delivery_request.pickup_requests],
            'total_units': delivery_request.total_units,
            'total_kilometers': delivery_request.total_kilometers,
            'status': delivery_request.status,
            'annex7_file': None if delivery_request.annex7_file is None else delivery_request.annex7_file.file_hash,
            'status': delivery_request.status,
            'pickup_rowcontrol': [{
                'pickup_request_id': x.id,
                'weight_tc': x.weight_tc,
                'wtf_file': None if x.wtf_file is None else x.wtf_file.file_hash

            } for x in delivery_request.pickup_requests]
        }

    @transaction_decorator
    def delivery_request_assign_to_courier(self, delivery_request_id):
        """
        called when the key-user assigns a delivery request to a carrier
        """
        delivery_request = self.get_delivery_request(delivery_request_id)
        delivery_request.status = PvcDeliveryRequest.STATUS_ASSIGNED

        for pr in delivery_request.pickup_requests:
            pr.status = PvcPickupRequest.STATUS_PENDING

        # send notification to courier
        nc = NotificationController(self.sa_session, request=self.request, account_id=self.get_account().id)
        notification = nc.set_notification(_('New delivery request'), _('A delivery request has been assigned to you.'), instance_class=PvcPickupRequestNotification, callback={
            'js_module': 'pvc_delivery_request_details',
            'parameters': [delivery_request_id]
        })
        recipients = delivery_request.carrier.default_group
        nc.create_notification(notification, recipients)

    @transaction_decorator
    def delivery_request_accept_transport(self, delivery_request_id):
        """
        called when the carrier has accepted a delivery request
        """
        delivery_request = self.get_delivery_request(delivery_request_id)
        delivery_request.status = PvcDeliveryRequest.STATUS_ACCEPTED

        for pr in delivery_request.pickup_requests:
            pr.status = PvcPickupRequest.STATUS_ACCEPTED

    @transaction_decorator
    def delivery_request_complete(self, delivery_request_id):
        """
        called when the carrier has completed the delivery request. all documents are uploaded
        """
        delivery_request = self.get_delivery_request(delivery_request_id)
        delivery_request.status = PvcDeliveryRequest.STATUS_COMPLETED

        # notification to key-user
        nc = NotificationController(self.sa_session, request=self.request, account_id=self.get_account().id)
        notification = nc.set_notification(_('Delivery request completed'), _('A delivery request has been completed'),
                                           instance_class=PvcPickupRequestNotification, callback={
            'js_module': 'pvc_delivery_request_details',
            'parameters': [delivery_request.id]
        })
        recipients = self.get_groups_in_domain_with_permissions('pvc_logistics_key_user')
        nc.create_notification(notification, recipients)

        for pickup_request in delivery_request.pickup_requests:

            # update status
            pickup_request.status = PvcPickupRequest.STATUS_COMPLETED

            # as the system, send a notification to holder
            nc = NotificationController(self.sa_session, request=self.request)
            notification = nc.set_notification(_('Pickup request completed'), _('A pickup request has been completed. Please fill in the survey so we can provide you the best service.'),
                                               instance_class=PvcPickupRequestNotification, callback={
                'js_module': 'pvc_logistics_feedback_panel'
            })

            recipients = pickup_request.holder.default_group
            nc.create_system_notification(notification, recipients)

    @transaction_decorator
    def approve_delivery_request(self, delivery_request_id):
        """
        the key user approves the delivery request
        """
        delivery_request = self.get_delivery_request(delivery_request_id)
        delivery_request.status = PvcDeliveryRequest.STATUS_APPROVED


    #
    # report
    #

    def _get_abstract_request_report(self, obj, status_list):
        # data used for Dynamic Table widget

        def get_count(status):
            return self.sa_session.query(obj).filter(obj.status == status).count()

        rv = [[None, _('Requests by status')]]
        for r in status_list:
            rv.append( (r[1], get_count(r[0]) ) )

        return rv

    def get_pickup_request_report(self):
        return self._get_abstract_request_report(PvcPickupRequest, [
            (PvcPickupRequest.STATUS_DRAFT, _('Draft') ),
            (PvcPickupRequest.STATUS_PENDING, _('Pending') ),
            (PvcPickupRequest.STATUS_REJECTED, _('Rejected') ),
            (PvcPickupRequest.STATUS_ACCEPTED, _('Accepted') ),
            (PvcPickupRequest.STATUS_INVALID, _('Invalid') ),
            (PvcPickupRequest.STATUS_COMPLETED, _('Completed') ),
        ])

    def get_delivery_request_report(self):
        return self._get_abstract_request_report(PvcDeliveryRequest, [
            (PvcDeliveryRequest.STATUS_DRAFT, _('Draft') ),
            (PvcDeliveryRequest.STATUS_ASSIGNED, _('Assigned') ),
            (PvcDeliveryRequest.STATUS_REJECTED, _('Rejected') ),
            (PvcDeliveryRequest.STATUS_ACCEPTED, _('Accepted') ),
            (PvcDeliveryRequest.STATUS_COMPLETED, _('Completed') ),
        ])


