# Copyright (c) 2015 BaseSoft B.V.
# See LICENSE for details.
from basesoft.web.contrib.license.models import License
from basesoft.web.core.forms.handler import FormValidator
from basesoft.web.core.input.validation import plugins
from pvcycle.logistics.controllers import PvcLogisticsController
from pvcycle.logistics.models import PvcDeliveryRequest


class LogisticsRpcHandler:

    def __init__(self, request, **kwargs):

        self._request = request
        self._sa_session = request.Session()
        self._account_id = request.session['user_id']
        self.controller = PvcLogisticsController(self._sa_session, account_id=self._account_id, request=self._request)

    def get_pickup_requests(self):
        return self.controller.get_pickup_request_report()

    def get_delivery_requests(self):
        return self.controller.get_delivery_request_report()

    def get_license_master_data_state_readonly(self, _id):

        # FIXME: should go into iqportal with a proper controller
        _license = self._sa_session.query(License).filter(License.id == _id).one()
        lc = _license.controller(self._sa_session, account_id=self._account_id, request=self._request)
        return lc.load_master_data_state(_license)


class PickupRequestHandler:

    def __init__(self, request, **kwargs):
        self.sa_session = request.Session()
        self.account_id = request.session['user_id']
        self._controller = PvcLogisticsController(self.sa_session, account_id=self.account_id, request=request)

    def get_holders(self):
        return [ {'label': x.name, 'value': x.id} for x in self._controller.get_holders() ]

    def validate_save_pickup_request(self, args):
        return FormValidator({
            'holder_id': [(plugins.Required, {})],
            'technology': [(plugins.Required, {})],
            'condition': [(plugins.Required, {})],
            'number_modules': [(plugins.Required, {})],
            'weight_holder': [(plugins.Required, {})],
        }).validate(args)

    def save_or_update_pickup_request(self, pickup_request_id, values):
        return self._controller.save_or_update_pickup_request(pickup_request_id, values).id

    def load_pickup_request(self, pickup_request_id):
        return self._controller.load_pickup_request(pickup_request_id)

    def submit_pickup_request(self, pickup_request_id):
        return self._controller.submit_pickup_request(pickup_request_id)

class DeliveryRequestHandler:

    def __init__(self, request, **kwargs):
        self.sa_session = request.Session()
        self.account_id = request.session['user_id']
        self._controller = PvcLogisticsController(self.sa_session, account_id=self.account_id, request=request)

    def get_carriers(self):
        return [ {'label': x.name, 'value': x.id} for x in self._controller.get_carriers() ]

    def get_treatment_centers(self):
        return [ {'label': x['name'], 'value': x['dbID']} for x in self._controller.get_treatment_centers_state() ]

    def get_pending_pickup_requests(self):
        return [{
            'label': '%03d - %s' % (x.id, x.holder.name),
            'value': x.id
        } for x in self._controller.get_pending_pickup_requests() ]

    def get_pickup_requests_from_delivery(self, delivery_request_id):

        if delivery_request_id is None:
            return []

        return [{
            'label': '%03d - %s' % (x.id, x.holder.name),
            'value': x.id
        } for x in self._controller.get_pickup_requests_from_delivery(delivery_request_id) ]

    def validate_save_delivery_request(self, args, delivery_request_id=None):

        v = {
            'carrier_id': [(plugins.Required, {})],
            'treatment_center_id': [(plugins.Required, {})],
            'pickup_requests': [(plugins.Required, {})],
            'total_units': [(plugins.Required, {})],
            #'weight_holder': [(plugins.Required, {})],
        }

        if delivery_request_id is not None:
            delivery_request = self._controller.get_delivery_request(delivery_request_id)
            if delivery_request.status == PvcDeliveryRequest.STATUS_ACCEPTED:
                v['pickup_request_id'] = [(plugins.Required, {})]
                v['weight_tc'] = [(plugins.Required, {})]
                v['wtf_upload'] = [(plugins.Required, {})]
                v['annex7_file'] = [(plugins.Required, {})]
                v['total_kilometers'] = [(plugins.Required, {})]

        return FormValidator(v).validate(args)

    def save_or_update_delivery_request(self, delivery_request_id, values):
        return self._controller.save_or_update_delivery_request(delivery_request_id, values).id

    def load_delivery_request(self, delivery_request_id):
        return self._controller.load_delivery_request(delivery_request_id)

    def get_pickup_totals(self, pickup_request_ids):
        pickup_requests = [self._controller.get_pickup_request(int(x)) for x in pickup_request_ids]
        return {
            'total_number_modules': sum( [0 if x.number_modules is None else x.number_modules for x in pickup_requests] ),
            'total_weight_holder': sum( [0 if x.weight_holder is None else x.weight_holder for x in pickup_requests] ),
            'total_weight_tc': sum( [0 if x.weight_tc is None else x.weight_tc for x in pickup_requests] ),
        }

    def delivery_request_assign_to_courier(self, delivery_request_id):
        return self._controller.delivery_request_assign_to_courier(delivery_request_id)

    def delivery_request_accept_transport(self, delivery_request_id):
        return self._controller.delivery_request_accept_transport(delivery_request_id)

    def complete_delivery_request(self, delivery_request_id):
        return self._controller.delivery_request_complete(delivery_request_id)

    def approve_delivery_request(self, delivery_request_id):
        return self._controller.approve_delivery_request(delivery_request_id)


handlers = (
    LogisticsRpcHandler,
    PickupRequestHandler,
    DeliveryRequestHandler,
)
