# Copyright (c) 2015 BaseSoft B.V.
# See LICENSE for details.

import datetime

from sqlalchemy.orm import mapper, relation
from sqlalchemy.schema import UniqueConstraint
from sqlalchemy import Integer, Column, UnicodeText, DateTime, ForeignKey, SmallInteger, Boolean, Numeric, Date, Unicode

from basesoft.sqlalchemy.table import IQTable
from basesoft.web.contrib.file.models import File
from basesoft.web.contrib.license import models as license_models

from pvcycle.logistics.models import PvcPickupRequest, PvcDeliveryRequest, PvcHolderLicense

__mapping_defined = False
__tables_defined = False
tables = {}

def reset():
    global __mapping_defined
    global __tables_defined

    __mapping_defined = False
    __tables_defined = False

def get_tables():
    return tables

def do_tables(metadata):
    """
    This function is called by Django middleware when needed.
    """
    global __tables_defined
    if __tables_defined:
        return

    tblPvcHolderLicense = IQTable('tblPvcHolderLicense', metadata,
        Column('license_id', Integer, ForeignKey('tblLicense.id', ondelete='CASCADE'), primary_key=True),
        Column('holder_type', SmallInteger, nullable=False, default=PvcHolderLicense.TYPE_COLLECTION_POINT)
    )

    tblPvcCarrierLicense = IQTable('tblPvcCarrierLicense', metadata,
        Column('license_id', Integer, ForeignKey('tblLicense.id', ondelete='CASCADE'), primary_key=True)
    )

    tblPvcPickupRequest = IQTable('tblPvcPickupRequest', metadata,
        Column('id', Integer, primary_key=True, autoincrement=True),
        #Column('tbs_id', Integer, ForeignKey('tblPvcTbsLicense.license_id'), nullable=True),
        Column('holder_id', Integer, ForeignKey('tblPvcHolderLicense.license_id'), nullable=False),  # TODO, this can be null if its a public request
        Column('delivery_request_id', Integer, ForeignKey('tblPvcDeliveryRequest.id'), nullable=True),  # not assigned when created
        Column('technology', SmallInteger, nullable=False, default=PvcPickupRequest.TECH_ASI),
        Column('condition', SmallInteger, nullable=False, default=PvcPickupRequest.CONDITION_NEW),
        Column('number_modules', Integer, nullable=False),
        Column('weight_holder', Integer, nullable=False),
        Column('weight_tc', Integer, nullable=True),
        Column('status', SmallInteger, nullable=False, default=PvcDeliveryRequest.STATUS_DRAFT),
        Column('photo_file_id', Integer, ForeignKey('tblFile.id', ondelete="CASCADE"), nullable=True),
        Column('wtf_file_id', Integer, ForeignKey('tblFile.id', ondelete="CASCADE"), nullable=True),
        Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
        Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now, onupdate=datetime.datetime.now),
    )

    tblPvcDeliveryRequest = IQTable('tblPvcDeliveryRequest', metadata,
        Column('id', Integer, primary_key=True, autoincrement=True),
        Column('carrier_id', Integer, ForeignKey('tblPvcCarrierLicense.license_id'), nullable=False),
        Column('treatment_center_id', Integer, ForeignKey('tblPvcTreatmentCenter.company_id'), nullable=False),
        Column('total_units', Integer, nullable=False), # filled in by key-user
        Column('total_kilometers', Integer, nullable=True), # filled in by carrier when route is completed
        Column('annex7_file_id', Integer, ForeignKey('tblFile.id', ondelete="CASCADE"), nullable=True),
        Column('status', SmallInteger, nullable=False, default=PvcPickupRequest.STATUS_DRAFT),
        Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
        Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now, onupdate=datetime.datetime.now),
    )

    tblPvcTreatmentCenter = IQTable('tblPvcTreatmentCenter', metadata,
        Column('company_id', Integer, ForeignKey('tblCompany.id', ondelete='CASCADE'), primary_key=True)
    )


    tables['tblPvcHolderLicense'] = tblPvcHolderLicense
    tables['tblPvcCarrierLicense'] = tblPvcCarrierLicense

    tables['tblPvcPickupRequest'] = tblPvcPickupRequest
    tables['tblPvcDeliveryRequest'] = tblPvcDeliveryRequest

    tables['tblPvcTreatmentCenter'] = tblPvcTreatmentCenter

    __tables_defined = True

def do_mappings():
    """
    This function is called by Django middleware when needed.
    """
    global __mapping_defined
    if __mapping_defined:
        return

    from pvcycle.core import models as core_models
    from pvcycle.logistics import models as logistics_models
    from basesoft.web.contrib.contact import models as contact_models
    from basesoft.web.contrib.notification import models as notification_models
    from basesoft.web.contrib.file.sa_datamapper import get_tables as file_tables

    mapper(logistics_models.PvcHolderLicense, tables['tblPvcHolderLicense'],
        polymorphic_identity='PvcHolderLicense',
        inherits=license_models.License
    )

    mapper(logistics_models.PvcCarrierLicense, tables['tblPvcCarrierLicense'],
        polymorphic_identity='PvcCarrierLicense',
        inherits=license_models.License
    )

    mapper(logistics_models.PvcTreatmentCenter, tables['tblPvcTreatmentCenter'],
        polymorphic_identity='PvcTreatmentCenter',
        inherits=contact_models.Company
    )

    mapper(logistics_models.PvcPickupRequest, tables['tblPvcPickupRequest'],
        properties = {
            #'tbs': relation(core_models.PvcTbsLicense),
            'holder': relation(logistics_models.PvcHolderLicense),
            'delivery_request': relation(logistics_models.PvcDeliveryRequest, backref='pickup_requests'),
            'photo_file': relation(File, primaryjoin=file_tables()['tblFile'].c.id==tables['tblPvcPickupRequest'].c.photo_file_id),
            'wtf_file': relation(File, primaryjoin=file_tables()['tblFile'].c.id==tables['tblPvcPickupRequest'].c.wtf_file_id),
        }
    )

    mapper(logistics_models.PvcDeliveryRequest, tables['tblPvcDeliveryRequest'],
        properties = {
            #'tbs': relation(core_models.PvcTbsLicense),
            'carrier': relation(logistics_models.PvcCarrierLicense),
            'treatment_center': relation(logistics_models.PvcTreatmentCenter),
            'annex7_file': relation(File),
        }
    )

    mapper(logistics_models.PvcPickupRequestNotification,
        inherits=notification_models.Notification,
        polymorphic_identity='PvcPickupRequestNotification'
    )

    mapper(logistics_models.PvcDeliveryRequestNotification,
        inherits=notification_models.Notification,
        polymorphic_identity='PvcDeliveryRequestNotification'
    )

    __mapping_defined = True

def fill_data(conn):
    pass

def app_permissions_category():
    """
    Returns a category for all permissions defined for this app
    """
    return license_models.PermissionCategory(u'PVC Logistics')

def app_permissions():
    from django.utils.translation import ugettext as _
    return [

        license_models.Permission(u'pvc_logistics', u"""
            Temporary permission to access the PV CYCLE Logistics demo
            """, "PVC Logistics Demo"),

        license_models.Permission(u'pvc_logistics_key_user', u"""
            Temporary permission to access the PV CYCLE Logistics demo
            """, "PVC Logistics Demo (key user)"),

        license_models.Permission(u'pvc_logistics_holder', u"""
            Temporary permission to access the PV CYCLE Logistics demo
            """, "PVC Logistics Demo (holder)"),

        license_models.Permission(u'pvc_logistics_carrier', u"""
            Temporary permission to access the PV CYCLE Logistics demo
            """, "PVC Logistics Demo (carrier)"),

    ]
