'''
Created on Oct 8, 2013

@author: marten
'''
import datetime, math
from decimal import Decimal
from basesoft.web.contrib.i18n.models import Country
from basesoft.web.contrib.notification.controllers import NotificationController
from django.utils.translation import ugettext as _, ugettext_lazy
from sqlalchemy.sql.expression import and_, extract, not_, or_
from dateutil.relativedelta import relativedelta

from basesoft.util import virtual_date
from basesoft.util.virtual_date import today, now
from basesoft.web.contrib.order.models import OrderPaymentStatus, OrderRuleCondition, Order, OrderLineRule, \
    OrderRuleAction
from basesoft.web.core.exceptions import SystemException, PublicUserException, SecurityException,\
    PublicSecurityException
from basesoft.web.contrib.order.controllers import AbstractOrderController, TaxController
from basesoft.web.core.util.controllers import LicenseConfigController, transaction_decorator, lazy_assert

from iqportal.iqube.core.controllers import IQubeAggregator
from iqportal.weee.core.models import WEEEProductLevel

from pvcycle.core.controllers import AbstractPvcController, PvcNotificationController, PvcTBSController, \
    PvcMemberController
from pvcycle.order.models import PvcOrder, PvcOrderLineAdministrationFee, PvcDeclarationOrder, PvcDeclarationOrderLine,\
    PvcOrderLineMembershipFee, PvcDeclarationReminderMail, PvcDeclarationOrderLineMinimumCF, PvcDeclarationNotification, \
    PvcContributionOrder, PvcContributionOrderLine
from pvcycle.core.models import PvcTbsMember, PvcTbsLicense,\
    PvcTbsMemberCertificate, PvcOperatorLicense, PvcMemberLicense, PvcTbsServiceLevel, PvcTbsMemberServiceLevel
from pvcycle.order.rule.conditions import PvcRuleConditionContributionIsCompliant, PvcRuleConditionContributionTbsMember, \
    PvcRuleConditionContributionLineTbsMemberProduct, PvcRuleConditionTbs, PvcRuleConditionMember
from pvcycle.order.rule.conditions import PvcRuleConditionContributionYear
from pvcycle.order.rule.conditions import PvcRuleConditionContributionQuarter
from pvcycle.order.rule.conditions import PvcRuleConditionContributionMonth
from pvcycle.order.rule.conditions import PvcRuleConditionContributionProducer
from pvcycle.order.rule.conditions import PvcRuleConditionIsFreeRider
from pvcycle.order.rule.conditions import PvcRuleConditionContributionIsForecast
from pvcycle.order.rule.actions import PvcRuleActionApplyMemberCountryVatRate
from pvcycle.order.rule.actions import PvcRuleActionContributionB2bPercentage
from pvcycle.product.controllers import PvcProductController
from pvcycle.product.models import PvcTbsProduct, PvcTbsMemberProduct, PvcTbsProductFee
from pvcycle.pvmodule.models import PvcPVModule, PvcTbsMemberPVModule
from pvcycle.core.exceptions import PvcPVModuleNotApprovedException, PvcServiceLevelException, \
    PvcServiceLevelNoChangeException
from pvcycle.pvmodule.controllers import PvcPVModuleController
from pvcycle.subrogation.models import PvcSubrogation

class PvcOrderController(AbstractOrderController, AbstractPvcController):
    _instance_class = PvcOrder

    def _load_lines_state(self, lines, for_export=False):
        if for_export:
            return self.__load_lines_state_for_export(lines)
        return super(PvcOrderController, self)._load_lines_state(lines)

    def __load_lines_state_for_export(self, lines):
        rv = []
        for line in lines:
            props = []
            if isinstance(line, PvcOrderLineAdministrationFee):
                if line.tbs_member_service_level.service_level is None:
                    name = line.tbs_member_service_level.pv_service_level.name
                else:
                    name = line.tbs_member_service_level.service_level.name
                props.append({
                    'key': 'service_level',
                    'value_str': name,
                })
            elif isinstance(line, PvcOrderLineMembershipFee):
                props.append({
                    'key': 'membership',
                    'value_str': line.tbs_member_membership.membership.name,
                })

            rv.append({
                'index': line.index,
                'polymorphic_type': line.polymorphic_type,
                'description': line.description,
                'price_per_unit': line.price_per_unit,
                'quantity': line.quantity,
                'vat_rate': line.vat_rate,
                'gl_account': line.gl_account_number,
                'line_amount_excluding_vat': line.line_amount_excluding_vat,
                'line_amount_including_vat': line.line_amount_including_vat,
                'properties': props,
            })
        return rv

    def _get_available_rule_actions(self, order_type):

        actions = super(PvcOrderController, self)._get_available_rule_actions(order_type)
        actions += [
            PvcRuleActionApplyMemberCountryVatRate(self),
        ]
        return actions

    def _get_available_rule_conditions(self, order_type):

        conditions = super(PvcOrderController, self)._get_available_rule_conditions(order_type)

        # override merchant / client
        conditions[0] = PvcRuleConditionTbs(self)
        conditions[1] = PvcRuleConditionMember(self)

        conditions += [
            PvcRuleConditionIsFreeRider(self),
        ]
        return conditions

    def set_additional_fields(self):
        self._set_additional_fields({
            self._instance_class: ['tbs_member']
        }, {
            PvcOrderLineAdministrationFee: ['tbs_member_service_level'],
            PvcOrderLineMembershipFee: ['tbs_member_membership'],
            OrderLineRule: ['tax_class_code', 'gl_account_code'],  # FIXME this should be added through parent!
        })

    @transaction_decorator
    def trigger_administration_fee_order(self, tbs_member):
        """
        This is triggered on registration of a new TBS Member; OR in case the member upgrades their Service Level.
        """

        assert isinstance(tbs_member, PvcTbsMember)

        if tbs_member.status != PvcTbsMember.STATUS_APPROVED:
            raise SystemException('Trying to trigger an Administration/Membership Order for an unapproved TBS Member')

        # we get if this TBS uses Memberships instead of Service Levels
        lcc = LicenseConfigController(self.sa_session, scope=tbs_member.tbs)
        use_membership = lcc.get('pvc.tbs.use.membership.instead.of.service.level')

        active_level = None
        if use_membership:
            fee_type = _('Membership Fee')
            type_title = _('Membership')
            fl = PvcOrderLineMembershipFee # Memberships use another ORDER line type.
            active_level = tbs_member.active_membership
            years = lcc.get('pvc.tbs.membership.valid.for.years')
            year_check = extract('year', PvcOrder.display_date) + years > today().year

        else:
            years = 1
            fee_type = _('Administration Fee')
            type_title = _('Service Level')
            fl = PvcOrderLineAdministrationFee
            active_level = tbs_member.active_service_level
            year_check = extract('year', PvcOrder.display_date) == today().year
            if active_level is None and not tbs_member.is_producer:
                return None # COMMISSIONERS do not have Service Levels.

        if active_level is None:
            raise PublicUserException(_('Cannot approve a member who has no active %(membership_type)s.' % {'membership_type': type_title}))

        #current_fee = getattr(active_level, active_level.BACKREF).get_current_administration_fee()
        lines = []
        level_fees = active_level.get_current_administration_fees()
        if len(level_fees) == 0:
            raise PublicUserException(_('Cannot approve a member without an %(fee_type)s set on %(membership_type)s: %(membership_name)s.' % {
                'fee_type' : fee_type,
                'membership_type': type_title,
                'membership_name': getattr(active_level, active_level.BACKREF).name
            }))

        for line in level_fees:

            line_title, price, period = line

            # check if the member already has a AF for this year (or period for memberships)
            q = self.sa_session.query(PvcOrder).join(fl) \
                .filter(
                    year_check,
                    getattr(fl, fl.BACKREF)==active_level,
                )
            if len(q.all()) > 0:
                raise PvcAdminFeeAlreadyExists('%(fee_type)s already exists for year %(year)s' % {'fee_type': fee_type, 'year': today().year})

            #period = getattr(active_level, active_level.BACKREF).get_current_fee_start_year()

            if years > 1:
                period = '%s - %s' % (str(period), str(period+years-1))
                description = _('%(fee_type)s %(line_title)s for period %(period)s' % {
                    'fee_type': fee_type,
                    'line_title': line_title,
                    'period': period
                })
            else:
                description = _('%(fee_type)s %(line_title)s for year %(year)s' % {
                    'fee_type': fee_type,
                    'line_title': line_title,
                    'year': period
                })

            # create the AF order
            lines.append({
                'polymorphic_type': fl,
                'description': description,
                'quantity': 1,
                'price': price,
                fl.BACKREF: active_level,
            })

        self.logger.info('trigger_administration_fee_order: Created new Administration/Membership Fee Order for TBS Member [%s]' % tbs_member.id)
        order = self._save_or_update_order(None, lines, merchant=tbs_member.tbs, scope=tbs_member.member.default_group, tbs_member=tbs_member, display_date=now())

        # if the new order has 0 amount, trigger the certificate directly
        if order is not None and order.invoiced_amount == 0:
            self.create_certificate(order)

        return order

    def _is_administrative_fee_order(self, order):

        # check if its a Administrative/Membership Fee order
        is_af_order = False
        for line in order.lines:
            if line.polymorphic_type == 'PvcOrderLineAdministrationFee':
                is_af_order = True
                break
            elif line.polymorphic_type == 'PvcOrderLineMembershipFee':
                is_af_order = True
                break

        return is_af_order

    @transaction_decorator
    def trigger_yearly_administrative_fees(self, tbs):
        """
        Called by the pvcycle YearlyOrderTask to trigger new AF / Membership orders
        """

        lcc = LicenseConfigController(self.sa_session, scope=tbs)
        fee_day = lcc.get('pvc.tbs.fee.day')
        if today().month != 1 or today().day != fee_day:
            self.logger.info('today (%s) is not the pvc.tbs.fee.day (%s)' % (today(), fee_day))
            return False

        # query all TbsMembers that are supposed to receive a new AF / Membership order (status STATUS_APPROVED)
        tbs_members = self.sa_session.query(PvcTbsMember) \
            .filter(PvcTbsMember.tbs == tbs) \
            .filter(PvcTbsMember.status == PvcTbsMember.STATUS_APPROVED) \
            .all()

        for tbs_member in tbs_members:
            try:
                self.trigger_administration_fee_order(tbs_member)

            # its possible that a fee already exists
            # - the task is run twice
            # - membership fee's are only created once per 'pvc.tbs.membership.valid.for.years'
            except PvcAdminFeeAlreadyExists, e:
                self.logger.info("trigger_yearly_administrative_fees: skipping order for tbs_member [%s] because [%s]" % (tbs_member.id, e.message))

            # ignore public user exceptions
            # - member does not have active server/membership level
            except PublicUserException, e:
                self.logger.info("trigger_yearly_administrative_fees: skipping order for tbs_member [%s] because [%s]" % (tbs_member.id, e.message))

        # secondly, update everyone that is STATUS_RESIGNING to STATUS_RESIGNED
        tbs_members = self.sa_session.query(PvcTbsMember) \
            .filter(PvcTbsMember.tbs == tbs) \
            .filter(PvcTbsMember.status == PvcTbsMember.STATUS_RESIGNING) \
            .all()
        for tbs_member in tbs_members:
            tbs_member.status = PvcTbsMember.STATUS_RESIGNED
            self.logger.info("trigger_yearly_administrative_fees: updated tbs_member [%s] to status resigned" % tbs_member)

        return True

    def trigger_change_service_level__administrative_fees(self, tbs_member, new_service_level, difference):
        assert isinstance(tbs_member, PvcTbsMember)
        assert isinstance(new_service_level, PvcTbsMemberServiceLevel)
        assert difference > 0

        # make a new AF order with the difference
        lines = [{
            'polymorphic_type': PvcOrderLineAdministrationFee,
            'description': _('Administration Fee for service level upgrade in year %s' % today().year),
            'quantity': 1,
            'price': difference,
            'tbs_member_service_level': new_service_level,
        }]

        self.logger.info('change_service_level: Created new Administration Fee Order for TBS Member [%s]' % tbs_member.id)
        order = self._save_or_update_order(None, lines, merchant=tbs_member.tbs, scope=tbs_member.member.default_group, tbs_member=tbs_member, display_date=now())

    @transaction_decorator
    def payment_status_change_cb(self, order):
        """
        Called by the payment controller when the order payment changes.
        """

        if order.get_current_payment_status() == OrderPaymentStatus.PAYMENT_COMPLETED:

            if self._is_administrative_fee_order(order):
                self.create_certificate(order)

    def create_certificate(self, order):
        """
        Gives the owner of the order a certificate. Called by payment_status_change_cb() or when the order amount is 0.
        """
        assert self._is_administrative_fee_order(order)

        # create a new certificate for this member
        tbs_member = self.sa_session.query(PvcTbsMember) \
            .filter(PvcTbsMember.tbs == order.merchant, PvcTbsMember.member == order.client).one()

        year = order.display_date.year

        cert = PvcTbsMemberCertificate()
        cert.year = year
        cert.tbs_member = tbs_member
        cert.status = PvcTbsMemberCertificate.STATUS_PAID
        cert.hash = self.generate_unique_hash()
        self.sa_session.add(cert)

        self.logger.info('payment_status_change_cb: Gave PVC certificate to tbs_member_id [%s] and year [%s]' % (tbs_member.id, year) )

    def generate_unique_hash(self):
        """
        Generates a unique hash.
        """
        import random
        from hashlib import sha256 as sha
        while True:
            _hash = sha(str(random.random())).hexdigest()[:8]
            if self.sa_session.query(PvcTbsMemberCertificate).filter(PvcTbsMemberCertificate.hash == _hash).count() == 0:
                break
        return _hash


class PvcAdminFeeAlreadyExists(PublicUserException): pass


class PvcAbstractContributionController(AbstractOrderController, AbstractPvcController):

    def _get_original_order(self, order):
        if order.parent_order is None:
            return order
        return self._get_original_order(order.parent_order)

    def _is_undeclared_period(self, tbs_member, year, quarter, month, producer, return_order=False):

        order = self.sa_session.query(self._instance_class).filter(
            and_(
                self._instance_class.tbs_member==tbs_member,
                self._instance_class.year==year,
                self._instance_class.quarter==quarter,
                self._instance_class.month==month,
                self._instance_class.producer==producer
            )
        ).first()

        if order is None:
            return True
        if return_order:
            return order
        return False

    def _get_possible_periods_for_yearly_declarations(self, min_year):
        td = today()
        max_year = td.year-1
        if td.month == 1 and td.day == 1:
            max_year = max_year-1
        if max_year < min_year:
            return None

        years = range(min_year, max_year+1)
        years.reverse()

        rv = {}
        for year in years:
            rv[year] = {}

        return rv

    def _get_possible_periods_for_quarterly_declarations(self, min_year):
        td = today()
        max_year = td.year
        import math
        max_q = int(math.floor((td.month-1)/3))
        if td.month%3 == 1 and td.day == 1:
            max_q = max_q-1
        if max_q < 1:
            max_q = 4
            max_year = max_year-1
        if max_year < min_year:
            return None

        years = range(min_year, max_year+1)
        years.reverse()

        rv = {}
        for year in years:
            quarters = []
            while max_q > 0:
                quarters.insert(0, max_q)
                max_q -= 1
            rv[year] = {'quarters': quarters}
            max_q = 4

        return rv

    def _get_possible_periods_for_monthly_declarations(self, min_year):
        td = today()
        max_year = td.year
        max_month = td.month-1 if td.day > 1 else td.month-2
        if max_month < 1:
            max_month = 12
            max_year = max_year-1
        if max_year < min_year:
            return None

        years = range(min_year, max_year+1)
        years.reverse()

        rv = {}
        for year in years:
            months = []
            while max_month > 0:
                months.insert(0, max_month)
                max_month -= 1
            rv[year] = {'months': months}
            max_month = 12

        return rv

    @transaction_decorator
    def request_for_resubmission(self, _id):
        self.assert_permissions('pvc_declare_declarations')

        order = self.get(_id, priv=self._instance_class.PRIV_VIEW).pop()
        if not self.has_privileges_on_object(order, self._instance_class.PRIV_UPDATE):
            if order.tbs_member != order.producer:
                raise PublicUserException(_('A resubmit request can only be done by the producer.'))
            raise PublicUserException(_('You cannot make a resubmit request for this order.'))

        if isinstance(order, PvcContributionOrder) and order.is_forecast:
            raise PublicUserException(_('You cannot make a resubmit request for forecasts.'))

        if order.status != self._instance_class.STATUS_SUBMITTED:
            raise PublicUserException(_('You cannot make a resubmit request for an unsubmitted order.'))

        if order.resubmit_request is not None:
            raise PublicUserException(_('No resubmit request possible for this order.'))

        order.resubmit_request = self._instance_class.RESUBMIT_REQUEST_PENDING
        self.sa_session.add(order)

        period = self.__get_period_label(order)

        nc = PvcNotificationController(self.sa_session, account_id=self.get_account().id)
        operators = self.sa_session.query(PvcOperatorLicense).filter(PvcOperatorLicense.domain==order.tbs_member.tbs.domain).all()
        recipients = self.get_groups_in_license_with_permissions('pvc_tbs_member_manager', license_id=order.tbs_member.tbs_id)
        recipients = recipients + [o.default_group for o in operators]

        subject = _('Resubmit request by member "%(member)s" at TBS "%(tbs)s".' %
                    {'member': order.tbs_member.member.name, 'tbs': order.tbs_member.tbs.name})
        message = _('The member "%(member)s" at TBS "%(tbs)s" has made a request for resubmission for period "%(period)s".' %
                    {'member': order.tbs_member.member.name, 'tbs': order.tbs_member.tbs.name, 'period': period})
        callback = {
            'js_module': 'pvc_contribution_manager' if isinstance(order, PvcContributionOrder) else 'pvc_declaration_manager',
            'parameters': [order.id]
        }
        notification = nc.set_notification(subject, message, instance_class=PvcDeclarationNotification, callback=callback)
        nc.create_system_notification(notification, recipients, sender=order.tbs_member.member.default_group, for_license=True)

        self.logger.info('request_for_resubmission: requested for order %s' % order.id)

    @transaction_decorator
    def approve_resubmit_request(self, _id):
        self.assert_permissions('pvc_approve_declaration_resubmit')
        order = self.get(_id, priv=self._instance_class.PRIV_VIEW).pop()

        if order.resubmit_request != self._instance_class.RESUBMIT_REQUEST_PENDING:
            raise PublicUserException(_('This order (#%s) does not have an outstanding resubmit request.' % order.id))

        order.resubmit_request = self._instance_class.RESUBMIT_REQUEST_APPROVED
        self.sa_session.add(order)

        period = self.__get_period_label(order)

        nc = PvcNotificationController(self.sa_session, account_id=self.get_account().id)
        recipients = self.get_groups_in_license_with_permissions('pvc_member_license', license_id=order.tbs_member.member_id)

        subject = _('Resubmit request approved')
        msg = _('Your request for resubmission at TBS "%(tbs)s" for period "%(period)s" has been approved. '
                'Please navigate to the declaration overview to make your resubmission.' %
                {'tbs': order.tbs_member.tbs.name, 'period': period})
        callback = {
            'js_module': 'pvc_declaration_manager',
            'parameters': [order.id]
        }
        notification = nc.set_notification(subject, msg, instance_class=PvcDeclarationNotification, callback=callback)
        nc.create_notification(notification, recipients, for_license=True)

        self.logger.info('approve_resubmit_request: approved for order %s' % order.id)

    @transaction_decorator
    def reject_resubmit_request(self, _id, reason):
        self.assert_permissions('pvc_approve_declaration_resubmit')
        order = self.get(_id, priv=self._instance_class.PRIV_VIEW).pop()

        if order.resubmit_request != self._instance_class.RESUBMIT_REQUEST_PENDING:
            raise PublicUserException(_('This order (#%s) does not have an outstanding resubmit request.' % order.id))

        order.resubmit_request = None
        self.sa_session.add(order)

        period = self.__get_period_label(order)

        nc = PvcNotificationController(self.sa_session, account_id=self.get_account().id)
        recipients = self.get_groups_in_license_with_permissions('pvc_member_license', license_id=order.tbs_member.member_id)

        subject = _('Resubmit request rejected')
        msg = _('Your request for resubmission at TBS "%(tbs)s" for period "%(period)s" has been rejected with the following reason: %(reason)s.' %
                {'tbs': order.tbs_member.tbs.name,
                 'period': period,
                 'reason': reason})

        notification = nc.set_notification(subject, msg, instance_class=PvcDeclarationNotification)
        nc.create_notification(notification, recipients, for_license=True)

        self.logger.info('reject_resubmit_request: rejected for order %s' % order.id)


    @transaction_decorator
    def cancel_resubmit_request(self, _id):
        self.assert_permissions('pvc_approve_declaration_resubmit')
        order = self.get(_id, priv=self._instance_class.PRIV_VIEW).pop()

        if order.resubmit_request != self._instance_class.RESUBMIT_REQUEST_APPROVED:
            raise PublicUserException(_('This order (#%s) has no approved resubmit request.' % order.id))

        order.resubmit_request = None
        self.sa_session.add(order)

        self.logger.info('cancel_resubmit_request: canceled for order %s' % order.id)


    def __get_period_label(self, order):
        period = order.year
        if order.month is not None:
            period = '%s-%s' % (order.month, order.year)
        elif order.quarter is not None:
            period = 'Q%s-%s' % (order.quarter, order.year)
        return period

    @staticmethod
    def is_declaration_period(periodicity, closing_day):
        import math
        delta = datetime.timedelta(days=closing_day)
        t = today()
        if periodicity == PvcTbsLicense.DECLARATION_PERIOD_YEARLY:
            m = 1
        elif periodicity == PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY:
            m = int(math.floor((t.month-1)/3)*3+1)
        else: # monthly
            m = t.month

        return (datetime.date(t.year, m, 1)+delta) > t

class PvcDeclarationOrderController(PvcAbstractContributionController):
    """
    Legacy controller. Used only for viewing old declarations.
    """

    _instance_class = PvcDeclarationOrder

    merchant_uses_weight = None

    def _assert_valid_line_removal(self, order, line):
        super(PvcDeclarationOrderController, self)._assert_valid_line_removal(order, line)
        # rejected by commissioner is valid for removal.
        if line.tbs_member_module.subrogation is not None and line.tbs_member_module.commissioner_approval_status == PvcTbsMemberPVModule.COMMISSIONER_APPROVAL_STATUS_REJECTED:
            return
        if line.tbs_member_module.module.approval_status == PvcPVModule.PV_MODULE_STATUS_APPROVED:
            raise PublicUserException(_('You cannot remove a line with an approved PV Module. Please declare 0 instead and deactive from the PV Module overview for future declarations.'))

    def set_additional_fields(self):
        self._set_additional_fields({
            self._instance_class: ['tbs_member', 'year', 'month', 'quarter', 'producer']
        }, {
            PvcDeclarationOrderLine: ['tbs_member_module'],
        })

    def load_state(self, _id, subrogations=False):
        order = self.get(_id).pop()

        field = None

        lcc = LicenseConfigController(self.sa_session, scope=order.merchant)
        if lcc.get('pvc.tbs.declaration.use.weight.for.amount.calculations'):
            field = 'weight'

        rv = super(PvcDeclarationOrderController, self).load_state(order)
        rv['year'] = order.year
        rv['month'] = order.month
        rv['quarter'] = order.quarter
        rv['amount_calculation_field'] = field
        rv['tbs_member_id'] = order.tbs_member.id
        rv['producer_id'] = order.producer.id
        if subrogations:
            rv['subrogations'] = [{'value': s.id, 'label': s.commissioner.member.name} for s in self.get_available_subrogations_for_order(_id)]
            rv['can_pay'] = len(self.get(_id, priv=PvcDeclarationOrder.PRIV_PAY)) == 1
            rv['can_declare'] = len(self.get(_id, priv=PvcDeclarationOrder.PRIV_UPDATE)) == 1

        return rv

    def _load_lines_state(self, lines, for_export=False):
        if for_export:
            return self.__load_lines_state_for_export(lines)
        rv = []
        additional = []
        for line in lines:
            if isinstance(line, PvcDeclarationOrderLineMinimumCF):
                additional = super(PvcDeclarationOrderController, self)._load_lines_state([line])
                continue
            module = line.tbs_member_module.module
            rv.append({
                'id': line.id,
                'index': line.index,
                'description': line.description,
                'price_per_unit': line.price_per_unit,
                'quantity': line.quantity,
                'vat_rate': line.vat_rate,
                'gl_account': line.gl_account_number,
                'line_amount_excluding_vat': line.line_amount_excluding_vat,
                'line_amount_including_vat': line.line_amount_including_vat,
                'supplier': module.supplier.name,
                'type': module.type.name,
                'cell_type': module.cell_type.name,
                'nominal_power': module.nominal_power,
                'weight': module.weight,
                'country': line.tbs_member_module.country.name,
                'surface': module.width*module.length,
                'module_status': module.approval_status,
                'vat_rate': line.vat_rate,
            })

        rv = sorted(rv, key=lambda k: (k['country'], k['index']))
        return rv + additional

    def _load_state_for_export(self, order):
        rv = super(PvcDeclarationOrderController, self)._load_state_for_export(order)
        rv['properties'] = [{
            'key': 'year',
            'value_int': order.year
        },{
            'key': 'month',
            'value_int': order.month
        },{
            'key': 'quarter',
            'value_int': order.quarter
        },{
            'key': 'producer_id',
            'value_int': order.producer.member.id
        }]
        return rv

    def __load_lines_state_for_export(self, lines):
        rv = []
        for line in lines:
            if isinstance(line, PvcDeclarationOrderLineMinimumCF):
                rv += super(PvcDeclarationOrderController, self)._load_lines_state([line])
                continue

            module = line.tbs_member_module.module
            rv.append({
                'index': line.index,
                'polymorphic_type': line.polymorphic_type,
                'description': line.description,
                'price_per_unit': line.price_per_unit,
                'quantity': line.quantity,
                'vat_rate': line.vat_rate,
                'gl_account': line.gl_account_number,
                'line_amount_excluding_vat': line.line_amount_excluding_vat,
                'line_amount_including_vat': line.line_amount_including_vat,
                'properties': [{
                        'key': 'supplier',
                        'value_str': module.supplier.name,
                    },{
                        'key': 'type',
                        'value_str': module.type.name,
                    },{
                        'key': 'cell_type',
                        'value_str': module.cell_type.name,
                    },{
                        'key': 'nominal_power',
                        'value_decimal': module.nominal_power,
                    },{
                        'key': 'weight',
                        'value_decimal': module.weight,
                    },{
                        'key': 'country_name',
                        'value_str': line.tbs_member_module.country.name,
                    },{
                        'key': 'country_iso_3166_a2',
                        'value_str': line.tbs_member_module.country.iso_3166_a2,
                    },{
                        'key': 'country_iso_3166_a3',
                        'value_str': line.tbs_member_module.country.iso_3166_a3,
                    },{
                        'key': 'surface',
                        'value_decimal': module.width*module.length,
                    }],
            })

        return rv

    @transaction_decorator
    def _subrogate_order(self, order, quantities, subrogate, additional_lines=None):
        orders = {}

        line_modules = {}
        for line in order.lines:
            line_modules['line_%s' % line.index] = line

        for line, subrogation_id in subrogate.iteritems():
            if not subrogation_id in [s.id for s in order.tbs_member.subrogations]:
                raise PublicSecurityException(_('You do not have access to the supplied Subrogation.'))

            subrogation = [s for s in order.tbs_member.subrogations if s.id == subrogation_id].pop()

            if not subrogation.status == PvcSubrogation.STATUS_ACTIVE:
                raise PublicSecurityException(_('The supplied Subrogation is not active.'))

            if subrogation.commissioner is None:
                self.logger.warn('Approved subrogation without a commissioner, something went wrong, order: %s, subrogation: %s' % (order.id, subrogation.id))
                raise SecurityException('Approved subrogation without a commissioner, something went wrong!')

            if not orders.has_key(subrogation.id):
                orders[subrogation.id] = {'tbs_member': subrogation.commissioner, 'producer': order.tbs_member, 'lines': []}

            orders[subrogation.id]['lines'].append({
                'polymorphic_type': PvcDeclarationOrderLine,
                'description': '%s' % line_modules[line].description,
                'quantity': int(quantities[line]),
                'price': line_modules[line].price_per_unit,
                'tbs_member_module': self.__get_pv_modules_for_subrogation(line_modules[line].tbs_member_module, subrogation)
            })
            del quantities[line]

        fields = self._additional_order_fields[self._instance_class]
        kwargs = {}
        for f in fields:
            kwargs[f] = getattr(order, f)

        rv = []
        for subrogation_id, local_order in orders.iteritems():
            kwargs['producer'] = local_order['producer']
            kwargs['tbs_member'] = local_order['tbs_member']

            if not self._is_undeclared_period(**kwargs):
                raise PublicUserException(_('A declaration for this period already exists at commissioner %(commissioner)s, please use the resubmit option.' % {'commissioner': kwargs['tbs_member'].member.name}))

            o = self._save_or_update_order(None, local_order['lines'], merchant=order.merchant, scope=local_order['tbs_member'].member.default_group, **kwargs)
            o.status = self._instance_class.STATUS_PENDING
            self.sa_session.add(o)
            rv.append(o)

            self.sa_session.add(o)

        # remove lines from original order.
        if len(quantities) == 0:
            # all lines are moved to subrogations.
            self.sa_session.delete(order)
            self.sa_session.flush()
            return rv

        lines = []
        new_quantities = {}
        i = 0
        for line, quantity in quantities.iteritems():
            new_quantities['line_%d' % i] = int(quantity)
            lines.append({
                'polymorphic_type': PvcDeclarationOrderLine,
                'description': '%s' % line_modules[line].description,
                'quantity': int(quantity),
                'price': line_modules[line].price_per_unit,
                'tbs_member_module': line_modules[line].tbs_member_module,
            })
            i += 1
        order = self._save_or_update_order(order.id, lines, merchant=order.tbs_member.tbs, scope=order.tbs_member.member.default_group, \
                                          tbs_member=order.tbs_member, year=order.year, month=order.month, quarter=order.quarter, producer=order.producer)
        order.status = self._instance_class.STATUS_PENDING
        self.sa_session.add(order)

        order = self.update_declaration(order.id, new_quantities, additional_lines=additional_lines)

        rv.append(order)

        self.sa_session.flush()
        return rv

    def __get_pv_modules_for_subrogation(self, module, subrogation):
        """
        Get or create the PV Module supplied for this subrogation.
        """
        c = PvcPVModuleController(self.sa_session, account_id=self.get_account().id, request=self.request)
        return c.save_or_update_module_for_tbs_member(module.tbs_member, module.country, module.module, subrogation_id=subrogation)

    @transaction_decorator
    def save_default_privilege_object_group(self, order, target_group):
        """
        Override this method in order to assign privs based on Subrogation.
        """
        # normal behavior, no subrogation
        if order.tbs_member == order.producer:
            return super(PvcDeclarationOrderController, self).save_default_privilege_object_group(order, target_group=target_group)

        # give producer the right privs.
        super(PvcDeclarationOrderController, self).save_default_privilege_object_group(order, target_group=order.producer.member.default_group, privileges=order.producer_privileges)

        # give the commissioner the right privs.
        self._grant_or_revoke(order, privileges=order.tbs_member_privileges, grant=True, group=True, target_group=order.tbs_member.member.default_group, check_own_privs=False)

    @transaction_decorator
    def update_declaration(self, _id, quantities, additional_lines=None, subrogate=None):
        self.assert_permissions('pvc_declare_declarations')
        order = self.get(_id, priv=self._instance_class.PRIV_UPDATE).pop()
        if order.status == self._instance_class.STATUS_CREDITED:
            raise PublicSecurityException(_('You cannot update an already Credited order.'))

        if order.status not in [self._instance_class.STATUS_NEW, self._instance_class.STATUS_PENDING]:
            if order.resubmit_request != self._instance_class.RESUBMIT_REQUEST_APPROVED:
                raise PublicUserException(_('You cannot resubmit this order.'))
            return self.resubmit_declaration(order, quantities, additional_lines, subrogate=subrogate)

        if order.type == self._instance_class.TYPE_CREDIT:
            raise PublicSecurityException(_('You cannot update a Credit Order.'))

        # We are dealing with a normal Declare (or Redeclare)
        if len(quantities) < len(order.lines):
            raise PublicUserException(_('Number of lines for Order #%s does not match with supplied quantities.' % order.id))

        if subrogate is not None:
            if order.tbs_member != order.producer:
                raise PublicUserException(_('Cannot subrogate an Order which is already subrogated.'))

            return self._subrogate_order(order, quantities, subrogate, additional_lines)

        lines = []
        used_modules = []
        for line in order.lines:
            quantity = int(quantities['line_%s' % line.index])
            if quantity < 0:
                raise PublicUserException(_('All quantities must be larger or equal to zero.'))

            lines.append({
                'polymorphic_type': PvcDeclarationOrderLine,
                'description': '%s' % line.description,
                'quantity': quantity,
                'price': line.price_per_unit,
                'tbs_member_module': line.tbs_member_module
            })
            used_modules.append(line.tbs_member_module.id)

        # Additional lines as per ticket #156
        if additional_lines is not None and len(additional_lines) > 0:
            af = self._get_price(order.producer)

            possible = {}
            for m in order.producer.modules:
                if m.id in used_modules or m.module.approval_status == PvcPVModule.PV_MODULE_STATUS_REJECTED:
                    continue
                if m.subrogation is not None and m.commissioner_approval_status == PvcTbsMemberPVModule.COMMISSIONER_APPROVAL_STATUS_REJECTED:
                    continue
                possible['m_%s' % m.id] = m

            additional = [value for key, value in additional_lines.iteritems() if key[:18]=='tbs_member_module_']
            # each line should be unique on tbs_member_module level.
            for nl in additional:

                if nl in used_modules:
                    raise PublicUserException(_('New line contains a module which is already part of this Declaration.'))
                if not possible.has_key('m_%s' % nl):
                    raise PublicUserException(_('New line contains a module which is not active for this Declaration.'))

                module = possible['m_%s' % nl]
                quantity = int(additional_lines['quantity_additional_%s' % nl])
                if quantity < 0:
                    raise PublicUserException(_('All quantities must be larger or equal to zero.'))

                lines.append({
                    'polymorphic_type': PvcDeclarationOrderLine,
                    'description': '%s - %s' % (module.module.supplier.name, module.module.type.name),
                    'quantity': quantity,
                    'price': af,
                    'tbs_member_module': module,
                })

        order = self._save_or_update_order(order.id, lines, order.merchant, order.client.default_group)
        order.status = self._instance_class.STATUS_PENDING
        order.is_compliant = self._instance_class.IS_COMPLIANT # if the declaration got updated, they are compliant

        self.sa_session.add(order)
        return order

    @transaction_decorator
    def remove_line(self, _id, line_id, disable_module=None):
        if isinstance(_id, self._instance_class):
            order = _id
        else:
            orders = self.get(_id, priv=self._instance_class.PRIV_UPDATE)
            if len(orders) < 1:
                raise PublicSecurityException(_('Cannot update Order #%s' % _id))
            order = orders.pop()

        lines = [l for l in order.lines if l.id == line_id]
        if len(lines) == 0:
            raise PublicUserException(_('Cannot find line within order.'))

        line = lines[0]
        self._assert_valid_line_removal(order, line)
        if line.tbs_member_module.module.approval_status == PvcPVModule.PV_MODULE_STATUS_REJECTED:
            disable_module = PvcTbsMemberPVModule.IS_OBSOLETE

        if disable_module is not None:
            c = PvcPVModuleController(self.sa_session, account_id=self.account_id, request=self.request)
            c.change_tbs_member_module_active_state(line.tbs_member_module, disable_module)

        return super(PvcDeclarationOrderController, self).remove_line(order, line_id)

    @transaction_decorator
    def resubmit_declaration(self, order, quantities, additional_lines=None, subrogate=None):
        assert isinstance(order, self._instance_class)
        if not self.has_different_quantities(order, quantities) and additional_lines is None and subrogate is None:
            # make sure the order is set to compliant to prevent issue with minimum contribution fee
            order.is_compliant = PvcDeclarationOrder.IS_COMPLIANT
            self.sa_session.add(order)
            return order

        order.resubmit_request = None
        self.sa_session.add(order)

        credit = self._credit_order(order)
        order = self._create_child_order(credit, exclude_line_type=PvcDeclarationOrderLineMinimumCF)
        return self.update_declaration(order.id, quantities, additional_lines, subrogate=subrogate)

    def _assert_ready_for_approval(self, order):
        """
        If one or more PV Modules are not approved, we block this Declaration. See ticket #152.
        """
        for line in order.lines:
            if isinstance(line, PvcDeclarationOrderLineMinimumCF):
                continue
            module = line.tbs_member_module.module
            if line.tbs_member_module.subrogation is not None and line.tbs_member_module.commissioner_approval_status != PvcTbsMemberPVModule.COMMISSIONER_APPROVAL_STATUS_APPROVED:
                raise PvcPVModuleNotApprovedException(_('The PV Module "%(supplier)s - %(type)s" for member "%(member)s" at TBS "%(tbs)s" is not approved by the Commissioner.' % {
                    'supplier': module.supplier.name,
                    'type': module.type.name,
                    'member': order.client.name,
                    'tbs': order.merchant.name
                }))
            if module.approval_status != PvcPVModule.PV_MODULE_STATUS_APPROVED:
                raise PvcPVModuleNotApprovedException(_('The PV Module "%(supplier)s - %(type)s" for member "%(member)s" at TBS "%(tbs)s" is not approved.' % {
                    'supplier': module.supplier.name,
                    'type': module.type.name,
                    'member': order.client.name,
                    'tbs': order.merchant.name
                }))

    def _get_price(self, tbs_member):
        asl = tbs_member.active_service_level
        if asl is None:
            raise PublicUserException(_('Cannot add declarations for a Member without an active Service Level.'))
        af = asl.service_level.get_current_contribution_fee()
        if af is None:
            raise PublicUserException(_('Cannot add declarations for a Member without a Contribution Fee for this Service Level'))
        return af

    def get_new_lines(self, _id, module_id):
        self.assert_permissions('pvc_declare_declarations')
        order = self.get(_id, priv=self._instance_class.PRIV_UPDATE).pop()
        modules = order.producer.modules

        declaration_modules = [l.tbs_member_module for l in order.lines if hasattr(l, 'tbs_member_module')]

        _new = [module for module in modules if module not in declaration_modules]
        # dealing with subrogation
        if order.producer != order.tbs_member:
            # filtering only for the correct commmissioner
            _new = [module for module in _new if module.subrogation is not None and module.subrogation.commissioner == order.tbs_member]
        else:
            # filtering only for own producer modules (subrogation should be None)
            _new = [module for module in _new if module.subrogation is None]

        if len(_new) == 0:
            return []

        rv = []
        for tbs_member_module in _new:

            module = tbs_member_module.module
            if module.approval_status == PvcPVModule.PV_MODULE_STATUS_REJECTED:
                continue
            if tbs_member_module.subrogation is not None and tbs_member_module.commissioner_approval_status == PvcTbsMemberPVModule.COMMISSIONER_APPROVAL_STATUS_REJECTED:
                continue
            price = self._get_price(order.producer)
            tax_rate, vat_fee = self._get_vat_rate(False, self._instance_class, order.merchant, price)

            assert tax_rate is not None
            assert vat_fee is not None

            rv.append({
                'tbs_member_module_id': tbs_member_module.id,
                'description': '%s - %s' % (module.supplier.name, module.type.name),
                'price_per_unit': price,
                'quantity': 0,
                'vat_rate': tax_rate,
                'gl_account': None,
                'line_amount_excluding_vat': 0,
                'line_amount_including_vat': 0,
                'supplier': module.supplier.name,
                'type': module.type.name,
                'cell_type': module.cell_type.name,
                'nominal_power': module.nominal_power,
                'weight': module.weight,
                'country': tbs_member_module.country.name,
                'surface': module.width*module.length,
                'module_status': module.approval_status,
            })
        return rv

    def get_available_subrogations_for_order(self, _id):
        """
        This returns a list with Subrogations which do not yet have a Declaration for the same period as the supplied order.
        """
        orders = self.get(_id, priv=self._instance_class.PRIV_UPDATE)
        if len(orders) == 0: # this happens when the load state requester is the COMMISSIONER in case of subrogation
            return []
        order = orders.pop()

        if order.producer != order.tbs_member:
            return []

        if not self.has_active_subrogations(order.tbs_member):
            return []

        rv = []
        for subrogation in order.tbs_member.subrogations:
            if subrogation.status != PvcSubrogation.STATUS_ACTIVE:
                continue

            if not self._is_undeclared_period(subrogation.commissioner, order.year, order.quarter, order.month, order.producer):
                continue
            rv.append(subrogation)

        return rv

    def get_undeclared_periods(self, tbs_member):
        lcc = LicenseConfigController(self.sa_session, scope=tbs_member.tbs)
        period = lcc.get('pvc.tbs.declaration.period')
        if period == PvcTbsLicense.DECLARATION_PERIOD_YEARLY:
            min_year = tbs_member.created.year
            possible = self._get_possible_periods_for_yearly_declarations(min_year)
            if possible is None:
                return {}

            declarations = self.sa_session.query(PvcDeclarationOrder).filter(PvcDeclarationOrder.tbs_member==tbs_member).all()
            for d in declarations:
                if not possible.has_key(d.year):
                    continue

                del possible[d.year]
            return possible

        if period == PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY:
            min_year = tbs_member.created.year
            possible = self._get_possible_periods_for_quarterly_declarations(min_year)
            if possible is None:
                return {}

            declarations = self.sa_session.query(PvcDeclarationOrder).filter(PvcDeclarationOrder.tbs_member==tbs_member).all()
            for d in declarations:
                if not possible.has_key(d.year):
                    continue
                if d.quarter not in possible[d.year]['quarters']:
                    continue
                possible[d.year]['quarters'] = filter(lambda a: a != d.quarter, possible[d.year]['quarters'])
            rm_y = []
            for year, quarters in possible.iteritems():
                if len(quarters['quarters']) == 0:
                    rm_y.append(year)
            for year in rm_y:
                del possible[year]
            return possible

        if period == PvcTbsLicense.DECLARATION_PERIOD_MONTHLY:

            min_year = tbs_member.created.year
            possible = self._get_possible_periods_for_monthly_declarations(min_year)
            if possible is None:
                return {}

            declarations = self.sa_session.query(PvcDeclarationOrder).filter(PvcDeclarationOrder.tbs_member==tbs_member).all()
            for d in declarations:
                if not possible.has_key(d.year):
                    continue
                if d.month not in possible[d.year]['months']:
                    continue
                possible[d.year]['months'] = filter(lambda a: a != d.month, possible[d.year]['months'])
            rm_y = []
            for year, months in possible.iteritems():
                if len(months['months']) == 0:
                    rm_y.append(year)
            for year in rm_y:
                del possible[year]
            return possible

        raise SecurityException('Incorrect period set for TBS: %s' % tbs_member.tbs)

    @transaction_decorator
    def add_open_declaration(self, tbs_member, year, month=None, quarter=None, producer=None):
        """
        This will add a new Declaration for the provided TBS Member in the year supplied.
        Checks on month first quarter depending on the TBS setting.
        Raises an exception in case there already is a Declaration.

        Credits/Resubmits are done through other methods.
        """

        self.logger.info('add_open_declaration: tbs_member: %s, year=%s, month=%s, quarter=%s' % (tbs_member, year, month, quarter))

        lcc = LicenseConfigController(self.sa_session, scope=tbs_member.tbs)
        period = lcc.get('pvc.tbs.declaration.period')

        try:
            assert year is not None
            if month == '':
                month = None
            if quarter == '':
                quarter = None
            if period == PvcTbsLicense.DECLARATION_PERIOD_YEARLY:
                assert month is None and quarter is None
            elif period == PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY:
                assert month is None and quarter is not None
            elif period == PvcTbsLicense.DECLARATION_PERIOD_MONTHLY:
                assert quarter is None and month is not None
            else:
                raise PublicUserException(_('Incorrect period provided for TBS member'))
        except AssertionError:
            raise PublicUserException(_('Incorrect period provided for TBS member'))

        # TODO: can only add for current year and previous periods OR previous year if period is YEARLY
        if producer is None:
            producer = tbs_member

        order = self.sa_session.query(PvcDeclarationOrder).filter(and_(
            PvcDeclarationOrder.tbs_member==tbs_member,
            PvcDeclarationOrder.year==year,
            PvcDeclarationOrder.quarter==quarter,
            PvcDeclarationOrder.month==month,
            PvcDeclarationOrder.producer==producer,
        )).first()
        if order is not None:
            raise PublicSecurityException(_('Declaration already exists for provided TBS Member and period.'))

        af = self._get_price(tbs_member)

        orders = {}

        lines = []
        for module in tbs_member.modules:
            if module.is_active != PvcTbsMemberPVModule.IS_ACTIVE: # we only add active modules
                continue

            s = module.subrogation
            if s is not None and module.commissioner_approval_status != PvcTbsMemberPVModule.COMMISSIONER_APPROVAL_STATUS_APPROVED:
                continue

            if not orders.has_key(s):
                orders[s] = {
                    'merchant': tbs_member.tbs,
                    'scope': tbs_member.member.default_group if s is None else module.subrogation.commissioner.member.default_group,
                    'tbs_member': tbs_member if s is None else module.subrogation.commissioner,
                    'year': year,
                    'month': month,
                    'quarter': quarter,
                    'producer': tbs_member,
                    'lines': []
                }

            orders[s]['lines'].append({
                'polymorphic_type': PvcDeclarationOrderLine,
                'description': '%s - %s' % (module.module.supplier.name, module.module.type.name),
                'quantity': 0,
                'price': af,
                'tbs_member_module': module,
            })

        rv = []
        for s, order in orders.iteritems():
            if len(order['lines']) == 0:
                self.logger.warn('add_open_declaration: NO ACTIVE PV MODULES for tbs_member: %s, year=%s, month=%s, quarter=%s, subrogation=%s' % (tbs_member, year, month, quarter, s))
                continue

            self.logger.info('add_open_declaration: DECLARATION ADDED for tbs_member: %s, year=%s, month=%s, quarter=%s, subrogation=%s' % (tbs_member, year, month, quarter, s))
            rv.append(self._save_or_update_order(None, **order))

        if len(rv) == 0:
            return None

        if len(rv) == 1:
            return rv[0]

        return rv

    def _get_line_unit(self, quantity, order, line_polymorphic_type, **kwargs):

        if self.merchant_uses_weight is None:
            self.merchant_uses_weight = {}

        if not self.merchant_uses_weight.has_key(order.merchant.id):
            lcc = LicenseConfigController(self.sa_session, scope=order.merchant)
            self.merchant_uses_weight[order.merchant.id] = lcc.get('pvc.tbs.declaration.use.weight.for.amount.calculations')

        if self.merchant_uses_weight[order.merchant.id] and kwargs.has_key('tbs_member_module'):
            # if based on weight, the weight should be multiplied by the number of units
            return kwargs['tbs_member_module'].module.weight*quantity

        return quantity

    def get_next_closing_date(self, tbs):

        today = virtual_date.today()

        lcc = LicenseConfigController(self.sa_session, scope=tbs)
        period = lcc.get('pvc.tbs.declaration.period')
        closing_day = lcc.get('pvc.tbs.declaration.closing.day')
        if closing_day is None:
            raise Exception('Closing day is None')

        closing_date = None

        if period == PvcTbsLicense.DECLARATION_PERIOD_YEARLY:
            closing_date = datetime.date(today.year, 1, 1) + datetime.timedelta(days=closing_day)

            if closing_date < today:
                closing_date = closing_date.replace(year=today.year+1)

        elif period == PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY:

            # get current quarter
            q = range(4)
            q.reverse()
            for i in q:
                x = datetime.date(today.year, (i*3)+1, 1)
                if x <= today:
                    break
            closing_date = datetime.date(x.year, x.month, 1) + datetime.timedelta(days=closing_day)

            if closing_date < today:
                closing_date = closing_date + relativedelta(months=3)

        elif period == PvcTbsLicense.DECLARATION_PERIOD_MONTHLY:
            closing_date = datetime.date(today.year, today.month, 1) + datetime.timedelta(days=closing_day)
            if closing_date < today:
                closing_date = closing_date + relativedelta(months=1)

        assert closing_date != None
        return closing_date

    def _close_declarations(self, open_orders):
        closed = []
        for order in open_orders:
            try:
                self.submit_order(order)
                closed.append(order)
                self.logger.info('_close_declarations: closing declaration %s' % order)
            except PvcPVModuleNotApprovedException as e:
                self.logger.info(u"_close_declarations: not closing declaration %s, unapproved module: %s" % (order, unicode(e)))
                nc = PvcNotificationController(self.sa_session)
                note = nc.set_notification(_('Unapproved PV Module blocked Declaration from closing.'), unicode(e),
                                           instance_class=PvcDeclarationNotification,
                                           callback={
                                               'js_module': 'pvc_pv_module_manager'
                                           })
                groups = self.get_groups_in_domain_with_permissions('pvc_approve_pv_modules', domain_id=order.merchant.domain.id, anonymous=True)
                if len(groups) == 0:
                    self.logger.info(u"_close_declarations: unable to send system notification for order %s" % order)
                    continue
                nc.create_system_notification(note, groups, sender=order.tbs_member.tbs.default_group, for_group=True)
        return closed

    @transaction_decorator
    def close_declarations(self, tbs):

        today = virtual_date.today()

        if (today != self.get_next_closing_date(tbs)):
            self.logger.debug('today is not closing day for tbs %s' % tbs)
            return None

        _where = self._instance_class.status.in_([self._instance_class.STATUS_NEW, self._instance_class.STATUS_PENDING])
        _where = and_(_where, self._instance_class.type!=self._instance_class.TYPE_CREDIT)
        _where = and_(_where, self._instance_class.merchant==tbs)

        open_orders = self.sa_session.query(self._instance_class).filter(_where).all()
        closed = self._close_declarations(open_orders)
        return closed

    @transaction_decorator
    def close_declarations_created_before_specified_date(self, tbs, before):
        self.assert_permissions('pvc_approve_declaration_resubmit', 'pvc_tbs_license')
        today = virtual_date.today()
        max_dt = datetime.date(today.year, today.month, today.day)
        if before.date() >= max_dt:
            raise PublicUserException('Invalid period for closing')

        _where = self._instance_class.status.in_([self._instance_class.STATUS_NEW, self._instance_class.STATUS_PENDING])
        _where = and_(_where, self._instance_class.type!=self._instance_class.TYPE_CREDIT)
        _where = and_(_where, self._instance_class.merchant==tbs)
        _where = and_(_where, self._instance_class.created<before)

        self.logger.info('manual trigger for closing all declarations for TBS %s before %s' % (tbs, before))

        open_orders = self.sa_session.query(self._instance_class).filter(_where).all()
        closed = self._close_declarations(open_orders)
        return closed

    @transaction_decorator
    def update_with_new_fee(self, service_level, new_fee, old_fee):
        # complexity:
        #  * There is no direct relationship between service levels and declarations
        #  * We will only update Declarations for Producers who have the provided Service Level set as their current SL.

        tbs = service_level.tbs
        tbs_members = self.sa_session.query(PvcTbsMember).filter(and_(PvcTbsMember.tbs==tbs, PvcTbsMember.is_producer==True)).all()
        _in = []
        for tm in tbs_members:
            if tm.active_service_level.service_level==service_level:
                _in.append(tm.id)

        # get all declarations for this TBS where the TBS Members have the selected service level
        _where = self._instance_class.status==self._instance_class.STATUS_NEW
        _where = and_(_where, self._instance_class.type!=self._instance_class.TYPE_CREDIT)
        _where = and_(_where, self._instance_class.merchant==tbs)
        _where = and_(_where, self._instance_class.tbs_member_id.in_(_in))

        orders = self.sa_session.query(self._instance_class).filter(_where).all()
        affected = []
        for order in orders:
            for line in order.lines:
                if line.price_per_unit != old_fee:
                    raise PublicUserException(_('Incorrect original fee found for order #%(order)d on line %(line)d' % {'order': order.id, 'line': line.id}))
                if line.quantity != 0:
                    raise PublicUserException(_('Found quantities on order #%(order)d on line %(line)d with status new' % {'order': order.id, 'line': line.id}))
                line.price_per_unit = new_fee
                self.sa_session.add(line)
                self.logger.info('altering price per unit for order #%d on line %s from fee %s to %s' % (order.id, line.id, str(old_fee), str(new_fee)))
            affected.append(order)
        return affected

    def trigger_new_declarations(self, tbs):
        """
        triggers new declarations for the given tbs by checking the declaration period
        """

        today = virtual_date.today()
        if today.day != 1: # we trigger only on the first
            return None

        lcc = LicenseConfigController(self.sa_session, scope=tbs)
        period = lcc.get('pvc.tbs.declaration.period')

        year = None
        month = None
        quarter = None

        if period == PvcTbsLicense.DECLARATION_PERIOD_YEARLY:
            if not (today.month == 1 and today.day == 1):
                return None
            year = today.year-1

        elif period == PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY:

            if not (today.month in (1, 4, 7, 10) and today.day == 1):
                return None

            # get previous quarter
            q = None
            r = range(4)
            r.reverse()
            for i in r:
                x = datetime.date(today.year, (i*3)+1, 1)
                if x < today:
                    q = x
                    year = today.year
                    quarter = i+1
                    break
            if q is None:
                q = datetime.date(today.year-1, 10, 1)
                year = today.year-1
                quarter = 4

        elif period == PvcTbsLicense.DECLARATION_PERIOD_MONTHLY:
            month = today.month -1
            year = today.year

            if month == 0:
                month = 12
                year = today.year -1

        else:
            raise SecurityException('Incorrect period set for TBS: %s' % tbs)

        tbs_members = self.sa_session.query(PvcTbsMember)\
            .filter(PvcTbsMember.tbs == tbs,
                    PvcTbsMember.is_producer == True,
                    PvcTbsMember.status.in_( (PvcTbsMember.STATUS_APPROVED, PvcTbsMember.STATUS_RESIGNING, ) )).all()

        orders = []
        for tbs_member in tbs_members:

            try:
                order = self.add_open_declaration(tbs_member, year, month, quarter)
                if order is not None:
                    orders.append(order)

            # This is trigger from the scheduler and we do not want public security exceptions to block the process
            # catch public user exceptions for 'Cannot add declarations for a Member without an active Service Level
            except (PublicUserException, PublicSecurityException), e:
                self.logger.warn('add_open_declaration: %s' % e.message)

        return orders

    @transaction_decorator
    def submit_order(self, _id):
        """
        extends super.submit_order() and sets the display_date
        ticket http://trac.office.basesoft.com/iqportal/ticket/3974
        """
        order = super(PvcDeclarationOrderController, self).submit_order(_id)
        order.display_date = now()

        if order.is_compliant is None:
            compliant = PvcDeclarationOrder.IS_NON_COMPLIANT
            if self.get_account() is not None or order.total_amount_excluding_vat > 0:
                compliant = PvcDeclarationOrder.IS_COMPLIANT

            order.is_compliant = compliant
        self.sa_session.add(order)
        order = self.add_minimum_contribution_fee(order)
        return order

    @transaction_decorator
    def add_minimum_contribution_fee(self, order):
        #get original order and if order is compliant we don't need to recalculate the min CF
        if self._get_original_order(order).is_compliant:
            return order

        lcc = LicenseConfigController(self.sa_session, scope=order.merchant)
        minimum_fee = lcc.get('pvc.tbs.minimum.contribution.fee')
        if minimum_fee == 0:
            return order
        #if declared amount is bigger than minimum Contribution fee forfeit fee is canceled
        #if declared amount is smaller a new line is created
        #with the difference between the min Contribution fee and declared amount
        if order.total_amount_excluding_vat > minimum_fee:
            return order
        minimum_fee = minimum_fee - order.total_amount_excluding_vat

        country = order.merchant.company.addresses[0].address.country
        client_country = order.client.company.addresses[0].address.country

        apply_vat = client_country == country
        self._save_or_update_order_line(order, len(order.lines), order.merchant, 'Minimum contribution fee', quantity=1, price=minimum_fee, polymorphic_type=PvcDeclarationOrderLineMinimumCF, apply_vat=apply_vat)

        order.total_amount_excluding_vat = sum([l.line_amount_excluding_vat for l in order.lines])
        order.total_amount_including_vat = sum([l.line_amount_including_vat for l in order.lines])
        order.invoiced_amount = self._get_invoiced_amount(order)
        self.sa_session.add(order)
        return order

    def send_declaration_reminder_email(self, order_ids):

        self.assert_permissions('pvc_manage_declaration_mail')

        from basesoft.web.contrib.mailing.schedule import HtmlMailTask
        mail_task = HtmlMailTask()

        for order_id in order_ids:

            declaration = self.get_order(order_id)[0]
            mail = self.get_object_by_priv_in_hierarchy(declaration.client.default_group, PvcDeclarationReminderMail, PvcDeclarationReminderMail.PRIV_USE_FOR_SENDING)
            if mail is None:
                raise PublicUserException(_('No declaration reminder mail is available for sending.'))

            license_ = declaration.client
            person = declaration.client.company.contact_person

            msg, result = mail_task.build_message(person, mail, context={
                'license': license_,
                'declaration': declaration
            }, mail_from=mail.mail_from())
            mail_task.send_mail(person, msg, result, mail.mail_from())


class PvcContributionOrderController(PvcAbstractContributionController):
    _instance_class = PvcContributionOrder

    def set_additional_fields(self):
        self._set_additional_fields({
            self._instance_class: ['tbs_member', 'year', 'month', 'quarter', 'producer', 'is_forecast', 'is_guarantee_forecast']
        }, {
            PvcContributionOrderLine: ['tbs_member_product', 'weight', 'custom_unit', 'is_b2b', 'brand'],
            OrderLineRule: ['tax_class_code', 'gl_account_code'],  # FIXME this should be added through parent!
        })

    def load_state(self, _id, subrogations=False, submit_dry_run=False):

        if isinstance(_id, self._instance_class):
            order = _id
        else:
            order = self.get(_id).pop()

        field = None

        rv = super(PvcContributionOrderController, self).load_state(order, submit_dry_run)
        rv['year'] = order.year
        rv['month'] = order.month
        rv['quarter'] = order.quarter
        rv['amount_calculation_field'] = field
        rv['tbs_member_id'] = order.tbs_member.id
        rv['producer_id'] = order.producer.id
        rv['tbs_id'] = order.merchant.id
        rv['is_forecast'] = order.is_forecast
        rv['is_guarantee_forecast'] = order.is_guarantee_forecast

        # check for B2B/B2C settings
        rv['has_end_user_distinction'] = order.merchant.get_config('pvc.tbs.require.end.user.distinction')
        rv['has_auto_end_user_distinction'] = order.merchant.get_config('pvc.tbs.contribution.automatic.end.user.distinction')
        rv['has_brands'] = order.merchant.get_config('pvc.tbs.contribution.split.by.brand')

        # the RoE TBS can not duplicate brand lines because a Country selection need to be made. They use Add Line instead.
        rv['can_duplicate_brands'] = False if not rv['has_brands'] else order.merchant.country is not None

        rv['subrogations'] = []
        if not order.is_guarantee_forecast and subrogations:
            rv['subrogations'] = [{'value': s.id, 'label': s.commissioner.member.name} for s in self.get_available_subrogations_for_order(_id)]

        rv['can_pay'] = len(self.get(_id, priv=PvcDeclarationOrder.PRIV_PAY)) == 1

        min_amount = order.merchant.get_config('basesoft.web.contrib.order.minimum_invoiceable_amount')
        if order.status == PvcContributionOrder.STATUS_SUBMITTED and order.total_amount_excluding_vat <= min_amount:
            rv['can_pay'] = False

        rv['can_declare'] = len(self.get(_id, priv=PvcDeclarationOrder.PRIV_UPDATE)) == 1
        if order.status == PvcContributionOrder.STATUS_SUBMITTED:
            # if there is no approved resubmit request on the submitted order, it can never be declared.
            if order.resubmit_request != PvcContributionOrder.RESUBMIT_REQUEST_APPROVED:
                rv['can_declare'] = False
        # if this is not a pending or new order, we cannot declare.
        elif order.status not in (PvcContributionOrder.STATUS_PENDING, PvcContributionOrder.STATUS_NEW, ):
            rv['can_declare'] = False

        return rv

    def _load_lines_state(self, lines, for_export=False):

        if for_export:
            return self.__load_lines_state_for_export(lines)
        rv = []
        additional = []
        for line in lines:
            if not isinstance(line, PvcContributionOrderLine):
                additional = super(PvcContributionOrderController, self)._load_lines_state([line])
                continue
            product = line.tbs_member_product.tbs_product

            rv.append({
                'id': line.id,
                'index': line.index,
                'country': line.tbs_member_product.country.name,
                'description': line.description,
                'price_per_unit': line.price_per_unit,
                'quantity': line.quantity,
                'weight': line.weight,
                'custom_unit': line.custom_unit,
                'product_custom_unit': product.custom_unit, # for the client so it can lookup the unit (m2/Wp)
                'vat_rate': line.vat_rate,
                'gl_account': line.gl_account_number,
                'line_amount_excluding_vat': line.line_amount_excluding_vat,
                'line_amount_including_vat': line.line_amount_including_vat,
                'product': '%s - %s' % (product.label, product.description),
                'basis_for_invoicing': product.basis_for_invoicing,
                'is_b2b': line.is_b2b,
                'brand': None if line.brand is None else line.brand.name,
                'tbs_product_id': product.id,
            })

        #rv = sorted(rv, key=lambda k: (k['country'], k['index']))
        return rv + additional

    def _load_state_for_export(self, order):
        rv = super(PvcContributionOrderController, self)._load_state_for_export(order)
        rv['properties'] = [{
            'key': 'year',
            'value_int': order.year
        },{
            'key': 'month',
            'value_int': order.month
        },{
            'key': 'quarter',
            'value_int': order.quarter
        },{
            'key': 'producer_id',
            'value_int': order.producer.member.id
        },{
            'key': 'has_proof_of_financial_guarantee',
            'value_int': 0 if order.producer.guarantee_file is None else 1
        },{
            'key': 'is_forecast',
            'value_int': 1 if order.is_forecast else 0
        }]
        return rv

    def __load_lines_state_for_export(self, lines):
        rv = []

        for line in lines:

            state_line = {
                'index': line.index,
                'polymorphic_type': line.polymorphic_type,
                'description': line.description,
                'price_per_unit': line.price_per_unit,
                'quantity': line.quantity,
                'vat_rate': line.vat_rate,
                'gl_account': line.gl_account_number,
                'line_amount_excluding_vat': line.line_amount_excluding_vat,
                'line_amount_including_vat': line.line_amount_including_vat,
                'properties': []
            }

            if line.polymorphic_type == 'PvcContributionOrderLine':

                state_line['properties'] += [{
                        'key': 'weight',
                        'value_decimal': line.weight,
                    }, {
                        'key': 'basis_for_invoicing',
                        'value_int': line.tbs_member_product.tbs_product.basis_for_invoicing,
                    },{
                        'key': 'country_name',
                        'value_str': line.tbs_member_product.country.name,
                    },{
                        'key': 'country_iso_3166_a2',
                        'value_str': line.tbs_member_product.country.iso_3166_a2,
                    },{
                        'key': 'country_iso_3166_a3',
                        'value_str': line.tbs_member_product.country.iso_3166_a3,
                    },{
                        'key': 'product_label',
                        'value_str': line.tbs_member_product.tbs_product.label,
                    },{
                        'key': 'product_description',
                        'value_str': line.tbs_member_product.tbs_product.description,
                    }]

                if line.tbs_member_product.tbs_product.custom_unit is not None:
                    state_line['properties'].append({
                        'key': 'custom_unit_label',
                        'value_str': unicode(PvcTbsProduct.CUSTOM_UNIT_LABEL_MAP[line.tbs_member_product.tbs_product.custom_unit]),
                    })
                    state_line['properties'].append({
                        'key': 'custom_unit_unit',
                        'value_str': unicode(PvcTbsProduct.CUSTOM_UNIT_UNIT_MAP[line.tbs_member_product.tbs_product.custom_unit]),
                    })
                    state_line['properties'].append({
                        'key': 'custom_unit_value',
                        'value_decimal': line.custom_unit,
                    })

                if line.is_b2b is not None:
                    state_line['properties'].append({
                        'key': 'dimension',
                        'value_str': 'b2b' if line.is_b2b else 'b2c'
                    })

                if line.brand is not None:
                    state_line['properties'].append({
                        'key': 'brand',
                        'value_str': line.brand.name
                    })


            rv.append(state_line)

        return rv

    def _get_line_unit(self, quantity, order, line_polymorphic_type, **kwargs):

        # check for non-contribution lines (from rules), they don't need custom units
        if line_polymorphic_type != PvcContributionOrderLine:
            return super(PvcContributionOrderController, self)._get_line_unit(quantity, order, line_polymorphic_type, **kwargs)

        if kwargs.has_key('line'):
            l = kwargs['line']
            tbs_product = l.tbs_member_product.tbs_product
            weight = l.weight
            custom_unit = l.custom_unit
        else:
            tbs_product = kwargs['tbs_member_product'].tbs_product
            weight = kwargs.get('weight')
            custom_unit = kwargs.get('custom_unit')

        if tbs_product.basis_for_invoicing == PvcTbsProduct.INVOICE_BASED_ON_WEIGHT:
            assert weight is not None
            return weight

        if tbs_product.basis_for_invoicing == PvcTbsProduct.INVOICE_BASED_ON_CUSTOM_UNIT:
            assert custom_unit is not None
            return custom_unit

        return int(quantity)  # always force quantity to be an int

    def get_available_tbs_member_products(self, tbs_member_id):
        if isinstance(tbs_member_id, PvcTbsMember):
            tbs_member = tbs_member_id
        else:
            tbs_member = self._get_tbs_member_licenses(tbs_member_id, permissions=['pvc_declare_declarations']).pop()

        return self.sa_session.query(PvcTbsProduct).filter(and_(
            PvcTbsProduct.status == PvcTbsProduct.STATUS_USE_IN_DECLARATION,
            PvcTbsProduct.tbs == tbs_member.tbs
        )).all()

    def get_active_tbs_member_products(self, tbs_member_id, subrogation=None):
        if isinstance(tbs_member_id, PvcTbsMember):
            tbs_member = tbs_member_id
        else:
            tbs_member = self._get_tbs_member_licenses(tbs_member_id, permissions=['pvc_declare_declarations']).pop()

        return self.sa_session.query(PvcTbsMemberProduct).filter(and_(
            PvcTbsMemberProduct.is_active == True,
            PvcTbsMemberProduct.tbs_member == tbs_member,
            PvcTbsMemberProduct.subrogation == subrogation
        )).all()

    def get_available_tbs_products(self, _id, additional_exclude=None):
        order = self.get(_id)[0]

        has_end_user_distinction = order.merchant.get_config('pvc.tbs.require.end.user.distinction')
        has_brands = order.merchant.get_config('pvc.tbs.contribution.split.by.brand')
        has_guarantee_forecast = order.merchant.get_config('pvc.tbs.guarantee.forecast')

        if not additional_exclude:
            additional_exclude = []

        exclude = []
        if not has_end_user_distinction and not has_brands:
            exclude = [l.tbs_member_product.tbs_product.id for l in order.lines if isinstance(l, PvcContributionOrderLine)]+additional_exclude
        elif not has_brands and has_end_user_distinction:  # exclude all the lines that already have both b2b and b2c
            exclude = []  # FIXME: #734

        _filter = and_(
            PvcTbsProduct.status == PvcTbsProduct.STATUS_USE_IN_DECLARATION,
            PvcTbsProduct.tbs == order.merchant
        )

        if len(exclude) > 0:
            _filter = and_(_filter, not_(PvcTbsProduct.id.in_(exclude)))

        if has_guarantee_forecast and not order.is_guarantee_forecast:
            types = []
            sl = order.producer.get_service_level_for_date(order.get_period())
            if sl.service_level is not None:
                types.append(sl.service_level.product_type)
            if sl.pv_service_level is not None:
                types.append(sl.pv_service_level.product_type)
            if len(types) == 0:
                raise PublicUserException(_('Unable to find any active service levels'))
            _filter = and_(_filter, PvcTbsProduct.product_type.in_(types))

        return self.sa_session.query(PvcTbsProduct).filter(_filter).all()

    def get_available_subrogations_for_order(self, _id):
        """
        This returns a list with Subrogations which do not yet have a Declaration for the same period as the supplied order.
        """
        orders = self.get(_id, priv=self._instance_class.PRIV_UPDATE)
        if len(orders) == 0: # this happens when the load state requester is the COMMISSIONER in case of subrogation
            return []
        order = orders.pop()

        if order.producer != order.tbs_member:
            return []

        if not self.has_active_subrogations(order.tbs_member):
            return []

        rv = []
        for subrogation in order.tbs_member.subrogations:
            if subrogation.status != PvcSubrogation.STATUS_ACTIVE:
                continue

            if not self._is_undeclared_period(subrogation.commissioner, order.year, order.quarter, order.month, order.producer):
                continue
            rv.append(subrogation)

        return rv

    @transaction_decorator
    def save_default_privilege_object_group(self, order, target_group):
        """
        Override this method in order to assign privs based on Subrogation.
        """
        # normal behavior, no subrogation
        if order.tbs_member == order.producer:
            return super(PvcContributionOrderController, self).save_default_privilege_object_group(order, target_group=target_group)

        # give producer the right privs.
        super(PvcContributionOrderController, self).save_default_privilege_object_group(order, target_group=order.producer.member.default_group, privileges=order.producer_privileges)

        # give the commissioner the right privs.
        self._grant_or_revoke(order, privileges=order.tbs_member_privileges, grant=True, group=True, target_group=order.tbs_member.member.default_group, check_own_privs=False)

    def get_new_lines(self, _id, products, is_b2b=None, countries=None):
        self.assert_permissions('pvc_declare_declarations')
        order = self.get(_id, priv=self._instance_class.PRIV_UPDATE).pop()

        rv = []

        if countries is None or len(countries) == 0:
            countries = [order.producer.tbs.country]
        else:
            countries = [c for c in self.sa_session.query(Country).filter(Country.id.in_(countries)).all()]

        if order.producer.tbs.country is None and (countries == [None] or len(countries) == 0):
            raise PublicUserException(_('Select at least one country.'))

        _map = {}
        for p in self.get_available_tbs_products(_id):
            _map[p.id] = p

        _split = []
        if not order.merchant.get_config('pvc.tbs.require.end.user.distinction'):
            if order.merchant.get_config('pvc.tbs.contribution.automatic.end.user.distinction'):
                _split.append(False)
            else:
                _split.append(None)
        else:
            assert is_b2b is not None
            if is_b2b & PvcContributionOrderLine.END_USER_IS_B2B == PvcContributionOrderLine.END_USER_IS_B2B:
                _split.append(True)
            if is_b2b & PvcContributionOrderLine.END_USER_IS_B2C == PvcContributionOrderLine.END_USER_IS_B2C:
                _split.append(False)

        period = order.get_period()
        period = '%s-%02d-01' % (period.year, period.month)

        if not order.is_guarantee_forecast:
            msl = order.producer.get_service_level_for_date(order.get_period())
            service_level_id = None if msl.service_level is None else msl.service_level.id
            pv_service_level_id = service_level_id if msl.pv_service_level is None else msl.pv_service_level.id

        # if we are dealing with a TBS which requires by brand declaration, this means we cannot check here for dupes.
        assert_dupes = False

        if not order.merchant.get_config('pvc.tbs.contribution.split.by.brand'):
            assert_dupes = True
            _dupes = []
            for line in order.lines:
                if not isinstance(line, PvcContributionOrderLine):
                    continue
                _dupes.append((line.tbs_member_product.tbs_product.id, line.is_b2b, ))

        apply_vat = self._apply_vat_rate(order)

        for tbs_product_id in products:
            assert _map.has_key(int(tbs_product_id))

            product = _map[int(tbs_product_id)]

            for country in countries:
                for b2b in _split:
                    if assert_dupes and (product.id, b2b, ) in _dupes:
                        continue
                    dimension = PvcTbsProductFee.DIMENSION_NONE
                    if b2b is not None:
                        dimension = PvcTbsProductFee.DIMENSION_IS_B2B if b2b else PvcTbsProductFee.DIMENSION_IS_B2C

                    if order.is_guarantee_forecast:
                        price = 0
                    else:
                        product_type = product.product_type
                        if product_type == PvcTbsProduct.TYPE_PV_RELATED:
                            price = product.get_fee(period, pv_service_level_id, dimension)
                        elif product_type == PvcTbsProduct.TYPE_OTHER:
                            price = product.get_fee(period, service_level_id, dimension)
                        else:
                            raise Exception('invalid product_type: %s' % product_type)

                    if price is None:
                        raise PublicUserException(_('There is no price set for "%(product)s" for period %(period)s.') % {
                            'product': product.label,
                            'period': period
                        })

                    tax_rate, vat_fee = self._get_vat_rate(apply_vat, PvcContributionOrderLine.tax_class_code, order.merchant, price)
                    assert tax_rate is not None
                    assert vat_fee is not None

                    rv.append({
                        'tbs_product_id': product.id,
                        'description': (product.label, product.description),
                        'price_per_unit': price,
                        'country': country.name,
                        'country_id': country.id,
                        'quantity': 0,
                        'weight': 0,
                        'custom_unit': None if product.custom_unit is None else 0,
                        'product_custom_unit': product.custom_unit, # for the client so it can lookup the unit (m2/Wp)
                        'gl_account': None,
                        'line_amount_excluding_vat': 0,
                        'line_amount_including_vat': 0,
                        'product': '%s - %s' % (product.label, product.description),
                        'basis_for_invoicing': product.basis_for_invoicing,
                        'vat_rate': tax_rate,
                        'is_b2b': b2b
                    })

        return rv

    @transaction_decorator
    def remove_line(self, _id, line_id, disable_product=None):
        if isinstance(_id, self._instance_class):
            order = _id
        else:
            orders = self.get(_id, priv=self._instance_class.PRIV_UPDATE)
            if len(orders) < 1:
                raise PublicSecurityException(_('Cannot update Order #%s' % _id))
            order = orders.pop()

        lines = [l for l in order.lines if l.id == line_id]
        if len(lines) == 0:
            raise PublicUserException(_('Cannot find line within order.'))

        line = lines[0]
        # check if the status is new or pending.
        self._assert_valid_line_removal(order, line)

        if disable_product is not None:
            line.tbs_member_product.is_active = False
            self.sa_session.add(line.tbs_member_product)

        return super(PvcContributionOrderController, self).remove_line(order, line_id)

    @transaction_decorator
    def resubmit_contribution(self, order, quantities, additional_lines=None, subrogate=None):
        assert isinstance(order, self._instance_class)

        self.logger.info('PvcContributionOrderController: resubmit_contribution() order=%s' % (order.id))

        if not self.has_different_quantities(order, quantities) and additional_lines is None and subrogate is None:
            self.logger.info(' - order is not changed')
            # make sure the order is set to compliant
            order.is_compliant = PvcDeclarationOrder.IS_COMPLIANT
            self.sa_session.add(order)
            return order

        order.resubmit_request = None
        self.sa_session.add(order)

        credit = self._credit_order(order)
        credit.is_forecast = False
        exclude_lines_with_property_value = None

        # if B2B/B2C is split automatically, we should not duplicate the lines with is_b2b=True
        if order.merchant.get_config('pvc.tbs.contribution.automatic.end.user.distinction'):
            exclude_lines_with_property_value = {'is_b2b': True}

        order = self._create_child_order(credit, include_line_types=[PvcContributionOrderLine], exclude_lines_with_property_value=exclude_lines_with_property_value)

        quantities = self._modify_indexes_of_supplied_values(order, quantities)

        return self.update_contribution(order.id, quantities, additional_lines, subrogate=subrogate)

    def _modify_indexes_of_supplied_values(self, order, quantities):

        original_lines = order.parent_order.lines
        values = {}
        for oline in original_lines:
            if not isinstance(oline, PvcContributionOrderLine):
                continue
            if not quantities.has_key('line_%d_quantity' % oline.index):
                continue
            for nline in order.lines:
                if oline.tbs_member_product != nline.tbs_member_product:
                    continue
                if oline.brand != nline.brand:
                    continue
                if oline.is_b2b != nline.is_b2b:
                    continue
                values['line_%d_quantity' % nline.index] = quantities['line_%d_quantity' % oline.index]
                values['line_%d_weight' % nline.index] = quantities['line_%d_weight' % oline.index]
                if quantities.has_key('line_%d_custom_unit' % oline.index):
                    values['line_%d_custom_unit' % nline.index] = quantities['line_%d_custom_unit' % oline.index]
                if quantities.has_key('line_%d_brand' % oline.index):
                    values['line_%d_brand' % nline.index] = quantities['line_%d_brand' % oline.index]
                if quantities.has_key('line_%d_is_b2b' % oline.index):
                    values['line_%d_is_b2b' % nline.index] = quantities['line_%d_is_b2b' % oline.index]
                break

        return values

    def get_child_order_price_for_line(self, order, line, is_credit):
        """
        resubmits should get the new/update fee for that period in case it has changed
        """

        # credits should always use the original price per unit.
        if is_credit:
            return line.price_per_unit

        # guarantee forecasts do not have a price per unit.
        if order.is_guarantee_forecast:
            return Decimal('0')

        period = order.get_period()
        period = '%s-%02d-01' % (period.year, period.month)

        dimension = PvcTbsProductFee.DIMENSION_NONE
        if line.is_b2b is not None:
            dimension = PvcTbsProductFee.DIMENSION_IS_B2B if line.is_b2b else PvcTbsProductFee.DIMENSION_IS_B2C

        msl = order.producer.get_service_level_for_date(order.get_period())
        service_level_id = None if msl.service_level is None else msl.service_level.id
        pv_service_level_id = service_level_id if msl.pv_service_level is None else msl.pv_service_level.id

        if line.tbs_member_product.tbs_product.product_type == PvcTbsProduct.TYPE_PV_RELATED:
            service_level_id = pv_service_level_id

        if service_level_id is None:
            raise PublicUserException(_('No active service level for period %s' % period))

        price = line.tbs_member_product.tbs_product.get_fee(period, service_level_id, dimension)

        if price is None:

            sl = self.sa_session.query(PvcTbsServiceLevel).filter(PvcTbsServiceLevel.id == service_level_id).one()

            raise PublicUserException(_('There is no price set for "%(product)s" for period %(period)s, service level "%(service_level)s" and dimension %(dimension)s.') % {
                'product': line.tbs_member_product.tbs_product.label,
                'period': period,
                'service_level': sl.name,
                'dimension': 'None' if dimension is PvcTbsProductFee.DIMENSION_NONE else
                    'B2B' if dimension is PvcTbsProductFee.DIMENSION_IS_B2B else 'B2C'
            })

        return price


    def has_different_quantities(self, order, values):
        """
        Overrides super and also checks for weight/etc
        """
        if len(order.lines) != len(values.values()): # different sizes means different quantities.
            return True

        # if any of the user fields (quantity, weight or custom_unit) changes, make a credit regardless which field
        # is used for calculation
        for line in order.lines:
            for field in line.CREDIT_FIELDS:
                if hasattr(line, field):
                    v = int(values['line_%s_%s' % (line.index, field)])
                    if getattr(line, field) != v:
                        return True

        return False

    def get_outstanding_forecast_requirement(self):
        """
        This validates whether or not the producer has an order (forecast) at any of the TBSs.
        :return:
        """
        rv = []

        tbs_members = self._get_tbs_member_licenses(permissions=['pvc_declare_declarations'])
        for tm in tbs_members:
            if not tm.is_producer:
                continue
            if tm.status not in (PvcTbsMember.STATUS_APPROVED, PvcTbsMember.STATUS_RESIGNING):
                continue
            if not tm.tbs.get_config('pvc.tbs.advance.invoicing'):
                continue

            country = None if tm.tbs.country is None else tm.tbs.country.name

            if self.sa_session.query(PvcContributionOrder).filter(PvcContributionOrder.tbs_member == tm).first() is None:
                rv.append((tm.id, None, tm.tbs.get_config('pvc.tbs.require.end.user.distinction'), tm.tbs.get_config('pvc.tbs.contribution.split.by.brand'), country))
                continue
            order = self._get_outstanding_forecast_order(tm)
            if order is not None:
                rv.append((tm.id, order.id, None, None, country))

        return rv

    def _get_outstanding_forecast_order(self, tm):
        return self.sa_session.query(PvcContributionOrder).filter(and_(
            PvcContributionOrder.tbs_member == tm,
            PvcContributionOrder.is_forecast == True,
            PvcContributionOrder.status.in_([PvcContributionOrder.STATUS_NEW, PvcContributionOrder.STATUS_PENDING])
        )).first()

    def get_guarantee_forecast_next_year(self):
        tbs_members = self._get_tbs_member_licenses(permissions=['pvc_declare_declarations'])

        rv = []

        for tm in tbs_members:
            if tm.status not in (PvcTbsMember.STATUS_APPROVED, ):
                continue

            if not tm.is_producer:
                continue

            if not tm.tbs.get_config('pvc.tbs.guarantee.forecast'):
                continue

            # check if there is one for the current year else make one
            if today().year > 2015:
                gf_order = self._get_guarantee_forecast_order(tm, today().year)

                if not gf_order:
                    rv.append((tm, today().year))

            # skip months before trigger month
            if today().month < tm.tbs.get_config('pvc.tbs.guarantee.forecast.trigger.month'):
                continue

            # try to get existing forecast, and return result if not yet found
            gf_order = self._get_guarantee_forecast_order(tm, today().year + 1)

            if not gf_order:
                rv.append((tm, today().year+1))

        return rv

    def _get_guarantee_forecast_order(self, tm, year):
        assert isinstance(tm, PvcTbsMember)

        return self.sa_session.query(PvcContributionOrder).filter(and_(
            PvcContributionOrder.tbs_member == tm,
            PvcContributionOrder.is_guarantee_forecast == True,
            PvcContributionOrder.year == year
        )).first()

    def get_outstanding_guarantee_forecast(self, tm):
        assert isinstance(tm, PvcTbsMember)

        if not tm.tbs.get_config('pvc.tbs.guarantee.forecast'):
            return False

        if not tm.is_producer:
            return False

        year = today().year

        if self._get_guarantee_forecast_order(tm, year) is None:
            return year


    def get_undeclared_periods(self, tbs_member, subrogation_id=None):
        """
        We get the undeclared periods based on the TBS Member periodicity.

        Rules per periodicity are listed inline.

        :param tbs_member:
        :return:
        """

        period = tbs_member.declaration_period

        tbs_member = tbs_member
        producer = tbs_member
        if subrogation_id is not None:
            subrogation = self.sa_session.query(PvcSubrogation).filter(PvcSubrogation.id == subrogation_id).first()
            if subrogation is None:
                raise PublicUserException(_('Subrogation not found.'))
            if subrogation.subrogator != tbs_member:
                raise PublicUserException(_('Incorrect subrogation supplied'))
            tbs_member = subrogation.commissioner

        declarations = self.sa_session.query(PvcContributionOrder).filter(and_(
            PvcContributionOrder.tbs_member == tbs_member,
            PvcContributionOrder.producer == producer,
            PvcContributionOrder.is_guarantee_forecast == False
        )).all()

        # yearly periodicity will show any year since the obligation start date without a declaration (until current year)
        if period == PvcTbsLicense.DECLARATION_PERIOD_YEARLY:
            min_year = tbs_member.obligation_start_date.year
            possible = self._get_possible_periods_for_yearly_declarations(min_year)
            if possible is None:  # no years available
                return {}

            for d in declarations:
                if not possible.has_key(d.year):
                    continue
                # remove this year from possibilities.
                del possible[d.year]
            return possible

        # quarterly declarations will list any quarters which do not have declarations yet.
        if period == PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY:
            min_year = tbs_member.obligation_start_date.year
            possible = self._get_possible_periods_for_quarterly_declarations(min_year)
            if possible is None:
                return {}  # no periods available

            for d in declarations:
                # if the following check resolves as True, it means periodicity has changed from yearly to quarterly.
                if d.quarter is None and d.month is None:
                    # skip any years for which yearly declarations were made.
                    del possible[d.year]
                    continue
                if not possible.has_key(d.year):
                    continue
                # if the following check resolves as True, it means periodicity has changed from monthly to quarterly.
                if d.month is not None:
                    # remove quarters corresponding with month.
                    q = ((d.month-1) / 3)+1
                    if q not in possible[d.year]['quarters']:
                        continue
                    possible[d.year]['quarters'].remove(q)
                    continue

                if d.quarter not in possible[d.year]['quarters']:
                    continue
                possible[d.year]['quarters'] = filter(lambda a: a != d.quarter, possible[d.year]['quarters'])

            rm_y = []
            # remove all the years for which there are no quarters available to declare.
            for year, quarters in possible.iteritems():
                if len(quarters['quarters']) == 0:
                    rm_y.append(year)
            for year in rm_y:
                del possible[year]
            return possible

        # monthly declarations will list any months which do not have declarations yet.
        if period == PvcTbsLicense.DECLARATION_PERIOD_MONTHLY:

            min_year = tbs_member.obligation_start_date.year
            possible = self._get_possible_periods_for_monthly_declarations(min_year)
            if possible is None:
                return {}  # no periods available

            for d in declarations:
                # the year is not available, no further checks necessary.
                if not possible.has_key(d.year):
                    continue

                # if the following check resolves as True it means periodicity has changed from yearly to monthly.
                if d.month is None and d.quarter is None:
                    # skip the whole year for monthly declarations.
                    del possible[d.year]
                    continue

                # if the following check resolves as True it means periodicity has changed from quarterly to monthly.
                if d.quarter is not None:
                    for m in range(3):
                        m = (d.quarter*3)-m
                        # remove any months for previously declared quarters.
                        if m in possible[d.year]['months']:
                            possible[d.year]['months'].remove(m)

                if d.month not in possible[d.year]['months']:
                    continue
                possible[d.year]['months'] = filter(lambda a: a != d.month, possible[d.year]['months'])

            rm_y = []
            # remove all the years for which there are no months available to declare
            for year, months in possible.iteritems():
                if len(months['months']) == 0:
                    rm_y.append(year)
            for year in rm_y:
                del possible[year]
            return possible

        raise SecurityException('Incorrect period set for TBS: %s' % tbs_member.tbs)

    @staticmethod
    def get_first_available_period_for_forecast(tbs_member):
        t = today()
        year = t.year

        # period is per tbs member, it automatically falls back to the TBS period.
        period = tbs_member.declaration_period

        if period == PvcTbsLicense.DECLARATION_PERIOD_YEARLY:
            return year+1, None, None

        if period == PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY:
            import math
            q = int(math.floor((t.month-1)/3)+1)
            return year, None, q

        return year, t.month, None

    def _do_periodicity_change(self, order):
        tm = order.producer
        # There is no upcoming periodicity change, no need to do anything here.
        if tm.upcoming_periodicity_change is None:
            return False

        # we do not change periods on forecast submissions
        if order.is_forecast:
            return False

        # TODO Verify that this order is from THIS period (No resubmits)

        self.logger.info('TBS member %s has an upcoming periodicity change' % tm)

        monthly = PvcTbsLicense.DECLARATION_PERIOD_MONTHLY
        quarterly = PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY
        yearly = PvcTbsLicense.DECLARATION_PERIOD_YEARLY

        cp = tm.declaration_period
        np = tm.upcoming_periodicity_change

        # Monthly to Quarterly
        if cp == monthly and np == quarterly:
            return order.month in (3, 6, 9, 12,)

        # Monthly to Yearly
        if cp == monthly and np == yearly:
            return order.month == 12

        # Quarterly to Yearly
        if cp == quarterly and np == yearly:
            return order.quarter == 4

        # Yearly to Quarterly or Monthly
        if cp == yearly and np in (quarterly, monthly, ):
            return order.month is None and order.quarter is None

        # Quarterly to Monthly
        return order.month is None

    @transaction_decorator
    def verify_periodicity_change(self, tm, new_periodicity, member_level_change=True):
        # the member_level_change flag is used to determine whether the periodicity level change is triggered through
        # the member details or on tbs level.
        cp = tm.declaration_period

        # the TBS Member has an overridden periodicity and we are changing the TBS setting.
        if not member_level_change and tm.declaration_periodicity != PvcTbsLicense.DECLARATION_PERIOD_DEFAULT:
            return tm

        # this means it got triggered from the TBS config
        if not member_level_change:
            # the config was already updated, so we grab what was the previous periodicity
            cp = tm.tbs.get_config('pvc.tbs.previous.declaration.period')

        if member_level_change:
            tm.declaration_periodicity = cp
            self.sa_session.add(tm)

        # we do not update periodicity in case they are the same
        if cp == new_periodicity:
            return tm

        # check for any declarations, no declarations mean new periodicity can always be set
        d = self.sa_session.query(PvcContributionOrder).filter(PvcContributionOrder.producer==tm).first()
        if d is None:
            if member_level_change:
                tm.declaration_periodicity = new_periodicity
            tm.upcoming_periodicity_change = None
            self.sa_session.add(tm)
            return tm

        # advance invoicing will always be triggered at closing day
        if tm.tbs.get_config('pvc.tbs.advance.invoicing'):
            tm.upcoming_periodicity_change = new_periodicity
            self.sa_session.add(tm)
            return tm

        t = today()
        np = new_periodicity

        if cp == PvcTbsLicense.DECLARATION_PERIOD_YEARLY:
            # we switch from yearly to monthly, the month has to be january still.
            if np == PvcTbsLicense.DECLARATION_PERIOD_MONTHLY:
                if t.month != 1:
                    tm.upcoming_periodicity_change = new_periodicity
                    self.sa_session.add(tm)
                    return tm
            # we switch from yearly to monthly, the month cannot be april or further yet.
            elif np == PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY:
                if t.month >= 4:
                    tm.upcoming_periodicity_change = new_periodicity
                    self.sa_session.add(tm)
                    return tm

            y = t.year-1
            q = None
            m = None

        elif cp == PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY:
            # we switch from quarterly to monthly, the month has to be january, april, july, october
            if np == PvcTbsLicense.DECLARATION_PERIOD_MONTHLY:
                if t.month not in (1, 4, 7, 10, ):
                    tm.upcoming_periodicity_change = new_periodicity
                    self.sa_session.add(tm)
                    return tm
            # we switch from quarterly to yearly, the month has to be january.
            elif np == PvcTbsLicense.DECLARATION_PERIOD_YEARLY:
                if t.month != 1:
                    tm.upcoming_periodicity_change = new_periodicity
                    self.sa_session.add(tm)
                    return tm

            y = t.year-1 if t.month == 1 else t.year
            q = 4 if t.month == 1 else (t.month / 3)
            m = None

        elif cp == PvcTbsLicense.DECLARATION_PERIOD_MONTHLY:
            # we switch from monthly to quarterly, the month has to be january, april, july, october
            if np == PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY:
                if t.month not in (1, 4, 7, 10, ):
                    tm.upcoming_periodicity_change = new_periodicity
                    self.sa_session.add(tm)
                    return tm
            elif np == PvcTbsLicense.DECLARATION_PERIOD_YEARLY:
                if t.month != 1:
                    tm.upcoming_periodicity_change = new_periodicity
                    self.sa_session.add(tm)
                    return tm

            y = t.year-1 if t.month == 1 else t.year
            q = None
            m = 12 if t.month == 1 else t.month-1

        d = self.sa_session.query(PvcContributionOrder).filter(and_(
            PvcContributionOrder.producer==tm,
            PvcContributionOrder.parent_order==None,
            PvcContributionOrder.status.in_((PvcContributionOrder.STATUS_NEW, PvcContributionOrder.STATUS_PENDING, )),
            PvcContributionOrder.year==y,
            PvcContributionOrder.month==m,
            PvcContributionOrder.quarter==q
        )).first()
        if d is None:
            if member_level_change:
                tm.declaration_periodicity = new_periodicity
            tm.upcoming_periodicity_change = None
            self.sa_session.add(tm)
            return tm

        tm.upcoming_periodicity_change = new_periodicity
        self.sa_session.add(tm)
        return tm

    @transaction_decorator
    def submit_order(self, _id, dry_run=False):

        self.logger.info('PvcContributionOrderController: submit_order: %s' % _id)

        if isinstance(_id, self._instance_class):
            order = _id
        else:
            order = self.get(_id).pop()

        order.display_date = now()  # FIXME: do this in super?

        if order.is_forecast:
            order.is_compliant = PvcDeclarationOrder.IS_COMPLIANT

        if order.is_compliant is None:
            compliant = PvcDeclarationOrder.IS_NON_COMPLIANT
            if self.get_account() is not None or order.total_amount_excluding_vat > 0:
                order.add_log('Contribution order submitted, either by user or total amount is larger than 0. Setting compliant flag.')
                compliant = PvcDeclarationOrder.IS_COMPLIANT
            order.is_compliant = compliant

        if order.is_guarantee_forecast:

            total_pv = 0
            total_other = 0

            for line in order.lines:

                product_type = line.tbs_member_product.tbs_product.product_type

                if product_type == PvcTbsProduct.TYPE_PV_RELATED:
                    #905: some PV products don't have a custom_unit, then assume 0 (batteries / invertors)
                    total_pv += 0 if line.custom_unit is None else line.custom_unit
                elif product_type == PvcTbsProduct.TYPE_OTHER:
                    assert line.weight >= 0
                    total_other += line.weight
                else:
                    raise Exception('Invalid product type: %s' % product_type)

            service_level_other = None
            service_level_pv = None

            if total_pv > 0:
                service_level_pv = self._get_forecast_guarantee_service_level(order.merchant.id, product_type=PvcTbsProduct.TYPE_PV_RELATED, total=total_pv)
            if total_other > 0:
                service_level_other = self._get_forecast_guarantee_service_level(order.merchant.id, product_type=PvcTbsProduct.TYPE_OTHER, total=total_other)

            if service_level_other is None and service_level_pv is None:
                raise PublicUserException(
                    _('The minimum amount for PV related products is %(total_pv)s, for other products is %(total_other)s. \
                    Please remove the products from the list if you sell 0 and contact PV CYCLE for assistance if necessary.') % {
                        'total_pv': self._get_forecast_guarantee_minimal_value(order.merchant.id, PvcTbsProduct.TYPE_PV_RELATED),
                        'total_other': self._get_forecast_guarantee_minimal_value(order.merchant.id, PvcTbsProduct.TYPE_OTHER),
                    })

            # create a new SL
            member_service_level = PvcTbsMemberServiceLevel()

            # ticket #1004: update existing SL if member is still pending
            if order.producer.status == PvcTbsMember.STATUS_PENDING and order.producer.active_service_level is not None:
                member_service_level = order.producer.active_service_level

            member_service_level.tbs_member = order.producer
            member_service_level.service_level = service_level_other
            member_service_level.pv_service_level = service_level_pv

            # #1014: check if SL hasnt changed, then dont make a new SL
            if order.producer.active_service_level and \
                order.producer.active_service_level.service_level == member_service_level.service_level and \
                    order.producer.active_service_level.pv_service_level == member_service_level.pv_service_level:
                member_service_level = None

            # if the GF order is a resubmit and member isn't pending (#1004), then perform a SL change
            if order.parent_order is not None and order.producer.status != PvcTbsMember.STATUS_PENDING:

                # raise an error if its a resubmit and the current year is larger than the guarantee_forecast.year
                if order.display_date.year < today().year:
                    raise PublicUserException('Cannot resubmit a guarantee forecast from previous years')

                if member_service_level is not None:
                    member_service_level.status = PvcTbsMemberServiceLevel.STATUS_PENDING

                mc = PvcMemberController(self.sa_session, account_id=self.get_account().id, request=self.request)

                try:
                    if order.year > today().year:
                        sl_date = order.get_period()
                    else:
                        # #1002: service level date is taken from the end of the year if its in the past
                        sl_date = datetime.date(order.year, 12, 31)

                    if member_service_level is not None:
                        member_service_level = mc.forecast_guarantee_resubmit_approve_change_service_level(member_service_level, sl_date)

                    # no change, remove any invalid previous member SLs as the SL changed back to the previous created
                    if member_service_level is None and order.year > today().year:
                        sls = self.sa_session.query(PvcTbsMemberServiceLevel).filter(
                            PvcTbsMemberServiceLevel.tbs_member == order.producer,
                            PvcTbsMemberServiceLevel.start_date >= datetime.date(order.year, 1, 1)
                        )
                        for sl in sls:
                            self.sa_session.delete(sl)
                            self.logger.info('submit_order: removed PvcTbsMemberServiceLevel %s' % sl.id)

                    nc = NotificationController(self.sa_session)
                    operators = self.sa_session.query(PvcOperatorLicense).all()
                    recipients = [o.default_group for o in operators]
                    nc.create_system_notification(nc.set_notification(
                        'Service level changed',
                        '"%(producer)s" has changed their service level through a Guarantee Forecast resubmit.' % {
                            'producer': order.producer.member.name
                        },
                    ), recipients)

                except PvcServiceLevelNoChangeException:
                    # forecast_guarantee_resubmit_approve_change_service_level gave a SL not changed exception, unset the new SL
                    member_service_level = None

            elif member_service_level is not None:
                # not a resubmit (or member pending (#1004): activate member_sl directly

                member_service_level.status = PvcTbsMemberServiceLevel.STATUS_APPROVED

                if order.year == today().year:
                    # guarantee forecast of this year, sl start date is now
                    member_service_level.start_date = datetime.date(2000, 1, 1) if order.producer.active_service_level is None else today()
                elif order.year == today().year + 1:
                    # guarantee forecast for next year, starts next year
                    member_service_level.start_date = datetime.date(today().year + 1, 1, 1)

                self.logger.info('submit_order: is_guarantee_forecast: added member service level sl=%s, pv-sl=%s to tbs_member %s, start-date=%s' %
                                 (service_level_other, service_level_pv, order.producer.id, member_service_level.start_date))

            if not dry_run:

                if member_service_level is not None:
                    self.sa_session.add(member_service_level)
                self._disable_all_products_not_in_guarantee_forecast(order)

        # send to super, which processes rules and changes status
        order = super(PvcContributionOrderController, self).submit_order(_id, dry_run)

        if dry_run:
            return order

        self.add_forecast_from_actuals(order)

        return order

    def _get_forecast_guarantee_service_level(self, tbs_id, product_type, total):
        lazy_assert(tbs_id, int)
        lazy_assert(product_type, int)
        assert total > 0

        service_levels = self.sa_session.query(PvcTbsServiceLevel).filter(
            PvcTbsServiceLevel.tbs_id == tbs_id,
            PvcTbsServiceLevel.status == PvcTbsServiceLevel.STATUS_ACTIVE,
            PvcTbsServiceLevel.product_type == product_type
        ).order_by(PvcTbsServiceLevel.guarantee_forecast_min_value.desc())

        selected_sl = None
        for pv_sl in service_levels:
            if total >= pv_sl.guarantee_forecast_min_value:
                selected_sl = pv_sl
                break

        if selected_sl is None:
            raise PublicUserException(_('No service level found'))

        return selected_sl

    def _get_forecast_guarantee_minimal_value(self, tbs_id, product_type):
        lazy_assert(tbs_id, int)
        lazy_assert(product_type, int)

        service_level = self.sa_session.query(PvcTbsServiceLevel).filter(
            PvcTbsServiceLevel.tbs_id == tbs_id,
            PvcTbsServiceLevel.status == PvcTbsServiceLevel.STATUS_ACTIVE,
            PvcTbsServiceLevel.product_type == product_type
        ).order_by(PvcTbsServiceLevel.guarantee_forecast_min_value.asc()).first()

        if service_level is None:
            raise PublicUserException(_('No active TBS service level found'))

        return service_level.guarantee_forecast_min_value

    @transaction_decorator
    def add_forecast_from_actuals(self, original):

        if original.is_forecast:  # we do not add forecasts from forecasts
            return None

        if original.status != self._instance_class.STATUS_SUBMITTED:
            raise PublicSecurityException(_('Trying to create a forecast from an unsubmitted order.'))

        period_change = self._do_periodicity_change(original)
        cp = original.producer.declaration_period
        np = original.producer.upcoming_periodicity_change

        if period_change:
            # the new periodicity change was made on TBS level.
            if original.producer.declaration_periodicity is not None:
                original.producer.declaration_periodicity = np
            original.producer.upcoming_periodicity_change = None
            self.sa_session.add(original.producer)

        if not original.merchant.get_config('pvc.tbs.advance.invoicing'):
            return None

        monthly = PvcTbsLicense.DECLARATION_PERIOD_MONTHLY
        quarterly = PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY
        yearly = PvcTbsLicense.DECLARATION_PERIOD_YEARLY

        divider = 1
        multiplier = 1

        period = original.producer.declaration_period
        # get the next period.
        y = original.year
        q = original.quarter
        m = original.month

        self.logger.info('Making forecast order from actuals %d' % original.id)
        if period_change:
            self.logger.info('Periodicity changed for producer %s from %d to %d' % (original.producer, cp, np))

            if cp == monthly and np == quarterly:
                multiplier = 3  # we multiple the inputs by 3 (3 months per quarter)
                q = ((original.month-1) / 3) + 1
            elif cp == monthly and np == yearly:
                multiplier = 12  # we multiple the inputs by 12 (12 months per year)
            elif cp == quarterly and np == yearly:
                multiplier = 4  # we multiple the inputs by 4 (4 quarters per year)
            elif cp == quarterly and np == monthly:
                divider = 3  # we divide the inputs by 3 (3 months per quarter)
                m = q*3
            elif cp == yearly and np == quarterly:  # only possible at end of yearly cycle.
                divider = 4  # we divide the inputs by 4 (4 quarters per year)
                q = 4
            elif cp == yearly and np == monthly:  # only possible at end of yearly cycle.
                divider = 12  # we divide the inputs by 12 (12 months per year)
                m = 12

        if period == quarterly:
            assert q is not None
            q += 1
            m = None
            if q > 4:
                q = 1
                y += 1
        elif period == monthly:
            assert m is not None
            q = None
            m += 1
            if m > 12:
                m = 1
                y += 1
        else:  # yearly
            m = None
            q = None
            y += 1

        # we do not add new forecasts from resigning members at year change.
        if original.year != y and original.producer.status == PvcTbsMember.STATUS_RESIGNING:
            return None

        def get_date(y, q, m):
            # if yearly, add a year to check against
            if m is None and q is None:
                return datetime.date(y+1, 1, 1)
            # if quarterly
            if m is None:
                # check next quarter
                q += 1
                # we only have 4 quarters
                if q > 4:
                    q = 1
                    y += 1
                m = (q*3)+1
                if m > 12:
                    y += 1
                    m = 1
            else:
                m += 1
                if m > 12:
                    m = 1
                    y += 1
            return datetime.date(y, m, 1)

        # Dont add forecasts in the past, we validate against the start date of the NEXT period.
        if today() >= get_date(y, q, m):
            return None

        # check if there is not yet a contribution order for that period (this can happen if the original is a resubmit)
        if not self._is_undeclared_period(original.tbs_member, y, q, m, original.producer):
            return None

        lines = []
        for line in original.lines:
            if not isinstance(line, PvcContributionOrderLine):
                continue
            lines.append({
                'tbs_member_product': line.tbs_member_product,
                'quantity': (int(line.quantity)*multiplier)/divider,
                'weight': (line.weight*multiplier)/divider,
                'custom_unit': None if line.custom_unit is None else (line.custom_unit*multiplier)/divider,
                'brand': line.brand,
                'is_b2b': line.is_b2b,
            })
        order = self._add_contribution_order(original.tbs_member, lines, year=y, month=m, quarter=q, producer=original.producer, is_forecast=True)
        return self.submit_order(order)

    def _add_common_forecast_order(self, tbs_member, products, is_b2b, y, m, q, is_forecast=False, is_guarantee_forecast=False, brand=None):

        pc = PvcProductController(self.sa_session, account_id=self.get_account().id, request=self.request)

        _split = []
        if is_b2b is None:
            _split.append(None)
        else:
            if is_b2b & PvcContributionOrderLine.END_USER_IS_B2B == PvcContributionOrderLine.END_USER_IS_B2B:
                _split.append(True)
            if is_b2b & PvcContributionOrderLine.END_USER_IS_B2C == PvcContributionOrderLine.END_USER_IS_B2C:
                _split.append(False)

        newlines = []
        for tbs_product_id in products:
            mp = pc.get_or_save_tbs_member_product(tbs_member, tbs_product_id)
            b = None
            if brand is not None:
                b = pc.get_or_save_brand(brand, mp)

            for b2b in _split:
                # store the product as member product in case it isn't yet.
                newlines.append({
                    'tbs_member_product': mp,
                    'quantity': 0,
                    'weight': Decimal('0.0'),
                    'brand': b,
                    'custom_unit': None if mp.tbs_product.custom_unit is None else Decimal('0.0'),
                    'is_b2b': b2b,
                })

        # create a new order.
        return self._add_contribution_order(tbs_member, newlines, y, m, q, is_forecast=is_forecast, is_guarantee_forecast=is_guarantee_forecast)

    @transaction_decorator
    def add_forecast_order(self, tbs_member_id, products, is_b2b=None, brand=None):
        # This is a new Declaration, except it is being used for advance invoicing and requires a resubmit before the period end.

        members = self._get_tbs_member_licenses(tbs_member_id, ['pvc_declare_declarations'])
        if len(members) == 0:
            raise PublicUserException(_('TBS member not available for forecast.'))
        tbs_member = members.pop()

        outstanding = self.get_outstanding_forecast_requirement()

        tm_id = None
        for settings in outstanding:
            if tbs_member.id != settings[0]:
                continue
            tm_id, order, distinction, brands, country = settings
            break

        if tm_id is None:
            raise PublicUserException(_('The provided TBS member is not eligible for a forecast declaration.'))

        if order is not None:
            raise PublicUserException(_('A forecast order already exists for this TBS member.'))

        # get the first upcoming (or current) declaration period.
        period = PvcContributionOrderController.get_first_available_period_for_forecast(tbs_member)

        y, m, q = period

        return self._add_common_forecast_order(tbs_member, products, is_b2b, y, m, q, is_forecast=True, brand=brand)

    def add_guarantee_forecast_order(self, tbs_member_id, year, products, is_b2b=None, brand=None):

        members = self._get_tbs_member_licenses(tbs_member_id, ['pvc_declare_declarations'])
        if len(members) == 0:
            raise PublicUserException(_('TBS member not available for forecast.'))
        tbs_member = members.pop()

        if not tbs_member.tbs.get_config('pvc.tbs.guarantee.forecast'):
            raise Exception('guarantee.forecast not enabled')

        if year == today().year:
            pass
        elif year == today().year +1:
            trigger_month = tbs_member.tbs.get_config('pvc.tbs.guarantee.forecast.trigger.month')
            assert today().month >= trigger_month
        else:
            raise Exception('invalid year: ' + year)

        y = today().year if year is None else year
        m = None
        q = None

        return self._add_common_forecast_order(tbs_member, products, is_b2b, y, m, q, is_guarantee_forecast=True, brand=brand)

    @transaction_decorator
    def _disable_all_products_not_in_guarantee_forecast(self, order):
        if not order.is_guarantee_forecast:
            return
        if order.year != today().year:
            return

        keep = [line.tbs_member_product.id for line in order.lines]
        products = self.sa_session.query(PvcTbsMemberProduct).filter(and_(
            not_(PvcTbsMemberProduct.id.in_(keep)),
            PvcTbsMemberProduct.tbs_member == order.producer,
            PvcTbsMemberProduct.is_active == True
        )).all()
        for mp in products:
            mp.is_active = False
            self.logger.info('Disabling TBS Member product %s - it is not selected in Guarantee Forecast.' % mp)
            self.sa_session.add(mp)

    @transaction_decorator
    def update_contribution(self, _id, values, additional_lines=None, subrogate=None):
        self.assert_permissions('pvc_declare_declarations')

        order = self.get(_id, priv=self._instance_class.PRIV_UPDATE).pop()
        if order.status == self._instance_class.STATUS_CREDITED:
            raise PublicSecurityException(_('You cannot update an already credited order.'))

        if order.status not in [self._instance_class.STATUS_NEW, self._instance_class.STATUS_PENDING]:
            if order.resubmit_request != self._instance_class.RESUBMIT_REQUEST_APPROVED:
                raise PublicUserException(_('You cannot resubmit this declaration.'))
            return self.resubmit_contribution(order, values, additional_lines, subrogate=subrogate)

        if order.type == self._instance_class.TYPE_CREDIT:
            raise PublicSecurityException(_('You cannot update a credit declaration.'))

        if len(values) < len(order.lines) * 2:
            raise PublicUserException(_('Number of lines for Order #%s does not match with supplied quantities.' % order.id))

        if subrogate is not None:
            if order.tbs_member != order.producer:
                raise PublicUserException(_('You cannot subrogate a declaration which is already subrogated.'))

            if order.is_guarantee_forecast:
                raise PublicUserException(_('You cannot subrogate a guarantee forecast order.'))

            return self._subrogate_order(order, values, subrogate, additional_lines)

        subrogation = None
        if order.tbs_member != order.producer:
            subrogation = self.sa_session.query(PvcSubrogation).filter(and_(
                PvcSubrogation.subrogator == order.producer,
                PvcSubrogation.commissioner == order.tbs_member
            )).first()

        pc = PvcProductController(self.sa_session, account_id=self.get_account().id, request=self.request)

        has_end_user_distinction = order.merchant.get_config('pvc.tbs.require.end.user.distinction')
        automatic_end_user_distinction = order.merchant.get_config('pvc.tbs.contribution.automatic.end.user.distinction')
        by_brand = order.merchant.get_config('pvc.tbs.contribution.split.by.brand')
        gl_account = None if not order.is_forecast else PvcContributionOrderLine.gl_forecast_account_code

        products = []
        lines = []

        for line in order.lines:
            l = self._line_from_value(line, values, has_end_user_distinction, automatic_end_user_distinction, by_brand)
            l['gl_account_code'] = gl_account
            lines.append(l)
            products.append(line.tbs_member_product.tbs_product.id)

        # Additional lines as per ticket #156
        if additional_lines is not None and len(additional_lines) > 0:

            # get period for fee.
            period = order.get_period()
            period = '%s-%02d-01' % (period.year, period.month)

            service_level_id = None
            pv_service_level_id = None
            if not order.is_guarantee_forecast:
                # get active service level for TBS member.
                msl = order.producer.get_service_level_for_date(order.get_period())
                service_level_id = None if msl.service_level is None else msl.service_level.id
                pv_service_level_id = service_level_id if msl.pv_service_level is None else msl.pv_service_level.id

            al = additional_lines
            # make a list out of the values
            for key, tbs_product_id in al.iteritems():
                if key[:12] != 'tbs_product_':
                    continue
                if not has_end_user_distinction and not by_brand:
                    if tbs_product_id in products:
                        raise PublicUserException(_('New line contains a product which is already part of this Declaration.'))

                value = key[12:]  # extra_%index

                product = pc.get_or_save_tbs_member_product(order.producer, tbs_product_id, country_id=al['line_%s_country_id' % value], subrogation=subrogation)

                sl = service_level_id
                if product.tbs_product.product_type == PvcTbsProduct.TYPE_PV_RELATED:
                    sl = pv_service_level_id

                is_b2b = None
                if automatic_end_user_distinction:
                    is_b2b = False
                elif has_end_user_distinction:
                    is_b2b = al['line_%s_is_b2b' % value]

                brand = None
                if by_brand:
                    brand = pc.get_or_save_brand(al['line_%s_brand' % value], product)

                # check if the is_b2b value for that product is the same and raise exception else save it
                for line in order.lines:
                    if line.tbs_member_product == product and line.is_b2b == is_b2b and line.brand == brand:
                        raise PublicUserException(_("New line has the same selections as product %s." % product.tbs_product.label))

                quantity = al['line_%s_quantity' % value]
                weight = al['line_%s_weight' % value]

                custom_unit = None
                if product.tbs_product.custom_unit is not None:
                    custom_unit = al['line_%s_custom_unit' % value]

                if weight > 0 or quantity > 0 or (custom_unit is not None and custom_unit > 0):
                    if weight == 0 or quantity == 0 or (custom_unit is not None and custom_unit == 0):
                        raise PublicUserException(_('Supplied values must either all be 0 or larger than 0 per line.'))

                dimension = PvcTbsProductFee.DIMENSION_NONE
                if is_b2b is not None:
                    dimension = PvcTbsProductFee.DIMENSION_IS_B2B if is_b2b else PvcTbsProductFee.DIMENSION_IS_B2C

                # guarantee forecast orders always have 0 prices
                if order.is_guarantee_forecast:
                    price = Decimal('0')
                else:
                    price = product.tbs_product.get_fee(period, sl, dimension)

                line = {
                    'polymorphic_type': PvcContributionOrderLine,
                    'description': '%s' % product.tbs_product.label,
                    'quantity': quantity,
                    'weight': weight,
                    'custom_unit': custom_unit,
                    'price': price,
                    'tbs_member_product': product,
                    'is_b2b': is_b2b,
                    'gl_account_code': gl_account,
                    'brand': brand,
                }
                self._validate_line_duplicate(line, lines, has_end_user_distinction or automatic_end_user_distinction, by_brand)
                lines.append(line)

        order = self._save_or_update_order(order.id, lines, order.merchant, order.client.default_group)
        order.status = self._instance_class.STATUS_PENDING
        order.is_compliant = self._instance_class.IS_COMPLIANT  # if the declaration got updated, they are compliant
        order.add_log('Contribution order updated by user. Setting compliant flag.')

        self.sa_session.add(order)
        self._reindex_order_lines(order)
        return order


    @transaction_decorator
    def save_contribution(self, _id, args):
        """
        moved from rpc originally
        """
        additional = args.get('additional', None)
        declaration = args['declaration']
        agreement = args['agreement']
        if not agreement.get('agreement', False):
            raise PublicUserException(_('You must declare that the information you provided is true.'))

        subrogations = None
        if args.has_key('subrogation'):
            subrogations = {}
            for k, v in args['subrogation'].iteritems():
                if v == '':
                    continue
                subrogations[k[12:]] = int(v)

        return self.update_contribution(_id, declaration, additional_lines=additional, subrogate=subrogations)

    @transaction_decorator
    def save_contribution_and_submit(self, _id, args):
        """
        moved from rpc originally
        """
        order = self.save_contribution(_id, args)
        if type(order) != list:
            self.submit_order(order.id)
            return order.id
        [self.submit_order(o.id) for o in order]

    def _line_from_value(self, line, values, has_end_user_distinction, automatic_end_user_distinction, by_brand, subrogation=None):
        # TODO: use this also for "additional line" stuff

        pc = PvcProductController(self.sa_session, account_id=self.get_account().id, request=self.request)
        quantity = int(values['line_%d_quantity' % line.index])
        weight = values['line_%d_weight' % line.index]

        custom_unit = None
        if line.tbs_member_product.tbs_product.custom_unit is not None:
            custom_unit = values['line_%d_custom_unit' % line.index]

        if quantity < 0 or weight < 0:
            raise PublicUserException(_('All quantities and weights must be larger or equal to zero.'))

        if weight > 0 or quantity > 0 or (custom_unit is not None and custom_unit > 0):
            if weight == 0 or quantity == 0 or (custom_unit is not None and custom_unit == 0):
                raise PublicUserException(_('Supplied values must either all be 0 or larger than 0 per line.'))

        is_b2b = None
        if automatic_end_user_distinction:
            is_b2b = False
        elif has_end_user_distinction:
            if values['line_%d_is_b2b' % line.index] is None:
                raise PublicUserException(_('All declaration lines require B2B/B2C distinction for this TBS.'))
            else:
                is_b2b = values['line_%d_is_b2b' % line.index]

        mp = line.tbs_member_product if subrogation is None else pc.get_subrogated_product(line.tbs_member_product, subrogation)

        brand = None
        if by_brand:
            brand = pc.get_or_save_brand(values['line_%d_brand' % line.index], mp)
        return {
            'polymorphic_type': PvcContributionOrderLine,
            'description': '%s' % line.description,
            'quantity': quantity,
            'weight': weight,
            'custom_unit': custom_unit,
            'price': line.price_per_unit,
            'tbs_member_product': mp,
            'is_b2b': is_b2b,
            'brand': brand,
        }

    def _validate_line_duplicate(self, validate, lines, end_user_distinction, by_brand):
        country = validate['tbs_member_product'].country
        product = validate['tbs_member_product'].tbs_product
        brand = validate['brand']
        is_b2b = validate['is_b2b']
        for line in lines:
            if product != line['tbs_member_product'].tbs_product:
                continue

            if country != line['tbs_member_product'].country:
                continue

            if not end_user_distinction and not by_brand:
                raise PublicUserException(_('New line contains a product %s which is already part of this Declaration.' % product.label))

            if end_user_distinction and not by_brand:
                if is_b2b == line['is_b2b']:
                    raise PublicUserException(_("New line has the same value for B2B/B2C for product %s." % product.label))
                continue

            if not end_user_distinction and by_brand:
                if brand == line['brand']:
                    raise PublicUserException(_("New line has the same brand %(brand)s for product %(product)s." % {
                        'brand': brand,
                        'product': product.label
                    }))
                continue

            # only option left is both end_user_distinction and by_brand
            if is_b2b == line['is_b2b'] and brand == line['brand']:
                raise PublicUserException(_("New line has the same brand %(brand)s and value for B2B/B2C for product %(product)s." % {
                    'brand': brand,
                    'product': product.label
                }))

    @transaction_decorator
    def _subrogate_order(self, order, values, subrogate, additional_lines=None):
        orders = {}

        line_products = {}
        for line in order.lines:
            line_products['line_%s' % line.index] = line

        pc = PvcProductController(self.sa_session, account_id=self.get_account().id, request=self.request)

        has_end_user_distinction = order.merchant.get_config('pvc.tbs.require.end.user.distinction')
        automatic_end_user_distinction = order.merchant.get_config('pvc.tbs.contribution.automatic.end.user.distinction')
        by_brand = order.merchant.get_config('pvc.tbs.contribution.split.by.brand')

        used_indexes = []
        for line_index, subrogation_id in subrogate.iteritems():
            if not subrogation_id in [s.id for s in order.tbs_member.subrogations]:
                raise PublicSecurityException(_('You do not have access to the supplied Subrogation.'))

            line = line_products[line_index]
            subrogation = [s for s in order.tbs_member.subrogations if s.id == subrogation_id].pop()

            if subrogation.status != PvcSubrogation.STATUS_ACTIVE:
                raise PublicSecurityException(_('The supplied Subrogation is not active.'))

            if subrogation.commissioner is None:
                self.logger.warn('Approved subrogation without a commissioner, something went wrong, order: %s, subrogation: %s' % (order.id, subrogation.id))
                raise SecurityException('Approved subrogation without a commissioner, something went wrong!')

            if not orders.has_key(subrogation.id):
                orders[subrogation.id] = {'tbs_member': subrogation.commissioner, 'producer': order.tbs_member, 'lines': []}

            l = self._line_from_value(line, values, has_end_user_distinction, automatic_end_user_distinction, by_brand, subrogation=subrogation)
            orders[subrogation.id]['lines'].append(l)
            used_indexes.append(line.index)

        fields = self._additional_order_fields[self._instance_class]
        kwargs = {}
        for f in fields:
            kwargs[f] = getattr(order, f)

        rv = []
        for subrogation_id, local_order in orders.iteritems():
            kwargs['producer'] = local_order['producer']
            kwargs['tbs_member'] = local_order['tbs_member']
            kwargs['is_forecast'] = order.is_forecast

            if not self._is_undeclared_period(local_order['tbs_member'], order.year, order.quarter, order.month, local_order['producer']):
                raise PublicUserException(_('A declaration for this period already exists at commissioner %(commissioner)s, please use the resubmit option.' % {'commissioner': kwargs['tbs_member'].member.name}))

            o = self._save_or_update_order(None, local_order['lines'], merchant=order.merchant, scope=local_order['tbs_member'].member.default_group, **kwargs)
            o.status = self._instance_class.STATUS_PENDING
            o.is_compliant = True
            self.sa_session.add(o)
            rv.append(o)

            self.sa_session.add(o)

        # remove lines from original order.
        if len(used_indexes) == len(order.lines):
            # all lines are moved to subrogations.
            self.sa_session.delete(order)
            self.sa_session.flush()
            return rv

        lines = []
        new_values = {}
        i = 0
        for line in order.lines:
            if line.index in used_indexes:
                continue
            l = self._line_from_value(line, values, has_end_user_distinction, automatic_end_user_distinction, by_brand)
            lines.append(l)
            new_values['line_%d_quantity' % i] = l['quantity']
            new_values['line_%d_weight' % i] = l['weight']
            new_values['line_%d_custom_unit' % i] = l['custom_unit']
            new_values['line_%d_is_b2b' % i] = l['is_b2b']
            new_values['line_%d_brand' % i] = None if l['brand'] is None else l['brand'].name

            i += 1

        order = self._save_or_update_order(order.id, lines, merchant=order.tbs_member.tbs, scope=order.tbs_member.member.default_group,
                                          tbs_member=order.tbs_member, year=order.year, month=order.month, quarter=order.quarter, producer=order.producer)
        order.status = self._instance_class.STATUS_PENDING
        self.sa_session.add(order)

        order = self.update_contribution(order.id, new_values, additional_lines=additional_lines)

        rv.append(order)

        self.sa_session.flush()
        return rv

    @transaction_decorator
    def _add_contribution_order(self, tbs_member, lines, year, month=None, quarter=None, producer=None, is_forecast=False, is_guarantee_forecast=False):

        self.logger.info('_add_contribution_order: tbs_member: %s, year=%s, month=%s, quarter=%s' % (tbs_member, year, month, quarter))

        if month is None:
            if quarter is None:
                month = 1
            else:
                month = (quarter-1) * 3 + 1

        period = datetime.date(int(year), int(month), 1)

        # Check if this period is within the obligation period of the TBS member.
        if tbs_member.obligation_start_date > period:
            raise PublicUserException(_('Incorrect period (before obligation start date) supplied for TBS member.'))

        # get active service level for TBS member.
        producer = tbs_member if producer is None else producer

        periodicity = tbs_member.declaration_period
        if periodicity != PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY:
            quarter = None
        if periodicity != PvcTbsLicense.DECLARATION_PERIOD_MONTHLY:
            month = None

        if is_guarantee_forecast:
            # guarantee forecast orders always go pear year
            service_level_id = None
            quarter = None
            month = None
        else:
            msl = producer.get_service_level_for_date(period)
            if msl is None:
                raise PublicUserException(_('No active service level for period %s' % period))
            service_level_id = None if msl.service_level is None else msl.service_level.id
            pv_service_level_id = service_level_id if msl.pv_service_level is None else msl.pv_service_level.id

        # check if order for period/tbs_member does not exist yet.
        order = self.sa_session.query(PvcContributionOrder).filter(and_(
            PvcContributionOrder.tbs_member == tbs_member,
            PvcContributionOrder.year == year,
            PvcContributionOrder.month == month,
            PvcContributionOrder.quarter == quarter,
            PvcContributionOrder.producer == producer,
            PvcContributionOrder.is_guarantee_forecast == is_guarantee_forecast,
        )).first()

        if order is not None:
            raise PublicSecurityException(_('Declaration already exists for provided TBS Member and period.'))

        order = {
            'merchant': tbs_member.tbs,
            'scope': tbs_member.member.default_group,
            'tbs_member': tbs_member,
            'year': year,
            'month': month,
            'quarter': quarter,
            'producer': producer,
            'is_forecast': is_forecast,
            'is_guarantee_forecast': is_guarantee_forecast,
            'lines': []
        }

        period = '%s-01-01' % year
        if month is not None:
            period = '%s-%02d-01' % (year, month)
        elif quarter is not None:
            import math
            m = ((quarter-1)*3)+1
            period = '%s-%02d-01' % (year, m)

        by_brand = tbs_member.tbs.get_config('pvc.tbs.contribution.split.by.brand')
        has_end_user_distinction = tbs_member.tbs.get_config('pvc.tbs.require.end.user.distinction')
        automatic_end_user_distinction = tbs_member.tbs.get_config('pvc.tbs.contribution.automatic.end.user.distinction')

        # for each line we get the price per unit, the basis for invoicing and the status of the tbs product.
        for line in lines:
            quantity = line['quantity']
            weight = line['weight']
            custom_unit = line['custom_unit']
            gl_account = None if not is_forecast else PvcContributionOrderLine.gl_forecast_account_code

            assert type(quantity) == int
            assert quantity >= 0

            assert isinstance(weight, Decimal)
            assert weight >= 0

            if custom_unit is not None:
                assert isinstance(custom_unit, Decimal)
                assert custom_unit >= 0

            # verify that if one of the fields has a value larger than 0, they all should.
            if weight > 0 or quantity > 0 and (custom_unit is None or custom_unit > 0):
                if weight == 0 or quantity == 0 or (custom_unit is not None and custom_unit == 0):
                    raise PublicUserException(_('Supplied values must either all be 0 or larger than 0 per line.'))

            p = line['tbs_member_product']
            # check the TBS product status
            if p.tbs_product.status != PvcTbsProduct.STATUS_USE_IN_DECLARATION:
                self.logger.info('_add_contribution_order: member product skipped due to inactivity on TBS level: %s' % p)
                continue

            is_b2b = None
            if automatic_end_user_distinction:
                is_b2b = False
            elif has_end_user_distinction:
                if not line.has_key('is_b2b') or line['is_b2b'] is None:
                    raise PublicUserException(_('All declaration lines require B2B/B2C distinction for this TBS.'))
                else:
                    is_b2b = line['is_b2b']

            brand = None
            if by_brand:
                brand = line['brand']
                assert brand is not None

            dimension = PvcTbsProductFee.DIMENSION_NONE
            if is_b2b is not None:
                dimension = PvcTbsProductFee.DIMENSION_IS_B2B if is_b2b else PvcTbsProductFee.DIMENSION_IS_B2C

            # get the fee
            # guarantee forecast orders always have price 0
            if is_guarantee_forecast:
                price = Decimal('0')
            else:

                if p.tbs_product.product_type == PvcTbsProduct.TYPE_PV_RELATED:
                    service_level_id = pv_service_level_id

                if service_level_id is None:
                    raise PvcServiceLevelException(_('No service level active for product %s' % p.tbs_product.label ))

                price = p.tbs_product.get_fee(period, service_level_id, dimension)

                if price is None:

                    sl = self.sa_session.query(PvcTbsServiceLevel).filter(PvcTbsServiceLevel.id == service_level_id).one()

                    raise PublicUserException(_('There is no price set for "%(product)s" for period %(period)s, service level "%(service_level)s" and dimension %(dimension)s.') % {
                        'product': p.tbs_product.label,
                        'period': period,
                        'service_level': sl.name,
                        'dimension': 'None' if dimension is PvcTbsProductFee.DIMENSION_NONE else
                            'B2B' if dimension is PvcTbsProductFee.DIMENSION_IS_B2B else 'B2C'
                    })

            assert price is not None

            line = {
                'polymorphic_type': PvcContributionOrderLine,
                'description': p.tbs_product.label,
                'quantity': quantity,
                'price': price,
                'tbs_member_product': p,
                'weight': weight,
                'custom_unit': custom_unit,
                'is_b2b': is_b2b,
                'gl_account_code': gl_account,
                'brand': brand
            }

            self._validate_line_duplicate(line, order['lines'], has_end_user_distinction or automatic_end_user_distinction, by_brand)
            order['lines'].append(line)

        self.logger.info('_add_contribution_order: DECLARATION ADDED for tbs_member: %s, year=%s, month=%s, quarter=%s' % (tbs_member, year, month, quarter))
        order = self._save_or_update_order(None, **order)
        self._reindex_order_lines(order)

        return order

    def trigger_new_contributions(self, tbs):
        """
        triggers new declarations for the given tbs
        """

        today = virtual_date.today()
        if today.day != 1:  # we trigger only on the first
            return None

        has_forecast = tbs.get_config('pvc.tbs.advance.invoicing')

        # get all the TBS members (active, producers) with an obligation start date before today
        tms = self.sa_session.query(PvcTbsMember).filter(
            PvcTbsMember.tbs == tbs,
            PvcTbsMember.status.in_((PvcTbsMember.STATUS_APPROVED, PvcTbsMember.STATUS_RESIGNING, )),
            PvcTbsMember.is_producer == True,
            PvcTbsMember.obligation_start_date <= today
        ).all()

        quarterly = today.month in (4, 7, 10, 1)
        yearly = today.month == 1

        q = today.month/3 if today.month/3 > 0 else 4

        year = today.year if today.month > 1 else today.year - 1
        month = today.month-1 if today.month > 1 else 12
        quarter = None if not quarterly else q

        orders = []
        for tm in tms:

            periodicity = tm.declaration_period
            if periodicity == PvcTbsLicense.DECLARATION_PERIOD_YEARLY and not yearly:
                continue
            if periodicity == PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY and not quarterly:
                continue

            m = month
            q = quarter
            # it is the first of -a- month, so monthly setting always triggers new declaration.
            if periodicity != PvcTbsLicense.DECLARATION_PERIOD_MONTHLY:
                m = None
            if periodicity != PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY:
                q = None

            producer = tm
            tms = [tm]
            if len(tm.subrogations) > 0:
                tms += [s.commissioner for s in tm.subrogations if s.status == PvcSubrogation.STATUS_ACTIVE]

            for tbs_member in tms:
                try:
                    if has_forecast:
                        order = self._enable_resubmit_on_forecast(tbs_member, year, m, q, producer)
                    else:
                        order = self.add_open_contribution_order(tbs_member, year, m, q, producer)

                    if order is not None:
                        orders.append(order)
                # This gets called from the task runner, we do not want it to exit on Public exceptions.
                except (PublicUserException, PublicSecurityException), e:
                    self.logger.warn('trigger_new_contributions: %s' % e.message)

        return orders

    def add_open_contribution_order(self, tm, year, month, quarter, producer=None, subrogation_id=None):

        periodicity = tm.declaration_period
        try:
            assert year is not None
            if periodicity == PvcTbsLicense.DECLARATION_PERIOD_YEARLY:
                assert month is None and quarter is None
            elif periodicity == PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY:
                assert month is None and quarter is not None
            elif periodicity == PvcTbsLicense.DECLARATION_PERIOD_MONTHLY:
                assert month is not None and quarter is None
        except AssertionError:
            raise PublicUserException(_('Incorrect period provided for new contribution.'))

        has_end_user_distinction = tm.tbs.get_config('pvc.tbs.require.end.user.distinction')
        by_brand = tm.tbs.get_config('pvc.tbs.contribution.split.by.brand')
        is_b2b = (True, False, ) if has_end_user_distinction else (None, )
        subrogation = None
        if producer is None:
            producer = tm
        elif producer != tm:
            subrogation = [s for s in producer.subrogations if s.commissioner == tm].pop()
        if subrogation_id is not None:
            subrogation = self.sa_session.query(PvcSubrogation).filter(PvcSubrogation.id == subrogation_id).first()
        if subrogation is not None:
            assert subrogation.status is PvcSubrogation.STATUS_ACTIVE
            tm = subrogation.commissioner
            producer = subrogation.subrogator

        lines = []
        products = self.get_active_tbs_member_products(producer, subrogation)
        if len(products) == 0:  # this member did not add any products to their library yet
            self.logger.info('%s - TBS member has not yet added any products to their library at %s' % (producer, tm))
            return None
        for mp in products:
            brands = [None]
            if by_brand:
                brands = [b.brand for b in mp.brands if b.is_active]
                if len(brands) == 0:
                    self.logger.info('%s - No brands are active for TBS Member product %s' % (producer, mp))
                    continue
            for b2b in is_b2b:
                for b in brands:
                    # a new declaration will be made with all the TBS member products.
                    lines.append({
                        'tbs_member_product': mp,
                        'quantity': 0,
                        'weight': Decimal('0.0'),
                        'brand': b,
                        'custom_unit': None if mp.tbs_product.custom_unit is None else Decimal('0.0'),
                        'is_b2b': b2b
                    })
        if len(lines) == 0:
            self.logger.info('%s - Unable to add a new Contribution order - no valid lines found %s' % (producer, tm))
            return None

        return self._add_contribution_order(tm, lines, year, month, quarter, producer=producer)

    add_open_declaration = add_open_contribution_order

    def _enable_resubmit_on_forecast(self, tbs_member, year, month, quarter, producer):
        forecast = self._is_undeclared_period(tbs_member, year, quarter, month, producer, return_order=True)
        if not isinstance(forecast, PvcContributionOrder):  # this mean we do not have a forecast yet (new member...)
            self.logger.info('No forecast found for tbs_member %s and producer %s' % (tbs_member, producer))
            return None

        forecast.resubmit_request = PvcContributionOrder.RESUBMIT_REQUEST_APPROVED
        self.logger.info('Automatic resubmit approval on forecast error at the beginning of period for order %s' % forecast)

        return forecast

    def _get_available_rule_conditions(self, order_type):

        conditions = super(PvcContributionOrderController, self)._get_available_rule_conditions(order_type)

        # override merchant / client
        conditions[0] = PvcRuleConditionTbs(self)
        conditions[1] = PvcRuleConditionMember(self)

        conditions += [
            PvcRuleConditionContributionProducer(self),
            PvcRuleConditionContributionIsCompliant(self),
            PvcRuleConditionContributionIsForecast(self),
            PvcRuleConditionContributionYear(self),
            PvcRuleConditionContributionQuarter(self),
            PvcRuleConditionContributionMonth(self),
            PvcRuleConditionIsFreeRider(self),
            PvcRuleConditionContributionLineTbsMemberProduct(self)
        ]
        return conditions

    def _get_available_rule_actions(self, order_type):

        actions = super(PvcContributionOrderController, self)._get_available_rule_actions(order_type)
        actions += [
            PvcRuleActionApplyMemberCountryVatRate(self),
            PvcRuleActionContributionB2bPercentage(self),
        ]
        return actions

    def get_next_closing_date(self, tm):

        today = virtual_date.today()

        period = tm.declaration_period
        closing_day = tm.closing_day

        closing_date = None

        if period == PvcTbsLicense.DECLARATION_PERIOD_YEARLY:
            closing_date = datetime.date(today.year, 1, 1) + datetime.timedelta(days=closing_day)

            if closing_date < today:
                closing_date = closing_date.replace(year=today.year+1)

        elif period == PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY:

            # get current quarter
            q = range(4)
            q.reverse()
            for i in q:
                x = datetime.date(today.year, (i*3)+1, 1)
                if x <= today:
                    break
            closing_date = datetime.date(x.year, x.month, 1) + datetime.timedelta(days=closing_day)

            if closing_date < today:
                closing_date = closing_date + relativedelta(months=3)

        elif period == PvcTbsLicense.DECLARATION_PERIOD_MONTHLY:
            closing_date = datetime.date(today.year, today.month, 1) + datetime.timedelta(days=closing_day)
            if closing_date < today:
                closing_date = closing_date + relativedelta(months=1)

        assert closing_date != None
        return closing_date

    def _close_contributions(self, open_orders):
        closed = []
        for order in open_orders:
            if today() != self.get_next_closing_date(order.tbs_member):
                continue

            self.submit_order(order)
            closed.append(order)
            self.logger.info('_close_contributions: closing contribution %s' % order)

        return closed

    @transaction_decorator
    def close_contributions(self, tbs):
        _where = self._instance_class.status.in_([self._instance_class.STATUS_NEW, self._instance_class.STATUS_PENDING])
        _where = and_(_where, self._instance_class.type!=self._instance_class.TYPE_CREDIT)
        _where = and_(_where, self._instance_class.merchant==tbs)
        _where = and_(_where, self._instance_class.is_guarantee_forecast==False)

        open_orders = self.sa_session.query(self._instance_class).filter(_where).all()
        closed = self._close_contributions(open_orders)
        return closed

    @transaction_decorator
    def update_with_new_fee(self, product_fee):
        """
        In case the fee for a service level-product-dimension combination gets altered we change the contribution lines
        for any non-saved/non-submitted declaration.

        There is no direct relation between service levels and contribution orders, so we will get all the TBS members
        with the supplied service level as active and with NEW declarations which carry the product/dimension combination.

        This is done in the background as it can take a bit of time.
        """
        assert isinstance(product_fee, PvcTbsProductFee)
        self.logger.info('Verifying alteration of Contribution orders due to fee change.')

        # get NEW orders with the referenced product/dimension combination

        # get products affected (child products of the tbs_product which do not have any individual fee set)
        products = []

        def get_children(tbs_product):
            products.append(tbs_product)
            for p in tbs_product.get_child_products_with_inherited_fee():
                get_children(p)

        get_children(product_fee.tbs_product)

        _where = and_(
            PvcContributionOrder.status == PvcContributionOrder.STATUS_NEW,
            PvcTbsMemberProduct.tbs_product_id.in_([p.id for p in products]),
        )
        if product_fee.dimension != PvcTbsProductFee.DIMENSION_NONE:
            _where = and_(_where, PvcContributionOrderLine.is_b2b == (product_fee.dimension == PvcTbsProductFee.DIMENSION_IS_B2B))

        # get periods affected:
        pc = PvcProductController(self.sa_session)
        yearly, quarterly, monthly = pc.get_periods_for_fee(product_fee)
        # if there are no periods affected, we dont need to update anything.
        if len(yearly) == 0 and len(quarterly) == 0 and len(monthly) == 0:
            return

        _periods = None
        for y in yearly:
            w = and_(PvcContributionOrder.year == y, PvcContributionOrder.month == None, PvcContributionOrder.quarter == None)
            if _periods is None:
                _periods = w
            else:
                _periods = or_(_periods, w)

        for quarter in quarterly:
            y, q = quarter
            w = and_(PvcContributionOrder.year == y, PvcContributionOrder.month == None, PvcContributionOrder.quarter == q)
            if _periods is None:
                _periods = w
            else:
                _periods = or_(_periods, w)

        for month in monthly:
            y, m = month
            w = and_(PvcContributionOrder.year == y, PvcContributionOrder.month == m, PvcContributionOrder.quarter == None)
            if _periods is None:
                _periods = w
            else:
                _periods = or_(_periods, w)

        _where = and_(_where, _periods)

        affected = []
        orders = self.sa_session.query(PvcContributionOrder).join(
            PvcContributionOrderLine).join(PvcTbsMemberProduct).filter(_where).all()
        for order in orders:

            period = order.get_period()

            # the producer is not using this service level.

            if product_fee.tbs_product.product_type == PvcTbsProduct.TYPE_PV_RELATED:
                if order.producer.get_service_level_for_date(period).pv_service_level != product_fee.service_level:
                    continue
            elif product_fee.tbs_product.product_type == PvcTbsProduct.TYPE_OTHER:
                if order.producer.get_service_level_for_date(period).service_level != product_fee.service_level:
                    continue
            else:
                raise Exception('invalid product type')

            o = self._update_order_with_new_fees(order)
            if o is None:
                continue
            affected.append(o)

        # TODO sent notification to Sien

        return affected

    @transaction_decorator
    def _update_order_with_new_fees(self, order):
        # we cannot update orders that are not NEW.
        if order.status != PvcContributionOrder.STATUS_NEW:
            return None

        for line in order.lines:
            line.price_per_unit = self.get_child_order_price_for_line(order, line, is_credit=False)
            self.recalculate_line(order, line)

        self._update_order_totals(order)
        self.logger.info('Updated order #%d with new fees' % order.id)
        return order

class PvcDeclarationAggregator(IQubeAggregator):
    _table = "aggPvcDeclaration"
    _affected_cubes = ({'datasource': 'PvcOrderData', 'cube': 'PvcDeclarationDataCube'}, )
    _columns = [
        'order_id',
        'order_line_id',
        'order_type',
        'order_status',
        'invoice_type',
        'country_id',
        'year',
        'quarter',
        'month',
        'tbs_id',
        'member_id',
        'producer_id',
        'pv_module_id',
        'supplier_id',
        'submitted_time_id',
        'payment_status',
        'payment_method',
        'payment_time_id',
        'gl_account',
        'quantity',
        'weight',
        'power',
        'line_amount_excluding_vat',
        'line_amount_including_vat',
        'price_per_unit',
        'last_updated'
    ]

    _select_sql = """

        SELECT
            o.id as order_id,
            ol.id as order_line_id,
            o.polymorphic_type as order_type,
            o.status as order_status,
            IF(o.type=0, 'Debit', 'Credit') as invoice_type,
            mpvm.country_id as country_id,
            do.year as `year`,
            IF(ISNULL(do.quarter) AND NOT ISNULL(do.month), IF(do.month < 4, 1, IF(do.month < 7, 2, IF(do.month < 10, 3, 4))), do.quarter) as `quarter`,
            do.month as `month`,
            ptm.member_id as producer_id,
            o.client_id as member_id,
            o.merchant_id as tbs_id,
            pvm.id as pv_module_id,
            pvms.id as supplier_id,
            IF(ISNULL(o.display_date), NULL, DATE_FORMAT(o.display_date, '%%Y%%m%%d')) as submitted_time_id,
            iqGetCurrentOrderPaymentStatus(o.id) as payment_status,
            iqGetCurrentOrderPaymentMethod(o.id) as payment_method,
            IF(ISNULL(iqGetCurrentOrderPaymentDate(o.id)), NULL, DATE_FORMAT(o.display_date, '%%Y%%m%%d')) as payment_time_id,
            ol.gl_account_number as gl_account,
            ol.quantity as quantity,
            (ol.quantity * pvm.weight) as weight,
            (ol.quantity * pvm.nominal_power) as power,
            ol.line_amount_excluding_vat as line_amount_excluding_vat,
            ol.line_amount_including_vat as line_amount_including_vat,
            ol.price_per_unit as price_per_unit,
            o.last_updated as last_updated
        FROM tblOrder as o
        INNER JOIN tblOrderLine as ol
            ON o.id = ol.order_id
        INNER JOIN tblPvcDeclarationOrder as do
            ON o.id = do.order_id
        INNER JOIN tblPvcDeclarationOrderLine as dol
            ON ol.id = dol.order_line_id
        INNER JOIN tblPvcTbsMember as ptm
            ON do.producer_id = ptm.id
        INNER JOIN tblPvcTbsMemberPVModule as mpvm
            ON dol.tbs_member_module_id = mpvm.id
        INNER JOIN tblPvcPVModule as pvm
            ON mpvm.module_id = pvm.id
        INNER JOIN tblPvcPVModuleSupplier as pvms
            ON pvm.supplier_id = pvms.id
        WHERE o.last_updated > %(max_last_updated)s
        AND o.deleted = 0
        ORDER BY o.id
    """

    def run(self, loop_validator=None):

        affected_rows = 0

        max_last_updated = 0
        rows = self.db_engine_ro.execute("""
            SELECT IFNULL(MAX(a.last_updated), 0)
            FROM %s as a""" % self._table)

        row = rows.fetchone()
        rows.close()

        if row is not None:
            max_last_updated = row[0]

        self.db_engine.execute("""
            DELETE FROM %(tbl)s WHERE order_id IN (
                SELECT id
                FROM tblOrder
                WHERE last_updated > '%(max_last_updated)s'
            )""" % {
            'tbl': self._table,
            'max_last_updated': max_last_updated
        })

        affected_rows = affected_rows + self._work({
            'max_last_updated': max_last_updated,
        }, loop_validator)

        return affected_rows

class PvcContributionAggregator(IQubeAggregator):
    _table = "aggPvcContribution"
    _affected_cubes = ({'datasource': 'PvcOrderData', 'cube': 'PvcContributionDataCube'}, )

    _columns = [
        'order_id',
        'order_line_id',
        'order_type',
        'order_status',
        'invoice_type',
        'country_id',
        'year',
        'quarter',
        'month',
        'tbs_id',
        'member_id',
        'producer_id',
        'commissioner_id',
        'submitted_time_id',
        'payment_status',
        'payment_method',
        'payment_time_id',
        'gl_account',
        'product_id',
        'local_product_id',
        'brand_id',
        'quantity',
        'weight',
        'nominal_power',
        'surface',
        'line_amount_excluding_vat',
        'line_amount_including_vat',
        'price_per_unit',
        'last_updated'
    ]

    _select_sql = """
    SELECT
            o.id as order_id,
            ol.id as order_line_id,
            IF(co.is_forecast = 1, 'Forecast', IF(co.is_guarantee_forecast, 'Guarantee Forecast', 'Declaration')) as order_type,
            o.status as order_status,
            IF(o.type=0, 'Debit', 'Credit') as invoice_type,
            mp.country_id as country_id,
            co.year as `year`,
            IF(ISNULL(co.quarter) AND NOT ISNULL(co.month), IF(co.month < 4, 1, IF(co.month < 7, 2, IF(co.month < 10, 3, 4))), co.quarter) as `quarter`,
            co.month as `month`,
            o.merchant_id as tbs_id,
            o.client_id as member_id,
            ptm.member_id as producer_id,
            IF(co.tbs_member_id = co.producer_id, NULL, ctm.member_id) as commissioner_id,
            IF(ISNULL(o.display_date), NULL, DATE_FORMAT(o.display_date, '%%Y%%m%%d')) as submitted_time_id,
            iqGetCurrentOrderPaymentStatus(o.id) as payment_status,
            iqGetCurrentOrderPaymentMethod(o.id) as payment_method,
            IF(ISNULL(iqGetCurrentOrderPaymentDate(o.id)), NULL, DATE_FORMAT(o.display_date, '%%Y%%m%%d')) as payment_time_id,
            ol.gl_account_number as gl_account,
            p.product_id as product_id,
            p.id as local_product_id,
            col.brand_id as brand_id,
            ol.quantity as quantity,
            col.weight as weight,
            IF(p.custom_unit = """+str(PvcTbsProduct.CUSTOM_UNIT_NOMINAL_POWER)+""", col.custom_unit, NULL) as nominal_power,
            IF(p.custom_unit = """+str(PvcTbsProduct.CUSTOM_UNIT_SURFACE)+""", col.custom_unit, NULL) as surface,
            ol.line_amount_excluding_vat as line_amount_excluding_vat,
            ol.line_amount_including_vat as line_amount_including_vat,
            ol.price_per_unit as price_per_unit,
            o.last_updated as last_updated
        FROM tblOrder as o
        INNER JOIN tblOrderLine as ol
            ON o.id = ol.order_id
        INNER JOIN tblPvcContributionOrder as co
            ON o.id = co.order_id
        INNER JOIN tblPvcContributionOrderLine as col
            ON ol.id = col.order_line_id
        INNER JOIN tblPvcTbsMember as ptm
            ON co.producer_id = ptm.id
        INNER JOIN tblPvcTbsMember as ctm
            ON co.tbs_member_id = ctm.id
        INNER JOIN tblPvcTbsMemberProduct as mp
            ON col.tbs_member_product_id = mp.id
        INNER JOIN tblPvcTbsProduct as p
            ON mp.tbs_product_id = p.id
        WHERE o.last_updated > %(max_last_updated)s
        AND o.deleted = 0
        ORDER BY o.id

    """

    def run(self, loop_validator=None):

        affected_rows = 0

        max_last_updated = 0
        rows = self.db_engine_ro.execute("""
            SELECT IFNULL(MAX(a.last_updated), 0)
            FROM %s as a""" % self._table)

        row = rows.fetchone()
        rows.close()

        if row is not None:
            max_last_updated = row[0]

        self.db_engine.execute("""
            DELETE FROM %(tbl)s WHERE order_id IN (
                SELECT id
                FROM tblOrder
                WHERE last_updated > '%(max_last_updated)s'
            )""" % {
            'tbl': self._table,
            'max_last_updated': max_last_updated
        })

        affected_rows = affected_rows + self._work({
            'max_last_updated': max_last_updated,
        }, loop_validator)

        return affected_rows
