# Copyright (c) 2014 BaseSoft B.V.
# See LICENSE for details.

from django.conf import settings
from django.utils.translation import ugettext as _

from basesoft.web.core.management.commands.iqubesync import ProjectBase, CubeBase, RoleBase
from basesoft.web.core.util.controllers import AbstractController
from basesoft.web.core.exceptions import PublicSecurityException
from basesoft.logging import getLogger
from basesoft.web.contrib.order.iqube_sync import OrderCube

from iqportal.iqube.core.models import DataSource
from iqportal.iqube.core.controllers import IQubeRoleController, IQubeRole

from pvcycle.order.models import PvcDeclarationOrder, PvcOrder
from pvcycle.core.models import PvcOperatorLicense, PvcTbsLicense, PvcMemberLicense

logger = getLogger('pvcycle.order.iqube_sync')

class PvcOrderDataProject(ProjectBase):
    project_name = 'PvcOrderData'
    world_privileges = DataSource.PRIV_VIEW
    description = 'Pvc Order Data cubes project'
    role_level = DataSource.ROLE_LEVEL_DB

    def __init__(self, sa_session, controller):
        ProjectBase.__init__(self, sa_session, controller)
        self._db_from_connstr(settings.DB_CONNSTR_READONLY)

    def get_cubes(self):
        return PvcDeclarationDataCube, PvcOrderCube, PvcContributionDataCube

    def get_roles(self):
        return PvcDeclarationDataCubeRole, PvcOrderCubeRole, PvcContributionDataCubeRole

class PvcDeclarationDataCube(CubeBase):
    name = 'PvcDeclarationDataCube'

    def _xml(self):
        return """
    <Cube name="%(name)s" defaultMeasure="Amount (excl. vat)">

        <Table name="aggPvcDeclaration" />

        <Dimension name="Order" foreignKey="order_id">
            <Hierarchy hasAll="true" allMemberName="All Orders" primaryKey="id" primaryKeyTable="tblOrder">
                    <Table name="tblOrder"/>
                <Level name="Order" table="tblOrder" column="id" type="Integer" />
            </Hierarchy>
        </Dimension>

        <Dimension name="Order Line" foreignKey="order_line_id">
            <Hierarchy hasAll="true" allMemberName="All Order Lines" primaryKey="id" primaryKeyTable="tblOrderLine">
                    <Table name="tblOrderLine"/>
                <Level name="Order Line" table="tblOrderLine" column="id" type="Integer" />
            </Hierarchy>
        </Dimension>

        <Dimension name="Order Type">
            <Hierarchy hasAll="true" allMemberName="All Order Types">
                <Level name="Order Type" column="order_type"/>
            </Hierarchy>
        </Dimension>

        <Dimension name="Invoice Type">
            <Hierarchy hasAll="true" allMemberName="All Invoice Types">
                <Level name="Invoice Type" column="invoice_type" />
            </Hierarchy>
        </Dimension>

        <Dimension name="TBS" foreignKey="tbs_id">
            <Hierarchy hasAll="true" allMemberName="All TBSs" primaryKey="license_id" primaryKeyTable="tblPvcTbsLicense">
                <Join leftKey="license_id" rightKey="id">
                    <Table name="tblPvcTbsLicense"/>
                    <Table name="tblLicense" alias="tbs" />
                </Join>
                <Level name="TBS" table="tbs" column="id" captionColumn="name" type="Integer" ordinalColumn="name" />
            </Hierarchy>
        </Dimension>

        <Dimension name="Member" foreignKey="member_id">
            <Hierarchy hasAll="true" allMemberName="All Members" primaryKey="license_id" primaryKeyTable="tblPvcMemberLicense">
                <Join leftKey="license_id" rightKey="id">
                    <Table name="tblPvcMemberLicense"/>
                    <Table name="tblLicense" alias="member" />
                </Join>
                <Level name="Member" table="member" column="id" captionColumn="name" type="Integer" ordinalColumn="name"/>
            </Hierarchy>
        </Dimension>

        <Dimension name="Producer" foreignKey="producer_id">
            <Hierarchy hasAll="true" allMemberName="All TBS" primaryKey="license_id" primaryKeyTable="tblPvcMemberLicense">
                <Join leftKey="license_id" rightKey="id">
                    <Table name="tblPvcMemberLicense"/>
                    <Table name="tblLicense" alias="producer" />
                </Join>
                <Level name="Producer" table="producer" column="id" captionColumn="name" type="Integer" ordinalColumn="name"/>
            </Hierarchy>
        </Dimension>

        <Dimension name="Country" foreignKey="country_id">
            <Hierarchy hasAll="true" allMemberName="All Countries" primaryKey="id" primaryKeyTable="tblCountry">
                <Table name="tblCountry" />
                <Level name="Country" table="tblCountry" column="id" captionColumn="name" type="Integer" ordinalColumn="name"/>
            </Hierarchy>
        </Dimension>

        <Dimension name="Declaration Year">
            <Hierarchy hasAll="true" allMemberName="All Declaration Years">
                <Level name="Year" column="year"/>
            </Hierarchy>
        </Dimension>

        <Dimension name="Declaration Quarter">
            <Hierarchy hasAll="true" allMemberName="All Declaration Quarters">
                <Level name="Quarter" column="quarter">
                    <CaptionExpression>
                        <SQL dialect="mysql">
                            CONCAT('Q', `aggPvcDeclaration`.`quarter`)
                        </SQL>
                    </CaptionExpression>
                </Level>
            </Hierarchy>
        </Dimension>

        <Dimension name="Declaration Month">
            <Hierarchy hasAll="true" allMemberName="All Declaration Months">
                <Level name="Month" column="month">
                    <CaptionExpression>
                        <SQL dialect="mysql">
                            MONTHNAME(STR_TO_DATE(`aggPvcDeclaration`.`month`, '%%m'))
                        </SQL>
                    </CaptionExpression>
                </Level>
            </Hierarchy>
        </Dimension>

        <Dimension name="Brand" foreignKey="supplier_id">
            <Hierarchy hasAll="true" allMemberName="All Brands" primaryKey="id" primaryKeyTable="tblPvcPVModuleSupplier">
                <Table name="tblPvcPVModuleSupplier" />
                <Level name="Brand" table="tblPvcPVModuleSupplier" column="id" captionColumn="name" type="Integer" ordinalColumn="name"/>
            </Hierarchy>
        </Dimension>

        <Dimension name="Submit Period" foreignKey="submitted_time_id" type="TimeDimension">
            <Hierarchy name="Quarter" hasAll="true" primaryKey="id" >
                <Table name="aggTimeQuarter"/>
                <Level name="Year" column="year" type="Integer" uniqueMembers="true" levelType="TimeYears" ordinalColumn="year" />
                <Level name="Quarter" column="quarter" uniqueMembers="false" levelType="TimeQuarters"  ordinalColumn="quarter" />
                <Level name="Month" column="month" type="Integer" uniqueMembers="false" levelType="TimeMonths" ordinalColumn="month" />
                <Level name="Day" column="day" type="Integer" uniqueMembers="false" levelType="TimeDays" ordinalColumn="day" />
            </Hierarchy>
            <Hierarchy name="WeekISO8601" hasAll="true" primaryKey="id" caption="Week">
                <Table name="aggTimeWeekISO8601"/>
                <Level name="Year" column="year" type="Integer" uniqueMembers="true" levelType="TimeYears" ordinalColumn="year" />
                <Level name="Week" column="week" uniqueMembers="false" levelType="TimeWeeks" ordinalColumn="week" />
                <Level name="Day" column="day" type="Integer" uniqueMembers="false" levelType="TimeDays" ordinalColumn="day" />
            </Hierarchy>
            <Hierarchy name="WeekUS" hasAll="true" primaryKey="id" caption="Week">
                <Table name="aggTimeWeekUS"/>
                <Level name="Year" column="year" type="Integer" uniqueMembers="true" levelType="TimeYears" ordinalColumn="year" />
                <Level name="Week" column="week" uniqueMembers="false" levelType="TimeWeeks" ordinalColumn="week" />
                <Level name="Day" column="day" type="Integer" uniqueMembers="false" levelType="TimeDays" ordinalColumn="day" />
            </Hierarchy>
        </Dimension>

        <Dimension name="Payment Period" foreignKey="payment_time_id" type="TimeDimension">
            <Hierarchy hasAll="true" primaryKey="id" name="Quarter">
                <Table name="aggTimeQuarter"/>
                <Level name="Year" column="year" type="Integer" uniqueMembers="true" levelType="TimeYears" ordinalColumn="year" />
                <Level name="Quarter" column="quarter" uniqueMembers="false" levelType="TimeQuarters"  ordinalColumn="quarter" />
                <Level name="Month" column="month" type="Integer" uniqueMembers="false" levelType="TimeMonths" ordinalColumn="month" />
                <Level name="Day" column="day" type="Integer" uniqueMembers="false" levelType="TimeDays" ordinalColumn="day" />
            </Hierarchy>
            <Hierarchy name="WeekISO8601" hasAll="true" primaryKey="id" caption="Week">
                <Table name="aggTimeWeekISO8601"/>
                <Level name="Year" column="year" type="Integer" uniqueMembers="true" levelType="TimeYears" ordinalColumn="year" />
                <Level name="Week" column="week" uniqueMembers="false" levelType="TimeWeeks" ordinalColumn="week" />
                <Level name="Day" column="day" type="Integer" uniqueMembers="false" levelType="TimeDays" ordinalColumn="day" />
            </Hierarchy>
            <Hierarchy name="WeekUS" hasAll="true" primaryKey="id" caption="Week">
                <Table name="aggTimeWeekUS"/>
                <Level name="Year" column="year" type="Integer" uniqueMembers="true" levelType="TimeYears" ordinalColumn="year" />
                <Level name="Week" column="week" uniqueMembers="false" levelType="TimeWeeks" ordinalColumn="week" />
                <Level name="Day" column="day" type="Integer" uniqueMembers="false" levelType="TimeDays" ordinalColumn="day" />
            </Hierarchy>
        </Dimension>

        <Dimension name="Payment Status">
            <Hierarchy hasAll="true" allMemberName="All Payment Statuses">
                <Level name="Payment Status" column="payment_status" type="Integer">
                    <CaptionExpression>
                        <SQL dialect="mysql">
                            CASE `aggPvcDeclaration`.`payment_status`
                                WHEN 0 THEN 'Pending'
                                WHEN 1 THEN 'Submitted'
                                WHEN 2 THEN 'Partial'
                                WHEN 3 THEN 'Completed'
                                WHEN 4 THEN 'Rejected'
                                ELSE '-' END
                        </SQL>
                    </CaptionExpression>
                </Level>
            </Hierarchy>
        </Dimension>

        <Dimension name="Payment Method">
            <Hierarchy hasAll="true" allMemberName="All Payment Methods">
                <Level name="Payment Method" column="payment_method" />
            </Hierarchy>
        </Dimension>

        <Dimension name="GL Account">
            <Hierarchy hasAll="true" allMemberName="All GL Accounts">
                <Level name="GL Account" column="gl_account" />
            </Hierarchy>
        </Dimension>

        <Dimension name="Is Full Member" foreignKey="order_id">
            <Hierarchy hasAll="true" primaryKey="order_id" primaryKeyTable="tblPvcDeclarationOrder">
                <Join leftKey="tbs_member_id" rightAlias="tbsmember" rightKey="id">
                    <Table name="tblPvcDeclarationOrder" />
                    <Join leftKey="is_full_member" rightKey="is_full_member">
                        <Table name="tblPvcTbsMember"  alias="tbsmember"/>
                        <Table name="viewPvcIsFullMember"/>
                    </Join>
                </Join>
                <Level name="Is Full Member" table="viewPvcIsFullMember" column="is_full_member" captionColumn="is_full_member_label">
                </Level>
            </Hierarchy>
        </Dimension>

        <Dimension name="Has Voting Rights" foreignKey="order_id">
            <Hierarchy hasAll="true" primaryKey="order_id" primaryKeyTable="tblPvcDeclarationOrder">
                <Join leftKey="tbs_member_id" rightAlias="voting" rightKey="id">
                    <Table name="tblPvcDeclarationOrder" />
                    <Join leftKey="has_voting_rights" rightKey="has_voting_rights">
                        <Table name="tblPvcTbsMember"  alias="voting"/>
                        <Table name="viewPvcHasVotingRights"/>
                    </Join>
                </Join>
                <Level name="Has Voting Rights" table="viewPvcHasVotingRights" column="has_voting_rights" captionColumn="has_voting_rights_label">
                </Level>
            </Hierarchy>
        </Dimension>

        <Measure name="Weight" column="weight" aggregator="sum" datatype="Numeric" formatString="#,##0.00" />
        <Measure name="Power" column="power" aggregator="sum" datatype="Numeric" formatString="#,##0.00" />
        <Measure name="Quantity" column="quantity" aggregator="sum" datatype="Integer" />
        <Measure name="Average quantity" column="quantity" aggregator="avg" datatype="Integer" />
        <Measure name="Amount (excl. vat)" column="line_amount_excluding_vat" aggregator="sum" datatype="Numeric" formatString="#,##0.00" />
        <Measure name="Amount (incl. vat)" column="line_amount_including_vat" aggregator="sum" datatype="Numeric" formatString="#,##0.00" />
        <Measure name="Price per unit" column="price_per_unit" aggregator="sum" datatype="Numeric" formatString="#,##0.00" />

        <Measure name="Number of members" column="member_id" aggregator="distinct-count" datatype="Integer" />

        <CalculatedMember
           name="Average weight"
           dimension="Measures"
           formula="[Measures].[Weight] / [Measures].[Quantity]"
           visible="true">
           <CalculatedMemberProperty name="DATATYPE" value="Numeric"/>
           <CalculatedMemberProperty name="SOLVE_ORDER" value="2"/>
           <CalculatedMemberProperty name="FORMAT_STRING" value="#,##0.00"/>
       </CalculatedMember>

        <CalculatedMember
           name="Average power"
           dimension="Measures"
           formula="[Measures].[Power] / [Measures].[Quantity]"
           visible="true">
           <CalculatedMemberProperty name="DATATYPE" value="Numeric"/>
           <CalculatedMemberProperty name="SOLVE_ORDER" value="2"/>
           <CalculatedMemberProperty name="FORMAT_STRING" value="#,##0.00"/>
       </CalculatedMember>

        <CalculatedMember
           name="Power (MW)"
           dimension="Measures"
           formula="[Measures].[Power] / 1000000"
           visible="true">
           <CalculatedMemberProperty name="DATATYPE" value="Numeric"/>
           <CalculatedMemberProperty name="SOLVE_ORDER" value="3"/>
           <CalculatedMemberProperty name="FORMAT_STRING" value="#,##0.00"/>
       </CalculatedMember>

       <CalculatedMember
           name="Average power (MW)"
           dimension="Measures"
           formula="[Measures].[Power (MW)] / [Measures].[Quantity]"
           visible="true">
           <CalculatedMemberProperty name="DATATYPE" value="Numeric"/>
           <CalculatedMemberProperty name="SOLVE_ORDER" value="4"/>
           <CalculatedMemberProperty name="FORMAT_STRING" value="#,##0.00"/>
       </CalculatedMember>

    </Cube>
        """ % {'name': self.name}

    xml = property(_xml)

class PvcOrderCube(OrderCube):
    name = 'PvcOrderCube'

    MERCHANT_LABEL = 'TBS'
    CLIENT_LABEL = 'Member'


class PvcAbstractDeclarationDataCubeRole(RoleBase):

    def get_role_list(self, account_id):
        """
        """

        logger.info("get_role_list: %s" % account_id)

        # check required permissions
        ac = AbstractController(self.sa_session, account_id=account_id, check_active=False)
        ac.assert_permissions('pvc_view_declarations')

        current_license = ac.get_default_group().license

        rc = IQubeRoleController(self.sa_session, account_id, self.ds, self.CUBE)
        rc.grant_cube(IQubeRole.ACCESS_ALL)

        # FIXME: WEEE Product levels are privileged objects!

        has_access = False

        if isinstance(current_license, PvcOperatorLicense):
            return rc.to_list()

        elif isinstance(current_license, PvcTbsLicense) or isinstance(current_license, PvcMemberLicense):

            tbss = []
            members = []
            producers = []

            for order in ac.get_privileged_object(PvcDeclarationOrder, priv=PvcDeclarationOrder.PRIV_VIEW):
                if order.merchant.id not in tbss:
                    rc.grant_member('[TBS].[%s]' % order.merchant.id, IQubeRole.ACCESS_ALL)
                    tbss.append(order.merchant.id)
                if order.client.id not in members:
                    rc.grant_member('[Member].[%s]' % order.client.id, IQubeRole.ACCESS_ALL)
                    members.append(order.client.id)
                if order.producer.member.id not in producers:
                    rc.grant_member('[Producer].[%s]' % order.producer.member.id, IQubeRole.ACCESS_ALL)
                    producers.append(order.producer.member.id)
                has_access = True

        else:
            raise PublicSecurityException(_('Invalid license type'))

        if has_access:
            rc.grant_hierarchy('[TBS]', IQubeRole.ACCESS_CUSTOM, IQubeRole.ROLLUP_POLICY_PARTIAL)
            rc.grant_hierarchy('[Member]', IQubeRole.ACCESS_CUSTOM, IQubeRole.ROLLUP_POLICY_PARTIAL)
            rc.grant_hierarchy('[Producer]', IQubeRole.ACCESS_CUSTOM, IQubeRole.ROLLUP_POLICY_PARTIAL)
        else:
            rc.grant_hierarchy('[TBS]', IQubeRole.ACCESS_NONE, IQubeRole.ROLLUP_POLICY_HIDDEN)
            rc.grant_hierarchy('[Member]', IQubeRole.ACCESS_NONE, IQubeRole.ROLLUP_POLICY_HIDDEN)
            rc.grant_hierarchy('[Producer]', IQubeRole.ACCESS_NONE, IQubeRole.ROLLUP_POLICY_HIDDEN)

        return rc.to_list()


class PvcDeclarationDataCubeRole(PvcAbstractDeclarationDataCubeRole):
    CUBE = 'PvcDeclarationDataCube'


class PvcOrderCubeRole(RoleBase):

    def get_role_list(self, account_id):
        """
        FIXME: haz to be optimized by splitting into cubes per domain
        """

        ac = AbstractController(self.sa_session, account_id=account_id, check_active=False)
        ac.assert_permissions('pvc_view_orders')

        has_order_access = False

        rc = IQubeRoleController(self.sa_session, account_id, self.ds, 'PvcOrderCube')
        rc.grant_cube(IQubeRole.ACCESS_ALL)

        clients = []
        merchants = []

        for order in ac.get_privileged_object(PvcOrder):
            if order.client.id not in clients:
                rc.grant_member('[Member].[%s]' % order.client.id, IQubeRole.ACCESS_ALL)
                clients.append(order.client.id)
            if order.merchant.id not in merchants:
                rc.grant_member('[TBS].[%s]' % order.merchant.id, IQubeRole.ACCESS_ALL)
                merchants.append(order.merchant.id)
            has_order_access = True

        if has_order_access:
            rc.grant_hierarchy('[Member]', IQubeRole.ACCESS_CUSTOM, IQubeRole.ROLLUP_POLICY_PARTIAL)
            rc.grant_hierarchy('[TBS]', IQubeRole.ACCESS_CUSTOM, IQubeRole.ROLLUP_POLICY_PARTIAL)
        else:
            rc.grant_hierarchy('[Member]', IQubeRole.ACCESS_NONE, IQubeRole.ROLLUP_POLICY_HIDDEN)
            rc.grant_hierarchy('[TBS]', IQubeRole.ACCESS_NONE, IQubeRole.ROLLUP_POLICY_HIDDEN)

        return rc.to_list()


class PvcContributionDataCube(CubeBase):
    name = 'PvcContributionDataCube'

    def _xml(self):
        return """
            <Cube name="%(name)s" defaultMeasure="Amount (excl. vat)">

                <Table name="aggPvcContribution" />

                <Dimension name="Order" foreignKey="order_id">
                    <Hierarchy hasAll="true" allMemberName="All Orders" primaryKey="id" primaryKeyTable="tblOrder">
                            <Table name="tblOrder"/>
                        <Level name="Order" table="tblOrder" column="id" type="Integer" />
                    </Hierarchy>
                </Dimension>

                <Dimension name="Order Line" foreignKey="order_line_id">
                    <Hierarchy hasAll="true" allMemberName="All Order Lines" primaryKey="id" primaryKeyTable="tblOrderLine">
                            <Table name="tblOrderLine"/>
                        <Level name="Order Line" table="tblOrderLine" column="id" type="Integer" />
                    </Hierarchy>
                </Dimension>

                <Dimension name="Order Type">
                    <Hierarchy hasAll="true" allMemberName="All Order Types">
                        <Level name="Order Type" column="order_type"/>
                    </Hierarchy>
                </Dimension>

                <Dimension name="Invoice Type">
                    <Hierarchy hasAll="true" allMemberName="All Invoice Types">
                        <Level name="Invoice Type" column="invoice_type" />
                    </Hierarchy>
                </Dimension>

                <Dimension name="TBS" foreignKey="tbs_id">
                    <Hierarchy hasAll="true" allMemberName="All TBSs" primaryKey="license_id" primaryKeyTable="tblPvcTbsLicense">
                        <Join leftKey="license_id" rightKey="id">
                            <Table name="tblPvcTbsLicense"/>
                            <Table name="tblLicense" alias="tbs" />
                        </Join>
                        <Level name="TBS" table="tbs" column="id" captionColumn="name" type="Integer" ordinalColumn="name" />
                    </Hierarchy>
                </Dimension>

                <Dimension name="Member" foreignKey="member_id">
                    <Hierarchy hasAll="true" allMemberName="All Members" primaryKey="license_id" primaryKeyTable="tblPvcMemberLicense">
                        <Join leftKey="license_id" rightKey="id">
                            <Table name="tblPvcMemberLicense"/>
                            <Table name="tblLicense" alias="member" />
                        </Join>
                        <Level name="Member" table="member" column="id" captionColumn="name" type="Integer" ordinalColumn="name"/>
                    </Hierarchy>
                </Dimension>

                <!-- TODO: this should only show is_producer=1 -->
                <Dimension name="Producer" foreignKey="producer_id">
                    <Hierarchy hasAll="true" allMemberName="All Producers" primaryKey="license_id" primaryKeyTable="tblPvcMemberLicense">
                        <Join leftKey="license_id" rightKey="id">
                            <Table name="tblPvcMemberLicense"/>
                            <Table name="tblLicense" alias="producer" />
                        </Join>
                        <Level name="Producer" table="producer" column="id" captionColumn="name" type="Integer" ordinalColumn="name"/>
                    </Hierarchy>
                </Dimension>

                <!-- TODO: this should only show is_commissioner=1 -->
                <Dimension name="Commissioner" foreignKey="commissioner_id">
                    <Hierarchy hasAll="true" allMemberName="All Commissioners" primaryKey="license_id" primaryKeyTable="tblPvcMemberLicense">
                        <Join leftKey="license_id" rightKey="id">
                            <Table name="tblPvcMemberLicense"/>
                            <Table name="tblLicense" alias="commissioner" />
                        </Join>
                        <Level name="Commissioner" table="commissioner" column="id" captionColumn="name" type="Integer" ordinalColumn="name"/>
                    </Hierarchy>
                </Dimension>

                <Dimension name="Country" foreignKey="country_id">
                    <Hierarchy hasAll="true" allMemberName="All Countries" primaryKey="id" primaryKeyTable="tblCountry">
                        <Table name="tblCountry" />
                        <Level name="Country" table="tblCountry" column="id" captionColumn="name" type="Integer" ordinalColumn="name"/>
                    </Hierarchy>
                </Dimension>

                <Dimension name="Declaration Year">
                    <Hierarchy hasAll="true" allMemberName="All Declaration Years">
                        <Level name="Year" column="year"/>
                    </Hierarchy>
                </Dimension>

                <Dimension name="Declaration Quarter">
                    <Hierarchy hasAll="true" allMemberName="All Declaration Quarters">
                        <Level name="Quarter" column="quarter">
                            <CaptionExpression>
                                <SQL dialect="mysql">
                                    CONCAT('Q', `aggPvcContribution`.`quarter`)
                                </SQL>
                            </CaptionExpression>
                        </Level>
                    </Hierarchy>
                </Dimension>

                <Dimension name="Declaration Month">
                    <Hierarchy hasAll="true" allMemberName="All Declaration Months">
                        <Level name="Month" column="month">
                            <CaptionExpression>
                                <SQL dialect="mysql">
                                    MONTHNAME(STR_TO_DATE(`aggPvcContribution`.`month`, '%%m'))
                                </SQL>
                            </CaptionExpression>
                        </Level>
                    </Hierarchy>
                </Dimension>

                <Dimension name="Submit Period" foreignKey="submitted_time_id" type="TimeDimension">
                    <Hierarchy name="Quarter" hasAll="true" primaryKey="id" >
                        <Table name="aggTimeQuarter"/>
                        <Level name="Year" column="year" type="Integer" uniqueMembers="true" levelType="TimeYears" ordinalColumn="year" />
                        <Level name="Quarter" column="quarter" uniqueMembers="false" levelType="TimeQuarters"  ordinalColumn="quarter" />
                        <Level name="Month" column="month" type="Integer" uniqueMembers="false" levelType="TimeMonths" ordinalColumn="month" />
                        <Level name="Day" column="day" type="Integer" uniqueMembers="false" levelType="TimeDays" ordinalColumn="day" />
                    </Hierarchy>
                    <Hierarchy name="WeekISO8601" hasAll="true" primaryKey="id" caption="Week">
                        <Table name="aggTimeWeekISO8601"/>
                        <Level name="Year" column="year" type="Integer" uniqueMembers="true" levelType="TimeYears" ordinalColumn="year" />
                        <Level name="Week" column="week" uniqueMembers="false" levelType="TimeWeeks" ordinalColumn="week" />
                        <Level name="Day" column="day" type="Integer" uniqueMembers="false" levelType="TimeDays" ordinalColumn="day" />
                    </Hierarchy>
                    <Hierarchy name="WeekUS" hasAll="true" primaryKey="id" caption="Week">
                        <Table name="aggTimeWeekUS"/>
                        <Level name="Year" column="year" type="Integer" uniqueMembers="true" levelType="TimeYears" ordinalColumn="year" />
                        <Level name="Week" column="week" uniqueMembers="false" levelType="TimeWeeks" ordinalColumn="week" />
                        <Level name="Day" column="day" type="Integer" uniqueMembers="false" levelType="TimeDays" ordinalColumn="day" />
                    </Hierarchy>
                </Dimension>

                <Dimension name="Payment Period" foreignKey="payment_time_id" type="TimeDimension">
                    <Hierarchy hasAll="true" primaryKey="id" name="Quarter">
                        <Table name="aggTimeQuarter"/>
                        <Level name="Year" column="year" type="Integer" uniqueMembers="true" levelType="TimeYears" ordinalColumn="year" />
                        <Level name="Quarter" column="quarter" uniqueMembers="false" levelType="TimeQuarters"  ordinalColumn="quarter" />
                        <Level name="Month" column="month" type="Integer" uniqueMembers="false" levelType="TimeMonths" ordinalColumn="month" />
                        <Level name="Day" column="day" type="Integer" uniqueMembers="false" levelType="TimeDays" ordinalColumn="day" />
                    </Hierarchy>
                    <Hierarchy name="WeekISO8601" hasAll="true" primaryKey="id" caption="Week">
                        <Table name="aggTimeWeekISO8601"/>
                        <Level name="Year" column="year" type="Integer" uniqueMembers="true" levelType="TimeYears" ordinalColumn="year" />
                        <Level name="Week" column="week" uniqueMembers="false" levelType="TimeWeeks" ordinalColumn="week" />
                        <Level name="Day" column="day" type="Integer" uniqueMembers="false" levelType="TimeDays" ordinalColumn="day" />
                    </Hierarchy>
                    <Hierarchy name="WeekUS" hasAll="true" primaryKey="id" caption="Week">
                        <Table name="aggTimeWeekUS"/>
                        <Level name="Year" column="year" type="Integer" uniqueMembers="true" levelType="TimeYears" ordinalColumn="year" />
                        <Level name="Week" column="week" uniqueMembers="false" levelType="TimeWeeks" ordinalColumn="week" />
                        <Level name="Day" column="day" type="Integer" uniqueMembers="false" levelType="TimeDays" ordinalColumn="day" />
                    </Hierarchy>
                </Dimension>

                <Dimension name="Payment Status">
                    <Hierarchy hasAll="true" allMemberName="All Payment Statuses">
                        <Level name="Payment Status" column="payment_status" type="Integer">
                            <CaptionExpression>
                                <SQL dialect="mysql">
                                    CASE `aggPvcContribution`.`payment_status`
                                        WHEN 0 THEN 'Pending'
                                        WHEN 1 THEN 'Submitted'
                                        WHEN 2 THEN 'Partial'
                                        WHEN 3 THEN 'Completed'
                                        WHEN 4 THEN 'Rejected'
                                        ELSE '-' END
                                </SQL>
                            </CaptionExpression>
                        </Level>
                    </Hierarchy>
                </Dimension>

                <Dimension name="Payment Method">
                    <Hierarchy hasAll="true" allMemberName="All Payment Methods">
                        <Level name="Payment Method" column="payment_method" />
                    </Hierarchy>
                </Dimension>

                <Dimension name="GL Account">
                    <Hierarchy hasAll="true" allMemberName="All GL Accounts">
                        <Level name="GL Account" column="gl_account" />
                    </Hierarchy>
                </Dimension>

                <Dimension name="Is Full Member" foreignKey="order_id">
                    <Hierarchy hasAll="true" primaryKey="order_id" primaryKeyTable="tblPvcDeclarationOrder">
                        <Join leftKey="tbs_member_id" rightAlias="tbsmember" rightKey="id">
                            <Table name="tblPvcDeclarationOrder" />
                            <Join leftKey="is_full_member" rightKey="is_full_member">
                                <Table name="tblPvcTbsMember"  alias="tbsmember"/>
                                <Table name="viewPvcIsFullMember"/>
                            </Join>
                        </Join>
                        <Level name="Is Full Member" table="viewPvcIsFullMember" column="is_full_member" captionColumn="is_full_member_label">
                        </Level>
                    </Hierarchy>
                </Dimension>

                <Dimension name="Has Voting Rights" foreignKey="order_id">
                    <Hierarchy hasAll="true" primaryKey="order_id" primaryKeyTable="tblPvcDeclarationOrder">
                        <Join leftKey="tbs_member_id" rightAlias="voting" rightKey="id">
                            <Table name="tblPvcDeclarationOrder" />
                            <Join leftKey="has_voting_rights" rightKey="has_voting_rights">
                                <Table name="tblPvcTbsMember"  alias="voting"/>
                                <Table name="viewPvcHasVotingRights"/>
                            </Join>
                        </Join>
                        <Level name="Has Voting Rights" table="viewPvcHasVotingRights" column="has_voting_rights" captionColumn="has_voting_rights_label">
                        </Level>
                    </Hierarchy>
                </Dimension>

                <Dimension name="Brand" foreignKey="brand_id">
                    <Hierarchy hasAll="true" allMemberName="All Brands" primaryKey="id" primaryKeyTable="tblPvcBrand">
                        <Table name="tblPvcBrand" />
                        <Level name="Brand" table="tblPvcBrand" column="id" captionColumn="name" type="Integer" ordinalColumn="name"/>
                    </Hierarchy>
                </Dimension>

                <Dimension name="Product" foreignKey="product_id">
                    <Hierarchy hasAll="true" primaryKey="id" allMemberName="All Products" primaryKeyTable="tblWEEEProductLevel">
                        <Table name="tblWEEEProductLevel" />
                        <Level name="WEEE Category" table="tblWEEEProductLevel" column="id" captionColumn="label" type="Integer" ordinalColumn="label"/>
                    </Hierarchy>
                </Dimension>

                <Dimension name="Local Product" foreignKey="local_product_id">
                    <Hierarchy hasAll="true" primaryKey="id" allMemberName="All Local Products" primaryKeyTable="tblPvcTbsProduct">
                        <Table name="tblPvcTbsProduct" />
                        <Level name="Product" table="tblPvcTbsProduct" column="id" captionColumn="label" type="Integer" ordinalColumn="label"/>
                    </Hierarchy>
                </Dimension>

                <Measure name="Weight" column="weight" aggregator="sum" datatype="Numeric" formatString="#,##0.00" />
                <Measure name="Power" column="nominal_power" aggregator="sum" datatype="Numeric" formatString="#,##0.00" />
                <Measure name="Surface" column="surface" aggregator="sum" datatype="Numeric" formatString="#,##0.00" />
                <Measure name="Quantity" column="quantity" aggregator="sum" datatype="Integer" />
                <Measure name="Average quantity" column="quantity" aggregator="avg" datatype="Integer" />
                <Measure name="Amount (excl. vat)" column="line_amount_excluding_vat" aggregator="sum" datatype="Numeric" formatString="#,##0.00" />
                <Measure name="Amount (incl. vat)" column="line_amount_including_vat" aggregator="sum" datatype="Numeric" formatString="#,##0.00" />
                <Measure name="Price per unit" column="price_per_unit" aggregator="sum" datatype="Numeric" formatString="#,##0.00" />

                <Measure name="Number of members" column="member_id" aggregator="distinct-count" datatype="Integer" />

                <CalculatedMember
                   name="Average weight"
                   dimension="Measures"
                   formula="[Measures].[Weight] / [Measures].[Quantity]"
                   visible="true">
                   <CalculatedMemberProperty name="DATATYPE" value="Numeric"/>
                   <CalculatedMemberProperty name="SOLVE_ORDER" value="2"/>
                   <CalculatedMemberProperty name="FORMAT_STRING" value="#,##0.00"/>
               </CalculatedMember>

                <CalculatedMember
                   name="Average power"
                   dimension="Measures"
                   formula="[Measures].[Power] / [Measures].[Quantity]"
                   visible="true">
                   <CalculatedMemberProperty name="DATATYPE" value="Numeric"/>
                   <CalculatedMemberProperty name="SOLVE_ORDER" value="2"/>
                   <CalculatedMemberProperty name="FORMAT_STRING" value="#,##0.00"/>
               </CalculatedMember>

                <CalculatedMember
                   name="Power (MW)"
                   dimension="Measures"
                   formula="[Measures].[Power] / 1000000"
                   visible="true">
                   <CalculatedMemberProperty name="DATATYPE" value="Numeric"/>
                   <CalculatedMemberProperty name="SOLVE_ORDER" value="3"/>
                   <CalculatedMemberProperty name="FORMAT_STRING" value="#,##0.00"/>
               </CalculatedMember>

               <CalculatedMember
                   name="Average power (MW)"
                   dimension="Measures"
                   formula="[Measures].[Power (MW)] / [Measures].[Quantity]"
                   visible="true">
                   <CalculatedMemberProperty name="DATATYPE" value="Numeric"/>
                   <CalculatedMemberProperty name="SOLVE_ORDER" value="4"/>
                   <CalculatedMemberProperty name="FORMAT_STRING" value="#,##0.00"/>
               </CalculatedMember>

            </Cube>
        """ % {'name': self.name}

    xml = property(_xml)


class PvcContributionDataCubeRole(PvcAbstractDeclarationDataCubeRole):
    CUBE = 'PvcContributionDataCube'


projects = (PvcOrderDataProject, )
