'''
Created on Oct 8, 2013

@author: marten
'''
import datetime
from basesoft.util.ordered_dict import OrderedDict
from django.utils.translation import ugettext_lazy as _
from django.utils.translation import ugettext
from sqlalchemy.orm.session import Session

from basesoft.web.contrib.notification.models import Notification
from basesoft.web.contrib.order.models import Order, OrderLine
from basesoft.web.contrib.mailing.models import Mail

def get_operator_and_accountant_groups(obj):
    from pvcycle.core.models import PvcOperatorLicense
    db = Session.object_session(obj)
    rv = []
    operators = db.query(PvcOperatorLicense).filter(PvcOperatorLicense.domain==obj.merchant.domain).all()
    for operator in operators:
        rv.append(operator.default_group)
    for ta in obj.merchant.tbs_accountants:
        rv.append(ta.accountant.default_group)

    return rv


class PvcOrder(Order):
    """
    @DynamicAttrs
    """
    auto_submit = True
    permissions = ('pvc_view_orders', )
    js_details_module = 'pvc_order_details'

    name_i18n = _('Administrative')

    def get_name_i18n(self):
        return ugettext('Administrative')

    def get_order_controller(self):
        from pvcycle.order.controllers import PvcOrderController
        return PvcOrderController

    def get_invoice_view_privileges(self):
        return Order.get_invoice_view_privileges(self) + get_operator_and_accountant_groups(self)


class PvcOrderLineAdministrationFee(OrderLine):
    """
    This is used for membership registration
    @DynamicAttrs
    """
    tax_class_code = 'TAX_ADMIN_FEE'
    gl_account_code = 'GL_ADMIN_FEE'

    BACKREF = 'tbs_member_service_level'

class PvcOrderLineMembershipFee(OrderLine):
    """
    This is used for membership registration in the UK
    @DynamicAttrs
    """
    tax_class_code = 'TAX_MEMBER_FEE'
    gl_account_code = 'GL_MEMBER_FEE'

    BACKREF = 'tbs_member_membership'

class PvcDeclarationOrder(Order):
    """
    @DynamicAttrs
    """
    RESUBMIT_REQUEST_PENDING = 1
    RESUBMIT_REQUEST_APPROVED = 2

    IS_NON_COMPLIANT = 0
    IS_COMPLIANT = 1

    readonly = True  # read only from PV CYCLE 2.0 onwards

    auto_submit = False
    permissions = ('pvc_view_declarations', )
    js_details_module = 'pvc_declaration_details'

    name_i18n = _('Contribution (legacy)')

    all_privileges = (Order.PRIV_VIEW, Order.PRIV_UPDATE, Order.PRIV_ARCHIVE, Order.PRIV_SUBMIT, Order.PRIV_PAY)
    producer_privileges = (Order.PRIV_VIEW, Order.PRIV_UPDATE, Order.PRIV_ARCHIVE, Order.PRIV_SUBMIT) # given to the Producer of the modules
    tbs_member_privileges = (Order.PRIV_VIEW, Order.PRIV_PAY) # Given to the license who pays for it, either same as Producer or Commissioner

    def get_order_controller(self):
        from pvcycle.order.controllers import PvcDeclarationOrderController
        return PvcDeclarationOrderController

    def get_name_i18n(self):
        return ugettext('Contribution (legacy)')

    def get_invoice_view_privileges(self):
        return Order.get_invoice_view_privileges(self) + get_operator_and_accountant_groups(self)


class PvcDeclarationOrderLine(OrderLine):
    """
    @DynamicAttrs
    """
    tax_class_code = 'TAX_CONTRIB_FEE'
    gl_account_code = 'GL_CONTRIB_FEE'
    MAX_LINE_QUANTITY = 1000 * 1000 # Ticket #319


class PvcDeclarationOrderLineMinimumCF(OrderLine):
    """
    @DynamicAttrs
    """
    tax_class_code = 'TAX_FORFEIT_FEE'
    gl_account_code = 'GL_FORFEIT_FEE'
    MAX_LINE_QUANTITY = 1000 * 1000 # Ticket #319


class PvcDeclarationReminderMail(Mail):
    PRIV_USE_FOR_SENDING = 64

    NAME_I18N = _("Declaration reminder mail")

    def get_privileges(self):
        return Mail.get_privileges(self) + [
            (self.PRIV_USE_FOR_SENDING, _('Use for sending'))
        ]

    @staticmethod
    def replacement_fields():
        x = Mail.replacement_fields()
        x.update({
            ugettext('Company'): {
                'company.name': ugettext('Name'),
            },
            ugettext('Declaration'): {
                'declaration.year': ugettext('Year'),
                'declaration.quarter': ugettext('Quarter'),
                'declaration.month': ugettext('Month'),
            },
        })
        return x

    def _replace_field(self, person, field_name, context):

        tokens = field_name.split('.')

        obj = None

        if tokens[0] == 'company':
            obj = context['license']
            assert tokens[1] in ['name']

        if tokens[0] == 'declaration':
            obj = context['declaration']
            assert tokens[1] in ['year', 'month', 'quarter']

        if obj is not None:
            x = getattr(obj, tokens[1])
            if x is None: x = '-'
            return unicode(x)

        return Mail._replace_field(self, person, field_name, context)


class PvcDeclarationNotification(Notification):
    """
    @DynamicAttrs
    """

    name_i18n = _('Declaration')

    def __repr__(self):
        return '<PvcDeclarationNotification id=%s>' % self.id


class PvcContributionOrder(Order):
    """
    @DynamicAttrs
    """

    RESUBMIT_REQUEST_PENDING = 1
    RESUBMIT_REQUEST_APPROVED = 2

    IS_NON_COMPLIANT = 0
    IS_COMPLIANT = 1

    auto_submit = False
    permissions = ('pvc_view_declarations', )
    js_details_module = 'pvc_contribution_details'

    name_i18n = _('Contribution')

    all_privileges = (Order.PRIV_VIEW, Order.PRIV_UPDATE, Order.PRIV_ARCHIVE, Order.PRIV_SUBMIT, Order.PRIV_PAY)
    producer_privileges = (Order.PRIV_VIEW, Order.PRIV_UPDATE, Order.PRIV_ARCHIVE, Order.PRIV_SUBMIT) # given to the Producer of the modules
    tbs_member_privileges = (Order.PRIV_VIEW, Order.PRIV_PAY) # Given to the license who pays for it, either same as Producer or Commissioner

    def get_order_controller(self):
        from pvcycle.order.controllers import PvcContributionOrderController
        return PvcContributionOrderController

    def get_name_i18n(self):
        return ugettext('Contribution')

    def get_period(self):
        month = self.month
        if month is None:
            if self.quarter is None:
                month = 1
            else:
                month = (self.quarter-1) * 3 + 1

        return datetime.date(int(self.year), int(month), 1)

    def get_submit_dry_run_clone_args(self):
        rv = super(PvcContributionOrder, self).get_submit_dry_run_clone_args()
        rv['reference_children'].append('tbs_member')
        rv['reference_children'].append('producer')
        rv['clone_children']['lines']['reference_children'].append('tbs_member_product')
        rv['clone_children']['lines']['reference_children'].append('brand')

        # FIXME: temporary hack
        rv['clone_children']['lines']['reference_children'].append('custom_unit')
        rv['clone_children']['lines']['reference_children'].append('weight')
        rv['clone_children']['lines']['reference_children'].append('is_b2b')

        return rv

    @staticmethod
    def get_simulate_fields():

        from pvcycle.order.rule import conditions

        # notice the order: tbs_member must be executed before producer
        return [
            conditions.PvcRuleConditionTbs,
            conditions.PvcRuleConditionContributionTbsMember,
            conditions.PvcRuleConditionContributionProducer,
            conditions.PvcRuleConditionContributionYear,
            conditions.PvcRuleConditionContributionMonth,
            conditions.PvcRuleConditionContributionQuarter,
            conditions.PvcRuleConditionContributionIsForecast,
        ]

    @staticmethod
    def get_simulate_order_line_types():
        return [
            ('Contribution line', 'PvcContributionOrderLine'),
        ]

    def get_invoice_view_privileges(self):
        return Order.get_invoice_view_privileges(self) + get_operator_and_accountant_groups(self)


class PvcContributionOrderLine(OrderLine):
    """
    @DynamicAttrs
    """
    tax_class_code = 'TAX_CONTRIB_FEE'
    gl_account_code = 'GL_CONTRIB_FEE'
    gl_forecast_account_code = 'GL_FORECAST_FEE'
    MAX_LINE_QUANTITY = 1000 * 1000  # Ticket #319
    MAX_LINE_DECIMALS = 1000 * 1000 * 1000  # Ticket #911, Weight & Custom Unit
    CREDIT_FIELDS = ('quantity', 'weight', 'custom_unit')

    END_USER_IS_B2B = 1
    END_USER_IS_B2C = 2

    def get_sort(self):
        return lambda k: (k.tbs_member_product.tbs_product.label, None if k.brand is None else k.brand.name, k.is_b2b)

    @staticmethod
    def get_simulate_fields():

        from pvcycle.order.rule import conditions
        rv = OrderLine.get_simulate_fields()

        rv += [
            conditions.PvcRuleConditionContributionLineTbsMemberProduct,
            conditions.PvcRuleConditionContributionLineWeight,
            conditions.PvcRuleConditionContributionLineCustomUnit,
            conditions.PvcRuleConditionContributionLineIsB2b,
            conditions.PvcRuleConditionContributionLineBrand,
        ]

        return rv



