from decimal import Decimal

from basesoft.web.contrib.order.controllers import TaxController
from basesoft.web.contrib.order.models import OrderRuleAction, OrderLineRule
from basesoft.web.contrib.order.rule.actions import RuleActionController
from django.utils.translation import ugettext_lazy
from pvcycle.order.models import PvcContributionOrderLine
from pvcycle.product.models import PvcTbsProductFee


class PvcRuleActionApplyMemberCountryVatRate(RuleActionController):

    name = 'pvc_apply_member_country_vat_rate'
    label = ugettext_lazy('Apply member country VAT rate')
    type = OrderRuleAction.TYPE_BOOLEAN

    def execute(self, order, rule, action, dry_run):
        if not action.value_bool:
            return # we do nothing

        country = order.client.default_group.license.company.addresses[0].address.country
        tax_controller = TaxController(self._order_controller.sa_session, request=self._order_controller.request)

        for line in order.lines:
            tax_class_code = line.tax_class_code
            if line.tax_class_code is None:  # we cant allocate a tax rate if we do not have a tax class code
                continue
            tax_rate = tax_controller.get_current_rate(country, tax_class_code, order.merchant.default_group)
            multiplier = 1 + (tax_rate / 100)

            line.vat_rate = tax_rate
            line.line_amount_including_vat = line.line_amount_excluding_vat * multiplier

        order.add_log('New VAT rate applied to order: %s' % tax_rate)


class PvcRuleActionContributionB2bPercentage(RuleActionController):
    """
    This action will automatically split Pvc Contribution Lines into b2b/b2c based on the supplied b2b percentage.
     * B2B quantities are rounded down.
     * B2C quantities are rounded up.
    """

    name = 'pvc_contribution_b2b_percentage'
    label = ugettext_lazy('B2B percentage of total')
    type = OrderRuleAction.TYPE_INTEGER

    def execute(self, order, rule, action, dry_run):
        if action.value_int == 0:
            return  # we do nothing TODO: can action values be lower than 0?

        assert order.merchant.get_config('pvc.tbs.contribution.automatic.end.user.distinction')

        lines = []

        oc = self._order_controller
        b2b_multiplier = action.value_int/Decimal('100.0')
        index = len(order.lines)

        service_level_id = order.tbs_member.active_service_level.service_level.id

        period = order.get_period()
        period = '%s-%02d-01' % (period.year, period.month)

        _fee_map = {}
        for line in order.lines:
            if not isinstance(line, PvcContributionOrderLine):
                continue
            _fee_map[line.tbs_member_product.tbs_product.id] = line.tbs_member_product.tbs_product.get_fee(period, service_level_id, PvcTbsProductFee.DIMENSION_IS_B2B)

        for line in order.lines:
            if not isinstance(line, PvcContributionOrderLine):
                lines.append(line)
                continue

            # the lines should have already been set to b2c
            assert line.is_b2b is False
            b2b_weight = None
            b2b_custom_unit = None
            b2b_quantity = int(round(line.quantity*b2b_multiplier, 0))
            line.quantity -= b2b_quantity
            if line.weight is not None:
                b2b_weight = line.weight*b2b_multiplier
                line.weight -= b2b_weight
            if line.custom_unit is not None:
                b2b_custom_unit = line.custom_unit*b2b_multiplier
                line.custom_unit -= b2b_custom_unit

            # we alter the existing line (b2c)
            lines.append(oc.recalculate_line(order, line))

            # we add a new line
            newline = line.clone()
            newline.is_b2b = True
            newline.quantity = b2b_quantity
            newline.weight = b2b_weight
            newline.custom_unit = b2b_custom_unit
            newline.price_per_unit = _fee_map[newline.tbs_member_product.tbs_product.id]
            newline.index = index
            lines.append(oc.recalculate_line(order, newline))
            index += 1

        order.lines = lines

        order.add_log('Order split in b2b (%d%%) vs. b2c (rest).' % action.value_int)
        oc._reindex_order_lines(order)
