from decimal import Decimal
from types import NoneType
from django.utils.translation import ugettext as _
from basesoft.web.contrib.order.models import OrderRuleCondition
from basesoft.web.contrib.order.rule.conditions import RuleConditionController
from basesoft.web.core.util.controllers import lazy_assert
from django.utils.translation import ugettext_lazy
from pvcycle.core.controllers import PvcTBSController, PvcMemberController
from pvcycle.core.models import PvcTbsLicense, PvcTbsMember, PvcMemberLicense
from pvcycle.order.models import PvcContributionOrder
from pvcycle.product.models import PvcTbsProduct, PvcTbsMemberProduct, PvcBrand


class PvcRuleConditionContributionIsCompliant(RuleConditionController):
    name = 'is_compliant'
    label = ugettext_lazy('Is compliant')

    type = OrderRuleCondition.TYPE_INTEGER

    show_select = True

    def matches(self, order, rule, condition):
        assert condition.value_int in (PvcContributionOrder.IS_NON_COMPLIANT, PvcContributionOrder.IS_COMPLIANT)
        return order.is_compliant == condition.value_int

    def get_select_options(self, form_values):
        return [
            {'label': _('Is non compliant'), 'value': PvcContributionOrder.IS_NON_COMPLIANT},
            {'label': _('Is compliant'), 'value': PvcContributionOrder.IS_COMPLIANT},
        ]


class PvcRuleConditionTbs(RuleConditionController):

    name = 'merchant'
    label = ugettext_lazy('TBS')

    type = OrderRuleCondition.TYPE_INTEGER

    show_select = True

    def matches(self, order, rule, condition):
        return order.merchant.id == condition.value_int

    def get_select_options(self, form_values):
        controller = PvcTBSController(self._order_controller.sa_session, account_id=self._order_controller.account_id, request=self._order_controller.request)
        return [{'value': t.id, 'label': t.name} for t in controller.get_take_back_schemes()]

    def simulate_to_object(self, tbs_id, order_state):
        lazy_assert(tbs_id, int)

        # FIXME: use controller?
        tbs = self._order_controller.sa_session.query(PvcTbsLicense).filter(PvcTbsLicense.id == tbs_id).one()
        return tbs

class PvcRuleConditionMember(RuleConditionController):
    name = 'client'
    label = ugettext_lazy('Member')

    type = OrderRuleCondition.TYPE_INTEGER

    show_select = True

    def matches(self, order, rule, condition):
        return order.client.id == condition.value_int

    def get_select_options(self, form_values):
        return [{'value': t.id, 'label': t.name} for t in self._order_controller._get_member_licenses()]

    def simulate_to_object(self, tbs_id, order_state):
        lazy_assert(tbs_id, int)

        # FIXME: use controller?
        tbs = self._order_controller.sa_session.query(PvcMemberLicense).filter(PvcMemberLicense.id == tbs_id).one()
        return tbs

class PvcRuleConditionContributionYear(RuleConditionController):
    name = 'year'
    label = ugettext_lazy('Year')

    type = OrderRuleCondition.TYPE_INTEGER

    show_select = True

    def matches(self, order, rule, condition):
        assert 2000 < condition.value_int < 2100, 'got: %s' % condition.value_int
        return order.year == condition.value_int

    def get_select_options(self, form_values):
        return [{'label': x, 'value': x} for x in range(2000, 2100)]


class PvcRuleConditionContributionQuarter(RuleConditionController):
    name = 'quarter'
    label = ugettext_lazy('Quarter')
    type = OrderRuleCondition.TYPE_INTEGER
    show_select = True
    simulation_required = False

    def matches(self, order, rule, condition):
        assert 1 <= condition.value_int <= 4
        return order.quarter == condition.value_int

    def get_select_options(self, form_values):
        return [{'label': x, 'value': x} for x in range(1, 5)]


class PvcRuleConditionContributionMonth(RuleConditionController):
    name = 'month'
    label = ugettext_lazy('Month')
    type = OrderRuleCondition.TYPE_INTEGER
    show_select = True
    simulation_required = False

    def matches(self, order, rule, condition):
        assert 1 <= condition.value_int <= 12
        return order.month == condition.value_int

    def get_select_options(self, form_values):
        return [{'label': x, 'value': x} for x in range(1, 13)]


class PvcRuleConditionContributionTbsMember(RuleConditionController):

    name = 'tbs_member'
    label = ugettext_lazy('TbsMember')

    type = OrderRuleCondition.TYPE_INTEGER

    show_select = True

    def matches(self, order, rule, condition):
        return order.tbs_member.id == condition.value_int

    def get_select_options(self, form_values):

        if form_values.get('merchant', '') == '':
            return []

        members = self._order_controller.sa_session.query(PvcTbsMember) \
            .join(PvcMemberLicense) \
            .filter(PvcTbsMember.tbs_id == form_values['merchant']) \
            .order_by(PvcMemberLicense.name)

        return [{'label': x.member.name, 'value': x.id} for x in members]

    def simulate_to_object(self, tbs_member_id, order_state):

        if tbs_member_id is None:
            # tbs_member selection can remain None, then take the producer
            tbs_member = self._order_controller.sa_session.query(PvcTbsMember) \
                .filter(PvcTbsMember.id == order_state['producer']).one()
        else:
            # else subrogation
            tbs_member = self._order_controller.sa_session.query(PvcTbsMember) \
                .filter(PvcTbsMember.id == tbs_member_id).one()

        return tbs_member


class PvcRuleConditionContributionProducer(RuleConditionController):
    name = 'producer'
    label = ugettext_lazy('Producer')
    type = OrderRuleCondition.TYPE_INTEGER
    show_select = True

    def matches(self, order, rule, condition):
        return order.producer.id == condition.value_int

    def get_select_options(self, form_values):

        if form_values.get('merchant', '') == '':
            return []

        members = self._order_controller.sa_session.query(PvcTbsMember) \
            .join(PvcMemberLicense) \
            .filter(PvcTbsMember.tbs_id == form_values['merchant']) \
            .order_by(PvcMemberLicense.name)

        return [{'label': x.member.name, 'value': x.id} for x in members]

    def simulate_to_object(self, producer_id, order_state):
        if producer_id in ('', None):
            return None

        lazy_assert(producer_id, int)

        tbs_member = self._order_controller.sa_session.query(PvcTbsMember) \
            .filter(PvcTbsMember.id == producer_id).one()

        return tbs_member


class PvcRuleConditionIsFreeRider(RuleConditionController):
    name = 'is_free_rider'
    label = ugettext_lazy('Is free rider')

    type = OrderRuleCondition.TYPE_BOOLEAN

    def matches(self, order, rule, condition):
        if isinstance(order, PvcContributionOrder):
            return order.producer.is_free_rider == condition.value_bool

        tm = self._order_controller._get_tbs_member(order.merchant.id, order.client.id)
        return tm.is_free_rider == condition.value_bool

class PvcRuleConditionContributionIsForecast(RuleConditionController):

    name = 'is_forecast'
    label = ugettext_lazy('Is forecast')
    type = OrderRuleCondition.TYPE_INTEGER

    show_select = True

    def matches(self, order, rule, condition):
        assert condition.value_int in (0, 1)
        return order.is_forecast == condition.value_int

    def get_select_options(self, form_values):
        return [
            {'label': _('Is a forecast'), 'value': 1},
            {'label': _('Is not a forecast'), 'value': 0},
        ]

#
# LINE CONDITIONS
#

class PvcRuleConditionContributionLineTbsMemberProduct(RuleConditionController):

    name = 'tbs_member_product'
    label = ugettext_lazy('Line TbsMemberProduct')
    type = OrderRuleCondition.TYPE_INTEGER
    show_select = True

    def simulate_to_object(self, tbs_member_product_id, order_state):
        lazy_assert(tbs_member_product_id, int)

        tbs_member_product = self._order_controller.sa_session.query(PvcTbsMemberProduct) \
            .filter(PvcTbsMemberProduct.id == tbs_member_product_id).one()

        return tbs_member_product

    def matches(self, order, rule, condition):
        lazy_assert(condition.value_int, int)

        tbs_member_product = self._order_controller.sa_session.query(PvcTbsMemberProduct) \
            .filter(PvcTbsMemberProduct.id == condition.value_int).one()

        for line in order.lines:
            if line.tbs_member_product == tbs_member_product:
                return True
        return False

    def get_select_options(self, form_values):

        if not form_values.has_key('tbs_member') or form_values['tbs_member'] == '':
            return []

        tmps = self._order_controller.sa_session.query(PvcTbsMemberProduct).filter(
            PvcTbsMemberProduct.tbs_member_id == form_values['tbs_member']
        )
        return [{
            'label': x.tbs_product.label,
            'value': x.id
        } for x in tmps]




class PvcRuleConditionContributionLineWeight(RuleConditionController):
    name = 'weight'
    label = ugettext_lazy('Line weight')
    type = OrderRuleCondition.TYPE_DECIMAL
    show_operators = True

    def matches(self, order, rule, condition):
        lazy_assert(condition.value_decimal, Decimal)

        for line in order.lines:
            if self._rule_condition_operator(line.weight, condition.operator, condition.value_decimal):
                return True
        return False

class PvcRuleConditionContributionLineCustomUnit(RuleConditionController):
    name = 'custom_unit'
    label = ugettext_lazy('Line custom unit')
    type = OrderRuleCondition.TYPE_DECIMAL
    show_operators = True

    def matches(self, order, rule, condition):
        lazy_assert(condition.value_decimal, Decimal)
        for line in order.lines:
            if self._rule_condition_operator(line.custom_unit, condition.operator, condition.value_decimal):
                return True
        return False


class PvcRuleConditionContributionLineIsB2b(RuleConditionController):
    name = 'is_b2b'
    label = ugettext_lazy('Line is b2b')
    type = OrderRuleCondition.TYPE_BOOLEAN

    def matches(self, order, rule, condition):
        lazy_assert(condition.value_decimal, Decimal)
        for line in order.lines:
            if line.is_b2b == condition.value_bool:
                return True
        return False


class PvcRuleConditionContributionIsSubrogated(RuleConditionController):
    name = 'is_subrogated'
    label = ugettext_lazy('Is subrogated')
    type = OrderRuleCondition.TYPE_BOOLEAN

    def matches(self, order, rule, condition):
        if condition.value_bool:
            return order.producer != order.tbs_member
        return order.producer == order.tbs_member


class PvcRuleConditionContributionLineBrand(RuleConditionController):
    name = 'brand'
    label = ugettext_lazy('Line brand')
    type = OrderRuleCondition.TYPE_INTEGER
    show_select = True
    simulation_required = False

    def get_select_options(self, form_values):

        brands = self._order_controller.sa_session.query(PvcBrand).order_by(PvcBrand.name)
        return [{
            'label': x.name,
            'value': x.id
        } for x in brands]


    def simulate_to_object(self, brand_id, order_state):

        if brand_id == '': brand_id = None

        lazy_assert(brand_id, (int, NoneType))

        if brand_id is None:
            return None

        brand = self._order_controller.sa_session.query(PvcBrand) \
            .filter(PvcBrand.id == brand_id).one()

        return brand

    def matches(self, order, rule, condition):
        lazy_assert(condition.value_int, int)

        brand = self._order_controller.sa_session.query(PvcBrand) \
            .filter(PvcBrand.id == condition.value_int).one()

        for line in order.lines:
            if line.brand == brand:
                return True
        return False

