# Copyright (c) 2015 BaseSoft B.V.
# See LICENSE for details.
from types import NoneType
import datetime
from basesoft.web.contrib.i18n.models import Country
from pvcycle.subrogation.models import PvcSubrogation

from sqlalchemy import and_, func

from django.utils.translation import ugettext as _

from basesoft.util.virtual_date import today
from basesoft.web.core.exceptions import PublicSecurityException, PublicUserException, SystemException
from basesoft.web.core.util.controllers import transaction_decorator, lazy_assert

from iqportal.weee.core.controllers import WEEEController
from iqportal.weee.core.models import WEEEProductLevelGroup, WEEEProductLevel

from pvcycle.core.models import PvcTbsMember, PvcTbsServiceLevel
from pvcycle.order.models import PvcContributionOrderLine
from pvcycle.core.controllers import AbstractPvcController
from pvcycle.product.models import PvcTbsProduct, PvcTbsMemberProduct, PvcTbsProductFee, PvcBrand, \
    PvcTbsMemberProductBrand


class PvcProductController(WEEEController, AbstractPvcController):

    def get_available_products_for_tbs(self, tbs_id, parent_level_id=None, recursive=False, include_parent=False):
        """
        This method retrieves all WEEE Categories for possible availability for the provided TBS.
        The WEEE product levels are privileged objects, default world WEEE categories are always available.
        PV CYCLE added additional Categories are saved per domain.
        :return list of dictionaries
        """
        rv = []

        _filter, fo = self.get_query_filter_on_privileges(WEEEProductLevel, WEEEProductLevelGroup, WEEEProductLevel.PRIV_VIEW)
        if tbs_id is not None:
            fo = fo.outerjoin(PvcTbsProduct, and_(WEEEProductLevel.id == PvcTbsProduct.product_id, PvcTbsProduct.tbs_id == tbs_id))
            q = self.sa_session.query(WEEEProductLevel, PvcTbsProduct)
        else:
            q = self.sa_session.query(WEEEProductLevel)

        if include_parent:
            _filter = and_(_filter, WEEEProductLevel.id == parent_level_id)
        else:
            _filter = and_(_filter, WEEEProductLevel.parent_level_id == parent_level_id)

        q = q.select_from(fo).filter(_filter).order_by(WEEEProductLevel.parent_level_id, WEEEProductLevel.position)

        for product in q.all():
            if isinstance(product, WEEEProductLevel):
                tbs_product = None
            else:
                product, tbs_product = product

            children = []
            if recursive:
                children = self.get_available_products_for_tbs(tbs_id, product.id, True)

            lvl = {
                'id': product.id,
                'position': product.position,
                'description': product.description if tbs_product is None else tbs_product.description,
                'label': product.label if tbs_product is None else tbs_product.label,
                'default_description': product.description,
                'default_label': product.label,
                'status': None if tbs_product is None else tbs_product.status,
                'has_sub_levels': len(product.sublevels) > 0,
                'basis_for_invoicing': None if tbs_product is None else tbs_product.basis_for_invoicing,
                'custom_unit': None if tbs_product is None else tbs_product.custom_unit,
                'product_type': PvcTbsProduct.TYPE_OTHER if tbs_product is None else tbs_product.product_type,
                'children': children,
            }
            rv.append(lvl)

        return rv

    def get_products_for_tbs(self, tbs_id, product_id=None):
        _filter, fo = self.get_query_filter_on_privileges(WEEEProductLevel, WEEEProductLevelGroup, WEEEProductLevel.PRIV_VIEW)
        fo = fo.outerjoin(PvcTbsProduct, and_(WEEEProductLevel.id == PvcTbsProduct.product_id, PvcTbsProduct.tbs_id == tbs_id))
        q = self.sa_session.query(WEEEProductLevel, PvcTbsProduct)

        if product_id is not None:
            _filter = and_(_filter, WEEEProductLevel.id == product_id)

        q = q.select_from(fo).filter(_filter)
        rv = []

        for product, tbs_product in q.all():

            label = product.label if tbs_product is None else tbs_product.label
            description = product.description if tbs_product is None else tbs_product.description

            rv.append({
                'id': product.id,
                'concat_label': '%s - %s' % (label, description),
                'label': label,
                'description': description,
                'basis_for_invoicing': None if tbs_product is None else tbs_product.basis_for_invoicing,
            })

        return rv

    def get_tbs_product(self, _id):
        lazy_assert(_id, int)
        tbss = [t.id for t in self._get_take_back_schemes(permissions='manage_pvc_product_levels')]
        return self.sa_session.query(PvcTbsProduct).filter(and_(PvcTbsProduct.id == _id, PvcTbsProduct.tbs_id.in_(tbss))).first()

    def get_tbs_product_by_weee_product(self, product_id, tbs_id):
        return self.sa_session.query(PvcTbsProduct).filter(and_(
            PvcTbsProduct.tbs_id == tbs_id, PvcTbsProduct.product_id == product_id)).first()

    def get_declarable_tbs_products(self, tbs_id):
        return self.sa_session.query(PvcTbsProduct).filter(and_(
            PvcTbsProduct.tbs_id == tbs_id,
            PvcTbsProduct.status == PvcTbsProduct.STATUS_USE_IN_DECLARATION
        )).all()

    @transaction_decorator
    def save_or_update_tbs_product(self, _id, label, description, tbs_id, parent_level_id=None, is_active=None,
                                   product_id=None, basis_for_invoicing=None, custom_unit=None,
                                   weee_label=None, weee_description=None, product_type=PvcTbsProduct.TYPE_OTHER):
        self.assert_permissions('manage_pvc_product_levels')

        lazy_assert(custom_unit, (int, NoneType))
        lazy_assert(product_type, int)

        weee_label = label if weee_label is None else weee_label
        weee_description = label if weee_description is None else weee_description

        if product_id is not None:
            product = self.get_product_level(product_id).pop()
        else:
            product = self.add_product_level(weee_label, weee_description, parent_level_id)

        if product.parent_level is not None:
            self.get_or_save_parent_product(product.parent_level.id, tbs_id)

        if basis_for_invoicing is None:
            basis_for_invoicing = PvcTbsProduct.INVOICE_BASED_ON_WEIGHT

        if basis_for_invoicing == PvcTbsProduct.INVOICE_BASED_ON_CUSTOM_UNIT and custom_unit is None:
            raise PublicSecurityException(_('Custom unit must be selected if used for invoicing basis'))

        if _id is not None:
            tp = self.get_tbs_product(_id)
            if tp is None:
                raise PublicSecurityException(_('You do not have access to the provided product.'))
        else:
            tp = self.get_tbs_product_by_weee_product(product_id, tbs_id)
            if tp is None:
                tp = PvcTbsProduct()

        tp.tbs = self._get_take_back_schemes(tbs_id, 'manage_pvc_product_levels').pop()
        tp.product = product
        tp.label = label
        tp.description = description
        tp.status = PvcTbsProduct.STATUS_USE_IN_DECLARATION if is_active is None else is_active
        tp.basis_for_invoicing = basis_for_invoicing
        tp.custom_unit = custom_unit
        tp.product_type = product_type
        self.sa_session.add(tp)
        return tp

    @transaction_decorator
    def get_or_save_parent_product(self, parent_level_id, tbs_id):
        """
        This method is called when a product is activated on TBS level.
        It ensures that the parent level is also activated for FEE management.
        :param parent_level_id:
        :param tbs_id:
        :return:
        """
        tp = self.get_tbs_product_by_weee_product(parent_level_id, tbs_id)
        if tp is not None:
            return tp

        product = self.get_product_level(parent_level_id).pop()
        tp = self.save_or_update_tbs_product(None, product.label, product.description, tbs_id,
                                        is_active=PvcTbsProduct.STATUS_INACTIVE, product_id=product.id)

        return tp

    @transaction_decorator
    def save_products_for_tbs(self, tbs_id, products):
        self.assert_permissions('manage_pvc_product_levels')

        for product_id, values in products.iteritems():
            values['tbs_id'] = tbs_id
            values['product_id'] = product_id
            self.save_or_update_tbs_product(None, **values)

        return True

    @transaction_decorator
    def get_or_save_tbs_member_product(self, tbs_member, tbs_product_id, is_active=None, subrogation=None, country_id=None, brands=None):

        if not isinstance(tbs_member, PvcTbsMember):
            tbs_member = self._get_tbs_member_licenses(tbs_member).pop()

        tbs_product = self.sa_session.query(PvcTbsProduct).filter(PvcTbsProduct.id == tbs_product_id).one()
        assert tbs_member.tbs == tbs_product.tbs

        if country_id is None:
            if tbs_member.tbs.country is None:
                raise PublicUserException(_('Please provide at least one country where this product is put on the market.'))
            country = tbs_member.tbs.country
        else:
            country = self.sa_session.query(Country).filter(Country.id == country_id).one()

        if subrogation is not None:
            assert subrogation.subrogator is tbs_member

        mp = self.sa_session.query(PvcTbsMemberProduct).filter(and_(
            PvcTbsMemberProduct.tbs_product == tbs_product,
            PvcTbsMemberProduct.tbs_member == tbs_member,
            PvcTbsMemberProduct.subrogation == subrogation,
            PvcTbsMemberProduct.country == country,
        )).first()

        if mp is None:

            mp = PvcTbsMemberProduct()
            mp.is_active = True if is_active is None else is_active
            mp.tbs_product = tbs_product
            mp.tbs_member = tbs_member
            mp.subrogation = subrogation
            mp.country = country
            self.sa_session.add(mp)

        if brands is not None:
            for b in brands:
                self.get_or_save_brand(b, mp)

        return mp

    @transaction_decorator
    def save_contribution_fee(self, tbs_product_id, service_level_id, fee, start_date, dimension=None):
        lazy_assert(tbs_product_id, int)
        lazy_assert(service_level_id, int)
        lazy_assert(start_date, datetime.date)

        self.assert_permissions('manage_pvc_product_levels')

        if dimension is None:
            dimension = PvcTbsProductFee.DIMENSION_NONE

        tbs_product = self.get_tbs_product(tbs_product_id)
        service_level = self.sa_session.query(PvcTbsServiceLevel).filter(PvcTbsServiceLevel.id == service_level_id).first()

        # check if there is a fee set for the current tbs_product_id, dimension and start_date
        pf = self.sa_session.query(PvcTbsProductFee).filter(and_(
            PvcTbsProductFee.tbs_product == tbs_product,
            PvcTbsProductFee.service_level == service_level,
            PvcTbsProductFee.start_date == start_date,
            PvcTbsProductFee.dimension == dimension,
        )).first()

        if pf is None:
            pf = PvcTbsProductFee()
            pf.tbs_product = tbs_product
            pf.service_level = service_level
            pf.start_date = start_date
            pf.dimension = dimension
            self.logger.info('PvcProductController: setting fee `%s` for product "%s", service_level "%s" and period "%s"' % (fee, tbs_product, service_level, start_date))
        else:
            self.logger.info('PvcProductController: updating fee `%s` for product "%s", service_level "%s" and period "%s"' % (fee, tbs_product, service_level, start_date))

        pf.fee = fee
        pf.process_status = PvcTbsProductFee.STATUS_PROCESS
        self.sa_session.add(pf)

        return pf

    def get_brand(self, _id):
        return self.sa_session.query(PvcBrand).get(_id)

    def get_brands_by_name(self, _like, exact_match=False):
        if exact_match:
            return self.sa_session.query(PvcBrand).filter(PvcBrand.name == _like).first()
        return self.sa_session.query(PvcBrand).filter(PvcBrand.name.ilike('%%%s%%' % _like)).all()

    @transaction_decorator
    def update_brand_name(self, _id, name):
        self.assert_permissions('manage_pvc_brands')
        name = name.strip()

        brand = self.get_brand(_id)
        b = self.get_brands_by_name(name, True)
        if b is not None:
            if brand != b:
                raise PublicUserException(_('A brand with the name %(brand)s already exists, please use the Merge option.'))

        brand = self.get_brand(_id)
        brand.name = name
        self.sa_session.add(brand)
        return brand

    @transaction_decorator
    def get_or_save_brand(self, name, tbs_member_product):

        assert type(name) in (str, unicode, )
        name = name.strip()
        if name == '':
            raise PublicUserException(_('Cannot add a brand with an empty name.'))

        brand = self._get_or_save_brand(name)
        mbrand = self._get_tbs_member_product_brand(brand, tbs_member_product)
        if mbrand is not None:
            # if we find an existing member brand for this product, we always reactivate it (#887).
            mbrand.is_active = True
            self.sa_session.add(mbrand)
            return brand

        mbrand = PvcTbsMemberProductBrand()
        mbrand.brand = brand
        mbrand.tbs_member_product = tbs_member_product
        self.sa_session.add(mbrand)

        return brand

    @transaction_decorator
    def _get_or_save_brand(self, name):
        brand = self.sa_session.query(PvcBrand).filter(PvcBrand.name.ilike(name)).first()
        if brand is None:
            brand = PvcBrand()
            brand.name = name
            self.sa_session.add(brand)
        return brand

    def _get_tbs_member_product_brand(self, brand, tbs_member_product):
        return self.sa_session.query(PvcTbsMemberProductBrand).filter(and_(
            PvcTbsMemberProductBrand.brand == brand,
            PvcTbsMemberProductBrand.tbs_member_product == tbs_member_product
        )).first()

    @transaction_decorator
    def merge_brand_with(self, brand_id, target_brand_id):
        """
        The target brand will be the one kept.
        :param brand_id:
        :param target_brand_id:
        :return:
        """
        self.assert_permissions('manage_pvc_brands')
        target = self.get_brand(target_brand_id)
        # we cannot set a merge if the target is already set to be merged.
        if target.merge_brand is not None:
            raise PublicUserException(_('Target brand "%(target)s" is already set to be merged with another brand "%(merge)s"' % {
                'target': target.name,
                'merge': target.merge_brand.name,
            }))

        brand = self.get_brand(brand_id)
        if brand.merge_status is not None and brand.merge_status != PvcBrand.MERGE_STATUS_PENDING:
            raise PublicUserException(_('This brand is already being processed for merging.'))

        brand.merge_status = PvcBrand.MERGE_STATUS_PENDING
        brand.merge_brand = target
        self.sa_session.add(brand)
        return brand

    @transaction_decorator
    def mark_merge_brands_as_processing(self):
        # get all brands to be merged:
        brands = self.sa_session.query(PvcBrand).filter(PvcBrand.merge_status == PvcBrand.MERGE_STATUS_PENDING).all()
        if len(brands) == 0:
            return None

        for brand in brands:
            if brand.merge_brand is None:
                self.logger.info('Brand marked for Merge but target brand no longer available.')
                brand.merge_status = None
                self.sa_session.add(brand)
                continue
            if brand.merge_brand.merge_status is not None:
                # This should not happen
                raise SystemException('Cannot merge brand (%d) due to merge clash.' % brand.id)

            brand.merge_status = PvcBrand.MERGE_STATUS_PROCESSING
            self.sa_session.add(brand)

        return brands

    @transaction_decorator
    def merge_brands(self, brands):
        for brand in brands:
            assert brand.merge_status == PvcBrand.MERGE_STATUS_PROCESSING

            self.logger.info('Start merging of brand (%s) into brand (%s)' % (brand, brand.merge_brand))
            # update all contribution lines to the target brand
            self.logger.info(' - Processing contribution order lines')
            lines = self.sa_session.query(PvcContributionOrderLine).filter(PvcContributionOrderLine.brand == brand).all()
            for line in lines:
                line.brand = brand.merge_brand
                line.order.add_log('Updated brand "%s" to "%s" on order line "%d" due to brand merge.' % (
                    brand.name, brand.merge_brand.name, line.index
                ))
                self.sa_session.add(line)

            # update all member brands
            self.logger.info(' - Processing tbs member product brands')
            member_brands = self.sa_session.query(PvcTbsMemberProductBrand).filter(PvcTbsMemberProductBrand.brand == brand).all()
            for mbrand in member_brands:
                # check if the member doesnt already have the target brand selected.
                tbrand = self._get_tbs_member_product_brand(brand.merge_brand, mbrand.tbs_member_product)
                if tbrand is not None:
                    if mbrand.is_active:
                        tbrand.is_active = True
                    self.sa_session.delete(mbrand)
                    continue
                mbrand.brand = brand.merge_brand
                self.sa_session.add(mbrand)

            self.logger.info(' - Deleting original brand')
            self.sa_session.delete(brand)

    @transaction_decorator
    def cancel_merge(self, brand_id):
        self.assert_permissions('manage_pvc_brands')
        brand = self.get_brand(brand_id)
        if brand.merge_status is None:
            return True

        if brand.merge_status == PvcBrand.MERGE_STATUS_PROCESSING:
            raise PublicUserException(_('The merge for this brand is already processing and cannot be undone.'))

        brand.merge_brand = None
        brand.merge_status = None

        self.logger.info('Pending brand merge request cancelled for: %s' % brand)
        self.sa_session.add(brand)
        return brand

    @transaction_decorator
    def get_subrogated_product(self, tbs_member_product, subrogation):
        return self.get_or_save_tbs_member_product(tbs_member_product.tbs_member, tbs_member_product.tbs_product.id,
                                                   subrogation=subrogation, country_id=tbs_member_product.country.id)

    def get_product_fees(self, tbs_id, service_level_id, period=None):
        if period is None:
            period = today()

        tbs = self._get_take_back_schemes(tbs_id, ['manage_pvc_product_levels']).pop()

        # CHECK FOR B2B split
        has_dimension_split = tbs.get_config('pvc.tbs.require.end.user.distinction') or tbs.get_config('pvc.tbs.contribution.automatic.end.user.distinction')

        rv = []
        unordered = {}
        result = self.sa_session.query(PvcTbsProduct).join(WEEEProductLevel).filter(
            PvcTbsProduct.tbs == tbs
        ).order_by(WEEEProductLevel.parent_level_id, WEEEProductLevel.position).all()
        # get all products for this TBS which are used in Declarations
        for tp in result:
            p = {
                'id': tp.id,
                'product_id': tp.product.id,
                'position': tp.product.position,
                'label': tp.label,
                'basis_for_invoicing': tp.basis_for_invoicing,
                'custom_unit': tp.custom_unit,
                'description': tp.description,
                'fee': None if has_dimension_split else tp.get_fee(period, service_level_id, PvcTbsProductFee.DIMENSION_NONE, False),
                'start_date': None if has_dimension_split else tp.get_fee_start_date(period, service_level_id, PvcTbsProductFee.DIMENSION_NONE, False),
                'fee_b2b': None if not has_dimension_split else tp.get_fee(period, service_level_id, PvcTbsProductFee.DIMENSION_IS_B2B, False),
                'b2b_start_date': None if not has_dimension_split else tp.get_fee_start_date(period, service_level_id, PvcTbsProductFee.DIMENSION_IS_B2B, False),
                'fee_b2c': None if not has_dimension_split else tp.get_fee(period, service_level_id, PvcTbsProductFee.DIMENSION_IS_B2C, False),
                'b2c_start_date': None if not has_dimension_split else tp.get_fee_start_date(period, service_level_id, PvcTbsProductFee.DIMENSION_IS_B2C, False),
                'inherited_fee': None if has_dimension_split else tp.get_fee(period, service_level_id, PvcTbsProductFee.DIMENSION_NONE, True),
                'inherited_fee_b2b': None if not has_dimension_split else tp.get_fee(period, service_level_id, PvcTbsProductFee.DIMENSION_IS_B2B, True),
                'inherited_fee_b2c': None if not has_dimension_split else tp.get_fee(period, service_level_id, PvcTbsProductFee.DIMENSION_IS_B2C, True),
                'fees': [], # all fees for all periods
                'children': []
            }

            if tp.product.parent_level is None:
                rv.append(p)
                continue
            if not unordered.has_key(tp.product.parent_level.id):
                unordered[tp.product.parent_level.id] = []

            unordered[tp.product.parent_level.id].append(p)

        def append_children_to_rv(unordered, parent_id, parent):
            if not unordered.has_key(parent_id):
                return
            parent['children'] = sorted(unordered[parent_id], key=lambda k: k['position'])
            for child in parent['children']:
                append_children_to_rv(unordered, child['product_id'], child)

        for parent in rv:
            append_children_to_rv(unordered, parent['product_id'], parent)

        return {
            'tbs_name': tbs.name,
            'has_dimension_split': has_dimension_split,
            'products': rv
        }

    @transaction_decorator
    def save_product_fees(self, tbs_id, service_level_id, period, values):
        lazy_assert(service_level_id, int)

        tbs = self._get_take_back_schemes(tbs_id, ['manage_pvc_product_levels']).pop()
        has_dimension_split = tbs.get_config('pvc.tbs.require.end.user.distinction') or tbs.get_config('pvc.tbs.contribution.automatic.end.user.distinction')

        result = self.sa_session.query(PvcTbsProduct).join(WEEEProductLevel).filter(
            PvcTbsProduct.tbs == tbs
        ).order_by(WEEEProductLevel.parent_level_id, WEEEProductLevel.position).all()
        for product in result:

            if has_dimension_split:
                if values['fee_b2b_%d' % product.id] != '' and values['fee_b2b_%d' % product.id] is not None:
                    self._save_or_update_product_fee(product, service_level_id, period, values['fee_b2b_%d' % product.id], PvcTbsProductFee.DIMENSION_IS_B2B)
                if values['fee_b2c_%d' % product.id] != '' and values['fee_b2c_%d' % product.id] is not None:
                    self._save_or_update_product_fee(product, service_level_id, period, values['fee_b2c_%d' % product.id], PvcTbsProductFee.DIMENSION_IS_B2C)
                continue

            if values['fee_%d' % product.id] == '' or values['fee_%d' % product.id] is None:
                continue

            self._save_or_update_product_fee(product, service_level_id, period, values['fee_%d' % product.id], PvcTbsProductFee.DIMENSION_NONE)

    def _save_or_update_product_fee(self, product, service_level_id, period, fee, dimension):

        pf = self.sa_session.query(PvcTbsProductFee).filter(
            PvcTbsProductFee.tbs_product == product,
            PvcTbsProductFee.service_level_id == service_level_id,
            PvcTbsProductFee.start_date == period.date(),
            PvcTbsProductFee.dimension == dimension
        ).first()

        if pf:
            # The fee has not changed.
            if pf.fee == fee:
                return pf

            self.logger.info('Fee for product %d with service_level %d and dimension %d for period %s has been altered from %s to %s' % (
                product.id,
                service_level_id,
                dimension,
                period.date(),
                pf.fee,
                fee
            ))

            # alter declarations affected.

        else:
            pf = PvcTbsProductFee()
            pf.tbs_product = product
            pf.service_level_id = service_level_id
            pf.start_date = period.date()
            pf.dimension = dimension

        pf.process_status = PvcTbsProductFee.STATUS_PROCESS
        pf.fee = fee
        self.sa_session.add(pf)
        return pf

    def get_periods_for_fee(self, product_fee):

        # if the start date of this fee is in the future we do not process.
        if product_fee.start_date > today():
            return [], [], [],

        import math

        yearly = []
        quarterly = []
        monthly = []

        m = product_fee.start_date.month
        y = product_fee.start_date.year
        if m == 1:
            yearly.append(y)
            quarterly.append((y, 1, ))
        if m in (4, 7, 10):
            quarterly.append((y, m/3+1, ))
        monthly.append((y, m, ))

        # difference between periods to get all possible periods
        next_period = self.sa_session.query(func.pvcGetNextProductContributionFeeDate(
            product_fee.tbs_product.id,
            product_fee.service_level.id,
            product_fee.start_date,
            product_fee.dimension
        )).first()

        if next_period is not None:
            next_period = next_period[0]

        # if next period is in the future, we only alter stuff until today (the rest should not be triggered/available yet).
        if next_period is None or next_period > today():
            next_period = today()

        # if we are still in the month to which the fee was set, we don't need to process any other months
        if next_period.month == m:
            return yearly, quarterly, monthly,

        def diff_month(d1, d2):
            return (d1.year - d2.year)*12 + d1.month - d2.month

        num = diff_month(next_period, product_fee.start_date)
        cm = m
        year_inc = 0
        for i in range(num):
            nm = cm+i+1

            # we go to the next year
            if nm > 12:
                year_inc = nm/12
                nm -= 12

            # the next year means yearly declarations are included.
            if nm == 1 and year_inc > 0:
                yearly.append(y+year_inc)

            # quarterly declarations are only involved with these months.
            if nm in (1, 4, 7, 10):
                quarterly.append((y+year_inc, nm/3+1, ))

            monthly.append((y+year_inc, nm, ))

        return yearly, quarterly, monthly,

    def get_tbs_member_products(self, tbs_id, member_id, subrogation_id=None, validate_types=False):
        """
        Returns all available TBS products to be used in Declarations and which ones are selected by the TBS member.
        The validate types parameter is optional as this method is also used for UPCOMING guarantee forecasts, which
        should have all the products available.
        :param tbs_id:
        :param member_id:
        :return:
        """
        tbs_products = self.get_declarable_tbs_products(tbs_id)

        tm = self._get_tbs_member(tbs_id, member_id)

        tmp = self.sa_session.query(PvcTbsMemberProduct).filter(and_(
            PvcTbsMemberProduct.tbs_member == tm,
            PvcTbsMemberProduct.tbs_product_id.in_([p.id for p in tbs_products]),
            PvcTbsMemberProduct.is_active == True,
            PvcTbsMemberProduct.subrogation_id == subrogation_id
        )).all()

        tm_products = list(set([p.tbs_product for p in tmp]))

        # if we have to validate types, check if this TBS has guarantee forecast on.
        if validate_types and tm.tbs.get_config('pvc.tbs.guarantee.forecast'):
            types = []

            # check the type of all of the active service levels of the member
            if tm.active_service_level.service_level is not None:
                types.append(tm.active_service_level.service_level.product_type)
            if tm.active_service_level.pv_service_level_id is not None:
                types.append(tm.active_service_level.pv_service_level.product_type)
            # filter for only those, in both tbs_products as well as the tbs member products.
            tbs_products = [p for p in tbs_products if p.product_type in types]
            tm_products = [p for p in tbs_products if p.product_type in types]

        return tbs_products, tm_products

    @transaction_decorator
    def save_tbs_member_products(self, tbs_id, member_id, products, countries=None, subrogation_id=None, brands=None):

        tm = self._get_tbs_member(tbs_id, member_id)
        if tm is None:
            raise PublicUserException(_('TBS member not found'))
        tbs_products = self.get_declarable_tbs_products(tbs_id)

        if brands is not None:
            brands = [b['brand'] for b in brands if b['brand'].strip() != '']

        if countries is None or len(countries) == 0:
            countries = [None]

        subrogation = None
        if subrogation_id not in (None, "", ):
            subrogation = self.sa_session.query(PvcSubrogation).filter(and_(
                PvcSubrogation.id == subrogation_id,
                PvcSubrogation.subrogator_id == tm.id
            )).one()

        rv = []
        for pid in products:
            for country_id in countries:
                rv.append(self.get_or_save_tbs_member_product(tm, pid, is_active=True, country_id=country_id, subrogation=subrogation, brands=brands))

        return rv

    @transaction_decorator
    def disable_tbs_member_product(self, tbs_member, tbs_product_id, country_id=None, subrogation=None):
        if country_id is None:
            country_id = tbs_member.tbs.country.id

        tmp = self.sa_session.query(PvcTbsMemberProduct).filter(and_(
            PvcTbsMemberProduct.tbs_member == tbs_member,
            PvcTbsMemberProduct.tbs_product_id == tbs_product_id,
            PvcTbsMemberProduct.country_id == country_id,
            PvcTbsMemberProduct.subrogation == subrogation
        )).first()
        if tmp is None:
            return None

        tmp.is_active = False
        self.sa_session.add(tmp)
        return tmp

    @transaction_decorator
    def deactivate_product_brand(self, tbs_member_product_id, brand_id):
        tmp = self.sa_session.query(PvcTbsMemberProduct).filter(PvcTbsMemberProduct.id == tbs_member_product_id).one()
        self._get_tbs_member_licenses(tmp.tbs_member.id, ['pvc_member_license'])  # permission check

        b = [mb for mb in tmp.brands if mb.brand.id == brand_id]
        if len(b) == 0:
            return None
        b = b.pop()
        b.is_active = False
        self.sa_session.add(b)
        return b

    def deactivate_tbs_member_product(self, tbs_member_product_id):
        return self._set_tbs_member_product_active_state(tbs_member_product_id, False)

    def reactivate_tbs_member_product(self, tbs_member_product_id):
        return self._set_tbs_member_product_active_state(tbs_member_product_id, True)

    @transaction_decorator
    def _set_tbs_member_product_active_state(self, tbs_member_product_id, is_active):
        tmp = self.sa_session.query(PvcTbsMemberProduct).filter(PvcTbsMemberProduct.id == tbs_member_product_id).one()
        self._get_tbs_member_licenses(tmp.tbs_member.id, ['pvc_member_license'])  # permission check
        tmp.is_active = is_active
        self.sa_session.add(tmp)
        return tmp