# Copyright (c) 2015 BaseSoft B.V.
# See LICENSE for details.
from basesoft.web.contrib.i18n.models import Country
from pvcycle.subrogation.models import PvcSubrogation
from sqlalchemy.orm.util import aliased
from sqlalchemy.sql.expression import and_, func

from django.utils.translation import ugettext as _

from basesoft.web.contrib.license.models import License
from basesoft.web.core.util.custom_query import CustomQueryModel, CustomQueryField, DATA_TYPE_BOOLEAN
from iqportal.weee.core.models import WEEEProductLevel

from pvcycle.core.custom_query import AbstractPvcCustomQuery
from pvcycle.core.models import PvcTbsLicense, PvcTbsMember, PvcMemberLicense
from pvcycle.product.models import PvcTbsProduct, PvcTbsMemberProduct, PvcBrand, PvcTbsMemberProductBrand


class TbsProductCustomQuery(AbstractPvcCustomQuery):
    permissions = ['manage_pvc_product_levels']

    def tables(self):
        return (
            CustomQueryModel(PvcTbsProduct),
            CustomQueryModel(WEEEProductLevel, priv=None),
            CustomQueryModel(PvcTbsLicense),
        )

    def fields(self):
        return (
            CustomQueryField(PvcTbsProduct.id, 'ID', key='id'),
            CustomQueryField(PvcTbsProduct.label, _('Label'), key='label', is_filter=True),
            CustomQueryField(WEEEProductLevel.label, _('Default label'), key='default_label', is_filter=True),
            CustomQueryField(PvcTbsProduct.description, _('Description'), key='description', is_filter=True),

            CustomQueryField(PvcTbsProduct.status, _('Status'), key='status', is_filter=True, filter_options=[{
                'value': PvcTbsProduct.STATUS_USE_IN_DECLARATION,
                'label': _("Use in declaration"),
            }, {
                'value': PvcTbsProduct.STATUS_INACTIVE,
                'label': _("Inactive"),
            }]),
            CustomQueryField(PvcTbsProduct.basis_for_invoicing, _('Basis for invoicing'), key='basis_for_invoicing', is_filter=True, filter_options=[{
                'value': PvcTbsProduct.INVOICE_BASED_ON_WEIGHT,
                'label': _("Weight"),
            }, {
                'value': PvcTbsProduct.INVOICE_BASED_ON_QUANTITY,
                'label': _("Quantity"),
            }, {
                'value': PvcTbsProduct.INVOICE_BASED_ON_CUSTOM_UNIT,
                'label': _("Custom unit"),
            }]),
            CustomQueryField(PvcTbsLicense.name, _('Take-back scheme'), key='tbs_name', is_filter=True)
        )

    def set_where(self):
        self.where = PvcTbsProduct.tbs_id.in_([x.id for x in self._get_take_back_schemes(permissions=['manage_pvc_product_levels'])])
        self.where = and_(PvcTbsProduct.status != 0, self.where)


TbsLicense = aliased(License)
CommissionerMember = aliased(PvcTbsMember)
CommissionerLicense = aliased(License)

class TbsMemberProductCustomQuery(AbstractPvcCustomQuery):
    permissions = ('pvc_member_license', )

    def _get_member_column(self):
        return PvcTbsMember.member_id

    def tables(self):
        return (
            CustomQueryModel(PvcTbsMember),
            CustomQueryModel(PvcMemberLicense, onclause=PvcTbsMember.member_id == PvcMemberLicense.license_id),
            CustomQueryModel(TbsLicense, onclause=PvcTbsMember.tbs_id == TbsLicense.id),
            CustomQueryModel(PvcTbsMemberProduct),
            CustomQueryModel(PvcTbsProduct),
            CustomQueryModel(Country, onclause=PvcTbsMemberProduct.country_id == Country.id),
            CustomQueryModel(PvcTbsMemberProductBrand, onclause=and_(
                PvcTbsMemberProduct.id == PvcTbsMemberProductBrand.tbs_member_product_id,
                PvcTbsMemberProductBrand.is_active == True
            ), use_outerjoin=True),
            CustomQueryModel(PvcBrand, onclause=PvcTbsMemberProductBrand.brand_id == PvcBrand.id, use_outerjoin=True),
            CustomQueryModel(PvcSubrogation, onclause=PvcTbsMemberProduct.subrogation_id == PvcSubrogation.id, use_outerjoin=True),
            CustomQueryModel(CommissionerMember, onclause=PvcSubrogation.commissioner_id == CommissionerMember.id, use_outerjoin=True),
            CustomQueryModel(CommissionerLicense, onclause=CommissionerMember.member_id == CommissionerLicense.id, use_outerjoin=True),
        )

    def fields(self):
        return (
            CustomQueryField(PvcTbsMemberProduct.id, _('ID'), key='id'),
            CustomQueryField(PvcTbsMember.id, _('ID'), key='tbs_member_id'),
            CustomQueryField(TbsLicense.name, _('TBS'), key='tbs_name', is_filter=True),
            CustomQueryField(PvcMemberLicense.name, _('Member'), key='member_name', is_filter=True),
            CustomQueryField(CommissionerLicense.name, _('Commissioner'), key='commissioner', is_filter=True),
            CustomQueryField(False, _('Brand ID'), 'brand_id', use_when_separated=PvcBrand.id),
            CustomQueryField(func.group_concat(PvcBrand.name), _('Brands'), key='brands', is_filter=True, use_when_separated=PvcBrand.name),
            CustomQueryField(Country.name, _('Country'), key='countries', is_filter=True),
            CustomQueryField(PvcTbsProduct.label, _('Product label'), key='product_label', is_filter=True),
            CustomQueryField(PvcTbsProduct.description, _('Product'), key='product_desc', is_filter=True),
            CustomQueryField(PvcTbsMemberProduct.is_active, _('Active'), key='is_active', is_filter=True, type=DATA_TYPE_BOOLEAN),
        )

    def set_where(self):
        self.where = PvcTbsMember.id.in_([x.id for x in self._get_tbs_member_licenses()])

    def set_group_by(self):
        if self.separated:
            self.group_by = (PvcTbsMemberProduct.id, PvcBrand.id, )
        else:
            self.group_by = (PvcTbsMemberProduct.id, )

class BrandCustomQuery(AbstractPvcCustomQuery):
    permissions = ['manage_pvc_product_levels']

    def tables(self):
        return (
            CustomQueryModel(PvcBrand),
            CustomQueryModel(PvcTbsMemberProductBrand, use_outerjoin=True),  # used to show usage
        )

    def fields(self):
        return (
            CustomQueryField(PvcBrand.id, _('ID'), key='brand_id'),
            CustomQueryField(PvcBrand.name, _('Name'), key='brand_name', is_filter=True),
            CustomQueryField(PvcBrand.merge_status, _('Merge status'), key='status'),
            CustomQueryField(func.IF(func.isnull(func.count(PvcTbsMemberProductBrand.brand_id)), 0, 1), _('Used'), key='used', type=DATA_TYPE_BOOLEAN),
        )

    group_by = (PvcBrand.id, )

custom_queries = (
    TbsProductCustomQuery,
    TbsMemberProductCustomQuery,
    BrandCustomQuery,
)
