# Copyright (c) 2015 BaseSoft B.V.
# See LICENSE for details.
from basesoft.util.virtual_date import today
from basesoft.web.core.util.controllers import lazy_assert
from django.utils.translation import ugettext_lazy
from iqportal.weee.core.models import WEEEProductLevel
from sqlalchemy import func, and_
from sqlalchemy.orm.session import Session


class PvcTbsProduct(object):
    """
    @DynamicAttrs
    """

    TYPE_OTHER = 1
    TYPE_PV_RELATED = 2

    INVOICE_BASED_ON_WEIGHT = 1
    INVOICE_BASED_ON_QUANTITY = 2
    INVOICE_BASED_ON_CUSTOM_UNIT = 3

    # this product level is available to the Members for Declaration.
    STATUS_USE_IN_DECLARATION = 1
    # this product level is no longer/not in use/only used for fee.
    STATUS_INACTIVE = 2

    CUSTOM_UNIT_NOMINAL_POWER = 1
    CUSTOM_UNIT_SURFACE = 2

    # make sure this matches product-debug.js
    CUSTOM_UNIT_LABEL_MAP = {
        CUSTOM_UNIT_NOMINAL_POWER: ugettext_lazy('Nominal power'),
        CUSTOM_UNIT_SURFACE: ugettext_lazy('Surface'),
    }
    CUSTOM_UNIT_UNIT_MAP = {
        CUSTOM_UNIT_NOMINAL_POWER: ugettext_lazy('Wp'),
        CUSTOM_UNIT_SURFACE: ugettext_lazy('m2'),
    }

    def get_fee(self, period, service_level_id, dimension=None, propagate=True):
        """
        dimension is used when b2b/b2c is enabled on tbs level.
        """
        lazy_assert(service_level_id, int)

        if dimension is None:
            dimension = PvcTbsProductFee.DIMENSION_NONE

        session = Session.object_session(self)
        fee = session.query(func.pvcGetCurrentProductContributionFee(self.id, service_level_id, period, dimension)).one()[0]

        if fee is not None:
            return fee

        if not propagate:
            return None

        if self.product.parent_level is None:
            return None

        # we get the parent product
        p = session.query(PvcTbsProduct).filter(and_(
            PvcTbsProduct.tbs == self.tbs,
            PvcTbsProduct.product == self.product.parent_level
        )).first()

        if p is None:
            return None
        return p.get_fee(period, service_level_id, dimension)

    def get_fee_start_date(self, period, service_level_id, dimension=None, propagate=True):
        if dimension is None:
            dimension = PvcTbsProductFee.DIMENSION_NONE

        session = Session.object_session(self)
        fee = session.query(func.pvcGetCurrentProductContributionFeeDate(self.id, service_level_id, period, dimension)).one()[0]

        if fee is not None:
            return fee

        if not propagate:
            return None

        if self.product.parent_level is None:
            return None

        # we get the parent product
        p = session.query(PvcTbsProduct).filter(and_(
            PvcTbsProduct.tbs == self.tbs,
            PvcTbsProduct.product == self.product.parent_level
        )).first()

        if p is None:
            return None
        return p.get_fee_start_date(period, service_level_id, dimension)

    def get_current_fee(self, service_level_id, dimension=None):
        return self.get_fee(today(), service_level_id, dimension)

    def get_child_products_with_inherited_fee(self):
        session = Session.object_session(self)

        return session.query(PvcTbsProduct).join(WEEEProductLevel).outerjoin(PvcTbsProductFee).filter(and_(
            WEEEProductLevel.parent_level == self.product,
            PvcTbsProduct.tbs == self.tbs,
            PvcTbsProductFee.service_level == None
        )).all()


    def __repr__(self):
        return '<PvcTbsProduct id="%d" tbs="%s" product="%s">' % (self.id, self.tbs, self.product)

class PvcTbsProductFee(object):
    """
    @DynamicAttrs
    """
    DIMENSION_NONE = 0
    DIMENSION_IS_B2B = 1
    DIMENSION_IS_B2C = 2

    STATUS_DONE = 0
    STATUS_PROCESS = 1
    STATUS_PROCESSING = 2

    def __repr__(self):
        return '<PvcTbsProductFee tbs_product=%s>' % self.tbs_product


class PvcTbsMemberProduct(object):
    """
    @DynamicAttrs
    """
    def __repr__(self):
        return '<PvcTbsMemberProduct id=%d>' % self.id


class PvcBrand(object):
    """
    @DynamicAttrs
    """
    MERGE_STATUS_PENDING = 1
    MERGE_STATUS_PROCESSING = 2

    def __repr__(self):
        return '<PvcBrand id=%d name="%s">' % (self.id, self.name)


class PvcTbsMemberProductBrand(object):
    """
    @DynamicAttrs
    """
    def __repr__(self):
        return '<PvcBrand tbs_member_product=%s brand="%s">' % (self.tbs_member_product, self.brand)

