# Copyright (c) 2015 BaseSoft B.V.
# See LICENSE for details.

import datetime

from sqlalchemy.orm import mapper, relation
from sqlalchemy.schema import UniqueConstraint
from sqlalchemy import Integer, Column, UnicodeText, DateTime, ForeignKey, SmallInteger, Boolean, Numeric, Date, Unicode

from basesoft.sqlalchemy.table import IQTable
from basesoft.web.contrib.license import models as license_models

from pvcycle.product.models import PvcTbsProduct, PvcTbsMemberProduct, PvcTbsProductFee, PvcBrand, \
    PvcTbsMemberProductBrand

__mapping_defined = False
__tables_defined = False
tables = {}


def reset():
    global __mapping_defined
    global __tables_defined

    __mapping_defined = False
    __tables_defined = False


def get_tables():
    return tables


def do_tables(metadata):
    """
    This function is called by Django middleware when needed.
    """
    global __tables_defined
    if __tables_defined:
        return

    # This is used in the reporting cubes, TODO #1007
    closurePvcTbsProduct = IQTable('closurePvcTbsProduct', metadata,
        Column('parent_id', Integer, nullable=False),
        Column('product_id', Integer, index=True, nullable=False),
        Column('distance', Integer, nullable=False),
        UniqueConstraint('parent_id', 'product_id')
    )

    tblPvcTbsProduct = IQTable('tblPvcTbsProduct', metadata,
        Column('id', Integer, primary_key=True, autoincrement=True),
        Column('tbs_id', Integer, ForeignKey('tblPvcTbsLicense.license_id'), nullable=False),
        Column('product_id', Integer, ForeignKey('tblWEEEProductLevel.id'), nullable=False),
        Column('product_type', SmallInteger, nullable=False, server_default=str(PvcTbsProduct.TYPE_OTHER)),
        Column('status', SmallInteger, nullable=False, default=PvcTbsProduct.STATUS_USE_IN_DECLARATION),
        Column('basis_for_invoicing', SmallInteger, nullable=False, default=PvcTbsProduct.INVOICE_BASED_ON_WEIGHT),
        Column('custom_unit', SmallInteger, nullable=True, default=None),
        Column('label', UnicodeText, nullable=False),
        Column('description', UnicodeText, nullable=False),
        Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
        Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now, onupdate=datetime.datetime.now),
        UniqueConstraint('tbs_id', 'product_id')
    )

    tblPvcTbsProductFee = IQTable('tblPvcTbsProductFee', metadata,
        Column('tbs_product_id', Integer, ForeignKey('tblPvcTbsProduct.id', ondelete='CASCADE'), primary_key=True),
        Column('service_level_id', Integer, ForeignKey('tblPvcTbsServiceLevel.id', ondelete='CASCADE'), primary_key=True),
        Column('start_date', Date(), primary_key=True),
        Column('dimension', SmallInteger, primary_key=True, autoincrement=False),
        Column('fee', Numeric(10, 5), nullable=False),
        Column('process_status', SmallInteger, nullable=False),
        Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
        Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now, onupdate=datetime.datetime.now)
    )

    tblPvcTbsMemberProduct = IQTable('tblPvcTbsMemberProduct', metadata,
        Column('id', Integer, primary_key=True, autoincrement=True),
        Column('country_id', Integer, ForeignKey('tblCountry.id'), nullable=False),
        Column('tbs_product_id', Integer, ForeignKey('tblPvcTbsProduct.id'), nullable=False),
        Column('tbs_member_id', Integer, ForeignKey('tblPvcTbsMember.id'), nullable=False),
        Column('subrogation_id', Integer, ForeignKey('tblPvcSubrogation.id'), nullable=True),

        # used to determine whether or not next declaration has these products added by default in case of no Forecast.
        Column('is_active', Boolean, nullable=False, default=True),

        Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
        Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now, onupdate=datetime.datetime.now),

        UniqueConstraint('country_id', 'tbs_product_id', 'tbs_member_id', 'subrogation_id')
    )

    tblPvcBrand = IQTable('tblPvcBrand', metadata,
        Column('id', Integer, primary_key=True, autoincrement=True),
        Column('name', Unicode(255), unique=True, nullable=False),

        Column('merge_brand_id', Integer, ForeignKey('tblPvcBrand.id'), nullable=True, default=None),
        Column('merge_status', Integer, nullable=True, default=None),

        Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
        Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now, onupdate=datetime.datetime.now),
    )

    tblPvcTbsMemberProductBrand = IQTable('tblPvcTbsMemberProductBrand', metadata,
        Column('tbs_member_product_id', Integer, ForeignKey('tblPvcTbsMemberProduct.id'), primary_key=True),
        Column('brand_id', Integer, ForeignKey('tblPvcBrand.id'), primary_key=True),

        # used to determine whether or not next declaration has these brands added by default in case of no Forecast.
        Column('is_active', Boolean, nullable=False, default=True),

        Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
        Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now, onupdate=datetime.datetime.now),
    )

    tables['closurePvcTbsProduct'] = closurePvcTbsProduct

    tables['tblPvcTbsProduct'] = tblPvcTbsProduct
    tables['tblPvcTbsProductFee'] = tblPvcTbsProductFee
    tables['tblPvcTbsMemberProduct'] = tblPvcTbsMemberProduct
    tables['tblPvcBrand'] = tblPvcBrand
    tables['tblPvcTbsMemberProductBrand'] = tblPvcTbsMemberProductBrand

    __tables_defined = True


def do_mappings():
    """
    This function is called by Django middleware when needed.
    """
    global __mapping_defined
    if __mapping_defined:
        return

    from basesoft.web.contrib.i18n import models as i18n_models
    from iqportal.weee.core import models as weee_models
    from pvcycle.core import models as core_models
    from pvcycle.subrogation import models as subrogation_models

    mapper(PvcTbsProduct, tables['tblPvcTbsProduct'],
        properties = {
            'tbs': relation(core_models.PvcTbsLicense),
            'product': relation(weee_models.WEEEProductLevel),
        }
    )

    mapper(PvcTbsProductFee, tables['tblPvcTbsProductFee'],
        properties = {
            'tbs_product': relation(PvcTbsProduct),
            'service_level': relation(core_models.PvcTbsServiceLevel),
        }
    )

    mapper(PvcTbsMemberProduct, tables['tblPvcTbsMemberProduct'],
        properties = {
            'tbs_product': relation(PvcTbsProduct),
            'tbs_member': relation(core_models.PvcTbsMember),
            'subrogation': relation(subrogation_models.PvcSubrogation),
            'country': relation(i18n_models.Country),
        }
    )

    mapper(PvcBrand, tables['tblPvcBrand'],
        properties = {
            'merge_brand': relation(PvcBrand, uselist=False, remote_side=[tables['tblPvcBrand'].c.id]),
        }
    )

    mapper(PvcTbsMemberProductBrand, tables['tblPvcTbsMemberProductBrand'],
        properties = {
            'tbs_member_product': relation(PvcTbsMemberProduct, backref="brands"),
            'brand': relation(PvcBrand),
        }
    )

    __mapping_defined = True


def fill_data(conn):
    pass


def app_permissions_category():
    """
    Returns a category for all permissions defined for this app
    """
    return license_models.PermissionCategory(u'PVC Product')


def app_permissions():
    from django.utils.translation import ugettext as _
    return [
        license_models.Permission(u'manage_pvc_product_levels', u"""
            This gives the right to manage the PV CYCLE product levels.
            """, _("Manage PVC product levels")),
        license_models.Permission(u'manage_pvc_brands', u"""
            This gives the right to manage the PV CYCLE brands.
            """, _("Manage PVC brands")),

    ]
