'''
Created on Aug 26, 2013

@author: marten
'''
import os
import xmlrpclib
from basesoft.web.contrib.notification.controllers import NotificationController

from django.utils import simplejson
from django.utils.translation import ugettext as _
from django.conf import settings

from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.sql.expression import and_

from twisted.web.xmlrpc import Fault

from decimal import Decimal

from basesoft.web.core.util.controllers import AbstractController, transaction_decorator
from basesoft.web.core.exceptions import PublicSecurityException, PublicUserException
from basesoft.web.contrib.file.models import ImportFile, File, ImportFileType
from basesoft.web.core.jsonrpc.dispatcher import json_object_hook
from basesoft.web.contrib.file.controllers import AnonymousFileController, ProcessFileController, UploadFileController

from pvcycle.pvmodule.models import PvcPVModule, PvcPVModuleSupplier, PvcPVModuleType, PvcPVModuleCellType, PvcTbsMemberPVModule
from pvcycle.core.controllers import AbstractPvcController, PvcMemberController, PvcNotificationController
from pvcycle.core.models import PvcMemberLicense, PvcTbsMember, PvcTbsLicense, PvcOperatorLicense
from pvcycle.pvmodule.importers import PhotonDBXlsxImporter, PhotonDBCsvImporter, PhotonDBImportException, FailureFileExporter
from pvcycle.subrogation.models import PvcSubrogation

class PvcPVModuleController(AbstractController, AbstractPvcController):
    """
    The PV Module controller is for managing all the Photon Voltaic modules for the PV Cycle project.
    """

    def _get(self, obj, _id, name, limit=None, contains=None):

        _where = None

        if _id is not None:
            if type(_id) != list:
                _id = [_id]
            _where = obj.id.in_(_id)

        if name is not None:
            if _where is None:
                _where = obj.name==name
            else:
                _where = and_(_where, obj.name==name)
        elif contains is not None:
            if _where is None:
                _where = obj.name.like('%%%s%%' % contains)
            else:
                _where = and_(_where, obj.name.like('%%%s%%' % contains))

        q = self.sa_session.query(obj).filter(_where).order_by(obj.name)
        if limit is not None:
            q = q.limit(limit)
        return q.all()

    def get_suppliers(self, _id=None, name=None, limit=None, contains=None):
        return self._get(PvcPVModuleSupplier, _id, name, limit=limit, contains=contains)

    def get_module_types(self, _id=None, name=None, supplier=None, limit=None, contains=None):
        if supplier is None:
            return self._get(PvcPVModuleType, _id, name)

        _where = PvcPVModuleSupplier.name==supplier
        if _id is not None:
            if type(_id) != list:
                _id = [_id]
            _where = PvcPVModuleType.id.in_(_id)

        if name is not None:
            _where = and_(_where, PvcPVModuleType.name==name)
        elif contains is not None:
            _where = and_(_where, PvcPVModuleType.name.like('%%%s%%' % contains))

        q = self.sa_session.query(PvcPVModuleType).join(
            PvcPVModule, PvcPVModule.type_id==PvcPVModuleType.id).join(
            PvcPVModuleSupplier, PvcPVModuleSupplier.id==PvcPVModule.supplier_id).filter(_where).order_by(PvcPVModuleType.name)

        if limit is not None:
            q = q.limit(limit)

        return q.all()

    def get_cell_types(self, _id=None, name=None, limit=None, contains=None):
        return self._get(PvcPVModuleCellType, _id, name, limit=limit, contains=contains)

    def get_modules(self, _id=None, supplier_name=None, type_name=None):
        _where = None

        q = self.sa_session.query(PvcPVModule)

        if _id is not None:
            if type(_id) != list:
                _id = [_id]
            _where = PvcPVModule.id.in_(_id)

        if supplier_name is not None:
            q = q.join(PvcPVModuleSupplier, PvcPVModule.supplier_id==PvcPVModuleSupplier.id)
            _w = PvcPVModuleSupplier.name==supplier_name
            if _where is None:
                _where = _w
            else:
                _where = and_(_where, _w)

        if type_name is not None:
            q = q.join(PvcPVModuleType, PvcPVModule.type_id==PvcPVModuleType.id)
            _w = PvcPVModuleType.name==type_name
            if _where is None:
                _where = _w
            else:
                _where = and_(_where, _w)

        return q.filter(_where).all()

    def get_modules_for_tbs(self, tbs, _id=None):

        #if type(tbs) == int:
        #    tbs = get_tbs()

        return []

    def load_module_state(self, _id):
        m = self.get_modules(_id).pop()
        return {
            'type': m.type.name,
            'supplier': m.supplier.name,
            'cell_type': m.cell_type.name,
            'nominal_power': m.nominal_power,
            'weight': m.weight,
            'width': m.width,
            'length': m.length,
        }

    @transaction_decorator
    def approve_module(self, _id):
        self.assert_permissions('pvc_approve_pv_modules')

        module = self.get_modules(_id).pop()

        if module.approval_status == PvcPVModule.PV_MODULE_STATUS_APPROVED:
            raise PublicUserException(_('This PV Module is already approved.'))

        # no first approval was done, the status still wont change.
        if module.first_approval_by is None:
            module.first_approval_by = self.get_account()
            self.sa_session.add(module)
            return module

        # we need two different parties to approve the PV Module.
        if module.first_approval_by == self.get_account():
            raise PublicSecurityException(_('You have already approved this PV Module, a secondary check by another party is required.'))

        # if the second party approves, the whole module is approved.
        module.second_approval_by = self.get_account()
        module.approval_status = PvcPVModule.PV_MODULE_STATUS_APPROVED

        self.sa_session.add(module)
        return module

    @transaction_decorator
    def reject_module(self, _id):
        self.assert_permissions('pvc_approve_pv_modules')

        module = self.get_modules(_id).pop()
        module.first_approval_by = None
        module.second_approval_by = None
        module.approval_status = PvcPVModule.PV_MODULE_STATUS_REJECTED

        self.logger.info('PV Module `%s` rejected by account `%s`' % (module.id, self.get_account().name))

        from basesoft.web.contrib.order.models import Order
        from pvcycle.order.models import PvcDeclarationOrderLine

        from pvcycle.order.controllers import PvcDeclarationOrderController
        dc = PvcDeclarationOrderController(self.sa_session, account_id=self.get_account().id, request=self.request)

        line = self.sa_session.query(PvcDeclarationOrderLine).join(
                Order, PvcDeclarationOrderLine.order_id==Order.id).join(
                PvcTbsMemberPVModule, PvcDeclarationOrderLine.tbs_member_module_id==PvcTbsMemberPVModule.id).filter(
                    and_(
                        Order.status.in_([Order.STATUS_NEW, Order.STATUS_PENDING]),
                        PvcTbsMemberPVModule.module==module
            )).first()

        # no declaration lines with this module.
        if line is None:
            self.sa_session.add(module)
            return module

        # remove lines from unclosed declarations
        while True:
            self.logger.info('Removing line (%s) from Declaration (%s) due to PV Module rejection' % (line.id, line.order.id))
            dc.remove_line(line.order, line.id)

            line = self.sa_session.query(PvcDeclarationOrderLine).join(
                Order, PvcDeclarationOrderLine.order_id==Order.id).join(
                PvcTbsMemberPVModule, PvcDeclarationOrderLine.tbs_member_module_id==PvcTbsMemberPVModule.id).filter(
                    and_(
                        Order.status.in_([Order.STATUS_NEW, Order.STATUS_PENDING]),
                        PvcTbsMemberPVModule.module==module
            )).first()
            if line is None:
                break

        self.sa_session.add(module)
        return module

    @transaction_decorator
    def save_or_update_module(self, _id, supplier, _type, cell_type, nominal_power, weight, width, length, photon_id=None):
        if _id is None:
            self.assert_permissions('pvc_create_pv_modules')
        else:
            self.assert_permissions('pvc_manage_pv_modules')

        module = self._save_or_update_module(_id, supplier, _type, cell_type, nominal_power, weight, width, length, photon_id)
        if self.has_permissions('pvc_approve_pv_modules'):
            module.first_approval_by = self.get_account()
        else:
            self._notify_for_approval(module)
        self.sa_session.add(module)
        return module

    @transaction_decorator
    def _save_or_update_module(self, _id, supplier, _type, cell_type, nominal_power, weight, width, length, photon_id=None):
        """
        Creates or updates a Photo Voltaic module definition.
        """
        nominal_power = Decimal(str(nominal_power))
        weight = Decimal(str(weight))
        width = Decimal(str(width))
        length = Decimal(str(length))

        assert nominal_power >= 0 # Some modules have 0 power output #462
        assert weight > 0
        assert width > 0
        assert length > 0

        if _id is None:
            module = PvcPVModule()
            self.logger.info('Creation of new PV Module with supplier `%s` and type `%s`' % (supplier, _type))
        else:
            module = self.sa_session.query(PvcPVModule).filter(PvcPVModule.id==_id).one()
            module.approval_status = PvcPVModule.PV_MODULE_STATUS_PENDING # reset approval status
            module.first_approval_by = None
            module.second_approval_by = None
            self.logger.info('Update of PV Module: %s' % _id)

        mods = self.get_modules(supplier_name=supplier, type_name=_type)
        if len(mods) != 0 and mods[0].id != module.id:
            raise PublicUserException(_('Combination for Brand and Type already exists, use the merge option to merge two or more PV Modules.'))

        module.supplier = self._get_or_create(PvcPVModuleSupplier, supplier)
        module.type = self._get_or_create(PvcPVModuleType, _type)
        module.cell_type = self._get_or_create(PvcPVModuleCellType, cell_type)
        module.nominal_power = nominal_power
        module.weight = weight # TODO: should weight be alterable as it may be used for amount calculations?
        module.width = width
        module.length = length
        module.photon_id = photon_id
        self.sa_session.add(module)
        return module

    def _notify_for_approval(self, module):
        nc = PvcNotificationController(self.sa_session, account_id=self.get_account().id)
        recipients = self.get_groups_in_domain_with_permissions('pvc_approve_pv_modules')

        notification = nc.set_notification(_('New PV Module registered'), _('A new PV Module has been registered by Member "%(member_name)s".' % {'member_name': self.get_default_group().license.name}))
        nc.create_system_notification(notification, recipients)

    def get_or_create_module(self, supplier, _type, cell_type, nominal_power, weight, width, length, photon_id=None, abstract=False):
        module = self.get_modules(supplier_name=supplier, type_name=_type)
        if len(module) > 0:
            return module.pop()
        if abstract:
            return self._save_or_update_module(None, supplier, _type, cell_type, nominal_power, weight, width, length, photon_id=photon_id)
        return self.save_or_update_module(None, supplier, _type, cell_type, nominal_power, weight, width, length, photon_id=photon_id)

    @transaction_decorator
    def _get_or_create(self, instance_class, _filter):
        if isinstance(_filter, instance_class):
            return _filter

        if type(_filter) == int:
            return self.sa_session.query(instance_class).filter(instance_class.id==_filter).one()

        try:
            instance = self.sa_session.query(instance_class).filter(instance_class.name==_filter).one()
        except NoResultFound:
            instance = instance_class()
            instance.name = _filter
            self.sa_session.add(instance)

        return instance

    @transaction_decorator
    def save_or_update_module_for_countries(self, member_id, countries, module, subrogation_id=None):
        member = None
        for g in self.get_active_groups():
            if g.license.id == member_id:
                member = g.license
                break
        if member is None:
            raise PublicUserException(_('Incorrect member supplied.'))

        if not isinstance(member, PvcMemberLicense):
            raise PublicUserException(_('You have to finish the Registration process before selecting PV Modules.'))

        if len(countries) == 0:
            raise PublicUserException(_('No countries provided.'))

        if module.approval_status == PvcPVModule.PV_MODULE_STATUS_REJECTED:
            raise PublicUserException(_('The provided module is not available for selection.'))

        rv = []
        c = PvcMemberController(self.sa_session, account_id=self.get_account().id)
        country_map = c.get_member_countries(member.id, True)
        for cid in countries:
            if not country_map.has_key(cid):
                raise PublicSecurityException(_('This member has no access to the provided country.'))
            tm, country = country_map[cid]
            mm = self.save_or_update_module_for_tbs_member(tm, country, module, subrogation_id=subrogation_id)

            if mm.subrogation is not None and mm.commissioner_approval_status == PvcTbsMemberPVModule.COMMISSIONER_APPROVAL_STATUS_REJECTED:
                raise PublicUserException(_('The provided module is not available for selection.'))

            rv.append(mm)

        return rv

    @transaction_decorator
    def save_or_update_module_for_tbs_member(self, tbs_member, country, module, is_active=PvcTbsMemberPVModule.IS_ACTIVE, subrogation_id=None):

        if tbs_member.tbs.country is None:
            assert country is not None
        else:
            assert country is tbs_member.tbs.country

        subrogation = None
        if subrogation_id is not None:

            if not isinstance(subrogation_id, PvcSubrogation):
                subrogation = self.sa_session.query(PvcSubrogation).filter(PvcSubrogation.id==subrogation_id).first()
            else:
                subrogation = subrogation_id
            if subrogation is None:
                raise PublicUserException(_('Subrogation not found.'))
            if subrogation.subrogator != tbs_member:
                raise PublicUserException(_('Incorrect subrogation supplied.'))
            if subrogation.status != PvcSubrogation.STATUS_ACTIVE:
                raise PublicUserException(_('You can only add PV Modules at an active subrogation.'))

        try:
            mm = self.sa_session.query(PvcTbsMemberPVModule).filter(and_(
                PvcTbsMemberPVModule.tbs_member==tbs_member,
                PvcTbsMemberPVModule.module==module,
                PvcTbsMemberPVModule.country==country,
                PvcTbsMemberPVModule.subrogation==subrogation,
            )).one()
        except NoResultFound:
            mm = PvcTbsMemberPVModule()
            mm.tbs_member = tbs_member
            mm.module = module
            mm.country = country
            mm.subrogation = subrogation
            mm.commissioner_approval_status = None if subrogation is None else PvcTbsMemberPVModule.COMMISSIONER_APPROVAL_STATUS_PENDING

        mm.is_active = is_active

        self.logger.info('PV Module (%s) selected by TBS Member (%s) for Country (%s). Used with Subrogation: %s' % (module.id, tbs_member.id, country.id, subrogation))

        self.sa_session.add(mm)
        return mm

    def _get_tbs_member_module(self, _id):
        return self.sa_session.query(PvcTbsMemberPVModule).filter(PvcTbsMemberPVModule.id==_id).one()

    def get_tbs_member_modules_by_tbs_member(self, module_id, tbs_member_id):
        return self.sa_session.query(PvcTbsMemberPVModule).filter(PvcTbsMemberPVModule.module_id==module_id, PvcTbsMemberPVModule.tbs_member_id==tbs_member_id).all()

    def get_tbs_member_modules_per_country_state(self, module_id, tbs_member_id):

        tbs_member = self._get_tbs_member_licenses(tbs_member_id)[0]

        modules = self.get_tbs_member_modules_by_tbs_member(module_id, tbs_member_id)
        rv = self.load_module_state(module_id)
        rv['member_id'] = tbs_member.member.id
        rv['countries'] = []
        for m in modules:
            rv['countries'].append(m.country_id)
        return rv

    @transaction_decorator
    def change_tbs_member_module_active_state(self, member_module, is_active=PvcTbsMemberPVModule.IS_INACTIVE):
        if not isinstance(member_module, PvcTbsMemberPVModule):
            member_module = self._get_tbs_member_module(member_module)

        member_module.is_active = is_active
        self.sa_session.add(member_module)
        self.sa_session.flush()
        return member_module

    def get_by_photon_id(self, photon_id):
        """
        Get a PV Module by the provided Photon ID.
        """
        return self.sa_session.query(PvcPVModule).filter(PvcPVModule.photon_id==photon_id).first()

    def save_or_update_from_photon_db(self, photon_id, supplier, _type, cell_type, nominal_power, weight, width, length):
        # check if there already is a module in our DB for this Photon ID
        module = self.get_by_photon_id(photon_id)

        # if there isn't get or create module
        if module is None:
            module = self.get_or_create_module(supplier, _type, cell_type, nominal_power, weight, width, length, photon_id, abstract=True)
            # make sure there wasn't already a photon_id for this SUPPLIER/TYPE.
            if photon_id != module.photon_id:
                raise PhotonDBImportException('Photon ID in import file does not match with Photon ID in database for supplier/module type combination.')
            # Update
            module = self._save_or_update_module(module.id, supplier, _type, cell_type, nominal_power, weight, width, length, photon_id)
            # auto approve:
            module.approval_status = PvcPVModule.PV_MODULE_STATUS_APPROVED
            self.sa_session.add(module)
            return module

        # Supplier and Module Type should be matched.
        if module.supplier.name != supplier:
            raise PhotonDBImportException('Brand in import file does not match with the Brand for this Photon ID in database.')
        if module.type.name != _type:
            raise PhotonDBImportException('Module type in import file does not match with the Module type for this Photon ID in database.')

        # update possible changes.
        module = self._save_or_update_module(module.id, supplier, _type, cell_type, nominal_power, weight, width, length, photon_id)
        # auto approve:
        module.approval_status = PvcPVModule.PV_MODULE_STATUS_APPROVED
        self.sa_session.add(module)

    @transaction_decorator
    def save_import_file(self, tmp_file_hash):
        self.assert_permissions('pvc_import_pv_modules')

        tfc = UploadFileController(self.sa_session)
        tmp_file = tfc.get_hash_temp_files(tmp_file_hash).pop()
        pfc = ProcessFileController(self.sa_session, account_id=self.get_account().id)
        category = pfc.get_or_create_file_category('PV Module imports', self.get_default_group())

        _type = self.sa_session.query(ImportFileType).filter(ImportFileType.name=='pvcycle.pvmodule').one()

        pfc.save_files_from_temp([tmp_file], category, ImportFile, {
            'status': ImportFile.STATUS_NEW,
            'settings': '{}',
            'type': _type
        })
        # Trigger task
        client = xmlrpclib.Server(settings.SCHEDULER_SERVER_XMLRPC, use_datetime=True, allow_none=True)
        try:
            client.start_task('pvcycle.pvmodule.schedule.ImportPVModuleTask', {})
        except Fault, e:
            raise PublicUserException(e.faultString)

        return True

    @transaction_decorator
    def import_pv_modules(self, import_file, dryrun=True):
        """
        Use importer to insert/match/update PV Modules as provided from the PHOTON.INFO database.
        Called from Scheduler
        """
        importer = self._get_importer(import_file)
        file_path = os.path.join(settings.FILE_REPOSITORY, import_file.file_hash)

        _settings = simplejson.loads(import_file.settings, object_hook=json_object_hook)

        modules, failures = importer.import_file(self, file_path, _settings, dryrun)
        if dryrun:
            raise Exception('Rollback')

        validation_file = None
        if len(failures) > 0:
            validation_file = self._write_failure_files(import_file, importer.get_column_headers(), failures)

        return modules, validation_file

    def _write_failure_files(self, import_file, column_headers, failures):
        ffe = FailureFileExporter()
        column_headers.append('failure')
        failures.insert(0, column_headers)
        file_path, file_name = ffe.export(import_file.file_name, failures)
        with file(file_path) as f:
            data = f.read()

        afc = AnonymousFileController(self.sa_session)
        return afc.save_file_for_target_group(data, file_name, 'Photon Import Failures', import_file.groups[0].group, File)

    def set_import_result(self, import_file, status, result, failure_file=None):
        import_file.status = status
        import_file.result = unicode(result)
        import_file.failure_file = failure_file
        self.sa_session.add(import_file)
        self.sa_session.flush()

    def _get_importer(self, tmp_file):
        assert isinstance(tmp_file, ImportFile)

        file_path = os.path.join(settings.FILE_REPOSITORY, tmp_file.file_hash)
        assert os.path.exists(file_path), 'file not found: %s' % file_path

        base_filename, file_extension = os.path.splitext(tmp_file.file_name)

        importer = None
        if file_extension in ('.csv', '.txt'):
            importer = PhotonDBCsvImporter()

        elif file_extension == '.xlsx':
            importer = PhotonDBXlsxImporter()

        else:
            raise PublicUserException(_('Cannot import for extension: %s') % file_extension)

        return importer

    def get_countries_for_module(self, module_id, tbs_member_id, subrogation_id, active=PvcTbsMemberPVModule.IS_ACTIVE):
        members = self._get_tbs_member_licenses()

        assert tbs_member_id in [m.id for m in members]

        modules = self.sa_session.query(PvcTbsMemberPVModule).filter(
            and_(
                PvcTbsMemberPVModule.module_id==module_id,
                PvcTbsMemberPVModule.tbs_member_id==tbs_member_id,
                PvcTbsMemberPVModule.is_active==active,
                PvcTbsMemberPVModule.subrogation_id==subrogation_id
            )
        ).all()

        return [m.country for m in modules]

    @transaction_decorator
    def deactivate_module(self, module_id, tbs_member_id, subrogation_id, countries):
        members = self._get_tbs_member_licenses()

        assert tbs_member_id in [m.id for m in members]

        modules = self.sa_session.query(PvcTbsMemberPVModule).filter(
            and_(
                PvcTbsMemberPVModule.module_id==module_id,
                PvcTbsMemberPVModule.tbs_member_id==tbs_member_id,
                PvcTbsMemberPVModule.subrogation_id==subrogation_id,
                PvcTbsMemberPVModule.country_id.in_(countries)
            )
        ).all()

        assert len(modules) == len(countries)

        for module in modules:
            module.is_active = PvcTbsMemberPVModule.IS_INACTIVE
            self.sa_session.add(module)

        self.sa_session.flush()

    @transaction_decorator
    def disable_pv_modules_for_all_non_eu_countries(self):
        """
        This will check the current list of selected EU countries.
        All countries not in this list selected for TBS ROE will be disabled.
        """
        all = self.get_efta_countries()
        current = self.get_roe_countries()
        disable = []
        for country in all:
            if country not in current:
                disable.append(country)

        if len(disable) == 0:
            return

        tbss = self.sa_session.query(PvcTbsLicense).filter(PvcTbsLicense.country==None).all()

        # we use a text query to disable them in one go.
        sql = """
            UPDATE tblPvcTbsMemberPVModule as mpvm
            INNER JOIN tblPvcTbsMember as tm
              ON mpvm.tbs_member_id = tm.id
                AND tm.tbs_id IN (%(tbs_id)s)
            SET mpvm.is_active = 0
            WHERE mpvm.country_id IN (%(countries)s)
        """ % {
            'tbs_id': ','.join(['%s' % tbs.id for tbs in tbss]),
            'countries': ','.join(['%s' % c.id for c in disable])
        }
        self.logger.info('Disabling PV Modules for the following countries at TBS ROE %(countries)s' % {
            'countries': ', '.join([c.name for c in disable])
        })

        self.sa_session.execute(sql)

        # when done -> send notification to operator.
        nc = NotificationController(self.sa_session)
        operators = self.sa_session.query(PvcOperatorLicense).all()
        recipients = [o.default_group for o in operators]
        nc.create_system_notification(nc.set_notification(
            'Automatic disabling of RoE removed countries complete',
            'The PV Modules for the following countries at TBS ROE were disabled: %(countries)s' % {
                'countries': ', '.join([c.name for c in disable])
            }
        ), recipients)
