'''
Created on Aug 27, 2013

@author: marten
'''
from django.utils.translation import ugettext as _

from sqlalchemy import func, case, and_
from sqlalchemy.types import Integer
from sqlalchemy.orm.util import aliased
from sqlalchemy.sql.expression import cast

from basesoft.web.core.util.custom_query import CustomQueryField, CustomQueryModel, CustomQuery, DATA_TYPE_INT,\
    DATA_TYPE_BOOLEAN
from basesoft.web.contrib.account.models import Account
from basesoft.web.contrib.i18n.models import Country
from basesoft.web.contrib.file.custom_query import AbstractFileImportCustomQuery
from basesoft.web.core.util.db import group_concat
from basesoft.web.contrib.license.models import License
from basesoft.web.contrib.order.models import Currency, OrderPaymentStatus

from pvcycle.pvmodule.models import PvcPVModule, PvcPVModuleSupplier, PvcPVModuleType, PvcPVModuleCellType, PvcTbsMemberPVModule
from pvcycle.core.models import PvcMemberLicense, PvcTbsMember
from pvcycle.core.custom_query import AbstractPvcCustomQuery
from pvcycle.order.models import PvcDeclarationOrder, PvcDeclarationOrderLine
from pvcycle.subrogation.models import PvcSubrogation

SecondAccount = aliased(Account)

class PVModuleCustomQuery(CustomQuery):

    permissions = ['pvc_manage_pv_modules']

    def tables(self):
        return (
            CustomQueryModel(PvcPVModule),
            CustomQueryModel(PvcPVModuleSupplier),
            CustomQueryModel(PvcPVModuleType),
            CustomQueryModel(PvcPVModuleCellType),
            CustomQueryModel(Account, onclause=Account.id==PvcPVModule.first_approval_by_id, use_outerjoin=True),
            CustomQueryModel(SecondAccount, onclause=SecondAccount.id==PvcPVModule.second_approval_by_id, use_outerjoin=True),
        )

    def fields(self):
        return (
            CustomQueryField(PvcPVModule.id, _('ID'), key='id'),
            CustomQueryField(PvcPVModule.photon_id, _('Photon ID'), key='photon_id', is_filter=True),
            CustomQueryField(PvcPVModule.id, _('Details'), key='details'),
            CustomQueryField(PvcPVModule.approval_status, _('Status'), key='status', is_filter=True, filter_options=[{
                'value': PvcPVModule.PV_MODULE_STATUS_PENDING,
                'label': _("Pending"),
            },{
                'value': PvcPVModule.PV_MODULE_STATUS_APPROVED,
                'label': _("Approved"),
            },{
                'value': PvcPVModule.PV_MODULE_STATUS_REJECTED,
                'label': _("Rejected"),
            }]),
            CustomQueryField(PvcPVModuleType.name, _('Module type'), key='type', is_filter=True),
            CustomQueryField(PvcPVModuleSupplier.name, _('Brand'), key='supplier', is_filter=True),
            CustomQueryField(PvcPVModuleCellType.name, _('Cell type'), key='cell_type', is_filter=True),
            CustomQueryField(PvcPVModule.nominal_power, _('Nominal power'), key='nominal_power', is_filter=True, type=DATA_TYPE_INT),
            CustomQueryField(PvcPVModule.weight, _('Weight'), key='weight', is_filter=True, type=DATA_TYPE_INT),
            CustomQueryField(PvcPVModule.width, _('Width'), key='width', is_filter=True, type=DATA_TYPE_INT),
            CustomQueryField(PvcPVModule.length, _('Length'), key='length', is_filter=True, type=DATA_TYPE_INT),
            CustomQueryField(Account.name, _('First approval'), key='first_approval'),
            CustomQueryField(SecondAccount.name, _('Second approval'), key='second_approval'),
        )

Commissioner = aliased(PvcTbsMember)
CommissionerLicense = aliased(PvcMemberLicense)

class PVModuleCountryCustomQuery(AbstractPvcCustomQuery):

    permissions = ['pvc_select_pv_modules', 'pvc_manage_pv_modules']

    def validate_access(self):
        self.assert_any_permissions(*self.permissions)

    def tables(self):
        return (
            CustomQueryModel(PvcTbsMemberPVModule),
            CustomQueryModel(PvcPVModule),
            CustomQueryModel(PvcPVModuleSupplier),
            CustomQueryModel(PvcPVModuleType),
            CustomQueryModel(PvcPVModuleCellType),
            CustomQueryModel(PvcTbsMember, onclause=PvcTbsMemberPVModule.tbs_member_id==PvcTbsMember.id),
            CustomQueryModel(PvcMemberLicense, onclause=PvcTbsMember.member_id==PvcMemberLicense.id),
            CustomQueryModel(Country, onclause=PvcTbsMemberPVModule.country_id==Country.id),
            CustomQueryModel(PvcSubrogation, onclause=PvcTbsMemberPVModule.subrogation_id==PvcSubrogation.id, use_outerjoin=True),
            CustomQueryModel(Commissioner, onclause=PvcSubrogation.commissioner_id==Commissioner.id, use_outerjoin=True),
            CustomQueryModel(CommissionerLicense, onclause=Commissioner.member_id==CommissionerLicense.id, use_outerjoin=True),
        )

    def fields(self):
        from basesoft.web.contrib.i18n.sa_datamapper import get_tables as i18n_tables
        return (
            CustomQueryField(PvcPVModule.id, _('ID'), key='id'),
            CustomQueryField(PvcMemberLicense.name, _('Member'), key='member', is_filter=True, is_group_by=True),
            CustomQueryField(PvcTbsMemberPVModule.subrogation_id, 'Subrogation', key='subrogation_id'),
            CustomQueryField(CommissionerLicense.name, _('Commissioner'), key='commissioner', is_filter=True),
            CustomQueryField(PvcTbsMember.id, 'tbs member id', key='tbs_member_id'),
            CustomQueryField(PvcPVModuleType.name, _('Module type'), key='type', is_filter=True, is_group_by=True),
            CustomQueryField(group_concat(i18n_tables()['tblCountry'].c.name, i18n_tables()['tblCountry'].c.name), _('Countries'), key='countries'),
            CustomQueryField(Country.id, _('Country'), key='country_id', in_select=False, is_filter=True, filter_options=Country.name, filter_is_unique=True),
            CustomQueryField(PvcPVModuleSupplier.name, _('Brand'), key='supplier', is_filter=True, is_group_by=True),
            CustomQueryField(PvcPVModuleCellType.name, _('Cell type'), key='cell_type', is_filter=True, is_group_by=True),
            CustomQueryField(PvcPVModule.nominal_power, _('Nominal power'), key='nominal_power', is_filter=True, type=DATA_TYPE_INT),
            CustomQueryField(PvcPVModule.weight, _('Weight'), key='weight', is_filter=True, type=DATA_TYPE_INT),
            CustomQueryField(PvcPVModule.width, _('Width'), key='width', is_filter=True, type=DATA_TYPE_INT),
            CustomQueryField(PvcPVModule.length, _('Length'), key='length', is_filter=True, type=DATA_TYPE_INT),
        )

    def set_where(self):
        self.where = and_(PvcMemberLicense.id.in_([x.id for x in self._get_member_licenses()]), PvcTbsMemberPVModule.is_active==PvcTbsMemberPVModule.IS_ACTIVE)

    group_by = (PvcPVModule.id, PvcTbsMember.member_id, PvcTbsMemberPVModule.subrogation_id, )

class FileImportCustomQuery(AbstractFileImportCustomQuery):
    IMPORT_FILE_TYPE ='pvcycle.pvmodule'


Tbs = aliased(License)

class PVModuleDeclarationCustomQuery(AbstractPvcCustomQuery):
    permissions = ['pvc_manage_pv_modules', 'pvc_view_declarations']

    # this shows an overview per pv module #411
    def tables(self):
        return (
            CustomQueryModel(PvcTbsMemberPVModule),
            CustomQueryModel(PvcPVModule),
            CustomQueryModel(PvcPVModuleSupplier),
            CustomQueryModel(PvcPVModuleType),
            CustomQueryModel(PvcPVModuleCellType),
            CustomQueryModel(PvcTbsMember, onclause=PvcTbsMemberPVModule.tbs_member_id==PvcTbsMember.id),
            CustomQueryModel(PvcMemberLicense, onclause=PvcTbsMember.member_id==PvcMemberLicense.id),
            CustomQueryModel(Country, onclause=PvcTbsMemberPVModule.country_id==Country.id),
            CustomQueryModel(Tbs, onclause=PvcTbsMember.tbs_id==Tbs.id),
            CustomQueryModel(PvcDeclarationOrderLine, onclause=PvcTbsMemberPVModule.id==PvcDeclarationOrderLine.tbs_member_module_id, use_outerjoin=True),
            CustomQueryModel(PvcDeclarationOrder, onclause=PvcTbsMember.id==PvcDeclarationOrder.tbs_member_id, use_outerjoin=True),
            CustomQueryModel(Currency, onclause=PvcDeclarationOrder.currency_id==Currency.id, use_outerjoin=True),
        )

    def fields(self):
        return (
            CustomQueryField(PvcMemberLicense.name, _('Member'), key='member', is_filter=True, is_group_by=True),
            CustomQueryField(Tbs.name, _('TBS'), key='tbs', is_filter=True, is_group_by=True),
            CustomQueryField(PvcPVModule.photon_id, _('Photon ID'), key='photon_id', is_filter=True),
            CustomQueryField(PvcPVModuleSupplier.name, _('Brand'), key='supplier', is_filter=True, is_group_by=True),
            CustomQueryField(PvcPVModuleType.name, _('Module type'), key='type', is_filter=True, is_group_by=True),
            CustomQueryField(PvcPVModuleCellType.name, _('Cell type'), key='cell_type', is_filter=True, is_group_by=True),
            CustomQueryField(Country.name, _('Country'), key='country', is_filter=True, filter_options=Country.name, filter_is_unique=True),
            CustomQueryField(cast(func.sum(PvcDeclarationOrderLine.quantity), Integer), _('Total quantity'), key='quantity', type=DATA_TYPE_INT),
            CustomQueryField(func.sum(PvcDeclarationOrderLine.line_amount_excluding_vat), _('Total amount excl. VAT'), key='total_amount'),
            CustomQueryField(func.ifnull(Currency.symbol, Currency.code), _('Currency'), 'currency'),
            CustomQueryField(case([(PvcDeclarationOrder.id==None, 0)], else_=1), _('In declaration'), key='in_declaration', is_filter=True, is_group_by=True, type=DATA_TYPE_BOOLEAN),
            CustomQueryField(PvcDeclarationOrder.year, _('Year'), key='year', is_filter=True, is_group_by=True),
            CustomQueryField(PvcDeclarationOrder.month, _('Month'), key='month', is_filter=True, is_group_by=True),
            CustomQueryField(PvcDeclarationOrder.quarter, _('Quarter'), key='quarter', is_filter=True, is_group_by=True),
            CustomQueryField(func.iqGetCurrentOrderPaymentStatus(PvcDeclarationOrder.id), _('Payment status'), key='payment_status', is_filter=True, filter_options=[{
                'value': OrderPaymentStatus.PAYMENT_PENDING,
                'label': _("Pending"),
            },{
                'value': OrderPaymentStatus.PAYMENT_SUBMITTED,
                'label': _("Submitted"),
            },{
                'value': OrderPaymentStatus.PAYMENT_PARTIALLY_RECEIVED,
                'label': _("Partially received"),
            },{
                'value': OrderPaymentStatus.PAYMENT_COMPLETED,
                'label': _("Completed"),
            },{
                'value': OrderPaymentStatus.PAYMENT_REJECTED,
                'label': _("Rejected"),
            }]),
            CustomQueryField(PvcPVModule.approval_status, _('Approval status'), key='status', is_filter=True, filter_options=[{
                'value': PvcPVModule.PV_MODULE_STATUS_PENDING,
                'label': _("Pending"),
            },{
                'value': PvcPVModule.PV_MODULE_STATUS_APPROVED,
                'label': _("Approved"),
            },{
                'value': PvcPVModule.PV_MODULE_STATUS_REJECTED,
                'label': _("Rejected"),
            }]),
        )

    def set_where(self):
        self.where = PvcMemberLicense.id.in_([x.id for x in self._get_member_licenses()])
        self.where = and_(self.where, PvcTbsMember.status == PvcTbsMember.STATUS_APPROVED)
        if not self.has_permissions('view_domain') and self.has_permissions('pvc_tbs_member_manager'):
            self.where = and_(self.where, PvcTbsMember.tbs_id.in_([x.id for x in self._get_take_back_schemes(permissions=['pvc_tbs_member_manager'])]))

    group_by = (PvcPVModule.id, Tbs.id, Country.id, PvcMemberLicense.id, )

custom_queries = (PVModuleCustomQuery, PVModuleCountryCustomQuery, FileImportCustomQuery, PVModuleDeclarationCustomQuery, )
