'''
Created on Dec 3, 2013

@author: marten
'''
import datetime
import csv, codecs, os, cStringIO

from django.conf import settings

class PhotonDBImportException(Exception):
    pass

class PhotonDBAbstractImporter:
    """
    Based on Lucid Importers.
    """
    _failure_lines = []
    _column_headers = []

    def import_file(self, controller, file_path, import_options, dryrun=False):
        self.controller = controller

        self.reset_failure_lines()
        self.set_column_mapping(file_path, import_options)

        modules = self._import_file(file_path, import_options, dryrun);
        return modules, self._failure_lines

    def get_column_headers(self):
        return self._column_headers

class PhotonDBLineImporter(PhotonDBAbstractImporter):

    REQUIRED_COLUMNS = ('ID Module', 'Supplier', 'Module type', 'Cell type', 'Nominal power', 'Weight', 'Width', 'Length', )

    def reset_failure_lines(self):
        self._failure_lines = []

    def set_column_mapping(self, file_path, import_options):
        sample = self.get_sample(file_path, import_options, lines=1)
        self._column_headers = sample[0]

        self._column_mapping = {}

        col_i = 0
        for column in sample[0]:
            column = column.strip(u'\ufeff')
            if column in PhotonDBLineImporter.REQUIRED_COLUMNS:
                self._column_mapping[ column ] = col_i
            col_i += 1

        # assert we have ALL columns
        for column in PhotonDBLineImporter.REQUIRED_COLUMNS:
            assert column in self._column_mapping.keys(), 'Required column not found "%s"' % column

    def _import_file(self, file_path, import_options, dryrun):
        row_i = 0
        rv = []
        for row in self._read_row(file_path, import_options):
            if row_i == 0: # Dont import the header line.
                row_i += 1
                continue
            try:
                rv.append(self._import_row(row, dryrun))
            except PhotonDBImportException, e:
                self._failure_lines.append(row+[e.message])
            row_i += 1
        return rv

    def _import_row(self, row, dryrun):
        # validate row:
        args = []
        for column in self.REQUIRED_COLUMNS:
            index = self._column_mapping[column]
            try:
                value = row[index].strip()
            except IndexError:
                value = ''
            if value == '':
                raise PhotonDBImportException('Column "%s" is empty.' % column)
            args.append(value)

        return self.controller.save_or_update_from_photon_db(*args)

    def get_sample(self, file_path, import_options, lines):

        rv = []
        count = 0
        for row in self._read_row(file_path, import_options):
            rv.append(row)

            count = count + 1

            if count == lines:
                return rv

        return rv

class PhotonDBCsvImporter(PhotonDBLineImporter):
    """
    Importer for CSV
    """
    class UTF8Recoder:
        def __init__(self, f, encoding):
            self.reader = codecs.getreader(encoding)(f)

        def __iter__(self):
            return self

        def next(self):
            return self.reader.next().encode("utf-8")

    class UnicodeCsvReader:

        def __init__(self, f, dialect=csv.excel, encoding="utf-8", **kwds):
            f = PhotonDBCsvImporter.UTF8Recoder(f, encoding)
            self.reader = csv.reader(f, dialect=dialect, **kwds)

        def next(self):
            row = self.reader.next()
            return [unicode(s, "utf-8") for s in row]

        def __iter__(self):
            return self

    def _read_row(self, file_path, import_options):

        if import_options is None:
            import_options = {}

        csvfile = open(file_path, "rb")

        dialect = None
        if import_options.has_key('dialect'):
            dialect = import_options['dialect']
        else:
            # sniff 10MB. it goes wrong if a value has a comma in it which can be way up ahead the file
            dialect = csv.Sniffer().sniff(csvfile.read(1024 * 1024 * 10))
            csvfile.seek(0)

        reader = PhotonDBCsvImporter.UnicodeCsvReader(csvfile, dialect=dialect)

        for row in reader:
            yield row

class PhotonDBXlsxImporter(PhotonDBLineImporter):
    """
    Importer for XLSX
    """

    def _read_row(self, file_path, import_options):

        from openpyxl.reader.excel import load_workbook

        wb = load_workbook(filename = file_path, use_iterators = True)
        ws = wb.get_active_sheet()

        for row in ws.iter_rows():
            values = []

            for cell in row:
                value = cell.internal_value
                if  value is None:
                    value = ''

                values.append(unicode(value))

            yield values


class FailureFileExporter:
    """
    csv export controller

    see UnicodeWriter from http://docs.python.org/library/csv.html
    """

    _writer = None

    def export(self, file_name, lines):
        file_name = 'failures_%s.csv' % file_name
        file_path = os.path.join(settings.TMP_FILE_REPOSITORY, file_name)
        file_ = open(file_path, 'wb')

        self._queue = cStringIO.StringIO()
        self._writer = csv.writer(self._queue, delimiter="\t", quotechar='"', quoting=csv.QUOTE_ALL)
        self._stream = file_
        self._encoder = codecs.getincrementalencoder('utf-8')()

        for line in lines:
            self._write_row(line)

        file_.close()
        return file_path, file_name

    def _write_row(self, row):
        self._writer.writerow(['' if s is None else unicode(s).encode("utf-8") for s in row])
        data = self._queue.getvalue()
        data = data.decode("utf-8")
        data = self._encoder.encode(data)
        self._stream.write(data)
        self._queue.truncate(0)
