'''
Created on Aug 26, 2013

@author: marten
'''
from django.utils.translation import ugettext as _

from basesoft.web.core.forms.handler import FormValidator
from basesoft.web.core.input.validation import plugins
from basesoft.web.core.exceptions import PublicUserException

from pvcycle.pvmodule.controllers import PvcPVModuleController

AUTOCOMPLETE_LIMIT = 40

class PVModuleHandler:
    def __init__(self, request, **kwargs):

        self._request = request
        self._sa_session = request.Session()
        self._account_id = request.session['user_id']

        self._controller = PvcPVModuleController(self._sa_session, account_id=self._account_id, request=request)

    def get_pv_module_by_name(self, name):
        module = self._controller.get_modules(name=name).pop()
        return module.id

    def get_modules(self, exclude=None):
        if exclude is None:
            exclude = []
        return [{'label': i.name, 'value': i.id} for i in self._controller.get_modules() if i.name not in exclude]

    def get_suppliers(self, contains=None, limit=AUTOCOMPLETE_LIMIT):
        return [i.name for i in self._controller.get_suppliers(limit=limit, contains=contains)]

    def get_types(self, contains=None, supplier=None, limit=AUTOCOMPLETE_LIMIT):
        return [i.name for i in self._controller.get_module_types(supplier=supplier, limit=limit, contains=contains)]

    def get_cell_types(self, contains=None, limit=AUTOCOMPLETE_LIMIT):
        return [i.name for i in self._controller.get_cell_types(limit=limit, contains=contains)]

    def load_pv_module(self, _id):
        return self._controller.load_module_state(_id)

    def load_pv_module_by_supplier_and_type(self, supplier, _type):
        module = self._controller.get_modules(supplier_name=supplier, type_name=_type)
        if len(module) > 0:
            return self._controller.load_module_state(module.pop().id)
        return False

    def save_pv_module(self, _id, args):
        if not self._controller.has_permissions('pvc_version_1_access'):
            raise DeprecationWarning('This is no longer available (PV CYCLE 2.0)')
        args['_type'] = args['type']
        del args['type']
        self._controller.save_or_update_module(_id, **args)
        return True

    def save_country_pv_module(self, member_id, countries, module, subrogation_id):
        if not self._controller.has_permissions('pvc_version_1_access'):
            raise DeprecationWarning('This is no longer available (PV CYCLE 2.0)')
        if countries is None:
            raise PublicUserException(_('Please select at least one country'))
        module['_type'] = module['type']
        del module['type']
        module = self._controller.get_or_create_module(**module)
        self._controller.save_or_update_module_for_countries(member_id, [int(c) for c in countries], module, subrogation_id)
        return {'approval_status': module.approval_status, 'id': module.id}

    def load_country_pv_module(self, tbs_member_id, module_id):
        return self._controller.get_tbs_member_modules_per_country_state(module_id, tbs_member_id)

    def approve_pv_module(self, _id):
        if not self._controller.has_permissions('pvc_version_1_access'):
            raise DeprecationWarning('This is no longer available (PV CYCLE 2.0)')
        self._controller.approve_module(_id)
        return True

    def reject_pv_module(self, _id):
        if not self._controller.has_permissions('pvc_version_1_access'):
            raise DeprecationWarning('This is no longer available (PV CYCLE 2.0)')
        self._controller.reject_module(_id)
        return True

    def validate_new_pv_module(self, args):
        return self.validate_pv_module(args, new=True)

    def validate_pv_module(self, args, new=False, checks=None):
        check = {
            'supplier': [(plugins.Required, {})],
            'type': [(plugins.Required, {})],
            'cell_type': [(plugins.Required, {})],
            'nominal_power': [(plugins.Required, {}), (plugins.Number, {'min_value': 0, 'inclusive': True})],
            'weight': [(plugins.Required, {}), (plugins.Number, {'min_value': 0})],
            'length': [(plugins.Required, {}), (plugins.Number, {'min_value': 0})],
            'width': [(plugins.Required, {}), (plugins.Number, {'min_value': 0})],
        }
        if checks is not None:
            check = dict(check.items()+checks.items())

        form_handler = FormValidator(check)
        return form_handler.validate(args)

    def validate_pv_module_select(self, args):
        return self.validate_pv_module(args, checks={'countries': [(plugins.Required, {}), ]})

    def import_photon_db(self, file_hash):
        if not self._controller.has_permissions('pvc_version_1_access'):
            raise DeprecationWarning('This is no longer available (PV CYCLE 2.0)')
        self._controller.save_import_file(file_hash)
        return True

    def get_unapproved_commissioner_pv_modules_count(self):
        return self._controller.get_unapproved_commissioner_pv_modules_count()

    def get_selected_countries(self, module_id, tbs_member_id, subrogation_id):
        countries = self._controller.get_countries_for_module(module_id, tbs_member_id, subrogation_id)
        return [{'value': c.id, 'label': c.name} for c in countries]

    def deactivate_module(self, module_id, tbs_member_id, subrogation_id, countries):
        if not self._controller.has_permissions('pvc_version_1_access'):
            raise DeprecationWarning('This is no longer available (PV CYCLE 2.0)')
        self._controller.deactivate_module(module_id, tbs_member_id, subrogation_id, countries)
        return True

handlers = (PVModuleHandler, )
