# Copyright (c) 2014 BaseSoft B.V.
# See LICENSE for details.
from django.utils.translation import ugettext as _
from sqlalchemy.exc import IntegrityError

from basesoft.util.virtual_date import today
from basesoft.util.ordered_dict import OrderedDict
from basesoft.web.core.util.controllers import AbstractController, transaction_decorator, LicenseConfigController
from basesoft.web.core.exceptions import PublicUserException, SecurityException
from basesoft.web.contrib.i18n.models import Country, EFTA_COUNTRIES

from iqportal.iqube.core.controllers import DataSourceController, MondrianMemberSerialize

from pvcycle.core.controllers import AbstractPvcController
from pvcycle.report.models import PvcMemberYearlyAggregatedData

class PvcReportController(AbstractController, AbstractPvcController):
    def get_efta_countries(self, datasource, cube, config, values):
        dsc = DataSourceController(self.sa_session, account_id=self.account_id)
        ds = dsc.get_datasource(datasource)

        client = ds.connect(self.sa_session, self.get_account().id)

        countries = self.sa_session.query(Country).filter(Country.iso_3166_a2.in_(EFTA_COUNTRIES)).all()

        def _serialize_member(x):
            # separators used to be compact like the JS one.
            return {'label': x['member_caption'], 'value': MondrianMemberSerialize(x)}

        rv = OrderedDict()
        members = []

        for country in countries:
            member = client.get_member_tree_op(cube, '[Country].[%d]' % country.id, client.MDTREEOP_SELF, "MEMBER_UNIQUE_NAME").pop()
            members.append(_serialize_member(member))

        rv.update([('Country', members, )])

        return rv

class PvcHistoricalDataController(AbstractController, AbstractPvcController):

    @transaction_decorator
    def save_or_update_historical_data(self, _id, tbs_member_id, year, country_id, megawatt, quantity, weight, b2b, b2c, _type=PvcMemberYearlyAggregatedData.TYPE_MANUAL):

        country = self.sa_session.query(Country).filter(Country.id==country_id).first()
        permission = 'pvc_create_historical_data' if _id is None else 'pvc_update_historical_data'
        tbs_member = self._get_tbs_member_licenses(tbs_member_id, [permission])
        if len(tbs_member) == 0:
            raise SecurityException('No access to TBS Member')
        tbs_member = tbs_member.pop()

        lcc = LicenseConfigController(self.sa_session, scope=tbs_member.tbs)
        minimum_year = lcc.get('pvc.tbs.minimum.year.for.historical.data')
        year = int(year)
        if year < minimum_year:
            raise PublicUserException(_('The supplied year is not available at this TBS.'))

        accepted_countries = self._get_tbs_member_countries(tbs_member)
        if country not in accepted_countries:
            raise PublicUserException(_('The supplied country is not a valid country for this TBS Member.'))

        quantity = int(quantity)

        assert megawatt >= 0, 'No negative power allowed'
        assert quantity >= 0, 'No negative quantity allowed'
        assert weight >= 0, 'No negative weight allowed'
        assert b2b >= 0, 'No negative b2b distribution allowed'
        assert b2c >= 0, 'No negative b2c distribution allowed'

        # if the quantity is zero we make sure ALL fields are 0
        if quantity == 0:
            megawatt = 0
            weight = 0
            b2b = 0
            b2c = 0

        # member should be approved at least at one TBS.
        if _id is None:
            self.assert_any_permissions('pvc_manage_historical_data', 'pvc_create_historical_data')
            # TODO: verify the supplied year is allowed for inserting if _type is manual.
            if year >= tbs_member.created.year:
                raise PublicUserException(_('The supplied year is not allowed for this TBS Member.'))

            hd = PvcMemberYearlyAggregatedData()
            hd.tbs_member = tbs_member
            hd.year = year
            hd.country = country
            hd.type = _type
        else:
            self.assert_any_permissions('pvc_manage_historical_data', 'pvc_update_historical_data')
            hd = self.sa_session.query(PvcMemberYearlyAggregatedData).filter(PvcMemberYearlyAggregatedData.id==_id).one()
            if hd.tbs_member != tbs_member:
                raise PublicUserException(_('The supplied TBS Member does not match the historical data.'))
            if hd.year != year:
                raise PublicUserException(_('The supplied year does not match the historical data.'))
            if hd.country != country:
                raise PublicUserException(_('The supplied country does not match the historical data.'))

        if quantity != 0 and (b2b + b2c != 100):
            raise PublicUserException(_('The supplied B2B- and B2C-distribution do not add up to 100%.'))

        # Do not save megawatt, quantity and weight for automated aggregation if a user is updating this, that should only happen in a task
        if hd.type != PvcMemberYearlyAggregatedData.TYPE_AUTOMATED or self.get_account() is not None:
            hd.total_megawatt = megawatt
            hd.total_quantity = quantity
            hd.total_weight = weight

        hd.distribution_b2b = b2b
        hd.distribution_b2c = b2c

        self.sa_session.add(hd)
        try:
            self.sa_session.flush()
        except IntegrityError:
            raise PublicUserException(_('An entry already exists for this TBS Member, year and country.'))

        return hd

    @transaction_decorator
    def delete_historical_data(self, _id):
        self.assert_any_permissions('pvc_manage_historical_data', 'pvc_delete_historical_data')

        hd = self.sa_session.query(PvcMemberYearlyAggregatedData).filter(PvcMemberYearlyAggregatedData.id==_id).one()
        self._get_tbs_member_licenses(hd.tbs_member.id, ['pvc_delete_historical_data'])

        self.sa_session.delete(hd)

    @transaction_decorator
    def save_from_state(self, tbs_member_id, year, countries, values):
        rv = []
        for country_id in countries:
            props = {}
            for prop in ['megawatt', 'quantity', 'weight', 'b2b', 'b2c']:
                if not values.has_key('%s_%s' % (prop, country_id)):
                    raise PublicUserException(_('Incorrect values supplied for saving of Historical Data.'))
                props[prop] = values['%s_%s' % (prop, country_id)]

            rv.append(self.save_or_update_historical_data(None, tbs_member_id, year, country_id, **props))
        return rv

    def load_historical_data_state(self, _id):
        self.assert_any_permissions('pvc_manage_historical_data', 'pvc_update_historical_data')

        hd = self.sa_session.query(PvcMemberYearlyAggregatedData).filter(PvcMemberYearlyAggregatedData.id==_id).one()

        tbs_member = self._get_tbs_member_licenses(hd.tbs_member.id, ['pvc_update_historical_data'])
        if len(tbs_member) == 0:
            raise SecurityException('No access to TBS Member')

        tbs_member = tbs_member.pop()

        return {
            'member_id': tbs_member.member.id,
            'tbs_id': tbs_member.tbs.id,
            'year': hd.year,
            'country_id': hd.country.id,
            'megawatt': hd.total_megawatt,
            'quantity': hd.total_quantity,
            'weight': hd.total_weight,
            'b2b': hd.distribution_b2b,
            'b2c': hd.distribution_b2c,
            'type': hd.type
        }

    def get_available_years_at_tbs(self, tbs_member, exclude_supplied_years=False):
        """
        The years available are the years between the TBS  pvc.tbs.minimum.year.for.historical.data and the tbs_member.created.year.

        We also exclude the years for which we already have
        """
        lcc = LicenseConfigController(self.sa_session, scope=tbs_member.tbs)
        minimum_year = lcc.get('pvc.tbs.minimum.year.for.historical.data')

        if minimum_year >= tbs_member.created.year:
            return []
        years = range(minimum_year, today().year)
        if exclude_supplied_years:
            # TODO: get all the years already submitted
            pass

        # the declared boolean is used to disable specific line inputs
        return [{'label': year, 'value': year, 'declared': year>=tbs_member.created.year} for year in years]

    def aggregate_declaration_data(self):
        """
        This aggregates Declaration data submitted by the user to make it available for reporting next to historical data.
        """
        # TODO: implement aggregation #409
        pass

