'''
Created on Jun 17, 2014

@author: rado
'''
from django.conf import settings

from basesoft.web.core.management.commands.iqubesync import ProjectBase, CubeBase, RoleBase
from basesoft.web.core.util.controllers import AbstractController
from basesoft.web.core.exceptions import PublicSecurityException

from iqportal.iqube.core.models import DataSource
from iqportal.iqube.core.controllers import IQubeRoleController, IQubeRole

from pvcycle.core.models import PvcOperatorLicense, PvcTbsLicense, PvcMemberLicense
from pvcycle.core.controllers import PvcTBSController, PvcMemberController


class HistoricalDataProject(ProjectBase):
    project_name = 'PVCHistoricalData'
    world_privileges = DataSource.PRIV_VIEW
    description = 'Historical Data cube project'
    role_level = DataSource.ROLE_LEVEL_DB

    def __init__(self, sa_session, controller):
        ProjectBase.__init__(self, sa_session, controller)
        self._db_from_connstr(settings.DB_CONNSTR_READONLY)

    def get_cubes(self):
        return (HistoricalDataCube, )

    def get_roles(self):
        return (HistoricalDataCubeRole, )


class HistoricalDataCube(CubeBase):
    name = 'HistoricalDataCube'
    def _xml(self):
        return """
    <Cube name="%(name)s" defaultMeasure="Quantity">

        <Table name="tblPvcMemberYearlyAggregatedData" />

        <Dimension name="TBS" foreignKey="tbs_member_id">
            <Hierarchy hasAll="true" allMemberName="All TBS" primaryKey="id" primaryKeyTable="tblPvcTbsMember">
                <Join leftKey="tbs_id" rightKey="id">
                    <Table name="tblPvcTbsMember"/>
                    <Table name="tblLicense"/>
                </Join>
                <Level name="TBS" table="tblLicense" column="id" captionColumn="name" type="Integer" />
            </Hierarchy>
        </Dimension>

        <Dimension name="Member" foreignKey="tbs_member_id">
            <Hierarchy hasAll="true" allMemberName="All Members" primaryKey="id" primaryKeyTable="tblPvcTbsMember">
                <Join leftKey="member_id" rightKey="id">
                    <Table name="tblPvcTbsMember"/>
                    <Table name="tblLicense"/>
                </Join>
                <Level name="Member" table="tblLicense" column="id" captionColumn="name" type="Integer" />
            </Hierarchy>
        </Dimension>

        <Dimension name="Year">
            <Hierarchy hasAll="true" allMemberName="All Years">
                <Level name="Year" column="year"/>
            </Hierarchy>
        </Dimension>

        <Dimension name="Country" foreignKey="country_id">
            <Hierarchy hasAll="true" allMemberName="All Countries" primaryKey="id" primaryKeyTable="tblCountry">
                <Table name="tblCountry" />
                <Level name="Country" table="tblCountry" column="id" captionColumn="name" type="Integer" />
            </Hierarchy>
        </Dimension>


        <Measure name="Weight" column="total_weight" aggregator="sum" datatype="Numeric" />
        <Measure name="Power" column="total_megawatt" aggregator="sum" datatype="Numeric" />
        <Measure name="Quantity" column="total_quantity" aggregator="sum" datatype="Integer" />

    </Cube>
        """ % {'name': self.name}

    xml = property(_xml)

class HistoricalDataCubeRole(RoleBase):

    def get_role_list(self, account_id):
        """
        """
        ac = AbstractController(self.sa_session, account_id=account_id, check_active=False)

        rc = IQubeRoleController(self.sa_session, account_id, self.ds, 'HistoricalDataCube')
        rc.grant_cube(IQubeRole.ACCESS_ALL)

        current_license = ac.get_default_group().license

        # operator has full access
        if isinstance(current_license, PvcOperatorLicense):
            ac.assert_permissions('pvc_manage_historical_data')
            return rc.to_list()

        if isinstance(current_license, PvcTbsLicense):
            ac.assert_permissions('pvc_manage_historical_data')

            tc = PvcTBSController(self.sa_session, account_id=account_id)
            tbss = tc.get_take_back_schemes(permissions=['pvc_manage_historical_data'])
            if len(tbss) == 0:
                raise PublicSecurityException('No access to historical data')

            queried = self.sa_session.execute("""
                SELECT tm.member_id, tm.tbs_id
                FROM tblPvcMemberYearlyAggregatedData as aggd
                INNER JOIN tblPvcTbsMember as tm
                    ON aggd.tbs_member_id = tm.id
                WHERE tm.tbs_id IN (%s)
                GROUP BY tm.member_id, tm.tbs_id
            """ % ', '.join([str(t.id) for t in tbss])).fetchall()
        elif isinstance(current_license, PvcMemberLicense):
            ac.assert_permissions('pvc_create_historical_data')

            mc = PvcMemberController(self.sa_session, account_id=account_id)
            members = mc.get_licenses(permissions=['pvc_create_historical_data'])

            if len(members) == 0:
                raise PublicSecurityException('No access to historical data')

            queried = self.sa_session.execute("""
                SELECT tm.member_id, tm.tbs_id
                FROM tblPvcMemberYearlyAggregatedData as aggd
                INNER JOIN tblPvcTbsMember as tm
                    ON aggd.tbs_member_id = tm.id
                WHERE tm.member_id IN (%s)
                GROUP BY tm.member_id, tm.tbs_id
            """ % ', '.join([str(m.id) for m in members])).fetchall()

        else:
            raise PublicSecurityException('No access to historical data')

        # no result is no access
        if len(queried) == 0:
            rc.grant_hierarchy('[TBS]', IQubeRole.ACCESS_NONE, IQubeRole.ROLLUP_POLICY_HIDDEN)
            rc.grant_hierarchy('[Member]', IQubeRole.ACCESS_NONE, IQubeRole.ROLLUP_POLICY_HIDDEN)
            return rc.to_list()

        tbss = []
        members = []

        for member_id, tbs_id in queried:
            if tbs_id not in tbss:
                rc.grant_member('[TBS].[%s]' % tbs_id, IQubeRole.ACCESS_ALL)
                tbss.append(tbs_id)
            if member_id not in members:
                rc.grant_member('[Member].[%s]' % member_id, IQubeRole.ACCESS_ALL)
                members.append(member_id)

        rc.grant_hierarchy('[TBS]', IQubeRole.ACCESS_CUSTOM, IQubeRole.ROLLUP_POLICY_PARTIAL)
        rc.grant_hierarchy('[Member]', IQubeRole.ACCESS_CUSTOM, IQubeRole.ROLLUP_POLICY_PARTIAL)
        return rc.to_list()

projects = (HistoricalDataProject, )
