# Copyright (c) 2014 BaseSoft B.V.
# See LICENSE for details.

from django.utils.translation import ugettext as _

from pvcycle.report.controllers import PvcHistoricalDataController, PvcReportController
from pvcycle.core.controllers import PvcMemberController

from basesoft.web.core.exceptions import PublicUserException
from basesoft.web.core.forms.handler import FormValidator
from basesoft.web.core.input.validation import plugins

class ReportDimensionHandler:
    def __init__(self, request, **kwargs):

        self._request = request
        self._sa_session = request.Session()
        self._account_id = request.session['user_id']

        self._prc = PvcReportController(self._sa_session, account_id=self._account_id, request=self._request)

    def get_efta_countries(self, datasource, cube, config, values):
        return self._prc.get_efta_countries(datasource, cube, config, values)

class HistoricalDataHandler:
    def __init__(self, request, **kwargs):

        self._request = request
        self._sa_session = request.Session()
        self._account_id = request.session['user_id']

        self._controller = PvcHistoricalDataController(self._sa_session, account_id=self._account_id, request=request)

    def load_historical_data(self, _id):
        return self._controller.load_historical_data_state(_id)

    def save_historical_data(self, _id, values):
        if not self._controller.has_permissions('pvc_version_1_access'):
            raise DeprecationWarning('This is no longer available (PV CYCLE 2.0)')

        member_id = values['member_id']
        del values['member_id']

        tbs_id = values['tbs_id']
        del values['tbs_id']

        mc = PvcMemberController(self._sa_session, account_id=self._account_id, request=self._request)
        tbs_member = [tm for tm in mc.get_active_for_tbs(member_id) if tm.tbs.id==int(tbs_id)].pop()
        hd = self._controller.save_or_update_historical_data(_id, tbs_member.id, **values)
        return hd.id

    def save_historical_data_batch(self, _id, values):
        if not self._controller.has_permissions('pvc_version_1_access'):
            raise DeprecationWarning('This is no longer available (PV CYCLE 2.0)')
        basic = values['basic']
        member_id = None if not basic.has_key('member_id') else basic['member_id']
        tbs_id = basic['tbs_id']
        mc = PvcMemberController(self._sa_session, account_id=self._account_id, request=self._request)
        tbs_member = [tm for tm in mc.get_active_for_tbs(member_id) if tm.tbs.id==int(tbs_id)].pop()

        year = basic['year']
        countries = basic['countries']

        self._controller.save_from_state(tbs_member.id, year, countries, values['lines'])
        return True

    def delete_historical_data(self, _id):
        if not self._controller.has_permissions('pvc_version_1_access'):
            raise DeprecationWarning('This is no longer available (PV CYCLE 2.0)')
        self._controller.delete_historical_data(_id)
        return True

    def get_available_years_at_tbs(self, member_id, tbs_id, exlude=False):
        # TODO: fix for multi-member stuff.
        if tbs_id is None or tbs_id == '':
            return []
        mc = PvcMemberController(self._sa_session, account_id=self._account_id, request=self._request)
        members = [tm for tm in mc.get_active_for_tbs(member_id) if tm.tbs.id==int(tbs_id)]
        if len(members) == 0:
            raise PublicUserException(_('This member is not active at the selected TBS'))
        tbs_member = members.pop()
        return self._controller.get_available_years_at_tbs(tbs_member, exclude_supplied_years=exlude)

    def validate_historical_data(self, args):
        form_handler = FormValidator({
            'license_id': [(plugins.Required, {}), (plugins.Integer, {'unsigned': True, 'allow_zero': False})],
            'tbs_id': [(plugins.Required, {}), (plugins.Integer, {'unsigned': True, 'allow_zero': False})],
            'year': [(plugins.Required, {}), ],
            'country_id': [(plugins.Required, {}), ],
            'megawatt': [(plugins.Required, {}), (plugins.Number, {'min_value': 0, 'inclusive': True})],
            'quantity': [(plugins.Required, {}), (plugins.Integer, {'unsigned': True, 'allow_zero': True})],
            'weight': [(plugins.Required, {}), (plugins.Number, {'min_value': 0, 'inclusive': True})],
            'b2b': [(plugins.Required, {}), (plugins.Number, {'max_value': 100, 'min_value': 0, 'inclusive': True})],
            'b2c': [(plugins.Required, {}), (plugins.Number, {'max_value': 100, 'min_value': 0, 'inclusive': True})],
        })
        return form_handler.validate(args)

    def validate_selection(self, args):
        form_handler = FormValidator({
            'license_id': [(plugins.Required, {}), ],
            'tbs_id': [(plugins.Required, {}), ],
            'year': [(plugins.Required, {}), ],
        })
        return form_handler.validate(args)

    def validate_lines(self, args):

        fields = {}
        for field in args:
            fields[field] = [(plugins.Required, {}), ]
            if field[:3] in ['b2b', 'b2c']:
                fields[field].append((plugins.Number, {'max_value': 100, 'min_value': 0, 'inclusive': True}))
            elif field == 'quantity':
                fields[field].append((plugins.Integer, {'unsigned': True, 'allow_zero': True}))
            else:
                fields[field].append( (plugins.Number, {'min_value': 0, 'inclusive': True}))


        form_handler = FormValidator(fields)
        return form_handler.validate(args)

handlers = (HistoricalDataHandler, ReportDimensionHandler, )
