# Copyright (c) 2014 BaseSoft B.V.
# See LICENSE for details.

from django.utils.translation import ugettext as _
from pvcycle.product.models import PvcTbsMemberProduct
from sqlalchemy.sql.expression import and_, or_

from basesoft.web.core.util.controllers import AbstractController, transaction_decorator
from basesoft.web.core.exceptions import PublicUserException, PublicSecurityException, SystemException

from pvcycle.core.controllers import AbstractPvcController, PvcMemberController, PvcNotificationController,\
    AbstractReplacementFieldsContext
from pvcycle.core.models import PvcTbsMember
from pvcycle.subrogation.models import PvcSubrogation, PvcSubrogationRejectionMail, PvcSubrogationSubrogatorMandateMail,\
    PvcSubrogationCommissionerMandateMail, PvcSubrogationNotification
from basesoft.web.contrib.license.models import License
from basesoft.web.contrib.contact.controllers import SimpleContactController
from sqlalchemy.orm.util import aliased
from basesoft.web.contrib.file.controllers import UploadFileController, ProcessFileController
from pvcycle.pvmodule.models import PvcTbsMemberPVModule

class MandateReplacementFieldsContext(AbstractReplacementFieldsContext):

    def __init__(self,subrogation=None):
        self._subrogation = subrogation

    def __getattr__(self, name):
        # TODO: mkpretty

        if name == 'subrogator':
            return self._subrogation.subrogator.member.company

        if name == 'subrogator_contact':
            return self._subrogation.subrogator.member.company.contact_person

        if name == 'subrogator_address':
            return self._subrogation.subrogator.member.company.addresses[0].address

        if name == 'subrogator_country':
            return self._subrogation.subrogator.member.company.addresses[0].address.country

        if name == 'commissioner':
            return self._subrogation.commissioner.member.company

        if name == 'commissioner_contact':
            return self._subrogation.commissioner.member.company.contact_person

        if name == 'commissioner_address':
            return self._subrogation.commissioner.member.company.addresses[0].address

        if name == 'commissioner_country':
            return self._subrogation.commissioner.member.company.addresses[0].address.country

        raise AttributeError

    def get_fields(self):
        rv = []
        rv += self._member_fields(prefix='subrogator', prefix_country=True, labels={
            'member': _('Producer'),
            'member_contact': _('Producer contact'),
            'member_address': _('Producer address')
        })
        rv += self._member_fields(prefix='commissioner', prefix_country=True, labels={
            'member': _('Commissioner'),
            'member_contact': _('Commissioner contact'),
            'member_address': _('Commissioner address')
        })

        return rv

class PvcSubrogationController(AbstractController, AbstractPvcController):

    def get_subrogations(self, subrogation_id=None):
        q = self.sa_session.query(PvcSubrogation).join(PvcTbsMember, PvcSubrogation.subrogator_id==PvcTbsMember.id)

        if self.has_permissions('pvc_manage_subrogations'):
            _filter = PvcTbsMember.tbs_id.in_([tbs.id for tbs in self._get_take_back_schemes(permissions=['pvc_manage_subrogations'])])
        else:
            Commissioner = aliased(PvcTbsMember)
            q.outerjoin(Commissioner, PvcSubrogation.commissioner_id==Commissioner.id)
            licenses = [l.id for l in self._get_member_licenses(None, ['pvc_view_subrogation'])]
            _filter = or_(PvcTbsMember.member_id.in_(licenses), Commissioner.member_id.in_(licenses))

        if subrogation_id is not None:
            if type(subrogation_id) != list:
                subrogation_id = [subrogation_id]
            _filter = and_(_filter, PvcSubrogation.id.in_(subrogation_id))
        return q.filter(_filter).all()

    @transaction_decorator
    def request_for_subrogation(self, tbs_member, commissioner_name):
        """
        The subrogator can make a request for subrogation by supplying a commissioner name.
        """
        self.assert_permissions('pvc_request_for_subrogation')

        if commissioner_name in [None, u'']:
            raise PublicUserException(_('Please provide a valid commissioner name.'))

        if not isinstance(tbs_member, PvcTbsMember):
            mc = PvcMemberController(self.sa_session, account_id=self.get_account().id, request=self.request)
            tbs_member = mc.get_tbs_member(int(tbs_member))

        if tbs_member.status != PvcTbsMember.STATUS_APPROVED:
            raise PublicUserException(_('Only approved producers can apply for subrogation.'))

        if not tbs_member.is_producer:
            raise PublicUserException(_('Only producers can apply for subrogation.'))

        self.logger.info('New subrogation request for %(producer)s at %(commissioner)s.' % {
                'producer': tbs_member.member.name,
                'commissioner': commissioner_name})

        subrogation = PvcSubrogation()
        subrogation.subrogator = tbs_member
        subrogation.request_for_commissioner = commissioner_name
        self.sa_session.add(subrogation)

        # send notification to TBS/Operator
        recipients = self._get_recipients_for_tbs_member(tbs_member)
        nc = PvcNotificationController(self.sa_session, account_id=self.get_account().id)
        notification = nc.set_notification(
            'New subrogation request by %(producer)s' % {'producer': tbs_member.member.name},
            'A new request for subrogation has been done by %(producer)s at commissioner %(commissioner)s.' % {
                'producer': tbs_member.member.name,
                'commissioner': commissioner_name
            },
            instance_class=PvcSubrogationNotification,
            callback={'js_module': 'pvc_subrogation_match', 'parameters': [subrogation.id]}
        )
        nc.create_system_notification(notification, recipients, for_license=True)
        return subrogation

    @transaction_decorator
    def match_commissioner(self, subrogation_id, commissioner_id):

        self.assert_permissions('pvc_manage_subrogations')

        assert type(subrogation_id) in (int, long, )
        subrogations = self.get_subrogations(subrogation_id)
        if len(subrogations) == 0:
            raise PublicUserException(_('Subrogation not found.'))

        subrogation = subrogations.pop()
        if subrogation.commissioner is not None or subrogation.status not in (PvcSubrogation.STATUS_NEW, PvcSubrogation.STATUS_REJECTED):
            raise PublicUserException(_('Subrogation already has a match.'))

        if commissioner_id == subrogation.subrogator.id:
            raise PublicUserException(_('Commissioner and Producer cannot be the same.'))

        commissioner = self.sa_session.query(PvcTbsMember).filter(
            and_(
                PvcTbsMember.id==commissioner_id,
                PvcTbsMember.tbs_id.in_([tbs.id for tbs in self._get_take_back_schemes(permissions=['pvc_manage_subrogations'])])
            )
        ).first()

        if commissioner is None:
            raise PublicUserException(_('Commissioner not found.'))

        if commissioner.status != PvcTbsMember.STATUS_APPROVED:
            self.logger.info('Trying to match non-approved commissioner [%s] for subrogation [%s].'% (commissioner.id, subrogation.id))
            raise PublicUserException(_('Only approved commissioners can be matched for subrogation.'))

        if subrogation.subrogator.tbs != commissioner.tbs:
            self.logger.info('Trying to match commissioner [%s] from non-matching TBS at subrogation [%s].' % (commissioner.id, subrogation.id))
            raise PublicSecurityException(_('Cannot assign a commissioner from a different TBS to subrogation.'))

        subrogation.commissioner = commissioner
        subrogation.mail_status |= PvcSubrogation.MAIL_STATUS_SUBROGATOR_PENDING
        subrogation.status = PvcSubrogation.STATUS_PENDING
        self.sa_session.add(subrogation)

        nc = PvcNotificationController(self.sa_session, account_id=self.get_account().id)
        notification = nc.set_notification(
            'Mandate signature required for subrogation at commissioner %(commissioner)s.' % {'commissioner': commissioner.member.name},
            'Commissioner %(commissioner)s was matched from your request for subrogation and you are now required to sign the mandate and upload it to the system.' % {'commissioner': commissioner.member.name},
            instance_class=PvcSubrogationNotification,
            callback={'js_module': 'pvc_subrogation_details', 'parameters': [subrogation.id]}
        )
        nc.create_system_notification(notification, [subrogation.subrogator.member.default_group], for_license=True)

        self.logger.info('Match with commissioner [%s] for subrogation [%s].' % (commissioner.id, subrogation.id))

        return subrogation

    def get_subrogation_state(self, _id, company_info=False):
        subrogation = self.get_subrogations(_id).pop()

        subrogator = subrogation.subrogator.member.name
        commissioner = subrogation.request_for_commissioner if subrogation.commissioner is None else subrogation.commissioner.member.name

        if company_info:
            scc = SimpleContactController(self.sa_session, account_id=self.get_account().id, request=self.request)
            subrogator = scc.load_company_state(subrogation.subrogator.member.company)
            if subrogation.commissioner is not None:
                commissioner = scc.load_company_state(subrogation.commissioner.member.company)

        rv = {
            'id': subrogation.id,
            'commissioner': commissioner,
            'request_for_commissioner': subrogation.request_for_commissioner,
            'subrogator': subrogator,
            'status': subrogation.status,
            'tbs': subrogation.subrogator.tbs.name,
        }

        return rv

    def get_available_commissioners_by_name(self, _id, commissioner_name):
        self.assert_permissions('pvc_manage_subrogations')
        subrogation = self.get_subrogations(_id).pop()

        # Get all APPROVED commissioners for the TBS linked to the Subrogator.
        result = self.sa_session.query(PvcTbsMember).join(License, PvcTbsMember.member_id==License.id).filter(
            and_(
                PvcTbsMember.status==PvcTbsMember.STATUS_APPROVED,
                PvcTbsMember.is_commissioner==True,
                License.name.like('%%%s%%' % commissioner_name),
                PvcTbsMember.tbs_id==subrogation.subrogator.tbs.id
            )
        ).all()

        rv = []
        for tm in result:
            rv.append({
                'id': tm.id,
                'member_id': tm.member.id,
                'member_name': tm.member.name,
            })
        return rv

    @transaction_decorator
    def approve_subrogation(self, _id):
        self.assert_permissions('pvc_manage_subrogations')
        subrogation = self.get_subrogations(_id).pop()

        # can only approve if the current status is COMMISSIONER APPROVED
        if subrogation.status != PvcSubrogation.STATUS_COMMISSIONER_APPROVED:
            raise PublicUserException(_('This subrogation is not ready for approval.'))

        if subrogation.status == PvcSubrogation.STATUS_ACTIVE:
            raise PublicUserException(_('This subrogation has already been approved.'))

        subrogation.status = PvcSubrogation.STATUS_ACTIVE

        self.logger.info('Subrogation approved.')

        self.sa_session.add(subrogation)
        self.sa_session.flush()

        # send notification to both commissioner and subrogator.
        nc = PvcNotificationController(self.sa_session, account_id=self.get_account().id)
        notification = nc.set_notification(
            'Subrogation now active between %(producer)s and %(commissioner)s.' % {
                'producer': subrogation.subrogator.member.name,
                'commissioner': subrogation.commissioner.member.name
            },
            'The subrogation between %(producer)s and commissioner %(commissioner)s has been approved and is now active.' % {
                'producer': subrogation.subrogator.member.name,
                'commissioner': subrogation.commissioner.member.name
            },
            instance_class=PvcSubrogationNotification,
            callback={'js_module': 'pvc_subrogation_details', 'parameters': [subrogation.id]}
        )
        nc.create_system_notification(notification, [subrogation.subrogator.member.default_group, subrogation.commissioner.member.default_group], for_license=True)
        return subrogation

    @transaction_decorator
    def reject_subrogation(self, _id, reject_reason):
        self.assert_permissions('pvc_manage_subrogations')
        subrogation = self.get_subrogations(_id).pop()

        if subrogation.status == PvcSubrogation.STATUS_REJECTED:
            raise PublicUserException(_('This subrogation has already been rejected.'))

        subrogation.reject_reason = reject_reason
        subrogation.status = PvcSubrogation.STATUS_REJECTED
        subrogation.mail_status |= PvcSubrogation.MAIL_STATUS_REJECTION_PENDING

        self.logger.info('Subrogation rejected for reason: "%s"' % reject_reason)

        self.sa_session.add(subrogation)
        self.sa_session.flush()
        return subrogation

    @transaction_decorator
    def suspend_subrogation(self, _id):
        """
        On suspension of Subrogation the subrogated PV Modules are set to inactive
        """
        self.assert_permissions('pvc_manage_subrogations')
        subrogation = self.get_subrogations(_id).pop()
        if subrogation.status == PvcSubrogation.STATUS_SUSPENDED:
            raise PublicUserException(_('This subrogation has already been suspended.'))
        if subrogation.status != PvcSubrogation.STATUS_ACTIVE:
            raise PublicUserException(_('Only active subrogations can be suspended.'))

        # get all pv modules for this subrogation.
        modules = self.sa_session.query(PvcTbsMemberPVModule).filter(PvcTbsMemberPVModule.subrogation_id==subrogation.id).all()
        for module in modules:
            if module.is_active != PvcTbsMemberPVModule.IS_ACTIVE:
                continue
            module.is_active = PvcTbsMemberPVModule.IS_INACTIVE
            self.sa_session.add(module)

        subrogation.status = PvcSubrogation.STATUS_SUSPENDED
        self.sa_session.add(subrogation)
        self.sa_session.flush()
        return subrogation

    def get_subrogation_rejection_mail(self, subrogation):
        group = subrogation.subrogator.member.default_group
        mail = self.get_object_by_priv_in_hierarchy(group, PvcSubrogationRejectionMail, PvcSubrogationRejectionMail.PRIV_USE_FOR_SENDING)

        targets = [subrogation.subrogator]
        if subrogation.commissioner is not None:
            targets.append(subrogation.commissioner)

        # mail object, attachment, targets
        return mail, targets

    def get_subrogator_mandate_mail(self, subrogation):
        if subrogation.commissioner is None:
            raise SystemException('Cannot send Subrogation Mandate mail for unmatched commissioner on Subrogation (%s)' % subrogation.id)

        group = subrogation.subrogator.member.default_group
        mail = self.get_object_by_priv_in_hierarchy(group, PvcSubrogationSubrogatorMandateMail, PvcSubrogationSubrogatorMandateMail.PRIV_USE_FOR_SENDING)

        targets = [subrogation.subrogator]

        return mail, targets

    def get_commissioner_mandate_mail(self, subrogation):
        if subrogation.commissioner is None:
            raise SystemException('Cannot send Subrogation Mandate mail for unmatched commissioner on Subrogation (%s)' % subrogation.id)

        group = subrogation.subrogator.member.default_group
        mail = self.get_object_by_priv_in_hierarchy(group, PvcSubrogationCommissionerMandateMail, PvcSubrogationCommissionerMandateMail.PRIV_USE_FOR_SENDING)

        targets = [subrogation.commissioner]

        return mail, targets

    @transaction_decorator
    def save_subrogation(self, _id, state):
        if self.has_permissions('pvc_view_subrogation'):
            return self._save_subrogation_mandate(_id, state)

        self.assert_permissions('pvc_manage_subrogations')
        subrogation = self.get_subrogations(_id).pop()

        if not state.has_key('status'):
            return

        if subrogation.status not in (PvcSubrogation.STATUS_ACTIVE, PvcSubrogation.STATUS_COMMISSIONER_APPROVED, PvcSubrogation.STATUS_SUSPENDED,):
            raise PublicUserException(_('The status for this subrogation cannot be changed.'))

        new_status = int(state['status'])
        assert new_status in (PvcSubrogation.STATUS_ACTIVE, PvcSubrogation.STATUS_SUSPENDED, )

        if new_status == subrogation.status:
            return

        if new_status == PvcSubrogation.STATUS_ACTIVE:
            return self.approve_subrogation(_id)

        return self.suspend_subrogation(_id)

    @transaction_decorator
    def _save_subrogation_mandate(self, _id, state):
        subrogation = self.get_subrogations(_id).pop()

        if not subrogation.commissioner or subrogation.status == PvcSubrogation.STATUS_NEW:
            """
            We can only add the mandate after the
            """
            raise PublicUserException(_('You can not update this Subrogation as the Commissioner is not yet matched.'))

        licenses = [l.id for l in self._get_member_licenses(None, ['pvc_view_subrogation'])]
        if subrogation.mandate_file is None:
            if subrogation.subrogator.member_id not in licenses:
                raise PublicUserException(_('Cannot save mandate for Subrogation as you are not the Producer.'))
            new_status = PvcSubrogation.STATUS_SUBROGATOR_APPROVED
            new_mail_status = PvcSubrogation.MAIL_STATUS_COMMISSIONER_PENDING
        else:
            if subrogation.commissioner.member_id not in licenses:
                raise PublicUserException(_('Cannot save mandate for Subrogation as you are not the Commissioner.'))

            if subrogation.status != PvcSubrogation.STATUS_SUBROGATOR_APPROVED:
                raise PublicUserException(_('Cannot save mandate for Subrogation which is not signed by the Producer.'))

            new_status = PvcSubrogation.STATUS_COMMISSIONER_APPROVED
            new_mail_status = None

        uploaded_file = UploadFileController().get_hash_temp_files(state['upload_mandate'])[0]
        file_controller = ProcessFileController(self.sa_session, account_id=self.account_id)
        fc = file_controller.get_category_by_label('subrogation mandates')
        if fc is None:
            fc = file_controller.save_or_update_file_category(None, u'subrogation mandates')
        self.sa_session.flush()
        subrogation.mandate_file = file_controller.save_files_from_temp([uploaded_file], fc)[0]
        subrogation.status = new_status
        if new_mail_status is not None:
            subrogation.mail_status |= new_mail_status
        self.sa_session.add(subrogation)

        self.logger.info('Mandate uploaded for subrogation [%s].' % subrogation.id)

        if new_mail_status is None:
            # send notification to the Operator/TBS.
            recipients = self._get_recipients_for_tbs_member(subrogation.commissioner)
            nc = PvcNotificationController(self.sa_session, account_id=self.get_account().id)
            notification = nc.set_notification(
                'Mandate signed by %(producer)s and commissioner %(commissioner)s' % {
                    'producer': subrogation.subrogator.member.name,
                    'commissioner': subrogation.commissioner.member.name
                },
                'A subrogation mandate has been signed by %(producer)s and commissioner %(commissioner)s.' % {
                    'producer': subrogation.subrogator.member.name,
                    'commissioner': subrogation.commissioner.member.name
                },
                instance_class=PvcSubrogationNotification,
                callback={'js_module': 'pvc_subrogation_details', 'parameters': [subrogation.id]}
            )
            nc.create_system_notification(notification, recipients, for_license=True)
        else:
            # send notification to the Commissioner.
            nc = PvcNotificationController(self.sa_session, account_id=self.get_account().id)
            notification = nc.set_notification(
                'Mandate signature required for subrogation from %(producer)s.' % {
                    'producer': subrogation.subrogator.member.name
                },
                'A subrogation request has been made by %(producer)s and a signature for the mandate is required.' % {
                    'producer': subrogation.subrogator.member.name
                },
                instance_class=PvcSubrogationNotification,
                callback={'js_module': 'pvc_subrogation_details', 'parameters': [subrogation.id]}
            )
            nc.create_system_notification(notification, [subrogation.commissioner.member.default_group], for_license=True)

        return subrogation

    def get_mandate_replacement_fields(self):
        return MandateReplacementFieldsContext().get_fields()

    def get_unapproved_commissioner_pv_modules_count(self):
        self.assert_permissions('pvc_view_subrogation')

        result = self.sa_session.execute("""
            SELECT COUNT(m.id) as cnt
            FROM tblPvcTbsMemberPVModule as m
            INNER JOIN tblPvcSubrogation as s
                ON m.subrogation_id = s.id
                    AND s.status = :subrogation_status

            WHERE m.commissioner_approval_status = :approval_status
                AND m.is_active = :active
                AND s.commissioner_id IN (:tbs_members)
        """, {
            'subrogation_status': PvcSubrogation.STATUS_ACTIVE,
            'approval_status': PvcTbsMemberPVModule.COMMISSIONER_APPROVAL_STATUS_PENDING,
            'active': PvcTbsMemberPVModule.IS_ACTIVE,
            'tbs_members': ', '.join([str(m.id) for m in self._get_tbs_member_licenses()])
        }).fetchone()
        return result['cnt']

    def _get_pv_module_for_commissioner(self, _id):
        if not self.is_commissioner():
            raise PublicUserException(_('You are not a commissioner.'))

        pvm = self.sa_session.query(PvcTbsMemberPVModule).filter(PvcTbsMemberPVModule.id==_id).first()
        if pvm is None or pvm.subrogation is None:
            raise PublicSecurityException(_('No access to PV Module.'))

        assert pvm.subrogation.status == PvcSubrogation.STATUS_ACTIVE
        assert pvm.is_active == PvcTbsMemberPVModule.IS_ACTIVE

        if pvm.subrogation.commissioner not in self._get_tbs_member_licenses():
            raise PublicSecurityException(_('Incorrect subrogated PV Module supplied.'))

        return pvm

    @transaction_decorator
    def approve_pv_module(self, _id):
        pvm = self._get_pv_module_for_commissioner(_id)

        pvm.commissioner_approval_status = PvcTbsMemberPVModule.COMMISSIONER_APPROVAL_STATUS_APPROVED
        self.sa_session.add(pvm)
        return pvm

    @transaction_decorator
    def reject_pv_module(self, _id, reason):
        pvm = self._get_pv_module_for_commissioner(_id)

        if reason is None or reason == '':
            raise PublicUserException(_('Please provide a reason for rejection.'))

        pvm.commissioner_approval_status = PvcTbsMemberPVModule.COMMISSIONER_APPROVAL_STATUS_REJECTED
        self.sa_session.add(pvm)

        # send a notification with the reject reason to the Subrogator.
        nc = PvcNotificationController(self.sa_session, account_id=self.get_account().id)
        recipients = [pvm.tbs_member.member.default_group]

        notification = nc.set_notification(
            'Subrogated PV Module rejected by Commissioner',
            'The commissioner (%(commissioner)s) has rejected a PV Module with the following reason: %(reason)s' % {
                'commissioner': pvm.subrogation.commissioner.member.name,
                'reason': reason,
            },
            instance_class=PvcSubrogationNotification
        )
        nc.create_system_notification(notification, recipients)

        self.logger.info('TBS Member PV Module `%s` rejected by Commissioner account `%s`' % (pvm.id, self.get_account().name))

        from basesoft.web.contrib.order.models import Order
        from pvcycle.order.models import PvcDeclarationOrderLine

        from pvcycle.order.controllers import PvcDeclarationOrderController
        dc = PvcDeclarationOrderController(self.sa_session, account_id=self.get_account().id, request=self.request)

        line = self.sa_session.query(PvcDeclarationOrderLine).join(
                Order, PvcDeclarationOrderLine.order_id==Order.id).filter(
                    and_(
                        Order.status.in_([Order.STATUS_NEW, Order.STATUS_PENDING]),
                        PvcDeclarationOrderLine.tbs_member_module_id==pvm.id
            )).first()

        # no declaration lines with this module.
        if line is None:
            return pvm

        # remove lines from unclosed declarations
        while True:
            self.logger.info('Removing line (%s) from Declaration (%s) due to TBS Member PV Module rejection by Commissioner' % (line.id, line.order.id))
            dc.remove_line(line.order, line.id)

            line = self.sa_session.query(PvcDeclarationOrderLine).join(
                Order, PvcDeclarationOrderLine.order_id==Order.id).filter(
                    and_(
                        Order.status.in_([Order.STATUS_NEW, Order.STATUS_PENDING]),
                        PvcDeclarationOrderLine.tbs_member_module_id==pvm.id
            )).first()
            if line is None:
                break

        return pvm

    def get_approved_subrogations_for_member(self, license_id, commissioner_id=None, tbs_id=None):
        subrogations = self.get_subrogations()
        rv = []
        for subrogation in subrogations:
            if subrogation.status != PvcSubrogation.STATUS_ACTIVE:
                continue
            if subrogation.subrogator.member.id == license_id:
                if commissioner_id is not None and commissioner_id != subrogation.commissioner.id:
                    continue
                if tbs_id is not None and subrogation.subrogator.tbs.id != tbs_id:
                    continue
                rv.append(subrogation)

        return rv

    @transaction_decorator
    def disable_commissioner(self, commissioner_id):
        self.assert_permissions('pvc_manage_subrogations')

        # get all subrogations for commissioner
        subrogations = self.sa_session.query(PvcSubrogation).filter(
            and_(
                PvcSubrogation.commissioner_id==commissioner_id,
                PvcSubrogation.status==PvcSubrogation.STATUS_ACTIVE
            )
        ).all()

        for subrogation in subrogations:
            self.suspend_subrogation(subrogation.id)

        return True

    @transaction_decorator
    def _activate_or_deactivate_product(self, _id, is_active):
        if not self.is_commissioner():
            raise PublicSecurityException(_('This is only available for the commissioner.'))

        p = self.sa_session.query(PvcTbsMemberProduct).filter(PvcTbsMemberProduct.id == _id).first()
        if p.subrogation is None:
            raise PublicSecurityException(_('This product is not part of a subrogation.'))

        c = self._get_tbs_member_licenses(p.subrogation.commissioner.id)
        if len(c) == 0:
            raise PublicSecurityException(_('You are not the commissioner for this subrogated product.'))

        if p.is_active == is_active:
            raise PublicSecurityException(_('The product is already %(is_active)s.' % {
                'is_active': _('active') if is_active else _('inactive')
            }))

        p.is_active = is_active
        self.sa_session.add(p)
        return p

    def deactivate_product(self, _id):
        return self._activate_or_deactivate_product(_id, False)

    def activate_product(self, _id):
        return self._activate_or_deactivate_product(_id, True)