# Copyright (c) 2014 BaseSoft B.V.
# See LICENSE for details.

from django.utils.translation import ugettext as _
from pvcycle.product.models import PvcTbsMemberProduct, PvcTbsProduct

from sqlalchemy.orm.util import aliased
from sqlalchemy.sql.expression import or_, and_

from basesoft.web.core.util.custom_query import CustomQueryModel, CustomQueryField, DATA_TYPE_BOOLEAN
from basesoft.web.contrib.license.models import License

from pvcycle.core.custom_query import AbstractPvcCustomQuery
from pvcycle.core.models import PvcTbsMember, PvcTbsLicense, PvcMemberLicense
from pvcycle.subrogation.models import PvcSubrogation
from basesoft.web.core.exceptions import PermissionDeniedException
from pvcycle.pvmodule.models import PvcTbsMemberPVModule, PvcPVModule, PvcPVModuleSupplier, PvcPVModuleType,\
    PvcPVModuleCellType
from basesoft.web.contrib.i18n.models import Country

Subrogator = aliased(PvcTbsMember)
SubrogatorLicense = aliased(PvcMemberLicense)
Commissioner = aliased(PvcTbsMember)
CommissionerLicense = aliased(PvcMemberLicense)

class SubrogationCustomQuery(AbstractPvcCustomQuery):

    permissions = ['pvc_request_for_subrogation', 'pvc_manage_subrogations'] # OR

    def _get_member_column(self):
        return Subrogator.member_id

    def tables(self):
        return (
            CustomQueryModel(PvcSubrogation),
            CustomQueryModel(Subrogator, onclause=PvcSubrogation.subrogator_id==Subrogator.id),
            CustomQueryModel(PvcTbsLicense, onclause=Subrogator.tbs_id==PvcTbsLicense.id),
            CustomQueryModel(SubrogatorLicense, onclause=Subrogator.member_id==SubrogatorLicense.id),
            CustomQueryModel(Commissioner, onclause=PvcSubrogation.commissioner_id==Commissioner.id, use_outerjoin=True),
            CustomQueryModel(CommissionerLicense, onclause=Commissioner.member_id==CommissionerLicense.id, use_outerjoin=True),
        )

    def fields(self):
        return (
            CustomQueryField(PvcSubrogation.id, 'ID', key='id'),
            CustomQueryField(SubrogatorLicense.id, 'subrogator_id', key='subrogator_id'),
            CustomQueryField(SubrogatorLicense.name, _('Producer'), key='subrogator', is_filter=True, is_group_by=True),
            CustomQueryField(CommissionerLicense.id, 'commissioner_id', key='commissioner_id'),
            CustomQueryField(CommissionerLicense.name, _('Commissioner'), key='commissioner', is_filter=True, is_group_by=True),
            CustomQueryField(PvcTbsLicense.name, _('TBS name'), is_filter=True, key='tbs_name', is_group_by=True),
            CustomQueryField(PvcSubrogation.request_for_commissioner, _('Commissioner requested'), key='commissioner_requested', is_filter=True),
            CustomQueryField(PvcSubrogation.id, _('Details'), key='details'),
            CustomQueryField(PvcSubrogation.status, _('Status'), key='status', is_filter=True, filter_options=[{
                'value': PvcSubrogation.STATUS_NEW,
                'label': _("New"),
            },{
                'value': PvcSubrogation.STATUS_PENDING,
                'label': _("Pending"),
            },{
                'value': PvcSubrogation.STATUS_SUBROGATOR_APPROVED,
                'label': _("Producer approved"),
            },{
                'value': PvcSubrogation.STATUS_COMMISSIONER_APPROVED,
                'label': _("Commissioner approved"),
            },{
                'value': PvcSubrogation.STATUS_ACTIVE,
                'label': _("Active"),
            },{
                'value': PvcSubrogation.STATUS_REJECTED,
                'label': _("Rejected"),
            },{
                'value': PvcSubrogation.STATUS_SUSPENDED,
                'label': _("Suspended"),
            }]),
            CustomQueryField(PvcSubrogation.id, _('Mandate'), key='mandate'),
            CustomQueryField(PvcSubrogation.created, _('Created'), key='created'),
            CustomQueryField(PvcSubrogation.last_updated, _('Last updated'), key='last_updated'),
        )

    def validate_access(self):
        self.assert_any_permissions(*self.permissions)

    def set_where(self):
        licenses = [l.id for l in self._get_member_licenses(None, ['pvc_view_subrogation'])]
        self.where = or_(Subrogator.member_id.in_(licenses), Commissioner.member_id.in_(licenses))
        if not self.has_permissions('view_domain') and self.has_permissions('pvc_tbs_member_manager'):
            self.where = and_(self.where, Subrogator.tbs_id.in_([x.id for x in self._get_take_back_schemes(permissions=['pvc_tbs_member_manager'])]))

    group_by = (PvcSubrogation.id, )

class SubrogatedPVModuleCustomQuery(AbstractPvcCustomQuery):

    permissions = ['pvc_view_subrogation']

    def tables(self):
        return (
            CustomQueryModel(PvcTbsMemberPVModule),
            CustomQueryModel(PvcPVModule),
            CustomQueryModel(PvcPVModuleSupplier),
            CustomQueryModel(PvcPVModuleType),
            CustomQueryModel(PvcPVModuleCellType),
            CustomQueryModel(Country, onclause=PvcTbsMemberPVModule.country_id==Country.id),
            CustomQueryModel(PvcSubrogation),
            CustomQueryModel(Commissioner, onclause=PvcSubrogation.commissioner_id==Commissioner.id),
            CustomQueryModel(CommissionerLicense, onclause=Commissioner.member_id==CommissionerLicense.id),
            CustomQueryModel(Subrogator, onclause=PvcSubrogation.subrogator_id==Subrogator.id),
            CustomQueryModel(SubrogatorLicense, onclause=Subrogator.member_id==SubrogatorLicense.id),
        )

    def fields(self):
        return (

            CustomQueryField(PvcTbsMemberPVModule.id, _('ID'), key='id'),
            CustomQueryField(SubrogatorLicense.name, _('Producer'), key='subrogator', is_filter=True),
            CustomQueryField(PvcPVModuleSupplier.name, _('Brand'), key='supplier', is_filter=True, is_group_by=True),
            CustomQueryField(PvcPVModuleType.name, _('Module type'), key='type', is_filter=True, is_group_by=True),
            CustomQueryField(PvcPVModuleCellType.name, _('Cell type'), key='cell_type', is_filter=True, is_group_by=True),
            CustomQueryField(Country.name, _('Country'), key='country', is_filter=True, filter_options=Country.name, filter_is_unique=True),
            CustomQueryField(PvcTbsMemberPVModule.commissioner_approval_status, _('Approval status'), key='approval_status'),
            CustomQueryField(PvcTbsMemberPVModule.id, _('Action'), key='action'),
        )

    def validate_access(self):
        # this interface is only for commissioners.
        if not self.is_commissioner():
            raise PermissionDeniedException('This custom query is only available for Commissioners')
        self.assert_permissions(*self.permissions)

    def set_where(self):

        licenses = [l.id for l in self._get_member_licenses(None, ['pvc_view_subrogation'])]
        self.where = Commissioner.member_id.in_(licenses)


class SubrogatedProductCustomQuery(AbstractPvcCustomQuery):
    permissions = ['pvc_view_subrogation']

    def tables(self):
        return (
            CustomQueryModel(PvcTbsMemberProduct),
            CustomQueryModel(PvcTbsProduct),
            CustomQueryModel(PvcSubrogation),
            CustomQueryModel(Commissioner, onclause=PvcSubrogation.commissioner_id==Commissioner.id),
            CustomQueryModel(CommissionerLicense, onclause=Commissioner.member_id==CommissionerLicense.id),
            CustomQueryModel(Subrogator, onclause=PvcSubrogation.subrogator_id==Subrogator.id),
            CustomQueryModel(SubrogatorLicense, onclause=Subrogator.member_id==SubrogatorLicense.id),
        )

    def fields(self):
        return (
            CustomQueryField(PvcTbsMemberProduct.id, _('ID'), key='id'),
            CustomQueryField(PvcTbsProduct.label, _('Product'), key='product'),
            CustomQueryField(SubrogatorLicense.name, _('Producer'), key='subrogator', is_filter=True),
            CustomQueryField(PvcTbsMemberProduct.is_active, _('Active'), key='is_active', is_filter=True, type=DATA_TYPE_BOOLEAN),
        )

    def validate_access(self):
        # this interface is only for commissioners.
        if not self.is_commissioner():
            raise PermissionDeniedException('This custom query is only available for Commissioners')
        self.assert_permissions(*self.permissions)

    def set_where(self):

        licenses = [l.id for l in self._get_member_licenses(None, ['pvc_view_subrogation'])]
        self.where = Commissioner.member_id.in_(licenses)


custom_queries = (
    SubrogationCustomQuery,
    SubrogatedPVModuleCustomQuery,
    SubrogatedProductCustomQuery,
)
