# Copyright (c) 2014 BaseSoft B.V.
# See LICENSE for details.

from django.utils.translation import ugettext, ugettext_lazy as _

from basesoft.web.contrib.mailing.models import Mail
from basesoft.web.contrib.notification.models import Notification


class PvcSubrogation(object):
    """
    @DynamicAttrs
    """
    STATUS_NEW = 0 # initial request started by Subrogator
    STATUS_PENDING = 1 # Commissioner matched by Operator/TBS
    STATUS_SUBROGATOR_APPROVED = 2 # Mandate signed and uploaded by Subrogator
    STATUS_COMMISSIONER_APPROVED = 3 # Mandate signed and uploaded by Commissioner
    STATUS_ACTIVE = 4 # Subrogation Approved by Operator/TBS
    STATUS_REJECTED = 5 # Subrogation Rejected by Operator/TBS
    STATUS_SUSPENDED = 6 # Subrogation Revoked/disbanded by Operator/TBS

    MAIL_STATUS_SUBROGATOR_PENDING = 1
    MAIL_STATUS_SUBROGATOR_SENDING = 2
    MAIL_STATUS_SUBROGATOR_SENT = 4
    MAIL_STATUS_COMMISSIONER_PENDING = 8
    MAIL_STATUS_COMMISSIONER_SENDING = 16
    MAIL_STATUS_COMMISSIONER_SENT = 32
    MAIL_STATUS_REJECTION_PENDING = 64
    MAIL_STATUS_REJECTION_SENDING = 128
    MAIL_STATUS_REJECTION_SENT = 256

    def __repr__(self):
        return '<PvcSubrogation id="%s">' % self.id

class PvcTbsSubrogationMandate(object):
    """
    @DynamicAttrs
    """
    pass

class AbstractPvcMandateMail(Mail):
    PRIV_USE_FOR_SENDING = 64

    JS_MODULE = 'iq_account_mail_plugin'

    def get_privileges(self):
        return Mail.get_privileges(self) + [
            (self.PRIV_USE_FOR_SENDING, ugettext('Use for sending'))
        ]

    @staticmethod
    def replacement_fields():
        x = Mail.replacement_fields()
        x.update({
            ugettext('Producer'): {
                'subrogator.name': ugettext('Producer name'),
            },
            ugettext('Commissioner'): {
                'commissioner.name': ugettext('Commissioner name'),
            }
        })
        return x

    def _replace_field(self, person, field_name, context):

        tokens = field_name.split('.')
        c = context['subrogation']
        if tokens[0] == 'subrogator':
            assert tokens[1] in ['name']
            c = c.subrogator.member
        elif tokens[0] == 'client':
            assert tokens[1] in ['name']
            if c.commissioner is None:
                return ''
            c = c.commissioner.member
        elif tokens[0] == 'person':
            c = None

        if hasattr(c, tokens[1]):
            return unicode(getattr(c, tokens[1]))

        return Mail._replace_field(self, person, field_name, c)

class PvcSubrogationSubrogatorMandateMail(AbstractPvcMandateMail):
    """
    @DynamicAttrs
    """
    NAME_I18N = _("Producer mandate mail")

class PvcSubrogationCommissionerMandateMail(AbstractPvcMandateMail):
    """
    @DynamicAttrs
    """
    NAME_I18N = _("Commissioner mandate mail")

class PvcSubrogationRejectionMail(AbstractPvcMandateMail):
    """
    @DynamicAttrs
    """
    NAME_I18N = _("Subrogation rejection mail")

    @staticmethod
    def replacement_fields():
        x = AbstractPvcMandateMail.replacement_fields()
        x.update({
            ugettext('Subrogation'): {
                'subrogation.reject_reason': ugettext('Reject reason'),
            }
        })
        return x

    def _replace_field(self, person, field_name, context):

        tokens = field_name.split('.')
        c = context['subrogation']
        if tokens[0] == 'subrogator':
            assert tokens[1] in ['name']
            c = c.subrogator.member
        elif tokens[0] == 'client':
            assert tokens[1] in ['name']
            if c.commissioner is None:
                return ''
            c = c.commissioner.member
        elif tokens[0] == 'person':
            c = None
        else:
            assert tokens[1] in ['reject_reason']

        if hasattr(c, tokens[1]):
            return unicode(getattr(c, tokens[1]))

        return Mail._replace_field(self, person, field_name, c)


class PvcSubrogationNotification(Notification):
    """
    @DynamicAttrs
    """

    name_i18n = _('Subrogation')

    def __repr__(self):
        return '<PvcSubrogationNotification id=%s>' % self.id