# Copyright (c) 2014 BaseSoft B.V.
# See LICENSE for details.

import datetime
from sqlalchemy.orm import mapper, relation
from sqlalchemy import Column, Integer, Unicode, DateTime, ForeignKey, SmallInteger, UnicodeText

from basesoft.sqlalchemy.table import IQTable
from basesoft.web.contrib.license import models as license_models

from pvcycle.subrogation import models as subrogation_models
from basesoft.web.contrib.license.hooks import PrivilegeObjectMapper
from basesoft.web.contrib.file.models import File

__mapping_defined = False
__tables_defined = False
tables = {}

def reset():
    global __mapping_defined
    global __tables_defined

    __mapping_defined = False
    __tables_defined = False

def get_tables():
    return tables

def do_tables(metadata):
    """
    This function is called by Django middleware when needed.
    """
    global __tables_defined
    if __tables_defined:
        return

    tblPvcSubrogation = IQTable('tblPvcSubrogation', metadata,
        Column('id', Integer, primary_key=True, autoincrement=True),
        Column('subrogator_id', Integer, ForeignKey('tblPvcTbsMember.id', ondelete='RESTRICT'), nullable=False),
        Column('commissioner_id', Integer, ForeignKey('tblPvcTbsMember.id', ondelete='RESTRICT'), nullable=True), # on validation by Operator/TBS this gets set Ticket #371
        Column('mandate_file_id', ForeignKey('tblFile.id', ondelete='RESTRICT'), nullable=True),
        Column('request_for_commissioner', Unicode(255), nullable=False), # Ticket #369
        Column('status', SmallInteger, nullable=False, default=subrogation_models.PvcSubrogation.STATUS_NEW),

        Column('reject_reason', UnicodeText, nullable=True),

        Column('mail_status', Integer, nullable=False, default=0),

        Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
        Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now, onupdate=datetime.datetime.now),
    )

    tblPvcSubrogationSubrogatorMandateMail = IQTable('tblPvcSubrogationSubrogatorMandateMail', metadata,
        Column('mail_id', Integer, ForeignKey('tblMail.id', ondelete='CASCADE'), primary_key=True),
        Column('fallback_locale_id', Integer, ForeignKey('tblLocale.id'), nullable=False),
    )

    tblPvcSubrogationCommissionerMandateMail = IQTable('tblPvcSubrogationCommissionerMandateMail', metadata,
        Column('mail_id', Integer, ForeignKey('tblMail.id', ondelete='CASCADE'), primary_key=True),
        Column('fallback_locale_id', Integer, ForeignKey('tblLocale.id'), nullable=False),
    )

    tblPvcSubrogationRejectionMail = IQTable('tblPvcSubrogationRejectionMail', metadata,
        Column('mail_id', Integer, ForeignKey('tblMail.id', ondelete='CASCADE'), primary_key=True),
        Column('fallback_locale_id', Integer, ForeignKey('tblLocale.id'), nullable=False),
    )

    tables['tblPvcSubrogation'] = tblPvcSubrogation
    tables['tblPvcSubrogationSubrogatorMandateMail'] = tblPvcSubrogationSubrogatorMandateMail
    tables['tblPvcSubrogationCommissionerMandateMail'] = tblPvcSubrogationCommissionerMandateMail
    tables['tblPvcSubrogationRejectionMail'] = tblPvcSubrogationRejectionMail

    __tables_defined = True

def do_mappings():
    """
    This function is called by Django middleware when needed.
    """
    global __mapping_defined
    if __mapping_defined:
        return

    from basesoft.web.contrib.i18n import models as i18n_models
    from basesoft.web.contrib.mailing import models as mailing_models
    from basesoft.web.contrib.mailing.sa_datamapper import get_tables as mail_tables
    from pvcycle.core import models as core_models
    from pvcycle.core.sa_datamapper import get_tables as core_tables
    from basesoft.web.contrib.notification import models as notification_models

    mapper(subrogation_models.PvcSubrogation, tables['tblPvcSubrogation'],
        properties = {
            'subrogator': relation(core_models.PvcTbsMember, backref="subrogations", primaryjoin=tables['tblPvcSubrogation'].c.subrogator_id==core_tables()['tblPvcTbsMember'].c.id),
            'commissioner': relation(core_models.PvcTbsMember, primaryjoin=tables['tblPvcSubrogation'].c.commissioner_id==core_tables()['tblPvcTbsMember'].c.id),
            'mandate_file': relation(File)
        }
    )

    mapper(subrogation_models.PvcSubrogationSubrogatorMandateMail, tables['tblPvcSubrogationSubrogatorMandateMail'],
        extension = PrivilegeObjectMapper(),
        properties = {
            'fallback_locale': relation(i18n_models.Locale),
            'locales': relation(mailing_models.MailLocale, order_by=mail_tables()['tblMailLocale'].c.locale_id, passive_deletes=True)
        },
        polymorphic_identity='PvcSubrogationSubrogatorMandateMail',
        inherits=mailing_models.Mail
    )

    mapper(subrogation_models.PvcSubrogationCommissionerMandateMail, tables['tblPvcSubrogationCommissionerMandateMail'],
        extension = PrivilegeObjectMapper(),
        properties = {
            'fallback_locale': relation(i18n_models.Locale),
            'locales': relation(mailing_models.MailLocale, order_by=mail_tables()['tblMailLocale'].c.locale_id, passive_deletes=True)
        },
        polymorphic_identity='PvcSubrogationCommissionerMandateMail',
        inherits=mailing_models.Mail
    )

    mapper(subrogation_models.PvcSubrogationRejectionMail, tables['tblPvcSubrogationRejectionMail'],
        extension = PrivilegeObjectMapper(),
        properties = {
            'fallback_locale': relation(i18n_models.Locale),
            'locales': relation(mailing_models.MailLocale, order_by=mail_tables()['tblMailLocale'].c.locale_id, passive_deletes=True)
        },
        polymorphic_identity='PvcSubrogationRejectionMail',
        inherits=mailing_models.Mail
    )

    mapper(subrogation_models.PvcSubrogationNotification,
           inherits=notification_models.Notification,
           polymorphic_identity='PvcSubrogationNotification'
    )

    __mapping_defined = True

def fill_data(conn):
    pass

def app_permissions_category():
    """
    Returns a category for all permissions defined for this app
    """
    return license_models.PermissionCategory(u'PVC Subrogation')

def app_permissions():
    from django.utils.translation import ugettext as _
    return [
        license_models.Permission(u'pvc_manage_subrogations', u"""
            This gives the right to manage Subrogation
            """, _("Manage PVC subrogations")),

        license_models.Permission(u'pvc_request_for_subrogation', u"""
            This gives the right to do a request for Subrogation
            """, _("Request for PVC subrogations")),

        license_models.Permission(u'pvc_view_subrogation', u"""
            This gives the right to view Subrogations
            """, _("View PVC subrogations")),

        license_models.Permission(u'pvc_manage_subrogation_mail', u"""
            This gives the right to manage Subrogation related mails
            """, _("Manage PVC subrogation mails")),


    ]
