'''
Created on Sep 17, 2013

@author: marten
'''
from basesoft.web.contrib.notification.models import Notification
from django.conf import settings
from decimal import Decimal
from datetime import date
from pvcycle.logistics.controllers import PvcLogisticsController
from pvcycle.logistics.models import PvcTreatmentCenter, PvcPickupRequest, PvcDeliveryRequest

from sqlalchemy.sql.expression import and_
from sqlalchemy.exc import IntegrityError

from basesoft.web.core.util.controllers import AbstractController
from basesoft.web.contrib.license.controllers import LicenseController
from basesoft.web.contrib.license.models import Domain, Permission
from basesoft.web.contrib.contact.models import Person
from basesoft.web.contrib.i18n.models import Country, Locale
from basesoft.web.contrib.account.models import Account, GroupAccount
from basesoft.web.contrib.account.controllers import AccountController
from basesoft.web.contrib.order.controllers import TaxController, GLAccountController
from basesoft.web.contrib.order.models import TaxClassCountry
from basesoft.web.core.exceptions import PublicUserException, PublicSecurityException
from basesoft.util.virtual_date import set_date

from iqportal.iqube.core.models import DataSourceWidget

from pvcycle.core.models import PvcOperatorLicense, PvcTbsLicense, PvcMemberLicense, PvcTbsServiceLevel, PvcTbsMember, PvcAccountantLicense, \
    PvcROECountry
from pvcycle.core.controllers import PvcTBSController, PvcMemberController, PvcAccountantController
from pvcycle.pvmodule.models import PvcPVModule, PvcPVModuleSupplier, PvcPVModuleType
from pvcycle.pvmodule.controllers import PvcPVModuleController
from pvcycle.order.schedule import NewDeclarationTask, CloseDeclarationTask, AggregatePvcDeclarationTask
from pvcycle.order.controllers import PvcDeclarationOrderController
from pvcycle.order.models import PvcDeclarationOrder


ADMIN_TAX_CLASSES = (
    ('BE', Decimal(21,0)),
    ('IT', Decimal(17,0)),
    ('DE', Decimal(9,0)),
)
CONTRIB_TAX_CLASSES = (
    ('BE', Decimal(19,0)),
    ('IT', Decimal(19,0)),
    ('DE', Decimal(16,1)),
)

class PvcController(AbstractController):

    def populate_pvcycle_domain(self):

        set_date(date(2013, 1, 1))

        self.domain = self.sa_session.query(Domain).filter(Domain.name=='PV CYCLE').one()
        self.create_operator()
        self.create_pv_modules()
        self.create_tbss()
        self.create_accountant()
        self.create_members()
        self.create_widget_access()
        self.create_declarations()
        self.fill_roe()

        if 'pvcycle.logistics' in settings.INSTALLED_APPS:
            self._fill_logistics()

    def create_operator(self):
        lc = LicenseController(self.sa_session, account_id=self.get_account().id)
        lc.add_permissions_to_license(self.get_default_group().license, self._get_permissions('pvc_operator_license', 'manage_roles', 'view_license'))
        lc.add_permissions_to_group(self.get_default_group(), self._get_permissions('pvc_operator_license', 'manage_roles', 'view_license'))

        _license = self.sa_session.query(PvcOperatorLicense).filter(and_(PvcOperatorLicense.name=='Operator', PvcOperatorLicense.domain==self.domain)).first()

        if _license is None:
            state = self._get_operator_state()
            _license = lc.save_license_from_state(state)

        lc.grant_from_role(lc.get_role_by_name('PVC Operator'), _license, _license.default_group)
        if 'pvcycle.uat' in settings.INSTALLED_APPS:
            lc.add_permissions_to_license(_license, self._get_permissions('pvc_user_acceptance_testing', 'user_acceptance_trigger_tasks', 'set_virtual_date'))
            lc.add_permissions_to_group(_license.default_group, self._get_permissions('pvc_user_acceptance_testing', 'user_acceptance_trigger_tasks', 'set_virtual_date'))
        if 'pvcycle.logistics' in settings.INSTALLED_APPS:
            lc.add_permissions_to_license(_license, self._get_permissions('pvc_logistics', 'pvc_logistics_key_user'))
            lc.add_permissions_to_group(_license.default_group, self._get_permissions('pvc_logistics', 'pvc_logistics_key_user'))

        self._operator_account = self._create_account('operator', _license.default_group)
        self._operator_account2 = self._create_account('operator2', _license.default_group)

        # we should add TAX_CLASSES
        for country, tax_rate in ADMIN_TAX_CLASSES:
            self._create_tax_class('TAX_ADMIN_FEE', country, tax_rate)

        for country, tax_rate in CONTRIB_TAX_CLASSES:
            self._create_tax_class('TAX_CONTRIB_FEE', country, tax_rate)

    def create_pv_modules(self):
        self._modules = []
        modules = self._get_pv_modules()

        c = PvcPVModuleController(self.sa_session, account_id=self._operator_account.id)
        c2 = PvcPVModuleController(self.sa_session, account_id=self._operator_account2.id)


        for module in modules:
            m = self.sa_session.query(PvcPVModule).join(PvcPVModuleSupplier).join(PvcPVModuleType).filter(
                and_(PvcPVModuleSupplier.name==module[1], PvcPVModuleType.name==module[2])).first()
            if m is not None:
                module[0] = m.id
            m = c.save_or_update_module(*module)
            c2.approve_module(m.id)
            self._modules.append(m)

    def create_tbss(self):
        # Name, handle, Service Level Count, Declaration Period, Closing Day, Use Weight for calculations, Minimum invoiceable amount, Country, IBAN, SWIFT, Membership instead of SL.
        tbss = (
            ('TBS Europe', 'tbs_europe', 3, PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY, 20, False, 5, None, 'NL39RABO0300065264', 'ABNANL2AWSS', False, 50, ),
            ('TBS Italy', 'tbs_italy', 2, PvcTbsLicense.DECLARATION_PERIOD_YEARLY, 40, True, 10, self._get_country_id('IT'), 'IT40S0542811101000000123456', 'ICRAIT3FUF0', False, 150, ),
            ('TBS Germany', 'tbs_germany', 1, PvcTbsLicense.DECLARATION_PERIOD_MONTHLY, 10, False, 0, self._get_country_id('DE'), 'DE89370400440532013000', 'AARBDE5WCLE', False, 0, ),
            ('TBS UK DTS', 'tbs_uk_dts', 0, PvcTbsLicense.DECLARATION_PERIOD_MONTHLY, 10, False, 0, self._get_country_id('GB'), 'GB29RBOS60161331926819', 'INNDNL2UDDE', True, 0, ),
        )
        tbsc = PvcTBSController(self.sa_session, account_id=self._operator_account.id)
        glac = GLAccountController(self.sa_session, account_id=self._operator_account.id)

        for tbs in tbss:
            _tbs = self.sa_session.query(PvcTbsLicense).filter(and_(PvcTbsLicense.name==tbs[0], PvcTbsLicense.domain==self.domain)).first()
            if _tbs is None:
                state = self._get_tbs_state(tbs)
                _tbs = tbsc.save_tbs(state)

            account = self._create_account(tbs[1], _tbs.default_group)
            setattr(self, '_%s_account' % tbs[1], account)

            # we should add GL ACCOUNTS
            try:
                glac.save_gl_account_state(None, {'merchant_id': _tbs.id, 'code': 'GL_ADMIN_FEE', 'account_number': 'GL_%s' % tbs[1]})
                glac.save_gl_account_state(None, {'merchant_id': _tbs.id, 'code': 'GL_CONTRIB_FEE', 'account_number': 'GL_%s' % tbs[1]})
            except PublicUserException: # in case the account already exists (2nd run of populate)
                pass

            # we should add SERVICE_LEVELS
            for i in range(tbs[2]):
                try:
                    tbsc.save_service_level_state(None, {
                        'tbs_id': _tbs.id,
                        'name': '%s_sl_%s' % (tbs[1], i),
                        'level': i,
                        'contract_annex_text': "%s SL %s ANNEX TEXT" % (tbs[0], 1),

                        'administration_fees': [{
                            'start_year': 2013,
                            'fee': Decimal(150)*(i+1),
                        }],

                        'contribution_fees': [{
                            'start_date': date(2013, 1, 1),
                            'fee': Decimal('0.25')*(i+1),
                        }]
                    })
                except PublicUserException: # this SL already exists
                    pass


    def create_accountant(self):
        accountant = self.sa_session.query(PvcAccountantLicense).filter(and_(PvcAccountantLicense.name=='Accountant', PvcAccountantLicense.domain==self.domain)).first()
        if accountant is not None:
            return

        ac = PvcAccountantController(self.sa_session, account_id=self._operator_account.id)
        accountant = ac.save_accountant(self._get_accountant_state())
        self._create_account('accountant', accountant.default_group)

    def create_members(self):

        # member name, handle, country, (tbs, service level), is_produder, is_commissioner, IBAN, SWIFT
        members = (
            ('Member A', 'member_a', 'BE', (('tbs_europe', 0, ), ('tbs_italy', 1, ),), True, False, '', '', 'en_GB', {'BE': [1,5,7], 'IT': [1,2,3,4]}, ),
            ('Member B', 'member_b', 'NL', (('tbs_europe', 2, ),), True, True, '', '', 'nl_NL', {'NL': [9], 'BE': [12, 13]}, ),
            ('Member C', 'member_c', 'IT', (('tbs_italy', 0, ),), True, False, '', '', 'it_IT', {'IT': [5,6,7,8,10,14]}, ),
            ('Member D', 'member_d', 'DE', (('tbs_italy', 0, ), ('tbs_europe', 1, ), ('tbs_germany', 0, ) ,), True, False, '', '', 'de_DE', {'IT': [1], 'FR': [10], 'DE': [3,4,5]}, ),
            ('Member E', 'member_e', 'DE', (('tbs_germany', 0, ),), True, False, '', '', 'en_GB', {'DE': [1,2,3,4,5,6,7,8,9]}, ),
            ('Member F', 'member_f', 'IT', (('tbs_germany', 0, ), ('tbs_italy', 0, ),), True, False, '', '', 'it_IT', {'DE': [9], 'IT': [6,7,8]}, ),
            ('Member G', 'member_g', 'AT', (('tbs_europe', 2, ),), True, True, '', '', 'en_GB', {'AT': [9,10,11]}, ),
            ('Member H', 'member_h', 'DE', (('tbs_germany', 0, ), ('tbs_italy', 1, ),), True, False, '', '', 'de_DE', {'DE': [1,5,9,15], 'IT': [3]}, ),
            ('Member I', 'member_i', 'IT', (('tbs_europe', 1, ), ('tbs_italy', 0, ),), True, True, '', '', 'en_GB', {'GB': [1,2,3,4,5], 'IT': [8,9,11,14]}, ),
            ('Member J', 'member_j', 'DE', (('tbs_europe', 2, ), ('tbs_germany', 0, ),), False, True, '', '', 'en_GB', {'BE': [2,4,6,8], 'DE': [2,4,6,8]}, ),
        )

        lc = LicenseController(self.sa_session, account_id=self.get_account().id)

        member_accounts = []

        for settings in members:
            name, handle, country, tbss, is_producer, is_commissioner, iban, bic, locale, modules = settings

            member = self.sa_session.query(PvcMemberLicense).filter(and_(PvcMemberLicense.name==name, PvcMemberLicense.domain==self.domain)).first()
            if member is None: # we do not have this member yet.
                state = self._get_member_state(settings)
                member = lc.save_license_from_state(state)

            account = self._create_account(handle, member.default_group)
            member_accounts.append(account)
            setattr(self, '_%s_account' % handle, account)

            mc = PvcMemberController(self.sa_session, account_id=account.id)
            state = self._get_member_state(settings)
            state['dbID'] = member.id

            lc.add_permissions_to_license(member, self._get_permissions('pvc_member_license'))
            lc.add_permissions_to_group(member.default_group, self._get_permissions('pvc_member_license'))
            lc.grant_from_role(lc.get_role_by_name('PVC Simple Member'), member, member.default_group, link_role=False)
            lc.grant_from_role(lc.get_role_by_name('PVC Member Admin'), member, member.default_group)

            member = mc.save_member_license(state)
            if not mc.is_applied_at_tbs(member.id):

                for tbs_name, service_level in tbss:
                    sl = self.sa_session.query(PvcTbsServiceLevel).filter(PvcTbsServiceLevel.name=='%s_sl_%s' % (tbs_name, service_level)).one()
                    tbs_member = mc.apply_at_tbs(sl.tbs.id, member.id, None, sl.id, None, is_commissioner, is_producer, None)
                    tbs_member.status = PvcTbsMember.STATUS_PENDING
                    self.sa_session.add(tbs_member)
                    self.sa_session.flush()

                    tmc = PvcMemberController(self.sa_session, account_id=getattr(self, '_%s_account' % tbs_name).id)
                    tmc.save_tbs_member(tbs_member.id, {
                        'status': PvcTbsMember.STATUS_APPROVED,
                        'is_full_member': False, # TODO: some should have this?
                        'is_representative': False
                    })

            # modules
            pvmc = PvcPVModuleController(self.sa_session, account_id=account.id)
            for country, mods in modules.iteritems():
                countries = [self._get_country_id(country)]
                for m in mods:
                    module = self._modules[m-1]
                    try:
                        pvmc.save_or_update_module_for_countries(member.id, countries, module)
                    except PublicSecurityException:
                        pass


            self._member_accounts = member_accounts

    def _create_tax_class(self, tax_class, country, tax_rate):
        tc = TaxController(self.sa_session, account_id=self._operator_account.id)

        admin_tax = None
        for tcc in tc.get_tax_classes():
            if tcc.code == tax_class:
                admin_tax = tcc
                break

        if admin_tax is None:
            admin_tax = tc.save_tax_class_state(None, {'name': '%s' % tax_class, 'code': tax_class})

        cid = self._get_country_id(country)
        tcc = self.sa_session.query(TaxClassCountry).filter(TaxClassCountry.tax_class_id==admin_tax.id, TaxClassCountry.country_id==cid).first()
        if tcc is not None:
            return

        tc.save_country_tax_state(None, {
            'country_id': cid,
            'tax_class_id': admin_tax.id,
            'tax_rates': [{
                'start_date': date(2013, 1, 1),
                'rate': tax_rate
            }]
        })

    def _create_account(self, name, group):
        a = self.sa_session.query(Account).filter(Account.name==name).first()
        if a is None:
            a = Account()

            a.name = name
            a.person = Person()

        a.password = AccountController(self.sa_session).hash_password('test_password')
        self.sa_session.add(a)

        ga = GroupAccount()
        ga.is_default = True
        ga.active = True
        ga.account = a
        ga.group = group
        self.sa_session.add(ga)
        try:
            self.sa_session.flush()
        except IntegrityError: # we already are at this group :)
            pass
        return a

    def _get_member_state(self, settings):
        name, handle, country, tbss, is_producer, is_commissioner, iban, bic, locale, modules = settings

        return {
            'dbID': None,
            'license': {
                'name': u'%s' % name,
                'expiration_date': None,
                'is_default': False,
                'polymorphic_type': 'License',
            },
            'domain': {'name': 'PV CYCLE'},
            'company': {
                'name': name,
                'vat_no': '%s00%s' % (country, handle)
            },
            'address_data': {
                'street': 'Street %s' % handle,
                'city': 'Member City',
                'postcode': 'CODE',
                'country': self._get_country_id(country)
            },
            'bank_data': {
                'iban': iban,
                'bic': bic,
                'number': 6543210987,
            },
            'contact_person': {
                'last_name': '%s Last Name' % name,
                'first_name': '%s First Name' % name,
                'preferred_locale': self.sa_session.query(Locale).filter(Locale.locale==locale).one().id,
                'gender': 'Male',
                'phone': '234234235',
            },
            'invoice_person_data': {
                'last_name': '%s Invoice Last Name' % name,
                'first_name': '%s Invoice First Name' % name,
                'preferred_locale': self.sa_session.query(Locale).filter(Locale.locale==locale).one().id,
                'gender': 'Female',
                'phone': '234234235',
            },
            'group': {'name': u'admins'},
            'permissions': [],
            'usage_options': []
        }

    def _get_tbs_state(self, settings):
        name, user, sls, period, closing_day, use_weight, minimum_amount, country, iban, swift, use_membership, minimum_contrib_fee = settings

        return {
            'tbs': {
                'tbs_type': PvcTbsLicense.TBS_TYPE_SUBSIDIARY,
                'country': country,
                'contract_text': "%s CONTRACT TEXT" % name,
            },
            'configs': {
                'pvc.tbs.declaration.period': period,
                'pvc.tbs.declaration.closing.day': closing_day,
                'pvc.tbs.declaration.use.weight.for.amount.calculations': use_weight,
                'pvc.tbs.use.membership.instead.of.service.level': use_membership,
                'pvc.tbs.membership.valid.for.years': 3,
                'pvc.tbs.minimum.contribution.fee': minimum_contrib_fee,
                'basesoft.web.contrib.order.minimum_invoiceable_amount': minimum_amount,
            },
            'contact_person': {
                'first_name': u'First %s' % name,
                'last_name': u'Last %s' % name,
                'gender': Person.MALE,
                'email': u'%s@mail-test.basesoft.com' % name,
                'phone': '+31206847087',
                'birth_date': '1970-01-01',
                'address_data': self._get_address_data()
                },
            'company': {
                'name': name,
                'vat_no': 'NL810449493B01',
                'website': u'http://www.basesoft.com',
                },
            'company_address': {
                'address': self._get_address_data(),
                'is_invoice': True,
                'is_delivery': False
                },
            'company_bank_account': {
                'bank_account': {
                    'number': 9876543210,
                    'iban': iban,
                    'bic': swift,
                },
                'is_default': True
            },
        }

    def _get_pv_modules(self):
        return [
            [None, 'Supplier A', 'Type A', 'Cell Type A', 100, 150, 10, 15],
            [None, 'Supplier B', 'Type B', 'Cell Type B', 110, 140, 11, 14],
            [None, 'Supplier C', 'Type C', 'Cell Type A', 120, 160, 12, 13],
            [None, 'Supplier D', 'Type D', 'Cell Type C', 130, 140, 13, 12],
            [None, 'Supplier A', 'Type E', 'Cell Type D', 110, 160, 14, 11],
            [None, 'Supplier B', 'Type F', 'Cell Type E', 110, 150, 15, 10],
            [None, 'Supplier C', 'Type G', 'Cell Type F', 120, 140, 14, 11],
            [None, 'Supplier D', 'Type H', 'Cell Type A', 130, 150, 13, 12],
            [None, 'Supplier A', 'Type I', 'Cell Type F', 140, 160, 12, 13],
            [None, 'Supplier B', 'Type J', 'Cell Type F', 150, 160, 11, 12],
            [None, 'Supplier C', 'Type K', 'Cell Type E', 160, 150, 10, 11],
            [None, 'Supplier A', 'Type L', 'Cell Type D', 180, 140, 11, 12],
            [None, 'Supplier B', 'Type M', 'Cell Type C', 190, 160, 12, 13],
            [None, 'Supplier A', 'Type N', 'Cell Type B', 110, 140, 13, 14],
            [None, 'Supplier A', 'Type O', 'Cell Type A', 120, 160, 14, 15],
        ]

    def _get_permissions(self, *requires):
        return self.sa_session.query(Permission).filter(Permission.name.in_(requires)).all()

    def _get_accountant_state(self):
        tbss = self.sa_session.query(PvcTbsLicense).filter(PvcTbsLicense.domain==self.domain).all()
        tbs_licenses = [{'tbs_id': t.id, 'start_date': date(2013, 1, 1), 'end_date': None} for t in tbss]
        name = 'Accountant'
        return {
            'tbs_licenses': tbs_licenses,
            'contact_person': {
                'first_name': u'First %s' % name,
                'last_name': u'Last %s' % name,
                'gender': Person.MALE,
                'email': u'%s@mail-test.basesoft.com' % name,
                'phone': '+31206847087',
                'birth_date': '1970-01-01',
                'address_data': self._get_address_data()
                },
            'company': {
                'name': name,
                'vat_no': 'NL810449493B01',
                'website': u'http://www.basesoft.com',
                },
            'company_address': {
                'address': self._get_address_data(),
                'is_invoice': True,
                'is_delivery': False
                },
            'company_bank_account': {
                'bank_account': {
                    'number': 9876543210,
                    'iban': 'SM86U0322509800000000270100',
                    'bic': 'GEBABEBB08A',
                },
                'is_default': True
            },
        }

    def _get_operator_state(self):
        return {
            'license': {
                'name': u'Operator',
                'expiration_date': None,
                'is_default': True,
                'polymorphic_type': 'PvcOperatorLicense',
                },
            'domain': {'name': 'PV CYCLE'},
            'contact_person': {
                'first_name': u'Opera',
                'last_name': u'Tor',
                'gender': Person.MALE,
                'email': u'pcv-operator@mail-test.basesoft.com',
                'position': u'CEO',
                'phone': '+31206847087',
                'birth_date': '1970-01-01',
                'address_data': self._get_address_data()
                },
            'company': {
                'name': 'Operator Company',
                'vat_no': 'BE0893027827',
                'website': u'http://www.basesoft.com',
                },
            'company_address': {
                'address': self._get_address_data(),
                'is_invoice': True,
                'is_delivery': False
                },
            'company_bank_account': {
                'bank_account': {
                    'number': 1234567890,
                    'iban': 'BE35001553652737',
                    'bic':'GEBEBABB'},
                'is_default': True
                },
            'group': {'name': u'admins'},
            'permissions': [],
            'usage_options': []
        }

    def _get_address_data(self, country='BE'):
        return {
            'street': 'Test street 1',
            'city': 'Test city',
            'state': 'Test state',
            'postcode': '08123',
            'country': self._get_country_id(country),
        }

    def _get_country_id(self, country='BE'):
        return self.sa_session.query(Country).filter(Country.iso_3166_a2==country).one().id

    def create_widget_access(self):
        from basesoft.web.contrib.order.schedule import AggregateOrderTask
        AggregateOrderTask().execute()

        self._grant_configurable_widget_privs(self._operator_account.default_group)

    def _grant_configurable_widget_privs(self, group):
        license_c = LicenseController(self.sa_session, account_id=self.account_id)
        dws = self.sa_session.query(DataSourceWidget).filter(DataSourceWidget.category=='PVC Report').all()
        for dw in dws:
            license_c.grant_privilege_to(dw, [DataSourceWidget.PRIV_VIEW], domain=True, target_group=group)

    def create_declarations(self):
        NewDeclarationTask().execute()

        values = [123, 67, 34, 12, 8, 0]
        vi = 0
        for ac in self._member_accounts:
            dc = PvcDeclarationOrderController(self.sa_session, account_id=ac.id)
            for d in dc.get_open_orders(priv=PvcDeclarationOrder.PRIV_UPDATE):
                lines = {}
                for i in range(len(d.lines)):
                    lines['line_%d' % i] = values[vi]
                    vi += 1
                    if vi >= len(values):
                        vi = 0
                dc.update_declaration(d.id, lines)

        set_date(date(2013, 1, 21)) # move to Closing Date of TBS Europe

        CloseDeclarationTask().execute()
        AggregatePvcDeclarationTask().execute()

    def fill_roe(self):
        ac = PvcTBSController(self.sa_session)
        countries = ac.get_efta_countries()
        self.sa_session.query(PvcROECountry).delete()

        for country in countries:
            c = PvcROECountry()
            c.country = country
            self.sa_session.add(c)

    def _get_logistics_license_state(self, name, pt, first_name, last_name):
        return {
            'license': {
                'name': name,
                'expiration_date': None,
                'is_default': False,
                'polymorphic_type': pt,
                },
            'domain': {'name': 'PV CYCLE'},
            'contact_person': {
                'first_name': first_name,
                'last_name': last_name,
                'gender': Person.FEMALE,
                'email': u'pcv-@mail-test.basesoft.com',
                'position': u'CEO',
                'phone': '+31206847087',
                'birth_date': '1970-01-01',
                'address_data': self._get_address_data('FR')
                },
            'company': {
                'name': name,
                'vat_no': 'BE 0893.027.827 ',
                'website': u'http://www.basesoft.com',
                },
            'company_address': {
                'address': self._get_address_data(),
                'is_invoice': True,
                'is_delivery': False
                },
            'company_bank_account': {
                'bank_account': {
                    'number': 1234567890,
                    'iban': 'BE35001553652737',
                    'bic':'GEBEBABB'},
                'is_default': True
                },
            'group': {'name': u'admins'},
            'permissions': [
                'view_dashboard','save_dashboard', 'pvc_logistics',
                'view_notification', 'set_custom_query_preset',
                'view_file', 'create_file', 'update_file',
                'create_notification', 'view_notification',
                'send_to_domain_default_group',
                'view_license', 'update_license_masterdata' # FIXME: this was added to view license details via MasterDataPanel but shouldnt be needed
            ],
            'usage_options': []
        }

    def _fill_logistics(self):
        from basesoft.web.contrib.branding.models import Site
        from basesoft.web.contrib.contact.controllers import SimpleContactController

        # update tblSite set home app to #logistics
        if settings.DEBUG:
            # set new public home app
            site = self.sa_session.query(Site).filter(Site.host_match=='').one()
            site.home_app = 'logistics'
            self.sa_session.add(site)

        # Insert a new Key User
        key_user_a = self._create_account('key_user', self._operator_account.default_group)
        lc = LicenseController(self.sa_session, account_id=self.get_account().id)

        # Insert a few Holder Licenses
        state = self._get_logistics_license_state('Carrefour Paris', 'PvcHolderLicense', 'Fredric', 'La Flours')
        state['permissions'] += ['pvc_logistics_holder']
        state['permissions'] = [p.id for p in self._get_permissions(*state['permissions'])]
        holder1 = lc.save_license_from_state(state)
        holder1_a = self._create_account('holder_a', holder1.default_group)

        state = self._get_logistics_license_state('Brico Marche', 'PvcHolderLicense', 'Marie', 'Du Point')
        state['permissions'] += ['pvc_logistics_holder']
        state['permissions'] = [p.id for p in self._get_permissions(*state['permissions'])]
        holder2 = lc.save_license_from_state(state)
        holder2_a = self._create_account('holder_b', holder2.default_group)

        state = self._get_logistics_license_state('Lidl', 'PvcHolderLicense', 'Gaston', 'Chaufeux')
        state['permissions'] += ['pvc_logistics_holder']
        state['permissions'] = [p.id for p in self._get_permissions(*state['permissions'])]
        holder3 = lc.save_license_from_state(state)
        holder3.holder_type = 2  # voluntary
        holder3_a = self._create_account('holder_c', holder3.default_group)

        # Insert a few Carrier Licenses
        state = self._get_logistics_license_state('Marten Transport', 'PvcCarrierLicense', 'Franck', 'Du Jour')
        state['permissions'] += ['pvc_logistics_carrier']
        state['permissions'] = [p.id for p in self._get_permissions(*state['permissions'])]
        carrier1 = lc.save_license_from_state(state)
        carrier1_a = self._create_account('carrier_a', carrier1.default_group)

        state = self._get_logistics_license_state('Gefco Logistics', 'PvcCarrierLicense', 'Eric', 'James')
        state['permissions'] += ['pvc_logistics_carrier']
        state['permissions'] = [p.id for p in self._get_permissions(*state['permissions'])]
        carrier2 = lc.save_license_from_state(state)
        carrier2_a = self._create_account('carrier_b', carrier2.default_group)

        state = self._get_logistics_license_state('Mathez Group', 'PvcCarrierLicense', 'Silvia', 'Merit')
        state['permissions'] += ['pvc_logistics_carrier']
        state['permissions'] = [p.id for p in self._get_permissions(*state['permissions'])]
        carrier3 = lc.save_license_from_state(state)
        carrier3_a = self._create_account('carrier_c', carrier3.default_group)

        # FIXME: add treatment center address

        cc = SimpleContactController(self.sa_session, account_id=self.get_account().id)
        tc1 = cc._save_or_update_company(None, {
            'polymorphic_type': 'PvcTreatmentCenter',
            'name': 'Le Centre de Traitement 1'
        }, PvcTreatmentCenter)
        tc2 = cc._save_or_update_company(None, {
            'polymorphic_type': 'PvcTreatmentCenter',
            'name': 'Le Centre de Traitement 2'
        }, PvcTreatmentCenter)
        tc3 = cc._save_or_update_company(None, {
            'polymorphic_type': 'PvcTreatmentCenter',
            'name': 'Le Centre de Traitement 3'
        }, PvcTreatmentCenter)

        assign_pickups = []

        for i in range(100):

            if i % 4 == 1: # put 4 pickups in a delivery

                carrier = [carrier1, carrier2, carrier3][ i % 3 ]
                tc = [tc1, tc2, tc3][i % 3]

                assert len(assign_pickups) > 0

                lc = PvcLogisticsController( self.sa_session, account_id=key_user_a.id )
                delivery_request = lc.save_or_update_delivery_request(None, {
                    'carrier_id': carrier.id,
                    'treatment_center_id': tc.id,
                    'pickup_requests': [x.id for x in assign_pickups],
                    'total_units': 100 / (i % 5 + 1),
                    'total_kilometers': None
                })
                self.sa_session.flush()

                if i > 10:
                    lc.delivery_request_assign_to_courier(delivery_request.id)
                if i > 20:
                    lc.delivery_request_accept_transport(delivery_request.id)
                if i > 30:
                    lc.delivery_request_complete(delivery_request.id)

                assign_pickups = []

            holder = [holder1, holder2, holder3][ i % 3 ]
            holder_a = [holder1_a, holder2_a, holder3_a][ i % 3 ]
            technology = [PvcPickupRequest.TECH_ASI, PvcPickupRequest.TECH_CDTE, PvcPickupRequest.TECH_CIS, PvcPickupRequest.TECH_OPC] [ i % 4 ]
            condition = [PvcPickupRequest.CONDITION_NEW, PvcPickupRequest.CONDITION_USED, PvcPickupRequest.CONDITION_BURNED] [ i % 3 ]

            # holder A creates a PickupRequest
            lc = PvcLogisticsController( self.sa_session, account_id=holder_a.id )
            pickup_request = lc.save_or_update_pickup_request(None, {
                'holder_id': holder.license_id,
                'technology':technology ,
                'condition': condition,
                'number_modules': 1000 / (i % 5 + 1),
                'weight_holder': 10000 / (i % 8 + 1),
                'weight_tc': None,
                'photo_file': None
            })

            #if i % 4 > 0:
            lc.submit_pickup_request(pickup_request.id)
            assign_pickups.append(pickup_request)


        # all notifications read-only
        notifications = self.sa_session.query(Notification)
        for n in notifications:
            n.is_read = True

        self.sa_session.flush()




