# -*- coding: utf-8 -*-
# Copyright (c) 2015 BaseSoft B.V.
# See LICENSE for details.
import time

from datetime import date
from decimal import Decimal
from basesoft.util.ordered_dict import OrderedDict

from basesoft.util.virtual_date import set_date, today, now
from basesoft.web.contrib.i18n.models import Country
from basesoft.web.contrib.order.models import OrderRule
from pvcycle.core.controllers import PvcMemberController
from pvcycle.core.models import PvcTbsMember, PvcTbsLicense
from pvcycle.order.controllers import PvcContributionOrderController
from pvcycle.order.models import PvcContributionOrder, PvcContributionOrderLine

from pvcycle.product.controllers import PvcProductController
from pvcycle.product.models import PvcTbsProduct, PvcTbsMemberProduct, PvcTbsProductFee
from pvcycle.subrogation.controllers import PvcSubrogationController
from pvcycle.subrogation.models import PvcSubrogation
from pvcycle.test.selenium.pvc_selenium import PvcSeleniumTestCase

class TestCase(PvcSeleniumTestCase):

    def setUp(self):
        super(TestCase, self).setUp()
        self.connect_db()

        set_date(date(2013, 1, 1)) # set date to January 2013

        self.operator_account = self._create_operator_account()

        self._create_contribution_tax_rates(self.operator_account)

        state = self._get_tbs_license_state()
        state['tbs']['country'] = self.sa_session.query(Country).filter(Country.iso_3166_a2=='BE').first().id

        self.tbs_account = self._create_tbs_account(self.operator_account, state=state)
        self.tbs = self.tbs_account.default_group.license
        self.accountant = self._create_accountant_account(self.tbs_account.default_group.license, self.operator_account)

        self._administration_fee = Decimal('250.00')
        self._contribution_fee = Decimal('2.75')

        self.service_level = self._create_tbs_service_level(self.tbs_account, admin_fee=self._administration_fee, year=today().year)

        self.pc = PvcProductController(self.sa_session, account_id=self.operator_account.id)

        products = self.pc.get_available_products_for_tbs(self.tbs.id)
        self.tbs_product = self.pc.save_or_update_tbs_product(None, 'EEE-TBS-TEST', 'WEEE', self.tbs.id,
            parent_level_id=products[0]['id'], basis_for_invoicing=PvcTbsProduct.INVOICE_BASED_ON_WEIGHT
        )

        self.pc.save_contribution_fee(self.tbs_product.id, self.service_level.id, Decimal('0.015'), today())
        self.pc.save_contribution_fee(self.tbs_product.id, self.service_level.id, Decimal('0.015'), today(), PvcTbsProductFee.DIMENSION_IS_B2B)
        self.pc.save_contribution_fee(self.tbs_product.id, self.service_level.id, Decimal('0.015'), today(), PvcTbsProductFee.DIMENSION_IS_B2C)

        self.tbs_product2 = self.pc.save_or_update_tbs_product(None, 'EEE-TBS-TEST-2', 'WEEE', self.tbs.id,
            parent_level_id=products[1]['id'], basis_for_invoicing=PvcTbsProduct.INVOICE_BASED_ON_WEIGHT
        )

        self.pc.save_contribution_fee(self.tbs_product2.id, self.service_level.id, Decimal('0.035'), today())
        self.pc.save_contribution_fee(self.tbs_product2.id, self.service_level.id, Decimal('0.035'), today(), PvcTbsProductFee.DIMENSION_IS_B2B)
        self.pc.save_contribution_fee(self.tbs_product2.id, self.service_level.id, Decimal('0.035'), today(), PvcTbsProductFee.DIMENSION_IS_B2C)

        self.tbs_product3 = self.pc.save_or_update_tbs_product(None, 'EEE-TBS-TEST-3', 'WEEE', self.tbs.id,
            parent_level_id=products[2]['id'], basis_for_invoicing=PvcTbsProduct.INVOICE_BASED_ON_WEIGHT
        )

        self.pc.save_contribution_fee(self.tbs_product3.id, self.service_level.id, Decimal('0.025'), today())
        self.pc.save_contribution_fee(self.tbs_product3.id, self.service_level.id, Decimal('0.025'), today(), PvcTbsProductFee.DIMENSION_IS_B2B)
        self.pc.save_contribution_fee(self.tbs_product3.id, self.service_level.id, Decimal('0.025'), today(), PvcTbsProductFee.DIMENSION_IS_B2C)

        self.accountant_account = self._create_accountant_account(self.tbs_account.default_group.license, self.operator_account)

        set_date(date(2013, 6, 1)) # set date to June 2013

        self._trash_all_rules()

    def _create_simple_member_for_declaration(self, obligation_start_date=None, tbs_account=None, service_level=None):
        ma = self._create_member_account()

        # apply at TBS
        tbs_account = self.tbs_account if tbs_account is None else tbs_account
        service_level = self.service_level if service_level is None else service_level
        tm = self._member_apply_at_tbs(ma, tbs_account, service_level)

        obligation_start_date = date(2012, 1, 1) if obligation_start_date is None else obligation_start_date

        # activate member to automatically create an order entry for the Administration Fee
        mc = PvcMemberController(self.sa_session, account_id=tbs_account.id)
        mc.save_tbs_member(tm.id, {
            'status': PvcTbsMember.STATUS_APPROVED,
            'is_full_member': False,
            'is_representative': False,
            'obligation_start_date': obligation_start_date})
        return ma, tm

    def _get_member_product(self, tm, tbs_product=None, country=None):
        if tbs_product is None:
            tbs_product = self.tbs_product

        # TODO this should be added with a controller method
        product = PvcTbsMemberProduct()
        product.tbs_product = tbs_product
        product.tbs_member = tm
        product.is_active = True
        product.country = country if country is not None else tm.tbs.country
        self.sa_session.add(product)
        self.sa_session.flush()
        return product

    def _get_member_brand(self, brand, tbs_member_product):
        pc = PvcProductController(self.sa_session, account_id=self.operator_account.id)
        return pc.get_or_save_brand(brand, tbs_member_product)

    def _trash_all_rules(self):
        for rule in self.sa_session.query(OrderRule).filter(OrderRule.domain==self.operator_account.default_group.license.domain):
            self.sa_session.delete(rule)

    def _make_rule(self, name, **kwargs):

        controller = PvcContributionOrderController(self.sa_session, account_id=self.operator_account.id)

        controller.save_rule(None, {
            'order_type': kwargs.get('order_type', 'MockOrder'),
            'name': name,
            'enabled': kwargs.get('enabled', True),
            'execute_order': kwargs.get('execute_order', None),
            'description': kwargs.get('description', ''),
            'start_date': kwargs.get('start_date', now()),
            'end_date': kwargs.get('end_date', None),
            'match_all': kwargs.get('match_all', True),
            'stop_processing': kwargs.get('stop_processing', False),
            'conditions': kwargs.get('conditions', []),
            'actions': kwargs.get('actions', [])
        })

    def _january_declaration(self, ma=None, tm=None, b2b_or_b2c=None, weight=None):
        """
        b2b_or_b2c is None, 0 or 1. 0 is False, 1 is True
        """
        if ma is None or tm is None:
            ma, tm = self._create_simple_member_for_declaration()

        dc = PvcContributionOrderController(self.sa_session, account_id=ma.id)

        product = self._get_member_product(tm)

        # add an open declaration
        # FIXME this is using a protected method on the ContributionController
        lines = [{
            'tbs_member_product': product,
            'quantity': 0,
            'weight': Decimal('0.0'),
            'custom_unit': None,
        }]
        if b2b_or_b2c is not None:
            lines[0]['is_b2b'] = b2b_or_b2c

        if tm.tbs.get_config('pvc.tbs.contribution.split.by.brand'):
            lines[0]['brand'] = self._get_member_brand('Philips', product)

        # add an open declaration for this member.
        d = dc._add_contribution_order(tm, lines, 2013, month=1)

        self.login(ma.name, 'test_password')

        self.click_iconbar('icon-pvc-outstanding-contributions')
        self.wait_for_loading() # for the custom query

        self.click_button(panel_name='pvc-outstanding-contributions', text='Declare')
        self.wait_for_loading(stable_time=4) # for the details panel

        self.fill_form(panel_name='pvc-contribution-details-panel', values={
            'line_0_quantity': 200,
            'line_0_weight': '202.4' if weight is None else weight,
        })
        return d

    def _silly_honesty_check(self, panel_name='pvc-contribution-details-panel'):
        self.fill_form(panel_name=panel_name, values={
            'agreement': ('checkbox2', True)
        })

    def test_simple_contribution(self):
        d = self._january_declaration()
        self._silly_honesty_check()
        self.wait_for_loading()

        self.click_button(panel_name='pvc-contribution-details-panel', text='Save')
        self.wait_for_loading()

        self.close_panel(panel_name='pvc-outstanding-contributions')
        self.click_iconbar('icon-pvc-outstanding-contributions')
        self.wait_for_loading() # for the custom query

        self.filter_cq(panel_name='pvc-outstanding-contributions', filters=[
            ('Year', 'Equal', ('select', '2013')),
            ('Month', 'Equal', ('select', '1')),
        ])
        self.wait_for_loading() # for the custom query

        self.assert_cq(panel_name='pvc-outstanding-contributions', row_index=0, data={
            'action': 'Redeclare',
            'total_amount': u'€ 3.03'  # WEIGHT times FEE = 0.015 * 202.4
        })

    def test_contribution_additional_line(self):
        d = self._january_declaration()
        self.click_button(panel_name='pvc-contribution-details-panel', text='Add line')
        self.wait_for_loading()

        self.fill_form(panel_name='pvc-contribution-extra-line-product-select', values={
            'products': ('fancy-select', ['EEE-TBS-TEST-2 - WEEE'])
        })

        self.click_button(panel_name='pvc-contribution-extra-line-product-select', text='Add to declaration')
        self.wait_for_loading()

        self.fill_form(panel_name='pvc-contribution-details-panel', values={
            'line_extra_2_quantity': 1024,
            'line_extra_2_weight': '10,000,000.45',
        })

        # checkity check that the member aint a liar.
        self._silly_honesty_check()
        self.wait_for_loading()

        self.click_button(panel_name='pvc-contribution-details-panel', text='Save')
        self.wait_for_loading()

        self.close_panel(panel_name='pvc-outstanding-contributions')
        self.click_iconbar('icon-pvc-outstanding-contributions')
        self.wait_for_loading() # for the custom query

        self.filter_cq(panel_name='pvc-outstanding-contributions', filters=[
            ('Year', 'Equal', ('select', '2013')),
            ('Month', 'Equal', ('select', '1')),
        ])
        self.wait_for_loading() # for the custom query

        self.assert_cq(panel_name='pvc-outstanding-contributions', row_index=0, data={
            'action': 'Redeclare',
            'total_amount': u'€ 350,003.05'
        })

        self.sa_session.expire(d)
        self.assertEqual(len(d.lines), 2)

    def test_contribution_additional_line_roe(self):

        self.reset_roe()

        state = self._get_tbs_license_state()
        state['tbs']['country'] = None
        eu_tbs = self._create_tbs_account(self.operator_account, state)
        tbs = eu_tbs.default_group.license

        eu_sl = self._create_tbs_service_level(eu_tbs, admin_fee=Decimal('232.2'), year=today().year)

        ma, tm = self._create_simple_member_for_declaration(tbs_account=eu_tbs, service_level=eu_sl)

        dc = PvcContributionOrderController(self.sa_session, account_id=ma.id)

        countries = dc.get_efta_countries()
        products = self.pc.get_available_products_for_tbs(tbs.id)
        product = self.pc.save_or_update_tbs_product(None, 'EEE-TBS-TEST', 'WEEE', tbs.id,
            basis_for_invoicing=PvcTbsProduct.INVOICE_BASED_ON_WEIGHT
        )
        self.pc.save_contribution_fee(product.id, eu_sl.id, Decimal('0.015'), date(2012, 1, 1))
        product2 = self.pc.save_or_update_tbs_product(None, 'EEE-TBS-TEST-2', 'WEEE', tbs.id,
            basis_for_invoicing=PvcTbsProduct.INVOICE_BASED_ON_WEIGHT
        )
        self.pc.save_contribution_fee(product2.id, eu_sl.id, Decimal('0.115'), date(2012, 1, 1))

        product = self._get_member_product(tm, tbs_product=product, country=countries[0])

        # add an open declaration
        lines = [{
            'tbs_member_product': product,
            'quantity': 0,
            'weight': Decimal('0.0'),
            'custom_unit': None,
        }]

        # add an open declaration for this member.
        d = dc._add_contribution_order(tm, lines, 2013, month=1)

        self.login(ma.name, 'test_password')

        self.click_iconbar('icon-pvc-outstanding-contributions')
        self.wait_for_loading() # for the custom query

        self.click_button(panel_name='pvc-outstanding-contributions', text='Declare')
        self.wait_for_loading(stable_time=4) # for the details panel

        self.fill_form(panel_name='pvc-contribution-details-panel', values={
            'line_0_quantity': 200,
            'line_0_weight': '202.4',
        })

        self.click_button(panel_name='pvc-contribution-details-panel', text='Add line')
        self.wait_for_loading()

        self.fill_form(panel_name='pvc-contribution-extra-line-product-select', values={
            'products': ('fancy-select', ['EEE-TBS-TEST-2 - WEEE']),
            'countries': ('fancy-select', ['Estonia']),
        })

        self.click_button(panel_name='pvc-contribution-extra-line-product-select', text='Add to declaration')
        self.wait_for_loading()

        self.fill_form(panel_name='pvc-contribution-details-panel', values={
            'line_extra_2_quantity': 1024,
            'line_extra_2_weight': '10,000,000.45',
        })

        # checkity check that the member aint a liar.
        self._silly_honesty_check()
        self.wait_for_loading()

        self.click_button(panel_name='pvc-contribution-details-panel', text='Save')
        self.wait_for_loading()

        self.filter_cq(panel_name='pvc-outstanding-contributions', filters=[
            ('Year', 'Equal', ('select', '2013')),
            ('Month', 'Equal', ('select', '1')),
        ])
        self.wait_for_loading() # for the custom query

        self.assert_cq(panel_name='pvc-outstanding-contributions', row_index=0, data={
            'action': 'Redeclare',
            'total_amount': u'€ 1,150,003.08'
        })

        self.sa_session.expire(d)
        self.assertEqual(len(d.lines), 2)

    def test_contribution_additional_line_quantity_as_basis(self):
        d = self._january_declaration()

        products = self.pc.get_available_products_for_tbs(self.tbs.id)
        tbs_product_quantity = self.pc.save_or_update_tbs_product(None, 'EEE-TBS-TEST-QUANTITY', 'WEEE', self.tbs.id,
            parent_level_id=products[0]['id'], basis_for_invoicing=PvcTbsProduct.INVOICE_BASED_ON_QUANTITY
        )

        self.pc.save_contribution_fee(tbs_product_quantity.id, self.service_level.id, Decimal('0.235'), date(2013,1,1))

        self.click_button(panel_name='pvc-contribution-details-panel', text='Add line')
        self.wait_for_loading()

        self.fill_form(panel_name='pvc-contribution-extra-line-product-select', values={
            'products': ('fancy-select', ['EEE-TBS-TEST-QUANTITY - WEEE'])
        })

        self.click_button(panel_name='pvc-contribution-extra-line-product-select', text='Add to declaration')
        self.wait_for_loading()

        self.fill_form(panel_name='pvc-contribution-details-panel', values={
            'line_extra_2_quantity': "12213",
            'line_extra_2_weight': '123.45',
        })

        # checkity check that the member aint a liar.
        self._silly_honesty_check()
        self.wait_for_loading()

        self.click_button(panel_name='pvc-contribution-details-panel', text='Save')
        self.wait_for_loading()

        self.close_panel(panel_name='pvc-outstanding-contributions')
        self.click_iconbar('icon-pvc-outstanding-contributions')
        self.wait_for_loading() # for the custom query

        self.filter_cq(panel_name='pvc-outstanding-contributions', filters=[
            ('Year', 'Equal', ('select', '2013')),
            ('Month', 'Equal', ('select', '1')),
        ])
        self.wait_for_loading() # for the custom query

        self.assert_cq(panel_name='pvc-outstanding-contributions', row_index=0, data={
            'action': 'Redeclare',
            'total_amount': u'€ 2,873.09'
        })

        self.sa_session.expire(d)
        self.assertEqual(len(d.lines), 2)

    def test_contribution_submit_with_rule(self):

        self._make_rule('is compliant',
            order_type = 'PvcContributionOrder',
            description = 'discount cuz order iz compliant',
            conditions = [{
                'condition': 'is_compliant',
                'operator': None, 'value_text': None, 'value_bool': None,
                'value_int': PvcContributionOrder.IS_COMPLIANT,
                'value_decimal': None,
            },],
            actions = [{
                'action': 'fixed_amount_discount',
                'description': '',
                'operator': None, 'value_text': None, 'value_bool': None, 'value_int': None,
                'value_decimal': 1,
                'tax_class_id': None, 'gl_account_code': None
            }]
        )

        self.payment_methods = self._setup_payment_methods(self.operator_account, self.tbs_account)

        d = self._january_declaration()
        self._silly_honesty_check()

        self.click_button(panel_name='pvc-contribution-details-panel', text='Submit')
        self.wait_for_loading(stable_time=4)

        # assert invoice
        self.assertEquals(self._get_text_from_selector('div[name="pvc-contribution-details-panel-checkout"] #order_product_0'), 'EEE-TBS-TEST - WEEE')
        self.assertEquals(self._get_text_from_selector('div[name="pvc-contribution-details-panel-checkout"] #order_product_1'), 'discount cuz order iz compliant')

        self.assertEquals(self._get_text_from_selector('div[name="pvc-contribution-details-panel-checkout"] #order_line_amount_excluding_vat_0'), '3.03')
        self.assertEquals(self._get_text_from_selector('div[name="pvc-contribution-details-panel-checkout"] #order_line_amount_excluding_vat_1'), '-1.00')

        self.assertEquals(self._get_text_from_selector('div[name="pvc-contribution-details-panel-checkout"] .subtotal-amount'), u'€ 2.03')
        self.assertEquals(self._get_text_from_selector('div[name="pvc-contribution-details-panel-checkout"] .total-vat'), u'€ 0.00')
        self.assertEquals(self._get_text_from_selector('div[name="pvc-contribution-details-panel-checkout"] .total-amount'), u'€ 2.03')

        # select 'bank transfer' payment method and continue
        self.fill_form('pvc-contribution-details-panel-checkout', values={
            'payment_method': ('radio', self.payment_methods[0].id)
        })

        self.click_button(panel_name='pvc-contribution-details-panel-checkout', text='Submit')
        self.wait_for_loading(times=2)

        self.click_button('order-payment_method-panel', text='Mark as submitted')
        self.wait_for_loading()
        self.assert_cq_empty(panel_name='pvc-outstanding-contributions')

        # order should be exportable
        dc = PvcContributionOrderController(self.sa_session, account_id=self.accountant_account.id)
        orders = [o['id'] for o in dc.get_unexported_orders_for_merchant(self.tbs_account.default_group.license.id)]
        self.assertTrue(d.id in orders)

    def test_contribution_submit_with_zero_rule(self):

        self._make_rule('full discount',
            order_type = 'PvcContributionOrder',
            description = 'full discount',
            conditions = [],
            actions = [{
                'action': 'relative_amount_discount',
                'description': '',
                'operator': None, 'value_text': None, 'value_bool': None, 'value_int': None,
                'value_decimal': 100,
                'tax_class_id': None, 'gl_account_code': None
            }]
        )

        self.payment_methods = self._setup_payment_methods(self.operator_account, self.tbs_account)

        d = self._january_declaration()
        self._silly_honesty_check()

        self.click_button(panel_name='pvc-contribution-details-panel', text='Submit')
        self.wait_for_loading(stable_time=4)

        self.assertEquals(self._get_text_from_selector('div[name="pvc-contribution-details-panel-checkout"] #order_product_1'), 'full discount')
        self.assertEquals(self._get_text_from_selector('div[name="pvc-contribution-details-panel-checkout"] .total-amount'), u'€ 0.00')
        self.click_button(panel_name='pvc-contribution-details-panel-checkout', text='Submit')

        self.wait_for_loading()
        self.assert_cq_empty(panel_name='pvc-outstanding-contributions')

    def test_contribution_resubmit_with_rule_and_additional_line(self):

        self._make_rule('is compliant',
            order_type = 'PvcContributionOrder',
            description = 'discount cuz order iz compliant',
            conditions = [{
                'condition': 'is_compliant',
                'operator': None, 'value_text': None, 'value_bool': None,
                'value_int': PvcContributionOrder.IS_COMPLIANT,
                'value_decimal': None,
            },],
            actions = [{
                'action': 'fixed_amount_discount',
                'description': '',
                'operator': None, 'value_text': None, 'value_bool': None, 'value_int': None,
                'value_decimal': 1,
                'tax_class_id': None, 'gl_account_code': None
            }]
        )

        self.payment_methods = self._setup_payment_methods(self.operator_account, self.tbs_account)

        d = self._january_declaration()
        self._silly_honesty_check()

        self.click_button(panel_name='pvc-contribution-details-panel', text='Submit')
        self.wait_for_loading(stable_time=4)

        # assert invoice
        self.assertEquals(self._get_text_from_selector('div[name="pvc-contribution-details-panel-checkout"] #order_product_0'), 'EEE-TBS-TEST - WEEE')
        self.assertEquals(self._get_text_from_selector('div[name="pvc-contribution-details-panel-checkout"] #order_product_1'), 'discount cuz order iz compliant')

        self.assertEquals(self._get_text_from_selector('div[name="pvc-contribution-details-panel-checkout"] #order_line_amount_excluding_vat_0'), '3.03')
        self.assertEquals(self._get_text_from_selector('div[name="pvc-contribution-details-panel-checkout"] #order_line_amount_excluding_vat_1'), '-1.00')

        self.assertEquals(self._get_text_from_selector('div[name="pvc-contribution-details-panel-checkout"] .subtotal-amount'), u'€ 2.03')
        self.assertEquals(self._get_text_from_selector('div[name="pvc-contribution-details-panel-checkout"] .total-vat'), u'€ 0.00')
        self.assertEquals(self._get_text_from_selector('div[name="pvc-contribution-details-panel-checkout"] .total-amount'), u'€ 2.03')

        # select 'bank transfer' payment method and continue
        self.fill_form('pvc-contribution-details-panel-checkout', values={
            'payment_method': ('radio', self.payment_methods[0].id)
        })

        self.click_button(panel_name='pvc-contribution-details-panel-checkout', text='Submit')
        self.wait_for_loading(times=2)

        self.click_button('order-payment_method-panel', text='Mark as submitted')
        self.wait_for_loading()

        d.resubmit_request = PvcContributionOrder.RESUBMIT_REQUEST_APPROVED
        self.sa_session.add(d)
        self.sa_session.flush()

        self.close_panel('pvc-outstanding-contributions')

        self.click_iconbar('icon-pvc-contribution-manager', hover='icon-pvc-outstanding-contributions')
        self.wait_for_panel('pvc-contribution-manager')

        self.filter_cq(panel_name='pvc-contribution-manager', filters=[
            ('Order ID', 'Equal', d.id),
        ])
        self.wait_for_loading() # for the custom query

        self.assert_cq(panel_name='pvc-contribution-manager', row_index=0, data={
            'action': 'Resubmit'
        })

        self.click_button(panel_name='pvc-contribution-manager', text='Resubmit')
        self.wait_for_loading() # for the details panel

        self.click_button(panel_name='pvc-contribution-details-panel', text='Add line')
        self.wait_for_loading()

        self.fill_form(panel_name='pvc-contribution-extra-line-product-select', values={
            'products': ('fancy-select', ['EEE-TBS-TEST-2 - WEEE'])
        })

        self.click_button(panel_name='pvc-contribution-extra-line-product-select', text='Add to declaration')
        self.wait_for_loading()

        self.fill_form(panel_name='pvc-contribution-details-panel', values={
            'line_extra_3_quantity': 1024,
            'line_extra_3_weight': '123.45',
        })

        # checkity check that the member aint a liar.
        self._silly_honesty_check()
        self.wait_for_loading()

        self.click_button(panel_name='pvc-contribution-details-panel', text='Save')
        self.wait_for_loading(stable_time=2)

        d = self.sa_session.query(PvcContributionOrder).filter(PvcContributionOrder.parent_order==d).first()
        n = self.sa_session.query(PvcContributionOrder).filter(PvcContributionOrder.parent_order==d).first()
        line1 = n.lines[0]
        line2 = n.lines[1]
        self.assertTrue(isinstance(line1, PvcContributionOrderLine))
        self.assertTrue(isinstance(line2, PvcContributionOrderLine))

    def test_resubmit(self):
        """
        Tests a resubmission (currently done through the order overview).
        """

        self.payment_methods = self._setup_payment_methods(self.operator_account, self.tbs_account)

        ma, tm = self._create_simple_member_for_declaration()

        d = self._january_declaration(ma, tm)
        self._silly_honesty_check()

        self.click_button(panel_name='pvc-contribution-details-panel', text='Submit')
        self.wait_for_loading(times=6)

        self.fill_form('pvc-contribution-details-panel-checkout', values={
            'payment_method': ('radio', self.payment_methods[0].id)
        })
        self.wait_for_loading()

        self.click_button(panel_name='pvc-contribution-details-panel-checkout', text='Submit')
        self.wait_for_loading(times=2)

        self.click_button('order-payment_method-panel', text='Mark as submitted')
        self.wait_for_loading()

        self.close_panel(panel_name='pvc-outstanding-contributions')

        self.click_iconbar('icon-pvc-contribution-manager', hover='icon-pvc-outstanding-contributions')
        self.wait_for_panel('pvc-contribution-manager')

        self.filter_cq(panel_name='pvc-contribution-manager', filters=[
            ('Order ID', 'Equal', d.id),
        ])
        self.wait_for_loading() # for the custom query

        self.assert_cq(panel_name='pvc-contribution-manager', row_index=0, data={
            'action': 'Request'
        })

        self.click_button(panel_name='pvc-contribution-manager', text='Request')
        self.wait_for_loading() # for the details panel

        self.click_button(panel_name='dialog-question', text='Yes')
        self.wait_for_loading(stable_time=4) # for the details panel

        self.assert_cq(panel_name='pvc-contribution-manager', row_index=0, data={
            'action': 'Pending'
        })

        self.sa_session.expire(d)

        # LOGIN as actual operator to test interface.
        self.logout()
        self.login(self.operator_account.name, 'test_password')
        self.click_iconbar('icon-pvc-contribution-manager')

        self.filter_cq(panel_name='pvc-contribution-manager', filters=[
            ('Order ID', 'Equal', d.id),
        ])
        self.wait_for_loading()
        self.click_button(panel_name='pvc-contribution-manager', text='Approve')
        self.wait_for_loading()

        self.click_button(panel_name='dialog-question', text='Yes')
        self.wait_for_loading()

        self.logout()

        self.login(ma.name, 'test_password')

        self.click_iconbar('icon-pvc-contribution-manager', hover='icon-pvc-outstanding-contributions')
        self.wait_for_panel('pvc-contribution-manager')

        self.filter_cq(panel_name='pvc-contribution-manager', filters=[
            ('Order ID', 'Equal', d.id),
        ])
        self.wait_for_loading() # for the custom query

        self.assert_cq(panel_name='pvc-contribution-manager', row_index=0, data={
            'action': 'Resubmit'
        })

        self.click_button(panel_name='pvc-contribution-manager', text='Resubmit')
        self.wait_for_loading() # for the details panel

        self.fill_form(panel_name='pvc-contribution-details-panel', values={
            'line_0_quantity': 400,
            'line_0_weight': 404,
        })
        self._silly_honesty_check()

        self.click_button(panel_name='pvc-contribution-details-panel', text='Save')
        self.wait_for_loading(times=2)

    def test_contribution_two_additional_lines(self):
        d = self._january_declaration()
        self.click_button(panel_name='pvc-contribution-details-panel', text='Add line')
        self.wait_for_loading()

        self.fill_form(panel_name='pvc-contribution-extra-line-product-select', values={
            'products': ('fancy-select', ['EEE-TBS-TEST-2 - WEEE'])
        })

        self.click_button(panel_name='pvc-contribution-extra-line-product-select', text='Add to declaration')
        self.wait_for_loading()

        self.fill_form(panel_name='pvc-contribution-details-panel', values={
            'line_extra_2_quantity': 1024,
            'line_extra_2_weight': '123.45',
        })

        self.click_button(panel_name='pvc-contribution-details-panel', text='Add line')
        self.wait_for_loading()

        self.fill_form(panel_name='pvc-contribution-extra-line-product-select', values={
            'products': ('fancy-select', ['EEE-TBS-TEST-3 - WEEE'])
        })

        self.click_button(panel_name='pvc-contribution-extra-line-product-select', text='Add to declaration')
        self.wait_for_loading()

        self.assert_form(panel_name='pvc-contribution-details-panel', values={
            'line_extra_2_quantity': '1024',
            'line_extra_2_weight': '123.45',
        })

        self.fill_form(panel_name='pvc-contribution-details-panel', values={
            'line_extra_3_quantity': 1200,
            'line_extra_3_weight': '54.321',
        })

        # checkity check that the member aint a liar.
        self._silly_honesty_check()
        self.wait_for_loading()

        self.click_button(panel_name='pvc-contribution-details-panel', text='Save')
        self.wait_for_loading()

        self.close_panel(panel_name='pvc-outstanding-contributions')
        self.click_iconbar('icon-pvc-outstanding-contributions')
        self.wait_for_loading() # for the custom query

        self.filter_cq(panel_name='pvc-outstanding-contributions', filters=[
            ('Year', 'Equal', ('select', '2013')),
            ('Month', 'Equal', ('select', '1')),
        ])
        self.wait_for_loading() # for the custom query

        self.assert_cq(panel_name='pvc-outstanding-contributions', row_index=0, data={
            'action': 'Redeclare',
            'total_amount': u'€ 8.71'
        })

        self.sa_session.expire(d)
        self.assertEqual(len(d.lines), 3)

    def test_contribution_two_additional_lines_one_shot(self):
        d = self._january_declaration()
        self.click_button(panel_name='pvc-contribution-details-panel', text='Add line')
        self.wait_for_loading()

        self.fill_form(panel_name='pvc-contribution-extra-line-product-select', values={
            'products': ('fancy-select', ['EEE-TBS-TEST-2 - WEEE','EEE-TBS-TEST-3 - WEEE'])
        })

        self.click_button(panel_name='pvc-contribution-extra-line-product-select', text='Add to declaration')
        self.wait_for_loading()

        self.fill_form(panel_name='pvc-contribution-details-panel', values={
            'line_extra_2_quantity': 1024,
            'line_extra_2_weight': '123.45',
            'line_extra_3_quantity': 1200,
            'line_extra_3_weight': '54.321',
        })

        # checkity check that the member aint a liar.
        self._silly_honesty_check()
        self.wait_for_loading()

        self.click_button(panel_name='pvc-contribution-details-panel', text='Save')
        self.wait_for_loading()

        self.close_panel(panel_name='pvc-outstanding-contributions')
        self.click_iconbar('icon-pvc-outstanding-contributions')
        self.wait_for_loading() # for the custom query

        self.filter_cq(panel_name='pvc-outstanding-contributions', filters=[
            ('Year', 'Equal', ('select', '2013')),
            ('Month', 'Equal', ('select', '1')),
        ])
        self.wait_for_loading() # for the custom query

        self.assert_cq(panel_name='pvc-outstanding-contributions', row_index=0, data={
            'action': 'Redeclare',
            'total_amount': u'€ 8.71'
        })

        self.sa_session.expire(d)
        self.assertEqual(len(d.lines), 3)

    def test_contribution_remove_order_line(self):
        d = self._january_declaration()
        self.click_button(panel_name='pvc-contribution-details-panel', text='Add line')
        self.wait_for_loading()

        self.fill_form(panel_name='pvc-contribution-extra-line-product-select', values={
            'products': ('fancy-select', ['EEE-TBS-TEST-2 - WEEE'])
        })

        self.click_button(panel_name='pvc-contribution-extra-line-product-select', text='Add to declaration')
        self.wait_for_loading()

        self.fill_form(panel_name='pvc-contribution-details-panel', values={
            'line_extra_2_quantity': 1024,
            'line_extra_2_weight': '123.45',
        })

        # checkity check that the member aint a liar.
        self._silly_honesty_check()
        self.wait_for_loading()

        self.click_button(panel_name='pvc-contribution-details-panel', text='Save')
        self.wait_for_loading()

        self.close_panel(panel_name='pvc-outstanding-contributions')
        self.click_iconbar('icon-pvc-outstanding-contributions')
        self.wait_for_loading() # for the custom query

        self.filter_cq(panel_name='pvc-outstanding-contributions', filters=[
            ('Year', 'Equal', ('select', '2013')),
            ('Month', 'Equal', ('select', '1')),
        ])
        self.wait_for_loading() # for the custom query
        self.click_button(panel_name='pvc-outstanding-contributions', text='Redeclare')
        self.wait_for_loading(stable_time=4) # for the details panel
        self._silly_honesty_check()  # Should we allow removal of lines without the check?

        self.click_button(panel_name='pvc-contribution-details-panel', text='Remove line')

    def test_contribution_remove_additional_line(self):
        # this tests if the additional line (which has not yet been added to the order) can be removed.
        d = self._january_declaration()
        self.click_button(panel_name='pvc-contribution-details-panel', text='Add line')
        self.wait_for_loading()

        self.fill_form(panel_name='pvc-contribution-extra-line-product-select', values={
            'products': ('fancy-select', ['EEE-TBS-TEST-2 - WEEE','EEE-TBS-TEST-3 - WEEE'])
        })

        self.click_button(panel_name='pvc-contribution-extra-line-product-select', text='Add to declaration')
        self.wait_for_loading()

        self.fill_form(panel_name='pvc-contribution-details-panel', values={
            'line_extra_2_quantity': 1024,
            'line_extra_2_weight': '123.45',
            'line_extra_3_quantity': 1200,
            'line_extra_3_weight': '54.321',
        })

        self.click_button_by_id('remove_line_extra_2')
        self.wait_for_loading()

        self._silly_honesty_check()

        self.click_button(panel_name='pvc-contribution-details-panel', text='Save')
        self.wait_for_loading()

        self.sa_session.expire(d)
        self.assertEqual(len(d.lines), 2)
        products = [l.tbs_member_product.tbs_product.id for l in d.lines]
        self.assertTrue(self.tbs_product.id in products)
        self.assertTrue(self.tbs_product3.id in products)

        # Product 2 was removed;
        self.assertTrue(self.tbs_product2.id not in products)

    def test_custom_unit(self):

        products = self.pc.get_available_products_for_tbs(self.tbs.id)

        tbs_product_quantity = self.pc.save_or_update_tbs_product(None, 'product quantity', 'WEEE-3', self.tbs.id,
            parent_level_id=products[1]['id'], basis_for_invoicing=PvcTbsProduct.INVOICE_BASED_ON_QUANTITY
        )

        tbs_product_custom_unit = self.pc.save_or_update_tbs_product(None, 'product power', 'WEEE-3', self.tbs.id,
            parent_level_id=products[1]['id'], basis_for_invoicing=PvcTbsProduct.INVOICE_BASED_ON_CUSTOM_UNIT, custom_unit=PvcTbsProduct.CUSTOM_UNIT_NOMINAL_POWER
        )

        tbs_product_custom_unit2 = self.pc.save_or_update_tbs_product(None, 'product surface', 'WEEE-3', self.tbs.id,
            parent_level_id=products[1]['id'], basis_for_invoicing=PvcTbsProduct.INVOICE_BASED_ON_CUSTOM_UNIT, custom_unit=PvcTbsProduct.CUSTOM_UNIT_SURFACE
        )

        self.pc.save_contribution_fee(tbs_product_quantity.id, self.service_level.id, Decimal('0.016'), date(2013, 1, 1))
        self.pc.save_contribution_fee(tbs_product_custom_unit.id, self.service_level.id, Decimal('0.017'), date(2013, 1, 1))
        self.pc.save_contribution_fee(tbs_product_custom_unit2.id, self.service_level.id, Decimal('0.018'), date(2013, 1, 1))

        ma, tm = self._create_simple_member_for_declaration()
        cc = PvcContributionOrderController(self.sa_session, account_id=ma.id)

        product_quantity = self._get_member_product(tm, tbs_product=tbs_product_quantity)
        product_weight = self._get_member_product(tm, tbs_product=self.tbs_product)
        product_custom_unit = self._get_member_product(tm, tbs_product=tbs_product_custom_unit)

        lines = [{
            'tbs_member_product': product_quantity,
            'quantity': 10,
            'weight': Decimal(101),
            'custom_unit': None,
        }, {
            'tbs_member_product': product_weight,
            'quantity': 11,
            'weight': Decimal('111'),
            'custom_unit': None,
        }, {
            'tbs_member_product': product_custom_unit,
            'quantity': 12,
            'weight': Decimal(121),
            'custom_unit': Decimal('22'),
        }]
        order = cc._add_contribution_order(tm, lines, 2013, month=1)

        self.login(ma.name, 'test_password')

        self.click_iconbar('icon-pvc-outstanding-contributions')
        self.wait_for_loading()

        self.filter_cq(panel_name='pvc-outstanding-contributions', filters=[
            ('Year', 'Equal', ('select', '2013')),
            ('Month', 'Equal', ('select', '1')),
        ])

        self.click_button(panel_name='pvc-outstanding-contributions', text='Declare')
        self.wait_for_loading()

        self.fill_form(panel_name='pvc-contribution-details-panel', values={
            'line_1_quantity': 100,
            'line_1_weight': '200',
            'line_1_custom_unit': '300.666',
        })

        self.click_button(panel_name='pvc-contribution-details-panel', text='Add line')
        self.wait_for_loading()

        self.fill_form(panel_name='pvc-contribution-extra-line-product-select', values={
            'products': ('fancy-select', ['product surface - WEEE-3'])
        })

        self.click_button(panel_name='pvc-contribution-extra-line-product-select', text='Add to declaration')
        self.wait_for_loading()

        self.fill_form(panel_name='pvc-contribution-details-panel', values={
            'line_extra_4_quantity': 10,
            'line_extra_4_weight': '11',
            'line_extra_4_custom_unit': '12.5',
        })

        self._silly_honesty_check()

        self.click_button(panel_name='pvc-contribution-details-panel', text='Submit')
        self.wait_for_loading(times=6)

        css = 'div[name="pvc-contribution-details-panel-checkout"] '

        # assert invoice
        self.assertEquals(self._get_text_from_selector(css + '#order_product_0'), 'EEE-TBS-TEST - WEEE')
        self.assertEquals(self._get_text_from_selector(css + '#order_product_1'), 'product power - WEEE-3')
        self.assertEquals(self._get_text_from_selector(css + '#order_product_2'), 'product quantity - WEEE-3')
        self.assertEquals(self._get_text_from_selector(css + '#order_product_3'), 'product surface - WEEE-3')

        self.assertEquals(self._get_text_from_selector(css + '#order_row_0 .line_quantity'), '11')
        self.assertEquals(self._get_text_from_selector(css + '#order_row_1 .line_quantity'), '100')
        self.assertEquals(self._get_text_from_selector(css + '#order_row_2 .line_quantity'), '10')
        self.assertEquals(self._get_text_from_selector(css + '#order_row_3 .line_quantity'), '10')

        self.assertEquals(self._get_text_from_selector(css + '#order_row_0 .line_weight'), '111.00')
        self.assertEquals(self._get_text_from_selector(css + '#order_row_1 .line_weight'), '200.00')
        self.assertEquals(self._get_text_from_selector(css + '#order_row_2 .line_weight'), '101.00')
        self.assertEquals(self._get_text_from_selector(css + '#order_row_3 .line_weight'), '11.00')

        self.assertEquals(self._get_text_from_selector(css + '#order_row_0 .line_custom_unit'), '')
        self.assertEquals(self._get_text_from_selector(css + '#order_row_1 .line_custom_unit'), '300.66')
        self.assertEquals(self._get_text_from_selector(css + '#order_row_2 .line_custom_unit'), '')
        self.assertEquals(self._get_text_from_selector(css + '#order_row_3 .line_custom_unit'), '12.50')

        self.assertEquals(self._get_text_from_selector(css + '#order_row_0 .line_line_amount_excluding_vat'), '1.66')
        self.assertEquals(self._get_text_from_selector(css + '#order_row_1 .line_line_amount_excluding_vat'), '5.11')
        self.assertEquals(self._get_text_from_selector(css + '#order_row_2 .line_line_amount_excluding_vat'), '0.16')
        self.assertEquals(self._get_text_from_selector(css + '#order_row_3 .line_line_amount_excluding_vat'), '0.22')

        self.assertEquals(self._get_text_from_selector(css + '.subtotal-amount'), u'€ 7.16')
        self.assertEquals(self._get_text_from_selector(css + '.total-vat'), u'€ 0.00')
        self.assertEquals(self._get_text_from_selector(css + '.total-amount'), u'€ 7.16')

    def test_contribution_lines_is_b2b(self):
        ma, tm = self._create_simple_member_for_declaration()

        # enable tbs b2b setting
        tm.tbs.set_config('pvc.tbs.require.end.user.distinction', True)

        d = self._january_declaration(ma, tm, 1)

        self.click_button(panel_name='pvc-contribution-details-panel', text='Add line')
        self.wait_for_loading()

        self.fill_form(panel_name='pvc-contribution-extra-line-product-select', values={
            'products': ('fancy-select', ['EEE-TBS-TEST-2 - WEEE']),
            'is_b2b': ('radio', '1')
        })

        self.click_button(panel_name='pvc-contribution-extra-line-product-select', text='Add to declaration')
        self.wait_for_loading()

        self.fill_form(panel_name='pvc-contribution-details-panel', values={
            'line_extra_2_quantity': 1024,
            'line_extra_2_weight': '123.45',
        })

        # checkity check that the member aint a liar.
        self._silly_honesty_check()
        self.wait_for_loading()

        self.click_button(panel_name='pvc-contribution-details-panel', text='Save')
        self.wait_for_loading()

        self.click_cq(panel_name='pvc-outstanding-contributions', row_index=0, col_name='details')
        self.wait_for_loading() # for the panel

        # assert new line is saved
        self._assert_order_line_is_b2b('pvc-contribution-details-panel', 'order_is_b2b', [1], 'B2B')

        # checkity check that the member aint a liar.
        self._silly_honesty_check()
        self.wait_for_loading()

        self.click_button(panel_name='pvc-contribution-details-panel', text='Submit')
        self.wait_for_loading()

        # assert new line in checkout mode
        self._assert_order_line_is_b2b('pvc-contribution-details-panel-checkout', 'order_is_b2b', [1], 'B2B')

    def _assert_order_line_is_b2b(self, panel_name, cell_id_name, row_index, value, form_name=None):
        css = self.wait_for_panel_or_form(panel_name, form_name)
        for row in row_index:

            input_css = "%s td#%s_%d" % (css, cell_id_name, row)
            input_el = self._driver.find_element_by_css_selector(input_css)
            self.assertEquals(input_el.get_attribute("innerHTML"), value)

    def test_contribution_lines_b2c_for_already_listed_product(self):
        ma, tm = self._create_simple_member_for_declaration()

        # enable tbs b2b setting
        tm.tbs.set_config('pvc.tbs.require.end.user.distinction', True)

        d = self._january_declaration(ma, tm, 1)

        self.click_button(panel_name='pvc-contribution-details-panel', text='Add line')
        self.wait_for_loading()

        self.fill_form(panel_name='pvc-contribution-extra-line-product-select', values={
            'products': ('fancy-select', ['EEE-TBS-TEST - WEEE']),
            'is_b2b': ('radio', '2')
        })

        self.click_button(panel_name='pvc-contribution-extra-line-product-select', text='Add to declaration')
        self.wait_for_loading()

        self.fill_form(panel_name='pvc-contribution-details-panel', values={
            'line_extra_2_quantity': 1024,
            'line_extra_2_weight': '123.45',
        })

        # checkity check that the member aint a liar.
        self._silly_honesty_check()
        self.wait_for_loading()

        self.click_button(panel_name='pvc-contribution-details-panel', text='Save')
        self.wait_for_loading()

        self.click_cq(panel_name='pvc-outstanding-contributions', row_index=0, col_name='details')
        self.wait_for_loading()

        self.sa_session.expire(d)
        self.assertEqual(len(d.lines), 2)

    def test_contribution_lines_b2b_b2c_no_dupes(self):
        ma, tm = self._create_simple_member_for_declaration()

        # enable tbs b2b setting
        tm.tbs.set_config('pvc.tbs.require.end.user.distinction', True)

        d = self._january_declaration(ma, tm, 1)

        self.click_button(panel_name='pvc-contribution-details-panel', text='Add line')
        self.wait_for_loading()

        self.fill_form(panel_name='pvc-contribution-extra-line-product-select', values={
            'products': ('fancy-select', ['EEE-TBS-TEST - WEEE']),
            'is_b2b': ('radio', '3')
        })

        self.click_button(panel_name='pvc-contribution-extra-line-product-select', text='Add to declaration')
        self.wait_for_loading()

        self.fill_form(panel_name='pvc-contribution-details-panel', values={
            'line_extra_2_quantity': 1024,
            'line_extra_2_weight': '123.45',
        })

        # checkity check that the member aint a liar.
        self._silly_honesty_check()
        self.wait_for_loading()

        self.click_button(panel_name='pvc-contribution-details-panel', text='Save')
        self.wait_for_loading()

        self.click_cq(panel_name='pvc-outstanding-contributions', row_index=0, col_name='details')
        self.wait_for_loading()

        self.sa_session.expire(d)
        self.assertEqual(len(d.lines), 2)

    def test_order_with_brands(self):
        ma, tm = self._create_simple_member_for_declaration()

        mp = self._get_member_product(tm, self.tbs_product2)
        brand1 = self._get_member_brand('Samsung', mp)
        brand2 = self._get_member_brand('Sony', mp)

        # enable tbs b2b setting
        tm.tbs.set_config('pvc.tbs.contribution.split.by.brand', True)

        d = self._january_declaration(ma, tm)

        self.click_button(panel_name='pvc-contribution-details-panel', text='Add line')
        self.wait_for_loading()

        self.fill_form(panel_name='pvc-contribution-extra-line-product-select', values={
            'products': ('fancy-select', ['EEE-TBS-TEST - WEEE'])  # same product as previous selection
        })

        self.click_button(panel_name='pvc-contribution-extra-line-product-select', text='Add to declaration')
        self.wait_for_loading()

        self.fill_form(panel_name='pvc-contribution-details-panel', values={
            'line_extra_2_quantity': 1024,
            'line_extra_2_weight': '123.45',
            'line_extra_2_brand': brand1.name
        })

        # duplicate line and enter a random brand name

        self.click_button_by_id('line_extra_2_duplicate')
        self.wait_for_loading()

        self.fill_form(panel_name='pvc-contribution-details-panel', values={
            'line_extra_3_quantity': 987,
            'line_extra_3_weight': '666.77',
            'line_extra_3_brand': self._uniquestring('brand')
        })

        self._silly_honesty_check()
        self.click_button(panel_name='pvc-contribution-details-panel', text='Save')
        self.wait_for_loading()

        self.sa_session.expire(d)
        self.assertEqual(len(d.lines), 3)

    def test_order_with_brands_and_rule_and_resubmit(self):
        # test case for ticket #1023
        self._make_rule('is compliant',
            order_type = 'PvcContributionOrder',
            description = 'discount cuz order iz compliant',
            conditions = [{
                'condition': 'is_compliant',
                'operator': None, 'value_text': None, 'value_bool': None,
                'value_int': PvcContributionOrder.IS_COMPLIANT,
                'value_decimal': None,
            },],
            actions = [{
                'action': 'fixed_amount_discount',
                'description': '',
                'operator': None, 'value_text': None, 'value_bool': None, 'value_int': None,
                'value_decimal': 1,
                'tax_class_id': None, 'gl_account_code': None
            }]
        )

        self.payment_methods = self._setup_payment_methods(self.operator_account, self.tbs_account)

        ma, tm = self._create_simple_member_for_declaration()

        mp = self._get_member_product(tm, self.tbs_product2)
        brand1 = self._get_member_brand('Samsung', mp)
        brand2 = self._get_member_brand('Sony', mp)

        # enable tbs b2b setting
        tm.tbs.set_config('pvc.tbs.contribution.split.by.brand', True)

        d = self._january_declaration(ma, tm)

        self.click_button(panel_name='pvc-contribution-details-panel', text='Add line')
        self.wait_for_loading()

        self.fill_form(panel_name='pvc-contribution-extra-line-product-select', values={
            'products': ('fancy-select', ['EEE-TBS-TEST - WEEE'])  # same product as previous selection
        })

        self.click_button(panel_name='pvc-contribution-extra-line-product-select', text='Add to declaration')
        self.wait_for_loading()

        self.fill_form(panel_name='pvc-contribution-details-panel', values={
            'line_extra_2_quantity': 1024,
            'line_extra_2_weight': '123.45',
            'line_extra_2_brand': brand1.name
        })

        self._silly_honesty_check()
        self.click_button(panel_name='pvc-contribution-details-panel', text='Submit')
        self.wait_for_loading(stable_time=4)

        # select 'bank transfer' payment method and continue
        self.fill_form('pvc-contribution-details-panel-checkout', values={
            'payment_method': ('radio', self.payment_methods[0].id)
        })

        self.click_button(panel_name='pvc-contribution-details-panel-checkout', text='Submit')
        self.wait_for_loading(times=2)

        self.click_button('order-payment_method-panel', text='Mark as submitted')
        self.wait_for_loading()

        self.sa_session.expire(d)
        self.assertEqual(len(d.lines), 3)

        d.resubmit_request = PvcContributionOrder.RESUBMIT_REQUEST_APPROVED
        self.sa_session.add(d)
        self.sa_session.flush()

        self.close_panel('pvc-outstanding-contributions')

        self.click_iconbar('icon-pvc-contribution-manager', hover='icon-pvc-outstanding-contributions')
        self.wait_for_panel('pvc-contribution-manager')

        self.filter_cq(panel_name='pvc-contribution-manager', filters=[
            ('Order ID', 'Equal', d.id),
        ])
        self.wait_for_loading() # for the custom query

        self.assert_cq(panel_name='pvc-contribution-manager', row_index=0, data={
            'action': 'Resubmit'
        })

        self.click_button(panel_name='pvc-contribution-manager', text='Resubmit')
        self.wait_for_loading() # for the details panel

        self.fill_form(panel_name='pvc-contribution-details-panel', values={
            'line_0_quantity': 400,
            'line_0_weight': '400',
            'line_1_quantity': 500,
            'line_1_weight': '500',
        })
        self._silly_honesty_check()
        self.click_button(panel_name='pvc-contribution-details-panel', text='Save')
        self.wait_for_loading()

    def test_order_with_brands_remove_line(self):
        ma, tm = self._create_simple_member_for_declaration()

        mp = self._get_member_product(tm, self.tbs_product2)
        brand1 = self._get_member_brand('Samsung', mp)
        brand2 = self._get_member_brand('Sony', mp)

        # enable tbs b2b setting
        tm.tbs.set_config('pvc.tbs.contribution.split.by.brand', True)

        d = self._january_declaration(ma, tm)

        self.click_button(panel_name='pvc-contribution-details-panel', text='Add line')
        self.wait_for_loading()

        self.fill_form(panel_name='pvc-contribution-extra-line-product-select', values={
            'products': ('fancy-select', ['EEE-TBS-TEST - WEEE'])  # same product as previous selection
        })

        self.click_button(panel_name='pvc-contribution-extra-line-product-select', text='Add to declaration')
        self.wait_for_loading()

        self.fill_form(panel_name='pvc-contribution-details-panel', values={
            'line_extra_2_quantity': 1024,
            'line_extra_2_weight': '123.45',
            'line_extra_2_brand': brand1.name
        })

        # duplicate line and enter a random brand name

        self.click_button_by_id('line_extra_2_duplicate')
        self.wait_for_loading()

        self.fill_form(panel_name='pvc-contribution-details-panel', values={
            'line_extra_3_quantity': 987,
            'line_extra_3_weight': '666.77',
            'line_extra_3_brand': self._uniquestring('brand')
        })

        self.click_button_by_id('line_extra_3_duplicate')
        self.wait_for_loading()

        self.fill_form(panel_name='pvc-contribution-details-panel', values={
            'line_extra_4_quantity': 234,
            'line_extra_4_weight': '324.77',
            'line_extra_4_brand': self._uniquestring('arand')
        })

        self._silly_honesty_check()

        self.click_button(panel_name='pvc-contribution-details-panel', text='Remove line')
        self.click_button(panel_name='dialog-question', text='Yes')
        self.wait_for_loading()

        self.sa_session.expire(d)
        self.assertEqual(len(d.lines), 3)

    def test_order_with_automatic_b2b_split(self):
        ma, tm = self._create_simple_member_for_declaration()
        tm.tbs.set_config('pvc.tbs.contribution.automatic.end.user.distinction', True)

        payment_methods = self._setup_payment_methods(self.operator_account, self.tbs_account)

        self._make_rule('italy_b2b_split',
            order_type='PvcContributionOrder',
            conditions=[{
                'condition': 'merchant',
                'operator': None, 'value_text': None, 'value_bool': None, 'value_int': tm.tbs.id,
                'value_decimal': None,
            }],
            actions=[{
                'action': 'pvc_contribution_b2b_percentage',
                'description': '',
                'operator': None, 'value_text': None, 'value_bool': None, 'value_int': 10,
                'value_decimal': None,
                'tax_class_id': None, 'gl_account_code': None
            }]
        )

        product = self.pc.save_or_update_tbs_product(None, 'EEE-TBS-TEST-DIMENSION', 'WEEE', self.tbs.id, basis_for_invoicing=PvcTbsProduct.INVOICE_BASED_ON_WEIGHT)

        self.pc.save_contribution_fee(product.id, self.service_level.id, Decimal('0.015'), date(2012, 1, 1), dimension=PvcTbsProductFee.DIMENSION_IS_B2C)
        self.pc.save_contribution_fee(product.id, self.service_level.id, Decimal('0.075'), date(2012, 1, 1), dimension=PvcTbsProductFee.DIMENSION_IS_B2B)

        d = self._january_declaration(ma, tm)

        self.click_button(panel_name='pvc-contribution-details-panel', text='Add line')
        self.wait_for_loading()

        self.fill_form(panel_name='pvc-contribution-extra-line-product-select', values={
            'products': ('fancy-select', ['EEE-TBS-TEST-DIMENSION - WEEE'])
        })

        self.click_button(panel_name='pvc-contribution-extra-line-product-select', text='Add to declaration')
        self.wait_for_loading()

        self.fill_form(panel_name='pvc-contribution-details-panel', values={
            'line_extra_2_quantity': 1024,
            'line_extra_2_weight': '123.45'
        })

        self._silly_honesty_check()
        self.click_button(panel_name='pvc-contribution-details-panel', text='Submit')
        self.wait_for_loading()

        self._assert_order_line_is_b2b('pvc-contribution-details-panel-checkout', 'order_is_b2b', [1, 3], 'B2B')
        self._assert_order_line_is_b2b('pvc-contribution-details-panel-checkout', 'order_is_b2b', [0, 2], 'B2C')

        self.fill_form('pvc-contribution-details-panel-checkout', values={
            'payment_method': ('radio', payment_methods[0].id)
        })
        self.wait_for_loading()

        self.click_button(panel_name='pvc-contribution-details-panel-checkout', text='Submit')
        self.wait_for_loading(times=2)

        self.click_button('order-payment_method-panel', text='Mark as submitted')
        self.wait_for_loading()

        # now lets do a resubmit

        d.resubmit_request = PvcContributionOrder.RESUBMIT_REQUEST_APPROVED
        self.sa_session.add(d)
        self.sa_session.flush()

        self.close_panel('pvc-outstanding-contributions')
        self.click_iconbar('icon-pvc-contribution-manager', hover='icon-pvc-outstanding-contributions')
        self.wait_for_loading()

        self.click_button('pvc-contribution-manager', text="Resubmit")
        self.wait_for_loading()

        self.fill_form(panel_name='pvc-contribution-details-panel', values={
            'line_0_quantity': 400,
            'line_0_weight': '400',
            'line_2_quantity': 500,
            'line_2_weight': '500',
        })
        self._silly_honesty_check()
        self.click_button(panel_name='pvc-contribution-details-panel', text='Save')
        self.wait_for_loading()

    def test_cancel_submit_with_additional_line(self):
        ma, tm = self._create_simple_member_for_declaration()
        d = self._january_declaration(ma, tm)

        self.click_button(panel_name='pvc-contribution-details-panel', text='Add line')
        self.wait_for_loading()

        self.fill_form(panel_name='pvc-contribution-extra-line-product-select', values={
            'products': ('fancy-select', ['EEE-TBS-TEST-2 - WEEE'])
        })

        self.click_button(panel_name='pvc-contribution-extra-line-product-select', text='Add to declaration')
        self.wait_for_loading()

        self.fill_form(panel_name='pvc-contribution-details-panel', values={
            'line_extra_2_quantity': 1024,
            'line_extra_2_weight': '123.45'
        })

        self._silly_honesty_check()
        self.click_button(panel_name='pvc-contribution-details-panel', text='Submit')
        self.wait_for_loading(stable_time=3)

        self.click_button(panel_name='pvc-contribution-details-panel-checkout', text="Cancel")
        self.wait_for_loading()

        self.assert_form(panel_name='pvc-contribution-details-panel', values={
            'line_0_quantity': '200',
            'line_0_weight': '202.40000',
            'line_1_quantity': '1024',
            'line_1_weight': '123.45000'
        })

        self.fill_form(panel_name='pvc-contribution-details-panel', values={
            'line_0_quantity': 201,
            'line_1_quantity': 1023,
        })

        self.click_button(panel_name='pvc-contribution-details-panel', text='Save')
        self.wait_for_loading()

        self.sa_session.expire(d)

        self.assertEqual(d.lines[0].quantity, 201)
        self.assertEqual(d.lines[1].quantity, 1023)

    def test_forecast_first_time_login(self):
        ma, tm = self._create_simple_member_for_declaration()
        tm.tbs.set_config('pvc.tbs.advance.invoicing', True)
        tm.tbs.set_config('pvc.tbs.declaration.period', PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY)

        self.login(ma.name, 'test_password')
        self.wait_for_loading()

        self.click_button(panel_name='dialog-information', text='OK')

        self.fill_form(panel_name='pvc-contribution-extra-line-product-select', values={
            'products': ('fancy-select', ['EEE-TBS-TEST - WEEE', 'EEE-TBS-TEST-2 - WEEE'])
        })

        self.click_button(panel_name='pvc-contribution-extra-line-product-select', text='Add to declaration')
        self.wait_for_loading()

        self.assert_form(panel_name='pvc-contribution-details-panel', values={
            'line_0_quantity': '0',
            'line_0_weight': '0.00000',
            'line_1_quantity': '0',
            'line_1_weight': '0.00000',
        })

    def test_forecast_first_time_login_additional_line_manual_dimension_split(self):
        ma, tm = self._create_simple_member_for_declaration()
        tm.tbs.set_config('pvc.tbs.advance.invoicing', True)
        tm.tbs.set_config('pvc.tbs.require.end.user.distinction', True)
        tm.tbs.set_config('pvc.tbs.declaration.period', PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY)

        self.login(ma.name, 'test_password')
        self.wait_for_loading()

        self.click_button(panel_name='dialog-information', text='OK')

        self.fill_form(panel_name='pvc-contribution-extra-line-product-select', values={
            'products': ('fancy-select', ['EEE-TBS-TEST - WEEE']),
            'is_b2b': ('radio', '1')
        })

        self.click_button(panel_name='pvc-contribution-extra-line-product-select', text='Add to declaration')
        self.wait_for_loading()
        self.wait_for_panel(panel_name='pvc-contribution-details-panel')
        self.wait_for_loading()

        self.assert_form(panel_name='pvc-contribution-details-panel', values={
            'line_0_quantity': '0',
            'line_0_weight': '0.00000',
        })

        self.click_button(panel_name='pvc-contribution-details-panel', text='Add line')
        self.wait_for_loading()

        self.fill_form(panel_name='pvc-contribution-extra-line-product-select', values={
            'products': ('fancy-select', ['EEE-TBS-TEST-2 - WEEE']),
            'is_b2b': ('radio', '3')
        })

        self.click_button(panel_name='pvc-contribution-extra-line-product-select', text='Add to declaration')
        self.wait_for_loading()

        self.fill_form(panel_name='pvc-contribution-details-panel', values={
            'line_extra_2_quantity': 1024,
            'line_extra_2_weight': '123.45',
            'line_extra_3_quantity': 234,
            'line_extra_3_weight': '54.23',
        })


    def test_forecast_first_time_login_during_declaration_period(self):
        ma, tm = self._create_simple_member_for_declaration()
        tm.tbs.set_config('pvc.tbs.advance.invoicing', True)
        tm.tbs.set_config('pvc.tbs.declaration.closing.day', 10)
        tm.tbs.set_config('pvc.tbs.declaration.period', PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY)

        set_date(date(2013, 7, 2))  # Declaration period over Quarter 2, so forecast should be Quarter 3

        self.login(ma.name, 'test_password')
        self.wait_for_loading()

        self.click_button(panel_name='dialog-information', text='OK')

        self.fill_form(panel_name='pvc-contribution-extra-line-product-select', values={
            'products': ('fancy-select', ['EEE-TBS-TEST - WEEE', 'EEE-TBS-TEST-2 - WEEE'])
        })

        self.click_button(panel_name='pvc-contribution-extra-line-product-select', text='Add to declaration')
        self.wait_for_loading()

        self.assert_form(panel_name='pvc-contribution-details-panel', values={
            'line_0_quantity': '0',
            'line_0_weight': '0.00000',
            'line_1_quantity': '0',
            'line_1_weight': '0.00000',
        })

        self.fill_form(panel_name='pvc-contribution-details-panel', values={
            'line_0_quantity': '10',
            'line_0_weight': '2.50000',
            'line_1_quantity': '15',
            'line_1_weight': '9.75000',
        })

        self._silly_honesty_check()
        self.click_button(panel_name='pvc-contribution-details-panel', text='Submit')
        self.wait_for_loading()

        cc = PvcContributionOrderController(self.sa_session, account_id=ma.id)
        orders = cc.get()
        self.assertEqual(len(orders), 1)
        self.assertEqual(orders[0].quarter, 3)

    def test_forecast_first_time_login_split_manual_b2b_b2c(self):
        ma, tm = self._create_simple_member_for_declaration()
        tm.tbs.set_config('pvc.tbs.advance.invoicing', True)
        tm.tbs.set_config('pvc.tbs.require.end.user.distinction', True)
        tm.tbs.set_config('pvc.tbs.declaration.period', PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY)

        self.login(ma.name, 'test_password')
        self.wait_for_loading()

        self.click_button(panel_name='dialog-information', text='OK')

        self.fill_form(panel_name='pvc-contribution-extra-line-product-select', values={
            'products': ('fancy-select', ['EEE-TBS-TEST - WEEE', 'EEE-TBS-TEST-2 - WEEE']),
            'is_b2b': ('radio', '1')
        })

        self.click_button(panel_name='pvc-contribution-extra-line-product-select', text='Add to declaration')
        self.wait_for_loading()

        self.assert_form(panel_name='pvc-contribution-details-panel', values={
            'line_0_quantity': '0',
            'line_0_weight': '0.00000',
            'line_1_quantity': '0',
            'line_1_weight': '0.00000',
        })

        self._assert_order_line_is_b2b('pvc-contribution-details-panel', 'order_is_b2b', [0], 'B2B')
        self._assert_order_line_is_b2b('pvc-contribution-details-panel', 'order_is_b2b', [1], 'B2B')

    def test_forecast_first_time_login_split_manual_both_b2b_b2c(self):
        ma, tm = self._create_simple_member_for_declaration()
        tm.tbs.set_config('pvc.tbs.advance.invoicing', True)
        tm.tbs.set_config('pvc.tbs.require.end.user.distinction', True)
        tm.tbs.set_config('pvc.tbs.declaration.period', PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY)

        self.login(ma.name, 'test_password')
        self.wait_for_loading()

        self.click_button(panel_name='dialog-information', text='OK')

        self.fill_form(panel_name='pvc-contribution-extra-line-product-select', values={
            'products': ('fancy-select', ['EEE-TBS-TEST - WEEE', 'EEE-TBS-TEST-2 - WEEE']),
            'is_b2b': ('radio', '3')
        })

        self.click_button(panel_name='pvc-contribution-extra-line-product-select', text='Add to declaration')
        self.wait_for_loading()

        self.assert_form(panel_name='pvc-contribution-details-panel', values={
            'line_0_quantity': '0',
            'line_0_weight': '0.00000',
            'line_1_quantity': '0',
            'line_1_weight': '0.00000',
        })

        self._assert_order_line_is_b2b('pvc-contribution-details-panel', 'order_is_b2b', [0], 'B2C')
        self._assert_order_line_is_b2b('pvc-contribution-details-panel', 'order_is_b2b', [1], 'B2B')
        self._assert_order_line_is_b2b('pvc-contribution-details-panel', 'order_is_b2b', [2], 'B2C')
        self._assert_order_line_is_b2b('pvc-contribution-details-panel', 'order_is_b2b', [3], 'B2B')

    def test_forecast_first_time_login_automatic_split_b2b_b2c(self):
        ma, tm = self._create_simple_member_for_declaration()
        tm.tbs.set_config('pvc.tbs.advance.invoicing', True)
        tm.tbs.set_config('pvc.tbs.contribution.automatic.end.user.distinction', True)
        tm.tbs.set_config('pvc.tbs.declaration.period', PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY)

        self._make_rule('italy_b2b_split',
            order_type='PvcContributionOrder',
            conditions=[{
                'condition': 'merchant',
                'operator': None, 'value_text': None, 'value_bool': None, 'value_int': tm.tbs.id,
                'value_decimal': None,
            }],
            actions=[{
                'action': 'pvc_contribution_b2b_percentage',
                'description': '',
                'operator': None, 'value_text': None, 'value_bool': None, 'value_int': 10,
                'value_decimal': None,
                'tax_class_id': None, 'gl_account_code': None
            }]
        )

        self.login(ma.name, 'test_password')
        self.wait_for_loading()

        self.click_button(panel_name='dialog-information', text='OK')

        self.fill_form(panel_name='pvc-contribution-extra-line-product-select', values={
            'products': ('fancy-select', ['EEE-TBS-TEST - WEEE', 'EEE-TBS-TEST-2 - WEEE']),
        })

        self.click_button(panel_name='pvc-contribution-extra-line-product-select', text='Add to declaration')
        self.wait_for_loading()

        self.assert_form(panel_name='pvc-contribution-details-panel', values={
            'line_0_quantity': '0',
            'line_0_weight': '0.00000',
            'line_1_quantity': '0',
            'line_1_weight': '0.00000',
        })

        self._silly_honesty_check()
        self.click_button(panel_name='pvc-contribution-details-panel', text='Submit')
        self.wait_for_loading()

        self._assert_order_line_is_b2b('pvc-contribution-details-panel-checkout', 'order_is_b2b', [1, 3], 'B2B')
        self._assert_order_line_is_b2b('pvc-contribution-details-panel-checkout', 'order_is_b2b', [0, 2], 'B2C')

    def test_forecast_first_time_login_split_by_brand(self):
        ma, tm = self._create_simple_member_for_declaration()
        tm.tbs.set_config('pvc.tbs.advance.invoicing', True)
        tm.tbs.set_config('pvc.tbs.contribution.split.by.brand', True)
        tm.tbs.set_config('pvc.tbs.declaration.period', PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY)

        self.login(ma.name, 'test_password')
        self.wait_for_loading()

        self.click_button(panel_name='dialog-information', text='OK')

        self.fill_form(panel_name='pvc-contribution-extra-line-product-select', values={
            'products': ('fancy-select', ['EEE-TBS-TEST - WEEE', 'EEE-TBS-TEST-2 - WEEE']),
        })

        self.fill_form(panel_name='pvc-contribution-extra-line-product-select', values={
            'brand': 'Philips'
        })

        self.click_button(panel_name='pvc-contribution-extra-line-product-select', text='Add to declaration')
        self.wait_for_loading(stable_time=2)

        self.assert_form(panel_name='pvc-contribution-details-panel', values={
            'line_0_quantity': '0',
            'line_0_weight': '0.00000',
            'line_0_brand': 'Philips',
            'line_1_quantity': '0',
            'line_1_weight': '0.00000',
            'line_1_brand': 'Philips',
        })

        self.fill_form(panel_name='pvc-contribution-details-panel', values={
            'line_0_quantity': '4',
            'line_0_weight': '1.23',
            'line_0_brand': 'Philips',
            'line_1_quantity': '10',
            'line_1_weight': '3.86',
            'line_1_brand': 'Sony',
        })

        self._silly_honesty_check()
        self.click_button(panel_name='pvc-contribution-details-panel', text='Submit')
        self.wait_for_loading()

    def test_manual_declaration(self):

        ma, tm = self._create_simple_member_for_declaration()

        self._get_member_product(tm, self.tbs_product)  # add a product so we can actual trigger a declaration

        self.pc.save_contribution_fee(self.tbs_product.id, self.service_level.id, Decimal('0.015'), date(2012, 1, 1))

        tm.declaration_periodicity = PvcTbsLicense.DECLARATION_PERIOD_MONTHLY
        self.sa_session.add(tm)
        self.sa_session.flush()
        self.sa_session.expire(tm)

        self.login(ma.name, 'test_password')
        self.wait_for_loading()

        self.click_iconbar('icon-pvc-contribution-manager', hover='icon-pvc-outstanding-contributions')
        self.wait_for_loading()

        self.click_menu(panel_name='pvc-contribution-manager', text='Add declaration')
        self.wait_for_loading()

        self.fill_form(panel_name='contribution-select-panel', values={
            'tbs_id': ('select', 'value=%d' % tm.tbs.id)
        })
        self.wait_for_loading()

        self.fill_form(panel_name='contribution-select-panel', values={
            'year': ('select', '2012'),
        })
        self.fill_form(panel_name='contribution-select-panel', values={
            'month': ('select', 'value=9'),
        })

        self.click_button(panel_name='contribution-select-panel', text='Save')
        self.wait_for_loading()

        self.assert_form(panel_name='pvc-contribution-details-panel', values={
            'line_0_quantity': '0',
            'line_0_weight': '0.00000',
        })

    def test_manual_declaration_resigning_member(self):

        ma, tm = self._create_simple_member_for_declaration()

        self._get_member_product(tm, self.tbs_product)  # add a product so we can actual trigger a declaration

        self.pc.save_contribution_fee(self.tbs_product.id, self.service_level.id, Decimal('0.015'), date(2012, 1, 1))

        tm.declaration_periodicity = PvcTbsLicense.DECLARATION_PERIOD_MONTHLY
        tm.status = PvcTbsMember.STATUS_RESIGNING
        self.sa_session.add(tm)
        self.sa_session.flush()
        self.sa_session.expire(tm)

        self.login(ma.name, 'test_password')
        self.wait_for_loading()

        self.click_iconbar('icon-pvc-contribution-manager', hover='icon-pvc-outstanding-contributions')
        self.wait_for_loading()

        self.click_menu(panel_name='pvc-contribution-manager', text='Add declaration')
        self.wait_for_loading()

        self.fill_form(panel_name='contribution-select-panel', values={
            'tbs_id': ('select', 'value=%d' % tm.tbs.id)
        })
        self.wait_for_loading()

        self.fill_form(panel_name='contribution-select-panel', values={
            'year': ('select', '2012'),
        })
        self.fill_form(panel_name='contribution-select-panel', values={
            'month': ('select', 'value=9'),
        })

        self.click_button(panel_name='contribution-select-panel', text='Save')
        self.wait_for_loading()

        self.assert_form(panel_name='pvc-contribution-details-panel', values={
            'line_0_quantity': '0',
            'line_0_weight': '0.00000',
        })

    # - subrogation test cases - #
    def _create_commissioner(self, obligation_start_date=None):
        ma = self._create_member_account(company_name='commissioner')

        # apply at TBS
        tm = self._member_apply_at_tbs(ma, self.tbs_account, self.service_level, is_commissioner=True)

        obligation_start_date = date(2012, 1, 1) if obligation_start_date is None else obligation_start_date

        # activate member to automatically create an order entry for the Administration Fee
        mc = PvcMemberController(self.sa_session, account_id=self.tbs_account.id)
        mc.save_tbs_member(tm.id, {
            'status': PvcTbsMember.STATUS_APPROVED,
            'is_full_member': False,
            'is_representative': False,
            'obligation_start_date': obligation_start_date})
        return ma, tm

    def _create_subrogation(self, ma, tm, ctm):
        controller = PvcSubrogationController(self.sa_session, account_id=ma.id)
        subrogation = controller.request_for_subrogation(tm, 'commissioner_name')
        # Set subrogation status active.
        subrogation.status = PvcSubrogation.STATUS_ACTIVE
        subrogation.commissioner = ctm
        self.sa_session.add(subrogation)
        self.sa_session.flush()
        return subrogation

    def test_manual_declaration_for_subrogation(self):
        ma, tm = self._create_simple_member_for_declaration()
        self._get_member_product(tm)
        self._get_member_product(tm, self.tbs_product2)

        cma, ctm = self._create_commissioner()

        subrogation = self._create_subrogation(ma, tm, ctm)

        # login as producer
        self.login(ma.name, 'test_password')

        # open product manager
        self.click_iconbar('icon-pvc-tbs-member-product-manager')
        self.wait_for_loading()

        # add new product for subrogation
        self.click_menu(panel_name='pvc-tbs-member-product-manager', text="Add products")
        self.wait_for_loading()

        self.fill_form('pvc-tbs-member-product-select', values={
            'tbs_id': ('select', 'value=%d' % tm.tbs.id)
        })
        self.wait_for_loading()

        # select commissioner
        self.fill_form('pvc-tbs-member-product-select', values={
            'subrogation_id': ('select', 'value=%d' % subrogation.id)
        })
        self.wait_for_loading()

        # add the product
        self.fill_form('pvc-tbs-member-product-select', values={
            'products': ('fancy-select', ['EEE-TBS-TEST-2 - WEEE']),
        })
        self.click_button('pvc-tbs-member-product-select', text='Save')
        self.wait_for_loading()

        # add new declarations for subrogation.
        self.click_iconbar('icon-pvc-contribution-manager', hover='icon-pvc-outstanding-contributions')
        self.wait_for_loading()

        self.click_menu('pvc-contribution-manager', text="Add declaration")
        self.wait_for_loading()

        # select TBS,
        self.fill_form('contribution-select-panel', values={
            'tbs_id': ('select', 'value=%d' % tm.tbs.id)
        })
        self.wait_for_loading()

        # select commissioner
        self.fill_form('contribution-select-panel', values={
            'subrogation_id': ('select', 'value=%d' % subrogation.id)
        })
        self.wait_for_loading()

        self.fill_form('contribution-select-panel', values={
            'year': ('select', '2013')
        })
        self.wait_for_loading()
        self.fill_form('contribution-select-panel', values={
            'month': ('select', 'January')
        })

        self.click_button('contribution-select-panel', text='Save')
        self.wait_for_loading(stable_time=2)

        # test add line
        self.click_button(panel_name='pvc-contribution-details-panel', text='Add line')
        self.wait_for_loading()

        self.fill_form(panel_name='pvc-contribution-extra-line-product-select', values={
            'products': ('fancy-select', ['EEE-TBS-TEST-3 - WEEE'])
        })
        self.click_button(panel_name='pvc-contribution-extra-line-product-select', text='Add to declaration')
        self.wait_for_loading()

        self.fill_form(panel_name='pvc-contribution-details-panel', values={
            'line_extra_2_quantity': 1024,
            'line_extra_2_weight': '123.45'
        })

        # ????
        self._silly_honesty_check()
        self.click_button('pvc-contribution-details-panel', text="Save")
        self.wait_for_loading()

        # profit
        self.assert_cq('pvc-contribution-manager', row_index=0, data={
            'commissioner': ctm.member.name
        })
        products = self.sa_session.query(PvcTbsMemberProduct).filter(PvcTbsMemberProduct.subrogation == subrogation).all()
        self.assertEqual(len(products), 2)

    def test_subrogate_line(self):
        ma, tm = self._create_simple_member_for_declaration()
        self._get_member_product(tm)
        self._get_member_product(tm, self.tbs_product2)

        cma, ctm = self._create_commissioner()

        subrogation = self._create_subrogation(ma, tm, ctm)

        set_date(date(2013, 7, 1))

        pc = PvcContributionOrderController(self.sa_session)
        orders = pc.trigger_new_contributions(tm.tbs)
        self.assertEqual(len(orders), 1)
        d = orders.pop()

        dc = PvcContributionOrderController(self.sa_session, account_id=ma.id)

        self.login(ma.name, 'test_password')
        self.click_iconbar('icon-pvc-outstanding-contributions')
        self.wait_for_loading() # for the custom query

        self.filter_cq(panel_name='pvc-outstanding-contributions', filters=[
            ('Year', 'Equal', ('select', '2013')),
            ('Month', 'Equal', ('select', '6')),
        ])
        self.wait_for_loading() # for the custom query

        self.assert_cq(panel_name='pvc-outstanding-contributions', row_index=0, data={
            'action': 'Declare'
        })

        self.click_button(panel_name='pvc-outstanding-contributions', text='Declare')
        self.wait_for_loading(stable_time=4) # for the details panel

        self.fill_form(panel_name='pvc-contribution-details-panel', values={
            'line_0_quantity': 400,
            'line_0_weight': '123.45',
            'line_1_quantity': 500,
            'line_1_weight': '543.21',
        })

        self.fill_form(panel_name='pvc-contribution-details-panel', values={
            'subrogation_line_1': ('select', 'commissioner'),
        })
        self._silly_honesty_check()
        self.wait_for_loading()

        self.click_button(panel_name='pvc-contribution-details-panel', text='Save')
        self.wait_for_loading()

        self.assert_cq(panel_name='pvc-outstanding-contributions', row_index=0, data={
            'commissioner': tm.member.name
        })

        self.assert_cq(panel_name='pvc-outstanding-contributions', row_index=1, data={
            'commissioner': ctm.member.name
        })

    def test_multi_country_add_line_with_brands(self):
        set_date(date(2013,1,1))

        self.reset_roe()

        state = self._get_tbs_license_state()
        state['tbs']['country'] = None
        tbs_account = self._create_tbs_account(self.operator_account, state=state)
        tbs = tbs_account.default_group.license
        tbs.set_config('pvc.tbs.contribution.split.by.brand', True)
        self.assertEqual(tbs.country, None)

        products = self.pc.get_available_products_for_tbs(tbs.id)
        tbs_product = self.pc.save_or_update_tbs_product(None, 'EEE-TBS-TEST-2', 'WEEE', tbs.id,
            parent_level_id=products[1]['id'], basis_for_invoicing=PvcTbsProduct.INVOICE_BASED_ON_WEIGHT
        )
        tbs_product2 = self.pc.save_or_update_tbs_product(None, 'EEE-TBS-TEST-4', 'WEEE', tbs.id,
            parent_level_id=products[2]['id'], basis_for_invoicing=PvcTbsProduct.INVOICE_BASED_ON_WEIGHT
        )
        service_level = self._create_tbs_service_level(tbs_account, admin_fee=Decimal('200.2'), year=today().year)

        self.pc.save_contribution_fee(tbs_product.id, service_level.id, Decimal('0.035'), today())
        self.pc.save_contribution_fee(tbs_product2.id, service_level.id, Decimal('0.115'), today())

        ma, tm = self._create_simple_member_for_declaration(tbs_account=tbs_account, service_level=service_level)

        countries = self.pc.get_tbs_member_countries(tm)
        country = countries[0]

        cc = PvcContributionOrderController(self.sa_session, account_id=ma.id)
        tmp = self._get_member_product(tm, tbs_product=tbs_product, country=country)
        self._get_member_brand('Speculaas', tmp)

        set_date(date(2013, 7, 1))

        pc = PvcContributionOrderController(self.sa_session)
        orders = pc.trigger_new_contributions(tm.tbs)
        self.assertEqual(len(orders), 1)
        d = orders.pop()

        self.login(ma.name, 'test_password')
        self.click_iconbar('icon-pvc-outstanding-contributions')
        self.wait_for_loading() # for the custom query

        self.filter_cq(panel_name='pvc-outstanding-contributions', filters=[
            ('Year', 'Equal', ('select', '2013')),
            ('Month', 'Equal', ('select', '6')),
        ])
        self.wait_for_loading() # for the custom query
        self.assert_cq(panel_name='pvc-outstanding-contributions', row_index=0, data={
            'action': 'Declare'
        })

        self.click_button(panel_name='pvc-outstanding-contributions', text='Declare')
        self.wait_for_loading(stable_time=4) # for the details panel

        self.click_button(panel_name='pvc-contribution-details-panel', text='Add line')
        self.wait_for_loading()

        self.fill_form(panel_name='pvc-contribution-extra-line-product-select', values={
            'products': ('fancy-select', ['EEE-TBS-TEST-4 - WEEE']),
            'countries': ('fancy-select', ['Austria', 'Bulgaria', 'Croatia'])
        })
        self.click_button(panel_name='pvc-contribution-extra-line-product-select', text='Add to declaration')
        self.wait_for_loading()

        self.fill_form(panel_name='pvc-contribution-details-panel', values={
            'line_extra_2_quantity': 1024,
            'line_extra_2_weight': '32.00',
            'line_extra_2_brand': 'Brandy',
            'line_extra_3_quantity': 2048,
            'line_extra_3_weight': '16.00',
            'line_extra_3_brand': 'Brando',
            'line_extra_4_quantity': 4096,
            'line_extra_4_weight': '8.00',
            'line_extra_4_brand': 'Brawndo',
        })

        self._silly_honesty_check()
        self.wait_for_loading()

        self.click_button(panel_name='pvc-contribution-details-panel', text='Save')
        self.wait_for_loading()

        self.close_panel(panel_name='pvc-outstanding-contributions')
        self.click_iconbar('icon-pvc-outstanding-contributions')
        self.wait_for_loading() # for the custom query

        self.filter_cq(panel_name='pvc-outstanding-contributions', filters=[
            ('Year', 'Equal', ('select', '2013')),
            ('Month', 'Equal', ('select', '6')),
        ])
        self.wait_for_loading() # for the custom query

        self.assert_cq(panel_name='pvc-outstanding-contributions', row_index=0, data={
            'action': 'Redeclare',
            'total_amount': u'€ 6.44'
        })

        self.sa_session.expire(d)
        self.assertEqual(len(d.lines), 4)

    def test_order_simulate(self):

        ma, tm = self._create_simple_member_for_declaration()

        cc = PvcContributionOrderController(self.sa_session, account_id=ma.id)
        product_weight = self._get_member_product(tm, tbs_product=self.tbs_product)

        self._make_rule('is compliant',
            order_type = 'PvcContributionOrder',
            description = 'discount cuz order iz compliant',
            conditions = [{
                'condition': 'is_compliant',
                'operator': None, 'value_text': None, 'value_bool': None,
                'value_int': PvcContributionOrder.IS_COMPLIANT,
                'value_decimal': None,
            },],
            actions = [{
                'action': 'fixed_amount_discount',
                'description': '',
                'operator': None, 'value_text': None, 'value_bool': None, 'value_int': None,
                'value_decimal': 1,
                'tax_class_id': None, 'gl_account_code': None
            }]
        )

        self.login(self.operator_account.name,  'test_password')

        self.click_iconbar(title='Order simulation manager', hover='Declaration manager')
        self.click_menu(panel_name='order-simulation-manager', text='Create new Order simulation')

        self.fill_form(panel_name='order-simulation-details', values={
            'name': 'Order sim1',
            'polymorphic_type': ('select', 'Contribution'),
        })
        self.wait_for_loading()

        values = OrderedDict()
        values['merchant'] = ('autocomplete', tm.tbs.name)
        values['tbs_member'] = ('autocomplete', tm.member.name)
        values['producer'] = ('autocomplete', tm.member.name)
        values['year'] = ('autocomplete', '2014')
        #values['quarter'] = ('select', '1')
        values['month'] = ('autocomplete', '1')
        values['is_forecast'] = ('autocomplete', 'Is a forecast')

        for k,v in values.iteritems():
            self.fill_form(panel_name='order-simulation-details', values={k: v})
            self.wait_for_loading()

        self.fill_form(panel_name='order-simulation-details', rowcontrol_name='lines', rowcontrol_row_index=0, values={
            'polymorphic_type': ('select', 'Contribution line'),
        })

        self.wait_for_loading()

        self.fill_form(panel_name='order-simulation-details', rowcontrol_name='lines', rowcontrol_row_index=0, values=OrderedDict([
            ('tbs_member_product', ('autocomplete', 'EEE-TBS-TEST'))
        ]))
        self.wait_for_loading()

        self.fill_form(panel_name='order-simulation-details', rowcontrol_name='lines', rowcontrol_row_index=0, values=OrderedDict([
            ('description', 'Line 1'),
            ('quantity', '10'),
            ('price', '5'),
            ('weight', '20'),
            ('custom_unit', '30'),
        ]))

        self.click_button(panel_name='order-simulation-details', text='Simulate submit')
        self.wait_for_loading()

        # verify last amount golumns
        amounts = [
            '121.0000',
            '-1.0000',
            '99.0000',
            '21.0000',
            '120.0000',
        ]
        last_tds = self._driver.find_elements_by_css_selector('div[name="order-simulation-result"] td:last-child')
        for i in range(len(last_tds)):
            td = last_tds[i]
            self.assertEqual(td.text, amounts[i])

    def test_guarantee_forecast_next_year(self):

        set_date(date(2015, 11, 1) )

        ma, tm = self._create_simple_member_for_declaration()
        tm.tbs.set_config('pvc.tbs.guarantee.forecast', True)

        tbs_account = self.tbs_account
        operator_account = self.operator_account
        tbs = self.tbs

        service_level_pv_1 = self._create_tbs_service_level(tbs_account, name='SL PV 1',
            year=today().year, admin_fee=Decimal('300.00'), level=1, product_type=PvcTbsProduct.TYPE_PV_RELATED, guarantee_forecast_min_value=0)
        service_level_pv_2 = self._create_tbs_service_level(tbs_account, name='SL PV 2',
            year=today().year, admin_fee=Decimal('300.00'), level=2, product_type=PvcTbsProduct.TYPE_PV_RELATED, guarantee_forecast_min_value=100)

        # add tbs products
        pc = PvcProductController(self.sa_session, account_id=operator_account.id)
        tbs_pv_product = pc.save_or_update_tbs_product(None, 'tbs pv product', 'desc', tbs.id,
            basis_for_invoicing=PvcTbsProduct.INVOICE_BASED_ON_WEIGHT,
            custom_unit=PvcTbsProduct.CUSTOM_UNIT_NOMINAL_POWER
        )
        self.pc.save_contribution_fee(tbs_pv_product.id, self.service_level.id, Decimal('0.075'), date(2015, 1, 1))
        self.pc.save_contribution_fee(tbs_pv_product.id, service_level_pv_1.id, Decimal('0.055'), date(2015, 1, 1))
        self.pc.save_contribution_fee(tbs_pv_product.id, service_level_pv_2.id, Decimal('0.025'), date(2015, 1, 1))

        product = self._get_member_product(tm)

        self.login(ma.name, 'test_password')

        self.wait_for_loading()

        self.click_button(panel_name='dialog-information', text='OK')

        # also select the pv product
        self.fill_form(panel_name='pvc-contribution-extra-line-product-select', values={
            'products': ('fancy-select', ['tbs pv product - desc'])
        })

        self.click_button(panel_name='pvc-contribution-extra-line-product-select', text='Add to declaration')
        self.wait_for_loading()

        #
        # contribution panel
        #

        self.assertTrue(self.get_panel_title('pvc-contribution-details-panel').find('New declaration for 2016') == 0)

        self.fill_form(panel_name='pvc-contribution-details-panel', values={
            'line_0_quantity': 100,
            'line_0_weight': '4.0000',
            'line_1_quantity': 100,
            'line_1_weight': '0',
            'line_1_custom_unit': '0',
        })

        self.fill_form(panel_name='pvc-contribution-details-panel', values={
            'agreement': ('checkbox2', True)
        })
        self.wait_for_loading()

        self.click_button(panel_name='pvc-contribution-details-panel', text='Submit')
        self.wait_for_loading(assert_no_errors=False)
        self.assert_growl_errors_present()

        self.click_growl('critical')
        self.wait_for_loading()

        self.fill_form(panel_name='pvc-contribution-details-panel', values={
            'line_1_weight': '100',
            'line_1_custom_unit': '100',
        })
        self.click_button(panel_name='pvc-contribution-details-panel', text='Submit')
        self.wait_for_loading()

        self.assertEquals(self._get_text_from_selector('div[name="pvc-contribution-details-panel-checkout"] .total-amount'), u'€ 0.00')

        self.click_button(panel_name='pvc-contribution-details-panel-checkout', text='Submit')
        self.wait_for_loading(times=2)

        # time travel to the future to test Manual Declaration in same year (#928)
        set_date(date(2016, 5, 1) )

        self.click_iconbar('icon-pvc-contribution-manager', hover='icon-pvc-outstanding-contributions')
        self.wait_for_loading()

        self.click_menu(panel_name='pvc-contribution-manager', text='Add declaration')
        self.wait_for_loading()

        self.fill_form(panel_name='contribution-select-panel', values={
            'tbs_id': ('select', 'value=%d' % tm.tbs.id)
        })
        self.wait_for_loading()

        self.fill_form(panel_name='contribution-select-panel', values={
            'year': ('select', '2016'),
        })
        self.fill_form(panel_name='contribution-select-panel', values={
            'month': ('select', 'January'),
        })
        self.click_button(panel_name='contribution-select-panel', text='Save')
        self.wait_for_loading()

        self.assert_form(panel_name='pvc-contribution-details-panel', values={
            'line_0_quantity': '0',
            'line_0_weight': '0.00000',
        })


    def test_guarantee_forecast_next_year_in_2016(self):
        """
        tests a member registering in 2015 but logs in in 2016. expected is the FG from 2016 to show up
        """

        set_date(date(2015, 11, 1) )

        ma, tm = self._create_simple_member_for_declaration()
        tm.tbs.set_config('pvc.tbs.guarantee.forecast', True)

        tbs_account = self.tbs_account
        operator_account = self.operator_account
        tbs = self.tbs

        service_level_pv_1 = self._create_tbs_service_level(tbs_account, name='SL PV 1',
            year=today().year, admin_fee=Decimal('300.00'), level=1, product_type=PvcTbsProduct.TYPE_PV_RELATED, guarantee_forecast_min_value=0)
        service_level_pv_2 = self._create_tbs_service_level(tbs_account, name='SL PV 2',
            year=today().year, admin_fee=Decimal('300.00'), level=2, product_type=PvcTbsProduct.TYPE_PV_RELATED, guarantee_forecast_min_value=100)

        # add tbs products
        pc = PvcProductController(self.sa_session, account_id=operator_account.id)
        tbs_pv_product = pc.save_or_update_tbs_product(None, 'tbs pv product', 'desc', tbs.id,
            basis_for_invoicing=PvcTbsProduct.INVOICE_BASED_ON_WEIGHT,
            custom_unit=PvcTbsProduct.CUSTOM_UNIT_NOMINAL_POWER
        )

        product = self._get_member_product(tm)

        set_date(date(2016, 1, 1) )

        self.login(ma.name, 'test_password')

        self.wait_for_loading()

        self.click_button(panel_name='dialog-information', text='OK')

        # also select the pv product
        self.fill_form(panel_name='pvc-contribution-extra-line-product-select', values={
            'products': ('fancy-select', ['tbs pv product - desc'])
        })

        self.click_button(panel_name='pvc-contribution-extra-line-product-select', text='Add to declaration')
        self.wait_for_loading()

        #
        # contribution panel
        #

        self.assertTrue(self.get_panel_title('pvc-contribution-details-panel').find('New declaration for 2016') == 0)

        self.fill_form(panel_name='pvc-contribution-details-panel', values={
            'line_0_quantity': 100,
            'line_0_weight': '100',
            'line_1_quantity': 100,
            'line_1_weight': '100',
            'line_1_custom_unit': '100',
        })

        self.fill_form(panel_name='pvc-contribution-details-panel', values={
            'agreement': ('checkbox2', True)
        })
        self.wait_for_loading()

        self.click_button(panel_name='pvc-contribution-details-panel', text='Submit')
        self.wait_for_loading()

        self.assertEquals(self._get_text_from_selector('div[name="pvc-contribution-details-panel-checkout"] .total-amount'), u'€ 0.00')

        self.click_button(panel_name='pvc-contribution-details-panel-checkout', text='Submit')
        self.wait_for_loading(times=2)

    def test_guarantee_forecast_resubmit(self):

        """
        tests DE forecast guarantee
        """
        set_date(date(2015, 10, 1))

        self.sa_session.delete(self.service_level)

        service_level_pv_1 = self._create_tbs_service_level(self.tbs_account, name='SL PV 1',
            year=today().year, admin_fee=Decimal('300.00'), level=1, product_type=PvcTbsProduct.TYPE_PV_RELATED, guarantee_forecast_min_value=0)
        service_level_pv_2 = self._create_tbs_service_level(self.tbs_account, name='SL PV 2',
            year=today().year, admin_fee=Decimal('400.00'), level=2, product_type=PvcTbsProduct.TYPE_PV_RELATED, guarantee_forecast_min_value=100)

        service_level_other_1 = self._create_tbs_service_level(self.tbs_account, name='SL Other 1', year=today().year,
            admin_fee=Decimal('500.00'), level=1, guarantee_forecast_min_value=0)
        service_level_other_2 = self._create_tbs_service_level(self.tbs_account, name='SL Other 2', year=today().year,
            admin_fee=Decimal('600.00'), level=2, guarantee_forecast_min_value=100)

        # set TBS config for guarantee forecast.
        self.tbs.set_config('pvc.tbs.guarantee.forecast', True)

        member_account = self._create_member_account()

        tbs_member = self._member_apply_at_tbs(member_account, self.tbs_account, service_level=None)
        self.assertEqual(tbs_member.status, PvcTbsMember.STATUS_PENDING)

        tbs_pv_product = self.pc.save_or_update_tbs_product(None, 'my pv product', 'desc', self.tbs.id,
            product_type=PvcTbsProduct.TYPE_PV_RELATED, basis_for_invoicing=PvcTbsProduct.INVOICE_BASED_ON_WEIGHT,
            custom_unit=PvcTbsProduct.CUSTOM_UNIT_NOMINAL_POWER
        )
        tbs_batteries_product = self.pc.save_or_update_tbs_product(None, 'my batteries product', 'desc', self.tbs.id,
            product_type=PvcTbsProduct.TYPE_PV_RELATED, basis_for_invoicing=PvcTbsProduct.INVOICE_BASED_ON_WEIGHT
        )

        self.pc.save_contribution_fee(self.tbs_product.id, service_level_other_2.id, Decimal('0.025'), today(), PvcTbsProductFee.DIMENSION_NONE)
        self.pc.save_contribution_fee(tbs_pv_product.id, service_level_pv_2.id, Decimal('0.030'), today(), PvcTbsProductFee.DIMENSION_NONE)
        self.pc.save_contribution_fee(tbs_batteries_product.id, service_level_pv_2.id, Decimal('0.035'), today(), PvcTbsProductFee.DIMENSION_NONE)

        cc = PvcContributionOrderController(self.sa_session, account_id=member_account.id)
        order = cc.add_guarantee_forecast_order(tbs_member.id, today().year, [self.tbs_product.id, tbs_pv_product.id], is_b2b=None)

        self.assertEqual(order.total_amount_including_vat, Decimal('0'))
        self.assertEqual(order.is_guarantee_forecast, True)
        self.assertEqual(len(order.lines), 2)

        order = cc.update_contribution(order.id, {
            # other module
            'line_0_quantity': 1,
            'line_0_weight': Decimal('1.00'),

            # pv module
            'line_1_quantity': 1,
            'line_1_weight': Decimal('1.00'),
            'line_1_custom_unit': Decimal('1.00'),

        })

        self.assertEqual(order.total_amount_including_vat, Decimal('0'))
        self.assertEqual(order.is_guarantee_forecast, True)
        self.assertEqual(len(order.lines), 2)

        # submit should determine the new SL
        order = cc.submit_order(order.id)

        # expect two service levels to be active for this tbs_member
        self.assertEqual(tbs_member.active_service_level.service_level, service_level_other_1)
        self.assertEqual(tbs_member.active_service_level.pv_service_level, service_level_pv_1)
        # expect the start date of the service levels to be the 1st of October.
        self.assertEqual(tbs_member.active_service_level.start_date, date(2000, 1, 1))

        order.resubmit_request = PvcContributionOrder.RESUBMIT_REQUEST_APPROVED
        self.sa_session.add(order)
        self.sa_session.flush()

        self.login(member_account.name, 'test_password')
        self.wait_for_loading()


        self.click_iconbar('icon-pvc-contribution-manager', hover='icon-pvc-outstanding-contributions')
        self.wait_for_panel('pvc-contribution-manager')

        self.click_button(panel_name='pvc-contribution-manager', text='Resubmit')
        self.wait_for_loading() # for the details panel

        self.fill_form(panel_name='pvc-contribution-details-panel', values={
            'line_1_custom_unit': '500',
        })
        self._silly_honesty_check()

        self.click_button(panel_name='pvc-contribution-details-panel', text='Submit')

        # TODO: assert new SL

    def test_guarantee_forecast_change_products(self):
        """
        tests changing of products before a SL is active
        """

        set_date(date(2015, 1, 1))

        operator_account = self._create_operator_account()
        tbs_account = self._create_tbs_account(operator_account)
        tbs = tbs_account.default_group.license

        self._create_contribution_tax_rates(operator_account)

        service_level_pv_1 = self._create_tbs_service_level(tbs_account, name='SL PV 1',
            year=today().year, admin_fee=Decimal('300.00'), level=1, product_type=PvcTbsProduct.TYPE_PV_RELATED, guarantee_forecast_min_value=0)
        service_level_pv_2 = self._create_tbs_service_level(tbs_account, name='SL PV 2',
            year=today().year, admin_fee=Decimal('300.00'), level=2, product_type=PvcTbsProduct.TYPE_PV_RELATED, guarantee_forecast_min_value=100)

        service_level_other_1 = self._create_tbs_service_level(tbs_account, name='SL Other 1', year=today().year,
            admin_fee=Decimal('300.00'), level=1, guarantee_forecast_min_value=0)
        service_level_other_2 = self._create_tbs_service_level(tbs_account, name='SL Other 2', year=today().year,
            admin_fee=Decimal('300.00'), level=2, guarantee_forecast_min_value=100)

        # set TBS config for advance invoicing.
        tbs.set_config('pvc.tbs.guarantee.forecast', True)

        # add tbs products
        pc = PvcProductController(self.sa_session, account_id=operator_account.id)

        products = pc.get_available_products_for_tbs(tbs.id)
        tbs_product = pc.save_or_update_tbs_product(None, 'EEE-TBS-TEST', 'WEEE', tbs.id,
            parent_level_id=products[0]['id'], basis_for_invoicing=PvcTbsProduct.INVOICE_BASED_ON_WEIGHT
        )
        tbs_pv_product = pc.save_or_update_tbs_product(None, 'tbs pv product', 'desc', tbs.id,
            parent_level_id=products[0]['id'], product_type=PvcTbsProduct.TYPE_PV_RELATED, basis_for_invoicing=PvcTbsProduct.INVOICE_BASED_ON_WEIGHT,
            custom_unit=PvcTbsProduct.CUSTOM_UNIT_NOMINAL_POWER
        )

        pc.save_contribution_fee(tbs_product.id, service_level_other_2.id, Decimal('0.025'), today(), PvcTbsProductFee.DIMENSION_NONE)
        pc.save_contribution_fee(tbs_pv_product.id, service_level_pv_2.id, Decimal('0.030'), today(), PvcTbsProductFee.DIMENSION_NONE)

        member_account = self._create_member_account()
        tbs_member = self._member_apply_at_tbs(member_account, tbs_account, service_level=None)
        self.assertEqual(tbs_member.status, PvcTbsMember.STATUS_PENDING)

        cc = PvcContributionOrderController(self.sa_session, account_id=member_account.id)
        order = cc.add_guarantee_forecast_order(tbs_member.id, today().year, [tbs_pv_product.id], is_b2b=None)

        self.assertEqual(order.total_amount_including_vat, Decimal('0'))
        self.assertEqual(order.is_guarantee_forecast, True)

        order = cc.update_contribution(order.id, {
            # pv module
            'line_0_quantity': 1,
            'line_0_weight': Decimal('200.00'),
            'line_0_custom_unit': Decimal('250.00'),
        }, additional_lines={
        })

        # login as member
        self.login(member_account.name, 'test_password')

        self.click_iconbar('icon-pvc-outstanding-contributions')
        self.wait_for_loading()

        self.click_button(panel_name='pvc-outstanding-contributions', text='Redeclare')
        self.wait_for_loading()

        self._silly_honesty_check()

        self.click_button(panel_name='pvc-contribution-details-panel', text='Remove line')

        self.click_button(panel_name='dialog-question', text='Yes')
        self.wait_for_loading(stable_time=4) # for the details panel

        self.click_button(panel_name='pvc-contribution-details-panel', text='Add line')
        self.wait_for_loading()

        self.fill_form(panel_name='pvc-contribution-extra-line-product-select', values={
            'products': ('fancy-select', [
                'EEE-TBS-TEST - WEEE',
                'tbs pv product - desc'
            ])
        })

        self.click_button(panel_name='pvc-contribution-extra-line-product-select', text='Add to declaration')
        self.wait_for_loading()

        self.fill_form(panel_name='pvc-contribution-details-panel', values={
            'line_extra_1_quantity': 100,
            'line_extra_1_weight': '100',
            'line_extra_2_quantity': 200,
            'line_extra_2_weight': '200',
            'line_extra_2_custom_unit': '200',
        })

        self._silly_honesty_check()
        self.wait_for_loading()

        self.click_button(panel_name='pvc-contribution-details-panel', text='Submit')
        self.wait_for_loading()

        self.click_button(panel_name='pvc-contribution-details-panel-checkout', text='Submit')
        self.wait_for_loading()

        self.assertEqual(tbs_member.active_service_level.service_level, service_level_other_2)
        self.assertEqual(tbs_member.active_service_level.pv_service_level, service_level_pv_2)

    def test_guarantee_forecast_brand_remove_line(self):
        """
        """
        set_date(date(2015, 11, 1) )

        ma, tm = self._create_simple_member_for_declaration()
        tm.tbs.set_config('pvc.tbs.guarantee.forecast', True)
        tm.tbs.set_config('pvc.tbs.contribution.split.by.brand', True)

        tbs_account = self.tbs_account
        operator_account = self.operator_account
        tbs = self.tbs

        service_level_pv_1 = self._create_tbs_service_level(tbs_account, name='SL PV 1',
            year=today().year, admin_fee=Decimal('300.00'), level=1, product_type=PvcTbsProduct.TYPE_PV_RELATED, guarantee_forecast_min_value=0)
        service_level_pv_2 = self._create_tbs_service_level(tbs_account, name='SL PV 2',
            year=today().year, admin_fee=Decimal('300.00'), level=2, product_type=PvcTbsProduct.TYPE_PV_RELATED, guarantee_forecast_min_value=100)

        # add tbs products
        pc = PvcProductController(self.sa_session, account_id=operator_account.id)
        tbs_pv_product = pc.save_or_update_tbs_product(None, 'tbs pv product', 'desc', tbs.id,
            basis_for_invoicing=PvcTbsProduct.INVOICE_BASED_ON_WEIGHT,
            custom_unit=PvcTbsProduct.CUSTOM_UNIT_NOMINAL_POWER
        )
        self.pc.save_contribution_fee(tbs_pv_product.id, self.service_level.id, Decimal('0.075'), date(2015, 1, 1))
        self.pc.save_contribution_fee(tbs_pv_product.id, service_level_pv_1.id, Decimal('0.055'), date(2015, 1, 1))
        self.pc.save_contribution_fee(tbs_pv_product.id, service_level_pv_2.id, Decimal('0.025'), date(2015, 1, 1))

        product = self._get_member_product(tm)

        self.login(ma.name, 'test_password')

        self.wait_for_loading()

        self.click_button(panel_name='dialog-information', text='OK')

        # also select the pv product
        self.fill_form(panel_name='pvc-contribution-extra-line-product-select', values={
            'products': ('fancy-select', ['tbs pv product - desc']),
            'brand': 'My brand'
        })


        self.click_button(panel_name='pvc-contribution-extra-line-product-select', text='Add to declaration')
        self.wait_for_loading()

        #
        # contribution panel
        #

        self.assertTrue(self.get_panel_title('pvc-contribution-details-panel').find('New declaration for 2016') == 0)

        self.fill_form(panel_name='pvc-contribution-details-panel', values={
            'line_0_quantity': 100,
            'line_0_weight': '100',
            'line_1_quantity': 100,
            'line_1_weight': '100',
            'line_1_custom_unit': '100',
        })

        self.click_button(panel_name='pvc-contribution-details-panel', text='Add line')
        self.wait_for_loading()

        self.fill_form(panel_name='pvc-contribution-extra-line-product-select', values={
            'products': ('fancy-select', [
                'EEE-TBS-TEST - WEEE',
            ])
        })

        self.click_button(panel_name='pvc-contribution-extra-line-product-select', text='Add to declaration')
        self.wait_for_loading()


        self.fill_form(panel_name='pvc-contribution-details-panel', values={
            'line_extra_3_brand': 'My brand',
            'line_extra_3_quantity': '100',
            'line_extra_3_weight': '100',
        })

        self._silly_honesty_check()

        # remove extra line (third button), that goes without confirmation
        button = self._driver.find_elements_by_css_selector("div[name='pvc-contribution-details-panel'] #remove_line_extra_3 button")[0]
        button.click()

        # remove line 0
        self.click_button(panel_name='pvc-contribution-details-panel', text='Remove line')
        self.click_button(panel_name='dialog-question', text='Yes')
        self.wait_for_loading()

        # submit should go ok
        self._silly_honesty_check()
        self.click_button(panel_name='pvc-contribution-details-panel', text='Submit')
        self.wait_for_loading()


    def test_resubmit_with_brands_mix_of_custom_units(self):
        # this is a test case trying to reproduce the issue in #933

        payment_methods = self._setup_payment_methods(self.operator_account, self.tbs_account)
        payment_method = payment_methods[0]
        products = self.pc.get_available_products_for_tbs(self.tbs.id)

        tbs_product_quantity = self.pc.save_or_update_tbs_product(None, 'EEE-TBS-TEST-9', 'WEEE-3', self.tbs.id,
            parent_level_id=products[1]['id'], basis_for_invoicing=PvcTbsProduct.INVOICE_BASED_ON_QUANTITY
        )

        tbs_product_custom_unit = self.pc.save_or_update_tbs_product(None, 'product power', 'WEEE-3', self.tbs.id,
            parent_level_id=products[1]['id'], basis_for_invoicing=PvcTbsProduct.INVOICE_BASED_ON_CUSTOM_UNIT, custom_unit=PvcTbsProduct.CUSTOM_UNIT_NOMINAL_POWER
        )

        tbs_product_custom_unit2 = self.pc.save_or_update_tbs_product(None, 'product surface', 'WEEE-3', self.tbs.id,
            parent_level_id=products[1]['id'], basis_for_invoicing=PvcTbsProduct.INVOICE_BASED_ON_CUSTOM_UNIT, custom_unit=PvcTbsProduct.CUSTOM_UNIT_SURFACE
        )

        self.pc.save_contribution_fee(tbs_product_quantity.id, self.service_level.id, Decimal('0.016'), date(2013, 1, 1))
        self.pc.save_contribution_fee(tbs_product_custom_unit.id, self.service_level.id, Decimal('0.017'), date(2013, 1, 1))
        self.pc.save_contribution_fee(tbs_product_custom_unit2.id, self.service_level.id, Decimal('0.018'), date(2013, 1, 1))

        ma, tm = self._create_simple_member_for_declaration()

        mp = self._get_member_product(tm, self.tbs_product2)
        brand1 = self._get_member_brand('Samsung', mp)
        brand2 = self._get_member_brand('Sony', mp)

        mp1 = self._get_member_product(tm, tbs_product_custom_unit)
        brand3 = self._get_member_brand('Samsung', mp1)
        brand4 = self._get_member_brand('Toshiba', mp1)

        mp2 = self._get_member_product(tm, tbs_product_quantity)
        brand5 = self._get_member_brand('Samsung', mp2)

        mp3 = self._get_member_product(tm, tbs_product_custom_unit2)
        brand7 = self._get_member_brand('Philips', mp3)

        # enable tbs b2b setting
        tm.tbs.set_config('pvc.tbs.contribution.split.by.brand', True)

        set_date(date(2013, 7, 1))

        cc = PvcContributionOrderController(self.sa_session)
        d = cc.trigger_new_contributions(self.tbs)
        d = d[0]

        self.login(ma.name, 'test_password')

        self.click_iconbar('icon-pvc-outstanding-contributions')
        self.wait_for_loading() # for the custom query

        self.click_button(panel_name='pvc-outstanding-contributions', text='Declare')
        self.wait_for_loading()

        self.fill_form(panel_name='pvc-contribution-details-panel', values={
            'line_0_quantity': 100,
            'line_0_weight': '100',
            'line_1_quantity': 101,
            'line_1_weight': '101',
            'line_2_quantity': 102,
            'line_2_weight': '102',
            'line_3_quantity': 103,
            'line_3_weight': '103',
            'line_3_custom_unit': '103',
            'line_4_quantity': 104,
            'line_4_weight': '104',
            'line_4_custom_unit': '104',
            'line_5_quantity': 105,
            'line_5_weight': '105',
            'line_5_custom_unit': '105',
        })

        self.click_button_by_id('line_5_duplicate-button')
        self.wait_for_loading()
        self.fill_form(panel_name='pvc-contribution-details-panel', values={
            'line_extra_7_quantity': 106,
            'line_extra_7_weight': '106',
            'line_extra_7_brand': 'Zombie',
            'line_extra_7_custom_unit': '106'
        })

        self.click_button_by_id('line_3_duplicate-button')
        self.wait_for_loading()
        self.fill_form(panel_name='pvc-contribution-details-panel', values={
            'line_extra_8_quantity': 107,
            'line_extra_8_weight': '107',
            'line_extra_8_brand': 'Aap',
            'line_extra_8_custom_unit': '107'
        })

        self.click_button_by_id('line_1_duplicate-button')
        self.wait_for_loading()
        self.fill_form(panel_name='pvc-contribution-details-panel', values={
            'line_extra_9_quantity': 108,
            'line_extra_9_weight': '108',
            'line_extra_9_brand': 'Aap',
        })

        self.click_button_by_id('line_1_duplicate-button')
        self.wait_for_loading()
        self.fill_form(panel_name='pvc-contribution-details-panel', values={
            'line_extra_10_quantity': 109,
            'line_extra_10_weight': '109',
            'line_extra_10_brand': 'Beer',
        })

        self.click_button_by_id('line_1_duplicate-button')
        self.wait_for_loading()
        self.fill_form(panel_name='pvc-contribution-details-panel', values={
            'line_extra_11_quantity': 110,
            'line_extra_11_weight': '110',
            'line_extra_11_brand': 'Aaaap',
        })

        self.click_button_by_id('line_5_duplicate-button')
        self.wait_for_loading()
        self.fill_form(panel_name='pvc-contribution-details-panel', values={
            'line_extra_12_quantity': 111,
            'line_extra_12_weight': '111',
            'line_extra_12_brand': 'Aaaar',
            'line_extra_12_custom_unit': '111'
        })

        self.click_button_by_id('line_1_duplicate-button')
        self.wait_for_loading()
        self.fill_form(panel_name='pvc-contribution-details-panel', values={
            'line_extra_13_quantity': 112,
            'line_extra_13_weight': '112',
            'line_extra_13_brand': 'CCCCC',
        })

        self._silly_honesty_check()

        self.click_button(panel_name='pvc-contribution-details-panel', text='Submit')
        self.wait_for_loading(times=25)

        self.fill_form('pvc-contribution-details-panel-checkout', values={
            'payment_method': ('radio', payment_method.id)
        })
        self.wait_for_loading()

        self.click_button(panel_name='pvc-contribution-details-panel-checkout', text='Submit')
        self.wait_for_loading(times=4)

        self.click_button('order-payment_method-panel', text='Mark as submitted')
        self.wait_for_loading()

        self.close_panel(panel_name='pvc-outstanding-contributions')

        d.resubmit_request = PvcContributionOrder.RESUBMIT_REQUEST_APPROVED
        self.sa_session.add(d)
        self.sa_session.flush()

        self.click_iconbar('icon-pvc-contribution-manager', hover='icon-pvc-outstanding-contributions')
        self.wait_for_panel('pvc-contribution-manager')

        self.click_button(panel_name='pvc-contribution-manager', text='Resubmit')
        self.wait_for_loading() # for the details panel

        self.fill_form(panel_name='pvc-contribution-details-panel', values={
            'line_0_quantity': 700,
            'line_0_weight': '700',
            'line_1_quantity': 701,
            'line_1_weight': '701',
        })

        self._silly_honesty_check()

        self.click_button(panel_name='pvc-contribution-details-panel', text='Save')