# Copyright (c) 2015 BaseSoft B.V.
# See LICENSE for details.
from datetime import date
from decimal import Decimal
from twisted.trial.unittest import SkipTest

from basesoft.util.virtual_date import set_date, today
from basesoft.web.core.util.controllers import LicenseConfigController
from pvcycle.core.controllers import PvcMemberController
from pvcycle.core.models import PvcTbsMember
from pvcycle.order.controllers import PvcDeclarationOrderController
from pvcycle.pvmodule.controllers import PvcPVModuleController
from pvcycle.report.controllers import PvcHistoricalDataController
from pvcycle.test.selenium.pvc_selenium import PvcSeleniumTestCase
from selenium.common.exceptions import ElementNotVisibleException


class TestCase(PvcSeleniumTestCase):
    def setUp(self):
        super(TestCase, self).setUp()
        self.connect_db()

        self.operator_account = self._create_operator_account()

        set_date(date(2013, 1, 1)) # set date to January 2013

        self._create_contribution_tax_rates(self.operator_account)

        set_date(date(2013, 6, 1)) # set date to June 2013

        self._modules = self._fill_pv_modules(self.operator_account)

        self.tbs_account = self._create_tbs_account(self.operator_account)
        self.accountant_account = self._create_accountant_account(self.tbs_account.default_group.license, self.operator_account)

        lcc = LicenseConfigController(self.sa_session, scope=self.tbs_account.default_group.license)
        lcc.set('pvc.tbs.declaration.use.weight.for.amount.calculations', False)

        self.payment_methods = self._setup_payment_methods(self.operator_account, self.tbs_account)

        self._administration_fee = Decimal('250.00')
        self._contribution_fee = Decimal('2.75')

        self.service_level = self._create_tbs_service_level(self.tbs_account, admin_fee=self._administration_fee, contrib_fee=self._contribution_fee, year=today().year)
        self.reset_roe()

    def _create_simple_member_for_declaration(self, modules=None, countries=None, tbs_account=None, service_level=None, ma=None, tm=None):

        if ma is None:
            ma = self._create_member_account()

        # apply at TBS
        if service_level is None:
            service_level = self.service_level
        if tbs_account is None:
            tbs_account = self.tbs_account

        if tm is None:
            tm = self._member_apply_at_tbs(ma, tbs_account, service_level)

        # activate member to automatically create an order entry for the Administration Fee
        mc = PvcMemberController(self.sa_session, account_id=tbs_account.id)
        mc.save_tbs_member(tm.id, {'status': PvcTbsMember.STATUS_APPROVED, 'is_full_member': False, 'is_representative': False})

        ma.default_group.license.is_legacy_member = True
        self.sa_session.add(ma.default_group.license)
        self.sa_session.flush()

        if modules is None:
            modules = [self._modules[0], self._modules[1]]

        # now that the TBSMember is approved we should be able to save pv modules for this tbs member
        c = PvcPVModuleController(self.sa_session, account_id=ma.id)
        for module in modules:
            if countries is None:
                c.save_or_update_module_for_tbs_member(tm, tm.tbs.country, module, is_active=True)
                continue
            for country in countries:
                c.save_or_update_module_for_tbs_member(tm, country, module, is_active=True)

        return ma, tm

    def _january_declaration(self, ma=None, tm=None):

        if ma is None or tm is None:
            ma, tm = self._create_simple_member_for_declaration()

        dc = PvcDeclarationOrderController(self.sa_session, account_id=ma.id)

        # add an open declaration for this member.
        d = dc.add_open_declaration(tm, 2013, month=1)
        d = dc.update_declaration(d.id, {'line_0': 200, 'line_1': 150})
        # submit the declaration (for export)
        return dc.submit_order(d.id)

    def test_operator_read_only_declaration(self):

        self.login(self.operator_account.name, 'test_password')

        # add some declaration data.
        ma, tm = self._create_simple_member_for_declaration()
        self._january_declaration(ma, tm)

        # open declaration manager.
        self.click_iconbar('icon-pvc-declaration-manager', hover='icon-pvc-contribution-manager')
        self.wait_for_loading()  # for the custom query

        # verify that details are still viewable.
        self.filter_cq(panel_name='pvc-declaration-manager', filters=[
            ('Year', 'Equal', ('select', '2013')),
            ('Month', 'Equal', ('select', '1')),
        ])
        self.wait_for_loading()  # for the custom query
        self.click_cq('pvc-declaration-manager', row_index=0, col_name='details')
        self.wait_for_loading()

        # close the declaration details.
        self.click_button(panel_name='pvc-declaration-details-panel', text='Close')
        self.close_panel(panel_name='pvc-declaration-manager')

        # open order manager.
        self.click_iconbar('icon-pvc-order-manager', hover='icon-pvc-contribution-manager')
        self.wait_for_loading()  # for the custom query

    def test_operator_read_only_pv_module(self):
        ma, tm = self._create_simple_member_for_declaration()
        self.login(self.operator_account.name, 'test_password')

        self.click_iconbar('icon-pvc-pv-module-manager')
        self.wait_for_loading()

        self.assertRaises(ElementNotVisibleException, self.click_menu, panel_name='pvc-pv-module-manager', text='Create PV Module')

        self.click_cq(panel_name='pvc-pv-module-manager', row_index=0, col_name='details')
        self.wait_for_loading()

        self.click_button(panel_name='pvc-pv-module-details', text='Close')

    def test_operator_read_only_historical_data(self):
        ma, tm = self._create_simple_member_for_declaration()
        controller = PvcHistoricalDataController(self.sa_session, account_id=ma.id)
        hd = controller.save_or_update_historical_data(None, tm.id, 2012, 155, Decimal('100.1'), 200, Decimal('300.3'), Decimal('23.45'), Decimal('76.55'))

        self.login(self.operator_account.name, 'test_password')

        self.click_iconbar('icon-pvc-historical-data-manager')
        self.wait_for_loading()

        self.assertRaises(ElementNotVisibleException, self.click_menu, panel_name='pvc-historical-data-manager', text='Delete historical data')

        self.click_cq(panel_name='pvc-historical-data-manager', row_index=0, col_name='details')
        self.wait_for_loading()

        self.click_button(panel_name='pvc-historical-data-details', text='Close')

    def test_member_read_only_declaration(self):
        # add some declaration data.
        ma, tm = self._create_simple_member_for_declaration()
        self._january_declaration(ma, tm)

        self.login(ma.name, 'test_password')

        # open declaration manager.
        self.click_iconbar('icon-pvc-declaration-manager', hover='icon-pvc-outstanding-contributions')
        self.wait_for_loading()  # for the custom query

        # verify that details are still viewable.
        self.filter_cq(panel_name='pvc-declaration-manager', filters=[
            ('Year', 'Equal', ('select', '2013')),
            ('Month', 'Equal', ('select', '1')),
        ])
        self.wait_for_loading()  # for the custom query
        self.click_cq('pvc-declaration-manager', row_index=0, col_name='details')
        self.wait_for_loading()

        self.assert_cq(panel_name='pvc-declaration-manager', row_index=0, data={
            'action': u'\ue861'
        })

        # close the declaration details.
        self.click_button(panel_name='pvc-declaration-details-panel', text='Close')
        self.close_panel(panel_name='pvc-declaration-manager')

        # open order manager.
        self.click_iconbar('icon-pvc-order-manager', hover='icon-pvc-outstanding-contributions')
        self.wait_for_loading()  # for the custom query

        self.filter_cq(panel_name='pvc-order-manager', filters=[
            ('Order type', 'Equal', ('select', 'Contribution (legacy)')),
        ])
        self.wait_for_loading()

        self.assert_cq(panel_name='pvc-order-manager', row_index=0, data={
            'payment_checkout': u'\ue861'
        })

    def test_member_read_only_pv_module(self):
        raise SkipTest('legacy test')

        ma, tm = self._create_simple_member_for_declaration()

        self.login(ma.name, 'test_password')

        self.click_iconbar('icon-pvc-pv-module-country-manager')
        self.wait_for_loading()

        self.assertRaises(ElementNotVisibleException, self.click_menu, panel_name='pvc-pv-module-country-manager', text='Add PV Module')

    def test_member_read_only_historical_data(self):
        ma, tm = self._create_simple_member_for_declaration()
        controller = PvcHistoricalDataController(self.sa_session, account_id=ma.id)
        hd = controller.save_or_update_historical_data(None, tm.id, 2012, 155, Decimal('100.1'), 200, Decimal('300.3'), Decimal('23.45'), Decimal('76.55'))

        self.login(ma.name, 'test_password')

        self.click_iconbar('icon-pvc-historical-data-manager')
        self.wait_for_loading()

        self.assertRaises(ElementNotVisibleException, self.click_menu, panel_name='pvc-historical-data-manager', text='Add historical data')

        self.click_cq(panel_name='pvc-historical-data-manager', row_index=0, col_name='details')
        self.wait_for_loading()

        self.click_button(panel_name='pvc-historical-data-details', text='Close')

