# Copyright (c) 2015 BaseSoft B.V.
# See LICENSE for details.
from datetime import date
from decimal import Decimal

from basesoft.util.virtual_date import set_date, today, now
from basesoft.web.contrib.i18n.models import Country
from basesoft.web.contrib.notification.controllers import NotificationController
from basesoft.web.contrib.notification.models import Notification
from basesoft.web.contrib.order.controllers import TaxController
from basesoft.web.contrib.order.models import OrderRule
from basesoft.web.core.exceptions import PublicUserException, PublicSecurityException

from pvcycle.core.controllers import PvcMemberController, PvcTBSController
from pvcycle.core.exceptions import PvcServiceLevelException, PvcServiceLevelDowngradeException, \
    PvcServiceLevelNoChangeException
from pvcycle.core.models import PvcTbsMember, PvcTbsLicense, PvcTbsMemberServiceLevel
from pvcycle.order.controllers import PvcContributionOrderController, PvcOrderController
from pvcycle.order.models import PvcContributionOrder, PvcContributionOrderLine, PvcOrder
from pvcycle.product.controllers import PvcProductController
from pvcycle.product.models import PvcTbsProduct, PvcTbsMemberProduct, PvcTbsProductFee
from pvcycle.product.schedule import MergeBrandTask
from pvcycle.subrogation.controllers import PvcSubrogationController
from pvcycle.subrogation.models import PvcSubrogation
from pvcycle.test.unit import pvc_unittest

from sqlalchemy.sql.expression import and_
from sqlalchemy.orm.exc import ObjectDeletedError


class TestCase(pvc_unittest.PvcTestCase):
    """
    pvcycle.test.unit.test_contribution.TestCase
    """

    def setUp(self):
        self.connect_db()

        self.operator_account = self._create_operator_account()

        set_date(date(2013, 1, 1)) # set date to January 2013
        self.reset_roe()

        self._create_contribution_tax_rates(self.operator_account)

        state = self._get_tbs_license_state()
        state['tbs']['country'] = self.sa_session.query(Country).filter(Country.iso_3166_a2=='BE').first().id

        self.tbs_account = self._create_tbs_account(self.operator_account, state=state)
        self.tbs = self.tbs_account.default_group.license
        self.accountant = self._create_accountant_account(self.tbs_account.default_group.license, self.operator_account)

        self._administration_fee = Decimal('250.00')
        self._contribution_fee = Decimal('2.75')

        self.service_level = self._create_tbs_service_level(self.tbs_account, admin_fee=self._administration_fee, year=today().year)

        self.pc = PvcProductController(self.sa_session, account_id=self.operator_account.id)

        products = self.pc.get_available_products_for_tbs(self.tbs.id)
        self.tbs_product = self.pc.save_or_update_tbs_product(None, 'EEE-TBS-TEST', 'WEEE', self.tbs.id,
            parent_level_id=products[0]['id'], basis_for_invoicing=PvcTbsProduct.INVOICE_BASED_ON_WEIGHT
        )

        self.tbs_product2 = self.pc.save_or_update_tbs_product(None, 'EEE-TBS-TEST-2', 'WEEE-2', self.tbs.id,
            parent_level_id=products[1]['id'], basis_for_invoicing=PvcTbsProduct.INVOICE_BASED_ON_WEIGHT
        )

        self.pc.save_contribution_fee(self.tbs_product.id, self.service_level.id, Decimal('0.015'), today(), PvcTbsProductFee.DIMENSION_NONE)
        self.pc.save_contribution_fee(self.tbs_product2.id, self.service_level.id, Decimal('0.035'), today(), PvcTbsProductFee.DIMENSION_NONE)
        self.pc.save_contribution_fee(self.tbs_product.id, self.service_level.id, Decimal('0.015'), today(), PvcTbsProductFee.DIMENSION_IS_B2C)
        self.pc.save_contribution_fee(self.tbs_product2.id, self.service_level.id, Decimal('0.035'), today(), PvcTbsProductFee.DIMENSION_IS_B2C)
        self.pc.save_contribution_fee(self.tbs_product.id, self.service_level.id, Decimal('0.015'), today(), PvcTbsProductFee.DIMENSION_IS_B2B)
        self.pc.save_contribution_fee(self.tbs_product2.id, self.service_level.id, Decimal('0.035'), today(), PvcTbsProductFee.DIMENSION_IS_B2B)

        set_date(date(2013, 6, 1)) # set date to June 2013

        self._trash_all_rules()

    def _create_simple_member_for_declaration(self, obligation_start_date=None):
        ma = self._create_member_account()

        # apply at TBS
        tm = self._member_apply_at_tbs(ma, self.tbs_account, self.service_level)

        obligation_start_date = date(2012, 1, 1) if obligation_start_date is None else obligation_start_date

        # activate member to automatically create an order entry for the Administration Fee
        mc = PvcMemberController(self.sa_session, account_id=self.tbs_account.id)
        mc.save_tbs_member(tm.id, {
            'status': PvcTbsMember.STATUS_APPROVED,
            'is_full_member': False,
            'is_representative': False,
            'obligation_start_date': obligation_start_date})
        return ma, tm

    def _get_member_product(self, tm, tbs_product=None, country=None):

        if tbs_product is None:
            tbs_product = self.tbs_product

        # TODO this should be added with a controller method
        product = PvcTbsMemberProduct()
        product.tbs_product = tbs_product
        product.tbs_member = tm
        product.is_active = True
        product.country = tm.tbs.country if country is None else country
        self.sa_session.add(product)
        self.sa_session.flush()
        return product

    def test_obligation_start_date_failure(self):
        # create a member for declaration
        ma, tm = self._create_simple_member_for_declaration(obligation_start_date=date(2014, 1, 1))

        # set TBS config for advance invoicing.
        self.tbs.set_config('pvc.tbs.advance.invoicing', True)
        self.tbs.set_config('pvc.tbs.declaration.period', PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY)
        cc = PvcContributionOrderController(self.sa_session, account_id=ma.id)

        products = [self.tbs_product.id]

        self.assertRaises(PublicUserException, cc.add_forecast_order, tm.id, products)

    def test_available_forecasts(self):
        # create a member for declaration
        ma, tm = self._create_simple_member_for_declaration()

        # set TBS config for advance invoicing.
        self.tbs.set_config('pvc.tbs.advance.invoicing', True)
        # set TBS config to quarterly declarations, this means the first available period is Q2
        self.tbs.set_config('pvc.tbs.declaration.period', PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY)

        # assert that the forecast is required.
        cc = PvcContributionOrderController(self.sa_session, account_id=ma.id)
        tms = cc.get_outstanding_forecast_requirement()
        self.assertEqual(len(tms), 1)

        y, m, q = PvcContributionOrderController.get_first_available_period_for_forecast(tm)
        self.assertEqual(y, today().year)
        self.assertEqual(m, None)
        self.assertEqual(q, 2)

    def test_supply_forecast(self):
        # create a member for declaration
        ma, tm = self._create_simple_member_for_declaration()

        # set TBS config for advance invoicing.
        self.tbs.set_config('pvc.tbs.advance.invoicing', True)
        self.tbs.set_config('pvc.tbs.declaration.period', PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY)
        cc = PvcContributionOrderController(self.sa_session, account_id=ma.id)

        products = [self.tbs_product.id]

        order = cc.add_forecast_order(tm.id, products)
        line = order.lines[0]
        self.assertEqual(line.tbs_member_product.tbs_product, self.tbs_product)

        self.assertEqual(line.quantity, 0)
        self.assertEqual(line.weight, Decimal('0.0'))
        self.assertEqual(line.custom_unit, None)

        self.assertEqual(line.line_amount_excluding_vat, Decimal('0.0'))
        self.assertEqual(line.gl_account_number, 'GLF0R3CAS7')

        values = {
            'line_0_quantity': 210,
            'line_0_weight': Decimal('16.44')
        }
        extra = {
            'tbs_product_tpid_%d' % self.tbs_product2.id: self.tbs_product2.id,
            'line_tpid_%d_weight' % self.tbs_product2.id: Decimal('13.6'),
            'line_tpid_%d_quantity' % self.tbs_product2.id: 100,
            'line_tpid_%d_country_id' % self.tbs_product2.id: None,
        }
        order = cc.update_contribution(order.id, values, extra)
        self.assertEqual(order.lines[0].gl_account_number, 'GLF0R3CAS7')
        self.assertEqual(order.lines[1].gl_account_number, 'GLF0R3CAS7')

    def test_update_contribution(self):
        # create a member for declaration
        ma, tm = self._create_simple_member_for_declaration()

        product = self._get_member_product(tm)

        # FIXME this is using a protected method on the ContributionController
        cc = PvcContributionOrderController(self.sa_session, account_id=ma.id)
        lines = [{
            'tbs_member_product': product,
            'quantity': 200,
            'weight': Decimal('15.3'),
            'custom_unit': None,
        }]

        order = cc._add_contribution_order(tm, lines, 2013, month=3)
        # we have an order now.

        values = {'line_0_quantity': 210}
        # this will fail, cuz weight is missing.
        self.assertRaises(PublicUserException, cc.update_contribution, order.id, values)

        values['line_0_weight'] = Decimal('16.32')
        order = cc.update_contribution(order.id, values)
        self.assertEqual(order.status, PvcContributionOrder.STATUS_PENDING)

        order = cc.get(order.id).pop()
        l = order.lines[0]

        self.assertEqual(l.quantity, 210)
        self.assertEqual(l.weight, Decimal('16.32'))

    def test_update_contribution_after_submit(self):
        # create a member for declaration
        ma, tm = self._create_simple_member_for_declaration()

        product = self._get_member_product(tm)
        cc = PvcContributionOrderController(self.sa_session, account_id=ma.id)
        lines = [{
            'tbs_member_product': product,
            'quantity': 200,
            'weight': Decimal('15.3'),
            'custom_unit': None,
        }]

        order = cc._add_contribution_order(tm, lines, 2013, month=3)
        cc.submit_order(order.id)

        values = {'line_0_quantity': 210, 'line_0_weight': Decimal('16.44')}
        # this will fail, because no resubmit has been approved.
        self.assertRaises(PublicUserException, cc.update_contribution, order.id, values)

    def test_update_contribution_additional_lines(self):
        # create a member for declaration
        ma, tm = self._create_simple_member_for_declaration()

        product = self._get_member_product(tm)

        # FIXME this is using a protected method on the ContributionController
        cc = PvcContributionOrderController(self.sa_session, account_id=ma.id)
        lines = [{
            'tbs_member_product': product,
            'quantity': 200,
            'weight': Decimal('15.3'),
            'custom_unit': None,
        }]

        order = cc._add_contribution_order(tm, lines, 2013, month=3)
        # we have an order now.

        values = {'line_0_quantity': 210, 'line_0_weight': Decimal('20.5')}

        # we try to add an extra line with the same product.
        extra = {
            'tbs_product_tpid_%d' % self.tbs_product.id: self.tbs_product.id,
            'line_tpid_%d_weight' % self.tbs_product.id: Decimal('13.6'),
            'line_tpid_%d_quantity' % self.tbs_product.id: 100,
            'line_tpid_%d_country_id' % self.tbs_product.id: None,
        }
        self.assertRaises(PublicUserException, cc.update_contribution, order.id, values, additional_lines=extra)

        extra = {
            'tbs_product_tpid_%d' % self.tbs_product2.id: self.tbs_product2.id,
            'line_tpid_%d_weight' % self.tbs_product2.id: Decimal('13.6'),
            'line_tpid_%d_quantity' % self.tbs_product2.id: 100,
            'line_tpid_%d_country_id' % self.tbs_product2.id: None,
        }

        order = cc.update_contribution(order.id, values, additional_lines=extra)
        self.assertEqual(len(order.lines), 2)
        self.assertEqual(order.total_amount_excluding_vat, Decimal('0.78350'))

    def test_update_contribution_mixed_zeros_and_values(self):
        # create a member for declaration
        ma, tm = self._create_simple_member_for_declaration()

        product = self._get_member_product(tm)

        # FIXME this is using a protected method on the ContributionController
        cc = PvcContributionOrderController(self.sa_session, account_id=ma.id)
        lines = [{
            'tbs_member_product': product,
            'quantity': 0,
            'weight': Decimal('15.3'),
            'custom_unit': None,
        }]

        self.assertRaises(PublicUserException, cc._add_contribution_order, tm, lines, 2013, month=3)

        lines[0]['quantity'] = 200
        order = cc._add_contribution_order(tm, lines, 2013, month=3)
        # we have an order now.

        values = {'line_0_quantity': 210, 'line_0_weight': Decimal('20.5')}

        # we try to add an extra line with zero weight.
        extra = {
            'tbs_product_tpid_%d' % self.tbs_product2.id: self.tbs_product2.id,
            'line_tpid_%d_weight' % self.tbs_product2.id: Decimal('0.0'),
            'line_tpid_%d_quantity' % self.tbs_product2.id: 100,
            'line_tpid_%d_country_id' % self.tbs_product2.id: None,
        }
        self.assertRaises(PublicUserException, cc.update_contribution, order.id, values, additional_lines=extra)

        extra = {
            'tbs_product_tpid_%d' % self.tbs_product2.id: self.tbs_product2.id,
            'line_tpid_%d_weight' % self.tbs_product2.id: Decimal('13.6'),
            'line_tpid_%d_quantity' % self.tbs_product2.id: 100,
            'line_tpid_%d_country_id' % self.tbs_product2.id: None,
        }

        order = cc.update_contribution(order.id, values, additional_lines=extra)
        self.assertEqual(len(order.lines), 2)
        self.assertEqual(order.total_amount_excluding_vat, Decimal('0.78350'))

    def test_remove_line_from_contribution(self):
        # create a member for declaration
        ma, tm = self._create_simple_member_for_declaration()

        product = self._get_member_product(tm)
        product2 = self._get_member_product(tm, self.tbs_product2)

        # FIXME this is using a protected method on the ContributionController
        cc = PvcContributionOrderController(self.sa_session, account_id=ma.id)
        lines = [{
            'tbs_member_product': product,
            'quantity': 200,
            'weight': Decimal('15.3'),
            'custom_unit': None,
        }, {
            'tbs_member_product': product2,
            'quantity': 100,
            'weight': Decimal('9.5'),
            'custom_unit': None,
        }]

        order = cc._add_contribution_order(tm, lines, 2013, month=3)

        self.assertTrue(len(order.lines), 2)

        order = cc.remove_line(order.id, order.lines[0].id, True)
        self.assertTrue(len(order.lines), 1)

        line = order.lines[0]
        self.assertEqual(line.quantity, 100)
        self.assertEqual(product.is_active, False)

    def test_update_contribution_with_subrogation(self):
        pass

    def test_export(self):
        ma, tm = self._create_simple_member_for_declaration()
        product = self._get_member_product(tm)

        cc = PvcContributionOrderController(self.sa_session, account_id=ma.id)
        lines = [{
            'tbs_member_product': product,
            'quantity': 200,
            'weight': Decimal('15.3'),
            'custom_unit': None,
        }]
        order = cc._add_contribution_order(tm, lines, 2013, month=3)

        # assert that the display_date is not set untill its submitted
        self.assertEquals(order.display_date, None)

        cc.submit_order(order.id)

        # now it must be set
        self.assertEquals(order.display_date.year, now().year)

        # assert its ready for the accountant
        adc = PvcContributionOrderController(self.sa_session, account_id=self.accountant.id)
        states = adc.get_unexported_orders_for_merchant(tm.tbs.id)

        self.assertEquals(len(states), 1)
        self.assertEquals(states[0]['id'], order.id)
        self.assertEquals(states[0]['polymorphic_type'], 'PvcContributionOrder')
        self.assertEquals(states[0]['total_amount_excluding_vat'], Decimal('0.22950'))

        self.assertEquals(len(states[0]['lines']), 1)
        self.assertEquals(states[0]['lines'][0]['polymorphic_type'], 'PvcContributionOrderLine')
        self.assertEquals(states[0]['lines'][0]['price_per_unit'], Decimal('0.015'))
        self.assertEquals(states[0]['lines'][0]['quantity'], 200)
        self.assertEquals(states[0]['lines'][0]['vat_rate'], Decimal('0'))
        self.assertEquals(states[0]['lines'][0]['gl_account'], 'GL789012')
        self.assertEquals(states[0]['lines'][0]['line_amount_excluding_vat'], Decimal('0.22950'))
        self.assertEquals(states[0]['lines'][0]['properties'][0]['key'], 'weight')
        self.assertEquals(states[0]['lines'][0]['properties'][0]['value_decimal'], Decimal('15.3'))
        self.assertEquals(states[0]['lines'][0]['properties'][1]['key'], 'basis_for_invoicing')
        self.assertEquals(states[0]['lines'][0]['properties'][1]['value_int'], PvcTbsProduct.INVOICE_BASED_ON_WEIGHT)

    def test_save_and_update_contribution_lines_is_b2b(self):
        # create a member for declaration
        ma, tm = self._create_simple_member_for_declaration()

        product = self._get_member_product(tm)

        # FIXME this is using a protected method on the ContributionController
        cc = PvcContributionOrderController(self.sa_session, account_id=ma.id)
        lines = [{
            'tbs_member_product': product,
            'quantity': 200,
            'weight': Decimal('15.3'),
            'custom_unit': None,
        }]

        # save order for TBS with B2B setting off
        order = cc._add_contribution_order(tm, lines, 2013, month=3)
        self.assertEqual(order.lines[0].is_b2b, None)

        # enable tbs b2b setting
        tm.tbs.set_config('pvc.tbs.require.end.user.distinction', True)

        # this should fail because TBS setting is on and is_b2b is None
        self.assertRaises(PublicUserException, cc._add_contribution_order, tm, lines, 2013, 4)

        # save properly
        lines[0]['is_b2b'] = False
        order = cc._add_contribution_order(tm, lines, 2013, month=4)
        self.assertEqual(order.lines[0].is_b2b, False)

        # update
        values = {'line_0_quantity': 210,
                  'line_0_weight': Decimal('16.32'),
                  'line_0_is_b2b': True}

        order = cc.update_contribution(order.id, values)
        self.assertEqual(order.status, PvcContributionOrder.STATUS_PENDING)

        order = cc.get(order.id).pop()
        l = order.lines[0]

        self.assertEqual(l.quantity, 210)
        self.assertEqual(l.weight, Decimal('16.32'))
        self.assertEqual(l.is_b2b, True)

        # add line for same product same is_b2b value = should raise exception
        extra = {
            'tbs_product_tpid_%d' % self.tbs_product.id: self.tbs_product.id,
            'line_tpid_%d_weight' % self.tbs_product.id: Decimal('13.6'),
            'line_tpid_%d_quantity' % self.tbs_product.id: 100,
            'line_tpid_%d_is_b2b' % self.tbs_product.id: True,
            'line_tpid_%d_country_id' % self.tbs_product.id: None,
        }

        self.assertRaises(PublicUserException, cc.update_contribution, order.id, values, extra)

        # add line for same product different is_b2b value = should pass
        extra = {
            'tbs_product_tpid_%d' % self.tbs_product.id: self.tbs_product.id,
            'line_tpid_%d_weight' % self.tbs_product.id: Decimal('13.6'),
            'line_tpid_%d_quantity' % self.tbs_product.id: 100,
            'line_tpid_%d_is_b2b' % self.tbs_product.id: False,
            'line_tpid_%d_country_id' % self.tbs_product.id: None,
        }

        order = cc.update_contribution(order.id, values, additional_lines=extra)
        self.assertEqual(order.lines[0].is_b2b, False)  # reindexing has happened
        self.assertEqual(order.lines[1].is_b2b, True)

        # add additional lines
        extra = {
            'tbs_product_tpid_%d' % self.tbs_product2.id: self.tbs_product2.id,
            'line_tpid_%d_weight' % self.tbs_product2.id: Decimal('13.6'),
            'line_tpid_%d_quantity' % self.tbs_product2.id: 100,
            'line_tpid_%d_is_b2b' % self.tbs_product2.id: False,
            'line_tpid_%d_country_id' % self.tbs_product2.id: None,
        }

        values = {
            'line_0_quantity': 110,
            'line_0_weight': Decimal('22.32'),
            'line_0_is_b2b': False,
            'line_1_quantity': 210,
            'line_1_weight': Decimal('16.32'),
            'line_1_is_b2b': True,
        }
        order = cc.update_contribution(order.id, values, additional_lines=extra)
        self.assertEqual(len(order.lines), 3)
        self.assertEqual(order.lines[2].is_b2b, False)

        # load declaration
        order = cc.load_state(order.id)
        self.assertEqual(order['lines'][0]['is_b2b'], False)
        self.assertEqual(order['lines'][1]['is_b2b'], True)
        self.assertEqual(order['lines'][2]['is_b2b'], False)

    def test_resubmit_contribution_lines_is_b2b(self):
        # create a member for declaration
        ma, tm = self._create_simple_member_for_declaration()

        product = self._get_member_product(tm)

        # FIXME this is using a protected method on the ContributionController
        cc = PvcContributionOrderController(self.sa_session, account_id=ma.id)
        lines = [{
            'tbs_member_product': product,
            'quantity': 200,
            'weight': Decimal('15.3'),
            'custom_unit': None,
            'is_b2b': True
        }]

        # enable tbs b2b setting
        tm.tbs.set_config('pvc.tbs.require.end.user.distinction', True)

        # save order for TBS with B2B/B2C setting
        order = cc._add_contribution_order(tm, lines, 2013, month=3)
        self.assertEqual(order.lines[0].is_b2b, True)

        # submit the declaration (for export)
        order = cc.submit_order(order.id)

        # make a request for resubmission
        cc.request_for_resubmission(order.id)

        # now the operator approves the resubmit.
        cod = PvcContributionOrderController(self.sa_session, account_id=self.operator_account.id)
        cod.approve_resubmit_request(order.id)

        # now that we have an approved resubmit request on a submitted declaration, an update on this declaration will do a resubmit.
        resubmit = cc.update_contribution(order.id, {'line_0_quantity': 225, 'line_0_weight': Decimal('15.4'), 'line_0_is_b2b': False})
        self.sa_session.expire(order)
        self.assertEqual(resubmit.lines[0].is_b2b, False)

    def _trash_all_rules(self):
        for rule in self.sa_session.query(OrderRule).filter(OrderRule.domain==self.operator_account.default_group.license.domain):
            self.sa_session.delete(rule)

    def _make_rule(self, name, **kwargs):

        controller = PvcContributionOrderController(self.sa_session, account_id=self.operator_account.id)

        controller.save_rule(None, {
            'order_type': kwargs.get('order_type', 'MockOrder'),
            'name': name,
            'enabled': kwargs.get('enabled', True),
            'execute_order': kwargs.get('execute_order', None),
            'description': kwargs.get('description', ''),
            'start_date': kwargs.get('start_date', now()),
            'end_date': kwargs.get('end_date', None),
            'match_all': kwargs.get('match_all', True),
            'stop_processing': kwargs.get('stop_processing', False),
            'conditions': kwargs.get('conditions', []),
            'actions': kwargs.get('actions', [])
        })

    def _make_rule_order(self, tm, **kwargs):

        pc = PvcProductController(self.sa_session, account_id=self.operator_account.id)
        mp = pc.get_or_save_tbs_member_product(tm, self.tbs_product.id)

        lines = [{
            'description': 'Some product',
            'polymorphic_type': PvcContributionOrderLine,
            'quantity': kwargs.get('quantity', 1),
            'price': Decimal('10.0'),
            'tbs_member_product': mp,
            'weight': Decimal('10'),
            'custom_unit': None
        }]

        year = kwargs.get('year', now().year)
        month = kwargs.get('month', None)
        quarter = kwargs.get('quarter', None)

        controller = PvcContributionOrderController(self.sa_session, account_id=self.operator_account.id)
        order = controller._add_contribution_order(tm, lines, year, month, quarter)

        return order

    def test_rule_is_compliant(self):

        ma, tm = self._create_simple_member_for_declaration()
        cc = PvcContributionOrderController(self.sa_session, account_id=ma.id)

        self._make_rule('is compliant',
            order_type = 'PvcContributionOrder',
            conditions = [{
                'condition': 'is_compliant',
                'operator': None, 'value_text': None, 'value_bool': None,
                'value_int': PvcContributionOrder.IS_COMPLIANT,
                'value_decimal': None,
            },{
                'condition': 'merchant',
                'operator': None, 'value_text': None, 'value_bool': None,
                'value_int': tm.tbs_id,
                'value_decimal': None,
            },],
            actions = [{
                'action': 'fixed_amount_penalty',
                'description': '',
                'operator': None, 'value_text': None, 'value_bool': None, 'value_int': None,
                'value_decimal': 10,
                'tax_class_id': None, 'gl_account_code': None
            }]
        )

        order = self._make_rule_order(tm)
        order.is_compliant = True

        # test dry-run
        order_state = cc.load_state(order.id, submit_dry_run=True)
        self.assertEqual(order_state['lines'][0]['weight'], Decimal('10') )
        self.assertEqual(order_state['total_amount_excluding_vat'], Decimal('10.15') )

        # re-load the order, make sure the rules haven't yet been executed
        order = cc.get(order.id)[0]
        self.assertEqual(order.total_amount_excluding_vat, Decimal('0.15000') )

        # submit final
        # 10 units of 0.015, plus a penalty of 10 = 10.150
        cc.submit_order(order.id)
        self.assertEquals(order.total_amount_excluding_vat, Decimal('10.15') )

    def test_rule_period_quarterly(self):

        ma, tm = self._create_simple_member_for_declaration()
        tm.tbs.set_config('pvc.tbs.declaration.period', PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY)
        cc = PvcContributionOrderController(self.sa_session, account_id=ma.id)

        self._make_rule('rule period',
            order_type = 'PvcContributionOrder',
            conditions = [{
                'condition': 'year',
                'operator': None, 'value_text': None, 'value_bool': None,
                'value_int': now().year,
                'value_decimal': None,
            }, {
                'condition': 'quarter',
                'operator': None, 'value_text': None, 'value_bool': None,
                'value_int': now().month // 3 + 1,
                'value_decimal': None,
            }],
            actions = [{
                'action': 'fixed_amount_penalty',
                'description': '',
                'operator': None, 'value_text': None, 'value_bool': None, 'value_int': None,
                'value_decimal': 10,
                'tax_class_id': None, 'gl_account_code': None
            }]
        )

        order = self._make_rule_order(tm, year=now().year, quarter = now().month // 3 + 1, month = None)
        cc.submit_order(order.id)

        self.assertEquals(len(order.lines), 2 )

    def test_rule_period_monthly(self):
        ma, tm = self._create_simple_member_for_declaration()
        tm.tbs.set_config('pvc.tbs.declaration.period', PvcTbsLicense.DECLARATION_PERIOD_MONTHLY)
        cc = PvcContributionOrderController(self.sa_session, account_id=ma.id)

        self._make_rule('rule period',
            order_type = 'PvcContributionOrder',
            conditions = [{
                'condition': 'year',
                'operator': None, 'value_text': None, 'value_bool': None,
                'value_int': now().year,
                'value_decimal': None,
            }, {
                'condition': 'month',
                'operator': None, 'value_text': None, 'value_bool': None,
                'value_int': 5,
                'value_decimal': None,
            }],
            actions = [{
                'action': 'fixed_amount_penalty',
                'description': '',
                'operator': None, 'value_text': None, 'value_bool': None, 'value_int': None,
                'value_decimal': 10,
                'tax_class_id': None, 'gl_account_code': None
            }]
        )

        order = self._make_rule_order(tm, year=now().year, quarter=None, month=5)
        cc.submit_order(order.id)

        self.assertEquals(len(order.lines), 2 )

    def test_rule_producer(self):

        ma, tm = self._create_simple_member_for_declaration()
        cc = PvcContributionOrderController(self.sa_session, account_id=ma.id)

        self._make_rule('producer',
            order_type = 'PvcContributionOrder',
            conditions = [{
                'condition': 'producer',
                'operator': None, 'value_text': None, 'value_bool': None,
                'value_int': tm.id,
                'value_decimal': None,
            }],
            actions = [{
                'action': 'fixed_amount_penalty',
                'description': '',
                'operator': None, 'value_text': None, 'value_bool': None, 'value_int': None,
                'value_decimal': 10,
                'tax_class_id': None, 'gl_account_code': None
            }]
        )

        order = self._make_rule_order(tm, year=now().year, quarter = now().month // 3 + 1, month = now().month)
        cc.submit_order(order.id)
        self.assertEquals(len(order.lines), 2 )

        condition = [x for x in cc.get_available_rule_conditions('PvcContributionOrder') if x.name == 'month'][0]
        self.assertTrue(len(condition.get_select_options({})) > 0)

    def test_rule_is_free_rider(self):

        ma, tm = self._create_simple_member_for_declaration()
        cc = PvcContributionOrderController(self.sa_session, account_id=ma.id)

        tm.is_free_rider = True

        self._make_rule('producer',
            order_type = 'PvcContributionOrder',
            conditions = [{
                'condition': 'is_free_rider',
                'operator': None, 'value_text': None, 'value_bool': True, 'value_int': None,
                'value_decimal': None,
            }],
            actions = [{
                'action': 'fixed_amount_penalty',
                'description': '',
                'operator': None, 'value_text': None, 'value_bool': None, 'value_int': None,
                'value_decimal': 10,
                'tax_class_id': None, 'gl_account_code': None
            }]
        )

        order = self._make_rule_order(tm)
        cc.submit_order(order.id)
        self.assertEquals(len(order.lines), 2 )

    def test_rule_client_override_by_member(self):

        ma, tm = self._create_simple_member_for_declaration()
        cc = PvcContributionOrderController(self.sa_session, account_id=ma.id)

        self._make_rule('producer',
            order_type = 'PvcContributionOrder',
            conditions = [{
                'condition': 'client',
                'operator': None, 'value_text': None, 'value_bool': None, 'value_int': tm.member.id,
                'value_decimal': None,
            }],
            actions = [{
                'action': 'fixed_amount_penalty',
                'description': '',
                'operator': None, 'value_text': None, 'value_bool': None, 'value_int': None,
                'value_decimal': 10,
                'tax_class_id': None, 'gl_account_code': None
            }]
        )

        order = self._make_rule_order(tm)
        cc.submit_order(order.id)
        self.assertEquals(len(order.lines), 2 )

        # create a second one that shouldnt match
        ma, tm2 = self._create_simple_member_for_declaration()
        cc = PvcContributionOrderController(self.sa_session, account_id=ma.id)

        self._make_rule('producer',
            order_type = 'PvcContributionOrder',
            conditions = [{
                'condition': 'client',
                'operator': None, 'value_text': None, 'value_bool': None, 'value_int': tm.member.id,
                'value_decimal': None,
            }],
            actions = [{
                'action': 'fixed_amount_penalty',
                'description': '',
                'operator': None, 'value_text': None, 'value_bool': None, 'value_int': None,
                'value_decimal': 10,
                'tax_class_id': None, 'gl_account_code': None
            }]
        )

        order = self._make_rule_order(tm2)
        cc.submit_order(order.id)
        self.assertEquals(len(order.lines), 1 )

    def test_resubmit_with_rule(self):

        ma, tm = self._create_simple_member_for_declaration()
        cc = PvcContributionOrderController(self.sa_session, account_id=ma.id)
        product = self._get_member_product(tm)

        # rule that always matches gives 10 eurie penalty
        self._make_rule('penalty',
            order_type = 'PvcContributionOrder',
            conditions = [],
            actions = [{
                'action': 'fixed_amount_penalty',
                'description': '',
                'operator': None, 'value_text': None, 'value_bool': None, 'value_int': None,
                'value_decimal': 10,
                'tax_class_id': None, 'gl_account_code': 'GL_CONTRIB_FEE'
            }]
        )

        # insert new declaration
        lines = [{
            'tbs_member_product': product,
            'quantity': 200,
            'weight': Decimal('15.3'),
            'custom_unit': None,
        }]
        order = cc._add_contribution_order(tm, lines, 2013, month=3)

        # try to do a request for resubmission, this should fail as the declaration is still outstanding
        self.assertRaises(PublicUserException, cc.request_for_resubmission, order.id)

        # preview declaration should include penalty
        order_state = cc.load_order_state(order.id, submit_dry_run=True)
        self.assertEquals(order_state['total_amount_excluding_vat'], Decimal('10.2295') ) # includes 10 penalty

        # submit the declaration (for export)
        order = cc.submit_order(order.id)

        # try to resubmit the declaration, this should trigger an error
        self.assertRaises(PublicUserException, cc.update_contribution, order.id, {'line_0_quantity': 225, 'line_0_weight': Decimal('15.3')})

        # count notifications for operator
        nc = NotificationController(self.sa_session, account_id=self.operator_account.id)
        cnt = nc.get_unread_notifications_count()

        # make a request for resubmission
        cc.request_for_resubmission(order.id)

        # check that notification is sent and sender is as expected
        notif = self.sa_session.query(Notification).order_by(Notification.id.desc()).first()
        self.assertEqual(notif.sender, tm.member.default_group)

        self.sa_session.expire(order)
        self.assertEqual(order.resubmit_request, PvcContributionOrder.RESUBMIT_REQUEST_PENDING)

        # assert that the operator received a new notification.
        self.assertEqual(nc.get_unread_notifications_count(), cnt+1)

        # try to request another time, this should raise an exception
        self.assertRaises(PublicUserException, cc.request_for_resubmission, order.id)

        # get the notification count of the member account
        nc = NotificationController(self.sa_session, account_id=ma.id)
        cnt = nc.get_unread_notifications_count()

        # now the operator approves the resubmit.
        cod = PvcContributionOrderController(self.sa_session, account_id=self.operator_account.id)
        cod.approve_resubmit_request(order.id)

        # assert that the member received a new notification.
        self.assertEqual(nc.get_unread_notifications_count(), cnt+1)

        # now that we have an approved resubmit request on a submitted declaration, an update on this declaration will do a resubmit.
        resubmit = cc.update_contribution(order.id, {'line_0_quantity': 225, 'line_0_weight': Decimal('15.4')})
        self.sa_session.expire(order)
        self.assertEqual(order.resubmit_request, None)

        # the resubmit's parent is the credited order.
        credit = resubmit.parent_order
        self.assertEqual(credit.parent_order, order) # our initial order should be the same as the first submit
        self.assertEqual(credit.lines[0].quantity, -200)
        self.assertEqual(credit.lines[0].weight, Decimal('-15.3'))
        self.assertEqual(credit.lines[0].line_amount_excluding_vat, Decimal('-0.2295'))
        self.assertEqual(credit.lines[1].quantity, -1)
        self.assertEqual(credit.lines[1].line_amount_excluding_vat, Decimal('-10'))
        self.assertEqual(credit.lines[1].gl_account_number, 'GL789012')
        self.assertEqual(credit.status, PvcContributionOrder.STATUS_PENDING)
        self.assertEqual(credit.total_amount_excluding_vat, Decimal('-10.2295')) # includes penalty

        # the new resubmit shouldnt contain the rule lines
        self.assertEqual(resubmit.lines[0].quantity, 225)
        self.assertEqual(resubmit.lines[0].weight, Decimal('15.4'))
        self.assertEqual(resubmit.total_amount_excluding_vat, Decimal('0.2310')) # does not (yet?) include penalty
        self.assertEquals(len(resubmit.lines), 1)

        # submit the resubmit
        resubmit = cc.submit_order(resubmit.id)

        # now the credit and resubmit should be submitted
        self.assertEqual(credit.status, PvcContributionOrder.STATUS_SUBMITTED)
        self.assertEqual(resubmit.status, PvcContributionOrder.STATUS_SUBMITTED)

    def test_apply_member_vat_rate_rule_action(self):
        """
        The following scenario is applicable:
            Merchant/TBS = French
            Client/Member = Guadeloupenesian!!!?
            Normal scenario = Merchant.country != Client.country = no vat
            Exception scenario = Merchant == TBS France && Client.country == Guadeloupe = apply Guadelope VAT rate.
        :return:
        """

        ma, tm = self._create_simple_member_for_declaration()

        country = self.sa_session.query(Country).filter(Country.iso_3166_a2 == 'GP').one()

        # alter the country of the member to be different than the country of the TBS
        ma.default_group.license.company.addresses[0].address.country = country
        self.sa_session.add(ma)
        self.sa_session.flush()

        # set the VAT rate for the member country to something different than the TBS country VAT rate
        self._create_contribution_tax_rates(self.operator_account, country=country, rate=Decimal('7.00'))

        # create the rule, Merchant = TBS, member country, action.
        self._make_rule('guadelope_vat_rate',
            order_type = 'PvcContributionOrder',
            conditions = [{
                'condition': 'client_country',
                'operator': None, 'value_text': None, 'value_bool': None, 'value_int': country.id,
                'value_decimal': None,
            }],
            actions = [{
                'action': 'pvc_apply_member_country_vat_rate',
                'description': '',
                'operator': None, 'value_text': None, 'value_bool': True, 'value_int': None,
                'value_decimal': None,
                'tax_class_id': None, 'gl_account_code': None
            }]
        )

        cc = PvcContributionOrderController(self.sa_session, account_id=ma.id)
        product = self._get_member_product(tm)

        # insert new declaration
        lines = [{
            'tbs_member_product': product,
            'quantity': 200,
            'weight': Decimal('412.55'),
            'custom_unit': None,
        }]
        order = cc._add_contribution_order(tm, lines, 2013, month=3)

        # draai run.
        order_state = cc.load_state(order.id, submit_dry_run=True)

        # assert contribution order lines to have member country VAT rate
        line = order_state['lines'][0]
        self.assertEqual(line['vat_rate'], Decimal('7.00'))
        self.assertEqual(line['line_amount_excluding_vat'], Decimal('6.18825'))

        # 412.55(weight)*0.015(fee)*1.07(vat)
        self.assertEqual(order_state['total_amount_including_vat'], Decimal('6.6214275'))

        order = cc.submit_order(order.id)
        self.assertEqual(order.total_amount_excluding_vat, Decimal('6.18825'))
        self.assertEqual(order.total_amount_including_vat, Decimal('6.62143'))

    def test_apply_member_vat_rate_rule_action_and_discount(self):
        """
        The following scenario is applicable:
            Merchant/TBS = French
            Client/Member = Guadeloupenesian!!!?
            Normal scenario = Merchant.country != Client.country = no vat
            Exception scenario = Merchant == TBS France && Client.country == Guadeloupe = apply Guadelope VAT rate.
        :return:
        """

        ma, tm = self._create_simple_member_for_declaration()

        country = self.sa_session.query(Country).filter(Country.iso_3166_a2 == 'GP').one()

        # alter the country of the member to be different than the country of the TBS
        ma.default_group.license.company.addresses[0].address.country = country
        self.sa_session.add(ma)
        self.sa_session.flush()

        # set the VAT rate for the member country to something different than the TBS country VAT rate
        self._create_contribution_tax_rates(self.operator_account, country=country, rate=Decimal('7.00'))

        # create the rule, Merchant = TBS, member country, action.
        self._make_rule('guadelope_vat_rate',
            execute_order=2,
            order_type = 'PvcContributionOrder',
            conditions = [{
                'condition': 'client_country',
                'operator': None, 'value_text': None, 'value_bool': None, 'value_int': country.id,
                'value_decimal': None,
            }],
            actions = [{
                'action': 'pvc_apply_member_country_vat_rate',
                'description': '',
                'operator': None, 'value_text': None, 'value_bool': True, 'value_int': None,
                'value_decimal': None,
                'tax_class_id': None, 'gl_account_code': None
            }]
        )

        c = TaxController(self.sa_session, account_id = self.operator_account.id)

        self._make_rule('discounty',
            execute_order=1,
            order_type = 'PvcContributionOrder',
            conditions = [{
                'condition': 'merchant',
                'operator': None, 'value_text': None, 'value_bool': None, 'value_int': self.tbs.id,
                'value_decimal': None,
            }],
            actions = [{
                'action': 'relative_amount_discount',
                'description': 'Discount',
                'operator': None, 'value_text': None, 'value_bool': None, 'value_int': None,
                'value_decimal': 10,
                'tax_class_id': c.get_tax_classes().pop().id, 'gl_account_code': None
            }]
        )

        cc = PvcContributionOrderController(self.sa_session, account_id=ma.id)
        product = self._get_member_product(tm)

        # insert new declaration
        lines = [{
            'tbs_member_product': product,
            'quantity': 200,
            'weight': Decimal('412.55'),
            'custom_unit': None,
        }]
        order = cc._add_contribution_order(tm, lines, 2013, month=3)

        # draai run.
        order_state = cc.load_state(order.id, submit_dry_run=True)

        # assert contribution order lines to have member country VAT rate
        line = order_state['lines'][0]
        self.assertEqual(line['vat_rate'], Decimal('7.00'))
        self.assertEqual(line['line_amount_excluding_vat'], Decimal('6.18825'))

        ruleline = order_state['lines'][1]
        self.assertEqual(line['vat_rate'], Decimal('7.00'))

        # 412.55(weight)*0.015(fee)*1.07(vat)
        self.assertEqual(order_state['total_amount_including_vat'], Decimal('5.9592901000'))

        order = cc.submit_order(order.id)
        self.assertEqual(order.total_amount_excluding_vat, Decimal('5.56943'))
        self.assertEqual(order.total_amount_including_vat, Decimal('5.95929'))

    def test_b2b_percentage_rule_action(self):
        # enable setting on TBS!

        ma, tm = self._create_simple_member_for_declaration()
        # we set the countries to the same, so a VAT is applied (reproduce bug #949)
        tm.member.company.addresses[0].address.country = tm.tbs.company.addresses[0].address.country
        self.sa_session.add(tm.member.company.addresses[0].address)
        self.sa_session.flush()

        self.tbs.set_config('pvc.tbs.contribution.automatic.end.user.distinction', True)

        self._make_rule('italy_b2b_split',
            order_type = 'PvcContributionOrder',
            conditions = [{
                'condition': 'merchant',
                'operator': None, 'value_text': None, 'value_bool': None, 'value_int': self.tbs.id,
                'value_decimal': None,
            }],
            actions = [{
                'action': 'pvc_contribution_b2b_percentage',
                'description': '',
                'operator': None, 'value_text': None, 'value_bool': None, 'value_int': 10,
                'value_decimal': None,
                'tax_class_id': None, 'gl_account_code': None
            }
            ]
        )

        cc = PvcContributionOrderController(self.sa_session, account_id=ma.id)
        product = self._get_member_product(tm)
        product2 = self._get_member_product(tm, tbs_product=self.tbs_product2)

        # insert new declaration
        lines = [{
            'tbs_member_product': product,
            'quantity': 234,
            'weight': Decimal('1024.23'),
            'custom_unit': None,
        }, {
            'tbs_member_product': product2,
            'quantity': 111,
            'weight': Decimal('542.22'),
            'custom_unit': None,
        }]
        order = cc._add_contribution_order(tm, lines, 2013, month=3)
        self.assertEqual(len(order.lines), 2)
        self.assertFalse(order.lines[0].is_b2b)
        self.assertFalse(order.lines[1].is_b2b)

        # draai run.
        order_state = cc.load_state(order.id, submit_dry_run=True)
        self.assertEqual(len(order_state['lines']), 4)
        lines = order_state['lines']
        self.assertFalse(lines[0]['is_b2b'])
        self.assertTrue(lines[1]['is_b2b'])
        self.assertFalse(lines[2]['is_b2b'])
        self.assertTrue(lines[3]['is_b2b'])

        self.assertEqual(order_state['total_amount_excluding_vat'], Decimal('34.34114'))
        self.assertEqual(order_state['total_amount_including_vat'], Decimal('41.55278'))

        order = cc.submit_order(order.id)
        self.assertEqual(len(order.lines), 4)
        self.assertEqual(order.total_amount_including_vat, Decimal('41.55278'))
        self.assertEqual(order.lines[0].quantity, 211)
        self.assertEqual(order.lines[1].quantity, 23)
        self.assertEqual(order.lines[2].quantity, 100)
        self.assertEqual(order.lines[3].quantity, 11)

    def test_b2b_percentage_rule_action_resubmit(self):

        ma, tm = self._create_simple_member_for_declaration()

        self.tbs.set_config('pvc.tbs.contribution.automatic.end.user.distinction', True)

        self._make_rule('italy_b2b_split',
            order_type = 'PvcContributionOrder',
            conditions = [{
                'condition': 'merchant',
                'operator': None, 'value_text': None, 'value_bool': None, 'value_int': self.tbs.id,
                'value_decimal': None,
            }],
            actions = [{
                'action': 'fixed_amount_discount',
                'description': 'Discount',
                'operator': None, 'value_text': None, 'value_bool': None, 'value_int': None,
                'value_decimal': 10,
                'tax_class_id': None, 'gl_account_code': None
            }, {
                'action': 'pvc_contribution_b2b_percentage',
                'description': '',
                'operator': None, 'value_text': None, 'value_bool': None, 'value_int': 10,
                'value_decimal': None,
                'tax_class_id': None, 'gl_account_code': None
            }]
        )

        cc = PvcContributionOrderController(self.sa_session, account_id=ma.id)
        product = self._get_member_product(tm)
        product2 = self._get_member_product(tm, tbs_product=self.tbs_product2)

        # insert new declaration
        lines = [{
            'tbs_member_product': product,
            'quantity': 234,
            'weight': Decimal('1024.23'),
            'custom_unit': None,
        }, {
            'tbs_member_product': product2,
            'quantity': 111,
            'weight': Decimal('542.22'),
            'custom_unit': None,
        }]
        order = cc._add_contribution_order(tm, lines, 2013, month=3)
        self.assertEqual(len(order.lines), 2)

        order = cc.submit_order(order.id)

        self.sa_session.flush()

        self.assertEqual(len(order.lines), 5)
        self.assertEqual(order.total_amount_including_vat, Decimal('24.34114'))

        # make a request for resubmission
        cc.request_for_resubmission(order.id)

        # now the operator approves the resubmit.
        cod = PvcContributionOrderController(self.sa_session, account_id=self.operator_account.id)
        cod.approve_resubmit_request(order.id)

        # now that we have an approved resubmit request on a submitted declaration, an update on this declaration will do a resubmit.
        resubmit = cc.update_contribution(order.id, {
            'line_0_quantity': 643,
            'line_0_weight': Decimal('234.4'),
            'line_2_quantity': 35,
            'line_2_weight': Decimal('345.4'),
        })
        self.sa_session.expire(order)
        self.assertEqual(len(resubmit.parent_order.lines), 5)
        self.assertEqual(len(resubmit.lines), 2)

        line0 = resubmit.lines[0]
        line1 = resubmit.lines[1]
        self.assertEqual(line0.quantity, 643)
        self.assertFalse(line0.is_b2b)
        self.assertEqual(line1.quantity, 35)
        self.assertFalse(line1.is_b2b)

    def test_rule_simulate(self):
        from twisted.trial.unittest import SkipTest

        ma, tm = self._create_simple_member_for_declaration()
        product = self._get_member_product(tm)

        self._make_rule('penalty',
            order_type = 'PvcContributionOrder',
            match_all=True,
            conditions = [],
            actions = [ {
                'action': 'fixed_amount_penalty',
                'description': None,
                'operator': None, 'value_text': None, 'value_bool': None, 'value_int': None,
                'value_decimal': Decimal('10.0'),
                'gl_account_code': None,
                'tax_class_id': None,
            }, ]
        )

        order_state = {

            'polymorphic_type': 'PvcContributionOrder',
            'merchant': tm.tbs.id,
            'producer': tm.id,
            'tbs_member': None, # no subrogation

            'year': 2015,
            'month': 1,
            'quarter': 1,
            'is_forecast': False,

            'lines': [{
                    'polymorphic_type': 'PvcContributionOrderLine',
                    'description': 'line 1',
                    'quantity': 10,
                    'price': Decimal('1'),
                    'tbs_member_product': product.id,
                    'weight': 10,
                    'custom_unit': None,
                    'is_b2b': False,
                    'brand': None,
                }
            ],

        }

        # process the simulate order
        cc = PvcContributionOrderController(self.sa_session, account_id=self.operator_account.id)
        order_state = cc.simulate_submit(order_state)
        self.assertEqual(order_state['total_amount_excluding_vat'], Decimal('20') )

        #pp = PrettyPrinter(indent=4)
        #pp.pprint(order_state)

    def test_brand_split(self):
        ma, tm = self._create_simple_member_for_declaration()

        self.tbs.set_config('pvc.tbs.contribution.split.by.brand', True)

        pc = PvcProductController(self.sa_session, account_id=ma.id)
        cc = PvcContributionOrderController(self.sa_session, account_id=ma.id)
        product = self._get_member_product(tm)
        product2 = self._get_member_product(tm, tbs_product=self.tbs_product2)

        # insert new declaration
        lines = [{
            'tbs_member_product': product,
            'quantity': 234,
            'weight': Decimal('1024.23'),
            'custom_unit': None,
            'brand': pc.get_or_save_brand('Philips', product),
        }, {
            'tbs_member_product': product2,
            'quantity': 111,
            'weight': Decimal('542.22'),
            'custom_unit': None,
            'brand': pc.get_or_save_brand('Samsung', product),
        }]
        order = cc._add_contribution_order(tm, lines, 2013, month=3)
        self.assertEqual(len(order.lines), 2)

        order = cc.update_contribution(order.id, {
            'line_0_quantity': 643,
            'line_0_weight': Decimal('234.4'),
            'line_0_brand': "Philips",
            'line_1_quantity': 35,
            'line_1_weight': Decimal('345.4'),
            'line_1_brand': "Samsung",
        })
        self.sa_session.expire(order)

        self.assertEqual(len(order.lines), 2)
        lines = order.lines
        philips = lines[0].brand
        self.assertEqual(lines[0].brand.name, "Philips")
        self.assertEqual(lines[1].brand.name, "Samsung")

        # test case insensitive lookup
        order = cc.update_contribution(order.id, {
            'line_0_quantity': 643,
            'line_0_weight': Decimal('234.4'),
            'line_0_brand': "Sony",
            'line_1_quantity': 35,
            'line_1_weight': Decimal('345.4'),
            'line_1_brand': "pHiLiPS",
        })
        self.sa_session.expire(order)
        self.assertEqual(lines[0].brand.name, "Sony")
        self.assertEqual(lines[1].brand.name, "Philips")
        self.assertEqual(lines[1].brand, philips)

        order = cc.submit_order(order.id)

        # assert accountant
        adc = PvcContributionOrderController(self.sa_session, account_id=self.accountant.id)
        states = adc.get_unexported_orders_for_merchant(tm.tbs.id)

        self.assertEquals(states[0]['lines'][0]['properties'][7]['key'], 'brand')
        self.assertEquals(states[0]['lines'][0]['properties'][7]['value_str'], 'Sony')

    def test_brand_split_with_additional_line_and_reindex_lines(self):
        ma, tm = self._create_simple_member_for_declaration()

        self.tbs.set_config('pvc.tbs.contribution.split.by.brand', True)

        pc = PvcProductController(self.sa_session, account_id=ma.id)
        cc = PvcContributionOrderController(self.sa_session, account_id=ma.id)
        product = self._get_member_product(tm)
        product2 = self._get_member_product(tm, tbs_product=self.tbs_product2)

        # insert new declaration
        lines = [{
            'tbs_member_product': product,
            'quantity': 234,
            'weight': Decimal('1024.23'),
            'custom_unit': None,
            'brand': pc.get_or_save_brand('Samsung', product),
        }, {
            'tbs_member_product': product2,
            'quantity': 111,
            'weight': Decimal('542.22'),
            'custom_unit': None,
            'brand': pc.get_or_save_brand('Philips', product),
        }]
        order = cc._add_contribution_order(tm, lines, 2013, month=3)
        self.assertEqual(len(order.lines), 2)

        extra = {
            'tbs_product_tpid_%d' % self.tbs_product.id: self.tbs_product.id,
            'line_tpid_%d_weight' % self.tbs_product.id: Decimal('13.6'),
            'line_tpid_%d_quantity' % self.tbs_product.id: 100,
            'line_tpid_%d_brand' % self.tbs_product.id: 'Aaaa',
            'line_tpid_%d_country_id' % self.tbs_product.id: None,
        }

        order = cc.update_contribution(order.id, {
            'line_0_quantity': 643,
            'line_0_weight': Decimal('234.4'),
            'line_0_brand': "Samsung",
            'line_1_quantity': 35,
            'line_1_weight': Decimal('345.4'),
            'line_1_brand': "Philips",
        }, additional_lines=extra)
        self.sa_session.expire(order)

        self.assertEqual(len(order.lines), 3)
        # The lines should be index per product, per brand name
        # Samsung and Aaaa are both part of the same product so they should be 1 & 2
        self.assertEqual(order.lines[0].brand.name, 'Aaaa')
        self.assertEqual(order.lines[0].index, 0)
        self.assertEqual(order.lines[1].brand.name, 'Samsung')
        self.assertEqual(order.lines[1].index, 1)
        self.assertEqual(order.lines[2].brand.name, 'Philips')
        self.assertEqual(order.lines[2].index, 2)

    def test_brand_split_with_automatic_b2b_split(self):

        ma, tm = self._create_simple_member_for_declaration()

        self.tbs.set_config('pvc.tbs.contribution.split.by.brand', True)
        self.tbs.set_config('pvc.tbs.contribution.automatic.end.user.distinction', True)

        self._make_rule('italy_b2b_split',
            order_type = 'PvcContributionOrder',
            conditions = [{
                'condition': 'merchant',
                'operator': None, 'value_text': None, 'value_bool': None, 'value_int': self.tbs.id,
                'value_decimal': None,
            }],
            actions = [{
                'action': 'pvc_contribution_b2b_percentage',
                'description': '',
                'operator': None, 'value_text': None, 'value_bool': None, 'value_int': 75,
                'value_decimal': None,
                'tax_class_id': None, 'gl_account_code': None
            }]
        )

        pc = PvcProductController(self.sa_session, account_id=ma.id)
        cc = PvcContributionOrderController(self.sa_session, account_id=ma.id)
        product = self._get_member_product(tm)
        product2 = self._get_member_product(tm, tbs_product=self.tbs_product2)

        # insert new declaration
        lines = [{
            'tbs_member_product': product,
            'quantity': 5,
            'weight': Decimal('1024.23'),
            'custom_unit': None,
            'brand': pc.get_or_save_brand('Philips', product),
        }, {
            'tbs_member_product': product2,
            'quantity': 111,
            'weight': Decimal('542.22'),
            'custom_unit': None,
            'brand': pc.get_or_save_brand('Samsung', product),
        }]
        order = cc._add_contribution_order(tm, lines, 2013, month=3)
        self.assertEqual(len(order.lines), 2)

        order = cc.submit_order(order.id)
        self.assertEqual(len(order.lines), 4)
        self.assertEqual(order.total_amount_including_vat, Decimal('34.34115'))

        lines = order.lines
        self.assertEqual(lines[0].brand.name, "Philips")
        self.assertEqual(lines[0].quantity, 1)
        self.assertEqual(lines[0].weight, Decimal('256.05750'))
        self.assertEqual(lines[1].brand.name, "Philips")
        self.assertEqual(lines[1].quantity, 4)
        self.assertEqual(lines[1].weight, Decimal('768.17250'))
        self.assertEqual(lines[2].brand.name, "Samsung")
        self.assertEqual(lines[3].brand.name, "Samsung")

    def test_merge_brands(self):
        """
        This test is in contribution as most of the processing is related to the actual contribution orders (lines)
        :return:
        """
        self._reset_brands()
        # make a few brands
        ma, tm = self._create_simple_member_for_declaration()

        self.tbs.set_config('pvc.tbs.contribution.split.by.brand', True)

        pc = PvcProductController(self.sa_session, account_id=ma.id)
        cc = PvcContributionOrderController(self.sa_session, account_id=ma.id)
        product = self._get_member_product(tm)
        product2 = self._get_member_product(tm, tbs_product=self.tbs_product2)
        orig = pc.get_or_save_brand('Philips', product)

        # insert new declaration
        lines = [{
            'tbs_member_product': product,
            'quantity': 234,
            'weight': Decimal('1024.23'),
            'custom_unit': None,
            'brand': pc.get_or_save_brand('Samsung', product),
        }, {
            'tbs_member_product': product2,
            'quantity': 111,
            'weight': Decimal('542.22'),
            'custom_unit': None,
            'brand': pc.get_or_save_brand('Philllips', product),
        }]
        # add a few contributions with different brands
        order = cc._add_contribution_order(tm, lines, 2013, month=3)

        brand_a = order.lines[1].brand
        # member account does not have the permission to mark to be merged
        self.assertRaises(PublicSecurityException, pc.merge_brand_with, brand_a.id, orig.id)

        # mark brands to be merged
        pc = PvcProductController(self.sa_session, account_id=self.operator_account.id)
        pc.merge_brand_with(brand_a.id, orig.id)

        # merge brands
        MergeBrandTask().execute()

        # validate member product brands & contribution lines
        self.assertNotEqual(pc._get_tbs_member_product_brand(orig, product), None)
        self.assertRaises(ObjectDeletedError, getattr, brand_a, 'id')

        self.sa_session.expire(order)
        line = order.lines[1]
        self.assertEqual(line.brand, orig)

    def test_custom_unit(self):
        """
        """

        products = self.pc.get_available_products_for_tbs(self.tbs.id)

        tbs_product_quantity = self.pc.save_or_update_tbs_product(None, 'A-EEE-TBS-TEST-3', 'A-WEEE-3', self.tbs.id,
            parent_level_id=products[1]['id'], basis_for_invoicing=PvcTbsProduct.INVOICE_BASED_ON_QUANTITY
        )

        tbs_product_custom_unit = self.pc.save_or_update_tbs_product(None, 'EEE-TBS-TEST-4', 'WEEE-3', self.tbs.id,
            parent_level_id=products[1]['id'], basis_for_invoicing=PvcTbsProduct.INVOICE_BASED_ON_CUSTOM_UNIT, custom_unit=PvcTbsProduct.CUSTOM_UNIT_NOMINAL_POWER
        )

        self.pc.save_contribution_fee(tbs_product_quantity.id, self.service_level.id, Decimal('0.016'), date(2013, 1, 1))
        self.pc.save_contribution_fee(tbs_product_custom_unit.id, self.service_level.id, Decimal('0.017'), date(2013, 1, 1))

        ma, tm = self._create_simple_member_for_declaration()
        cc = PvcContributionOrderController(self.sa_session, account_id=ma.id)

        product_quantity = self._get_member_product(tm, tbs_product=tbs_product_quantity)
        product_weight = self._get_member_product(tm, tbs_product=self.tbs_product)
        product_custom_unit = self._get_member_product(tm, tbs_product=tbs_product_custom_unit)

        lines = [{
            'tbs_member_product': product_quantity,
            'quantity': 10,
            'weight': Decimal(101),
            'custom_unit': None,
        }, {
            'tbs_member_product': product_weight,
            'quantity': 11,
            'weight': Decimal('111'),
            'custom_unit': None,
        }, {
            'tbs_member_product': product_custom_unit,
            'quantity': 12,
            'weight': Decimal(121),
            'custom_unit': Decimal('0'),
        }]

        # check that zero custom unit raises an exception.
        self.assertRaises(PublicUserException, cc._add_contribution_order, tm, lines, 2013, month=3)
        lines[2]['custom_unit'] = Decimal('22')

        order = cc._add_contribution_order(tm, lines, 2013, month=3)

        self.assertEqual(order.lines[0].quantity, 10)
        self.assertEqual(order.lines[0].weight, 101)
        self.assertEqual(order.lines[0].custom_unit, None)

        self.assertEqual(order.lines[1].quantity, 11)
        self.assertEqual(order.lines[1].weight, Decimal('111'))
        self.assertEqual(order.lines[1].custom_unit, None)

        self.assertEqual(order.lines[2].quantity, 12)
        self.assertEqual(order.lines[2].weight, 121)
        self.assertEqual(order.lines[2].custom_unit, Decimal('22'))

        # submit
        order = cc.submit_order(order.id)

        # request for resubmission and approve
        cc.request_for_resubmission(order.id)
        cod = PvcContributionOrderController(self.sa_session, account_id=self.operator_account.id)
        cod.approve_resubmit_request(order.id)

        self.assertRaises(PublicUserException, cc.update_contribution, order.id, {
            'line_0_quantity': 30,
            'line_0_weight': 301,
            'line_0_custom_unit': None,
            'line_1_quantity': 31,
            'line_1_weight': Decimal(311),
            'line_1_custom_unit': None,
            'line_2_quantity': 32,
            'line_2_weight': 0,
            'line_2_custom_unit': 0,
        })

        # resubmit
        resubmit = cc.update_contribution(order.id, {
            'line_0_quantity': 30,
            'line_0_weight': 301,
            'line_0_custom_unit': None,
            'line_1_quantity': 31,
            'line_1_weight': Decimal(311),
            'line_1_custom_unit': None,
            'line_2_quantity': 32,
            'line_2_weight': Decimal(321),
            'line_2_custom_unit': Decimal('42'),
        })

        self.assertEqual(resubmit.lines[0].quantity, 30)
        self.assertEqual(resubmit.lines[0].weight, 301)
        self.assertEqual(resubmit.lines[0].custom_unit, None)

        self.assertEqual(resubmit.lines[1].quantity, 31)
        self.assertEqual(resubmit.lines[1].weight, Decimal('311'))
        self.assertEqual(resubmit.lines[1].custom_unit, None)

        self.assertEqual(resubmit.lines[2].quantity, 32)
        self.assertEqual(resubmit.lines[2].weight, 321)
        self.assertEqual(resubmit.lines[2].custom_unit, Decimal('42'))


        # assert its ready for the accountant
        adc = PvcContributionOrderController(self.sa_session, account_id=self.accountant.id)
        states = adc.get_unexported_orders_for_merchant(tm.tbs.id)

        self.assertEquals(states[0]['lines'][2]['properties'][7]['key'], 'custom_unit_label')
        self.assertEquals(states[0]['lines'][2]['properties'][7]['value_str'], 'Nominal power')

        self.assertEquals(states[0]['lines'][2]['properties'][8]['key'], 'custom_unit_unit')
        self.assertEquals(states[0]['lines'][2]['properties'][8]['value_str'], 'Wp')

        self.assertEquals(states[0]['lines'][2]['properties'][9]['key'], 'custom_unit_value')
        self.assertEquals(states[0]['lines'][2]['properties'][9]['value_decimal'], Decimal('22.00000'))


    def test_forecast_from_actuals(self):
        ma, tm = self._create_simple_member_for_declaration()

        self.tbs.set_config('pvc.tbs.advance.invoicing', True)
        self.tbs.set_config('pvc.tbs.declaration.period', PvcTbsLicense.DECLARATION_PERIOD_MONTHLY)

        pc = PvcProductController(self.sa_session, account_id=ma.id)
        cc = PvcContributionOrderController(self.sa_session, account_id=ma.id)
        product = self._get_member_product(tm)
        product2 = self._get_member_product(tm, tbs_product=self.tbs_product2)

        # insert new declaration
        lines = [{
            'tbs_member_product': product,
            'quantity': 234,
            'weight': Decimal('1024.23'),
            'custom_unit': None,
        }, {
            'tbs_member_product': product2,
            'quantity': 111,
            'weight': Decimal('542.22'),
            'custom_unit': None,
        }]
        order = cc._add_contribution_order(tm, lines, 2013, month=6)

        self.assertRaises(PublicSecurityException, cc.add_forecast_from_actuals, order)

        original = cc.submit_order(order)

        orders = cc.get()
        self.assertEqual(len(orders), 2)
        order_a, order_b = orders
        forecast = order_b
        if not forecast.is_forecast:
            forecast = order_a

        self.assertEqual(original.status, PvcContributionOrder.STATUS_SUBMITTED)
        self.assertEqual(forecast.status, PvcContributionOrder.STATUS_SUBMITTED)
        self.assertEqual(forecast.year, original.year)
        self.assertEqual(forecast.month, original.month+1)
        self.assertEqual(forecast.quarter, None)
        self.assertEqual(len(forecast.lines), 2)

        self.assertEqual(forecast.lines[0].gl_account_number, 'GLF0R3CAS7')
        self.assertEqual(forecast.lines[1].gl_account_number, 'GLF0R3CAS7')

    def test_forecast_from_actuals_resubmit(self):
        ma, tm = self._create_simple_member_for_declaration()
        cc = PvcContributionOrderController(self.sa_session, account_id=ma.id)
        product = self._get_member_product(tm)

        self.tbs.set_config('pvc.tbs.advance.invoicing', True)
        self.tbs.set_config('pvc.tbs.declaration.period', PvcTbsLicense.DECLARATION_PERIOD_MONTHLY)

        # insert new declaration
        lines = [{
            'tbs_member_product': product,
            'quantity': 200,
            'weight': Decimal('15.3'),
            'custom_unit': None,
        }]
        order = cc._add_contribution_order(tm, lines, 2013, month=6)

        # submit the declaration (for export)
        order = cc.submit_order(order.id)

        orders = cc.get()
        self.assertEqual(len(orders), 2)

        # make a request for resubmission
        cc.request_for_resubmission(order.id)

        self.sa_session.expire(order)
        self.assertEqual(order.resubmit_request, PvcContributionOrder.RESUBMIT_REQUEST_PENDING)

        # now the operator approves the resubmit.
        cod = PvcContributionOrderController(self.sa_session, account_id=self.operator_account.id)
        cod.approve_resubmit_request(order.id)

        # now that we have an approved resubmit request on a submitted declaration
        resubmit = cc.update_contribution(order.id, {'line_0_quantity': 225, 'line_0_weight': Decimal('15.4')})
        self.sa_session.expire(order)
        self.assertEqual(order.resubmit_request, None)

        orders = cc.get() # we should only have 2 more orders (CREDIT and RESUBMIT)
        self.assertEqual(len(orders), 4)
        self.assertEqual(len([o for o in orders if o.is_forecast]), 1)  # we should only have ONE forecast order

    def test_forecast_from_actuals_with_rule_lines_added(self):
        ma, tm = self._create_simple_member_for_declaration()
        cc = PvcContributionOrderController(self.sa_session, account_id=ma.id)
        product = self._get_member_product(tm)

        self.tbs.set_config('pvc.tbs.advance.invoicing', True)
        self.tbs.set_config('pvc.tbs.declaration.period', PvcTbsLicense.DECLARATION_PERIOD_MONTHLY)

        # rule that always matches gives 10 eurie penalty
        self._make_rule('penalty',
            order_type = 'PvcContributionOrder',
            conditions = [{
                'condition': 'is_forecast',
                'operator': None, 'value_text': None, 'value_bool': None, 'value_int': 0,
                'value_decimal': None,
            }],
            actions = [{
                'action': 'fixed_amount_penalty',
                'description': '',
                'operator': None, 'value_text': None, 'value_bool': None, 'value_int': None,
                'value_decimal': 10,
                'tax_class_id': None, 'gl_account_code': None
            }]
        )

        # insert new declaration
        lines = [{
            'tbs_member_product': product,
            'quantity': 200,
            'weight': Decimal('15.3'),
            'custom_unit': None,
        }]
        order = cc._add_contribution_order(tm, lines, 2013, month=6)
        original = cc.submit_order(order)
        self.assertEqual(len(original.lines), 2)  # one for the declaration, one for the roel

        orders = cc.get()
        self.assertEqual(len(orders), 2)
        order_a, order_b = orders
        forecast = order_b
        if not forecast.is_forecast:
            forecast = order_a

        self.assertEqual(len(forecast.lines), 1)

    def test_forecast_from_actuals_with_rule_for_forecast(self):
        ma, tm = self._create_simple_member_for_declaration()
        cc = PvcContributionOrderController(self.sa_session, account_id=ma.id)
        product = self._get_member_product(tm)

        tm.tbs.set_config('pvc.tbs.advance.invoicing', True)
        tm.tbs.set_config('pvc.tbs.declaration.period', PvcTbsLicense.DECLARATION_PERIOD_MONTHLY)

        # rule that always matches gives 10 eurie penalty
        self._make_rule('penalty',
            order_type = 'PvcContributionOrder',
            conditions = [{
                'condition': 'is_forecast',
                'operator': None, 'value_text': None, 'value_bool': None, 'value_int': 1,
                'value_decimal': None,
            }],
            actions = [{
                'action': 'fixed_amount_penalty',
                'description': '',
                'operator': None, 'value_text': None, 'value_bool': None, 'value_int': None,
                'value_decimal': 10,
                'tax_class_id': None, 'gl_account_code': None
            }]
        )

        # insert new declaration
        lines = [{
            'tbs_member_product': product,
            'quantity': 200,
            'weight': Decimal('15.3'),
            'custom_unit': None,
        }]
        order = cc._add_contribution_order(tm, lines, 2013, month=6)
        original = cc.submit_order(order)
        self.assertEqual(len(original.lines), 1)

        orders = cc.get()
        self.assertEqual(len(orders), 2)
        order_a, order_b = orders
        forecast = order_b
        if not forecast.is_forecast:
            forecast = order_a

        self.assertEqual(len(forecast.lines), 2)  # the forecast should have an extra line

    def test_forecast_from_actuals_year_end(self):
        ma, tm = self._create_simple_member_for_declaration()

        tm.tbs.set_config('pvc.tbs.advance.invoicing', True)
        tm.tbs.set_config('pvc.tbs.declaration.period', PvcTbsLicense.DECLARATION_PERIOD_MONTHLY)

        pc = PvcProductController(self.sa_session, account_id=ma.id)
        cc = PvcContributionOrderController(self.sa_session, account_id=ma.id)
        product = self._get_member_product(tm)
        product2 = self._get_member_product(tm, tbs_product=self.tbs_product2)

        # insert new declaration
        lines = [{
            'tbs_member_product': product,
            'quantity': 234,
            'weight': Decimal('1024.23'),
            'custom_unit': None,
        }, {
            'tbs_member_product': product2,
            'quantity': 111,
            'weight': Decimal('542.22'),
            'custom_unit': None,
        }]
        order = cc._add_contribution_order(tm, lines, 2013, month=12)

        self.assertRaises(PublicSecurityException, cc.add_forecast_from_actuals, order)

        original = cc.submit_order(order)

        orders = cc.get()
        self.assertEqual(len(orders), 2)
        order_a, order_b = orders
        forecast = order_b
        if not forecast.is_forecast:
            forecast = order_a

        self.assertEqual(original.status, PvcContributionOrder.STATUS_SUBMITTED)
        self.assertEqual(forecast.status, PvcContributionOrder.STATUS_SUBMITTED)
        self.assertEqual(forecast.year, original.year+1)
        self.assertEqual(forecast.month, 1)
        self.assertEqual(forecast.quarter, None)
        self.assertEqual(len(forecast.lines), 2)

    def test_forecast_from_actuals_in_the_past(self):
        """
        If the next period does not have a declaration yet, but is in the past, we should not automatically add a new
        forecast order.
        :return:
        """
        ma, tm = self._create_simple_member_for_declaration()

        tm.tbs.set_config('pvc.tbs.advance.invoicing', True)
        tm.tbs.set_config('pvc.tbs.declaration.period', PvcTbsLicense.DECLARATION_PERIOD_MONTHLY)

        pc = PvcProductController(self.sa_session, account_id=ma.id)
        cc = PvcContributionOrderController(self.sa_session, account_id=ma.id)
        product = self._get_member_product(tm)
        product2 = self._get_member_product(tm, tbs_product=self.tbs_product2)

        # insert new declaration
        lines = [{
            'tbs_member_product': product,
            'quantity': 234,
            'weight': Decimal('1024.23'),
            'custom_unit': None,
        }, {
            'tbs_member_product': product2,
            'quantity': 111,
            'weight': Decimal('542.22'),
            'custom_unit': None,
        }]
        order = cc._add_contribution_order(tm, lines, 2013, month=1)

        set_date(date(2013, 12, 1)) # somewhere in the future (monthly declarations)

        self.assertRaises(PublicSecurityException, cc.add_forecast_from_actuals, order)

        original = cc.submit_order(order)

        orders = cc.get()
        self.assertEqual(len(orders), 1)
        self.assertFalse(orders[0].is_forecast)

    def test_forecast_from_actuals_resigning_member(self):
        ma, tm = self._create_simple_member_for_declaration()

        tm.tbs.set_config('pvc.tbs.advance.invoicing', True)
        tm.tbs.set_config('pvc.tbs.declaration.period', PvcTbsLicense.DECLARATION_PERIOD_MONTHLY)

        pc = PvcProductController(self.sa_session, account_id=ma.id)
        cc = PvcContributionOrderController(self.sa_session, account_id=ma.id)
        product = self._get_member_product(tm)
        product2 = self._get_member_product(tm, tbs_product=self.tbs_product2)

        # insert new declaration
        lines = [{
            'tbs_member_product': product,
            'quantity': 234,
            'weight': Decimal('1024.23'),
            'custom_unit': None,
        }, {
            'tbs_member_product': product2,
            'quantity': 111,
            'weight': Decimal('542.22'),
            'custom_unit': None,
        }]
        order = cc._add_contribution_order(tm, lines, 2013, month=12)

        self.assertRaises(PublicSecurityException, cc.add_forecast_from_actuals, order)

        tm.status = PvcTbsMember.STATUS_RESIGNING
        self.sa_session.add(tm)
        self.sa_session.flush()

        original = cc.submit_order(order)

        orders = cc.get()
        self.assertEqual(len(orders), 1)
        self.assertFalse(orders[0].is_forecast)

    def test_trigger_new_declarations_no_products(self):
        ma, tm = self._create_simple_member_for_declaration()
        tm.tbs.set_config('pvc.tbs.declaration.period', PvcTbsLicense.DECLARATION_PERIOD_MONTHLY)

        # its the first of june, we should get an order for the TM

        cc = PvcContributionOrderController(self.sa_session)
        orders = cc.trigger_new_contributions(tm.tbs)
        self.assertEqual(len(orders), 0)

    def test_trigger_new_declarations_monthly(self):
        ma, tm = self._create_simple_member_for_declaration()
        product = self._get_member_product(tm)
        tm.tbs.set_config('pvc.tbs.declaration.period', PvcTbsLicense.DECLARATION_PERIOD_MONTHLY)

        # its the first of june, we should get an order for the TM

        cc = PvcContributionOrderController(self.sa_session)
        orders = cc.trigger_new_contributions(tm.tbs)
        self.assertEqual(len(orders), 1)

        order = orders.pop()
        self.assertEqual(order.year, today().year)
        self.assertEqual(order.month, 5)

    def test_trigger_new_declarations_january(self):
        ma, tm = self._create_simple_member_for_declaration()
        product = self._get_member_product(tm)
        tm.tbs.set_config('pvc.tbs.declaration.period', PvcTbsLicense.DECLARATION_PERIOD_MONTHLY)

        set_date(date(2014, 1, 1))
        # its the first of june, we should get an order for the TM

        cc = PvcContributionOrderController(self.sa_session)
        orders = cc.trigger_new_contributions(tm.tbs)
        self.assertEqual(len(orders), 1)

        order = orders.pop()
        self.assertEqual(order.year, today().year-1)
        self.assertEqual(order.month, 12)

    def test_trigger_new_declarations_quarterly(self):
        ma, tm = self._create_simple_member_for_declaration()
        product = self._get_member_product(tm)
        tm.tbs.set_config('pvc.tbs.declaration.period', PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY)

        # its the first of june, we should not get a quarterly order

        cc = PvcContributionOrderController(self.sa_session)
        orders = cc.trigger_new_contributions(tm.tbs)
        self.assertEqual(len(orders), 0)

        # now we go to july
        set_date(date(2014, 7, 1))

        orders = cc.trigger_new_contributions(tm.tbs)
        order = orders.pop()
        self.assertEqual(order.year, today().year)
        self.assertEqual(order.quarter, 2)
        self.assertEqual(order.month, None)

    def test_trigger_new_declarations_forecast_orders(self):
        ma, tm = self._create_simple_member_for_declaration()

        self.tbs.set_config('pvc.tbs.advance.invoicing', True)
        self.tbs.set_config('pvc.tbs.declaration.period', PvcTbsLicense.DECLARATION_PERIOD_MONTHLY)

        pc = PvcProductController(self.sa_session, account_id=ma.id)
        cc = PvcContributionOrderController(self.sa_session, account_id=ma.id)
        product = self._get_member_product(tm)
        product2 = self._get_member_product(tm, tbs_product=self.tbs_product2)

        # insert new declaration
        lines = [{
            'tbs_member_product': product,
            'quantity': 234,
            'weight': Decimal('1024.23'),
            'custom_unit': None,
        }, {
            'tbs_member_product': product2,
            'quantity': 111,
            'weight': Decimal('542.22'),
            'custom_unit': None,
        }]
        order = cc._add_contribution_order(tm, lines, 2013, month=6)

        self.assertRaises(PublicSecurityException, cc.add_forecast_from_actuals, order)

        original = cc.submit_order(order)

        orders = cc.get()
        self.assertEqual(len(orders), 2)
        order_a, order_b = orders
        forecast = order_b
        if not forecast.is_forecast:
            forecast = order_a

        self.assertEqual(original.status, PvcContributionOrder.STATUS_SUBMITTED)
        self.assertEqual(forecast.status, PvcContributionOrder.STATUS_SUBMITTED)
        self.assertEqual(forecast.year, original.year)
        self.assertEqual(forecast.month, original.month+1)
        self.assertEqual(forecast.quarter, None)
        self.assertEqual(len(forecast.lines), 2)

        set_date(date(2013, 8, 1))  # advance a month

        cc = PvcContributionOrderController(self.sa_session)
        orders = cc.trigger_new_contributions(tm.tbs)
        self.assertEqual(len(orders), 1)

        order = orders.pop()
        self.assertEqual(order.year, today().year)
        self.assertEqual(order.quarter, None)
        self.assertEqual(order.month, original.month+1)
        self.assertEqual(order.id, forecast.id)
        self.assertEqual(order.resubmit_request, PvcContributionOrder.RESUBMIT_REQUEST_APPROVED)

        cc = PvcContributionOrderController(self.sa_session, account_id=ma.id)
        order = cc.update_contribution(order.id, {
            'line_0_quantity': 210,
            'line_0_weight': Decimal('16.44'),
            'line_1_quantity': 250,
            'line_1_weight': Decimal('11.11')
        })
        cn = order.parent_order
        line = cn.lines[0]
        self.assertEqual(line.gl_account_number, 'GLF0R3CAS7')

        self.assertEqual(order.lines[0].gl_account_number, 'GL789012')

    def test_trigger_new_declarations_with_brands(self):
        ma, tm = self._create_simple_member_for_declaration()

        self.tbs.set_config('pvc.tbs.contribution.split.by.brand', True)

        pc = PvcProductController(self.sa_session, account_id=ma.id)
        cc = PvcContributionOrderController(self.sa_session, account_id=ma.id)
        product = self._get_member_product(tm)
        product2 = self._get_member_product(tm, tbs_product=self.tbs_product2)

        # insert new declaration
        lines = [{
            'tbs_member_product': product,
            'quantity': 234,
            'weight': Decimal('1024.23'),
            'custom_unit': None,
            'brand': pc.get_or_save_brand('Philips', product),
        }, {
            'tbs_member_product': product,
            'quantity': 34,
            'weight': Decimal('32.54'),
            'custom_unit': None,
            'brand': pc.get_or_save_brand('Sony', product),
        }, {
            'tbs_member_product': product2,
            'quantity': 111,
            'weight': Decimal('542.22'),
            'custom_unit': None,
            'brand': pc.get_or_save_brand('Samsung', product),
        }]
        order = cc._add_contribution_order(tm, lines, 2013, month=3)
        self.assertEqual(len(order.lines), 3)
        cc.submit_order(order.id)

        # Next period.

        set_date(date(2013, 7, 1))

        acc = PvcContributionOrderController(self.sa_session)
        orders = acc.trigger_new_contributions(tm.tbs)
        self.assertEqual(len(orders), 1)
        order = orders.pop()
        self.assertEqual(len(order.lines), 3)

        brands = ["Philips", "Sony", "Samsung"]
        for line in order.lines:
            self.assertTrue(line.brand.name in brands)
            brands.remove(line.brand.name)

    def test_undeclared_period_yearly_to_monthly(self):

        ma, tm = self._create_simple_member_for_declaration()
        self._get_member_product(tm, self.tbs_product)  # add a tbs member product
        cc = PvcContributionOrderController(self.sa_session, account_id=ma.id)

        self.pc.save_contribution_fee(self.tbs_product.id, self.service_level.id, Decimal('0.015'), date(2012, 1, 1))

        # member is yearly, start date = 01-01-2012
        tm.declaration_periodicity = PvcTbsLicense.DECLARATION_PERIOD_YEARLY
        self.sa_session.add(tm)
        self.sa_session.flush()

        # go to 02-01-2014
        set_date(date(2014, 1, 2))

        # available periods should be 2012 & 2013
        periods = cc.get_undeclared_periods(tm)
        self.assertEqual(sorted(periods.keys()), [2012, 2013])

        # declare for 2012, 2013 yearly
        cc.add_open_contribution_order(tm, 2012, None, None)
        cc.add_open_contribution_order(tm, 2013, None, None)

        # member changes periodicity to monthly
        tm.declaration_periodicity = PvcTbsLicense.DECLARATION_PERIOD_MONTHLY
        self.sa_session.add(tm)
        self.sa_session.flush()
        self.sa_session.expire(tm)

        # set date to 01-06-2014
        set_date(date(2014, 6, 1))

        # periods available should be months 1,2,3,4 of 2014 (May is excluded as its the first of the month)
        periods = cc.get_undeclared_periods(tm)
        self.assertEqual(len(periods), 1)  # only declarations for 2014
        self.assertEqual(sorted(periods[2014]['months']), [1, 2, 3, 4])

    def test_undeclared_period_yearly_to_quarterly(self):

        ma, tm = self._create_simple_member_for_declaration()
        self._get_member_product(tm, self.tbs_product)  # add a tbs member product
        cc = PvcContributionOrderController(self.sa_session, account_id=ma.id)

        self.pc.save_contribution_fee(self.tbs_product.id, self.service_level.id, Decimal('0.015'), date(2012, 1, 1))

        # member is yearly, start date = 01-01-2012
        tm.declaration_periodicity = PvcTbsLicense.DECLARATION_PERIOD_YEARLY
        self.sa_session.add(tm)
        self.sa_session.flush()

        # go to 02-01-2014
        set_date(date(2014, 1, 2))

        # available periods should be 2012 & 2013
        periods = cc.get_undeclared_periods(tm)
        self.assertEqual(sorted(periods.keys()), [2012, 2013])

        # declare for 2012, 2013 yearly
        cc.add_open_contribution_order(tm, 2012, None, None)
        cc.add_open_contribution_order(tm, 2013, None, None)

        # member changes periodicity to monthly
        tm.declaration_periodicity = PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY
        self.sa_session.add(tm)
        self.sa_session.flush()
        self.sa_session.expire(tm)

        # set date to 01-06-2014
        set_date(date(2014, 7, 2))

        # periods available should be quarter 1 and 2 of 2014
        periods = cc.get_undeclared_periods(tm)
        self.assertEqual(len(periods), 1)  # only declarations for 2014
        self.assertEqual(sorted(periods[2014]['quarters']), [1, 2])

    def test_undeclared_period_quarterly_to_monthly(self):

        ma, tm = self._create_simple_member_for_declaration()
        self._get_member_product(tm, self.tbs_product)  # add a tbs member product
        cc = PvcContributionOrderController(self.sa_session, account_id=ma.id)

        self.pc.save_contribution_fee(self.tbs_product.id, self.service_level.id, Decimal('0.015'), date(2012, 1, 1))

        # member is yearly, start date = 01-01-2012
        tm.declaration_periodicity = PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY
        self.sa_session.add(tm)
        self.sa_session.flush()

        # go to 02-01-2014
        set_date(date(2014, 1, 2))

        # available periods should be 2012 & 2013
        periods = cc.get_undeclared_periods(tm)
        self.assertEqual(sorted(periods.keys()), [2012, 2013])

        # declare for 2012, 2013 quarterly
        cc.add_open_contribution_order(tm, 2012, None, 1)
        cc.add_open_contribution_order(tm, 2012, None, 2)
        cc.add_open_contribution_order(tm, 2012, None, 3)
        cc.add_open_contribution_order(tm, 2012, None, 4)
        cc.add_open_contribution_order(tm, 2013, None, 1)
        cc.add_open_contribution_order(tm, 2013, None, 2)
        cc.add_open_contribution_order(tm, 2013, None, 3)

        # member changes periodicity to monthly
        tm.declaration_periodicity = PvcTbsLicense.DECLARATION_PERIOD_MONTHLY
        self.sa_session.add(tm)
        self.sa_session.flush()
        self.sa_session.expire(tm)

        # set date to 01-06-2014
        set_date(date(2014, 6, 1))

        # periods available should be months 1,2,3,4 of 2014 (May is excluded as its the first of the month)
        periods = cc.get_undeclared_periods(tm)
        self.assertEqual(len(periods), 2)  # last months of 2013 and first months of 2014
        self.assertEqual(sorted(periods[2013]['months']), [10, 11, 12])
        self.assertEqual(sorted(periods[2014]['months']), [1, 2, 3, 4])

    def test_close_contribution_non_compliant(self):
        ma, tm = self._create_simple_member_for_declaration()
        self._get_member_product(tm)

        dc = PvcContributionOrderController(self.sa_session, account_id=ma.id)
        d = dc.add_open_declaration(tm, 2013, month=1, quarter=None)

        pc = PvcContributionOrderController(self.sa_session)
        set_date(date(2013, 2, 11))
        pc.close_contributions(tm.tbs)

        self.sa_session.expire(d)
        self.assertEqual(d.status, PvcContributionOrder.STATUS_SUBMITTED)
        self.assertEqual(d.is_compliant, PvcContributionOrder.IS_NON_COMPLIANT)

    def test_close_contribution_compliant(self):
        ma, tm = self._create_simple_member_for_declaration()
        self._get_member_product(tm)

        dc = PvcContributionOrderController(self.sa_session, account_id=ma.id)
        d = dc.add_open_declaration(tm, 2013, month=1, quarter=None)

        d = dc.update_contribution(d.id, {'line_0_quantity': 210, 'line_0_weight': Decimal('16.44')})
        self.assertEqual(d.status, PvcContributionOrder.STATUS_PENDING)

        pc = PvcContributionOrderController(self.sa_session)
        set_date(date(2013, 2, 11))
        pc.close_contributions(tm.tbs)

        self.sa_session.expire(d)
        self.assertEqual(len(d.lines), 1)
        self.assertEqual(d.status, PvcContributionOrder.STATUS_SUBMITTED)
        self.assertEqual(d.is_compliant, PvcContributionOrder.IS_COMPLIANT)

    def test_close_contribution_different_closing_day(self):
        ma, tm = self._create_simple_member_for_declaration()
        self._get_member_product(tm)
        tm.declaration_closing_day = 15
        self.sa_session.add(tm)
        self.sa_session.flush()

        dc = PvcContributionOrderController(self.sa_session, account_id=ma.id)
        d = dc.add_open_declaration(tm, 2013, month=1, quarter=None)

        pc = PvcContributionOrderController(self.sa_session)
        set_date(date(2013, 2, 11))
        pc.close_contributions(tm.tbs)

        self.sa_session.expire(d)
        self.assertEqual(d.status, PvcContributionOrder.STATUS_NEW)

        set_date(date(2013, 2, 16))
        pc.close_contributions(tm.tbs)

        self.sa_session.expire(d)
        self.assertEqual(d.status, PvcContributionOrder.STATUS_SUBMITTED)
        self.assertEqual(d.is_compliant, PvcContributionOrder.IS_NON_COMPLIANT)

    # - subrogation test cases - #
    def _create_commissioner(self, obligation_start_date=None):
        ma = self._create_member_account(company_name='commissioner')

        # apply at TBS
        tm = self._member_apply_at_tbs(ma, self.tbs_account, self.service_level, is_commissioner=True)

        obligation_start_date = date(2012, 1, 1) if obligation_start_date is None else obligation_start_date

        # activate member to automatically create an order entry for the Administration Fee
        mc = PvcMemberController(self.sa_session, account_id=self.tbs_account.id)
        mc.save_tbs_member(tm.id, {
            'status': PvcTbsMember.STATUS_APPROVED,
            'is_full_member': False,
            'is_representative': False,
            'obligation_start_date': obligation_start_date})
        return ma, tm

    def _create_subrogation(self, ma, tm, ctm):
        controller = PvcSubrogationController(self.sa_session, account_id=ma.id)
        subrogation = controller.request_for_subrogation(tm, 'commissioner_name')
        # Set subrogation status active.
        subrogation.status = PvcSubrogation.STATUS_ACTIVE
        subrogation.commissioner = ctm
        self.sa_session.add(subrogation)
        self.sa_session.flush()
        return subrogation

    def test_subrogate_line(self):
        ma, tm = self._create_simple_member_for_declaration()
        self._get_member_product(tm)
        self._get_member_product(tm, self.tbs_product2)

        cma, ctm = self._create_commissioner()

        subrogation = self._create_subrogation(ma, tm, ctm)

        set_date(date(2013, 7, 1))

        pc = PvcContributionOrderController(self.sa_session)
        orders = pc.trigger_new_contributions(tm.tbs)
        self.assertEqual(len(orders), 1)
        d = orders.pop()

        dc = PvcContributionOrderController(self.sa_session, account_id=ma.id)

        declarations = dc.update_contribution(d.id, {
            'line_0_quantity': 225,
            'line_0_weight': Decimal('200.00'),
            'line_1_quantity': 175,
            'line_1_weight': Decimal('365.75'),
        }, subrogate={'line_1': subrogation.id})
        self.assertEqual(len(declarations), 2)

        subrogated = declarations[0]
        original = declarations[1]

        # both should have status pending
        self.assertEqual(subrogated.status, PvcContributionOrder.STATUS_PENDING)
        self.assertEqual(original.status, PvcContributionOrder.STATUS_PENDING)
        # both should be compliant
        self.assertEqual(subrogated.is_compliant, PvcContributionOrder.IS_COMPLIANT)
        self.assertEqual(original.is_compliant, PvcContributionOrder.IS_COMPLIANT)

        self.assertEqual(len(subrogated.lines), 1)
        self.assertEqual(len(original.lines), 1)

        sline = subrogated.lines[0]
        self.assertEqual(sline.quantity, 175)
        self.assertEqual(sline.weight, Decimal('365.75'))

        line = original.lines[0]
        self.assertEqual(line.quantity, 225)
        self.assertEqual(line.weight, Decimal('200.00'))

    def test_subrogate_line_after_resubmit(self):
        ma, tm = self._create_simple_member_for_declaration()
        self._get_member_product(tm)
        self._get_member_product(tm, self.tbs_product2)

        cma, ctm = self._create_commissioner()

        subrogation = self._create_subrogation(ma, tm, ctm)

        set_date(date(2013, 7, 1))

        pc = PvcContributionOrderController(self.sa_session)
        orders = pc.trigger_new_contributions(tm.tbs)
        self.assertEqual(len(orders), 1)
        d = orders.pop()

        dc = PvcContributionOrderController(self.sa_session, account_id=ma.id)

        d = dc.update_contribution(d.id, {
            'line_0_quantity': 225,
            'line_0_weight': Decimal('200.00'),
            'line_1_quantity': 175,
            'line_1_weight': Decimal('365.75'),
        })

        dc.submit_order(d.id)
        d.resubmit_request = PvcContributionOrder.RESUBMIT_REQUEST_APPROVED
        self.sa_session.add(d)
        self.sa_session.flush()

        declarations = dc.update_contribution(d.id, {
            'line_0_quantity': 225,
            'line_0_weight': Decimal('200.00'),
            'line_1_quantity': 234,
            'line_1_weight': Decimal('500.22'),
        }, subrogate={'line_1': subrogation.id})
        self.assertEqual(len(declarations), 2)

        subrogated = declarations[0]
        original = declarations[1]

        # both should have status pending
        self.assertEqual(subrogated.status, PvcContributionOrder.STATUS_PENDING)
        self.assertEqual(original.status, PvcContributionOrder.STATUS_PENDING)
        # both should be compliant
        self.assertEqual(subrogated.is_compliant, PvcContributionOrder.IS_COMPLIANT)
        self.assertEqual(original.is_compliant, PvcContributionOrder.IS_COMPLIANT)

        self.assertEqual(len(subrogated.lines), 1)
        self.assertEqual(len(original.lines), 1)

        sline = subrogated.lines[0]
        self.assertEqual(sline.quantity, 234)
        self.assertEqual(sline.weight, Decimal('500.22'))

        line = original.lines[0]
        self.assertEqual(line.quantity, 225)
        self.assertEqual(line.weight, Decimal('200.00'))

    def test_subrogate_line_roe(self):
        ma, tm = self._create_simple_member_for_declaration()

        self.reset_roe()

        self._get_member_product(tm)
        self._get_member_product(tm, self.tbs_product2)

        tm.tbs.country = None
        self.sa_session.add(tm.tbs)
        self.sa_session.flush()

        cma, ctm = self._create_commissioner()

        subrogation = self._create_subrogation(ma, tm, ctm)

        set_date(date(2013, 7, 1))

        pc = PvcContributionOrderController(self.sa_session)
        orders = pc.trigger_new_contributions(tm.tbs)
        self.assertEqual(len(orders), 1)
        d = orders.pop()

        dc = PvcContributionOrderController(self.sa_session, account_id=ma.id)

        declarations = dc.update_contribution(d.id, {
            'line_0_quantity': 225,
            'line_0_weight': Decimal('200.00'),
            'line_1_quantity': 175,
            'line_1_weight': Decimal('365.75'),
        }, subrogate={'line_1': subrogation.id})
        self.assertEqual(len(declarations), 2)

        subrogated = declarations[0]
        original = declarations[1]

        # both should have status pending
        self.assertEqual(subrogated.status, PvcContributionOrder.STATUS_PENDING)
        self.assertEqual(original.status, PvcContributionOrder.STATUS_PENDING)
        # both should be compliant
        self.assertEqual(subrogated.is_compliant, PvcContributionOrder.IS_COMPLIANT)
        self.assertEqual(original.is_compliant, PvcContributionOrder.IS_COMPLIANT)

        self.assertEqual(len(subrogated.lines), 1)
        self.assertEqual(len(original.lines), 1)

        sline = subrogated.lines[0]
        self.assertEqual(sline.quantity, 175)
        self.assertEqual(sline.weight, Decimal('365.75'))

        line = original.lines[0]
        self.assertEqual(line.quantity, 225)
        self.assertEqual(line.weight, Decimal('200.00'))

    def test_product_activation_and_deactivation(self):
        ma, tm = self._create_simple_member_for_declaration()
        mp1 = self._get_member_product(tm)
        mp2 = self._get_member_product(tm, self.tbs_product2)

        cma, ctm = self._create_commissioner()

        subrogation = self._create_subrogation(ma, tm, ctm)

        # first we try to disable the product with the commissioner (note, the product is not part of the subrogation yet)
        sc = PvcSubrogationController(self.sa_session, account_id=cma.id)
        self.assertRaises(PublicSecurityException, sc.deactivate_product, mp1.id)

        set_date(date(2013, 7, 1))

        pc = PvcContributionOrderController(self.sa_session)
        orders = pc.trigger_new_contributions(tm.tbs)
        self.assertEqual(len(orders), 1)
        d = orders.pop()

        dc = PvcContributionOrderController(self.sa_session, account_id=ma.id)

        declarations = dc.update_contribution(d.id, {
            'line_0_quantity': 225,
            'line_0_weight': Decimal('200.00'),
            'line_1_quantity': 175,
            'line_1_weight': Decimal('365.75'),
        }, subrogate={'line_1': subrogation.id})
        self.assertEqual(len(declarations), 2)

        subrogated = declarations[0]
        original = declarations[1]

        smp1 = subrogated.lines[0].tbs_member_product
        self.assertTrue(smp1.is_active)
        self.assertEqual(smp1.tbs_product, mp2.tbs_product)

        # cant activate an already active product
        self.assertRaises(PublicSecurityException, sc.activate_product, smp1.id)

        # try to deactivate it with the producer
        osc = PvcSubrogationController(self.sa_session, account_id=ma.id)
        self.assertRaises(PublicSecurityException, osc.deactivate_product, smp1.id)

        # deactivate it
        sc.deactivate_product(smp1.id)

        # try to deactivate it again
        self.assertRaises(PublicSecurityException, sc.deactivate_product, smp1.id)

        # reactivate it with producer
        self.assertRaises(PublicSecurityException, osc.activate_product, smp1.id)

        # reactivate it with commissioner
        sc.activate_product(smp1.id)

    def test_subrogation_with_forecast(self):
        # create a member for declaration
        ma, tm = self._create_simple_member_for_declaration()

        # set TBS config for advance invoicing.
        self.tbs.set_config('pvc.tbs.advance.invoicing', True)
        self.tbs.set_config('pvc.tbs.declaration.period', PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY)
        cc = PvcContributionOrderController(self.sa_session, account_id=ma.id)

        products = [self.tbs_product.id, self.tbs_product2.id]

        order = cc.add_forecast_order(tm.id, products)
        self.assertEqual(len(order.lines), 2)

        d = cc.update_contribution(order.id, {
            'line_0_quantity': 225,
            'line_0_weight': Decimal('200.00'),
            'line_1_quantity': 750,
            'line_1_weight': Decimal('650.00'),
        })

        cc.submit_order(d.id)

        # make subrogation
        cma, ctm = self._create_commissioner()

        subrogation = self._create_subrogation(ma, tm, ctm)

        # go to quarter 3 so quarter two can be resubmited
        set_date(date(2013, 7, 1))

        cc.trigger_new_contributions(tm.tbs)
        declarations = cc.resubmit_contribution(d, {
            'line_0_quantity': 225,
            'line_0_weight': Decimal('200.00'),
            'line_1_quantity': 1250,
            'line_1_weight': Decimal('550.00'),
        }, subrogate={'line_1': subrogation.id})

        # this should both trigger a forecast order
        cc.submit_order(declarations[0].id)
        cc.submit_order(declarations[1].id)

        forecast = self.sa_session.query(PvcContributionOrder).filter(and_(
            PvcContributionOrder.tbs_member==ctm,
            PvcContributionOrder.is_forecast==True
        )).first()
        line = forecast.lines[0]
        self.assertEqual(line.quantity, 1250)
        self.assertEqual(line.weight, Decimal('550.00'))

    def test_subrogation_with_brands(self):
        ma, tm = self._create_simple_member_for_declaration()
        tm.tbs.set_config('pvc.tbs.contribution.split.by.brand', True)
        mp = self._get_member_product(tm)
        mp1 = self._get_member_product(tm, self.tbs_product2)

        pc = PvcProductController(self.sa_session, account_id=ma.id)

        pc.get_or_save_brand("Philips", mp)
        pc.get_or_save_brand("Sony", mp1)

        cma, ctm = self._create_commissioner()

        subrogation = self._create_subrogation(ma, tm, ctm)

        set_date(date(2013, 7, 1))

        pc = PvcContributionOrderController(self.sa_session)
        orders = pc.trigger_new_contributions(tm.tbs)
        self.assertEqual(len(orders), 1)
        d = orders.pop()

        dc = PvcContributionOrderController(self.sa_session, account_id=ma.id)

        declarations = dc.update_contribution(d.id, {
            'line_0_quantity': 225,
            'line_0_brand': 'Producer brand',
            'line_0_weight': Decimal('200.00'),
            'line_1_quantity': 175,
            'line_1_brand': 'Subrogated brand',
            'line_1_weight': Decimal('365.75'),
        }, subrogate={'line_1': subrogation.id})
        self.assertEqual(len(declarations), 2)

        subrogated = declarations[0]
        original = declarations[1]
        self.assertEqual(subrogated.lines[0].brand.name, 'Subrogated brand')
        self.assertEqual(original.lines[0].brand.name, 'Producer brand')

    def test_subrogation_with_b2b_split(self):
        ma, tm = self._create_simple_member_for_declaration()
        tm.tbs.set_config('pvc.tbs.require.end.user.distinction', True)
        self._get_member_product(tm)
        self._get_member_product(tm, self.tbs_product2)

        cma, ctm = self._create_commissioner()

        subrogation = self._create_subrogation(ma, tm, ctm)

        set_date(date(2013, 7, 1))

        pc = PvcContributionOrderController(self.sa_session)
        orders = pc.trigger_new_contributions(tm.tbs)
        self.assertEqual(len(orders), 1)
        d = orders.pop()

        dc = PvcContributionOrderController(self.sa_session, account_id=ma.id)

        declarations = dc.update_contribution(d.id, {
            'line_0_quantity': 225,
            'line_0_is_b2b': False,
            'line_0_weight': Decimal('200.00'),
            'line_1_quantity': 225,
            'line_1_is_b2b': True,
            'line_1_weight': Decimal('200.00'),
            'line_2_quantity': 175,
            'line_2_is_b2b': False,
            'line_2_weight': Decimal('365.75'),
            'line_3_quantity': 175,
            'line_3_is_b2b': True,
            'line_3_weight': Decimal('365.75'),
        }, subrogate={'line_2': subrogation.id, 'line_3': subrogation.id})
        self.assertEqual(len(declarations), 2)

        subrogated = declarations[0]
        original = declarations[1]
        self.assertEqual(subrogated.lines[0].is_b2b, True)
        self.assertEqual(subrogated.lines[1].is_b2b, False)
        self.assertEqual(original.lines[0].is_b2b, False)
        self.assertEqual(original.lines[1].is_b2b, True)

    def test_subrogation_with_automatic_b2b_split(self):
        ma, tm = self._create_simple_member_for_declaration()
        tm.tbs.set_config('pvc.tbs.contribution.automatic.end.user.distinction', True)

        self._make_rule('italy_b2b_split',
            order_type = 'PvcContributionOrder',
            conditions = [{
                'condition': 'merchant',
                'operator': None, 'value_text': None, 'value_bool': None, 'value_int': tm.tbs.id,
                'value_decimal': None,
            }],
            actions = [{
                'action': 'pvc_contribution_b2b_percentage',
                'description': '',
                'operator': None, 'value_text': None, 'value_bool': None, 'value_int': 10,
                'value_decimal': None,
                'tax_class_id': None, 'gl_account_code': None
            }]
        )

        self._get_member_product(tm)
        self._get_member_product(tm, self.tbs_product2)

        cma, ctm = self._create_commissioner()

        subrogation = self._create_subrogation(ma, tm, ctm)

        set_date(date(2013, 7, 1))

        pc = PvcContributionOrderController(self.sa_session)
        orders = pc.trigger_new_contributions(tm.tbs)
        self.assertEqual(len(orders), 1)
        d = orders.pop()

        dc = PvcContributionOrderController(self.sa_session, account_id=ma.id)

        declarations = dc.update_contribution(d.id, {
            'line_0_quantity': 225,
            'line_0_weight': Decimal('200.00'),
            'line_1_quantity': 225,
            'line_1_weight': Decimal('200.00'),
        }, subrogate={'line_1': subrogation.id})
        self.assertEqual(len(declarations), 2)
        subrogated = declarations[0]
        original = declarations[1]
        self.assertEqual(len(subrogated.lines), 1)
        self.assertEqual(len(original.lines), 1)

        subrogated = dc.submit_order(subrogated.id)
        original = dc.submit_order(original.id)

        self.assertEqual(len(subrogated.lines), 2)
        self.assertEqual(len(original.lines), 2)

    def test_subrogation_with_custom_units(self):
        ma, tm = self._create_simple_member_for_declaration()

        products = self.pc.get_available_products_for_tbs(tm.tbs.id)

        tbs_product_custom_unit = self.pc.save_or_update_tbs_product(None, 'EEE-TBS-TEST-4', 'WEEE-3', tm.tbs.id,
            parent_level_id=products[1]['id'], basis_for_invoicing=PvcTbsProduct.INVOICE_BASED_ON_CUSTOM_UNIT, custom_unit=PvcTbsProduct.CUSTOM_UNIT_NOMINAL_POWER
        )

        self.pc.save_contribution_fee(tbs_product_custom_unit.id, self.service_level.id, Decimal('0.017'), date(2013, 1, 1))

        self._get_member_product(tm)
        self._get_member_product(tm, tbs_product_custom_unit)

        cma, ctm = self._create_commissioner()

        subrogation = self._create_subrogation(ma, tm, ctm)

        set_date(date(2013, 7, 1))

        pc = PvcContributionOrderController(self.sa_session)
        orders = pc.trigger_new_contributions(tm.tbs)
        self.assertEqual(len(orders), 1)
        d = orders.pop()

        dc = PvcContributionOrderController(self.sa_session, account_id=ma.id)

        declarations = dc.update_contribution(d.id, {
            'line_0_quantity': 225,
            'line_0_custom_unit': None,
            'line_0_weight': Decimal('200.00'),
            'line_1_quantity': 225,
            'line_1_custom_unit': Decimal('12.23'),
            'line_1_weight': Decimal('200.00'),
        }, subrogate={'line_1': subrogation.id})
        self.assertEqual(len(declarations), 2)

        subrogated = declarations[0]
        original = declarations[1]

        self.assertEqual(subrogated.lines[0].custom_unit, Decimal('12.23'))
        self.assertEqual(original.lines[0].custom_unit, None)

    def test_subrogation_trigger_new_declarations(self):
        ma, tm = self._create_simple_member_for_declaration()

        self._get_member_product(tm)
        self._get_member_product(tm, self.tbs_product2)

        cma, ctm = self._create_commissioner()

        subrogation = self._create_subrogation(ma, tm, ctm)

        set_date(date(2013, 7, 1))

        pc = PvcContributionOrderController(self.sa_session)
        orders = pc.trigger_new_contributions(tm.tbs)
        self.assertEqual(len(orders), 1)
        d = orders.pop()

        dc = PvcContributionOrderController(self.sa_session, account_id=ma.id)

        declarations = dc.update_contribution(d.id, {
            'line_0_quantity': 225,
            'line_0_weight': Decimal('200.00'),
            'line_1_quantity': 225,
            'line_1_weight': Decimal('200.00'),
        }, subrogate={'line_1': subrogation.id})
        self.assertEqual(len(declarations), 2)

        subrogated = declarations[0]

        self.assertEqual(subrogated.lines[0].tbs_member_product.subrogation, subrogation)

        set_date(date(2013, 10, 1))

        pc = PvcContributionOrderController(self.sa_session)
        orders = pc.trigger_new_contributions(tm.tbs)
        self.assertEqual(len(orders), 2)

    def test_subrogation_trigger_new_declarations_by_brand(self):
        ma, tm = self._create_simple_member_for_declaration()

        tm.tbs.set_config('pvc.tbs.contribution.split.by.brand', True)

        pc = PvcProductController(self.sa_session, account_id=ma.id)

        mp = self._get_member_product(tm)
        mp1 = self._get_member_product(tm, self.tbs_product2)
        pc.get_or_save_brand("Philips", mp)
        b1 = pc.get_or_save_brand("Sony", mp1)

        cma, ctm = self._create_commissioner()

        subrogation = self._create_subrogation(ma, tm, ctm)

        set_date(date(2013, 7, 1))

        pc = PvcContributionOrderController(self.sa_session)
        orders = pc.trigger_new_contributions(tm.tbs)
        self.assertEqual(len(orders), 1)
        d = orders.pop()

        dc = PvcContributionOrderController(self.sa_session, account_id=ma.id)

        declarations = dc.update_contribution(d.id, {
            'line_0_quantity': 225,
            'line_0_brand': 'Philips',
            'line_0_weight': Decimal('200.00'),
            'line_1_quantity': 225,
            'line_1_brand': 'Sony',
            'line_1_weight': Decimal('200.00'),
        }, subrogate={'line_1': subrogation.id})
        self.assertEqual(len(declarations), 2)

        subrogated = declarations[0]

        self.assertEqual(subrogated.lines[0].tbs_member_product.subrogation, subrogation)
        self.assertEqual(subrogated.lines[0].tbs_member_product.brands[0].brand, b1)

        set_date(date(2013, 10, 1))

        pc = PvcContributionOrderController(self.sa_session)
        orders = pc.trigger_new_contributions(tm.tbs)
        self.assertEqual(len(orders), 2)

    def test_subrogation_resubmission(self):
        # create a member for declaration
        ma, tm = self._create_simple_member_for_declaration()
        self._get_member_product(tm)
        self._get_member_product(tm, self.tbs_product2)

        cma, ctm = self._create_commissioner()

        subrogation = self._create_subrogation(ma, tm, ctm)

        set_date(date(2013, 7, 1))

        pc = PvcContributionOrderController(self.sa_session)
        orders = pc.trigger_new_contributions(tm.tbs)
        self.assertEqual(len(orders), 1)
        d = orders.pop()

        dc = PvcContributionOrderController(self.sa_session, account_id=ma.id)

        declarations = dc.update_contribution(d.id, {
            'line_0_quantity': 225,
            'line_0_weight': Decimal('200.00'),
            'line_1_quantity': 175,
            'line_1_weight': Decimal('365.75'),
        }, subrogate={'line_1': subrogation.id})
        self.assertEqual(len(declarations), 2)

        subrogated = declarations[0]
        original = declarations[1]

        dc.submit_order(subrogated.id)

        cc = PvcContributionOrderController(self.sa_session, account_id=cma.id)
        # try to do the resubmit request by the commissioner
        self.assertRaises(PublicUserException, cc.request_for_resubmission, subrogated.id)

        dc.request_for_resubmission(subrogated.id)

        self.sa_session.expire(subrogated)
        self.assertEqual(subrogated.resubmit_request, PvcContributionOrder.RESUBMIT_REQUEST_PENDING)

        # now the operator approves the resubmit.
        cod = PvcContributionOrderController(self.sa_session, account_id=self.operator_account.id)
        cod.approve_resubmit_request(subrogated.id)

        # now that we have an approved resubmit request on a submitted declaration, an update on this declaration will do a resubmit.
        resubmit = dc.update_contribution(subrogated.id, {'line_0_quantity': 180, 'line_0_weight': Decimal('77.4')})
        self.sa_session.expire(subrogated)
        self.assertEqual(subrogated.resubmit_request, None)

        self.assertEqual(resubmit.producer, tm)
        self.assertEqual(resubmit.tbs_member, ctm)

    def test_open_contribution_order_fee_change(self):
        ma, tm = self._create_simple_member_for_declaration()
        mp1 = self._get_member_product(tm)
        mp2 = self._get_member_product(tm, self.tbs_product2)

        cc = PvcContributionOrderController(self.sa_session)
        orders = cc.trigger_new_contributions(tm.tbs)
        order = orders[0]

        from datetime import datetime

        pc = PvcProductController(self.sa_session, account_id=self.operator_account.id)
        pf = pc._save_or_update_product_fee(self.tbs_product,
            self.service_level.id, datetime.combine(date(2013, 1, 1), datetime.min.time()),
            Decimal('0.75'), PvcTbsProductFee.DIMENSION_NONE)

        self.assertTrue(pf.process_status, PvcTbsProductFee.STATUS_PROCESS)

        affected = cc.update_with_new_fee(pf)
        self.assertEqual(len(affected), 1)
        affect = affected[0]
        self.assertEqual(affect.id, order.id)

        self.assertEqual(affect.lines[0].price_per_unit, Decimal('0.75'))
        self.assertEqual(affect.lines[1].price_per_unit, Decimal('0.035'))

    def test_closed_contribution_order_fee_change(self):
        ma, tm = self._create_simple_member_for_declaration()
        mp1 = self._get_member_product(tm)
        mp2 = self._get_member_product(tm, self.tbs_product2)

        cc = PvcContributionOrderController(self.sa_session)
        orders = cc.trigger_new_contributions(tm.tbs)
        order = orders[0]

        set_date(date(2013, 6, 11))

        cc.close_contributions(tm.tbs)

        from datetime import datetime

        pc = PvcProductController(self.sa_session, account_id=self.operator_account.id)
        pf = pc._save_or_update_product_fee(self.tbs_product,
            self.service_level.id, datetime.combine(date(2013, 1, 1), datetime.min.time()),
            Decimal('0.75'), PvcTbsProductFee.DIMENSION_NONE)

        self.assertTrue(pf.process_status, PvcTbsProductFee.STATUS_PROCESS)

        affected = cc.update_with_new_fee(pf)
        self.assertEqual(len(affected), 0)

        self.sa_session.expire(order)
        self.assertEqual(order.lines[0].price_per_unit, Decimal('0.015'))
        self.assertEqual(order.lines[1].price_per_unit, Decimal('0.035'))

    def test_open_contribution_fee_change_with_b2b_split(self):
        ma, tm = self._create_simple_member_for_declaration()

        tm.tbs.set_config('pvc.tbs.require.end.user.distinction', True)

        self.pc.save_contribution_fee(self.tbs_product.id, self.service_level.id, Decimal('1.015'), date(2013, 1, 1), PvcTbsProductFee.DIMENSION_IS_B2C)
        self.pc.save_contribution_fee(self.tbs_product2.id, self.service_level.id, Decimal('1.035'), date(2013, 1, 1), PvcTbsProductFee.DIMENSION_IS_B2C)
        self.pc.save_contribution_fee(self.tbs_product.id, self.service_level.id, Decimal('2.015'), date(2013, 1, 1), PvcTbsProductFee.DIMENSION_IS_B2B)
        self.pc.save_contribution_fee(self.tbs_product2.id, self.service_level.id, Decimal('2.035'), date(2013, 1, 1), PvcTbsProductFee.DIMENSION_IS_B2B)

        mp1 = self._get_member_product(tm)
        mp2 = self._get_member_product(tm, self.tbs_product2)

        cc = PvcContributionOrderController(self.sa_session)
        orders = cc.trigger_new_contributions(tm.tbs)
        order = orders[0]

        self.assertEqual(len(order.lines), 4)

        pf = self.pc.save_contribution_fee(self.tbs_product.id, self.service_level.id, Decimal('8.655'), date(2013, 1, 1), PvcTbsProductFee.DIMENSION_IS_B2B)

        self.assertTrue(pf.process_status, PvcTbsProductFee.STATUS_PROCESS)

        affected = cc.update_with_new_fee(pf)
        self.assertEqual(len(affected), 1)
        affect = affected[0]
        self.assertEqual(affect.id, order.id)

        self.assertEqual(affect.lines[0].price_per_unit, Decimal('1.015'))
        self.assertEqual(affect.lines[1].price_per_unit, Decimal('8.655'))
        self.assertEqual(affect.lines[2].price_per_unit, Decimal('1.035'))
        self.assertEqual(affect.lines[3].price_per_unit, Decimal('2.035'))

    def test_resubmit_with_fee_change_for_period(self):
        ma, tm = self._create_simple_member_for_declaration()

        product = self._get_member_product(tm)
        cc = PvcContributionOrderController(self.sa_session, account_id=ma.id)

        # insert new declaration
        lines = [{
            'tbs_member_product': product,
            'quantity': 200,
            'weight': Decimal('15.3'),
            'custom_unit': None,
        }]
        order = cc._add_contribution_order(tm, lines, 2013, month=3)

        # try to do a request for resubmission, this should fail as the declaration is still outstanding
        self.assertRaises(PublicUserException, cc.request_for_resubmission, order.id)

        # preview declaration should include penalty
        order_state = cc.load_order_state(order.id, submit_dry_run=True)
        self.assertEquals(order_state['total_amount_excluding_vat'], Decimal('0.2295') )

        # submit the declaration (for export)
        order = cc.submit_order(order.id)

        # try to resubmit the declaration, this should trigger an error
        self.assertRaises(PublicUserException, cc.update_contribution, order.id, {'line_0_quantity': 225, 'line_0_weight': Decimal('15.3')})

        # make a request for resubmission
        cc.request_for_resubmission(order.id)

        self.sa_session.expire(order)
        self.assertEqual(order.resubmit_request, PvcContributionOrder.RESUBMIT_REQUEST_PENDING)

        # now the operator approves the resubmit.
        cod = PvcContributionOrderController(self.sa_session, account_id=self.operator_account.id)
        cod.approve_resubmit_request(order.id)

        # update the fee for march 2013
        pf = self.pc.save_contribution_fee(self.tbs_product.id, self.service_level.id, Decimal('125.95'), date(2013, 3, 1))

        # now that we have an approved resubmit request on a submitted declaration, an update on this declaration will do a resubmit.
        resubmit = cc.update_contribution(order.id, {'line_0_quantity': 225, 'line_0_weight': Decimal('15.4')})
        self.sa_session.expire(order)
        self.assertEqual(order.resubmit_request, None)

        # the resubmit's parent is the credited order.
        credit = resubmit.parent_order
        self.assertEqual(credit.parent_order, order) # our initial order should be the same as the first submit
        self.assertEqual(credit.lines[0].quantity, -200)
        self.assertEqual(credit.lines[0].weight, Decimal('-15.3'))
        self.assertEqual(credit.lines[0].price_per_unit, Decimal('0.015'))

        # the new resubmit should contain the new fee
        self.assertEqual(resubmit.lines[0].quantity, 225)
        self.assertEqual(resubmit.lines[0].weight, Decimal('15.4'))
        self.assertEqual(resubmit.lines[0].price_per_unit, Decimal('125.95'))
        self.assertEqual(resubmit.total_amount_excluding_vat, Decimal('1939.63'))
        self.assertEquals(len(resubmit.lines), 1)

    def test_member_resubmit_previous_service_level(self):

        set_date(date(2015, 1, 01))

        service_level_2 = self._create_tbs_service_level(self.tbs_account, name='SL B', year=today().year, admin_fee=Decimal('300.00'), level=2)

        self.pc.save_contribution_fee(self.tbs_product.id, self.service_level.id, Decimal('0.010'), today())
        self.pc.save_contribution_fee(self.tbs_product.id, service_level_2.id, Decimal('0.020'), today())

        member_account, tbs_member = self._create_simple_member_for_declaration(obligation_start_date=today())

        # in june, member requests to change service level
        set_date(date(2015, 6, 1))
        upload_contract = self._create_uploaded_file('upload_contract.pdf', 'upload_contract contents')
        c = PvcMemberController(sa_session=self.sa_session, account_id=member_account.id)
        member_service_level = c.request_change_service_level(tbs_member.id, service_level_2.id, upload_contract)

        # operator approves the request
        c_o = PvcMemberController(sa_session=self.sa_session, account_id=self.operator_account.id)
        c_o.approve_change_service_level(member_service_level.id)

        # goto july cuz then the new SL kicks in
        set_date(date(2015, 7, 1))
        self.assertEqual(tbs_member.active_service_level, member_service_level)

        product = self._get_member_product(tbs_member, self.tbs_product)
        cc = PvcContributionOrderController(self.sa_session, account_id=member_account.id)

        # declaration for january.. sl1 should be active
        lines = [{
            'tbs_member_product': product,
            'quantity': 10,
            'weight': Decimal('10'),
            'custom_unit': None,
        }]
        order = cc._add_contribution_order(tbs_member, lines, 2015, month=1)
        self.assertEqual(order.lines[0].price_per_unit, Decimal('0.010'))

        # declaration for june.. sl2 should be active
        lines = [{
            'tbs_member_product': product,
            'quantity': 10,
            'weight': Decimal('10'),
            'custom_unit': None,
        }]
        order = cc._add_contribution_order(tbs_member, lines, 2015, month=7)
        self.assertEqual(order.lines[0].price_per_unit, Decimal('0.020'))

    def test_member_guarantee_forecast(self):
        """
        tests DE forecast guarantee
        """
        set_date(date(2015, 1, 1))

        self.sa_session.delete(self.service_level)

        service_level_pv_1 = self._create_tbs_service_level(self.tbs_account, name='SL PV 1',
            year=today().year, admin_fee=Decimal('300.00'), level=1, product_type=PvcTbsProduct.TYPE_PV_RELATED, guarantee_forecast_min_value=0)
        service_level_pv_2 = self._create_tbs_service_level(self.tbs_account, name='SL PV 2',
            year=today().year, admin_fee=Decimal('400.00'), level=2, product_type=PvcTbsProduct.TYPE_PV_RELATED, guarantee_forecast_min_value=100)

        service_level_other_1 = self._create_tbs_service_level(self.tbs_account, name='SL Other 1', year=today().year,
            admin_fee=Decimal('500.00'), level=1, guarantee_forecast_min_value=0)
        service_level_other_2 = self._create_tbs_service_level(self.tbs_account, name='SL Other 2', year=today().year,
            admin_fee=Decimal('600.00'), level=2, guarantee_forecast_min_value=100)

        # set TBS config for guarantee forecast.
        self.tbs.set_config('pvc.tbs.guarantee.forecast', True)

        member_account = self._create_member_account()

        tbs_member = self._member_apply_at_tbs(member_account, self.tbs_account, service_level=None)
        self.assertEqual(tbs_member.status, PvcTbsMember.STATUS_PENDING)

        tbs_pv_product = self.pc.save_or_update_tbs_product(None, 'my pv product', 'desc', self.tbs.id,
            product_type=PvcTbsProduct.TYPE_PV_RELATED, basis_for_invoicing=PvcTbsProduct.INVOICE_BASED_ON_WEIGHT,
            custom_unit=PvcTbsProduct.CUSTOM_UNIT_NOMINAL_POWER
        )

        self.pc.save_contribution_fee(self.tbs_product.id, service_level_other_2.id, Decimal('0.025'), today(), PvcTbsProductFee.DIMENSION_NONE)
        self.pc.save_contribution_fee(tbs_pv_product.id, service_level_pv_2.id, Decimal('0.030'), today(), PvcTbsProductFee.DIMENSION_NONE)

        cc = PvcContributionOrderController(self.sa_session, account_id=member_account.id)
        order = cc.add_guarantee_forecast_order(tbs_member.id, today().year, [self.tbs_product.id, tbs_pv_product.id], is_b2b=None)

        self.assertEqual(order.total_amount_including_vat, Decimal('0'))
        self.assertEqual(order.is_guarantee_forecast, True)
        self.assertEqual(len(order.lines), 2)

        # fill quantities and add extra line
        order = cc.update_contribution(order.id, {
            # other module
            'line_0_quantity': 1,
            'line_0_weight': Decimal('100.00'),
            # pv module
            'line_1_quantity': 1,
            'line_1_weight': Decimal('200.00'),
            'line_1_custom_unit': Decimal('250.00'),
        }, additional_lines={
            # extra product
            'tbs_product_tpid_%d' % self.tbs_product2.id: self.tbs_product2.id,
            'line_tpid_%d_weight' % self.tbs_product2.id: Decimal('300'),
            'line_tpid_%d_quantity' % self.tbs_product2.id: 300,
            'line_tpid_%d_country_id' % self.tbs_product2.id: None,
        })

        self.assertEqual(order.total_amount_including_vat, Decimal('0'))
        self.assertEqual(order.is_guarantee_forecast, True)
        self.assertEqual(len(order.lines), 3)

        # submit should determine the new SL
        cc.submit_order(order.id)

        # expect two service levels to be active for this tbs_member
        self.assertEqual(tbs_member.active_service_level.service_level, service_level_other_2)
        self.assertEqual(tbs_member.active_service_level.pv_service_level, service_level_pv_2)
        self.assertEqual(tbs_member.active_service_level.start_date, date(2000, 1, 1))

        # approve member
        mc = PvcMemberController(self.sa_session, account_id=self.tbs_account.id)
        mc.save_tbs_member(tbs_member.id, {'status': PvcTbsMember.STATUS_APPROVED, 'is_full_member': False, 'is_representative': False})

        # expect the AF order to contain 2 lines
        poc = PvcOrderController(self.sa_session, account_id=member_account.id)
        orders = poc.get()

        self.assertEqual(len(orders), 1)
        order = orders[0]
        self.assertTrue(isinstance(order, PvcOrder))

        self.assertEqual(len(order.lines), 2)
        self.assertEqual(order.lines[0].description, u'Administration Fee SL Other 2 for year 2015')
        self.assertEqual(order.lines[0].line_amount_excluding_vat, Decimal('600'))
        self.assertEqual(order.lines[1].description, u'Administration Fee SL PV 2 for year 2015')
        self.assertEqual(order.lines[1].line_amount_excluding_vat, Decimal('400'))

        member_other_product = self._get_member_product(tbs_member, self.tbs_product)
        member_pv_product = self._get_member_product(tbs_member, tbs_pv_product)

        # goto march and do a declaration
        set_date(date(2015, 3, 1))

        cc = PvcContributionOrderController(self.sa_session, account_id=member_account.id)
        lines = [{
            'tbs_member_product': member_other_product,
            'quantity': 10,
            'weight': Decimal('10'),
            'custom_unit': None,
        }, {
            'tbs_member_product': member_pv_product,
            'quantity': 10,
            'weight': Decimal('10'),
            'custom_unit': None,
        }]

        order = cc._add_contribution_order(tbs_member, lines, 2015, month=3)
        order = cc.submit_order(order.id)

        # assert pricies, expect different price unit prices for different products
        self.assertEqual(len(order.lines), 2)
        self.assertEqual(order.lines[0].price_per_unit, Decimal('0.025'))
        self.assertEqual(order.lines[0].description, 'EEE-TBS-TEST')
        self.assertEqual(order.lines[1].price_per_unit, Decimal('0.030'))
        self.assertEqual(order.lines[1].description, 'my pv product')

        # goto november and do a GF for next year

        set_date(date(today().year, 11, 1))

        order = cc.add_guarantee_forecast_order(tbs_member.id, today().year + 1, [self.tbs_product.id, tbs_pv_product.id], is_b2b=None)
        self.assertEqual(order.total_amount_including_vat, Decimal('0'))
        self.assertEqual(order.is_guarantee_forecast, True)

        # has minimal quantites so expect downgrade next year
        order = cc.update_contribution(order.id, {
            # other module
            'line_0_quantity': 1,
            'line_0_weight': Decimal('1'),
            # pv module
            'line_1_quantity': 1,
            'line_1_weight': Decimal('1'),
            'line_1_custom_unit': Decimal('1'),
        }, additional_lines={
        })

        self.assertEqual(order.total_amount_including_vat, Decimal('0'))
        self.assertEqual(order.is_guarantee_forecast, True)

        # submit should determine the new SL for next year
        guarantee_order = cc.submit_order(order.id)

        # current SL should not have changed
        self.assertEqual(tbs_member.active_service_level.service_level, service_level_other_2)
        self.assertEqual(tbs_member.active_service_level.pv_service_level, service_level_pv_2)

        # upcoming SL should start in jan
        self.assertEqual(tbs_member.upcoming_service_level.service_level, service_level_other_1)
        self.assertEqual(tbs_member.upcoming_service_level.pv_service_level, service_level_pv_1)
        self.assertEqual(tbs_member.upcoming_service_level.start_date, date(today().year+1, 1, 1))

    def test_member_guarantee_forecast_remove_product(self):
        """
        tests DE forecast guarantee
        """
        set_date(date(2015, 1, 1))

        self.sa_session.delete(self.service_level)

        service_level_pv_1 = self._create_tbs_service_level(self.tbs_account, name='SL PV 1',
            year=today().year, admin_fee=Decimal('300.00'), level=1, product_type=PvcTbsProduct.TYPE_PV_RELATED, guarantee_forecast_min_value=0)
        service_level_pv_2 = self._create_tbs_service_level(self.tbs_account, name='SL PV 2',
            year=today().year, admin_fee=Decimal('400.00'), level=2, product_type=PvcTbsProduct.TYPE_PV_RELATED, guarantee_forecast_min_value=100)

        service_level_other_1 = self._create_tbs_service_level(self.tbs_account, name='SL Other 1', year=today().year,
            admin_fee=Decimal('500.00'), level=1, guarantee_forecast_min_value=0)
        service_level_other_2 = self._create_tbs_service_level(self.tbs_account, name='SL Other 2', year=today().year,
            admin_fee=Decimal('600.00'), level=2, guarantee_forecast_min_value=100)

        # set TBS config for guarantee forecast.
        self.tbs.set_config('pvc.tbs.guarantee.forecast', True)

        member_account = self._create_member_account()

        tbs_member = self._member_apply_at_tbs(member_account, self.tbs_account, service_level=None)
        self.assertEqual(tbs_member.status, PvcTbsMember.STATUS_PENDING)

        tbs_pv_product = self.pc.save_or_update_tbs_product(None, 'my pv product', 'desc', self.tbs.id,
            product_type=PvcTbsProduct.TYPE_PV_RELATED, basis_for_invoicing=PvcTbsProduct.INVOICE_BASED_ON_WEIGHT,
            custom_unit=PvcTbsProduct.CUSTOM_UNIT_NOMINAL_POWER
        )

        self.pc.save_contribution_fee(self.tbs_product.id, service_level_other_2.id, Decimal('0.025'), today(), PvcTbsProductFee.DIMENSION_NONE)
        self.pc.save_contribution_fee(tbs_pv_product.id, service_level_pv_2.id, Decimal('0.030'), today(), PvcTbsProductFee.DIMENSION_NONE)

        cc = PvcContributionOrderController(self.sa_session, account_id=member_account.id)
        order = cc.add_guarantee_forecast_order(tbs_member.id, today().year, [self.tbs_product.id, tbs_pv_product.id], is_b2b=None)

        self.assertEqual(order.total_amount_including_vat, Decimal('0'))
        self.assertEqual(order.is_guarantee_forecast, True)
        self.assertEqual(len(order.lines), 2)

        # fill quantities and add extra line
        order = cc.update_contribution(order.id, {
            # other module
            'line_0_quantity': 1,
            'line_0_weight': Decimal('100.00'),
            # pv module
            'line_1_quantity': 1,
            'line_1_weight': Decimal('200.00'),
            'line_1_custom_unit': Decimal('250.00'),
        })

        cc.remove_line(order.id, order.lines[0].id)
        order = cc.submit_order(order.id)
        self.assertEqual(len(order.lines), 1)

        mps = self.sa_session.query(PvcTbsMemberProduct).filter(PvcTbsMemberProduct.tbs_member == order.producer).all()
        self.assertEqual(len(mps), 2)
        success = False
        for mp in mps:
            if mp.tbs_product == self.tbs_product:
                self.assertFalse(mp.is_active)
                success = True

        self.assertTrue(success)

    def test_member_guarantee_forecast_one_sl(self):
        """
        tests DE forecast guarantee with only having a PV SL
        """
        set_date(date(2015, 1, 1))

        self.sa_session.delete(self.service_level)

        service_level_pv_1 = self._create_tbs_service_level(self.tbs_account, name='SL PV 1',
            year=today().year, admin_fee=Decimal('100.00'), level=1, product_type=PvcTbsProduct.TYPE_PV_RELATED, guarantee_forecast_min_value=0)
        service_level_pv_2 = self._create_tbs_service_level(self.tbs_account, name='SL PV 2',
            year=today().year, admin_fee=Decimal('200.00'), level=2, product_type=PvcTbsProduct.TYPE_PV_RELATED, guarantee_forecast_min_value=100)
        service_level_pv_3 = self._create_tbs_service_level(self.tbs_account, name='SL PV 3',
            year=today().year, admin_fee=Decimal('300.00'), level=3, product_type=PvcTbsProduct.TYPE_PV_RELATED, guarantee_forecast_min_value=200)

        service_level_other_1 = self._create_tbs_service_level(self.tbs_account, name='SL Other 1', year=today().year,
            admin_fee=Decimal('300.00'), level=1, guarantee_forecast_min_value=0)
        service_level_other_2 = self._create_tbs_service_level(self.tbs_account, name='SL Other 2', year=today().year,
            admin_fee=Decimal('300.00'), level=2, guarantee_forecast_min_value=100)

        # set TBS config for guarantee forecast.
        self.tbs.set_config('pvc.tbs.guarantee.forecast', True)

        member_account = self._create_member_account()

        tbs_member = self._member_apply_at_tbs(member_account, self.tbs_account, service_level=None)
        self.assertEqual(tbs_member.status, PvcTbsMember.STATUS_PENDING)

        tbs_pv_product = self.pc.save_or_update_tbs_product(None, 'my pv product', 'desc', self.tbs.id,
            product_type=PvcTbsProduct.TYPE_PV_RELATED, basis_for_invoicing=PvcTbsProduct.INVOICE_BASED_ON_WEIGHT,
            custom_unit=PvcTbsProduct.CUSTOM_UNIT_NOMINAL_POWER
        )

        tbs_pv_product2 = self.pc.save_or_update_tbs_product(None, 'my other pv product', 'other desc', self.tbs.id,
            product_type=PvcTbsProduct.TYPE_PV_RELATED, basis_for_invoicing=PvcTbsProduct.INVOICE_BASED_ON_WEIGHT,
            custom_unit=PvcTbsProduct.CUSTOM_UNIT_NOMINAL_POWER
        )

        member_pv_product = self._get_member_product(tbs_member, tbs_pv_product)
        member_other_product = self._get_member_product(tbs_member, self.tbs_product)

        self.pc.save_contribution_fee(self.tbs_product.id, service_level_other_2.id, Decimal('0.025'), today(), PvcTbsProductFee.DIMENSION_NONE)
        self.pc.save_contribution_fee(tbs_pv_product.id, service_level_pv_2.id, Decimal('0.030'), today(), PvcTbsProductFee.DIMENSION_NONE)
        self.pc.save_contribution_fee(tbs_pv_product2.id, service_level_pv_2.id, Decimal('0.065'), today(), PvcTbsProductFee.DIMENSION_NONE)

        cc = PvcContributionOrderController(self.sa_session, account_id=member_account.id)
        order = cc.add_guarantee_forecast_order(tbs_member.id, today().year, [tbs_pv_product.id], is_b2b=None)

        self.assertEqual(order.total_amount_including_vat, Decimal('0'))
        self.assertEqual(order.is_guarantee_forecast, True)
        self.assertEqual(len(order.lines), 1)

        # fill quantities and add extra line
        order = cc.update_contribution(order.id, {
            # pv module
            'line_0_quantity': 100,
            'line_0_weight': Decimal('100.00'),
            'line_0_custom_unit': Decimal('100.00'),
        }, additional_lines={
        })

        self.assertEqual(order.total_amount_including_vat, Decimal('0'))
        self.assertEqual(order.is_guarantee_forecast, True)
        self.assertEqual(len(order.lines), 1)

        # submit should determine the new SL
        guarantee_order = cc.submit_order(order.id)

        # expect one service level to be active for this tbs_member
        self.assertEqual(tbs_member.active_service_level.service_level, None)
        self.assertEqual(tbs_member.active_service_level.pv_service_level.name, 'SL PV 2')

        # approve member
        mc = PvcMemberController(self.sa_session, account_id=self.tbs_account.id)
        mc.save_tbs_member(tbs_member.id, {'status': PvcTbsMember.STATUS_APPROVED, 'is_full_member': False, 'is_representative': False})

        # expect the AF order to contain 1 line
        poc = PvcOrderController(self.sa_session, account_id=member_account.id)
        orders = poc.get()

        self.assertEqual(len(orders), 1)
        order = orders[0]
        self.assertTrue(isinstance(order, PvcOrder))

        self.assertEqual(len(order.lines), 1)
        self.assertEqual(order.lines[0].description, u'Administration Fee SL PV 2 for year 2015')
        self.assertEqual(order.lines[0].line_amount_excluding_vat, Decimal('200'))

        #
        # do a contribution order
        #

        set_date(date(2015, 3, 1))

        cc = PvcContributionOrderController(self.sa_session, account_id=member_account.id)

        # assert that it's not possible to declare for a product where we don't have a SL for
        self.assertRaises(PvcServiceLevelException, cc._add_contribution_order, tbs_member, [{
            'tbs_member_product': member_other_product,
            'quantity': 10,
            'weight': Decimal('10'),
            'custom_unit': None,
        }, {
            'tbs_member_product': member_pv_product,
            'quantity': 10,
            'weight': Decimal('10'),
            'custom_unit': None,
        }], 2015, month=3)

        # but it is possible with only PV
        order = cc._add_contribution_order(tbs_member, [{
            'tbs_member_product': member_pv_product,
            'quantity': 10,
            'weight': Decimal('10'),
            'custom_unit': None,
        }], 2015, month=3)

        # UPDATE the contribution with an additional line (PV, ticket #983)
        order = cc.update_contribution(order.id, {
            # pv module
            'line_0_quantity': 300,
            'line_0_weight': Decimal('300'),
            'line_0_custom_unit': Decimal('300.00'),
        }, additional_lines={
            'tbs_product_tpid_%d' % tbs_pv_product2.id: tbs_pv_product2.id,
            'line_tpid_%d_weight' % tbs_pv_product2.id: Decimal('10'),
            'line_tpid_%d_quantity' % tbs_pv_product2.id: 100,
            'line_tpid_%d_custom_unit' % tbs_pv_product2.id: 100,
            'line_tpid_%d_country_id' % tbs_pv_product2.id: None,
        })

        self.assertEqual(len(order.lines), 2)
        self.assertEqual(order.lines[0].price_per_unit, Decimal('0.065'))
        self.assertEqual(order.lines[1].price_per_unit, Decimal('0.030'))

        #
        # test upgrading/downgrading SL
        #

        # goto march
        set_date(date(2015, 3, 1))

        # make a request for resubmission
        cc = PvcContributionOrderController(self.sa_session, account_id=member_account.id)
        cc.request_for_resubmission(guarantee_order.id)

        # now the operator approves the resubmit.
        cod = PvcContributionOrderController(self.sa_session, account_id=self.operator_account.id)
        cod.approve_resubmit_request(guarantee_order.id)

        # perform resubmit which will result in having a lower SL
        resubmit = cc.update_contribution(guarantee_order.id, {
            # pv module
            'line_0_quantity': 10,
            'line_0_weight': Decimal('10'),
            'line_0_custom_unit': Decimal('10.00'),
        })

        # submit the resubmit, assert that service level downgrade is not possible
        self.assertRaises(PvcServiceLevelDowngradeException, cc.submit_order, resubmit.id)

        # perform another resubmit which will result in having a higher SL
        cc.update_contribution(resubmit.id, {
            # pv module
            'line_0_quantity': 300,
            'line_0_weight': Decimal('300'),
            'line_0_custom_unit': Decimal('300.00'),
        })

        # submit the resubmit, assert that service level downgrade is not possible
        cc.submit_order(resubmit.id)

        # assert the new SL. this is an upgrade so new SL must start next month
        sl = tbs_member.upcoming_service_level
        self.assertEqual(sl.service_level, None)
        self.assertEqual(sl.pv_service_level.name, 'SL PV 3')
        self.assertEqual(sl.start_date, date(2015, 4, 1))

    def test_member_guarantee_forecast_batteries(self):
        """
        tests DE forecast guarantee
        """
        set_date(date(2015, 1, 1))

        self.sa_session.delete(self.service_level)

        service_level_pv_1 = self._create_tbs_service_level(self.tbs_account, name='SL PV 1',
            year=today().year, admin_fee=Decimal('300.00'), level=1, product_type=PvcTbsProduct.TYPE_PV_RELATED, guarantee_forecast_min_value=0)
        service_level_pv_2 = self._create_tbs_service_level(self.tbs_account, name='SL PV 2',
            year=today().year, admin_fee=Decimal('400.00'), level=2, product_type=PvcTbsProduct.TYPE_PV_RELATED, guarantee_forecast_min_value=100)

        service_level_other_1 = self._create_tbs_service_level(self.tbs_account, name='SL Other 1', year=today().year,
            admin_fee=Decimal('500.00'), level=1, guarantee_forecast_min_value=0)
        service_level_other_2 = self._create_tbs_service_level(self.tbs_account, name='SL Other 2', year=today().year,
            admin_fee=Decimal('600.00'), level=2, guarantee_forecast_min_value=100)

        # set TBS config for guarantee forecast.
        self.tbs.set_config('pvc.tbs.guarantee.forecast', True)

        member_account = self._create_member_account()

        tbs_member = self._member_apply_at_tbs(member_account, self.tbs_account, service_level=None)
        self.assertEqual(tbs_member.status, PvcTbsMember.STATUS_PENDING)

        tbs_pv_product = self.pc.save_or_update_tbs_product(None, 'my pv product', 'desc', self.tbs.id,
            product_type=PvcTbsProduct.TYPE_PV_RELATED, basis_for_invoicing=PvcTbsProduct.INVOICE_BASED_ON_WEIGHT,
            custom_unit=PvcTbsProduct.CUSTOM_UNIT_NOMINAL_POWER
        )
        tbs_batteries_product = self.pc.save_or_update_tbs_product(None, 'my batteries product', 'desc', self.tbs.id,
            product_type=PvcTbsProduct.TYPE_PV_RELATED, basis_for_invoicing=PvcTbsProduct.INVOICE_BASED_ON_WEIGHT
        )

        self.pc.save_contribution_fee(self.tbs_product.id, service_level_other_2.id, Decimal('0.025'), today(), PvcTbsProductFee.DIMENSION_NONE)
        self.pc.save_contribution_fee(tbs_pv_product.id, service_level_pv_2.id, Decimal('0.030'), today(), PvcTbsProductFee.DIMENSION_NONE)
        self.pc.save_contribution_fee(tbs_batteries_product.id, service_level_pv_2.id, Decimal('0.035'), today(), PvcTbsProductFee.DIMENSION_NONE)

        cc = PvcContributionOrderController(self.sa_session, account_id=member_account.id)
        order = cc.add_guarantee_forecast_order(tbs_member.id, today().year, [self.tbs_product.id, tbs_pv_product.id, tbs_batteries_product.id], is_b2b=None)

        self.assertEqual(order.total_amount_including_vat, Decimal('0'))
        self.assertEqual(order.is_guarantee_forecast, True)
        self.assertEqual(len(order.lines), 3)

        order = cc.update_contribution(order.id, {
            # other module
            'line_0_quantity': 1,
            'line_0_weight': Decimal('100.00'),

            # pv module
            'line_2_quantity': 1,
            'line_2_weight': Decimal('200.00'),
            'line_2_custom_unit': Decimal('250.00'),

            # batteries (pv module but without custom_unit)
            'line_1_quantity': 1,
            'line_1_weight': Decimal('200.00'),

        }, additional_lines={
        })

        self.assertEqual(order.total_amount_including_vat, Decimal('0'))
        self.assertEqual(order.is_guarantee_forecast, True)
        self.assertEqual(len(order.lines), 3)

        # submit should determine the new SL
        cc.submit_order(order.id)

        # expect two service levels to be active for this tbs_member
        self.assertEqual(tbs_member.active_service_level.service_level, service_level_other_2)
        self.assertEqual(tbs_member.active_service_level.pv_service_level, service_level_pv_2)

    def test_member_guarantee_forecast_resubmit(self):
        """
        tests DE forecast guarantee
        """
        set_date(date(2015, 10, 1))

        self.sa_session.delete(self.service_level)

        service_level_pv_1 = self._create_tbs_service_level(self.tbs_account, name='SL PV 1',
            year=today().year, admin_fee=Decimal('300.00'), level=1, product_type=PvcTbsProduct.TYPE_PV_RELATED, guarantee_forecast_min_value=0)
        service_level_pv_2 = self._create_tbs_service_level(self.tbs_account, name='SL PV 2',
            year=today().year, admin_fee=Decimal('400.00'), level=2, product_type=PvcTbsProduct.TYPE_PV_RELATED, guarantee_forecast_min_value=100)

        service_level_other_1 = self._create_tbs_service_level(self.tbs_account, name='SL Other 1', year=today().year,
            admin_fee=Decimal('500.00'), level=1, guarantee_forecast_min_value=0)
        service_level_other_2 = self._create_tbs_service_level(self.tbs_account, name='SL Other 2', year=today().year,
            admin_fee=Decimal('600.00'), level=2, guarantee_forecast_min_value=100)

        # set TBS config for guarantee forecast.
        self.tbs.set_config('pvc.tbs.guarantee.forecast', True)

        # setup tbs products
        tbs_pv_product = self.pc.save_or_update_tbs_product(None, 'my pv product', 'desc', self.tbs.id,
            product_type=PvcTbsProduct.TYPE_PV_RELATED, basis_for_invoicing=PvcTbsProduct.INVOICE_BASED_ON_WEIGHT,
            custom_unit=PvcTbsProduct.CUSTOM_UNIT_NOMINAL_POWER
        )
        tbs_batteries_product = self.pc.save_or_update_tbs_product(None, 'my batteries product', 'desc', self.tbs.id,
            product_type=PvcTbsProduct.TYPE_PV_RELATED, basis_for_invoicing=PvcTbsProduct.INVOICE_BASED_ON_WEIGHT
        )

        self.pc.save_contribution_fee(self.tbs_product.id, service_level_other_2.id, Decimal('0.025'), today(), PvcTbsProductFee.DIMENSION_NONE)
        self.pc.save_contribution_fee(tbs_pv_product.id, service_level_pv_2.id, Decimal('0.030'), today(), PvcTbsProductFee.DIMENSION_NONE)
        self.pc.save_contribution_fee(tbs_batteries_product.id, service_level_pv_2.id, Decimal('0.035'), today(), PvcTbsProductFee.DIMENSION_NONE)

        # setup member
        member_account = self._create_member_account()
        tbs_member = self._member_apply_at_tbs(member_account, self.tbs_account, service_level=None)
        self.assertEqual(tbs_member.status, PvcTbsMember.STATUS_PENDING)

        # initial Guarantee forecast
        cc = PvcContributionOrderController(self.sa_session, account_id=member_account.id)
        order = cc.add_guarantee_forecast_order(tbs_member.id, today().year, [self.tbs_product.id], is_b2b=None)
        self.assertEqual(order.total_amount_including_vat, Decimal('0'))
        self.assertEqual(order.is_guarantee_forecast, True)
        self.assertEqual(len(order.lines), 1)

        order = cc.update_contribution(order.id, {
            # other module
            'line_0_quantity': 1,
            'line_0_weight': Decimal('100.00'),

            # pv module
            #'line_1_quantity': 1,
            #'line_1_weight': Decimal('200.00'),
            #'line_1_custom_unit': Decimal('250.00'),

        })

        self.assertEqual(order.total_amount_including_vat, Decimal('0'))
        self.assertEqual(order.is_guarantee_forecast, True)
        self.assertEqual(len(order.lines), 1)

        # submit and expect one service level to be active for this tbs_member
        order = cc.submit_order(order.id)
        self.assertEqual(tbs_member.active_service_level.service_level, service_level_other_2)
        self.assertEqual(tbs_member.active_service_level.pv_service_level, None)

        # expect the start date of the service levels to be the 1st of January 2000.
        self.assertEqual(tbs_member.active_service_level.start_date, date(2000, 1, 1))

        # approve member
        mc = PvcMemberController(self.sa_session, account_id=self.tbs_account.id)
        mc.save_tbs_member(tbs_member.id, {'status': PvcTbsMember.STATUS_APPROVED, 'is_full_member': False, 'is_representative': False})

        # request/approve resubmit
        order.resubmit_request = PvcContributionOrder.RESUBMIT_REQUEST_APPROVED
        self.sa_session.add(order)
        self.sa_session.flush()

        # update Guarantee forecast, add PV module
        order = cc.update_contribution(order.id, {
            # other module
            'line_0_quantity': 1,
            'line_0_weight': Decimal('120.00'),
        }, additional_lines={
            'tbs_product_tpid_%d' % tbs_pv_product.id: tbs_pv_product.id,
            'line_tpid_%d_weight' % tbs_pv_product.id: Decimal('10'),
            'line_tpid_%d_quantity' % tbs_pv_product.id: 100,
            'line_tpid_%d_custom_unit' % tbs_pv_product.id: 100,
            'line_tpid_%d_country_id' % tbs_pv_product.id: None,
        })

        self.assertTrue(order.parent_order is not None)
        self.assertEqual(order.total_amount_including_vat, Decimal('0'))
        self.assertEqual(order.is_guarantee_forecast, True)
        self.assertEqual(len(order.lines), 2)
        self.assertEqual(order.lines[0].price_per_unit, Decimal('0'))
        self.assertEqual(order.lines[1].price_per_unit, Decimal('0'))
        order = cc.submit_order(order.id)

        # expect two service levels to be active for this tbs_member
        self.assertEqual(tbs_member.active_service_level.service_level, service_level_other_2)
        self.assertEqual(tbs_member.active_service_level.pv_service_level, None)

        # assert upcoming SL to be PV
        self.assertEqual(tbs_member.upcoming_service_level.service_level, service_level_other_2)
        self.assertEqual(tbs_member.upcoming_service_level.pv_service_level, service_level_pv_2)

    def test_member_guarantee_forecast_resubmit_same_sl(self):
        """
        resubmit of a GF where the final SL will remain the same
        """
        set_date(date(2015, 10, 1))

        self.sa_session.delete(self.service_level)

        service_level_pv_1 = self._create_tbs_service_level(self.tbs_account, name='SL PV 1',
            year=today().year, admin_fee=Decimal('300.00'), level=1, product_type=PvcTbsProduct.TYPE_PV_RELATED, guarantee_forecast_min_value=0)
        service_level_pv_2 = self._create_tbs_service_level(self.tbs_account, name='SL PV 2',
            year=today().year, admin_fee=Decimal('400.00'), level=2, product_type=PvcTbsProduct.TYPE_PV_RELATED, guarantee_forecast_min_value=100)

        service_level_other_1 = self._create_tbs_service_level(self.tbs_account, name='SL Other 1', year=today().year,
            admin_fee=Decimal('500.00'), level=1, guarantee_forecast_min_value=0)
        service_level_other_2 = self._create_tbs_service_level(self.tbs_account, name='SL Other 2', year=today().year,
            admin_fee=Decimal('600.00'), level=2, guarantee_forecast_min_value=100)

        # set TBS config for guarantee forecast.
        self.tbs.set_config('pvc.tbs.guarantee.forecast', True)

        # setup tbs products
        tbs_pv_product = self.pc.save_or_update_tbs_product(None, 'my pv product', 'desc', self.tbs.id,
            product_type=PvcTbsProduct.TYPE_PV_RELATED, basis_for_invoicing=PvcTbsProduct.INVOICE_BASED_ON_WEIGHT,
            custom_unit=PvcTbsProduct.CUSTOM_UNIT_NOMINAL_POWER
        )
        tbs_batteries_product = self.pc.save_or_update_tbs_product(None, 'my batteries product', 'desc', self.tbs.id,
            product_type=PvcTbsProduct.TYPE_PV_RELATED, basis_for_invoicing=PvcTbsProduct.INVOICE_BASED_ON_WEIGHT
        )

        self.pc.save_contribution_fee(self.tbs_product.id, service_level_other_2.id, Decimal('0.025'), today(), PvcTbsProductFee.DIMENSION_NONE)
        self.pc.save_contribution_fee(tbs_pv_product.id, service_level_pv_2.id, Decimal('0.030'), today(), PvcTbsProductFee.DIMENSION_NONE)
        self.pc.save_contribution_fee(tbs_batteries_product.id, service_level_pv_2.id, Decimal('0.035'), today(), PvcTbsProductFee.DIMENSION_NONE)

        # setup member
        member_account = self._create_member_account()
        tbs_member = self._member_apply_at_tbs(member_account, self.tbs_account, service_level=None)
        self.assertEqual(tbs_member.status, PvcTbsMember.STATUS_PENDING)

        # initial Guarantee forecast set SL for 2015
        cc = PvcContributionOrderController(self.sa_session, account_id=member_account.id)
        order = cc.add_guarantee_forecast_order(tbs_member.id, today().year, [tbs_pv_product.id], is_b2b=None)
        self.assertEqual(order.total_amount_including_vat, Decimal('0'))
        self.assertEqual(order.is_guarantee_forecast, True)
        self.assertEqual(len(order.lines), 1)

        order = cc.update_contribution(order.id, {
            # pv module
            'line_0_quantity': 101,
            'line_0_weight': Decimal('101'),
            'line_0_custom_unit': Decimal('101'),

        })

        self.assertEqual(order.total_amount_including_vat, Decimal('0'))
        self.assertEqual(order.is_guarantee_forecast, True)
        self.assertEqual(len(order.lines), 1)

        # submit and expect one service level to be active for this tbs_member
        order = cc.submit_order(order.id)
        self.assertEqual(tbs_member.active_service_level.service_level, None)
        self.assertEqual(tbs_member.active_service_level.pv_service_level, service_level_pv_2)

        # approve member
        mc = PvcMemberController(self.sa_session, account_id=self.tbs_account.id)
        mc.save_tbs_member(tbs_member.id, {'status': PvcTbsMember.STATUS_APPROVED, 'is_full_member': False, 'is_representative': False})

        # GF for 2016, do a SL downgrade

        set_date(date(today().year, 11, 1))

        order = cc.add_guarantee_forecast_order(tbs_member.id, today().year + 1, [tbs_pv_product.id], is_b2b=None)
        self.assertEqual(order.total_amount_including_vat, Decimal('0'))
        self.assertEqual(order.is_guarantee_forecast, True)

        # has minimal quantites so expect downgrade next year
        order = cc.update_contribution(order.id, {
            # pv module
            'line_0_quantity': 10,
            'line_0_weight': Decimal('10'),
            'line_0_custom_unit': Decimal('10'),

        }, additional_lines={
        })

        self.assertEqual(order.total_amount_including_vat, Decimal('0'))
        self.assertEqual(order.is_guarantee_forecast, True)

        # submit should determine the new SL for next year
        order = cc.submit_order(order.id)

        # current SL should stay the same,
        self.assertEqual(tbs_member.active_service_level.service_level, None)
        self.assertEqual(tbs_member.active_service_level.pv_service_level, service_level_pv_2)

        # next year should be sl1
        self.assertEqual(tbs_member.get_service_level_for_date( date(today().year + 1, 1, 1) ).service_level, None)
        self.assertEqual(tbs_member.get_service_level_for_date( date(today().year + 1, 1, 1) ).pv_service_level, service_level_pv_1)

        self.sa_session.flush()

        #
        # now resubmit the GF for next year
        #

        # request/approve resubmit
        order.resubmit_request = PvcContributionOrder.RESUBMIT_REQUEST_APPROVED
        self.sa_session.add(order)
        self.sa_session.flush()

        # same quantities as old GF
        order = cc.update_contribution(order.id, {
            'line_0_quantity': 10,
            'line_0_weight': Decimal('10'),
            'line_0_custom_unit': Decimal('10'),
        })

        # even though the SL hasn't changed for 2016, we still should be able to resubmit
        cc.submit_order(order.id)

        # assert that the SL for 2016 is still the same
        self.assertEqual(tbs_member.get_service_level_for_date( date(today().year + 1, 1, 1) ).service_level, None)
        self.assertEqual(tbs_member.get_service_level_for_date( date(today().year + 1, 1, 1) ).pv_service_level, service_level_pv_1)

    def test_member_guarantee_forecast_resubmit_future(self):
        """
        """
        set_date(date(2015, 12, 1))

        self.sa_session.delete(self.service_level)

        service_level_pv_1 = self._create_tbs_service_level(self.tbs_account, name='SL PV 1',
            year=today().year, admin_fee=Decimal('300.00'), level=1, product_type=PvcTbsProduct.TYPE_PV_RELATED, guarantee_forecast_min_value=0)
        service_level_pv_2 = self._create_tbs_service_level(self.tbs_account, name='SL PV 2',
            year=today().year, admin_fee=Decimal('400.00'), level=2, product_type=PvcTbsProduct.TYPE_PV_RELATED, guarantee_forecast_min_value=100)

        service_level_other_1 = self._create_tbs_service_level(self.tbs_account, name='SL Other 1', year=today().year,
            admin_fee=Decimal('500.00'), level=1, guarantee_forecast_min_value=0)
        service_level_other_2 = self._create_tbs_service_level(self.tbs_account, name='SL Other 2', year=today().year,
            admin_fee=Decimal('600.00'), level=2, guarantee_forecast_min_value=100)

        # set TBS config for guarantee forecast.
        self.tbs.set_config('pvc.tbs.guarantee.forecast', True)

        # setup tbs products
        tbs_pv_product = self.pc.save_or_update_tbs_product(None, 'my pv product', 'desc', self.tbs.id,
            product_type=PvcTbsProduct.TYPE_PV_RELATED, basis_for_invoicing=PvcTbsProduct.INVOICE_BASED_ON_WEIGHT,
            custom_unit=PvcTbsProduct.CUSTOM_UNIT_NOMINAL_POWER
        )
        tbs_batteries_product = self.pc.save_or_update_tbs_product(None, 'my batteries product', 'desc', self.tbs.id,
            product_type=PvcTbsProduct.TYPE_PV_RELATED, basis_for_invoicing=PvcTbsProduct.INVOICE_BASED_ON_WEIGHT
        )

        self.pc.save_contribution_fee(self.tbs_product.id, service_level_other_2.id, Decimal('0.025'), today(), PvcTbsProductFee.DIMENSION_NONE)
        self.pc.save_contribution_fee(tbs_pv_product.id, service_level_pv_2.id, Decimal('0.030'), today(), PvcTbsProductFee.DIMENSION_NONE)
        self.pc.save_contribution_fee(tbs_batteries_product.id, service_level_pv_2.id, Decimal('0.035'), today(), PvcTbsProductFee.DIMENSION_NONE)

        # setup member
        member_account = self._create_member_account()
        tbs_member = self._member_apply_at_tbs(member_account, self.tbs_account, service_level=None)
        self.assertEqual(tbs_member.status, PvcTbsMember.STATUS_PENDING)

        # initial Guarantee forecast set SL for 2015
        cc = PvcContributionOrderController(self.sa_session, account_id=member_account.id)
        order = cc.add_guarantee_forecast_order(tbs_member.id, today().year, [tbs_pv_product.id], is_b2b=None)

        order = cc.update_contribution(order.id, {
            # pv module
            'line_0_quantity': 101,
            'line_0_weight': Decimal('101'),
            'line_0_custom_unit': Decimal('101'),
        })

        # submit and expect one service level to be active for this tbs_member
        order = cc.submit_order(order.id)

        # approve member
        mc = PvcMemberController(self.sa_session, account_id=self.tbs_account.id)
        mc.save_tbs_member(tbs_member.id, {'status': PvcTbsMember.STATUS_APPROVED, 'is_full_member': False, 'is_representative': False})

        # GF for 2016, lower SL
        order = cc.add_guarantee_forecast_order(tbs_member.id, today().year + 1, [tbs_pv_product.id], is_b2b=None)
        order = cc.update_contribution(order.id, {
            # pv module
            'line_0_quantity': 10,
            'line_0_weight': Decimal('10'),
            'line_0_custom_unit': Decimal('10'),

        }, additional_lines={
        })

        # submit should determine the new SL for next year
        order = cc.submit_order(order.id)

        # current SL is pv2
        self.assertEqual(tbs_member.active_service_level.service_level, None)
        self.assertEqual(tbs_member.active_service_level.pv_service_level, service_level_pv_2)

        # next year should be sl1
        self.assertEqual(tbs_member.get_service_level_for_date( date(today().year + 1, 1, 1) ).service_level, None)
        self.assertEqual(tbs_member.get_service_level_for_date( date(today().year + 1, 1, 1) ).pv_service_level, service_level_pv_1)

        self.sa_session.flush()

        #
        # now resubmit the GF for next year
        #

        # request/approve resubmit
        order.resubmit_request = PvcContributionOrder.RESUBMIT_REQUEST_APPROVED
        self.sa_session.add(order)
        self.sa_session.flush()

        # going higher
        order = cc.update_contribution(order.id, {
            'line_0_quantity': 101,
            'line_0_weight': Decimal('101'),
            'line_0_custom_unit': Decimal('101'),
        })

        cc.submit_order(order.id)

        # assert the SL for 2016
        self.assertEqual(tbs_member.get_service_level_for_date( date(today().year + 1, 1, 1) ).service_level, None)
        self.assertEqual(tbs_member.get_service_level_for_date( date(today().year + 1, 1, 1) ).pv_service_level, service_level_pv_2)

        # assert 1 AF was triggered yet (for 2016)
        poc = PvcOrderController(self.sa_session, account_id=member_account.id)
        orders = poc.get()
        self.assertEqual(len(orders), 1)

        # expected that eventually only 1 SL was made starting at 2000-01-01
        sls = self.sa_session.query(PvcTbsMemberServiceLevel).filter(
            PvcTbsMemberServiceLevel.tbs_member_id == tbs_member.id
        ).all()
        self.assertEqual(len(sls), 1)
        self.assertEqual(sls[0].start_date, date(2000, 1, 1))

    def test_member_guarantee_forecast_resubmit_pending(self):
        """
        tests that a pending member does a guarantee forecast resubmit. expected is that the SL is updated
        """
        set_date(date(2016, 1, 1))

        self.sa_session.delete(self.service_level)

        service_level_pv_1 = self._create_tbs_service_level(self.tbs_account, name='SL PV 1',
            year=today().year, admin_fee=Decimal('300.00'), level=1, product_type=PvcTbsProduct.TYPE_PV_RELATED, guarantee_forecast_min_value=0)
        service_level_pv_2 = self._create_tbs_service_level(self.tbs_account, name='SL PV 2',
            year=today().year, admin_fee=Decimal('400.00'), level=2, product_type=PvcTbsProduct.TYPE_PV_RELATED, guarantee_forecast_min_value=100)

        service_level_other_1 = self._create_tbs_service_level(self.tbs_account, name='SL Other 1', year=today().year,
            admin_fee=Decimal('500.00'), level=1, guarantee_forecast_min_value=0)
        service_level_other_2 = self._create_tbs_service_level(self.tbs_account, name='SL Other 2', year=today().year,
            admin_fee=Decimal('600.00'), level=2, guarantee_forecast_min_value=100)

        # set TBS config for guarantee forecast.
        self.tbs.set_config('pvc.tbs.guarantee.forecast', True)

        # setup tbs products
        tbs_pv_product = self.pc.save_or_update_tbs_product(None, 'my pv product', 'desc', self.tbs.id,
            product_type=PvcTbsProduct.TYPE_PV_RELATED, basis_for_invoicing=PvcTbsProduct.INVOICE_BASED_ON_WEIGHT,
            custom_unit=PvcTbsProduct.CUSTOM_UNIT_NOMINAL_POWER
        )
        tbs_batteries_product = self.pc.save_or_update_tbs_product(None, 'my batteries product', 'desc', self.tbs.id,
            product_type=PvcTbsProduct.TYPE_PV_RELATED, basis_for_invoicing=PvcTbsProduct.INVOICE_BASED_ON_WEIGHT
        )

        self.pc.save_contribution_fee(self.tbs_product.id, service_level_other_2.id, Decimal('0.025'), today(), PvcTbsProductFee.DIMENSION_NONE)
        self.pc.save_contribution_fee(tbs_pv_product.id, service_level_pv_2.id, Decimal('0.030'), today(), PvcTbsProductFee.DIMENSION_NONE)
        self.pc.save_contribution_fee(tbs_batteries_product.id, service_level_pv_2.id, Decimal('0.035'), today(), PvcTbsProductFee.DIMENSION_NONE)

        # setup member
        member_account = self._create_member_account()
        tbs_member = self._member_apply_at_tbs(member_account, self.tbs_account, service_level=None)
        self.assertEqual(tbs_member.status, PvcTbsMember.STATUS_PENDING)

        # initial Guarantee forecast set SL for 2016
        cc = PvcContributionOrderController(self.sa_session, account_id=member_account.id)
        order = cc.add_guarantee_forecast_order(tbs_member.id, today().year, [tbs_pv_product.id], is_b2b=None)

        order = cc.update_contribution(order.id, {
            # pv module
            'line_0_quantity': 1,
            'line_0_weight': Decimal('1'),
            'line_0_custom_unit': Decimal('1'),
        })
        order = cc.submit_order(order.id)

        # current SL is pv1
        self.assertEqual(tbs_member.active_service_level.service_level, None)
        self.assertEqual(tbs_member.active_service_level.pv_service_level, service_level_pv_1)

        # while the member is still pending, do a resubmit..
        order.resubmit_request = PvcContributionOrder.RESUBMIT_REQUEST_APPROVED
        self.sa_session.add(order)
        self.sa_session.flush()

        # going higher
        order = cc.update_contribution(order.id, {
            'line_0_quantity': 101,
            'line_0_weight': Decimal('101'),
            'line_0_custom_unit': Decimal('101'),
        })
        cc.submit_order(order.id)

        # the new SL should be active already
        self.assertEqual(tbs_member.get_service_level_for_date(today()).service_level, None)
        self.assertEqual(tbs_member.get_service_level_for_date(today()).pv_service_level, service_level_pv_2)

    def test_member_guarantee_same_sl_af(self):
        """
        tests ticket #1014
        """
        set_date(date(2015, 12, 1))

        self.sa_session.delete(self.service_level)

        # create a certificate for this tbs for 2015, 2016
        tbsc = PvcTBSController(self.sa_session, account_id=self.operator_account.id)
        tbsc.save_tbs_certificate_state(None, {'tbs_id': self.tbs.default_group.license.id, 'year': today().year, 'text': 'certificate text 2015'})
        tbsc.save_tbs_certificate_state(None, {'tbs_id': self.tbs.default_group.license.id, 'year': today().year + 1, 'text': 'certificate text 2016'})

        service_level_pv_1 = self._create_tbs_service_level(self.tbs_account, name='SL PV 1',
            year=today().year, admin_fee=Decimal('0'), level=1, product_type=PvcTbsProduct.TYPE_PV_RELATED, guarantee_forecast_min_value=0)
        service_level_pv_2 = self._create_tbs_service_level(self.tbs_account, name='SL PV 2',
            year=today().year, admin_fee=Decimal('0'), level=2, product_type=PvcTbsProduct.TYPE_PV_RELATED, guarantee_forecast_min_value=100)

        service_level_other_1 = self._create_tbs_service_level(self.tbs_account, name='SL Other 1', year=today().year,
            admin_fee=Decimal('0'), level=1, guarantee_forecast_min_value=0)
        service_level_other_2 = self._create_tbs_service_level(self.tbs_account, name='SL Other 2', year=today().year,
            admin_fee=Decimal('0'), level=2, guarantee_forecast_min_value=100)

        # set TBS config for guarantee forecast.
        self.tbs.set_config('pvc.tbs.guarantee.forecast', True)

        # setup tbs products
        tbs_pv_product = self.pc.save_or_update_tbs_product(None, 'my pv product', 'desc', self.tbs.id,
            product_type=PvcTbsProduct.TYPE_PV_RELATED, basis_for_invoicing=PvcTbsProduct.INVOICE_BASED_ON_WEIGHT,
            custom_unit=PvcTbsProduct.CUSTOM_UNIT_NOMINAL_POWER
        )
        tbs_batteries_product = self.pc.save_or_update_tbs_product(None, 'my batteries product', 'desc', self.tbs.id,
            product_type=PvcTbsProduct.TYPE_PV_RELATED, basis_for_invoicing=PvcTbsProduct.INVOICE_BASED_ON_WEIGHT
        )

        self.pc.save_contribution_fee(self.tbs_product.id, service_level_other_2.id, Decimal('0.025'), today(), PvcTbsProductFee.DIMENSION_NONE)
        self.pc.save_contribution_fee(tbs_pv_product.id, service_level_pv_2.id, Decimal('0.030'), today(), PvcTbsProductFee.DIMENSION_NONE)
        self.pc.save_contribution_fee(tbs_batteries_product.id, service_level_pv_2.id, Decimal('0.035'), today(), PvcTbsProductFee.DIMENSION_NONE)

        # setup member
        member_account = self._create_member_account()
        tbs_member = self._member_apply_at_tbs(member_account, self.tbs_account, service_level=None)
        self.assertEqual(tbs_member.status, PvcTbsMember.STATUS_PENDING)

        # initial Guarantee forecast set SL for 2015
        cc = PvcContributionOrderController(self.sa_session, account_id=member_account.id)
        order = cc.add_guarantee_forecast_order(tbs_member.id, today().year, [tbs_pv_product.id], is_b2b=None)

        order = cc.update_contribution(order.id, {
            # pv module
            'line_0_quantity': 101,
            'line_0_weight': Decimal('101'),
            'line_0_custom_unit': Decimal('101'),
        })

        # submit and expect one service level to be active for this tbs_member
        order = cc.submit_order(order.id)

        # goto 2016
        set_date(date(2016, 1, 12))

        # approve member
        mc = PvcMemberController(self.sa_session, account_id=self.tbs_account.id)
        mc.save_tbs_member(tbs_member.id, {'status': PvcTbsMember.STATUS_APPROVED, 'is_full_member': False, 'is_representative': False})

        # GF for 2016, same SL
        order = cc.add_guarantee_forecast_order(tbs_member.id, today().year, [tbs_pv_product.id], is_b2b=None)
        order = cc.update_contribution(order.id, {
            # pv module
            'line_0_quantity': 101,
            'line_0_weight': Decimal('101'),
            'line_0_custom_unit': Decimal('101'),

        }, additional_lines={
        })
        order = cc.submit_order(order.id)

        # trigger new year FEE, this used to error
        PvcOrderController(self.sa_session).trigger_yearly_administrative_fees(self.tbs.default_group.license)

