# Copyright (c) 2014 BaseSoft B.V.
# See LICENSE for details.
from decimal import Decimal
from datetime import date

from basesoft.util.virtual_date import set_date, today

from pvcycle.test.unit import pvc_unittest
from pvcycle.core.controllers import PvcMemberController
from pvcycle.core.models import PvcTbsMember
from pvcycle.report.controllers import PvcHistoricalDataController
from basesoft.web.core.exceptions import PublicUserException, SecurityException, PermissionDeniedException
from pvcycle.report.models import PvcMemberYearlyAggregatedData

class TestCase(pvc_unittest.PvcTestCase):
    def setUp(self):
        self.connect_db()

        self.reset_roe()
        set_date(date(2014,1,1))

        self.operator = self._create_operator_account()
        self._create_contribution_tax_rates(self.operator)

        # this TBS has it's setting on MONTHLY!
        self.tbs_account = self._create_tbs_account(self.operator)
        self.accountant = self._create_accountant_account(self.tbs_account.default_group.license, self.operator)

        self._administration_fee = Decimal('250.00')
        self._contribution_fee = Decimal('2.75')

        self.service_level = self._create_tbs_service_level(self.tbs_account, admin_fee=self._administration_fee, contrib_fee=self._contribution_fee, year=today().year)

        self.ma = self._create_member_account()

        # apply at TBS
        self.tm = self._member_apply_at_tbs(self.ma, self.tbs_account, self.service_level)

        # activate member to automatically create an order entry for the Administration Fee
        mc = PvcMemberController(self.sa_session, account_id=self.tbs_account.id)
        mc.save_tbs_member(self.tm.id, {'status': PvcTbsMember.STATUS_APPROVED, 'is_full_member': False, 'is_representative': False})

        self.controller = PvcHistoricalDataController(self.sa_session, account_id=self.ma.id)

    def test_yearly_historical_data_failures(self):
        self._update_creation_date(self.tm, date(2014, 1, 1))

        # assert with wrong year
        self.assertRaises(PublicUserException, self.controller.save_or_update_historical_data, None, self.tm.id, 2005, 155, 1000, 2000, 3000, Decimal('23.45'), Decimal('76.55'))
        # assert with wrong country
        self.assertRaises(PublicUserException, self.controller.save_or_update_historical_data, None, self.tm.id, 2012, 1, 1000, 2000, 3000, Decimal('23.45'), Decimal('76.55'))
        # assert with wrong year
        self.assertRaises(PublicUserException, self.controller.save_or_update_historical_data, None, self.tm.id, 2014, 155, 1000, 2000, 3000, Decimal('23.45'), Decimal('76.55'))
        # assert with wrong b2b+b2c
        self.assertRaises(PublicUserException, self.controller.save_or_update_historical_data, None, self.tm.id, 2012, 155, 1000, 2000, 3000, Decimal('10.00'), Decimal('25.00'))
        # assert with negative power
        self.assertRaises(AssertionError, self.controller.save_or_update_historical_data, None, self.tm.id, 2012, 155, -1000, 2000, 3000, Decimal('23.45'), Decimal('76.55'))
        # assert with negative quantity
        self.assertRaises(AssertionError, self.controller.save_or_update_historical_data, None, self.tm.id, 2012, 155, 1000, -2000, 3000, Decimal('23.45'), Decimal('76.55'))
        # assert with negative weight
        self.assertRaises(AssertionError, self.controller.save_or_update_historical_data, None, self.tm.id, 2012, 155, 1000, 2000, -3000, Decimal('23.45'), Decimal('76.55'))
        # assert with negative b2b
        self.assertRaises(AssertionError, self.controller.save_or_update_historical_data, None, self.tm.id, 2012, 155, 1000, 2000, 3000, Decimal('-23.45'), Decimal('76.55'))
        # assert with negative b2c
        self.assertRaises(AssertionError, self.controller.save_or_update_historical_data, None, self.tm.id, 2012, 155, 1000, 2000, 3000, Decimal('23.45'), Decimal('-76.55'))

    def test_yearly_historical_data_creation(self):
        hd = self.controller.save_or_update_historical_data(None, self.tm.id, 2012, 155, Decimal('100.1'), 200, Decimal('300.3'), Decimal('23.45'), Decimal('76.55'))

        self.assertEqual(hd.year, 2012)
        self.assertEqual(hd.total_megawatt, Decimal('100.1'))
        self.assertEqual(hd.total_quantity, 200)
        self.assertEqual(hd.total_weight, Decimal('300.3'))
        self.assertEqual(hd.tbs_member, self.tm)
        self.assertEqual(hd.country.iso_3166_a2, 'NL')

        #test duplicate for key TBS MEMBER, YEAR, COUNTRY
        self.assertRaises(PublicUserException, self.controller.save_or_update_historical_data, None, self.tm.id, 2012, 155, 1000, 2000, 3000, Decimal('23.45'), Decimal('76.55'))

    def test_yearly_historical_data_update(self):
        hd = self.controller.save_or_update_historical_data(None, self.tm.id, 2012, 155, 1000, 2000, 3000, Decimal('23.45'), Decimal('76.55'))
        hd = self.controller.save_or_update_historical_data(hd.id, self.tm.id, 2012, 155, 1100, 2100, 3100, Decimal('23.45'), Decimal('76.55'))
        self.assertEqual(hd.total_megawatt, 1100)
        self.assertEqual(hd.total_quantity, 2100)
        self.assertEqual(hd.total_weight, 3100)

        # test wrong year at update
        self.assertRaises(PublicUserException, self.controller.save_or_update_historical_data, hd.id, self.tm.id, 2013, 155, 1100, 2100, 3100, Decimal('23.45'), Decimal('76.55'))

        ma = self._create_member_account()
        tm = self._member_apply_at_tbs(ma, self.tbs_account, self.service_level)

        # test without rights for TBS Member at update
        self.assertRaises(SecurityException, self.controller.save_or_update_historical_data, hd.id, tm.id, 2013, 155, 1100, 2100, 3100, Decimal('23.45'), Decimal('76.55'))

        tbs_account = self._create_tbs_account(self.operator)
        service_level = self._create_tbs_service_level(tbs_account, admin_fee=self._administration_fee, contrib_fee=self._contribution_fee, year=today().year)

        tm = self._member_apply_at_tbs(self.ma, tbs_account, service_level)
        # test wrong TBS Member at update
        self.assertRaises(PublicUserException, self.controller.save_or_update_historical_data, hd.id, tm.id, 2013, 155, 1100, 2100, 3100, Decimal('23.45'), Decimal('76.55'))

    def test_yearly_historical_data_update_by_tbs(self):
        hd = self.controller.save_or_update_historical_data(None, self.tm.id, 2012, 155, 1000, 2000, 3000, Decimal('23.45'), Decimal('76.55'))

        controller = PvcHistoricalDataController(self.sa_session, account_id=self.tbs_account.id)
        hd = controller.save_or_update_historical_data(hd.id, self.tm.id, 2012, 155, 1100, 2100, 3100, Decimal('23.45'), Decimal('76.55'))

        self.assertEqual(hd.total_megawatt, 1100)
        self.assertEqual(hd.total_quantity, 2100)
        self.assertEqual(hd.total_weight, 3100)

        # test no rights to member
        tbs_account = self._create_tbs_account(self.operator)
        service_level = self._create_tbs_service_level(tbs_account, admin_fee=self._administration_fee, contrib_fee=self._contribution_fee, year=today().year)

        tm = self._member_apply_at_tbs(self.ma, tbs_account, service_level)
        hd = self.controller.save_or_update_historical_data(None, tm.id, 2012, 155, 1000, 2000, 3000, Decimal('23.45'), Decimal('76.55'))
        self.assertRaises(SecurityException, controller.save_or_update_historical_data, hd.id, tm.id, 2012, 155, 1100, 2100, 3100, Decimal('23.45'), Decimal('76.55'))


    def test_yearly_historical_data_update_by_operator(self):
        hd = self.controller.save_or_update_historical_data(None, self.tm.id, 2012, 155, 1000, 2000, 3000, Decimal('23.45'), Decimal('76.55'))

        controller = PvcHistoricalDataController(self.sa_session, account_id=self.operator.id)
        hd = controller.save_or_update_historical_data(hd.id, self.tm.id, 2012, 155, 1100, 2100, 3100, Decimal('23.45'), Decimal('76.55'))

        self.assertEqual(hd.total_megawatt, 1100)
        self.assertEqual(hd.total_quantity, 2100)
        self.assertEqual(hd.total_weight, 3100)

    def test_save_historical_data_from_state(self):
        controller = PvcHistoricalDataController(self.sa_session, account_id=self.ma.id)
        data = controller.save_from_state(self.tm.id, 2012, [155], {
            'megawatt_155': 12,
            'quantity_155': 13,
            'weight_155': 14,
            'b2b_155': 75,
            'b2c_155': 25,
        })

        hd = data[0]
        self.assertEqual(hd.year, 2012)
        self.assertEqual(hd.total_megawatt, 12)
        self.assertEqual(hd.total_quantity, 13)
        self.assertEqual(hd.total_weight, 14)
        self.assertEqual(hd.distribution_b2b, Decimal('75.00'))

    def test_save_historical_data_from_state_multi_country(self):

        state = self._get_tbs_license_state()
        state['tbs']['country'] = None

        tbs_account = self._create_tbs_account(self.operator, state=state)
        service_level = self._create_tbs_service_level(tbs_account, admin_fee=self._administration_fee, contrib_fee=self._contribution_fee, year=today().year)

        tm = self._member_apply_at_tbs(self.ma, tbs_account, service_level)


        controller = PvcHistoricalDataController(self.sa_session, account_id=self.ma.id)
        data = controller.save_from_state(tm.id, 2012, [128, 136], {
            'megawatt_128': 12,
            'quantity_128': 13,
            'weight_128': 14,
            'b2b_128': 75,
            'b2c_128': 25,
            'megawatt_136': 120,
            'quantity_136': 130,
            'weight_136': 140,
            'b2b_136': 16,
            'b2c_136': 84,
        })
        self.assertEqual(len(data), 2)

    def test_delete_historical_data(self):
        hd = self.controller.save_or_update_historical_data(None, self.tm.id, 2012, 155, 1000, 2000, 3000, Decimal('23.45'), Decimal('76.55'))
        self.assertRaises(PermissionDeniedException, self.controller.delete_historical_data, hd.id)

        controller = PvcHistoricalDataController(self.sa_session, account_id=self.tbs_account.id)
        controller.delete_historical_data(hd.id)

        r = self.sa_session.query(PvcMemberYearlyAggregatedData).filter(PvcMemberYearlyAggregatedData.id==hd.id).first()
        self.assertEqual(r, None)


# below is for running the test from eclipse
if __name__ == '__main__':

    from twisted.scripts.trial import run
    import sys
    sys.argv = ['test_historical.py', 'pvcycle.test.unit.test_historical_data.TestCase.test_save_historical_data_from_state_multi_country']
    run()
    sys.exit()

