'''
Created on Jul 22, 2013

@author: marten
'''

from basesoft.web.contrib.license.controllers import LicenseController
from basesoft.web.core.exceptions import SecurityException, PublicSecurityException, PublicUserException

from pvcycle.test.unit import pvc_unittest
from pvcycle.core.models import PvcOperatorLicense

from pvcycle.pvmodule.controllers import PvcPVModuleController
from pvcycle.pvmodule.models import PvcPVModuleSupplier, PvcPVModuleType, PvcPVModuleCellType, PvcPVModule
from sqlalchemy.exc import IntegrityError

class TestCase(pvc_unittest.PvcTestCase):
    def setUp(self):
        self.connect_db()

        # we need an account (sysadmin) in order to create a new operator
        self.account = self._create_active_account()
        self.group = self.account.groups[0].group

        # this is all tested in IQPortal.
        self._attach_permissions_to_group(self.group,
            'view_license', 'view_all_domains', 'view_group', 'create_domain', 'create_license',
            'distribute_license_permissions', 'distribute_group_permissions'
        )

    def test_create_operator(self):
        """
        This tests the creation of the Operator license and it's underlying group(s).
        """
        lc = LicenseController(self.sa_session, account_id=self.account.id)

        self.assertRaises(SecurityException, lc.save_license_from_state, self._get_operator_license_state())
        self._attach_permissions_to_group(self.group, 'pvc_operator_license')

        _license = lc.save_license_from_state(self._get_operator_license_state())

        # assert that we have a new PvcOperatorLicensee
        self.assertTrue(isinstance(_license, PvcOperatorLicense))

    def test_create_pv_module(self):
        # create an operator
        account = self._create_operator_account()

        controller = PvcPVModuleController(self.sa_session, account_id=account.id)

        name = self._uniquestring('pvmodule')

        # let the operator create a new PV Module (should have the permissions from the role)
        module = controller.save_or_update_module(None, name, 'Sunny', 'Automated', 24, 12, 10, 20)

        self.assertTrue(isinstance(module.supplier, PvcPVModuleSupplier))
        self.assertEqual(module.supplier.name, name)

        self.assertTrue(isinstance(module.type, PvcPVModuleType))
        self.assertEqual(module.type.name, 'Sunny')

        self.assertTrue(isinstance(module.cell_type, PvcPVModuleCellType))
        self.assertEqual(module.cell_type.name, 'Automated')

        self.assertEqual(module.nominal_power, 24)
        self.assertEqual(module.weight, 12)
        self.assertEqual(module.width, 10)
        self.assertEqual(module.length, 20)

    def test_duplicate_pv_module(self):
        # create an operator
        account = self._create_operator_account()

        name = self._uniquestring('pvmodule')

        controller = PvcPVModuleController(self.sa_session, account_id=account.id)
        controller.save_or_update_module(None, name, 'Sunny', 'Automated', 24, 12, 10, 20)

        # try to save again with same name
        self.assertRaises(PublicUserException, controller.save_or_update_module, None, name, 'Sunny', 'Automated', 24, 12, 10, 20)

    def test_approval_of_pv_module(self):
        """
        Approval test of PV Module added by the Operator.
        """
        account = self._create_operator_account()

        name = self._uniquestring('pvmodule')

        controller = PvcPVModuleController(self.sa_session, account_id=account.id)
        module = controller.save_or_update_module(None, name, 'Sunny', 'Automated', 24, 12, 10, 20)

        # operator can approve, so first approval is automatically done.
        self.assertEqual(module.first_approval_by, account)
        self.assertEqual(module.approval_status, PvcPVModule.PV_MODULE_STATUS_PENDING)
        self.assertRaises(PublicSecurityException, controller.approve_module, module.id)

        new_account = self._create_operator_account()
        self.assertNotEqual(account, new_account)

        controller = PvcPVModuleController(self.sa_session, account_id=new_account.id)
        controller.approve_module(module.id)
        self.sa_session.expire(module)

        self.assertEqual(module.approval_status, PvcPVModule.PV_MODULE_STATUS_APPROVED)
        self.assertEqual(module.second_approval_by, new_account)





