'''
Created on Oct 8, 2013

@author: marten
'''

import time

from decimal import Decimal

from basesoft.util.virtual_date import today, set_date, now
from basesoft.web.contrib.notification.models import Notification

from pvcycle.test.unit import pvc_unittest
from pvcycle.core.controllers import PvcMemberController, PvcTBSController
from pvcycle.core.models import PvcTbsMember, PvcTbsMemberCertificate
from pvcycle.order.controllers import PvcOrderController
from pvcycle.order.models import PvcOrder, PvcOrderLineAdministrationFee, PvcOrderLineMembershipFee
from basesoft.web.contrib.i18n.models import Country
from basesoft.web.core.exceptions import PublicUserException
from basesoft.web.contrib.order.controllers import PaymentController
from basesoft.web.contrib.order.models import PaymentMethodBankTransfer, OrderPaymentStatus, OrderRule
import datetime

#import pdb; pdb.set_trace()

class TestCase(pvc_unittest.PvcTestCase):
    """
    pvcycle.test.unit.test_order.TestCase
    """

    def setUp(self):

        # jan 2014
        set_date(datetime.datetime(2014, 1, 1))

        # create operator, tbs's and their members
        self.connect_db()

        self.operator = self._create_operator_account()
        self.tbs = self._create_tbs_account(self.operator)

        # create an accountant for manual asserting with middleware app
        self.accountant = self._create_accountant_account(self.tbs.default_group.license, self.operator)

        self._administration_fee = Decimal('250.00')

        self.service_level = self._create_tbs_service_level(self.tbs, admin_fee=self._administration_fee, year=today().year)

        self._trash_all_rules()

    def _make_rule(self, name, **kwargs):

        controller = PvcOrderController(self.sa_session, account_id=self.operator.id)

        controller.save_rule(None, {
            'order_type': kwargs.get('order_type', 'MockOrder'),
            'name': name,
            'enabled': kwargs.get('enabled', True),
            'execute_order': kwargs.get('execute_order', None),
            'description': kwargs.get('description', ''),
            'start_date': kwargs.get('start_date', now()),
            'end_date': kwargs.get('end_date', None),
            'match_all': kwargs.get('match_all', True),
            'stop_processing': kwargs.get('stop_processing', False),
            'conditions': kwargs.get('conditions', []),
            'actions': kwargs.get('actions', [])
        })

    def test_member_administration_fee_order(self):
        """
        pvcycle.test.unit.test_order.TestCase.test_member_administration_fee_order
        """

        # create a member
        ma = self._create_member_account()

        # set the country to NL so that VAT is added. (The Country of the TBS is NL -> ticket #3953)
        ma.default_group.license.company.addresses[0].address.country = self.sa_session.query(Country).filter(Country.iso_3166_a2=='NL').one()
        self.sa_session.add(ma.default_group.license.company)
        self.sa_session.flush()

        # apply at TBS
        tm = self._member_apply_at_tbs(ma, self.tbs, self.service_level)

        # activate member to automatically create an order entry for the Administration Fee
        mc = PvcMemberController(self.sa_session, account_id=self.tbs.id)
        mc.save_tbs_member(tm.id, {'status': PvcTbsMember.STATUS_APPROVED, 'is_full_member': False, 'is_representative': False})

        # verify the order to be created
        poc = PvcOrderController(self.sa_session, account_id=ma.id)
        orders = poc.get()

        self.assertEqual(len(orders), 1)
        order = orders[0]
        self.assertTrue(isinstance(order, PvcOrder))

        self.assertEqual(len(order.lines), 1)
        self.assertTrue(isinstance(order.lines[0], PvcOrderLineAdministrationFee))
        self.assertEqual(order.lines[0].price_per_unit, self._administration_fee)
        self.assertEqual(order.total_amount_excluding_vat, self._administration_fee)

        self.assertEqual(order.lines[0].vat_rate, Decimal('21.0'))
        self.assertEqual(order.lines[0].line_amount_including_vat, Decimal('302.50000'))

        # verify its not possible to apply again
        self.assertRaises(PublicUserException, self._member_apply_at_tbs, ma, self.tbs, self.service_level)

        # apply at 2nd tbs
        tbs2 = self._create_tbs_account(self.operator)
        sl2 = self._create_tbs_service_level(tbs2, admin_fee=self._administration_fee, year=today().year)
        tm2 = self._member_apply_at_tbs(ma, tbs2, sl2)

        # activate member at 2nd tbs
        mc = PvcMemberController(self.sa_session, account_id=tbs2.id)
        mc.save_tbs_member(tm2.id, {'status': PvcTbsMember.STATUS_APPROVED, 'is_full_member': False, 'is_representative': False})

        # verfify there are 2 orders
        poc = PvcOrderController(self.sa_session, account_id=ma.id)
        orders = poc.get()
        self.assertEqual(len(orders), 2)

        # suspend account at 2nd tbs and activate again
        mc = PvcMemberController(self.sa_session, account_id=tbs2.id)
        mc.save_tbs_member(tm2.id, {'status': PvcTbsMember.STATUS_SUSPENDED, 'is_full_member': False, 'is_representative': False})
        mc.save_tbs_member(tm2.id, {'status': PvcTbsMember.STATUS_APPROVED, 'is_full_member': False, 'is_representative': False})

        # there should still be 2 orders
        poc = PvcOrderController(self.sa_session, account_id=ma.id)
        orders = poc.get()
        self.assertEqual(len(orders), 2)

        # goto 2015
        set_date(datetime.datetime(2015, 1, 12))

        # similar as the task, trigger new AF orders for current tbs
        PvcOrderController(self.sa_session).trigger_yearly_administrative_fees(self.tbs.default_group.license)

        # there should be 3 orders now
        self.assertEqual(len(poc.get()), 3)


    def test_member_administration_fee_order_with_rule(self):
        """
        pvcycle.test.unit.test_order.TestCase.test_member_administration_fee_order
        """

        self._make_rule('af_roel',
            order_type = 'PvcOrder',
            actions = [{
                'action': 'relative_amount_discount',
                'description': '',
                'operator': None, 'value_text': None, 'value_bool': None, 'value_int': None,
                'value_decimal': 100,
                'tax_class_id': None, 'gl_account_code': None
            }]
        )

        # create a member
        ma = self._create_member_account()

        # apply at TBS
        tm = self._member_apply_at_tbs(ma, self.tbs, self.service_level)

        # activate member to automatically create an order entry for the Administration Fee
        mc = PvcMemberController(self.sa_session, account_id=self.tbs.id)
        mc.save_tbs_member(tm.id, {'status': PvcTbsMember.STATUS_APPROVED, 'is_full_member': False, 'is_representative': False})

        # verify the order to be created
        poc = PvcOrderController(self.sa_session, account_id=ma.id)
        orders = poc.get()

        self.assertEqual(len(orders), 1)
        order = orders[0]
        self.assertTrue(isinstance(order, PvcOrder))

        self.assertEqual(len(order.lines), 2)
        self.assertEqual(order.total_amount_excluding_vat, Decimal('0'))

    def _trash_all_rules(self):
        for rule in self.sa_session.query(OrderRule).filter(OrderRule.domain==self.operator.default_group.license.domain):
            self.sa_session.delete(rule)

    def test_certificate(self):
        """
        pvcycle.test.unit.test_order.TestCase.test_certificate
        """

        # create a certificate for this tbs for 2014
        tbsc = PvcTBSController(self.sa_session, account_id=self.operator)
        tbsc.save_tbs_certificate_state(None, {'tbs_id': self.tbs.default_group.license.id, 'year': today().year, 'text': 'certificate text'})

        # create a member
        ma = self._create_member_account()

        # setup payment methods
        pc = PaymentController(self.sa_session, account_id=self.operator.id)
        pm1 = pc.save_or_update_payment_method_for_merchant(PaymentMethodBankTransfer, label='bank transfer', merchant=self.tbs.default_group.license)

        # set the country to NL so that VAT is added. (The Country of the TBS is NL -> ticket #3953)
        ma.default_group.license.company.addresses[0].address.country = self.sa_session.query(Country).filter(Country.iso_3166_a2=='NL').one()
        self.sa_session.add(ma.default_group.license.company)
        self.sa_session.flush()

        # apply at TBS
        tm = self._member_apply_at_tbs(ma, self.tbs, self.service_level)

        # activate member to automatically create an order entry for the Administration Fee
        mc = PvcMemberController(self.sa_session, account_id=self.tbs.id)
        mc.save_tbs_member(tm.id, {'status': PvcTbsMember.STATUS_APPROVED, 'is_full_member': False, 'is_representative': False})

        # there should be an order
        poc = PvcOrderController(self.sa_session, account_id=ma.id)
        orders = poc.get()
        self.assertEqual(len(orders), 1)
        order = orders[0]

        # assert no certificate is present yet
        certificate = self.sa_session.query(PvcTbsMemberCertificate).filter( PvcTbsMemberCertificate.tbs_member == tm, PvcTbsMemberCertificate.year == today().year ).first()
        self.assertTrue(certificate is None)

        # let the accountant approve the order
        accountant_pc = PaymentController(self.sa_session, account_id=self.accountant, payment_method=pm1)
        time.sleep(1); accountant_pc.complete_payment(order.id, "all good")
        self.assertEqual(order.get_current_payment_status(), OrderPaymentStatus.PAYMENT_COMPLETED)

        # there should be a certificate now
        certificate = self.sa_session.query(PvcTbsMemberCertificate).filter( PvcTbsMemberCertificate.tbs_member == tm, PvcTbsMemberCertificate.year == today().year ).first()
        self.assertTrue(certificate is not None)
        self.assertTrue(certificate.hash is not None)

        # verify that the member cannot yet download his certificate
        self.assertRaises(PublicUserException, mc.get_certificate, tm.id, today().year)

        # let the operator approve the certificate
        certificate = self.sa_session.query(PvcTbsMemberCertificate) \
            .filter(PvcTbsMemberCertificate.tbs_member == tm, PvcTbsMemberCertificate.year == today().year, PvcTbsMemberCertificate.status == PvcTbsMemberCertificate.STATUS_PAID).one()
        mc_op = PvcMemberController(self.sa_session, account_id=self.operator.id)
        mc_op.update_certificate_status(tm.id, today().year, PvcTbsMemberCertificate.STATUS_APPROVED)

        # check that notification is sent and sender is as expected
        notif = self.sa_session.query(Notification).order_by(Notification.id.desc()).first()
        self.assertEqual(notif.sender, tm.tbs.default_group)

        # now it should be possible to download the certificate
        mc.get_certificate(tm.id, today().year)

        #
        # goto 2015
        #
        set_date(datetime.datetime(2015, 1, 12))

        # create a certificate for 2015
        tbsc = PvcTBSController(self.sa_session, account_id=self.operator)
        tbsc.save_tbs_certificate_state(None, {'tbs_id': self.tbs.default_group.license.id, 'year': today().year, 'text': 'certificate text'})

        # similar as the task, trigger new AF orders for current tbs
        PvcOrderController(self.sa_session).trigger_yearly_administrative_fees(self.tbs.default_group.license)

        # there should be a new order
        poc = PvcOrderController(self.sa_session, account_id=ma.id)
        orders = poc.get()
        self.assertEqual(len(orders), 2)
        order = orders[1]

        # assert no certificate is present yet
        certificate = self.sa_session.query(PvcTbsMemberCertificate).filter( PvcTbsMemberCertificate.tbs_member == tm, PvcTbsMemberCertificate.year == today().year ).first()
        self.assertTrue(certificate is None)

        # let the accountant approve the order
        accountant_pc = PaymentController(self.sa_session, account_id=self.accountant, payment_method=pm1)
        time.sleep(1); accountant_pc.complete_payment(order.id, "all good")
        self.assertEqual(order.get_current_payment_status(), OrderPaymentStatus.PAYMENT_COMPLETED)

        # there should be a certificate now
        certificate = self.sa_session.query(PvcTbsMemberCertificate).filter( PvcTbsMemberCertificate.tbs_member == tm, PvcTbsMemberCertificate.year == today().year ).first()
        self.assertTrue(certificate is not None)
        self.assertTrue(certificate.hash is not None)

        # verify that the member cannot yet download his certificate
        self.assertRaises(PublicUserException, mc.get_certificate, tm.id, today().year)

        # let the operator approve the certificate
        certificate = self.sa_session.query(PvcTbsMemberCertificate) \
            .filter(PvcTbsMemberCertificate.tbs_member == tm, PvcTbsMemberCertificate.year == today().year, PvcTbsMemberCertificate.status == PvcTbsMemberCertificate.STATUS_PAID).one()
        mc_op = PvcMemberController(self.sa_session, account_id=self.operator.id)
        mc_op.update_certificate_status(tm.id, today().year, PvcTbsMemberCertificate.STATUS_APPROVED)

        # now it should be possible to download the certificate
        mc.get_certificate(tm.id, today().year)


    def test_certificate_zero_admin_fee(self):
        """
        pvcycle.test.unit.test_order.TestCase.test_certificate_zero_admin_fee
        """

        # create a certificate for this tbs for 2014
        tbsc = PvcTBSController(self.sa_session, account_id=self.operator)
        tbsc.save_tbs_certificate_state(None, {'tbs_id': self.tbs.default_group.license.id, 'year': today().year, 'text': 'certificate text'})

        # create a member
        ma = self._create_member_account()

        # zero fee
        service_level_zero_fee = self._create_tbs_service_level(self.tbs, name='zero fee', admin_fee=Decimal(0), year=today().year, level=2)

        # apply at TBS
        tm = self._member_apply_at_tbs(ma, self.tbs, service_level_zero_fee)

        # activate member to automatically create an order entry for the Administration Fee
        mc = PvcMemberController(self.sa_session, account_id=self.tbs.id)
        mc.save_tbs_member(tm.id, {'status': PvcTbsMember.STATUS_APPROVED, 'is_full_member': False, 'is_representative': False})

        # there should be an order
        poc = PvcOrderController(self.sa_session, account_id=ma.id)
        orders = poc.get()
        self.assertEqual(len(orders), 1)
        order = orders[0]

        # without approving the order there should be a certificate
        certificate = self.sa_session.query(PvcTbsMemberCertificate).filter( PvcTbsMemberCertificate.tbs_member == tm, PvcTbsMemberCertificate.year == today().year ).first()

        # there should be a certificate now with status paid (not yet approved)
        self.assertTrue(certificate is not None)
        self.assertTrue(certificate.hash is not None)
        self.assertEqual(certificate.status, PvcTbsMemberCertificate.STATUS_PAID)

        #
        # goto 2015
        #
        set_date(datetime.datetime(2015, 1, 12))

        # create a certificate for 2015
        tbsc = PvcTBSController(self.sa_session, account_id=self.operator)
        tbsc.save_tbs_certificate_state(None, {'tbs_id': self.tbs.default_group.license.id, 'year': today().year, 'text': 'certificate text'})

        # similar as the task, trigger new AF orders for current tbs
        PvcOrderController(self.sa_session).trigger_yearly_administrative_fees(self.tbs.default_group.license)

        # there should be an order
        poc = PvcOrderController(self.sa_session, account_id=ma.id)
        orders = poc.get()
        self.assertEqual(len(orders), 2)
        order = orders[1]

        # without approving the order there should be a certificate
        certificate = self.sa_session.query(PvcTbsMemberCertificate).filter( PvcTbsMemberCertificate.tbs_member == tm, PvcTbsMemberCertificate.year == today().year ).first()

        # there should be a certificate now with status paid (not yet approved)
        self.assertTrue(certificate is not None)
        self.assertTrue(certificate.hash is not None)
        self.assertEqual(certificate.status, PvcTbsMemberCertificate.STATUS_PAID)


    def test_membership_fee(self):
        """
        Membership fees are used for the UK DTS TBS.

        pvcycle.test.unit.test_order.TestCase.test_membership_fee
        """

        state = self._get_tbs_license_state()
        state['configs']['pvc.tbs.use.membership.instead.of.service.level'] = True
        state['configs']['pvc.tbs.auto.approve.member'] = True
        state['configs']['pvc.tbs.membership.valid.for.years'] = 3

        tbs_account = self._create_tbs_account(self.operator, state=state)
        accountant = self._create_accountant_account(tbs_account.default_group.license, self.operator)

        membership_fee = Decimal('375.00')
        memberships = self._create_memberships(tbs_account, fee=membership_fee)

        # setup payment methods
        pc = PaymentController(self.sa_session, account_id=self.operator.id)
        pm1 = pc.save_or_update_payment_method_for_merchant(PaymentMethodBankTransfer, label='bank transfer', merchant=tbs_account.default_group.license)

        # create a member
        ma = self._create_member_account()

        # apply at TBS
        tm = self._member_apply_at_tbs(ma, tbs_account, upload_contract=False, is_producer=False, membership=memberships[0])

        # expect 1 order to be made
        poc = PvcOrderController(self.sa_session, account_id=ma.id)
        orders = poc.get()
        self.assertEqual(len(orders), 1)

        order = orders.pop()
        self.assertTrue(isinstance(order.lines[0], PvcOrderLineMembershipFee))
        self.assertEqual(order.total_amount_excluding_vat, membership_fee)

        # assert no certificate is present yet
        certificate = self.sa_session.query(PvcTbsMemberCertificate).filter( PvcTbsMemberCertificate.tbs_member == tm, PvcTbsMemberCertificate.year == today().year ).first()
        self.assertTrue(certificate is None)

        # let the accountant approve the order
        accountant_pc = PaymentController(self.sa_session, account_id=accountant, payment_method=pm1)
        time.sleep(1); accountant_pc.complete_payment(order.id, "all good")
        self.assertEqual(order.get_current_payment_status(), OrderPaymentStatus.PAYMENT_COMPLETED)

        # there should be a certificate now
        certificate = self.sa_session.query(PvcTbsMemberCertificate).filter( PvcTbsMemberCertificate.tbs_member == tm, PvcTbsMemberCertificate.year == today().year ).first()
        self.assertTrue(certificate is not None)
        self.assertTrue(certificate.hash is not None)

        # goto 2015 and trigger new AF/Membership orders
        set_date(datetime.datetime(2015, 1, 12))
        PvcOrderController(self.sa_session).trigger_yearly_administrative_fees(tbs_account.default_group.license)

        # there should _NOT_ be a new order (this will be in 2014+3 = 2017)
        self.assertEqual(len(poc.get()), 1)

        # goto 2016 and trigger new AF/Membership orders
        set_date(datetime.datetime(2016, 1, 12))
        PvcOrderController(self.sa_session).trigger_yearly_administrative_fees(tbs_account.default_group.license)

        # there should _NOT_ be a new order (this will be in 2014+3 = 2017)
        self.assertEqual(len(poc.get()), 1)

        # goto 2017 and trigger new AF/Membership orders
        set_date(datetime.datetime(2017, 1, 12))
        PvcOrderController(self.sa_session).trigger_yearly_administrative_fees(tbs_account.default_group.license)

        # there should be a new order!
        self.assertEqual(len(poc.get()), 2)

        # # submit it
        # poc.submit_order(poc.get()[1].id)

        # assert accountant (asserts ticket #1024)
        adc = PvcOrderController(self.sa_session, account_id=accountant.id)
        states = adc.get_unexported_orders_for_merchant(tm.tbs.id)
        self.assertEqual(len(states), 2)


    def test_resign(self):
        """
        pvcycle.test.unit.test_order.TestCase.test_resign
        """

        # create a member
        ma = self._create_member_account()

        # apply at TBS
        tm = self._member_apply_at_tbs(ma, self.tbs, self.service_level)

        # activate member to automatically create an order entry for the Administration Fee
        mc = PvcMemberController(self.sa_session, account_id=self.tbs.id)
        mc.save_tbs_member(tm.id, {'status': PvcTbsMember.STATUS_APPROVED, 'is_full_member': False, 'is_representative': False})

        # verify the order to be created
        poc = PvcOrderController(self.sa_session, account_id=ma.id)
        orders = poc.get()
        self.assertEqual(len(orders), 1)

        # before going into 2015, resign
        mc.save_tbs_member(tm.id, {'status': PvcTbsMember.STATUS_RESIGNING, 'is_full_member': False, 'is_representative': False})

        # goto 2015
        set_date(datetime.datetime(2015, 1, 12))

        # similar as the task, trigger new AF orders for current tbs
        PvcOrderController(self.sa_session).trigger_yearly_administrative_fees(self.tbs.default_group.license)

        # the new tbs_member status should be 'resigned'
        tm = mc._get_tbs_member(tm.id)
        self.assertEqual(tm.status, PvcTbsMember.STATUS_RESIGNED)

        # there should be still only 1 order
        self.assertEqual(len(poc.get()), 1)

        # now approve the member again. expect that he gets the AF triggered still (#611)
        mc.save_tbs_member(tm.id, {'status': PvcTbsMember.STATUS_APPROVED, 'is_full_member': False, 'is_representative': False})
        self.assertEqual(len(poc.get()), 2)


# below is for running the test from eclipse
if __name__ == '__main__':

    from twisted.scripts.trial import run
    import sys
    sys.argv = ['test_member.py', 'pvcycle.test.unit.test_order.TestCase']
    run()
    sys.exit()

