# Copyright (c) 2015 BaseSoft B.V.
# See LICENSE for details.

from datetime import date
from decimal import Decimal

from basesoft.util.virtual_date import set_date, today, now
from pvcycle.core.controllers import PvcMemberController, PvcTBSController
from pvcycle.core.models import PvcTbsMember, PvcTbsLicense
from pvcycle.order.controllers import PvcContributionOrderController
from pvcycle.product.models import PvcTbsMemberProduct, PvcTbsProduct

from pvcycle.test.unit import pvc_unittest
from pvcycle.product.controllers import PvcProductController

class TestCase(pvc_unittest.PvcTestCase):

    def setUp(self):
        self.connect_db()

        self.operator_account = self._create_operator_account()

        set_date(date(2013, 1, 1)) # set date to January 2013
        self.reset_roe()

        self._create_contribution_tax_rates(self.operator_account)

        self.tbs_account = self._create_tbs_account(self.operator_account)
        self.tbs = self.tbs_account.default_group.license
        self.accountant = self._create_accountant_account(self.tbs_account.default_group.license, self.operator_account)

        self._administration_fee = Decimal('250.00')
        self._contribution_fee = Decimal('2.75')

        self.service_level = self._create_tbs_service_level(self.tbs_account, admin_fee=self._administration_fee, year=today().year)

        self.pc = PvcProductController(self.sa_session, account_id=self.operator_account.id)

        products = self.pc.get_available_products_for_tbs(self.tbs.id)
        self.tbs_product = self.pc.save_or_update_tbs_product(None, 'EEE-TBS-TEST', 'WEEE', self.tbs.id,
            parent_level_id=products[0]['id'], basis_for_invoicing=PvcTbsProduct.INVOICE_BASED_ON_WEIGHT
        )

        self.tbs_product2 = self.pc.save_or_update_tbs_product(None, 'EEE-TBS-TEST-2', 'WEEE-2', self.tbs.id,
            parent_level_id=products[1]['id'], basis_for_invoicing=PvcTbsProduct.INVOICE_BASED_ON_WEIGHT
        )

        self.pc.save_contribution_fee(self.tbs_product.id, self.service_level.id, Decimal('0.015'), today())
        self.pc.save_contribution_fee(self.tbs_product2.id, self.service_level.id, Decimal('0.035'), today())

        set_date(date(2013, 6, 1)) # set date to June 2013

    def _create_simple_member_for_declaration(self, obligation_start_date=None):
        ma = self._create_member_account()

        # apply at TBS
        tm = self._member_apply_at_tbs(ma, self.tbs_account, self.service_level)

        obligation_start_date = date(2012, 1, 1) if obligation_start_date is None else obligation_start_date

        # activate member to automatically create an order entry for the Administration Fee
        mc = PvcMemberController(self.sa_session, account_id=self.tbs_account.id)
        mc.save_tbs_member(tm.id, {
            'status': PvcTbsMember.STATUS_APPROVED,
            'is_full_member': False,
            'is_representative': False,
            'obligation_start_date': obligation_start_date})
        return ma, tm

    def _get_member_product(self, tm, tbs_product=None):

        if tbs_product is None:
            tbs_product = self.tbs_product

        # TODO this should be added with a controller method
        product = PvcTbsMemberProduct()
        product.tbs_product = tbs_product
        product.tbs_member = tm
        product.is_active = True
        product.country = tm.tbs.country
        self.sa_session.add(product)
        self.sa_session.flush()
        return product

    def test_periodicity_change_monthly_to_quarterly(self):
        ma, tm = self._create_simple_member_for_declaration()
        self._get_member_product(tm)

        tm.declaration_periodicity = PvcTbsLicense.DECLARATION_PERIOD_MONTHLY
        self.sa_session.add(tm)
        self.sa_session.flush()

        set_date(date(2014, 3, 1))

        mcc = PvcContributionOrderController(self.sa_session, account_id=ma.id)
        cc = PvcContributionOrderController(self.sa_session)

        # make january and february declarations
        cc.add_open_contribution_order(tm, 2014, 1, None)
        cc.add_open_contribution_order(tm, 2014, 2, None)

        set_date(date(2014, 3, 11))

        cc.close_contributions(tm.tbs)

        # set new periodicity
        mc = PvcMemberController(self.sa_session, account_id=self.tbs_account.id)
        mc.save_tbs_member(tm.id, {
            'status': tm.status,
            'is_full_member': tm.is_full_member,
            'is_representative': tm.is_representative,
            'declaration_periodicity': PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY
        })

        # verify that march get triggered correctly
        set_date(date(2014, 4, 1))
        orders = cc.trigger_new_contributions(tm.tbs)
        self.assertEqual(len(orders), 1)

        march = orders.pop()
        self.assertEqual(march.month, 3)

        # submit march declaration
        mcc.submit_order(march)

        # verify that new declaration period is quarterly
        self.sa_session.expire(tm)
        self.assertEqual(tm.declaration_period, PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY)
        self.assertEqual(tm.upcoming_periodicity_change, None)

        # now lets go to july
        set_date(date(2014, 7, 1))
        orders = cc.trigger_new_contributions(tm.tbs)
        self.assertEqual(len(orders), 1)

        quarter2 = orders.pop()
        self.assertEqual(quarter2.month, None)
        self.assertEqual(quarter2.quarter, 2)

    def test_periodicity_change_monthly_to_yearly(self):
        ma, tm = self._create_simple_member_for_declaration()
        self._get_member_product(tm)

        tm.declaration_periodicity = PvcTbsLicense.DECLARATION_PERIOD_MONTHLY
        self.sa_session.add(tm)
        self.sa_session.flush()

        set_date(date(2014, 12, 1))

        mcc = PvcContributionOrderController(self.sa_session, account_id=ma.id)
        cc = PvcContributionOrderController(self.sa_session)

        # make january-november declarations
        cc.add_open_contribution_order(tm, 2014, 1, None)
        cc.add_open_contribution_order(tm, 2014, 2, None)
        cc.add_open_contribution_order(tm, 2014, 3, None)
        cc.add_open_contribution_order(tm, 2014, 4, None)
        cc.add_open_contribution_order(tm, 2014, 5, None)
        cc.add_open_contribution_order(tm, 2014, 6, None)
        cc.add_open_contribution_order(tm, 2014, 7, None)
        cc.add_open_contribution_order(tm, 2014, 8, None)
        cc.add_open_contribution_order(tm, 2014, 9, None)
        cc.add_open_contribution_order(tm, 2014, 10, None)
        cc.add_open_contribution_order(tm, 2014, 11, None)

        set_date(date(2014, 12, 11))

        cc.close_contributions(tm.tbs)

        # set new periodicity
        mc = PvcMemberController(self.sa_session, account_id=self.tbs_account.id)
        mc.save_tbs_member(tm.id, {
            'status': tm.status,
            'is_full_member': tm.is_full_member,
            'is_representative': tm.is_representative,
            'declaration_periodicity': PvcTbsLicense.DECLARATION_PERIOD_YEARLY
        })

        # verify that december get triggered correctly
        set_date(date(2015, 1, 1))
        orders = cc.trigger_new_contributions(tm.tbs)
        self.assertEqual(len(orders), 1)

        december = orders.pop()
        self.assertEqual(december.month, 12)

        # submit december declaration
        mcc.submit_order(december)

        # verify that new declaration period is yearly
        self.sa_session.expire(tm)
        self.assertEqual(tm.declaration_period, PvcTbsLicense.DECLARATION_PERIOD_YEARLY)

        # now lets go to january 2016
        set_date(date(2016, 1, 1))
        orders = cc.trigger_new_contributions(tm.tbs)
        self.assertEqual(len(orders), 1)

        yearly = orders.pop()
        self.assertEqual(yearly.month, None)
        self.assertEqual(yearly.quarter, None)
        self.assertEqual(yearly.year, 2015)

    def test_periodicity_change_quarterly_to_yearly(self):
        ma, tm = self._create_simple_member_for_declaration()
        self._get_member_product(tm)

        tm.declaration_periodicity = PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY
        self.sa_session.add(tm)
        self.sa_session.flush()

        set_date(date(2014, 10, 1))

        mcc = PvcContributionOrderController(self.sa_session, account_id=ma.id)
        cc = PvcContributionOrderController(self.sa_session)

        # make q1-q3 declarations
        cc.add_open_contribution_order(tm, 2014, None, 1)
        cc.add_open_contribution_order(tm, 2014, None, 2)
        cc.add_open_contribution_order(tm, 2014, None, 3)

        set_date(date(2014, 10, 11))

        cc.close_contributions(tm.tbs)

        # set new periodicity
        mc = PvcMemberController(self.sa_session, account_id=self.tbs_account.id)
        mc.save_tbs_member(tm.id, {
            'status': tm.status,
            'is_full_member': tm.is_full_member,
            'is_representative': tm.is_representative,
            'declaration_periodicity': PvcTbsLicense.DECLARATION_PERIOD_YEARLY
        })

        # verify that q4 get triggered correctly
        set_date(date(2015, 1, 1))
        orders = cc.trigger_new_contributions(tm.tbs)
        self.assertEqual(len(orders), 1)

        quarter4 = orders.pop()
        self.assertEqual(quarter4.quarter, 4)

        # submit december declaration
        mcc.submit_order(quarter4)

        # verify that new declaration period is yearly
        self.sa_session.expire(tm)
        self.assertEqual(tm.declaration_period, PvcTbsLicense.DECLARATION_PERIOD_YEARLY)

        # now lets go to january 2016
        set_date(date(2016, 1, 1))
        orders = cc.trigger_new_contributions(tm.tbs)
        self.assertEqual(len(orders), 1)

        yearly = orders.pop()
        self.assertEqual(yearly.month, None)
        self.assertEqual(yearly.quarter, None)
        self.assertEqual(yearly.year, 2015)

    def test_periodicity_change_quarterly_to_monthly(self):
        ma, tm = self._create_simple_member_for_declaration()
        self._get_member_product(tm)

        tm.declaration_periodicity = PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY
        self.sa_session.add(tm)
        self.sa_session.flush()

        set_date(date(2014, 7, 1))

        mcc = PvcContributionOrderController(self.sa_session, account_id=ma.id)
        cc = PvcContributionOrderController(self.sa_session)

        # make q1-q3 declarations
        cc.add_open_contribution_order(tm, 2014, None, 1)
        cc.add_open_contribution_order(tm, 2014, None, 2)

        set_date(date(2014, 7, 11))

        cc.close_contributions(tm.tbs)

        set_date(date(2014, 10, 1))

        # open q3
        quarter3 = cc.add_open_contribution_order(tm, 2014, None, 3)

        # set new periodicity
        mc = PvcMemberController(self.sa_session, account_id=self.tbs_account.id)
        mc.save_tbs_member(tm.id, {
            'status': tm.status,
            'is_full_member': tm.is_full_member,
            'is_representative': tm.is_representative,
            'declaration_periodicity': PvcTbsLicense.DECLARATION_PERIOD_MONTHLY
        })

        mcc.submit_order(quarter3)
        # verify that new declaration period is monthly
        self.sa_session.expire(tm)
        self.assertEqual(tm.declaration_period, PvcTbsLicense.DECLARATION_PERIOD_MONTHLY)

        # verify that october get triggered correctly
        set_date(date(2014, 11, 1))
        orders = cc.trigger_new_contributions(tm.tbs)
        self.assertEqual(len(orders), 1)

        october = orders.pop()
        self.assertEqual(october.month, 10)
        self.assertEqual(october.quarter, None)

    def test_periodicity_change_yearly_to_monthly(self):
        ma, tm = self._create_simple_member_for_declaration()
        self._get_member_product(tm)

        tm.declaration_periodicity = PvcTbsLicense.DECLARATION_PERIOD_YEARLY
        self.sa_session.add(tm)
        self.sa_session.flush()

        set_date(date(2015, 1, 1))

        mcc = PvcContributionOrderController(self.sa_session, account_id=ma.id)
        cc = PvcContributionOrderController(self.sa_session)

        # make year declaration
        yearly = cc.add_open_contribution_order(tm, 2014, None, None)

        # set new periodicity
        mc = PvcMemberController(self.sa_session, account_id=self.tbs_account.id)
        mc.save_tbs_member(tm.id, {
            'status': tm.status,
            'is_full_member': tm.is_full_member,
            'is_representative': tm.is_representative,
            'declaration_periodicity': PvcTbsLicense.DECLARATION_PERIOD_MONTHLY
        })

        mcc.submit_order(yearly)

        # verify that new declaration period is monthly
        self.sa_session.expire(tm)
        self.assertEqual(tm.declaration_period, PvcTbsLicense.DECLARATION_PERIOD_MONTHLY)

        # verify that october get triggered correctly
        set_date(date(2015, 2, 1))
        orders = cc.trigger_new_contributions(tm.tbs)
        self.assertEqual(len(orders), 1)

        january = orders.pop()
        self.assertEqual(january.year, 2015)
        self.assertEqual(january.month, 1)
        self.assertEqual(january.quarter, None)

    def test_periodicity_change_yearly_to_quarterly(self):
        ma, tm = self._create_simple_member_for_declaration()
        self._get_member_product(tm)

        tm.declaration_periodicity = PvcTbsLicense.DECLARATION_PERIOD_YEARLY
        self.sa_session.add(tm)
        self.sa_session.flush()

        set_date(date(2015, 1, 1))

        mcc = PvcContributionOrderController(self.sa_session, account_id=ma.id)
        cc = PvcContributionOrderController(self.sa_session)

        # make year declaration
        yearly = cc.add_open_contribution_order(tm, 2014, None, None)

        # set new periodicity
        mc = PvcMemberController(self.sa_session, account_id=self.tbs_account.id)
        mc.save_tbs_member(tm.id, {
            'status': tm.status,
            'is_full_member': tm.is_full_member,
            'is_representative': tm.is_representative,
            'declaration_periodicity': PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY
        })

        mcc.submit_order(yearly)

        # verify that new declaration period is monthly
        self.sa_session.expire(tm)
        self.assertEqual(tm.declaration_period, PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY)

        # verify that q1 get triggered correctly
        set_date(date(2015, 4, 1))
        orders = cc.trigger_new_contributions(tm.tbs)
        self.assertEqual(len(orders), 1)

        quarter1 = orders.pop()
        self.assertEqual(quarter1.year, 2015)
        self.assertEqual(quarter1.month, None)
        self.assertEqual(quarter1.quarter, 1)

    def test_periodicity_change_monthly_to_quarterly_with_forecast_before_trigger(self):
        ma, tm = self._create_simple_member_for_declaration()

        tm.tbs.set_config('pvc.tbs.advance.invoicing', True)

        self._get_member_product(tm)

        tm.declaration_periodicity = PvcTbsLicense.DECLARATION_PERIOD_MONTHLY
        self.sa_session.add(tm)
        self.sa_session.flush()

        set_date(date(2014, 3, 1))

        mcc = PvcContributionOrderController(self.sa_session, account_id=ma.id)
        cc = PvcContributionOrderController(self.sa_session)

        # make january and february declarations
        cc.add_open_contribution_order(tm, 2014, 1, None)
        cc.add_open_contribution_order(tm, 2014, 2, None)

        set_date(date(2014, 3, 11))

        cc.close_contributions(tm.tbs)

        # set new periodicity
        mc = PvcMemberController(self.sa_session, account_id=self.tbs_account.id)
        mc.save_tbs_member(tm.id, {
            'status': tm.status,
            'is_full_member': tm.is_full_member,
            'is_representative': tm.is_representative,
            'declaration_periodicity': PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY
        })

        # verify that march get triggered correctly
        set_date(date(2014, 4, 1))
        orders = cc.trigger_new_contributions(tm.tbs)
        self.assertEqual(len(orders), 1)

        march = orders.pop()
        self.assertEqual(march.month, 3)

        # submit march declaration
        march = mcc.resubmit_contribution(march, {'line_0_quantity': 100, 'line_0_weight': Decimal('2334.23')})
        mcc.submit_order(march)

        # verify that new declaration period is quarterly
        self.sa_session.expire(tm)
        self.assertEqual(tm.declaration_period, PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY)

        # now lets go to july
        set_date(date(2014, 7, 1))
        orders = cc.trigger_new_contributions(tm.tbs)
        self.assertEqual(len(orders), 1)

        quarter2 = orders.pop()
        self.assertTrue(quarter2.is_forecast)
        self.assertEqual(quarter2.month, None)
        self.assertEqual(quarter2.quarter, 2)
        line = quarter2.lines[0]
        self.assertEqual(line.quantity, 300)
        self.assertEqual(line.weight, Decimal('7002.69'))

    def test_periodicity_change_monthly_to_quarterly_with_forecast_after_trigger(self):
        ma, tm = self._create_simple_member_for_declaration()

        tm.tbs.set_config('pvc.tbs.advance.invoicing', True)

        self._get_member_product(tm)

        tm.declaration_periodicity = PvcTbsLicense.DECLARATION_PERIOD_MONTHLY
        self.sa_session.add(tm)
        self.sa_session.flush()

        set_date(date(2014, 3, 1))

        mcc = PvcContributionOrderController(self.sa_session, account_id=ma.id)
        cc = PvcContributionOrderController(self.sa_session)

        # make january and february declarations
        cc.add_open_contribution_order(tm, 2014, 1, None)
        cc.add_open_contribution_order(tm, 2014, 2, None)

        set_date(date(2014, 3, 11))

        cc.close_contributions(tm.tbs)

        # verify that march get triggered correctly
        set_date(date(2014, 4, 1))
        orders = cc.trigger_new_contributions(tm.tbs)
        self.assertEqual(len(orders), 1)

        march = orders.pop()
        self.assertEqual(march.month, 3)

        # submit march declaration
        march = mcc.resubmit_contribution(march, {'line_0_quantity': 100, 'line_0_weight': Decimal('2334.23')})
        mcc.submit_order(march)

        # set new periodicity
        mc = PvcMemberController(self.sa_session, account_id=self.tbs_account.id)
        mc.save_tbs_member(tm.id, {
            'status': tm.status,
            'is_full_member': tm.is_full_member,
            'is_representative': tm.is_representative,
            'declaration_periodicity': PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY
        })

        # now lets go to july
        set_date(date(2014, 5, 1))
        orders = cc.trigger_new_contributions(tm.tbs)
        self.assertEqual(len(orders), 1)

        quarter2 = orders.pop()
        self.assertTrue(quarter2.is_forecast)
        self.assertEqual(quarter2.month, 4)
        self.assertEqual(quarter2.quarter, None)

        # verify that declaration period is still monthly
        self.sa_session.expire(tm)
        self.assertEqual(tm.declaration_period, PvcTbsLicense.DECLARATION_PERIOD_MONTHLY)

        # skip to july
        set_date(date(2014, 7, 1))

        cc.add_open_contribution_order(tm, 2014, 5, None)
        cc.add_open_contribution_order(tm, 2014, 6, None)

        set_date(date(2014, 7, 11))
        cc.close_contributions(tm.tbs)

        # verify that new declaration period is quarterly
        self.sa_session.expire(tm)
        self.assertEqual(tm.declaration_period, PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY)

        quarter3 = cc._is_undeclared_period(tm, 2014, 3, None, tm, True)
        self.assertTrue(quarter3.is_forecast)
        self.assertEqual(quarter3.month, None)
        self.assertEqual(quarter3.quarter, 3)

    def test_periodicity_change_monthly_to_yearly_with_forecast_before_trigger(self):
        ma, tm = self._create_simple_member_for_declaration()

        tm.tbs.set_config('pvc.tbs.advance.invoicing', True)

        self._get_member_product(tm)

        tm.declaration_periodicity = PvcTbsLicense.DECLARATION_PERIOD_MONTHLY
        self.sa_session.add(tm)
        self.sa_session.flush()

        set_date(date(2014, 12, 1))

        mcc = PvcContributionOrderController(self.sa_session, account_id=ma.id)
        cc = PvcContributionOrderController(self.sa_session)

        # make january - november
        cc.add_open_contribution_order(tm, 2014, 1, None)
        cc.add_open_contribution_order(tm, 2014, 2, None)
        cc.add_open_contribution_order(tm, 2014, 3, None)
        cc.add_open_contribution_order(tm, 2014, 4, None)
        cc.add_open_contribution_order(tm, 2014, 5, None)
        cc.add_open_contribution_order(tm, 2014, 6, None)
        cc.add_open_contribution_order(tm, 2014, 7, None)
        cc.add_open_contribution_order(tm, 2014, 8, None)
        cc.add_open_contribution_order(tm, 2014, 9, None)
        cc.add_open_contribution_order(tm, 2014, 10, None)
        cc.add_open_contribution_order(tm, 2014, 11, None)

        set_date(date(2014, 12, 11))

        cc.close_contributions(tm.tbs)

        # set new periodicity
        mc = PvcMemberController(self.sa_session, account_id=self.tbs_account.id)
        mc.save_tbs_member(tm.id, {
            'status': tm.status,
            'is_full_member': tm.is_full_member,
            'is_representative': tm.is_representative,
            'declaration_periodicity': PvcTbsLicense.DECLARATION_PERIOD_YEARLY
        })

        # verify that december get triggered correctly
        set_date(date(2015, 1, 1))
        orders = cc.trigger_new_contributions(tm.tbs)
        self.assertEqual(len(orders), 1)

        december = orders.pop()
        self.assertEqual(december.month, 12)

        # submit march declaration
        december = mcc.resubmit_contribution(december, {'line_0_quantity': 100, 'line_0_weight': Decimal('2334.23')})
        mcc.submit_order(december)

        # verify that new declaration period is yearly
        self.sa_session.expire(tm)
        self.assertEqual(tm.declaration_period, PvcTbsLicense.DECLARATION_PERIOD_YEARLY)

        # now lets go to next january
        set_date(date(2016, 1, 1))
        orders = cc.trigger_new_contributions(tm.tbs)
        self.assertEqual(len(orders), 1)

        year = orders.pop()
        self.assertTrue(year.is_forecast)
        self.assertEqual(year.month, None)
        self.assertEqual(year.quarter, None)
        self.assertEqual(year.year, 2015)
        line = year.lines[0]
        self.assertEqual(line.quantity, 1200)
        self.assertEqual(line.weight, Decimal('28010.76'))

    def test_periodicity_change_monthly_to_yearly_with_forecast_after_trigger(self):
        ma, tm = self._create_simple_member_for_declaration()

        tm.tbs.set_config('pvc.tbs.advance.invoicing', True)

        self._get_member_product(tm)

        tm.declaration_periodicity = PvcTbsLicense.DECLARATION_PERIOD_MONTHLY
        self.sa_session.add(tm)
        self.sa_session.flush()

        set_date(date(2014, 12, 1))

        mcc = PvcContributionOrderController(self.sa_session, account_id=ma.id)
        cc = PvcContributionOrderController(self.sa_session)

        # make january - november
        cc.add_open_contribution_order(tm, 2014, 1, None)
        cc.add_open_contribution_order(tm, 2014, 2, None)
        cc.add_open_contribution_order(tm, 2014, 3, None)
        cc.add_open_contribution_order(tm, 2014, 4, None)
        cc.add_open_contribution_order(tm, 2014, 5, None)
        cc.add_open_contribution_order(tm, 2014, 6, None)
        cc.add_open_contribution_order(tm, 2014, 7, None)
        cc.add_open_contribution_order(tm, 2014, 8, None)
        cc.add_open_contribution_order(tm, 2014, 9, None)
        cc.add_open_contribution_order(tm, 2014, 10, None)
        cc.add_open_contribution_order(tm, 2014, 11, None)

        set_date(date(2014, 12, 11))

        cc.close_contributions(tm.tbs)

        # verify that december get triggered correctly
        set_date(date(2015, 1, 1))
        orders = cc.trigger_new_contributions(tm.tbs)
        self.assertEqual(len(orders), 1)

        december = orders.pop()
        self.assertEqual(december.month, 12)

        # submit december declaration
        december = mcc.resubmit_contribution(december, {'line_0_quantity': 100, 'line_0_weight': Decimal('2334.23')})
        mcc.submit_order(december)

        # set new periodicity
        mc = PvcMemberController(self.sa_session, account_id=self.tbs_account.id)
        mc.save_tbs_member(tm.id, {
            'status': tm.status,
            'is_full_member': tm.is_full_member,
            'is_representative': tm.is_representative,
            'declaration_periodicity': PvcTbsLicense.DECLARATION_PERIOD_YEARLY
        })

        # now lets go to february
        set_date(date(2015, 2, 1))
        orders = cc.trigger_new_contributions(tm.tbs)
        self.assertEqual(len(orders), 1)

        january = orders.pop()
        self.assertTrue(january.is_forecast)
        self.assertEqual(january.month, 1)
        self.assertEqual(january.quarter, None)

        # verify that declaration period is still monthly
        self.sa_session.expire(tm)
        self.assertEqual(tm.declaration_period, PvcTbsLicense.DECLARATION_PERIOD_MONTHLY)

        # skip to next year
        set_date(date(2016, 1, 1))

        cc.add_open_contribution_order(tm, 2015, 3, None)
        cc.add_open_contribution_order(tm, 2015, 4, None)
        cc.add_open_contribution_order(tm, 2015, 5, None)
        cc.add_open_contribution_order(tm, 2015, 6, None)
        cc.add_open_contribution_order(tm, 2015, 7, None)
        cc.add_open_contribution_order(tm, 2015, 8, None)
        cc.add_open_contribution_order(tm, 2015, 9, None)
        cc.add_open_contribution_order(tm, 2015, 10, None)
        cc.add_open_contribution_order(tm, 2015, 11, None)
        cc.add_open_contribution_order(tm, 2015, 12, None)

        set_date(date(2016, 1, 11))
        cc.close_contributions(tm.tbs)

        # verify that new declaration period is yearly
        self.sa_session.expire(tm)
        self.assertEqual(tm.declaration_period, PvcTbsLicense.DECLARATION_PERIOD_YEARLY)

        year3 = cc._is_undeclared_period(tm, 2016, None, None, tm, True)
        self.assertTrue(year3.is_forecast)
        self.assertEqual(year3.month, None)
        self.assertEqual(year3.quarter, None)
        self.assertEqual(year3.year, 2016)

    def test_periodicity_change_quarterly_to_yearly_with_forecast_before_trigger(self):
        ma, tm = self._create_simple_member_for_declaration()

        tm.tbs.set_config('pvc.tbs.advance.invoicing', True)

        self._get_member_product(tm)

        tm.declaration_periodicity = PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY
        self.sa_session.add(tm)
        self.sa_session.flush()

        set_date(date(2014, 10, 1))

        mcc = PvcContributionOrderController(self.sa_session, account_id=ma.id)
        cc = PvcContributionOrderController(self.sa_session)

        # make q1 - q3
        cc.add_open_contribution_order(tm, 2014, None, 1)
        cc.add_open_contribution_order(tm, 2014, None, 2)
        cc.add_open_contribution_order(tm, 2014, None, 3)

        set_date(date(2014, 10, 11))

        cc.close_contributions(tm.tbs)

        # set new periodicity
        mc = PvcMemberController(self.sa_session, account_id=self.tbs_account.id)
        mc.save_tbs_member(tm.id, {
            'status': tm.status,
            'is_full_member': tm.is_full_member,
            'is_representative': tm.is_representative,
            'declaration_periodicity': PvcTbsLicense.DECLARATION_PERIOD_YEARLY
        })

        # verify that quarter 4 get triggered correctly
        set_date(date(2015, 1, 1))
        orders = cc.trigger_new_contributions(tm.tbs)
        self.assertEqual(len(orders), 1)

        quarter4 = orders.pop()
        self.assertEqual(quarter4.month, None)
        self.assertEqual(quarter4.quarter, 4)

        # submit march declaration
        quarter4 = mcc.resubmit_contribution(quarter4, {'line_0_quantity': 100, 'line_0_weight': Decimal('2334.23')})
        mcc.submit_order(quarter4)

        # verify that new declaration period is yearly
        self.sa_session.expire(tm)
        self.assertEqual(tm.declaration_period, PvcTbsLicense.DECLARATION_PERIOD_YEARLY)

        # now lets go to next january
        set_date(date(2016, 1, 1))
        orders = cc.trigger_new_contributions(tm.tbs)
        self.assertEqual(len(orders), 1)

        year = orders.pop()
        self.assertTrue(year.is_forecast)
        self.assertEqual(year.month, None)
        self.assertEqual(year.quarter, None)
        self.assertEqual(year.year, 2015)
        line = year.lines[0]
        self.assertEqual(line.quantity, 400)
        self.assertEqual(line.weight, Decimal('9336.92'))

    def test_periodicity_change_quarterly_to_yearly_with_forecast_after_trigger(self):
        ma, tm = self._create_simple_member_for_declaration()

        tm.tbs.set_config('pvc.tbs.advance.invoicing', True)

        self._get_member_product(tm)

        tm.declaration_periodicity = PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY
        self.sa_session.add(tm)
        self.sa_session.flush()

        set_date(date(2014, 10, 1))

        mcc = PvcContributionOrderController(self.sa_session, account_id=ma.id)
        cc = PvcContributionOrderController(self.sa_session)

        # make q1 - q3
        cc.add_open_contribution_order(tm, 2014, None, 1)
        cc.add_open_contribution_order(tm, 2014, None, 2)
        cc.add_open_contribution_order(tm, 2014, None, 3)

        set_date(date(2014, 10, 11))

        cc.close_contributions(tm.tbs)

        # verify that quarter 4 get triggered correctly
        set_date(date(2015, 1, 1))
        orders = cc.trigger_new_contributions(tm.tbs)
        self.assertEqual(len(orders), 1)

        quarter4 = orders.pop()
        self.assertEqual(quarter4.month, None)
        self.assertEqual(quarter4.month, None)

        # submit december declaration
        quarter4 = mcc.resubmit_contribution(quarter4, {'line_0_quantity': 100, 'line_0_weight': Decimal('2334.23')})
        mcc.submit_order(quarter4)

        # set new periodicity
        mc = PvcMemberController(self.sa_session, account_id=self.tbs_account.id)
        mc.save_tbs_member(tm.id, {
            'status': tm.status,
            'is_full_member': tm.is_full_member,
            'is_representative': tm.is_representative,
            'declaration_periodicity': PvcTbsLicense.DECLARATION_PERIOD_YEARLY
        })

        # now lets go to april
        set_date(date(2015, 4, 1))
        orders = cc.trigger_new_contributions(tm.tbs)
        self.assertEqual(len(orders), 1)

        quarter1 = orders.pop()
        self.assertTrue(quarter1.is_forecast)
        self.assertEqual(quarter1.month, None)
        self.assertEqual(quarter1.quarter, 1)

        # verify that declaration period is still quarterly
        self.sa_session.expire(tm)
        self.assertEqual(tm.declaration_period, PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY)

        # skip to next year
        set_date(date(2016, 1, 1))

        cc.add_open_contribution_order(tm, 2015, None, 2)
        cc.add_open_contribution_order(tm, 2015, None, 3)
        cc.add_open_contribution_order(tm, 2015, None, 4)

        set_date(date(2016, 1, 11))
        cc.close_contributions(tm.tbs)

        # verify that new declaration period is yearly
        self.sa_session.expire(tm)
        self.assertEqual(tm.declaration_period, PvcTbsLicense.DECLARATION_PERIOD_YEARLY)

        year3 = cc._is_undeclared_period(tm, 2016, None, None, tm, True)
        self.assertTrue(year3.is_forecast)
        self.assertEqual(year3.month, None)
        self.assertEqual(year3.quarter, None)
        self.assertEqual(year3.year, 2016)

    def test_periodicity_change_quarterly_to_monthly_with_forecast_before_trigger(self):
        ma, tm = self._create_simple_member_for_declaration()

        tm.tbs.set_config('pvc.tbs.advance.invoicing', True)

        self._get_member_product(tm)

        tm.declaration_periodicity = PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY
        self.sa_session.add(tm)
        self.sa_session.flush()

        set_date(date(2014, 10, 1))

        mcc = PvcContributionOrderController(self.sa_session, account_id=ma.id)
        cc = PvcContributionOrderController(self.sa_session)

        # make q1 - q3
        cc.add_open_contribution_order(tm, 2014, None, 1)
        cc.add_open_contribution_order(tm, 2014, None, 2)
        cc.add_open_contribution_order(tm, 2014, None, 3)

        set_date(date(2014, 10, 11))

        cc.close_contributions(tm.tbs)

        # set new periodicity
        mc = PvcMemberController(self.sa_session, account_id=self.tbs_account.id)
        mc.save_tbs_member(tm.id, {
            'status': tm.status,
            'is_full_member': tm.is_full_member,
            'is_representative': tm.is_representative,
            'declaration_periodicity': PvcTbsLicense.DECLARATION_PERIOD_MONTHLY
        })

        # verify that quarter 4 get triggered correctly
        set_date(date(2015, 1, 1))
        orders = cc.trigger_new_contributions(tm.tbs)
        self.assertEqual(len(orders), 1)

        quarter4 = orders.pop()
        self.assertEqual(quarter4.month, None)
        self.assertEqual(quarter4.quarter, 4)

        # submit march declaration
        quarter4 = mcc.resubmit_contribution(quarter4, {'line_0_quantity': 300, 'line_0_weight': Decimal('333.33')})
        mcc.submit_order(quarter4)

        # verify that new declaration period is monthly
        self.sa_session.expire(tm)
        self.assertEqual(tm.declaration_period, PvcTbsLicense.DECLARATION_PERIOD_MONTHLY)

        # now lets go to february
        set_date(date(2015, 2, 1))
        orders = cc.trigger_new_contributions(tm.tbs)
        self.assertEqual(len(orders), 1)

        year = orders.pop()
        self.assertTrue(year.is_forecast)
        self.assertEqual(year.month, 1)
        self.assertEqual(year.quarter, None)
        self.assertEqual(year.year, 2015)
        line = year.lines[0]
        self.assertEqual(line.quantity, 100)
        self.assertEqual(line.weight, Decimal('111.11'))

    def test_periodicity_change_quarterly_to_monthly_with_forecast_after_trigger(self):
        ma, tm = self._create_simple_member_for_declaration()

        tm.tbs.set_config('pvc.tbs.advance.invoicing', True)

        self._get_member_product(tm)

        tm.declaration_periodicity = PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY
        self.sa_session.add(tm)
        self.sa_session.flush()

        set_date(date(2014, 10, 1))

        mcc = PvcContributionOrderController(self.sa_session, account_id=ma.id)
        cc = PvcContributionOrderController(self.sa_session)

        # make q1 - q3
        cc.add_open_contribution_order(tm, 2014, None, 1)
        cc.add_open_contribution_order(tm, 2014, None, 2)
        cc.add_open_contribution_order(tm, 2014, None, 3)

        set_date(date(2014, 10, 11))

        cc.close_contributions(tm.tbs)

        # verify that quarter 4 get triggered correctly
        set_date(date(2015, 1, 1))
        orders = cc.trigger_new_contributions(tm.tbs)
        self.assertEqual(len(orders), 1)

        quarter4 = orders.pop()
        self.assertEqual(quarter4.month, None)
        self.assertEqual(quarter4.month, None)

        # submit december declaration
        quarter4 = mcc.resubmit_contribution(quarter4, {'line_0_quantity': 100, 'line_0_weight': Decimal('2334.23')})
        mcc.submit_order(quarter4)

        # set new periodicity
        mc = PvcMemberController(self.sa_session, account_id=self.tbs_account.id)
        mc.save_tbs_member(tm.id, {
            'status': tm.status,
            'is_full_member': tm.is_full_member,
            'is_representative': tm.is_representative,
            'declaration_periodicity': PvcTbsLicense.DECLARATION_PERIOD_MONTHLY
        })

        # now lets go to april
        set_date(date(2015, 4, 1))
        orders = cc.trigger_new_contributions(tm.tbs)
        self.assertEqual(len(orders), 1)

        quarter1 = orders.pop()
        self.assertTrue(quarter1.is_forecast)
        self.assertEqual(quarter1.month, None)
        self.assertEqual(quarter1.quarter, 1)

        # verify that declaration period is still quarterly
        self.sa_session.expire(tm)
        self.assertEqual(tm.declaration_period, PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY)

        quarter1 = mcc.resubmit_contribution(quarter1, {'line_0_quantity': 100, 'line_0_weight': Decimal('2334.23')})
        mcc.submit_order(quarter1)

        # verify that new declaration period is quarterly
        self.sa_session.expire(tm)
        self.assertEqual(tm.declaration_period, PvcTbsLicense.DECLARATION_PERIOD_MONTHLY)

        april = cc._is_undeclared_period(tm, 2015, None, 4, tm, True)
        self.assertTrue(april.is_forecast)
        self.assertEqual(april.month, 4)
        self.assertEqual(april.quarter, None)
        self.assertEqual(april.year, 2015)

    def test_periodicity_change_yearly_to_monthly_with_forecast_before_trigger(self):
        ma, tm = self._create_simple_member_for_declaration()

        tm.tbs.set_config('pvc.tbs.advance.invoicing', True)

        self._get_member_product(tm)

        tm.declaration_periodicity = PvcTbsLicense.DECLARATION_PERIOD_YEARLY
        self.sa_session.add(tm)
        self.sa_session.flush()

        set_date(date(2015, 1, 1))

        mcc = PvcContributionOrderController(self.sa_session, account_id=ma.id)
        cc = PvcContributionOrderController(self.sa_session)

        # make year 1 declaration
        cc.add_open_contribution_order(tm, 2014, None, None)

        set_date(date(2015, 1, 11))

        cc.close_contributions(tm.tbs)

        # set new periodicity
        mc = PvcMemberController(self.sa_session, account_id=self.tbs_account.id)
        mc.save_tbs_member(tm.id, {
            'status': tm.status,
            'is_full_member': tm.is_full_member,
            'is_representative': tm.is_representative,
            'declaration_periodicity': PvcTbsLicense.DECLARATION_PERIOD_MONTHLY
        })

        # verify that next year get triggered correctly
        set_date(date(2016, 1, 1))
        orders = cc.trigger_new_contributions(tm.tbs)
        self.assertEqual(len(orders), 1)

        year2 = orders.pop()
        self.assertEqual(year2.month, None)
        self.assertEqual(year2.quarter, None)
        self.assertEqual(year2.year, 2015)

        # submit march declaration
        year2 = mcc.resubmit_contribution(year2, {'line_0_quantity': 1200, 'line_0_weight': Decimal('2400.00')})
        mcc.submit_order(year2)

        # verify that new declaration period is monthly
        self.sa_session.expire(tm)
        self.assertEqual(tm.declaration_period, PvcTbsLicense.DECLARATION_PERIOD_MONTHLY)

        # now lets go to february
        set_date(date(2016, 2, 1))
        orders = cc.trigger_new_contributions(tm.tbs)
        self.assertEqual(len(orders), 1)

        year = orders.pop()
        self.assertTrue(year.is_forecast)
        self.assertEqual(year.month, 1)
        self.assertEqual(year.quarter, None)
        self.assertEqual(year.year, 2016)
        line = year.lines[0]
        self.assertEqual(line.quantity, 100)
        self.assertEqual(line.weight, Decimal('200.00'))

    def test_periodicity_change_yearly_to_monthly_with_forecast_after_trigger(self):
        ma, tm = self._create_simple_member_for_declaration()

        tm.tbs.set_config('pvc.tbs.advance.invoicing', True)

        self._get_member_product(tm)

        tm.declaration_periodicity = PvcTbsLicense.DECLARATION_PERIOD_YEARLY
        self.sa_session.add(tm)
        self.sa_session.flush()

        set_date(date(2015, 1, 1))

        mcc = PvcContributionOrderController(self.sa_session, account_id=ma.id)
        cc = PvcContributionOrderController(self.sa_session)

        # make year one
        cc.add_open_contribution_order(tm, 2014, None, None)

        set_date(date(2015, 1, 11))

        cc.close_contributions(tm.tbs)

        # verify that next year get triggered correctly
        set_date(date(2016, 1, 1))
        orders = cc.trigger_new_contributions(tm.tbs)
        self.assertEqual(len(orders), 1)

        year2 = orders.pop()
        self.assertEqual(year2.month, None)
        self.assertEqual(year2.month, None)

        # submit year2 declaration
        year2 = mcc.resubmit_contribution(year2, {'line_0_quantity': 100, 'line_0_weight': Decimal('2334.23')})
        mcc.submit_order(year2)

        # set new periodicity
        mc = PvcMemberController(self.sa_session, account_id=self.tbs_account.id)
        mc.save_tbs_member(tm.id, {
            'status': tm.status,
            'is_full_member': tm.is_full_member,
            'is_representative': tm.is_representative,
            'declaration_periodicity': PvcTbsLicense.DECLARATION_PERIOD_MONTHLY
        })
        self.sa_session.expire(tm)

        # now lets go to 2017
        set_date(date(2017, 1, 1))
        orders = cc.trigger_new_contributions(tm.tbs)
        self.assertEqual(len(orders), 1)

        year3 = orders.pop()
        self.assertTrue(year3.is_forecast)
        self.assertEqual(year3.month, None)
        self.assertEqual(year3.quarter, None)
        self.assertEqual(year3.year, 2016)

        # verify that declaration period is still yearly
        self.sa_session.expire(tm)
        self.assertEqual(tm.declaration_period, PvcTbsLicense.DECLARATION_PERIOD_YEARLY)

        year3 = mcc.resubmit_contribution(year3, {'line_0_quantity': 100, 'line_0_weight': Decimal('2334.23')})
        mcc.submit_order(year3)

        # verify that new declaration period is quarterly
        self.sa_session.expire(tm)
        self.assertEqual(tm.declaration_period, PvcTbsLicense.DECLARATION_PERIOD_MONTHLY)

        january = cc._is_undeclared_period(tm, 2017, None, 1, tm, True)
        self.assertTrue(january.is_forecast)
        self.assertEqual(january.month, 1)
        self.assertEqual(january.quarter, None)
        self.assertEqual(january.year, 2017)

    def test_periodicity_change_yearly_to_quarterly_with_forecast_before_trigger(self):
        ma, tm = self._create_simple_member_for_declaration()

        tm.tbs.set_config('pvc.tbs.advance.invoicing', True)

        self._get_member_product(tm)

        tm.declaration_periodicity = PvcTbsLicense.DECLARATION_PERIOD_YEARLY
        self.sa_session.add(tm)
        self.sa_session.flush()

        set_date(date(2015, 1, 1))

        mcc = PvcContributionOrderController(self.sa_session, account_id=ma.id)
        cc = PvcContributionOrderController(self.sa_session)

        # make year 1 declaration
        cc.add_open_contribution_order(tm, 2014, None, None)

        set_date(date(2015, 1, 11))

        cc.close_contributions(tm.tbs)

        # set new periodicity
        mc = PvcMemberController(self.sa_session, account_id=self.tbs_account.id)
        mc.save_tbs_member(tm.id, {
            'status': tm.status,
            'is_full_member': tm.is_full_member,
            'is_representative': tm.is_representative,
            'declaration_periodicity': PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY
        })

        # verify that next year get triggered correctly
        set_date(date(2016, 1, 1))
        orders = cc.trigger_new_contributions(tm.tbs)
        self.assertEqual(len(orders), 1)

        year2 = orders.pop()
        self.assertEqual(year2.month, None)
        self.assertEqual(year2.quarter, None)
        self.assertEqual(year2.year, 2015)

        # submit march declaration
        year2 = mcc.resubmit_contribution(year2, {'line_0_quantity': 400, 'line_0_weight': Decimal('2400.00')})
        mcc.submit_order(year2)

        # verify that new declaration period is quarterly
        self.sa_session.expire(tm)
        self.assertEqual(tm.declaration_period, PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY)

        # now lets go to april
        set_date(date(2016, 4, 1))
        orders = cc.trigger_new_contributions(tm.tbs)
        self.assertEqual(len(orders), 1)

        year = orders.pop()
        self.assertTrue(year.is_forecast)
        self.assertEqual(year.month, None)
        self.assertEqual(year.quarter, 1)
        self.assertEqual(year.year, 2016)
        line = year.lines[0]
        self.assertEqual(line.quantity, 100)
        self.assertEqual(line.weight, Decimal('600.00'))

    def test_periodicity_change_yearly_to_quarterly_with_forecast_after_trigger(self):
        ma, tm = self._create_simple_member_for_declaration()

        tm.tbs.set_config('pvc.tbs.advance.invoicing', True)

        self._get_member_product(tm)

        tm.declaration_periodicity = PvcTbsLicense.DECLARATION_PERIOD_YEARLY
        self.sa_session.add(tm)
        self.sa_session.flush()

        set_date(date(2015, 1, 1))

        mcc = PvcContributionOrderController(self.sa_session, account_id=ma.id)
        cc = PvcContributionOrderController(self.sa_session)

        # make year one
        cc.add_open_contribution_order(tm, 2014, None, None)

        set_date(date(2015, 1, 11))

        cc.close_contributions(tm.tbs)

        # verify that next year get triggered correctly
        set_date(date(2016, 1, 1))
        orders = cc.trigger_new_contributions(tm.tbs)
        self.assertEqual(len(orders), 1)

        year2 = orders.pop()
        self.assertEqual(year2.month, None)
        self.assertEqual(year2.month, None)

        # submit year2 declaration
        year2 = mcc.resubmit_contribution(year2, {'line_0_quantity': 100, 'line_0_weight': Decimal('2334.23')})
        mcc.submit_order(year2)

        # set new periodicity
        # set new periodicity
        mc = PvcMemberController(self.sa_session, account_id=self.tbs_account.id)
        mc.save_tbs_member(tm.id, {
            'status': tm.status,
            'is_full_member': tm.is_full_member,
            'is_representative': tm.is_representative,
            'declaration_periodicity': PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY
        })
        self.sa_session.expire(tm)

        # now lets go to 2017
        set_date(date(2017, 1, 1))
        orders = cc.trigger_new_contributions(tm.tbs)
        self.assertEqual(len(orders), 1)

        year3 = orders.pop()
        self.assertTrue(year3.is_forecast)
        self.assertEqual(year3.month, None)
        self.assertEqual(year3.quarter, None)
        self.assertEqual(year3.year, 2016)

        # verify that declaration period is still yearly
        self.sa_session.expire(tm)
        self.assertEqual(tm.declaration_period, PvcTbsLicense.DECLARATION_PERIOD_YEARLY)

        year3 = mcc.resubmit_contribution(year3, {'line_0_quantity': 100, 'line_0_weight': Decimal('2334.23')})
        mcc.submit_order(year3)

        # verify that new declaration period is quarterly
        self.sa_session.expire(tm)
        self.assertEqual(tm.declaration_period, PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY)

        quarter1 = cc._is_undeclared_period(tm, 2017, 1, None, tm, True)
        self.assertTrue(quarter1.is_forecast)
        self.assertEqual(quarter1.month, None)
        self.assertEqual(quarter1.quarter, 1)
        self.assertEqual(quarter1.year, 2017)

    def test_periodicity_change_after_closing_day(self):
        ma, tm = self._create_simple_member_for_declaration()
        self._get_member_product(tm)

        tm.declaration_periodicity = PvcTbsLicense.DECLARATION_PERIOD_MONTHLY
        self.sa_session.add(tm)
        self.sa_session.flush()

        set_date(date(2014, 4, 1))

        mcc = PvcContributionOrderController(self.sa_session, account_id=ma.id)
        cc = PvcContributionOrderController(self.sa_session)

        # make january and february declarations
        cc.add_open_contribution_order(tm, 2014, 1, None)
        cc.add_open_contribution_order(tm, 2014, 2, None)
        cc.add_open_contribution_order(tm, 2014, 3, None)

        set_date(date(2014, 4, 11))

        cc.close_contributions(tm.tbs)

        # set new periodicity
        mc = PvcMemberController(self.sa_session, account_id=self.tbs_account.id)
        mc.save_tbs_member(tm.id, {
            'status': tm.status,
            'is_full_member': tm.is_full_member,
            'is_representative': tm.is_representative,
            'declaration_periodicity': PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY
        })

        # verify that april does not get triggered for this TM
        set_date(date(2014, 5, 1))
        orders = cc.trigger_new_contributions(tm.tbs)
        self.assertEqual(len(orders), 0)

        # verify that quarter 2 does get triggered
        set_date(date(2014, 7, 1))
        orders = cc.trigger_new_contributions(tm.tbs)
        self.assertEqual(len(orders), 1)

    def test_periodicity_change_after_closing_day_with_forecast(self):
        ma, tm = self._create_simple_member_for_declaration()
        self._get_member_product(tm)

        tm.tbs.set_config('pvc.tbs.advance.invoicing', True)

        tm.declaration_periodicity = PvcTbsLicense.DECLARATION_PERIOD_MONTHLY
        self.sa_session.add(tm)
        self.sa_session.flush()

        set_date(date(2014, 4, 1))

        mcc = PvcContributionOrderController(self.sa_session, account_id=ma.id)
        cc = PvcContributionOrderController(self.sa_session)

        # make january and february declarations
        cc.add_open_contribution_order(tm, 2014, 1, None)
        cc.add_open_contribution_order(tm, 2014, 2, None)
        cc.add_open_contribution_order(tm, 2014, 3, None)

        set_date(date(2014, 4, 11))

        cc.close_contributions(tm.tbs)

        # set new periodicity
        mc = PvcMemberController(self.sa_session, account_id=self.tbs_account.id)
        mc.save_tbs_member(tm.id, {
            'status': tm.status,
            'is_full_member': tm.is_full_member,
            'is_representative': tm.is_representative,
            'declaration_periodicity': PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY
        })

        # verify that april does get triggered for this TM, as there should be a forecast order
        set_date(date(2014, 5, 1))
        orders = cc.trigger_new_contributions(tm.tbs)
        self.assertEqual(len(orders), 1)

    def test_periodicity_change_on_tbs_level(self):
        # set TBS periodicity to monthly
        self.tbs.set_config('pvc.tbs.declaration.period', PvcTbsLicense.DECLARATION_PERIOD_MONTHLY)
        # create a member, set its own periodicity
        ma1, tm1 = self._create_simple_member_for_declaration()
        tm1.declaration_periodicity = PvcTbsLicense.DECLARATION_PERIOD_YEARLY
        self.sa_session.add(tm1)
        self.sa_session.flush()

        # create a second member, no own periodicity
        ma2, tm2 = self._create_simple_member_for_declaration()

        tc = PvcTBSController(self.sa_session, account_id=self.operator_account.id)
        # this normally happens through the save_callback on the config change.
        self.tbs.set_config('pvc.tbs.declaration.period', PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY)
        self.tbs.set_config('pvc.tbs.previous.declaration.period', PvcTbsLicense.DECLARATION_PERIOD_MONTHLY)
        tc.update_periodicity_change(self.tbs.id, PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY, PvcTbsLicense.DECLARATION_PERIOD_MONTHLY)

        self.sa_session.expire(tm1)
        self.assertEqual(tm1.declaration_period, PvcTbsLicense.DECLARATION_PERIOD_YEARLY)
        self.assertEqual(tm1.upcoming_periodicity_change, None)

        self.sa_session.expire(tm2)

        self.assertEqual(tm2.declaration_period, PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY)
        self.assertEqual(tm2.upcoming_periodicity_change, None)
        self.assertEqual(tm2.declaration_periodicity, PvcTbsLicense.DECLARATION_PERIOD_DEFAULT)

    def test_periodicity_change_on_tbs_level_after_trigger(self):
        # set TBS periodicity to monthly
        self.tbs.set_config('pvc.tbs.declaration.period', PvcTbsLicense.DECLARATION_PERIOD_MONTHLY)
        # create a member, set its own periodicity
        ma1, tm1 = self._create_simple_member_for_declaration()
        tm1.declaration_periodicity = PvcTbsLicense.DECLARATION_PERIOD_YEARLY
        self.sa_session.add(tm1)
        self.sa_session.flush()
        self._get_member_product(tm1)

        # create a second member, no own periodicity
        ma2, tm2 = self._create_simple_member_for_declaration()
        self._get_member_product(tm2)

        # first of february should trigger declarations for tm2
        set_date(date(2013, 7, 1))

        cc = PvcContributionOrderController(self.sa_session)
        orders = cc.trigger_new_contributions(self.tbs)
        self.assertEqual(len(orders), 1)

        tc = PvcTBSController(self.sa_session, account_id=self.operator_account.id)
        # this normally happens through the save_callback on the config change.
        self.tbs.set_config('pvc.tbs.declaration.period', PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY)
        self.tbs.set_config('pvc.tbs.previous.declaration.period', PvcTbsLicense.DECLARATION_PERIOD_MONTHLY)
        tc.update_periodicity_change(self.tbs.id, PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY, PvcTbsLicense.DECLARATION_PERIOD_MONTHLY)

        self.sa_session.expire(tm1)
        self.assertEqual(tm1.declaration_period, PvcTbsLicense.DECLARATION_PERIOD_YEARLY)
        self.assertEqual(tm1.upcoming_periodicity_change, None)

        self.sa_session.expire(tm2)

        self.assertEqual(tm2.declaration_period, PvcTbsLicense.DECLARATION_PERIOD_MONTHLY)
        self.assertEqual(tm2.upcoming_periodicity_change, PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY)
        self.assertEqual(tm2.declaration_periodicity, PvcTbsLicense.DECLARATION_PERIOD_DEFAULT)

