# Copyright (c) 2015 BaseSoft B.V.
# See LICENSE for details.
import datetime, decimal
from decimal import Decimal
from sqlalchemy.orm.util import was_deleted

from basesoft.util.virtual_date import today, set_date
from basesoft.web.core.exceptions import PublicUserException
from pvcycle.product.controllers import PvcProductController
from pvcycle.product.models import PvcTbsProduct
from pvcycle.test.unit import pvc_unittest


class TestCase(pvc_unittest.PvcTestCase):
    def setUp(self):
        self.connect_db()

        self.operator = self._create_operator_account()

        self.controller = PvcProductController(self.sa_session, account_id=self.operator.id)

    def test_get_products_for_tbs(self):
        # by default the subset of available products for all tbs's should be the same as for a new tbs.
        products = self.controller.get_available_products_for_tbs(None)
        tbs_account = self._create_tbs_account(self.operator)

        tbs_products = self.controller.get_available_products_for_tbs(tbs_account.default_group.license.id)
        self.assertEqual(products, tbs_products)

    def test_save_or_update_tbs_product(self):

        tbs_account = self._create_tbs_account(self.operator)
        tbs = tbs_account.default_group.license

        products = self.controller.get_available_products_for_tbs(tbs.id)

        children = len(self.controller.get_available_products_for_tbs(tbs.id, products[0]['id']))

        tbs_product = self.controller.save_or_update_tbs_product(None, 'EEE-TBS-TEST', 'WEEE', tbs.id,
            parent_level_id=products[0]['id'], basis_for_invoicing=PvcTbsProduct.INVOICE_BASED_ON_CUSTOM_UNIT, custom_unit=PvcTbsProduct.CUSTOM_UNIT_SURFACE)

        self.assertEqual(tbs_product.product.parent_level.id, products[0]['id'])
        self.assertEqual(tbs_product.basis_for_invoicing, PvcTbsProduct.INVOICE_BASED_ON_CUSTOM_UNIT)
        self.assertEqual(children+1, len(self.controller.get_available_products_for_tbs(tbs.id, products[0]['id'])))

    def test_tbs_product_including_parent_levels(self):
        tbs_account = self._create_tbs_account(self.operator)
        tbs = tbs_account.default_group.license

        products = self.controller.get_available_products_for_tbs(tbs.id)
        self.assertEqual(products[0]['status'], None)
        children = len(self.controller.get_available_products_for_tbs(tbs.id, products[0]['id']))

        tbs_product = self.controller.save_or_update_tbs_product(None, 'EEE-TBS-TEST', 'WEEE', tbs.id,
            parent_level_id=products[0]['id'], basis_for_invoicing=PvcTbsProduct.INVOICE_BASED_ON_CUSTOM_UNIT, custom_unit=PvcTbsProduct.CUSTOM_UNIT_SURFACE)

        products = self.controller.get_available_products_for_tbs(tbs.id)
        self.assertEqual(products[0]['status'], PvcTbsProduct.STATUS_INACTIVE)

    def test_get_recursive_fee(self):

        set_date(datetime.date(2013, 1, 1))

        tbs_account = self._create_tbs_account(self.operator)
        sl = self._create_tbs_service_level(tbs_account)
        tbs = tbs_account.default_group.license

        products = self.controller.get_available_products_for_tbs(tbs.id)

        pp = products[0]

        tbs_product = self.controller.save_or_update_tbs_product(None, 'EEE-TBS-TEST', 'WEEE', tbs.id,
            parent_level_id=pp['id'], basis_for_invoicing=PvcTbsProduct.INVOICE_BASED_ON_CUSTOM_UNIT, custom_unit=PvcTbsProduct.CUSTOM_UNIT_SURFACE)

        tbs_product_parent = self.controller.get_tbs_product_by_weee_product(pp['id'], tbs.id)

        # add a fee for parent level only
        self.controller.save_contribution_fee(tbs_product_parent.id, sl.id, decimal.Decimal('2.015'), today())

        set_date(datetime.date(2013, 5, 1))

        fee = tbs_product.get_current_fee(sl.id)

        self.assertEqual(decimal.Decimal('2.015'), fee)

    def test_brand_merge(self):
        self._reset_brands()

        brand1 = self.controller._get_or_save_brand('Philips')
        brand2 = self.controller._get_or_save_brand('Samsumg')
        brand3 = self.controller._get_or_save_brand('Samsung')
        brand4 = self.controller._get_or_save_brand('Sony')
        brand5 = self.controller._get_or_save_brand('Phillllips')
        brand6 = self.controller._get_or_save_brand('Filips')

        self.controller.merge_brand_with(brand6.id, brand1.id)

        merge = self.controller.mark_merge_brands_as_processing()
        # this brand is already processing.
        self.assertRaises(PublicUserException, self.controller.merge_brand_with, brand6.id, brand5.id)
        # the target is already being merged
        self.assertRaises(PublicUserException, self.controller.merge_brand_with, brand5.id, brand6.id)

        self.controller.merge_brand_with(brand5.id, brand1.id)
        self.controller.merge_brand_with(brand2.id, brand3.id)

        merge += self.controller.mark_merge_brands_as_processing()
        self.assertRaises(AssertionError, self.controller.merge_brands, [brand1])
        self.controller.merge_brands(merge)

        self.assertTrue(was_deleted(brand2))
        self.assertTrue(was_deleted(brand5))
        self.assertTrue(was_deleted(brand6))

        self.assertEqual(brand1, self.controller._get_or_save_brand('Philips'))
        self.assertEqual(brand4, self.controller._get_or_save_brand('Sony'))
        self.assertEqual(brand3, self.controller._get_or_save_brand('Samsung'))

    def test_cancel_brand_merge(self):
        self._reset_brands()

        brand1 = self.controller._get_or_save_brand('Philips')
        brand2 = self.controller._get_or_save_brand('Filips')
        self.controller.merge_brand_with(brand2.id, brand1.id)

        brand = self.controller.cancel_merge(brand2.id)
        self.assertEqual(brand.merge_status, None)
        self.assertEqual(brand.merge_brand, None)

        self.controller.merge_brand_with(brand2.id, brand1.id)
        self.controller.mark_merge_brands_as_processing()

        self.assertRaises(PublicUserException, self.controller.cancel_merge, brand2.id)

    def test_brand_rename(self):
        self._reset_brands()

        self.controller._get_or_save_brand('Philips')
        brand2 = self.controller._get_or_save_brand('Filips')

        # we cant rename to an already existing name.
        self.assertRaises(PublicUserException, self.controller.update_brand_name, brand2.id, 'Philips')

        self.controller.update_brand_name(brand2.id, 'My First Rename')
        self.sa_session.expire(brand2)
        self.assertEqual(brand2.name, 'My First Rename')

    def test_product_fee_periods_no_next_fee_set(self):
        set_date(datetime.date(2013, 3, 1))

        tbs_account = self._create_tbs_account(self.operator)
        service_level = self._create_tbs_service_level(tbs_account, admin_fee=Decimal('10.2'), year=today().year)
        tbs = tbs_account.default_group.license

        products = self.controller.get_available_products_for_tbs(tbs.id)

        children = len(self.controller.get_available_products_for_tbs(tbs.id, products[0]['id']))

        tbs_product = self.controller.save_or_update_tbs_product(None, 'EEE-TBS-TEST', 'WEEE', tbs.id,
            parent_level_id=products[0]['id'], basis_for_invoicing=PvcTbsProduct.INVOICE_BASED_ON_CUSTOM_UNIT, custom_unit=PvcTbsProduct.CUSTOM_UNIT_SURFACE)

        # set fee for period.
        pf = self.controller.save_contribution_fee(tbs_product.id, service_level.id, fee=Decimal('9.72'), start_date=datetime.date(2013, 4, 1))

        # set new date 4 months in future
        set_date(datetime.date(2013, 8, 1))

        # get all possible periods
        yearly, quarterly, monthly = self.controller.get_periods_for_fee(pf)

        self.assertEqual(len(yearly), 0)
        self.assertEqual(len(quarterly), 2)
        self.assertEqual(len(monthly), 5)  # april, may, june, july and august

        self.assertEqual(quarterly[0], (2013, 2, ), (2013, 3, ), )
        self.assertEqual(monthly, [(2013, 4, ), (2013, 5, ), (2013, 6, ), (2013, 7, ), (2013, 8, ), ])

    def test_product_fee_periods_multiple_quarters(self):
        set_date(datetime.date(2013, 3, 1))

        tbs_account = self._create_tbs_account(self.operator)
        service_level = self._create_tbs_service_level(tbs_account, admin_fee=Decimal('10.2'), year=today().year)
        tbs = tbs_account.default_group.license

        products = self.controller.get_available_products_for_tbs(tbs.id)

        tbs_product = self.controller.save_or_update_tbs_product(None, 'EEE-TBS-TEST', 'WEEE', tbs.id,
            parent_level_id=products[0]['id'], basis_for_invoicing=PvcTbsProduct.INVOICE_BASED_ON_CUSTOM_UNIT, custom_unit=PvcTbsProduct.CUSTOM_UNIT_SURFACE)

        # set fee for period.
        pf = self.controller.save_contribution_fee(tbs_product.id, service_level.id, fee=Decimal('9.72'), start_date=datetime.date(2013, 4, 1))

        # set new date 4 months in future
        set_date(datetime.date(2013, 11, 1))

        # get all possible periods
        yearly, quarterly, monthly = self.controller.get_periods_for_fee(pf)

        self.assertEqual(len(yearly), 0)
        self.assertEqual(len(quarterly), 3)
        self.assertEqual(len(monthly), 8)  # april, may, june, july and august

        self.assertEqual(quarterly[0], (2013, 2, ))
        self.assertEqual(quarterly[1], (2013, 3, ))
        self.assertEqual(quarterly[2], (2013, 4, ))

    def test_product_fee_periods_multiple_years(self):
        set_date(datetime.date(2013, 3, 1))

        tbs_account = self._create_tbs_account(self.operator)
        service_level = self._create_tbs_service_level(tbs_account, admin_fee=Decimal('10.2'), year=today().year)
        tbs = tbs_account.default_group.license

        products = self.controller.get_available_products_for_tbs(tbs.id)

        tbs_product = self.controller.save_or_update_tbs_product(None, 'EEE-TBS-TEST', 'WEEE', tbs.id,
            parent_level_id=products[0]['id'], basis_for_invoicing=PvcTbsProduct.INVOICE_BASED_ON_CUSTOM_UNIT, custom_unit=PvcTbsProduct.CUSTOM_UNIT_SURFACE)

        # set fee for period.
        pf = self.controller.save_contribution_fee(tbs_product.id, service_level.id, fee=Decimal('9.72'), start_date=datetime.date(2013, 4, 1))

        # set new date 4 months in future
        set_date(datetime.date(2014, 7, 1))

        # get all possible periods
        yearly, quarterly, monthly = self.controller.get_periods_for_fee(pf)

        self.assertEqual(len(yearly), 1)
        self.assertEqual(len(quarterly), 6)
        self.assertEqual(len(monthly), 16)

        self.assertEqual(yearly[0], 2014)
        self.assertEqual(quarterly, [(2013, 2, ), (2013, 3, ), (2013, 4, ), (2014, 1, ), (2014, 2, ), (2014, 3, ), ])
        self.assertEqual(monthly, [
            (2013, 4, ), (2013, 5, ), (2013, 6, ), (2013, 7, ), (2013, 8, ), (2013, 9, ), (2013, 10, ), (2013, 11, ), (2013, 12, ),
            (2014, 1, ), (2014, 2, ), (2014, 3, ), (2014, 4, ), (2014, 5, ), (2014, 6, ), (2014, 7, ),
        ])

    def test_tbs_member_product_selection(self):
        tbs_account = self._create_tbs_account(self.operator)
        tbs = tbs_account.default_group.license

        products = self.controller.get_available_products_for_tbs(tbs.id)

        tbs_product = self.controller.save_or_update_tbs_product(None, 'EEE-TBS-TEST', 'WEEE', tbs.id,
            parent_level_id=products[0]['id'], basis_for_invoicing=PvcTbsProduct.INVOICE_BASED_ON_CUSTOM_UNIT, custom_unit=PvcTbsProduct.CUSTOM_UNIT_SURFACE)

        tbs_product2 = self.controller.save_or_update_tbs_product(None, 'EEE-TBS-TEST-2', 'WEEE2', tbs.id,
            parent_level_id=products[1]['id'], basis_for_invoicing=PvcTbsProduct.INVOICE_BASED_ON_CUSTOM_UNIT, custom_unit=PvcTbsProduct.CUSTOM_UNIT_SURFACE)

        service_level = self._create_tbs_service_level(tbs_account, admin_fee=Decimal('22.2'), year=today().year)

        ma = self._create_member_account()

        # apply at TBS
        tm = self._member_apply_at_tbs(ma, tbs_account, service_level)

        pc = PvcProductController(self.sa_session, account_id=ma.id)
        products = pc.save_tbs_member_products(tbs.id, tm.member.id, [tbs_product.id, tbs_product2.id])
        self.assertEqual(len(products), 2)

    def test_tbs_member_product_deactivation(self):
        tbs_account = self._create_tbs_account(self.operator)
        tbs = tbs_account.default_group.license

        products = self.controller.get_available_products_for_tbs(tbs.id)

        tbs_product = self.controller.save_or_update_tbs_product(None, 'EEE-TBS-TEST', 'WEEE', tbs.id,
            parent_level_id=products[0]['id'], basis_for_invoicing=PvcTbsProduct.INVOICE_BASED_ON_CUSTOM_UNIT, custom_unit=PvcTbsProduct.CUSTOM_UNIT_SURFACE)

        service_level = self._create_tbs_service_level(tbs_account, admin_fee=Decimal('22.2'), year=today().year)

        ma = self._create_member_account()

        # apply at TBS
        tm = self._member_apply_at_tbs(ma, tbs_account, service_level)

        pc = PvcProductController(self.sa_session, account_id=ma.id)
        product = pc.save_tbs_member_products(tbs.id, tm.member.id, [tbs_product.id]).pop()
        self.assertTrue(product.is_active)

        pc.deactivate_tbs_member_product(product.id)
        self.sa_session.expire(product)

        self.assertFalse(product.is_active)

    def test_tbs_member_brand_deactivation(self):
        tbs_account = self._create_tbs_account(self.operator)
        tbs = tbs_account.default_group.license
        tbs.set_config('pvc.tbs.contribution.split.by.brand', True)

        products = self.controller.get_available_products_for_tbs(tbs.id)

        tbs_product = self.controller.save_or_update_tbs_product(None, 'EEE-TBS-TEST', 'WEEE', tbs.id,
            parent_level_id=products[0]['id'], basis_for_invoicing=PvcTbsProduct.INVOICE_BASED_ON_CUSTOM_UNIT, custom_unit=PvcTbsProduct.CUSTOM_UNIT_SURFACE)

        service_level = self._create_tbs_service_level(tbs_account, admin_fee=Decimal('22.2'), year=today().year)

        ma = self._create_member_account()

        # apply at TBS
        tm = self._member_apply_at_tbs(ma, tbs_account, service_level)

        pc = PvcProductController(self.sa_session, account_id=ma.id)
        product = pc.get_or_save_tbs_member_product(tm.id, tbs_product.id, brands=['Philips', 'Sony'])
        self.assertEqual(len(product.brands), 2)

        brand = product.brands[0].brand

        pc.deactivate_product_brand(product.id, brand.id)

        self.sa_session.expire(product)
        mb = product.brands[0]
        self.assertFalse(mb.is_active)


