# Copyright (c) 2013 BaseSoft B.V.
# See LICENSE for details.
import time
from pvcycle.product.controllers import PvcProductController
from pvcycle.product.models import PvcTbsMemberProduct, PvcTbsProduct, PvcTbsProductFee
import suds
import datetime
import time

from django.conf import settings
from decimal import Decimal

from pvcycle.test.unit import pvc_unittest
from basesoft.util.virtual_date import today, set_date, now
from suds import WebFault
from basesoft.web.contrib.i18n.models import Country, Locale
from pvcycle.core.models import PvcTbsMember
from pvcycle.core.controllers import PvcMemberController
from pvcycle.order.controllers import PvcOrderController, PvcContributionOrderController
from basesoft.web.contrib.order.controllers import PaymentController
from basesoft.web.contrib.order.models import PaymentMethodBankTransfer,\
    OrderPaymentStatus, OrderRule, InvoiceFile


class TestCase(pvc_unittest.PvcTestCase):
    def setUp(self):
        self.connect_db()
        self._start_django()
        self.client = suds.client.Client('%spvcycle.core/api/wsdl' % settings.BASE_URL, cache=None)

        set_date(datetime.datetime.now().date())

    def test_login(self):
        """
        TODO: this should be IQPortal level.
        """
        account = self._create_active_account()
        self.assertRaises(WebFault, self.client.service.login, account.name, 'test_password') # no login permission

        self._attach_permissions_to_group(account.default_group, 'soap_login')

        # raises an exception on failure
        self.assertEqual(None, self.client.service.login(account.name, 'test_password'))


    def test_get_take_back_schemes(self):
        """
        This tests the method to get Take Back Schemes for the current accountant.
        """
        self.assertRaises(suds.WebFault, self.client.service.get_take_back_schemes)

        operator_account = self._create_operator_account()
        tbs_account = self._create_tbs_account(operator_account)
        tbs_license = tbs_account.default_group.license
        accountant = self._create_accountant_account(tbs_license, operator_account)

        self.client.service.login(accountant.name, 'test_password')
        rv = self.client.service.get_take_back_schemes()

        self.assertEqual(rv.item[0].name, tbs_license.name)
        self.assertEqual(rv.item[0].id, tbs_license.id)

    def test_get_tbs_master_data(self):
        """
        This test the export of the master data of all the TBS' for a specific Accountant.
        """
        operator_account = self._create_operator_account()
        tbs_account = self._create_tbs_account(operator_account)
        tbs_license = tbs_account.default_group.license
        accountant = self._create_accountant_account(tbs_license, operator_account)
        self.client.service.login(accountant.name, 'test_password')

        rv = self.client.service.get_tbs_master_data()
        tbs = rv.item[0]
        self.assertEqual(tbs.id, tbs_license.id)
        self.assertEqual(tbs.company.name, tbs_license.name)
        self.assertEqual(tbs.contact_person_data.first_name, tbs_license.company.contact_person.first_name)

    def test_get_member_master_data_content(self):
        """
        This test verifies all the fields for a member at a TBS.
        """
        operator_account = self._create_operator_account()
        tbs_account = self._create_tbs_account(operator_account)
        tbs_license = tbs_account.default_group.license
        sl = self._create_tbs_service_level(tbs_account, 'Some Service Level', Decimal('40.00'), Decimal('0.10'), today().year)

        state = {
            'dbID': None,
            'company': {
                'name': 'Member For SOAP',
                'website': 'http://soapmember.net',
                'vat_no': 'BE0893027827'
            },
            'address_data': {
                'street': 'SOAP Street 12',
                'city': 'Bathtub',
                'postcode': '90210',
                'country': self.sa_session.query(Country).filter(Country.iso_3166_a2=='GB').one().id
            },
            'bank_data': {
                'iban': "SOAP1010101010",
                'bic': "BIC974234",
                'number': "324234234",
            },
            'contact_person_data': {
                'last_name': 'SOAP Last Name',
                'first_name': 'SOAP First Name',
                'preferred_locale': self.sa_session.query(Locale).filter(Locale.locale=='en_GB').one().id,
                'gender': 'Male',
                'phone': '080012345678',
            },
            'invoice_person_data': {
                'last_name': 'SOAP INVOICE Last Name',
                'first_name': 'SOAP INVOICE First Name',
                'preferred_locale': self.sa_session.query(Locale).filter(Locale.locale=='en_GB').one().id,
                'gender': 'Female',
                'phone': '090012345678',
            }
        }
        member_account = self._create_member_account(state=state)
        member_license = member_account.default_group.license
        tbs_member = self._member_apply_at_tbs(member_account, tbs_account, sl)
        tbs_member.status = PvcTbsMember.STATUS_APPROVED
        self.sa_session.add(tbs_member)
        self.sa_session.flush()

        accountant = self._create_accountant_account(tbs_license, operator_account)
        self.client.service.login(accountant.name, 'test_password')

        rv = self.client.service.get_member_master_data(tbs_license.id)
        member = rv.item[0]

        self.assertEqual(member.id, member_license.id)
        self.assertEqual(member.company.name, state['company']['name'])
        self.assertEqual(member.address_data.street, state['address_data']['street'])
        self.assertEqual(member.address_data.country, state['address_data']['country'])
        self.assertEqual(member.address_data.country_iso_3166_a2, 'GB')
        self.assertEqual(member.address_data.country_iso_3166_a3, 'GBR')

        self.assertEqual(member.invoice_person_data.last_name, state['invoice_person_data']['last_name'])
        self.assertEqual(member.invoice_person_data.preferred_locale_str, 'en_GB')

    def test_get_member_master_data(self):
        """
        This test the export of the master data of multiple members at a specific TBS.
        """
        operator_account = self._create_operator_account()
        tbs_account = self._create_tbs_account(operator_account)

        sl1 = self._create_tbs_service_level(tbs_account, 'SL ONE', Decimal('40.00'), Decimal('0.10'), today().year)
        sl2 = self._create_tbs_service_level(tbs_account, 'SL TWO', Decimal('60.00'), Decimal('0.20'), today().year, level=2)
        sl3 = self._create_tbs_service_level(tbs_account, 'SL THREE', Decimal('80.00'), Decimal('0.30'), today().year, level=3)

        tbs_license = tbs_account.default_group.license
        self._create_multiple_producers_at_tbs(tbs_account, service_levels=[sl1, sl2, sl3])

        accountant = self._create_accountant_account(tbs_license, operator_account)
        self.client.service.login(accountant.name, 'test_password')

        # we need to supply a TBS ID.
        # TODO: this should be handled by nillable=False in the WSDL
        self.assertRaises(WebFault, self.client.service.get_member_master_data)

        rv = self.client.service.get_member_master_data(tbs_license.id)
        self.assertEqual(len(rv.item), 10)

    def test_get_administrative_order_data(self):
        """
        Simple test for getting Administrative Fee (Registration) order data from the SOAP server.
        """

        operator_account = self._create_operator_account()
        tbs_account = self._create_tbs_account(operator_account)

        sl1 = self._create_tbs_service_level(tbs_account, 'SERVICE', Decimal('40.00'), Decimal('0.10'), today().year)

        tbs_license = tbs_account.default_group.license


        accountant = self._create_accountant_account(tbs_license, operator_account)
        self.client.service.login(accountant.name, 'test_password')

        # create a member
        ma = self._create_member_account()

        # apply at TBS
        tm = self._member_apply_at_tbs(ma, tbs_account, sl1)

        # activate member to automatically create an order entry for the Administration Fee
        mc = PvcMemberController(self.sa_session, account_id=tbs_account.id)
        mc.save_tbs_member(tm.id, {'status': PvcTbsMember.STATUS_APPROVED, 'is_full_member': False, 'is_representative': False})

        # get order from soap
        rv = self.client.service.get_unexported_orders_for_merchant(tbs_license.id)
        orders = rv.item
        self.assertEqual(len(orders), 1)
        order = orders.pop()

        self.assertEqual(order.total_amount_excluding_vat, 40.0)
        self.assertEqual(order.currency_code, 'EUR')

        self.assertEqual(order.current_payment_method_id, None)

        self.assertEqual(len(order.lines), 1)
        self.assertEqual(order.merchant_id, tbs_license.id)
        self.assertEqual(order.client_id, tm.member.id)

        self.assertEqual(order.lines[0][0]['vat_rate'], 0)
        self.assertEqual(order.lines[0][0]['quantity'], 1)
        self.assertEqual(order.lines[0][0]['gl_account'], 'GL123456')
        self.assertEqual(order.lines[0][0]['properties'][0][0].value_str, 'SERVICE')

        # complete the order, expect payment information to have changed
        pc = PaymentController(self.sa_session, account_id=operator_account.id)
        pm1 = pc.save_or_update_payment_method_for_merchant(PaymentMethodBankTransfer, label='bank transfer', merchant=tbs_license)
        accountant_pc = PaymentController(self.sa_session, account_id=accountant.id, payment_method=pm1)
        accountant_pc.complete_payment(order.id, "all good")

        # get order from soap
        order = self.client.service.get_order_by_id(order.id)

        # assert updated payment info
        self.assertEqual(order.current_payment_method_id, pm1.id)
        self.assertEqual(order.current_payment_method_label, 'bank transfer')
        self.assertEqual(order.current_payment_status, OrderPaymentStatus.PAYMENT_COMPLETED)
        self.assertEqual(order.current_payment_date.year, datetime.datetime.now().year)


    def test_get_previous_exported_order(self):
        """
        This tests the re-getting of a previously exported order.
        """
        operator_account = self._create_operator_account()
        tbs_account = self._create_tbs_account(operator_account)

        sl1 = self._create_tbs_service_level(tbs_account, 'SERVICE', Decimal('40.00'), Decimal('0.10'), today().year)

        tbs_license = tbs_account.default_group.license

        accountant = self._create_accountant_account(tbs_license, operator_account)
        self.client.service.login(accountant.name, 'test_password')

        self.assertRaises(WebFault, self.client.service.get_order_by_id, 1) # this order does not exist.

        # create a member
        ma = self._create_member_account()

        # apply at TBS
        tm = self._member_apply_at_tbs(ma, tbs_account, sl1)

        # activate member to automatically create an order entry for the Administration Fee
        mc = PvcMemberController(self.sa_session, account_id=tbs_account.id)
        mc.save_tbs_member(tm.id, {'status': PvcTbsMember.STATUS_APPROVED, 'is_full_member': False, 'is_representative': False})

        # get the order.
        poc = PvcOrderController(self.sa_session, account_id=ma.id)
        order = poc.get().pop()

        self.assertRaises(WebFault, self.client.service.get_order_by_id, order.id) # this order is not available for export.

        rv = self.client.service.get_unexported_orders_for_merchant(tbs_license.id)
        orders = rv.item
        self.assertEqual(len(orders), 1)

        exported = self.client.service.get_order_by_id(order.id)
        self.assertEqual(exported.id, order.id)
        self.assertEqual(exported.merchant_id, tbs_license.id)
        self.assertEqual(exported.client_id, tm.member.id)

    def test_get_previously_exported_orders_by_date(self):

        operator_account = self._create_operator_account()
        tbs_account = self._create_tbs_account(operator_account)

        sl1 = self._create_tbs_service_level(tbs_account, 'SERVICE', Decimal('40.00'), Decimal('0.10'), today().year)

        tbs_license = tbs_account.default_group.license

        accountant = self._create_accountant_account(tbs_license, operator_account)
        self.client.service.login(accountant.name, 'test_password')

        # create a member
        ma = self._create_member_account()

        # apply at TBS
        tm = self._member_apply_at_tbs(ma, tbs_account, sl1)

        # activate member to automatically create an order entry for the Administration Fee
        mc = PvcMemberController(self.sa_session, account_id=tbs_account.id)
        mc.save_tbs_member(tm.id, {'status': PvcTbsMember.STATUS_APPROVED, 'is_full_member': False, 'is_representative': False})

        order1 = self.client.service.get_unexported_orders_for_merchant(tbs_license.id).item.pop()

        order2 = self.client.service.get_orders_by_export_date(tbs_license.id, today()).item.pop()
        self.assertEqual(order1.id, order2.id)

    def test_upload_invoice_document_for_order(self):
        operator_account = self._create_operator_account()
        tbs_account = self._create_tbs_account(operator_account)

        sl1 = self._create_tbs_service_level(tbs_account, 'SERVICE', Decimal('40.00'), Decimal('0.10'), today().year)

        tbs_license = tbs_account.default_group.license

        accountant = self._create_accountant_account(tbs_license, operator_account)
        self.client.service.login(accountant.name, 'test_password')

        # create a member
        ma = self._create_member_account()

        # apply at TBS
        tm = self._member_apply_at_tbs(ma, tbs_account, sl1)

        # activate member to automatically create an order entry for the Administration Fee
        mc = PvcMemberController(self.sa_session, account_id=tbs_account.id)
        mc.save_tbs_member(tm.id, {'status': PvcTbsMember.STATUS_APPROVED, 'is_full_member': False, 'is_representative': False})

        # get the order.
        poc = PvcOrderController(self.sa_session, account_id=ma.id)
        order = poc.get().pop()

        self.assertRaises(WebFault, self.client.service.attach_invoice_to_order, order.id, {'file_name': 'invoice.txt', 'payload': 'Some content'.encode('base64')}, False)

        self.client.service.get_unexported_orders_for_merchant(tbs_license.id)

        self.assertRaises(WebFault, self.client.service.attach_invoice_to_order, order.id, {'file_name': 'invoice.txt', 'payload': 'Some content'.encode('base64')}, True)
        self.client.service.attach_invoice_to_order(order.id, {'file_name': 'invoice.txt', 'payload': 'Some content'.encode('base64')}, False)

        self.sa_session.expire(order)
        # The TBS should have view privs on the invoice.
        mc.assert_privileges_on_object(order.invoice, InvoiceFile.PRIV_VIEW)

        # The Accountant should have view privs on the invoice.
        mc = PvcMemberController(self.sa_session, account_id=accountant.id)
        mc.assert_privileges_on_object(order.invoice, InvoiceFile.PRIV_VIEW)

        # The operator should have view privs on the invoice.
        mc = PvcMemberController(self.sa_session, account_id=operator_account.id)
        mc.assert_privileges_on_object(order.invoice, InvoiceFile.PRIV_VIEW)

    def test_upload_order_payment_status(self):

        operator_account = self._create_operator_account()
        tbs_account = self._create_tbs_account(operator_account)

        sl1 = self._create_tbs_service_level(tbs_account, 'SERVICE', Decimal('40.00'), Decimal('0.10'), today().year)

        tbs_license = tbs_account.default_group.license

        accountant = self._create_accountant_account(tbs_license, operator_account)

        ma = self._create_member_account()
        tm = self._member_apply_at_tbs(ma, tbs_account, sl1)

        # activate member to automatically create an order entry for the Administration Fee
        mc = PvcMemberController(self.sa_session, account_id=tbs_account.id)
        mc.save_tbs_member(tm.id, {'status': PvcTbsMember.STATUS_APPROVED, 'is_full_member': False, 'is_representative': False})

        # get the order.
        poc = PvcOrderController(self.sa_session, account_id=ma.id)
        order = poc.get().pop()

        # setup payment methods
        pc = PaymentController(self.sa_session, account_id=operator_account.id)
        pm1 = pc.save_or_update_payment_method_for_merchant(PaymentMethodBankTransfer, label='bank transfer', merchant=tbs_license)

        # let accountant complete the order
        self.client.service.login(accountant.name, 'test_password')

        # update payment status
        self.client.service.update_order_payment(order.id, OrderPaymentStatus.PAYMENT_COMPLETED, pm1.polymorphic_type, 'moneys received', now())

        # assert payment status
        self.assertEqual(order.get_current_payment_status(), OrderPaymentStatus.PAYMENT_COMPLETED)


    def test_bug_4124(self):
        """
        asserts bug #4124 : error raised on retrieval of order that has multiple payment statusses
        """

        operator_account = self._create_operator_account()
        tbs_account = self._create_tbs_account(operator_account)

        sl1 = self._create_tbs_service_level(tbs_account, 'SERVICE', Decimal('40.00'), Decimal('0.10'), today().year)

        tbs_license = tbs_account.default_group.license

        accountant = self._create_accountant_account(tbs_license, operator_account)

        ma = self._create_member_account()
        tm = self._member_apply_at_tbs(ma, tbs_account, sl1)

        # activate member to automatically create an order entry for the Administration Fee
        mc = PvcMemberController(self.sa_session, account_id=tbs_account.id)
        mc.save_tbs_member(tm.id, {'status': PvcTbsMember.STATUS_APPROVED, 'is_full_member': False, 'is_representative': False})

        # get the order.
        poc = PvcOrderController(self.sa_session, account_id=ma.id)
        order = poc.get().pop()
        order.export_date = datetime.datetime.now().date() # set as exported so we can attach an invoice

        # setup payment methods
        pc = PaymentController(self.sa_session, account_id=operator_account.id)
        pm1 = pc.save_or_update_payment_method_for_merchant(PaymentMethodBankTransfer, label='bank transfer', merchant=tbs_license)

        # instanciate the user payment specific controller
        user_pc = pc.get_payment_controller(pm1.id)(self.sa_session, account_id=ma.id, payment_method=pm1)

        # the user marks the order as paid
        time.sleep(1); user_pc.submit_payment(order.id, "transfer description")
        self.assertEqual(order.get_current_payment_status(), OrderPaymentStatus.PAYMENT_SUBMITTED)

        # let accountant complete the order
        self.client.service.login(accountant.name, 'test_password')

        # update payment status
        time.sleep(1); self.client.service.update_order_payment(order.id, OrderPaymentStatus.PAYMENT_COMPLETED, pm1.polymorphic_type, 'moneys received', now())

        # get order.. this used to trigger an error
        order_state = self.client.service.get_order_by_id(order.id)

        # payment status should be completed
        self.assertEqual(order_state['current_payment_status'], OrderPaymentStatus.PAYMENT_COMPLETED)

    def _get_member_product(self, tm, tbs_product=None, country=None):
        if tbs_product is None:
            tbs_product = self.tbs_product

        product = PvcTbsMemberProduct()
        product.tbs_product = tbs_product
        product.tbs_member = tm
        product.is_active = True
        product.country = country if country is not None else tm.tbs.country
        self.sa_session.add(product)
        self.sa_session.flush()
        return product


    def _trash_all_rules(self, operator_account):
        for rule in self.sa_session.query(OrderRule).filter(OrderRule.domain==operator_account.default_group.license.domain):
            self.sa_session.delete(rule)

    def _make_rule(self, operator_account, name, **kwargs):

        controller = PvcContributionOrderController(self.sa_session, account_id=operator_account.id)

        controller.save_rule(None, {
            'order_type': kwargs.get('order_type', 'MockOrder'),
            'name': name,
            'enabled': kwargs.get('enabled', True),
            'execute_order': kwargs.get('execute_order', None),
            'description': kwargs.get('description', ''),
            'start_date': kwargs.get('start_date', now()),
            'end_date': kwargs.get('end_date', None),
            'match_all': kwargs.get('match_all', True),
            'stop_processing': kwargs.get('stop_processing', False),
            'conditions': kwargs.get('conditions', []),
            'actions': kwargs.get('actions', [])
        })

    def test_contribution_order(self):

        set_date(datetime.date(2015, 1, 1))

        operator_account = self._create_operator_account()
        tbs_account = self._create_tbs_account(operator_account)
        service_level = self._create_tbs_service_level(tbs_account, 'SERVICE', Decimal('40.00'), Decimal('0.10'), today().year)
        tbs_license = tbs_account.default_group.license
        accountant = self._create_accountant_account(tbs_license, operator_account)

        tbs_account.default_group.license.set_config('pvc.tbs.contribution.automatic.end.user.distinction', True)
        tbs_account.default_group.license.set_config('pvc.tbs.contribution.split.by.brand', True)

        pc = PvcProductController(self.sa_session, account_id=operator_account.id)
        products = pc.get_available_products_for_tbs(tbs_license.id)

        tbs_product_weight = pc.save_or_update_tbs_product(None, 'EEE-TBS-TEST', 'WEEE', tbs_license.id,
            parent_level_id=products[0]['id'], basis_for_invoicing=PvcTbsProduct.INVOICE_BASED_ON_WEIGHT
        )

        tbs_product_custom_unit = pc.save_or_update_tbs_product(None, 'EEE-TBS-TEST-4', 'WEEE-3', tbs_license.id,
            parent_level_id=products[1]['id'],
            basis_for_invoicing=PvcTbsProduct.INVOICE_BASED_ON_CUSTOM_UNIT,
            custom_unit=PvcTbsProduct.CUSTOM_UNIT_NOMINAL_POWER
        )

        pc.save_contribution_fee(tbs_product_weight.id, service_level.id, Decimal('0.015'), today())
        pc.save_contribution_fee(tbs_product_weight.id, service_level.id, Decimal('0.016'), today(), PvcTbsProductFee.DIMENSION_IS_B2B)
        pc.save_contribution_fee(tbs_product_weight.id, service_level.id, Decimal('0.017'), today(), PvcTbsProductFee.DIMENSION_IS_B2C)

        pc.save_contribution_fee(tbs_product_custom_unit.id, service_level.id, Decimal('0.018'), today())
        pc.save_contribution_fee(tbs_product_custom_unit.id, service_level.id, Decimal('0.019'), today(), PvcTbsProductFee.DIMENSION_IS_B2B)
        pc.save_contribution_fee(tbs_product_custom_unit.id, service_level.id, Decimal('0.020'), today(), PvcTbsProductFee.DIMENSION_IS_B2C)


        ma = self._create_member_account()
        tm = self._member_apply_at_tbs(ma, tbs_account, service_level)

        self._trash_all_rules(operator_account)
        self._make_rule(operator_account, 'producer',
            order_type = 'PvcContributionOrder',
            conditions = [],
            actions = [{
                'action': 'fixed_amount_discount',
                'description': 'Discount',
                'operator': None, 'value_text': None, 'value_bool': None, 'value_int': None,
                'value_decimal': 10,
                'tax_class_id': None, 'gl_account_code': None
            }, {
                'action': 'pvc_contribution_b2b_percentage',
                'description': '',
                'operator': None, 'value_text': None, 'value_bool': None, 'value_int': 10,
                'value_decimal': None,
                'tax_class_id': None, 'gl_account_code': None
            }]
        )

        # activate member to automatically create an order entry for the Administration Fee
        mc = PvcMemberController(self.sa_session, account_id=tbs_account.id)
        mc.save_tbs_member(tm.id, {
            'status': PvcTbsMember.STATUS_APPROVED,
            'is_full_member': False,
            'is_representative': False,
            'obligation_start_date': datetime.date(2015, 1, 1)})

        product_weight = self._get_member_product(tm, tbs_product_weight)
        product_custom_unit = self._get_member_product(tm, tbs_product=tbs_product_custom_unit)

        cc = PvcContributionOrderController(self.sa_session, account_id=ma.id)
        lines = [{
            'tbs_member_product': product_weight,
            'quantity': 200,
            'weight': Decimal('1500.3'),
            'custom_unit': None,
            'brand': pc.get_or_save_brand('Brand 1', product_weight),
        }, {
            'tbs_member_product': product_custom_unit,
            'quantity': 12,
            'weight': Decimal(121),
            'custom_unit': Decimal('22'),
            'brand': pc.get_or_save_brand('Brand 2', product_custom_unit),
        }]
        order = cc._add_contribution_order(tm, lines, 2015, month=3)

        # member submits order
        cc.submit_order(order.id)

        # setup payment methods
        pc = PaymentController(self.sa_session, account_id=operator_account.id)
        pm1 = pc.save_or_update_payment_method_for_merchant(PaymentMethodBankTransfer, label='bank transfer', merchant=tbs_license)

        # instanciate the user payment specific controller
        user_pc = pc.get_payment_controller(pm1.id)(self.sa_session, account_id=ma.id, payment_method=pm1)

        # the user marks the order as paid
        time.sleep(1)
        user_pc.submit_payment(order.id, "transfer description")
        self.assertEqual(order.get_current_payment_status(), OrderPaymentStatus.PAYMENT_SUBMITTED)

        # let accountant complete the order
        self.client.service.login(accountant.name, 'test_password')

        # update payment status
        time.sleep(1)
        self.client.service.update_order_payment(order.id, OrderPaymentStatus.PAYMENT_COMPLETED, pm1.polymorphic_type, 'moneys received', now())

        rv = self.client.service.get_unexported_orders_for_merchant(tbs_license.id)
        order_cf = rv.item[0]
        order_af = rv.item[1]

        self.assertEqual(order_af.total_amount_excluding_vat, 40.0)
        self.assertEqual(order_cf.total_amount_excluding_vat, 15.79287)

        self.assertEqual(order_cf.properties.item[0].key, 'year')
        self.assertEqual(order_cf.properties.item[0].value_int, 2015)

        self.assertEqual(order_cf.properties.item[1].key, 'month')
        self.assertEqual(order_cf.properties.item[1].value_int, 3)

        self.assertEqual(order_cf.properties.item[2].key, 'quarter')
        self.assertEqual(order_cf.properties.item[2].value_int, None)

        self.assertEqual(order_cf.properties.item[3].key, 'producer_id')
        self.assertEqual(order_cf.properties.item[3].value_int, tm.member.license_id)

        line_0 = order_cf.lines.item[0]

        self.assertEqual(line_0.polymorphic_type, 'PvcContributionOrderLine')
        self.assertEqual(line_0.description, 'EEE-TBS-TEST')
        self.assertEqual(line_0.price_per_unit, 0.017)
        self.assertEqual(line_0.quantity, 180)

        self.assertEqual(line_0.properties.item[0].key, 'weight')
        self.assertEqual(line_0.properties.item[0].value_decimal, 1350.27)

        self.assertEqual(line_0.properties.item[1].key, 'basis_for_invoicing')
        self.assertEqual(line_0.properties.item[1].value_int, 1)

        # handy to check in middleware
        print 'accountant: %s' % accountant


# below is for running the test from eclipse
if __name__ == '__main__':

    from twisted.scripts.trial import run
    import sys
    sys.argv = ['test_member.py', 'pvcycle.test.unit.test_soap.TestCase.test_bug_4124']
    run()
    sys.exit()
