'''
Created on Jul 22, 2013

@author: marten
'''

from dateutil.relativedelta import relativedelta
from basesoft.web.contrib.account.controllers import AccountController
from pvcycle.product.models import PvcTbsProduct

from pvcycle.test.unit import pvc_unittest
from pvcycle.core.controllers import PvcTBSController, PvcMemberController,\
    PvcInvalidFeeException
from pvcycle.core.models import PvcTbsLicense, PvcTbsMember, PvcTbsServiceLevel
import datetime
from decimal import Decimal
from basesoft.web.core.exceptions import PublicUserException
from basesoft.util.virtual_date import today, set_date
from basesoft.web.core.util.controllers import LicenseConfigController

class TestCase(pvc_unittest.PvcTestCase):
    def setUp(self):

        self.connect_db()

        # reset virtual date
        set_date(datetime.datetime.now().date())

        # sysadmin
        self.sysadmin_account = self._create_active_account()
        self.sysadmin_group = self.sysadmin_account.groups[0].group
        self._attach_permissions_to_group(self.sysadmin_group,
            'view_license', 'view_all_domains', 'view_group', 'create_domain', 'create_license',
            'distribute_license_permissions',
            'pvc_operator_license'
        )

        # operator user
        self.operator_account = self._create_operator_account()


    def test_create_take_back_scheme(self):
        """
        This tests the creation of TBS License(s) by the operator.
        """

        tbs_controller = PvcTBSController(account_id=self.operator_account.id)

        # saving
        tbs_license = tbs_controller.save_tbs(self._get_tbs_license_state())
        self.assertTrue(isinstance(tbs_license, PvcTbsLicense))

        # loading
        tbs_state = tbs_controller.load_tbs(tbs_license.id)
        self.assertTrue(tbs_state['company']['name'].startswith('TbsCompany'))
        self.assertEquals(tbs_state['tbs']['tbs_type'], PvcTbsLicense.TBS_TYPE_SUBSIDIARY)
        self.assertEquals(tbs_state['tbs']['country'], 155)

        # tbs user (if needed)
        self.tbs_account = self._create_active_account(license=tbs_license)
        self.tbs_group = self.tbs_account.groups[0].group

    def test_tbs_fee(self):

        # set fixed date
        set_date(datetime.date(2013, 1, 1))

        feb1 = datetime.date(2013, 2, 1)
        jan1_2014 = datetime.date(2014, 1, 1).year

        tbs_controller = PvcTBSController(account_id=self.operator_account.id)
        tbs_license = tbs_controller.save_tbs(self._get_tbs_license_state())

        # create a service level
        service_level = tbs_controller.save_service_level_state(None, {
            'tbs_id': tbs_license.id,
            'name': 'tbs fee 1',
            'contract_annex_text': 'contract text',
            'administration_fees': [{
                'start_year': jan1_2014,
                'fee': Decimal('10.12345'),
            }],
            'contribution_fees': [{
                'start_date': feb1,
                'fee': Decimal('11.12345'),
            }],
        })

        self.assertDeepEqual(tbs_controller.get_service_level_state(service_level.id), {
            'tbs_id': service_level.tbs.id,
            'name': u'tbs fee 1',
            'level': 1,
            'contract_annex_text': 'contract text',
            'status': PvcTbsServiceLevel.STATUS_ACTIVE,
            'administration_fees': [{
                'fee': Decimal('10.12345'),
                'start_year': jan1_2014,
            }],
            'contribution_fees': [{
                'fee': Decimal('11.12345'),
                'start_date': feb1,
            }],
            'product_type': PvcTbsProduct.TYPE_OTHER,
            'guarantee_forecast_min_value': None
        })

        # move to next year
        set_date(today() + datetime.timedelta(days=365))

        # get current fees
        self.assertEqual(service_level.get_current_administration_fee(), Decimal('10.12345'))
        self.assertEqual(service_level.get_current_contribution_fee(), Decimal('11.12345'))

        # deleting shouldn't be possible
        self.assertRaises(PublicUserException, tbs_controller.delete_service_level, service_level.id)

        # edit this year
        service_level = tbs_controller.save_service_level_state(service_level.id, {
            'tbs_id': tbs_license.id,
            'name': 'tbs fee 1',
            'level': 1,
            'contract_annex_text': 'contract text',
            'administration_fees': [{
                'start_year': jan1_2014,
                'fee': Decimal('12.12345'),
            }],
            'contribution_fees': [{
                'start_date': feb1,
                'fee': Decimal('11.12345'),
            }],
        })

        self.assertEqual(service_level.get_current_administration_fee(), Decimal('12.12345'))

        # move to 2015
        set_date(today() + datetime.timedelta(days=365))

        # edit 2014, should throw exception
        self.assertRaises(PvcInvalidFeeException, tbs_controller.save_service_level_state, service_level.id, {
            'tbs_id': tbs_license.id,
            'name': 'tbs fee 1',
            'level': 1,
            'contract_annex_text': 'contract text',
            'administration_fees': [{
                'start_year': jan1_2014,
                'fee': Decimal('14.12345'),
            }],
            'contribution_fees': [{
                'start_date': feb1,
                'fee': Decimal('11.12345'),
            }],
        })

    def test_update_service_level_contribution_fee(self):
        set_date(datetime.date(2013, 1, 1))

        tbs_controller = PvcTBSController(account_id=self.operator_account.id)
        tbs_license = tbs_controller.save_tbs(self._get_tbs_license_state())

        lcc = LicenseConfigController(self.sa_session, scope=tbs_license)
        lcc.set('pvc.tbs.declaration.period', PvcTbsLicense.DECLARATION_PERIOD_MONTHLY)

        state = {'tbs_id': tbs_license.id, 'name': '', 'contract_annex_text': '',
            'administration_fees': [],
            'contribution_fees': [{ 'start_date': datetime.date(2013, 1, 1), 'fee': Decimal('1'), }],
        }

        service_level = tbs_controller.save_service_level_state(None, state)

        self.assertRaises(PublicUserException, tbs_controller.update_service_level_contribution_fee, service_level.id, datetime.date(2013, 3, 1), Decimal('2'))

        set_date(today() + datetime.timedelta(days=365))

        state['contribution_fees'].append({'start_date': datetime.date(2014, 2, 1), 'fee': Decimal('1'), })

        service_level = tbs_controller.save_service_level_state(service_level.id, state)

        set_date(datetime.date(2014, 2, 1))

        self.assertRaises(PublicUserException, tbs_controller.update_service_level_contribution_fee, service_level.id, datetime.date(2013, 1, 1), Decimal('2'))

        self.assertEqual([], tbs_controller.update_service_level_contribution_fee(service_level.id, datetime.date(2014, 2, 1), Decimal('2')))

    def test_update_service_level_administration_fee(self):
        set_date(datetime.date(2013, 1, 1))

        tbs_controller = PvcTBSController(account_id=self.operator_account.id)
        tbs_license = tbs_controller.save_tbs(self._get_tbs_license_state())

        lcc = LicenseConfigController(self.sa_session, scope=tbs_license)
        lcc.set('pvc.tbs.declaration.period', PvcTbsLicense.DECLARATION_PERIOD_MONTHLY)

        state = {'tbs_id': tbs_license.id, 'name': '', 'contract_annex_text': '',
            'administration_fees': [{ 'start_year': 2013, 'fee': Decimal('1'), }],
            'contribution_fees': [],
        }

        service_level = tbs_controller.save_service_level_state(None, state)

        self.assertRaises(PublicUserException, tbs_controller.update_service_level_administration_fee, service_level.id, 2014, Decimal('2'))

        set_date(today() + datetime.timedelta(days=365))

        state['administration_fees'].append({'start_year': 2014, 'fee': Decimal('1.1'), })

        service_level = tbs_controller.save_service_level_state(service_level.id, state)

        set_date(datetime.date(2014, 2, 1))

        self.assertRaises(PublicUserException, tbs_controller.update_service_level_administration_fee, service_level.id, 2013, Decimal('2'))

        self.assertTrue(tbs_controller.update_service_level_administration_fee(service_level.id, 2014, Decimal('2')))

    def test_save_service_level_contribution_fee_date(self):

        # set fixed date
        set_date(datetime.date(2013, 1, 1))

        tbs_controller = PvcTBSController(account_id=self.operator_account.id)
        tbs_license = tbs_controller.save_tbs(self._get_tbs_license_state())

        lcc = LicenseConfigController(self.sa_session, scope=tbs_license)
        lcc.set('pvc.tbs.declaration.period', PvcTbsLicense.DECLARATION_PERIOD_MONTHLY)

        jan1 = datetime.date(2013, 1, 1)
        feb1 = datetime.date(2013, 2, 1)
        feb2 = datetime.date(2013, 2, 2)

        # create a service level first time
        service_level = tbs_controller.save_service_level_state(None, {'tbs_id': tbs_license.id, 'name': '', 'contract_annex_text': '',
            'administration_fees': [],
            'contribution_fees': [{ 'start_date': jan1, 'fee': Decimal('1'), }],
        })

        # update fee without changes should go ok
        tbs_controller.save_service_level_state(service_level.id, {'tbs_id': tbs_license.id, 'name': '', 'contract_annex_text': '',
            'administration_fees': [],
            'contribution_fees': [{'start_date': jan1, 'fee': Decimal('1'), }],
        })

        # try to change fee to past should fail
        self.assertRaises(PvcInvalidFeeException, tbs_controller.save_service_level_state, service_level.id, {'tbs_id': tbs_license.id, 'name': '', 'contract_annex_text': '',
            'administration_fees': [],
            'contribution_fees': [ {'start_date': jan1 + relativedelta(days=-1), 'fee': Decimal('1'), }, ],
        })

        # try to change fee to a day later should fail
        self.assertRaises(PvcInvalidFeeException, tbs_controller.save_service_level_state, service_level.id, {'tbs_id': tbs_license.id, 'name': '', 'contract_annex_text': '',
            'administration_fees': [],
            'contribution_fees': [ {'start_date': jan1 + relativedelta(days=1), 'fee': Decimal('1'), }, ],
        })

        # update fee with a new scheduled fee for next month should go ok
        tbs_controller.save_service_level_state(service_level.id, {'tbs_id': tbs_license.id, 'name': '', 'contract_annex_text': '',
            'administration_fees': [],
            'contribution_fees': [
                {'start_date': jan1, 'fee': Decimal('1'), },
                {'start_date': feb1, 'fee': Decimal('1'), }
            ],
        })

        # schedule a new fee to the second of next month should fail
        self.assertRaises(PvcInvalidFeeException, tbs_controller.save_service_level_state, service_level.id, {'tbs_id': tbs_license.id, 'name': '', 'contract_annex_text': '',
            'administration_fees': [],
            'contribution_fees': [
                {'start_date': jan1, 'fee': Decimal('1'), },
                {'start_date': feb2 + relativedelta(days=1), 'fee': Decimal('1'), },
            ],
        })

        # change tbs declaration date to quarter
        lcc = LicenseConfigController(self.sa_session, scope=tbs_license)
        lcc.set('pvc.tbs.declaration.period', PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY)

        # scheduling for march 1 should fail
        self.assertRaises(PvcInvalidFeeException, tbs_controller.save_service_level_state, service_level.id, {'tbs_id': tbs_license.id, 'name': '', 'contract_annex_text': '',
            'administration_fees': [],
            'contribution_fees': [
                {'start_date': jan1, 'fee': Decimal('1'), },
                {'start_date': datetime.date(2013, 3, 1), 'fee': Decimal('1'), },
            ],
        })

        # scheduling for 1 april is ok
        tbs_controller.save_service_level_state(service_level.id, {'tbs_id': tbs_license.id, 'name': '', 'contract_annex_text': '',
            'administration_fees': [],
            'contribution_fees': [
                {'start_date': jan1, 'fee': Decimal('1'), },
                {'start_date': datetime.date(2013, 4, 1), 'fee': Decimal('1'), }
            ],
        })

        # change time to december
        set_date(datetime.date(2013, 12, 1))

        # scheduling for 1 jan 2014 is ok
        tbs_controller.save_service_level_state(service_level.id, {'tbs_id': tbs_license.id, 'name': '', 'contract_annex_text': '',
            'administration_fees': [],
            'contribution_fees': [
                {'start_date': jan1, 'fee': Decimal('1'), },
                {'start_date': datetime.date(2013, 4, 1), 'fee': Decimal('1'), },
                {'start_date': datetime.date(2014, 1, 1), 'fee': Decimal('1'), }
            ],
        })

        # change time jan 2014
        set_date(datetime.date(2014, 1, 1))

        # change tbs declaration date to yearly
        lcc = LicenseConfigController(self.sa_session, scope=tbs_license)
        lcc.set('pvc.tbs.declaration.period', PvcTbsLicense.DECLARATION_PERIOD_YEARLY)

        # scheduling for april 1 should fail
        self.assertRaises(PvcInvalidFeeException, tbs_controller.save_service_level_state, service_level.id, {'tbs_id': tbs_license.id, 'name': '', 'contract_annex_text': '',
            'administration_fees': [],
            'contribution_fees': [
                {'start_date': jan1, 'fee': Decimal('1'), },
                {'start_date': datetime.date(2013, 4, 1), 'fee': Decimal('1'), },
                {'start_date': datetime.date(2014, 1, 1), 'fee': Decimal('1'), },
                {'start_date': datetime.date(2014, 4, 1), 'fee': Decimal('1'), },
            ],
        })

        # but for jan 2015 is ok
        tbs_controller.save_service_level_state(service_level.id, {'tbs_id': tbs_license.id, 'name': '', 'contract_annex_text': '',
            'administration_fees': [],
            'contribution_fees': [
                {'start_date': jan1, 'fee': Decimal('1'), },
                {'start_date': datetime.date(2013, 4, 1), 'fee': Decimal('1'), },
                {'start_date': datetime.date(2014, 1, 1), 'fee': Decimal('1'), },
                {'start_date': datetime.date(2015, 1, 1), 'fee': Decimal('1'), },
            ],
        })

        # remove future fee should be ok
        tbs_controller.save_service_level_state(service_level.id, {'tbs_id': tbs_license.id, 'name': '', 'contract_annex_text': '',
            'administration_fees': [],
            'contribution_fees': [
                {'start_date': jan1, 'fee': Decimal('1'), },
                {'start_date': datetime.date(2013, 4, 1), 'fee': Decimal('1'), },
                {'start_date': datetime.date(2014, 1, 1), 'fee': Decimal('1'), },
            ],
        })

        # cannot remove the fee from today
        self.assertRaises(PvcInvalidFeeException, tbs_controller.save_service_level_state, service_level.id, {'tbs_id': tbs_license.id, 'name': '', 'contract_annex_text': '',
            'administration_fees': [],
            'contribution_fees': [
                {'start_date': jan1, 'fee': Decimal('1'), },
                {'start_date': datetime.date(2013, 4, 1), 'fee': Decimal('1'), },
            ],
        })



    def test_tbs_approve_member(self):

        operator_account = self._create_operator_account()
        tbs_account = self._create_tbs_account(operator_account)
        service_level = self._create_tbs_service_level(tbs_account, year=today().year)

        member_account = self._create_member_account()
        tbs_member = self._member_apply_at_tbs(member_account, tbs_account, service_level)

        # now, as the tbs: approve the member
        mc = PvcMemberController(self.sa_session, account_id = tbs_account.id)
        tbs_member_state = mc.load_tbs_member(tbs_member.id)

        # approve the tbs
        mc.save_tbs_member(tbs_member.id, {
            'status': PvcTbsMember.STATUS_APPROVED,
            'is_full_member': False,
            'is_representative': False,
            'is_representative': False
        })

        # assert new status
        self.assertEquals(tbs_member.status, PvcTbsMember.STATUS_APPROVED)

    def test_tbs_override_member_declaration_period(self):
        operator_account = self._create_operator_account()
        tbs_account = self._create_tbs_account(operator_account)
        service_level = self._create_tbs_service_level(tbs_account, year=today().year)

        tbs = tbs_account.default_group.license

        member_account = self._create_member_account()
        tbs_member = self._member_apply_at_tbs(member_account, tbs_account, service_level)

        self.assertEqual(tbs_member.declaration_period, tbs.get_config('pvc.tbs.declaration.period'))

        # now, as the tbs: approve the member
        mc = PvcMemberController(self.sa_session, account_id = tbs_account.id)
        tbs_member_state = mc.load_tbs_member(tbs_member.id)

        # approve the tbs
        mc.save_tbs_member(tbs_member.id, {
            'status': PvcTbsMember.STATUS_APPROVED,
            'is_full_member': False,
            'is_representative': False,
            'is_representative': False,
            'declaration_periodicity': PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY
        })

        # assert new status
        self.assertEquals(tbs_member.status, PvcTbsMember.STATUS_APPROVED)
        # No declarations yet, means the periodicity can be changed on the fly
        self.assertEqual(tbs_member.declaration_periodicity, PvcTbsLicense.DECLARATION_PERIOD_QUARTERLY)

    def test_tbs_approve_member_commissioner_only(self):

        operator_account = self._create_operator_account()
        tbs_account = self._create_tbs_account(operator_account)

        member_account = self._create_member_account()
        tbs_member = self._member_apply_at_tbs(member_account, tbs_account, service_level=None, upload_contract=False, is_producer=False, is_commissioner=True)

        # now, as the tbs: approve the member
        mc = PvcMemberController(self.sa_session, account_id = tbs_account.id)
        tbs_member_state = mc.load_tbs_member(tbs_member.id)

        # approve the tbs
        mc.save_tbs_member(tbs_member.id, {
            'status': PvcTbsMember.STATUS_APPROVED,
            'is_full_member': False,
            'is_representative': False,
        })

        # assert new status
        self.assertEquals(tbs_member.status, PvcTbsMember.STATUS_APPROVED)

    def test_tbs_reject_member(self):

        tbs_account = self._create_tbs_account()
        member_account = self._create_member_account()
        service_level = self._create_tbs_service_level(tbs_account)

        tbs_member = self._member_apply_at_tbs(member_account, tbs_account, service_level)

        # as the tbs, reject the member
        mc = PvcMemberController(self.sa_session, account_id = tbs_account.id)
        mc.save_tbs_member(tbs_member.id, {
            'status': PvcTbsMember.STATUS_REJECTED,
            'is_full_member': False,
            'reject_reason': 'reject reason',
            'is_representative': False
        })

        # assert new status
        self.assertEquals(tbs_member.status, PvcTbsMember.STATUS_REJECTED)

    def test_tbs_reject_member_then_back_to_pending(self):
        # This reproduces ticket #365 and validates the fix.

        tbs_account = self._create_tbs_account()
        member_account = self._create_member_account()
        service_level = self._create_tbs_service_level(tbs_account)

        tbs_member = self._member_apply_at_tbs(member_account, tbs_account, service_level)

        # as the tbs, reject the member
        mc = PvcMemberController(self.sa_session, account_id = tbs_account.id)
        mc.save_tbs_member(tbs_member.id, {
            'status': PvcTbsMember.STATUS_REJECTED,
            'is_full_member': False,
            'reject_reason': 'reject reason',
            'is_representative': False
        })

        # assert new status
        self.assertEquals(tbs_member.status, PvcTbsMember.STATUS_REJECTED)

        self.assertRaises(PublicUserException, mc.save_tbs_member, tbs_member.id, {
            'status': PvcTbsMember.STATUS_APPROVED,
            'is_full_member': False,
            'is_representative': False
        })

        mc.save_tbs_member(tbs_member.id, {
            'status': PvcTbsMember.STATUS_PENDING,
            'is_full_member': False,
            'is_representative': False
        })

        # assert new status
        self.assertEquals(tbs_member.status, PvcTbsMember.STATUS_PENDING)

    def test_membership_fee(self):
        set_date(datetime.date(2013, 1, 1))

        tbs_account = self._create_tbs_account()

        tbs = tbs_account.default_group.license

        c = PvcTBSController(self.sa_session, account_id=self.operator_account.id)
        self.assertRaises(PublicUserException, c.save_membership_state, None, {'tbs_id': tbs.id, 'name': 'test'})

        lcc = LicenseConfigController(self.sa_session, scope=tbs)
        lcc.set('pvc.tbs.use.membership.instead.of.service.level', True)

        m = c.save_membership_state(None, {'tbs_id': tbs.id, 'name': 'test',
            'membership_fees': [{
                'fee': Decimal('10.12345'),
                'start_year': 2013,
            }],
        })
        self.sa_session.expire(m)

        self.assertEqual(len(m.membership_fees), 1)
        self.assertEqual(m.membership_fees[0].start_year, 2013)
        self.assertEqual(m.membership_fees[0].fee, Decimal('10.12345'))

        m = c.save_membership_state(None, {'tbs_id': tbs.id, 'name': 'test2',
            'membership_fees': [{
                'fee': Decimal('10.12345'),
                'start_year': 2014,
            }, {
                'fee': Decimal('12.12345'),
                'start_year': 2017,
            }],
        })
        self.sa_session.expire(m)

        self.assertEqual(len(m.membership_fees), 2)
        self.assertEqual(m.membership_fees[0].start_year, 2014)
        self.assertEqual(m.membership_fees[0].fee, Decimal('10.12345'))
        self.assertEqual(m.membership_fees[1].start_year, 2017)
        self.assertEqual(m.membership_fees[1].fee, Decimal('12.12345'))

        # now we try to update a membership fee and remove the other
        m = c.save_membership_state(m.id, {'tbs_id': tbs.id, 'name': 'test2',
            'membership_fees': [{
                'fee': Decimal('11.12345'),
                'start_year': 2014,
            }],
        })
        self.sa_session.expire(m)
        self.assertEqual(len(m.membership_fees), 1)
        self.assertEqual(m.membership_fees[0].start_year, 2014)
        self.assertEqual(m.membership_fees[0].fee, Decimal('11.12345'))

    def test_membership_fee_failures(self):
        tbs_account = self._create_tbs_account()
        tbs = tbs_account.default_group.license

        lcc = LicenseConfigController(self.sa_session, scope=tbs)
        lcc.set('pvc.tbs.use.membership.instead.of.service.level', True)

        c = PvcTBSController(self.sa_session, account_id=self.operator_account.id)

        # verify that the years have to be at least 3 years apart.
        self.assertRaises(PublicUserException, c.save_membership_state, None, {'tbs_id': tbs.id, 'name': 'test',
            'membership_fees': [{
                'fee': Decimal('10.12345'),
                'start_year': 2013,
            },{
                'fee': Decimal('10.12345'),
                'start_year': 2014,
            }],
        })

        m = c.save_membership_state(None, {'tbs_id': tbs.id, 'name': 'test',
            'membership_fees': [{
                'fee': Decimal('10.12345'),
                'start_year': 2013,
            }],
        })

        # We cannot change a fee in the past for an existing membership
        self.assertRaises(PvcInvalidFeeException, c.save_membership_state, m.id, {'tbs_id': tbs.id, 'name': 'test_past',
            'membership_fees': [{
                'fee': Decimal('10.12345'),
                'start_year': 2011,
            }],
        })

    def test_tbs_switch_account(self):
        tbs_account = self._create_tbs_account()
        member_account = self._create_member_account()
        service_level = self._create_tbs_service_level(tbs_account)

        tm = self._member_apply_at_tbs(member_account, tbs_account, service_level)

        # second member
        other_member = self._create_member_account()

        ac = AccountController(self.sa_session, account_id=tbs_account.id)
        account = ac.login_as(None, tbs_account.name, member_account.name)
        self.assertEqual(account, member_account)

        self.assertFalse(ac.login_as(None, tbs_account.name, other_member.name))

        # suspend the member
        tm.status = PvcTbsMember.STATUS_SUSPENDED
        self.sa_session.add(tm)

        self.assertFalse(ac.login_as(None, tbs_account.name, member_account.name))



